/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <iostream>
#include "my_getopt.h"

#include <math.h>
#include <mapidefs.h>
#include <mapispi.h>
#include <mapix.h>
#include <mapiutil.h>

#include "IECServiceAdmin.h"
#include "IECUnknown.h"

#include "ECTags.h"
#include "ECGuid.h"
#include "ECABEntryID.h"
#include "CommonUtil.h"
#include "stringutil.h"
#include "ecversion.h"
#include "mapiext.h"
#include "Util.h"

#include <edkmdb.h>
#include <edkguid.h>
#include <libintl.h>

using namespace std;

enum modes {
	MODE_INVALID = 0, MODE_LIST_USERS, MODE_CREATE_PUBLIC,
	MODE_CREATE_USER, MODE_CREATE_STORE, MODE_HOOK_STORE,
	MODE_DELETE_STORE, MODE_REMOVE_STORE, MODE_UPDATE_USER, MODE_DELETE_USER,
	MODE_CREATE_GROUP, MODE_DELETE_GROUP,
	MODE_LIST_GROUP, MODE_ADDUSER_GROUP, MODE_DELETEUSER_GROUP,
	MODE_CREATE_COMPANY, MODE_UPDATE_COMPANY, MODE_DELETE_COMPANY, MODE_LIST_COMPANY,
	MODE_ADD_VIEW, MODE_DEL_VIEW, MODE_LIST_VIEW,
	MODE_ADD_ADMIN, MODE_DEL_ADMIN, MODE_LIST_ADMIN,
	MODE_ADD_USERQUOTA_RECIPIENT, MODE_DEL_USERQUOTA_RECIPIENT, MODE_LIST_USERQUOTA_RECIPIENT,
	MODE_ADD_COMPANYQUOTA_RECIPIENT, MODE_DEL_COMPANYQUOTA_RECIPIENT, MODE_LIST_COMPANYQUOTA_RECIPIENT,
	MODE_SYNC_USERS, MODE_DETAILS, MODE_LIST_SENDAS, MODE_HELP,
	MODE_SYSTEM_ADMIN, MODE_PURGE_SOFTDELETE, MODE_LIST_ORPHANS
};

enum {
	OPT_CREATE_STORE = 129,  // high to avoid clashes with modes
	OPT_DELETE_STORE,
	OPT_HOOK_STORE,
	OPT_REMOVE_STORE,
	OPT_HELP,
	OPT_HOST,
	OPT_SYNC_USERS,
	OPT_DETAILS,
	OPT_DETAILS_TYPE,
	OPT_USER_QUOTA_HARD,
	OPT_USER_QUOTA_SOFT,
	OPT_USER_QUOTA_WARN,
	OPT_USER_QUOTA_OVERRIDE,
	OPT_USER_DEFAULT_QUOTA_HARD,
	OPT_USER_DEFAULT_QUOTA_SOFT,
	OPT_USER_DEFAULT_QUOTA_WARN,
	OPT_USER_DEFAULT_QUOTA_OVERRIDE,
	OPT_LANG,
	OPT_MR_ACCEPT,
	OPT_MR_DECLINE_CONFLICT,
	OPT_MR_DECLINE_RECURRING,
	OPT_ADD_SENDAS,
	OPT_DEL_SENDAS,
	OPT_LIST_SENDAS,
	OPT_CREATE_COMPANY,
	OPT_UPDATE_COMPANY,
	OPT_DELETE_COMPANY,
	OPT_LIST_COMPANY,
	OPT_ADD_VIEW,
	OPT_DEL_VIEW,
	OPT_LIST_VIEW,
	OPT_ADD_ADMIN,
	OPT_DEL_ADMIN,
	OPT_LIST_ADMIN,
	OPT_SYSTEM_ADMIN,
	OPT_ADD_UQUOTA_RECIPIENT,
	OPT_DEL_UQUOTA_RECIPIENT,
	OPT_LIST_UQUOTA_RECIPIENT,
	OPT_ADD_CQUOTA_RECIPIENT,
	OPT_DEL_CQUOTA_RECIPIENT,
	OPT_LIST_CQUOTA_RECIPIENT,
	OPT_PURGE_SOFTDELETE,
	OPT_LIST_ORPHANS,
    OPT_CONFIG
};

struct option long_options[] = {
		{ "create-store", 1, NULL, OPT_CREATE_STORE },
		{ "delete-store", 1, NULL, OPT_DELETE_STORE },
		{ "hook-store", 1, NULL, OPT_HOOK_STORE },
		{ "remove-store", 1, NULL, OPT_REMOVE_STORE },
		{ "list-orphans", 0, NULL, OPT_LIST_ORPHANS },
		{ "details", 1, NULL, OPT_DETAILS },
		{ "type", 1, NULL, OPT_DETAILS_TYPE },
		{ "help", 0, NULL, OPT_HELP },
		{ "host", 1, NULL, OPT_HOST },
		{ "sync", 0, NULL, OPT_SYNC_USERS },
		{ "qh", 1, NULL, OPT_USER_QUOTA_HARD },
		{ "qs", 1, NULL, OPT_USER_QUOTA_SOFT },
		{ "qw", 1, NULL, OPT_USER_QUOTA_WARN },
		{ "qo", 1, NULL, OPT_USER_QUOTA_OVERRIDE },
		{ "udqh", 1, NULL, OPT_USER_DEFAULT_QUOTA_HARD },
		{ "udqs", 1, NULL, OPT_USER_DEFAULT_QUOTA_SOFT },
		{ "udqw", 1, NULL, OPT_USER_DEFAULT_QUOTA_WARN },
		{ "udqo", 1, NULL, OPT_USER_DEFAULT_QUOTA_OVERRIDE },
		{ "lang", 1, NULL, OPT_LANG },
		{ "mr-accept", 1, NULL, OPT_MR_ACCEPT },
		{ "mr-decline-conflict", 1, NULL, OPT_MR_DECLINE_CONFLICT },
		{ "mr-decline-recurring", 1, NULL, OPT_MR_DECLINE_RECURRING },
		{ "add-sendas", 1, NULL, OPT_ADD_SENDAS },
		{ "del-sendas", 1, NULL, OPT_DEL_SENDAS },
		{ "list-sendas", 1, NULL, OPT_LIST_SENDAS },
		{ "create-company", 1, NULL, OPT_CREATE_COMPANY },
		{ "update-company", 1, NULL, OPT_UPDATE_COMPANY },
		{ "delete-company", 1, NULL, OPT_DELETE_COMPANY },
		{ "list-companies", 0, NULL, OPT_LIST_COMPANY },
		{ "add-to-viewlist", 1, NULL, OPT_ADD_VIEW },
		{ "del-from-viewlist", 1, NULL, OPT_DEL_VIEW },
		{ "list-view", 0, NULL, OPT_LIST_VIEW },
		{ "add-to-adminlist", 1, NULL, OPT_ADD_ADMIN },
		{ "del-from-adminlist", 1, NULL, OPT_DEL_ADMIN },
		{ "list-admin", 0, NULL, OPT_LIST_ADMIN },
		{ "set-system-admin", 1, NULL, OPT_SYSTEM_ADMIN },
		{ "add-userquota-recipient", 1, NULL, OPT_ADD_UQUOTA_RECIPIENT },
		{ "del-userquota-recipient", 1, NULL, OPT_DEL_UQUOTA_RECIPIENT },
		{ "list-userquota-recipients", 0, NULL, OPT_LIST_UQUOTA_RECIPIENT },
		{ "add-companyquota-recipient", 1, NULL, OPT_ADD_CQUOTA_RECIPIENT },
		{ "del-companyquota-recipient", 1, NULL, OPT_DEL_CQUOTA_RECIPIENT },
		{ "list-companyquota-recipients", 0, NULL, OPT_LIST_CQUOTA_RECIPIENT },
		{ "purge-softdelete", 1, NULL, OPT_PURGE_SOFTDELETE },
		{ "config", 1, NULL, OPT_CONFIG },
		{ NULL, 0, NULL, 0 }
};

#ifndef PR_FREEBUSY_ENTRYIDS
#define PR_FREEBUSY_ENTRYIDS PROP_TAG(PT_MV_BINARY, 0x36e4)
#endif

#ifndef PR_FREEBUSY_NUM_MONTHS
#define PR_FREEBUSY_NUM_MONTHS PROP_TAG(PT_LONG, 0x6841)
#endif

// also define in freebusytags.h
#define PR_PROCESS_MEETING_REQUESTS				PROP_TAG(PT_BOOLEAN, 0x686d)
#define PR_DECLINE_CONFLICTING_MEETING_REQUESTS	PROP_TAG(PT_BOOLEAN, 0x686f)
#define PR_DECLINE_RECURRING_MEETING_REQUESTS	PROP_TAG(PT_BOOLEAN, 0x686e)

void print_help(char *name) {
	cout << "Usage:" << endl;
	cout << name << " [action] [options]" << endl << endl;
	cout << "Actions: [-s] | [[-c|-u|-d|-b|-B|--details] username] | [[-g|-G] groupname] | [-l|-L]" << endl;
	cout << "\t" << " -s" << "\t\t\t" << "create public store." << endl;
	cout << "\t" << " --sync" << "\t\t\t" << "synchronize users and groups with external source." << endl;
	cout << "\t" << " --purge-softdelete N" << "\t\t\t" << "Purge items in marked as softdeleted that are older than N days." << endl;
	cout << "\t" << " -l" << "\t\t\t" << "list users. Use -I to list users of a specific company, if applicable." << endl;
	cout << "\t" << " -L" << "\t\t\t" << "list groups. Use -I to list groups of a specific company, if applicable." << endl;
	cout << "\t" << " --list-sendas name" << "\t" << "List all users allowed to send-as 'name'." << endl;
	cout << "\t" << " --list-companies" << "\t" << "List all companies" << endl;
	cout << "\t" << " --list-view" << "\t\t" << "List all companies in the remote-view list." << endl;
	cout << "\t" << " --list-admin" << "\t\t" << "List all users in the remote-admin list." << endl;
	cout << "\t" << " --list-userquota-recipients" << "\t" << "List all additional recipients for a userquota warning email." << endl;
	cout << "\t" << " --list-companyquota-recipients" << "\t" << "List all additional recipients for a companyquota warning email." << endl;
	cout << "\t" << " --details name" << "\t\t" << "show object details, use --type to indicate the object type" << endl;
	cout << "\t" << " --type type" << "\t\t" << "Set object type for --details. Values can be \"user\", \"group\" or \"company\"." << endl;
	cout << endl;
	cout << "Additional Actions when using the DB user plugin:" << endl;
	cout << "\t" << " -c user" << "\t" << "create user, -p, -f, -e options required, -a and -n are optional. Quota options are optional." << endl;
	cout << "\t" << " -u user" << "\t" << "update user, -U, -p, -f, -e and -a optional. Quota options are optional." << endl;
	cout << "\t" << " -d user" << "\t" << "delete user." << endl;
	cout << "\t" << " -g group" << "\t" << "create group." << endl;
	cout << "\t" << " -G group" << "\t" << "delete group." << endl;
	cout << "\t" << " -b user" << "\t" << "add user to a group, -i required." << endl;
	cout << "\t" << " -B user" << "\t" << "delete user from a group, -i required." << endl;
	cout << endl;
	cout << "Additional Actions when using the Unix user plugin:" << endl;
	cout << "\t" << " -u user" << "\t" << "update user, -p, -e and -a optional. Quota options are optional." << endl;
	cout << endl;
	cout << "Additional Actions when using the DB user plugin in hosted mode:" << endl;
	cout << "\t" << " --create-company name" << "\t" << "Create company space" << endl;
	cout << "\t" << " --update-company name" << "\t" << "Update company space" << endl;
	cout << "\t" << " --delete-company name" << "\t" << "Delete company space" << endl;
	cout << "\t" << " --set-system-admin name" << "\t" << "Set system administrator for the company specified by -I (does not grant Admin privileges)" << endl;
	cout << "\t" << " --add-to-viewlist name" << "\t\t" << "Add company 'name' to remote-view list of company specified by -I" << endl;
	cout << "\t" << " --del-from-viewlist name" << "\t" << "Delete company 'name' from remote-view list of company specified by -I" << endl;
	cout << "\t" << " --add-to-adminlist name" << "\t" << "Add user 'name' to remote-admin list of company specified by -I" << endl;
	cout << "\t" << " --del-from-adminlist name" << "\t" << "Delete user 'name' from remote-admin list of company specified by -I" << endl;
	cout << "\t" << " --add-userquota-recipient user" << "\t\t" << "Add 'user' as recipient to userquota warning emails." << endl;
	cout << "\t" << " --del-userquota-recipient user" << "\t\t" << "Delete 'user' as recipient to userquota warning emails." << endl;
	cout << "\t" << " --add-companyquota-recipient user" << "\t" << "Add 'user' as recipient to companyquota warning emails." << endl;
	cout << "\t" << " --del-companyquota-recipient user" << "\t" << "Delete 'user' as recipient to companyquota warning emails." << endl;
	cout << endl;
	cout << "Options: [-U new username] [-P|-p password] [-f fullname] [-e emailaddress]" << endl;
	cout << "         [-a [y|n]] [-n [y|n]] [-h path] [-i group] [--qo [y|n]] [--qh value] [--qs value] [--qw value]" << endl;
	cout << "\t" << " -U new username" << "\t" << "rename username to new username" << endl;
	cout << "\t" << " -P" << "\t\t\t" << "prompt for password, can be substituted by '-p pass'" << endl;
	cout << "\t" << " -p pass" << "\t\t" << "set password to pass, can be substituted by '-P'" << endl;
	cout << "\t" << " -f full" << "\t\t" << "set fullname to full" << endl;
	cout << "\t" << " -e addr" << "\t\t" << "set email address to addr" << endl;
	cout << "\t" << " -a [y|n]" << "\t\t" << "set administrator level for user. yes/no, y/n or 2/1/0." << endl;
	cout << "\t" << " -n [y|n]" << "\t\t" << "create nonactive user. yes/no, y/n or 1/0. Can only be used with -c." << endl;
	cout << "\t" << " --qo [y|n]" << "\t\t" << "override server quota limits. yes/no, y/n or 1/0." << endl;
	cout << "\t" << " --qh hardquota" << "\t\t" << "set hardquota limit in Mb" << endl;
	cout << "\t" << " --qs softquota" << "\t\t" << "set softquota limit in Mb" << endl;
	cout << "\t" << " --qw warnquota" << "\t\t" << "set warnquota limit in Mb" << endl;
	cout << "\t" << " --udqo [y|n]" << "\t\t" << "override user default server quota limits for specific company. yes/no, y/n or 1/0." << endl;
	cout << "\t" << " --udqh hardquota" << "\t" << "set user default hardquota limit for specific company in Mb" << endl;
	cout << "\t" << " --udqs softquota" << "\t" << "set user default softquota limit for specific company in Mb" << endl;
	cout << "\t" << " --udqw warnquota" << "\t" << "set user default warnquota limit for specific company in Mb" << endl;
	cout << "\t" << " -i group" << "\t\t" << "name of the group" << endl;
	cout << "\t" << " -I company" << "\t\t" << "name of the company" << endl;
	cout << "\t" << " --mr-accept" << "\t\t" << "(resource) auto-accept meeting requests. yes/no" << endl;
	cout << "\t" << " --mr-decline-conflict" << "\t" << "(resource) decline meeting requests for conflicting times. yes/no" << endl;
	cout << "\t" << " --mr-decline-recurring" << "\t" << "(resource) decline meeting requests for all recurring items. yes/no" << endl;
	cout << "\t" << " --add-sendas name" << "\t" << "Add user 'name' to send-as list of user specified by -u" << endl;
	cout << "\t" << " --del-sendas name" << "\t" << "Delete user 'name' from send-as list of user specified by -u" << endl;
	cout << endl;
	cout << "The following functions are to control stores of users:" << endl;
	cout << "\t" << " --list-orphans" << "\t\t" << "list all users without stores and stores without users." << endl;
	cout << "\t" << " --remove-store storeguid" << "\t" << "delete orphaned store of user that is deleted from external source." << endl;
	cout << "\t" << " --hook-store storeguid" << "\t" << "hook orphaned store to a user." << endl;
	cout << "\t" << "   -u username" << "\t\t" << "update user to received orphaned store given in --hook-store." << endl;
	cout << endl;
	cout << "The following functions are for use from the create/delete user/group scripts:" << endl;
	cout << "\t" << " --create-store user" << "\t" << "create store for user that exists in external source." << endl;
	cout << "\t" << " --lang language" << "\t" << "create folders in a new store in this language (e.g. en_EN)." << endl;
	cout << endl;
	cout << "Global options: [-h|--host path]" << endl;
	cout << "\t" << " -h path" << "\t\t" << "connect through <path>, e.g. file:///var/run/socket" << endl;
	cout << "\t" << " -V Print version info." << endl;
	cout << "\t" << " --help" << "\t\t\t" << "show this help text." << endl;
	cout << endl;
}

char *get_password() {
	static char password[80] = {0};
	char *s = NULL;
	
	s = get_password("Type password:");
	if(s == NULL) {
		return NULL;
	}
		
	strncpy(password, s, sizeof(password)-1);
	
	s = get_password("Re-Type password:");
	if(s == NULL) {
		return NULL;
	}
	
	if(strcmp(password, s) != 0) {
		cerr << "Passwords don't match" << endl;
		return NULL;
	}
	
	return password;
}

int parse_yesno(char *opt) {
	if (opt[0] == 'y' || opt[0] == '1')
		return 1;
	return 0;
}

void print_quota(LPECQUOTA lpQuota, LPECQUOTASTATUS lpQuotaStatus)
{
	if (lpQuota) {
		// watch the not:
		cout << "Quota overrides:\t" << (!lpQuota->bUseDefaultQuota?"yes":"no") << endl;
		cout << "Warning level:\t\t" << str_storage(lpQuota->llWarnSize) << endl;
		cout << "Soft level:\t\t" << str_storage(lpQuota->llSoftSize) << endl;
		cout << "Hard level:\t\t" << str_storage(lpQuota->llHardSize) << endl;
	}

	if (lpQuotaStatus)
		cout << "Current store size:\t" << str_storage((unsigned long long)ceil((double)lpQuotaStatus->llStoreSize), false) << endl;
}

// Default freebusy publish months
#define ECFREEBUSY_DEFAULT_PUBLISH_MONTHS		6

/**
 * An enumeration for getting the localfreebusy from the calendar or from the free/busy data folder.
 *
 * @note it's also the array position of property PR_FREEBUSY_ENTRYIDS
 */
enum DGMessageType { 
	dgAssociated = 0,	/**< Localfreebusy message in default associated calendar folder */
	dgFreebusydata = 1	/**< Localfreebusy message in Free/busy data folder */
};

/**
 * Create a local free/busy message
 *
 * @param[in] lpFolder
 *				Destenation folder for creating the local free/busy message
 * @param[in] ulFlags
 *				MAPI_ASSOCIATED	for Localfreebusy message in default associated calendar folder
 * @param[out] lppMessage
 *				The localfreebusy message with the free/busy settings
 *
 * @todo move the code to a common file
 *
 */
HRESULT CreateLocalFreeBusyMessage(LPMAPIFOLDER lpFolder, ULONG ulFlags, LPMESSAGE *lppMessage)
{
	HRESULT hr = hrSuccess;
	LPMESSAGE lpMessage = NULL;
	SPropValue sPropValMessage[6];
	
	memset(sPropValMessage, 0, sizeof(SPropValue) * 6);

	if (lpFolder == NULL || lppMessage == NULL || (ulFlags&~MAPI_ASSOCIATED) != 0) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpFolder->CreateMessage(&IID_IMessage, (ulFlags&MAPI_ASSOCIATED), &lpMessage);
	if(hr != hrSuccess)
		goto exit;

	sPropValMessage[0].ulPropTag = PR_MESSAGE_CLASS;
	sPropValMessage[0].Value.lpszA = "IPM.Microsoft.ScheduleData.FreeBusy";

	sPropValMessage[1].ulPropTag = PR_SUBJECT;
	sPropValMessage[1].Value.lpszA = "LocalFreebusy";

	sPropValMessage[2].ulPropTag = PR_FREEBUSY_NUM_MONTHS;
	sPropValMessage[2].Value.ul = ECFREEBUSY_DEFAULT_PUBLISH_MONTHS;

	sPropValMessage[3].ulPropTag = PR_DECLINE_RECURRING_MEETING_REQUESTS;
	sPropValMessage[3].Value.b = false;

	sPropValMessage[4].ulPropTag = PR_DECLINE_CONFLICTING_MEETING_REQUESTS;
	sPropValMessage[4].Value.b = false;

	sPropValMessage[5].ulPropTag = PR_PROCESS_MEETING_REQUESTS;
	sPropValMessage[5].Value.b = false;

	hr = lpMessage->SetProps(6, sPropValMessage, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = lpMessage->SaveChanges(KEEP_OPEN_READWRITE);
	if(hr != hrSuccess)
		goto exit;

	hr = lpMessage->QueryInterface(IID_IMessage, (void**)lppMessage);

exit:

	if (lpMessage)
		lpMessage->Release();

	return hr;
}

/*
 * Get local free/busy message to get delegate information
 * 
 * @note There is a differents between Outlook before 2003 and from 2003. 
 *       The free/busy settings are written on another place.
 *       Outlook 2000 and 2002, a message in the calendar associated folder
 *       Outlook 2003 and higher, a message in the freebusydata folder
 *		 exchange uses always the same?
 *
 * @param[in]	ulSettingsLocation
 *					Type message to get information
 * @param[in]	lpMsgStore
 *					the delegate message store for getting delegate information.
 * @param[in]	bCreateIfMissing
 *					If not exist create the localfreebusy message
 * @param[out]	lppFBMessage
 *					The localfreebusy message with the free/busy settings
 *
 * @return MAPI_E_NOT_FOUND 
 *			Local free/busy message is not exist
 *
 * @todo move the code to a common file
*/

HRESULT OpenLocalFBMessage(DGMessageType eDGMsgType, IMsgStore *lpMsgStore, bool bCreateIfMissing, IMessage **lppFBMessage)
{
    HRESULT hr = hrSuccess;
    IMAPIFolder *lpRoot = NULL;
    IMAPIFolder *lpFolder = NULL;
    IMessage *lpMessage = NULL;
    ULONG ulType = 0;
    LPSPropValue lpPropFB = NULL;
    LPSPropValue lpPropFBNew = NULL;
	LPSPropValue lpPVFBFolder = NULL;
    LPSPropValue lpEntryID = NULL;
    LPSPropValue lpPropFBRef = NULL; // Non-free
	LPSPropValue lpAppEntryID = NULL;
    ULONG cbEntryIDInbox = 0;
    LPENTRYID lpEntryIDInbox = NULL;
    IMAPIFolder *lpInbox = NULL;
    char *lpszExplicitClass = NULL;
    
    hr = lpMsgStore->OpenEntry(0, NULL, &IID_IMAPIFolder, MAPI_MODIFY, &ulType, (IUnknown **) &lpRoot);
    if(hr != hrSuccess)
        goto exit;

	// Check if the freebusydata folder and LocalFreeBusy is exist. Create the folder and message if it is request.
    if((HrGetOneProp(lpRoot, PR_FREEBUSY_ENTRYIDS, &lpPropFB) != hrSuccess ||
        lpPropFB->Value.MVbin.cValues < 2 ||
		lpPropFB->Value.MVbin.lpbin[eDGMsgType].lpb == NULL ||
        lpMsgStore->OpenEntry(lpPropFB->Value.MVbin.lpbin[eDGMsgType].cb, (LPENTRYID)lpPropFB->Value.MVbin.lpbin[eDGMsgType].lpb, &IID_IMessage, MAPI_MODIFY, &ulType, (IUnknown **) &lpMessage) != hrSuccess)
        && bCreateIfMissing) {
        
		// Open the inbox
		hr = lpMsgStore->GetReceiveFolder("", 0, &cbEntryIDInbox, &lpEntryIDInbox, &lpszExplicitClass);
        if(hr != hrSuccess)
            goto exit;
            
        hr = lpMsgStore->OpenEntry(cbEntryIDInbox, lpEntryIDInbox, &IID_IMAPIFolder, MAPI_MODIFY, &ulType, (IUnknown **) &lpInbox);
        if(hr != hrSuccess)
            goto exit;

		if (eDGMsgType == dgFreebusydata) {
			// Create freebusydata Folder
			hr = lpRoot->CreateFolder(FOLDER_GENERIC, "Freebusy Data", "", &IID_IMAPIFolder, OPEN_IF_EXISTS, &lpFolder);
			if(hr != hrSuccess)
				goto exit;

			// Get entryid of freebusydata
			hr = HrGetOneProp(lpFolder, PR_ENTRYID, &lpPVFBFolder);
			if(hr != hrSuccess)
				goto exit;

		} else if (eDGMsgType == dgAssociated) {
			//Open default calendar
			hr = HrGetOneProp(lpInbox, PR_IPM_APPOINTMENT_ENTRYID, &lpAppEntryID);
			if(hr != hrSuccess)
				goto exit;
			   
			hr = lpMsgStore->OpenEntry(lpAppEntryID->Value.bin.cb, (LPENTRYID)lpAppEntryID->Value.bin.lpb, &IID_IMAPIFolder, MAPI_MODIFY, &ulType, (IUnknown **) &lpFolder);
			if(hr != hrSuccess)
				goto exit;

		}
        
		hr = CreateLocalFreeBusyMessage(lpFolder, (eDGMsgType == dgAssociated)?MAPI_ASSOCIATED : 0, &lpMessage);
		if(hr != hrSuccess)
			goto exit;
            
        hr = HrGetOneProp(lpMessage, PR_ENTRYID, &lpEntryID);
        if(hr != hrSuccess)
            goto exit;
        
		// Update Free/Busy entryid
        if(lpPropFB == NULL || lpPropFB->Value.MVbin.cValues < 2) {
            hr = MAPIAllocateBuffer(sizeof(SPropValue), (void **)&lpPropFBNew);
            if(hr != hrSuccess)
                goto exit;
                
            lpPropFBNew->ulPropTag = PR_FREEBUSY_ENTRYIDS;
            hr = MAPIAllocateMore(sizeof(SBinary) * 4, lpPropFB, (void **)&lpPropFBNew->Value.MVbin.lpbin);
            if(hr != hrSuccess)
                goto exit;
            
			memset(lpPropFBNew->Value.MVbin.lpbin, 0, sizeof(SBinary) * 4);
			

			if (eDGMsgType == dgFreebusydata) {

				if(lpPropFB && lpPropFB->Value.MVbin.cValues > 0) {
					lpPropFBNew->Value.MVbin.lpbin[0] = lpPropFB->Value.MVbin.lpbin[0];

					if (lpPropFB->Value.MVbin.cValues > 2)
						lpPropFBNew->Value.MVbin.lpbin[2] = lpPropFB->Value.MVbin.lpbin[2];
				}
	            
				lpPropFBNew->Value.MVbin.lpbin[1].cb = lpEntryID->Value.bin.cb;
				lpPropFBNew->Value.MVbin.lpbin[1].lpb = lpEntryID->Value.bin.lpb;

				lpPropFBNew->Value.MVbin.lpbin[3].cb = lpPVFBFolder->Value.bin.cb;
				lpPropFBNew->Value.MVbin.lpbin[3].lpb = lpPVFBFolder->Value.bin.lpb;


			} else if(eDGMsgType == dgAssociated) {
				lpPropFBNew->Value.MVbin.lpbin[0].cb = lpEntryID->Value.bin.cb;
				lpPropFBNew->Value.MVbin.lpbin[0].lpb = lpEntryID->Value.bin.lpb;
			}

			lpPropFBNew->Value.MVbin.cValues = 4; // no problem if the data is NULL

            lpPropFBRef = lpPropFBNew; // use this one later on
                    
        } else {

            lpPropFB->Value.MVbin.lpbin[eDGMsgType].cb = lpEntryID->Value.bin.cb;
            lpPropFB->Value.MVbin.lpbin[eDGMsgType].lpb = lpEntryID->Value.bin.lpb;
            
            lpPropFBRef = lpPropFB; // use this one later on
        }
        
        // Put the MV property in the root folder
        hr = lpRoot->SetProps(1, lpPropFBRef, NULL);
        if(hr != hrSuccess)
            goto exit;
            
		// Put the MV property in the inbox folder
        hr = lpInbox->SetProps(1, lpPropFBRef, NULL);
        if(hr != hrSuccess)
            goto exit;
    }
    
    if(lpMessage == NULL) {
        hr = MAPI_E_NOT_FOUND;
        goto exit;
    }
    
    // We now have a message lpMessage which is the LocalFreeBusy message.
    *lppFBMessage = lpMessage;
        
exit:
    if(lpszExplicitClass)
        MAPIFreeBuffer(lpszExplicitClass);
    
	if(lpAppEntryID)
		MAPIFreeBuffer(lpAppEntryID);
    if(lpRoot)
        lpRoot->Release();
    if(lpPropFB)
        MAPIFreeBuffer(lpPropFB);
    if(lpPropFBNew)
        MAPIFreeBuffer(lpPropFBNew);
    if(lpEntryID)
        MAPIFreeBuffer(lpEntryID);
    if(lpEntryIDInbox)
        MAPIFreeBuffer(lpEntryIDInbox);
    if(lpInbox)
        lpInbox->Release();

    return hr;
}

HRESULT setQuota(IECServiceAdmin *lpServiceAdmin, ULONG cbEid, LPENTRYID lpEid, int quota, bool udefault,
				 long long warn, long long soft, long long hard, bool print = false)
{
	HRESULT hr = hrSuccess;
	LPECQUOTASTATUS lpsQuotaStatus = NULL;
	LPECQUOTA lpsQuota = NULL;
	ECQUOTA sQuota;

	if (lpEid == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpServiceAdmin->GetQuota(cbEid, lpEid, &lpsQuota);
	if (hr != hrSuccess) {
		cerr << "Unable to update quota, probably not found." << endl;
		goto exit;
	}

	if (print) {
		cout << "Old quota settings:" << endl;
		print_quota(lpsQuota, NULL);
		cout << endl;
	}

	if (quota == -1)
		sQuota.bUseDefaultQuota = lpsQuota->bUseDefaultQuota;
	else
		sQuota.bUseDefaultQuota = !quota;
	sQuota.bIsUserDefaultQuota = udefault;
	sQuota.llHardSize = (hard >= 0) ? hard : lpsQuota->llHardSize;
	sQuota.llSoftSize = (soft >= 0) ? soft : lpsQuota->llSoftSize;
	sQuota.llWarnSize = (warn >= 0) ? warn : lpsQuota->llWarnSize;

	hr = lpServiceAdmin->SetQuota(cbEid, lpEid, &sQuota);
	if(hr != hrSuccess) {
		cerr << "Unable to update quota information." << endl;
		goto exit;
	}

	if (print) {
		hr = lpServiceAdmin->GetQuotaStatus(cbEid, lpEid, &lpsQuotaStatus);
		if(hr != hrSuccess) {
			cerr << "Unable to request updated quota information." << endl;
			goto exit;
		}

		cout << "New quota settings:" << endl;
		print_quota(&sQuota, lpsQuotaStatus);
		cout << endl;
	}

exit:
	if (lpsQuota)
		MAPIFreeBuffer(lpsQuota);

	if (lpsQuotaStatus)
		MAPIFreeBuffer(lpsQuotaStatus);

	return hr;
}
/**
 * Set proccessing meeting request options of a user
 *
 * Use these options if you are responsible for coordinating resources, such as conference rooms.
 *
 * @param[in] lpMsgStore user store to get the options
 * @param[out] bAutoAccept Automatically accept meeting requests and proccess cancellations
 * @param[out] bDeclineConflict Automatically decline conflicting meeting requests
 * @param[out] bDeclineRecurring Automatically decline recurring meeting requests
 *
 * @note because a unknown issue it will update two different free/busy messages, one for 
 *		outlook 2000/xp and one for outlook 2003/2007.
 *
 * @todo find out why outlook 2000/xp opened the wrong local free/busy message
 * @todo check, should the properties PR_SCHDINFO_BOSS_WANTS_COPY, PR_SCHDINFO_DONT_MAIL_DELEGATES, 
 *		PR_SCHDINFO_BOSS_WANTS_INFO on TRUE?
 */
HRESULT SetAutoAcceptSettings(IMsgStore *lpMsgStore, bool bAutoAccept, bool bDeclineConflict, bool bDeclineRecurring)
{
    HRESULT hr = hrSuccess;
    IMessage *lpLocalFBMessage = NULL;
    SPropValue FBProps[6];
        
    // Meaning of these values are unknown, but are always TRUE in cases seen until now
    FBProps[0].ulPropTag = PROP_TAG(PT_BOOLEAN, 0x6842); // PR_SCHDINFO_BOSS_WANTS_COPY
    FBProps[0].Value.b = TRUE;
    FBProps[1].ulPropTag = PROP_TAG(PT_BOOLEAN, 0x6843);//PR_SCHDINFO_DONT_MAIL_DELEGATES
    FBProps[1].Value.b = TRUE;
    FBProps[2].ulPropTag = PROP_TAG(PT_BOOLEAN, 0x684B); // PR_SCHDINFO_BOSS_WANTS_INFO
    FBProps[2].Value.b = TRUE;

    FBProps[3].ulPropTag = PR_PROCESS_MEETING_REQUESTS;
    FBProps[3].Value.b = bAutoAccept ? TRUE : FALSE;
    FBProps[4].ulPropTag = PR_DECLINE_CONFLICTING_MEETING_REQUESTS;
    FBProps[4].Value.b = bDeclineConflict ? TRUE : FALSE;
    FBProps[5].ulPropTag = PR_DECLINE_RECURRING_MEETING_REQUESTS;
    FBProps[5].Value.b = bDeclineRecurring ? TRUE : FALSE;

	// Save localfreebusy settings
	hr = OpenLocalFBMessage(dgFreebusydata, lpMsgStore, true, &lpLocalFBMessage);
    if(hr != hrSuccess)
        goto exit;

    hr = lpLocalFBMessage->SetProps(6, FBProps, NULL);
    if(hr != hrSuccess)
        goto exit;
        
    hr = lpLocalFBMessage->SaveChanges(0);
    if(hr != hrSuccess)
        goto exit;

	if(lpLocalFBMessage) {
        lpLocalFBMessage->Release();
		lpLocalFBMessage = NULL;
	}

	// Hack to support outlook 2000/2002 with resources
	hr = OpenLocalFBMessage(dgAssociated, lpMsgStore, true, &lpLocalFBMessage);
    if(hr != hrSuccess)
        goto exit;

    hr = lpLocalFBMessage->SetProps(6, FBProps, NULL);
    if(hr != hrSuccess)
        goto exit;
        
    hr = lpLocalFBMessage->SaveChanges(0);
    if(hr != hrSuccess)
        goto exit;

exit:
    if(lpLocalFBMessage)
        lpLocalFBMessage->Release();
        
    return hr;
}

/**
 * Get the proccessing meeting request options of a user
 *
 * Use these options if you are responsible for coordinating resources, such as conference rooms.
 *
 * @param[in] lpMsgStore user store to get the options
 * @param[out] lpbAutoAccept Automatically accept meeting requests and proccess cancellations
 * @param[out] lpbDeclineConflict Automatically decline conflicting meeting requests
 * @param[out] lpbDeclineRecurring Automatically decline recurring meeting requests
 *
 * @note you get the outlook 2003/2007 settings
 */
HRESULT GetAutoAcceptSettings(IMsgStore *lpMsgStore, bool *lpbAutoAccept, bool *lpbDeclineConflict, bool *lpbDeclineRecurring)
{
    HRESULT hr = hrSuccess;
    IMessage *lpLocalFBMessage = NULL;
    LPSPropValue lpProps = NULL;
    SizedSPropTagArray(3, sptaFBProps) = {3, {PR_PROCESS_MEETING_REQUESTS, PR_DECLINE_CONFLICTING_MEETING_REQUESTS, PR_DECLINE_RECURRING_MEETING_REQUESTS}};
    ULONG cValues = 0;
    
    bool bAutoAccept = false;
    bool bDeclineConflict = false;
    bool bDeclineRecurring = false;
    
    hr = OpenLocalFBMessage(dgFreebusydata, lpMsgStore, false, &lpLocalFBMessage);
    if(hr != hrSuccess) {
        // No FB -> all settings are FALSE
        hr = hrSuccess;
    } else {
        hr = lpLocalFBMessage->GetProps((LPSPropTagArray)&sptaFBProps, 0, &cValues, &lpProps);
        if(FAILED(hr))
            goto exit;
            
        if(lpProps[0].ulPropTag == PR_PROCESS_MEETING_REQUESTS)
            bAutoAccept = lpProps[0].Value.b;
        if(lpProps[1].ulPropTag == PR_DECLINE_CONFLICTING_MEETING_REQUESTS)
            bDeclineConflict = lpProps[1].Value.b;
        if(lpProps[2].ulPropTag == PR_DECLINE_RECURRING_MEETING_REQUESTS)
            bDeclineRecurring = lpProps[2].Value.b;
    }
    
    *lpbAutoAccept = bAutoAccept;
    *lpbDeclineConflict = bDeclineConflict;
    *lpbDeclineRecurring = bDeclineRecurring;
    
exit:
    if(lpProps)
        MAPIFreeBuffer(lpProps);
    if(lpLocalFBMessage)
        lpLocalFBMessage->Release();
        
    return hr;
}

/*
 * We hould try to be as informative as possible to the user,
 * try to get a nice descriptive message for the error and
 * in the last case just print the hex code.
 */
string getMapiCodeString(HRESULT hr, char* object = "object")
{
	switch (hr) {
	case MAPI_E_NO_SUPPORT:
		return "action not supported by server";
	case MAPI_E_INVALID_PARAMETER:
		return "missing or invalid argument";
	case MAPI_E_NOT_FOUND:
		return string(object) + " not found";
	case MAPI_E_COLLISION:
		return string(object) + " already exists";
	case MAPI_E_NO_ACCESS:
		return "no access to " + string(object);
	};

	return "unknown error (" + stringify(hr, true) + ")";
}

void print_companies(unsigned int cCompanies, LPECCOMPANY lpECCompanies, bool bList) {
	for (unsigned int i = 0; i< cCompanies; i++) {
		if (!bList)
			cout << ((i > 0) ? ", " : "");
		else
			cout << "\t";

		cout << lpECCompanies[i].lpszCompanyname;

		if (bList)
			cout << endl;
	}
}

void print_groups(unsigned int cGroups, LPECGROUP lpECGroups, bool bList) {
	for (unsigned int i = 0; i < cGroups; i++) {
		if (!bList)
			cout << ((i > 0) ? ", " : "");
		else
			cout << "\t";

		cout << lpECGroups[i].lpszGroupname;

		if (bList)
			cout << endl;
	}
}

void print_users(unsigned int cUsers, LPECUSER lpECUsers, bool bList, bool bShowHomeServer = false) {
	for (unsigned int i = 0; i < cUsers; i++) {
		if (!bList)
			cout << ((i > 0) ? ", " : "");
		else
			cout << "\t";

		cout << lpECUsers[i].lpszUsername << "\t\t" << lpECUsers[i].lpszFullName;
        if (bShowHomeServer && lpECUsers[i].lpszServername != NULL && strlen(lpECUsers[i].lpszServername) > 0)
            cout << "\t(" << lpECUsers[i].lpszServername << ")";

		if (bList)
			cout << endl;
	}
}

void print_company_settings(LPECCOMPANY lpECCompany, LPECUSER lpECAdministrator)
{
	cout << "Companyname:\t\t" << lpECCompany->lpszCompanyname << endl;
	cout << "Sysadmin:\t\t" << lpECAdministrator->lpszUsername << endl;
	if (lpECCompany->lpszServername != NULL && strlen(lpECCompany->lpszServername) > 0) {
		cout << "Home server:\t\t" << lpECCompany->lpszServername << endl;
	}
}

void print_group_settings(LPECGROUP lpECGroup)
{
	cout << "Groupname:\t\t" << lpECGroup->lpszGroupname << endl;
	cout << "Fullname:\t\t" << lpECGroup->lpszFullname << endl;
}

void print_user_settings(IMsgStore *lpStore, LPECUSER lpECUser, bool bAutoAccept, bool bDeclineConflict, bool bDeclineRecur)
{
    LPSPropValue lpProps = NULL;
    SizedSPropTagArray(2, sptaProps) = {2, { PR_LAST_LOGON_TIME, PR_LAST_LOGOFF_TIME } };
    ULONG cValues = 0;
    
    lpStore->GetProps((LPSPropTagArray)&sptaProps, 0, &cValues, &lpProps);
    
	cout << "Username:\t\t" << lpECUser->lpszUsername << endl;
	cout << "Fullname:\t\t" << lpECUser->lpszFullName << endl;
	cout << "Emailaddress:\t\t" << lpECUser->lpszFullEmail << endl;
	cout << "Active:\t\t\t" << ((lpECUser->ulIsNonActive==0) ? "yes" : "no") << endl;
	cout << "Administrator:\t\t" << ((lpECUser->ulIsAdmin >= 1) ? "yes" : "no") << ((lpECUser->ulIsAdmin == 2) ? " (system)" : "") << endl;
	cout << "Auto-accept meeting req:" << (bAutoAccept ? "yes" : "no") << endl;
	if (bAutoAccept) {
		cout << "Decline dbl meetingreq:\t" << (bDeclineConflict ? "yes" : "no") << endl;
		cout << "Decline recur meet.req:\t" << (bDeclineRecur ? "yes" : "no") << endl;
	}
	if (lpECUser->lpszServername != NULL && strlen(lpECUser->lpszServername) > 0) {
		cout << "Home server:\t\t" << lpECUser->lpszServername << endl;
	}
	if (lpProps) {
	    time_t logon = 0, logoff = 0;
	    char d[64];
	    
	    if(lpProps[0].ulPropTag == PR_LAST_LOGON_TIME)
    	    FileTimeToUnixTime(lpProps[0].Value.ft, &logon);
	    if(lpProps[1].ulPropTag == PR_LAST_LOGOFF_TIME)
    	    FileTimeToUnixTime(lpProps[1].Value.ft, &logoff);
    	    
        if(logon) {
            strftime(d, sizeof(d), "%x %X", localtime(&logon));
    	    cout << "Last logon:\t\t" << d << std::endl;
        }
        if(logoff) {
            strftime(d, sizeof(d), "%x %X", localtime(&logoff));
    	    cout << "Last logoff:\t\t" << d << std::endl;
        }
    }
    
    if(lpProps)
        MAPIFreeBuffer(lpProps);
}

HRESULT list_orphans(IECServiceAdmin *lpServiceAdmin)
{
	HRESULT hr = hrSuccess;
	ULONG i = 0;
	IMAPITable *lpTable = NULL;
	LPSRowSet lpRowSet = NULL;
	LPSPropValue lpStoreGuid = NULL;
	LPSPropValue lpUserName = NULL;
	std::string strUsername;
	bool bHeader = true;
	const static SizedSSortOrderSet(2, tableSort) =
		{ 2, 0, 0,
		  {
			  { PR_EC_USERNAME, TABLE_SORT_ASCEND },
			  { PR_EC_STOREGUID, TABLE_SORT_ASCEND },
		  }
		};

	hr = lpServiceAdmin->OpenUserStoresTable(0, &lpTable);
	if (hr != hrSuccess) {
		cerr << "Unable to open user/stores table" << endl;
		goto exit;
	}

	hr = lpTable->SortTable((LPSSortOrderSet)&tableSort, 0);
	if (hr != hrSuccess) {
		cerr << "Unable to sort user/stores table" << endl;
		goto exit;
	}

	// Because of the sort, we start with these stores
	cout << "Stores without users:" << endl;
	cout << "\tStore guid\t\t\t\tGuessed username" << endl;
	cout << "\t--------------------------------------------------------" << endl;

	while(TRUE) {
		hr = lpTable->QueryRows(50, 0, &lpRowSet);
		if(hr != hrSuccess) {
			cerr << "Unable to load user/stores table" << endl;
			goto exit;
		}

		if(lpRowSet->cRows == 0)
			break;

		for (i = 0; i < lpRowSet->cRows; i++) {
			lpStoreGuid = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_EC_STOREGUID);
			lpUserName = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_EC_USERNAME);

			if (lpStoreGuid && lpUserName)
				continue;

			if (!lpUserName) {
				// find "guessed" named
				lpUserName = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_DISPLAY_NAME);
				if (lpUserName)
					strUsername = lpUserName->Value.lpszA;
				else
					strUsername = "<unknown>";
			} else {
				// we had all stores without users, now the users without stores
				if (bHeader) {
					cout << endl << "Users without stores:" << endl;
					cout << "\tUsername" << endl;
					cout << "\t---------------" << endl;
					bHeader = false;
				}

				strUsername = lpUserName->Value.lpszA;
			}
			if (lpStoreGuid)
				cout << "\t" << Util::bin2hex(lpStoreGuid->Value.bin.cb, lpStoreGuid->Value.bin.lpb);
			cout << "\t" << strUsername << endl;
		}

		FreeProws(lpRowSet);
		lpRowSet = NULL;
	}

exit:
	if (lpRowSet)
		FreeProws(lpRowSet);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT print_details(LPMAPISESSION lpSession, IECUnknown *lpECMsgStore, IECServiceAdmin *lpServiceAdmin, userobject_type_t ulType, char *lpszName)
{
	HRESULT hr = hrSuccess;
	LPECUSER lpECUser = NULL;
	LPECGROUP lpECGroup = NULL;
	LPECCOMPANY lpECCompany = NULL;
	LPECQUOTASTATUS lpsQuotaStatus = NULL;
	LPECQUOTA lpsQuota = NULL;
	LPECGROUP lpECGroups = NULL;
	LPECUSER lpECUsers = NULL;
	LPECUSER lpECAdmins = NULL;
	LPECCOMPANY lpECViews = NULL;
	ULONG cGroups = 0;
	ULONG cUsers = 0;
	ULONG cAdmins = 0;
	ULONG cViews = 0;
	ULONG cbEntryID = 0;
	LPENTRYID lpEntryID = NULL;
	LPMDB lpStore = NULL;
	IExchangeManageStore *lpIEMS = NULL;
	bool bAutoAccept = false;
	bool bDeclineConflict = false;
	bool bDeclineRecurring = false;
	ULONG cbObjectId = 0;
	LPENTRYID lpObjectId = NULL;

	switch (ulType) {
	case USEROBJECT_TYPE_COMPANY:
		hr = lpServiceAdmin->ResolveCompanyName(lpszName, &cbObjectId, &lpObjectId);
		if (hr != hrSuccess) {
			cerr << "Unable to resolve company, " << getMapiCodeString(hr, lpszName) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetCompany(cbObjectId, lpObjectId, &lpECCompany);
		if (hr != hrSuccess) {
			cerr << "Unable to show company details, " << getMapiCodeString(hr) << endl;
			goto exit;
		}
		
		hr = lpECMsgStore->QueryInterface(IID_IExchangeManageStore, (void **)&lpIEMS);
		if (hr != hrSuccess) {
			cerr << "Unable to get admin interface." << endl;
			goto exit;
		}

		hr = lpIEMS->CreateStoreEntryID("", lpECCompany->lpszCompanyname, 0, &cbEntryID, &lpEntryID);
		if (hr != hrSuccess) {
			cerr << "Unable to get company store entry id. Company possibly has no store." << endl;
			goto exit;
		}

		hr = lpSession->OpenMsgStore(0, cbEntryID, lpEntryID, &IID_IMsgStore, MDB_WRITE, &lpStore);
		if (hr != hrSuccess) {
			cerr << "Unable to open company store." << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetUser(lpECCompany->sAdministrator.cb, (LPENTRYID)lpECCompany->sAdministrator.lpb, &lpECUser);
		if (hr != hrSuccess) {
			cerr << "Unable to resolve company administrator, " << getMapiCodeString(hr) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetRemoteAdminList(cbObjectId, lpObjectId, &cAdmins, &lpECAdmins);
		if (hr != hrSuccess) {
			cerr << "Unable to display remote-admin list, " << getMapiCodeString(hr) << endl;
			hr = hrSuccess; /* Don't make error fatal */
		}

		hr = lpServiceAdmin->GetRemoteViewList(cbObjectId, lpObjectId, &cViews, &lpECViews);
		if (hr != hrSuccess) {
			cerr << "Unable to display remote-view list, " << getMapiCodeString(hr) << endl;
			hr = hrSuccess; /* Don't make error fatal */
		}

		print_company_settings(lpECCompany, lpECUser);

		break;
	case USEROBJECT_TYPE_GROUP:
		hr = lpServiceAdmin->ResolveGroupName(lpszName, &cbObjectId, &lpObjectId);
		if (hr != hrSuccess) {
			cerr << "Unable to resolve group, " << getMapiCodeString(hr, lpszName) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetGroup(cbObjectId, lpObjectId, &lpECGroup);
		if (hr != hrSuccess) {
			cerr << "Unable to show group details, " << getMapiCodeString(hr) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetUserListOfGroup(cbObjectId, lpObjectId, &cUsers, &lpECUsers);
		if (hr != hrSuccess) {
			cerr << "Unable to request users for group, " << getMapiCodeString(hr) << endl;
			hr = hrSuccess; /* Don't make error fatal */
		}

		print_group_settings(lpECGroup);

		break;
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
	default:
		hr = lpServiceAdmin->ResolveUserName(lpszName, &cbObjectId, &lpObjectId);
		if (hr != hrSuccess) {
			cerr << "Unable to resolve user, " << getMapiCodeString(hr, lpszName) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetUser(cbObjectId, lpObjectId, &lpECUser);
		if (hr != hrSuccess) {
			cerr << "Unable to show user details, " << getMapiCodeString(hr) << endl;
			goto exit;
		}

		hr = lpECMsgStore->QueryInterface(IID_IExchangeManageStore, (void **)&lpIEMS);
		if (hr != hrSuccess) {
			cerr << "Unable to get admin interface." << endl;
			goto exit;
		}

		hr = lpIEMS->CreateStoreEntryID("", lpECUser->lpszUsername, 0, &cbEntryID, &lpEntryID);
		if (hr != hrSuccess) {
			cerr << "Unable to get user store entry id. User possibly has no store." << endl;
			goto exit;
		}

		hr = lpSession->OpenMsgStore(0, cbEntryID, lpEntryID, &IID_IMsgStore, MDB_WRITE, &lpStore);
		if (hr != hrSuccess) {
			cerr << "Unable to open user store." << endl;
			goto exit;
		}

		GetAutoAcceptSettings(lpStore, &bAutoAccept, &bDeclineConflict, &bDeclineRecurring);
		/* Ignore return value */

		hr = lpServiceAdmin->GetGroupListOfUser(cbObjectId, lpObjectId, &cGroups, &lpECGroups);
		if (hr != hrSuccess) {
			cerr << "Unable to request groups for user, " << getMapiCodeString(hr) << endl;
			hr = hrSuccess; /* Don't make error fatal */
		}

		print_user_settings(lpStore, lpECUser, bAutoAccept, bDeclineConflict, bDeclineRecurring);

		break;
	}

	/* Group quota is not completely implemented at this time on the server... */
	if (ulType != USEROBJECT_TYPE_GROUP) {
		hr = lpServiceAdmin->GetQuota(cbObjectId, lpObjectId, &lpsQuota);
		if (hr != hrSuccess) {
			cerr << "Unable to show object quota, " << getMapiCodeString(hr) << endl;
			hr = hrSuccess; /* Don't make error fatal */
		} else {
			hr = Util::HrGetQuotaStatus(lpStore, lpsQuota, &lpsQuotaStatus);
			if (hr != hrSuccess) {
				cerr << "Unable to show object quota information, " << getMapiCodeString(hr) << endl;
				hr = hrSuccess; /* Don't make error fatal */
			} else
				print_quota(lpsQuota, lpsQuotaStatus);
		}
	}

	if (cUsers) {
		cout << "Users (" << cUsers << "):" << endl;
		print_users(cUsers, lpECUsers, true, true);
		cout << endl;
	}

	if (cGroups) {
		cout << "Groups (" << cGroups << "):" << endl;
		print_groups(cGroups, lpECGroups, true);
		cout << endl;
	}

	if (cAdmins) {
		cout << "Remote admins (" << cAdmins << "):" << endl;
		print_users(cAdmins, lpECAdmins, true);
		cout << endl;
	}

	if (cViews) {
		cout << "Remote viewers (" << cViews << "):" << endl;
		print_companies(cViews, lpECViews, true);
		cout << endl;
	}

exit:
	if (lpECUser)
		MAPIFreeBuffer(lpECUser);
	if (lpECGroup)
		MAPIFreeBuffer(lpECGroup);
	if (lpECCompany)
		MAPIFreeBuffer(lpECCompany);

	if (lpECUsers)
		MAPIFreeBuffer(lpECUsers);
	if (lpECGroups)
		MAPIFreeBuffer(lpECGroups);
	if (lpECAdmins)
		MAPIFreeBuffer(lpECAdmins);
	if (lpECViews)
		MAPIFreeBuffer(lpECViews);

	if (lpsQuota)
		MAPIFreeBuffer(lpsQuota);
	if (lpsQuotaStatus)
		MAPIFreeBuffer(lpsQuotaStatus);

	if (lpIEMS)
		lpIEMS->Release();
	if (lpStore)
		lpStore->Release();
	if (lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	return hr;
}

HRESULT ListUsers(IECServiceAdmin *lpServiceAdmin, LPECCOMPANY lpCompany)
{
	HRESULT		hr = hrSuccess;
	ECUSER*		lpECUsers = NULL;
	ULONG		cUsers = 0;

	hr = lpServiceAdmin->GetUserList(lpCompany->sCompanyId.cb, (LPENTRYID)lpCompany->sCompanyId.lpb, &cUsers, &lpECUsers);
	if (hr != hrSuccess) {
		cerr << "Unable to list users, " << getMapiCodeString(hr) << endl;
		goto exit;
	}

	cout << "User list for " << lpCompany->lpszCompanyname << "("<< cUsers <<"):" << endl;
	cout << "\t" << "username" << "\t\t" << "fullname" << endl;
	cout << "\t---------------------------------------------" << endl;
	print_users(cUsers, lpECUsers, true, true);
	cout << endl;

exit:
	if (lpECUsers)
		MAPIFreeBuffer(lpECUsers);

	return hr;
}

HRESULT ListGroups(IECServiceAdmin *lpServiceAdmin, LPECCOMPANY lpCompany)
{
	HRESULT		hr = hrSuccess;
	ECGROUP*	lpECGroups = NULL;
	ULONG		cGroups = 0;

	hr = lpServiceAdmin->GetGroupList(lpCompany->sCompanyId.cb, (LPENTRYID)lpCompany->sCompanyId.lpb, &cGroups, &lpECGroups);
	if (hr != hrSuccess) {
		cerr << "Unable to list groups, " << getMapiCodeString(hr) << endl;
		goto exit;
	}

	cout << "Group list for " << lpCompany->lpszCompanyname << "("<< cGroups <<"):" << endl;
	cout << "\t" << "groupname" << "" << endl;
	cout << "\t-------------------------------------" << endl;
	print_groups(cGroups, lpECGroups, true);
	cout << endl;

exit:
	if (lpECGroups)
		MAPIFreeBuffer(lpECGroups);

	return hr;
}

HRESULT SyncUsers(IECServiceAdmin *lpServiceAdmin, LPECCOMPANY lpCompany)
{
	HRESULT hr = hrSuccess;

	hr = lpServiceAdmin->SyncUsers(lpCompany->sCompanyId.cb, (LPENTRYID)lpCompany->sCompanyId.lpb);
	if (hr != hrSuccess) {
		cerr << "User/group synchronization failed, " << getMapiCodeString(hr) << endl;
		goto exit;
	}

exit:
	return hr;
}

HRESULT ForEachCompany(IECServiceAdmin *lpServiceAdmin, char *lpszCompanyName, HRESULT (*lpWork)(IECServiceAdmin *, LPECCOMPANY))
{
	HRESULT hr = hrSuccess;
	ULONG cbCompanyId = 0;
	LPENTRYID lpCompanyId = NULL;
	ULONG cCompanies = 0;

	LPECCOMPANY lpECCompanies = NULL;
	LPECCOMPANY lpECCompaniesAlloc = NULL;
	ECCOMPANY sRootCompany = {{g_cbSystemEid, g_lpSystemEid}, "Default", NULL, {0, NULL}};
	
	if (lpszCompanyName) {
		hr = lpServiceAdmin->ResolveCompanyName(lpszCompanyName, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, lpszCompanyName) << endl;
			goto exit;
		}
			
		cCompanies = 1;
		sRootCompany.sCompanyId.cb = cbCompanyId;
		sRootCompany.sCompanyId.lpb = (LPBYTE)lpCompanyId;
		sRootCompany.lpszCompanyname = lpszCompanyName;
		lpECCompanies = &sRootCompany;
	} else {
		hr = lpServiceAdmin->GetCompanyList(&cCompanies, &lpECCompaniesAlloc);
		if (hr != hrSuccess) {
			cCompanies = 1;
			lpECCompanies = &sRootCompany;
			hr = hrSuccess;
		} else {
			lpECCompanies = lpECCompaniesAlloc;
		}
	}

	if (cCompanies == 0) {
		cerr << "No companies found." << endl;
		goto exit;
	}
		    
	for (unsigned int i = 0; i < cCompanies; i++) {
		hr = lpWork(lpServiceAdmin, &lpECCompanies[i]);
		if (hr != hrSuccess)
			goto exit;
	}

exit:
	if (lpECCompaniesAlloc)
		MAPIFreeBuffer(lpECCompaniesAlloc);

	if (lpCompanyId)
		MAPIFreeBuffer(lpCompanyId);

	return hr;
}

int main(int argc, char* argv[])
{
	HRESULT hr = hrSuccess;
	LPMAPISESSION lpSession = NULL;
	
	IECUnknown *lpECMsgStore = NULL;
	IMsgStore *lpMsgStore = NULL;

	IECServiceAdmin *lpServiceAdmin = NULL;
	ULONG cbUserId = 0;
	LPENTRYID lpUserId = NULL;
	ULONG cbGroupId = 0;
	LPENTRYID lpSenderId = NULL;
	ULONG cbSenderId = 0;
	LPENTRYID lpGroupId = NULL;
	ULONG cbCompanyId = 0;
	LPENTRYID lpCompanyId = NULL;
	ULONG cbSetCompanyId = 0;
	LPENTRYID lpSetCompanyId = NULL;
	ULONG ulDays = 0;
	LPSPropValue lpPropValue = NULL;
	ULONG cbGUID = 0;
	LPGUID lpGUID = NULL;
	
	ECUSER sECUser, *lpECUser = NULL;
	ULONG cbStoreId = 0;
	LPENTRYID lpStoreId = NULL;
	ULONG cbRootId = 0;
	LPENTRYID lpRootId = NULL;

	ECGROUP		sECGroup;
	LPECGROUP	lpECGroups = NULL;
	ULONG		cCompanies = 0;
	ULONG		cUsers = 0;

	ECCOMPANY sECCompany;
	LPECCOMPANY lpECCompanies = NULL;
	
	LPECQUOTASTATUS	lpsQuotaStatus = NULL;
	LPECQUOTA lpsQuota = NULL;

	ULONG cSenders = 0;
	LPECUSER lpSenders = NULL;

	userobject_type_t ulType;
	char *detailstype = NULL;
	char *username = NULL;
	char *groupname = NULL;
	char *companyname = NULL;
	char *set_companyname = NULL;
	char *password = NULL;
	char *emailadr = NULL;
	char *fullname = NULL;
	char *new_username = NULL;
	char *storeguid = NULL;
	char *path = NULL;
	char *lang = NULL;
	int quota = -1;
	long long quotahard = -1;
	long long quotasoft = -1;
	long long quotawarn = -1;
	int ud_quota = -1;
	long long ud_quotahard = -1;
	long long ud_quotasoft = -1;
	long long ud_quotawarn = -1;
	int isadmin = -1;
	int isnonactive = -1;
	int mr_accept = -1;
	int mr_decline_conflict = -1;
	int mr_decline_recurring = -1;
	char *sendas_user = NULL;
	int sendas_action = -1;
	modes mode = MODE_INVALID;
	int passprompt = 0;
	
	bool bAutoAccept = false, bDeclineConflict = false, bDeclineRecurring = false;
	ULONG cbEntryID = 0;
	LPENTRYID lpEntryID = NULL;
	LPMDB lpUserStore = NULL;
	IExchangeManageStore *lpIEMS = NULL;
    
    const configsetting_t lpDefaults[] = {
		{ "server_socket", NULL, CONFIGSETTING_UNUSED },
		{ "sslkey_file", "" },
		{ "sslkey_pass", "" },
		{ NULL, NULL },
	};
    ECConfig sConfig(lpDefaults);
    bool bExplicitConfig = false;
    
    char *szConfig = "/etc/zarafa/admin.cfg";

	// make sure the LANGUAGE environment variable is not set, seems this breaks translations on Debian.
	// putenv is also available on linux, but may break the program: see ECUserManagement.cpp in the server.
	unsetenv("LANGUAGE");
	// Get locale from LC_MESSAGES (ie fr_FR, de_DE, etc)
	setlocale(LC_MESSAGES,"");
	// Specify that we want strings in iso-8859-15 from gettext
	bind_textdomain_codeset("zarafa", "iso-8859-15");

	if(argc < 2) {
		print_help(argv[0]);
		return 1;
	}

	int c;
	while (1) {
		c = my_getopt_long(argc, argv, "VlLsc:u:d:U:Pp:f:e:a:h:g:G:b:B:i:I:n:", long_options, NULL);
		if (c == -1)
			break;
		switch (c) {
		case 'l':
			mode = MODE_LIST_USERS;
			break;
		case 's':
			mode = MODE_CREATE_PUBLIC;
			break;
		case 'c':
			mode = MODE_CREATE_USER;
			username = my_optarg;
			break;
		case 'u':
			if (mode != MODE_HOOK_STORE)
				mode = MODE_UPDATE_USER;
			username = my_optarg;
			break;
		case 'd':
			mode = MODE_DELETE_USER;
			username = my_optarg;
			break;
		case 'g':
			mode = MODE_CREATE_GROUP;
			groupname = my_optarg;
			break;
		case 'G':
			mode = MODE_DELETE_GROUP;
			groupname = my_optarg;
			break;
		case 'L':
			mode = MODE_LIST_GROUP;
			break;
		case 'b':
			mode = MODE_ADDUSER_GROUP;
			username = my_optarg;
			break;
		case 'B':
			mode = MODE_DELETEUSER_GROUP;
			username = my_optarg;
			break;

		case 'U':
			new_username = my_optarg;
			break;
		case 'P':
			passprompt = 1;
			break;
		case 'p':
			password = my_optarg;
			break;
		case 'f':
			fullname = my_optarg;
			break;
		case 'e':
			emailadr = my_optarg;
			break;

		case 'a':
			if (my_optarg[0] == '1' || my_optarg[0] == '2')
				isadmin = atoi(my_optarg);
			else
				isadmin = parse_yesno(my_optarg);
			break;
        case 'n':
            isnonactive = parse_yesno(my_optarg);
            break;
		case 'i':
			groupname = my_optarg;
			break;
		case 'I':
			companyname = my_optarg;
			break;

			// error handling?
		case '?':
			break;
		case OPT_HOST:
		case 'h':
			path = my_optarg;
			break;
		case OPT_HELP:
			mode = MODE_HELP;
			break;
		case OPT_CREATE_STORE:
			mode = MODE_CREATE_STORE;
			username = my_optarg;
			break;
		case OPT_DELETE_STORE:
			mode = MODE_DELETE_STORE;
			username = my_optarg;
			break;
		case OPT_HOOK_STORE:
			mode = MODE_HOOK_STORE;
			storeguid = my_optarg;
			break;
		case OPT_REMOVE_STORE:
			mode = MODE_REMOVE_STORE;
			storeguid = my_optarg;
			break;
		case OPT_SYNC_USERS:
			mode = MODE_SYNC_USERS;
			break;
		case OPT_DETAILS:
			mode = MODE_DETAILS;
			username = my_optarg;
			break;
		case OPT_DETAILS_TYPE:
			detailstype = my_optarg;
			break;
			// Make values from Mb to bytes which the server wants
		case OPT_USER_QUOTA_HARD:
			quotahard = _atoi64(my_optarg) *1024*1024;
			break;
		case OPT_USER_QUOTA_SOFT:
			quotasoft = _atoi64(my_optarg) *1024*1024;
			break;
		case OPT_USER_QUOTA_WARN:
			quotawarn = _atoi64(my_optarg) *1024*1024;
			break;
		case OPT_USER_QUOTA_OVERRIDE:
			quota = parse_yesno(my_optarg);
			break;
		case OPT_USER_DEFAULT_QUOTA_HARD:
			ud_quotahard = _atoi64(my_optarg) * 1024 * 1024;
			break;
		case OPT_USER_DEFAULT_QUOTA_SOFT:
			ud_quotasoft = _atoi64(my_optarg) * 1024 * 1024;
			break;
		case OPT_USER_DEFAULT_QUOTA_WARN:
			ud_quotawarn = _atoi64(my_optarg) * 1024 * 1024;
			break;
		case OPT_USER_DEFAULT_QUOTA_OVERRIDE:
			ud_quota = parse_yesno(my_optarg);
			break;
        case OPT_LANG:
            // Use alternate language
            lang = my_optarg;
            setlocale(LC_MESSAGES, my_optarg);
            break;
        case OPT_MR_ACCEPT:
            mr_accept = parse_yesno(my_optarg);
            break;
        case OPT_MR_DECLINE_CONFLICT:
            mr_decline_conflict = parse_yesno(my_optarg);
            break;
        case OPT_MR_DECLINE_RECURRING:
            mr_decline_recurring = parse_yesno(my_optarg);
            break;
		case OPT_LIST_SENDAS:
			mode = MODE_LIST_SENDAS;
			username = my_optarg;
			break;
		case OPT_ADD_SENDAS:
			sendas_user = my_optarg;
			sendas_action = 1;
			break;
		case OPT_DEL_SENDAS:
			sendas_user = my_optarg;
			sendas_action = 0;
			break;
		case OPT_CREATE_COMPANY:
            mode = MODE_CREATE_COMPANY;
			companyname = my_optarg;
	    	break;
		case OPT_UPDATE_COMPANY:
			mode = MODE_UPDATE_COMPANY;
			companyname = my_optarg;
			break;
		case OPT_DELETE_COMPANY:
            mode = MODE_DELETE_COMPANY;
			companyname = my_optarg;
	    	break;
		case OPT_LIST_COMPANY:
            mode = MODE_LIST_COMPANY;
			set_companyname = my_optarg;
			break;
		case OPT_ADD_VIEW:
			mode = MODE_ADD_VIEW;
			set_companyname = my_optarg;
			break;
		case OPT_DEL_VIEW:
			mode = MODE_DEL_VIEW;
			set_companyname = my_optarg;
			break;
		case OPT_LIST_VIEW:
			mode = MODE_LIST_VIEW;
			break;
		case OPT_ADD_ADMIN:
			mode = MODE_ADD_ADMIN;
			username = my_optarg;
			break;
		case OPT_DEL_ADMIN:
			mode = MODE_DEL_ADMIN;
			username = my_optarg;
			break;
		case OPT_LIST_ADMIN:
			mode = MODE_LIST_ADMIN;
			break;
		case OPT_SYSTEM_ADMIN:
			mode = MODE_SYSTEM_ADMIN;
			username = my_optarg;
			break;
		case OPT_ADD_UQUOTA_RECIPIENT:
			mode = MODE_ADD_USERQUOTA_RECIPIENT;
			username = my_optarg;
			break;
		case OPT_DEL_UQUOTA_RECIPIENT:
			mode = MODE_DEL_USERQUOTA_RECIPIENT;
			username = my_optarg;
			break;
		case OPT_LIST_UQUOTA_RECIPIENT:
			mode = MODE_LIST_USERQUOTA_RECIPIENT;
			break;
		case OPT_ADD_CQUOTA_RECIPIENT:
			mode = MODE_ADD_COMPANYQUOTA_RECIPIENT;
			username = my_optarg;
			break;
		case OPT_DEL_CQUOTA_RECIPIENT:
			mode = MODE_DEL_COMPANYQUOTA_RECIPIENT;
			username = my_optarg;
			break;
		case OPT_LIST_CQUOTA_RECIPIENT:
			mode = MODE_LIST_COMPANYQUOTA_RECIPIENT;
			break;
        case OPT_PURGE_SOFTDELETE:
            ulDays = atoui(my_optarg);
            mode = MODE_PURGE_SOFTDELETE;
            break;
		case OPT_LIST_ORPHANS:
			mode = MODE_LIST_ORPHANS;
			break;
        case OPT_CONFIG:
            szConfig = my_optarg;
            bExplicitConfig = true;
            break;
		case 'V':
			cout << "Product version:\t" <<  PROJECT_VERSION_PROFADMIN_STR << endl
				 << "File version:\t\t" << PROJECT_SVN_REV_STR << endl;
			return 1;
		default:
			break;
		};
	}

	// check empty input
	if (username && strcmp(username,"")==0) {
		cerr << "Username cannot be empty" << endl;
		return 1;
	}
	if (password && strcmp(password,"")==0) {
		cerr << "Password cannot be empty" << endl;
		return 1;
	}
	if (companyname && strcmp(companyname,"")==0){
		cerr << "Companyname cannot be empty" << endl;
		return 1;
	}
	if (groupname && strcmp(groupname,"")==0) {
		cerr << "Groupname cannot be empty" << endl;
		return 1;
	}
	if (fullname && strcmp(fullname,"")==0) {
		cerr << "Fullname cannot be empty" << endl;
		return 1;
	}
	if (emailadr && strcmp(emailadr,"")==0) {
		cerr << "Email address cannot be empty" << endl;
		return 1;
	}


	// check parameters
	if (my_optind < argc) {
		cerr << "Too many options given." << endl;
		return 1;
	}

	if (mode == MODE_INVALID) {
		cerr << "No correct command given." << endl;
		return 1;
	}

	if (mode == MODE_HELP) {
		print_help(argv[0]);
		cout << endl << "Please read zarafa-admin(1) for detailed information by typing 'man zarafa-admin'." << endl << endl;
		return 0;
	}

	      // For the following modes we need a companyname.
	if (!companyname &&
 		(mode == MODE_ADD_VIEW || mode == MODE_DEL_VIEW || mode == MODE_LIST_VIEW ||
		 mode == MODE_ADD_ADMIN || mode == MODE_DEL_ADMIN || mode == MODE_LIST_ADMIN ||
		 mode == MODE_SYSTEM_ADMIN)){
			cerr << "Missing companyname to perform action" << endl;
			return 1;
       }

	if (mode == MODE_DETAILS && username == NULL) {
		cerr << "Missing information to show user details." << endl;
		return 1;
	}

	if (mode == MODE_CREATE_USER && (username == NULL || (password == NULL && passprompt == 0 && isnonactive < 1) || emailadr == NULL || fullname == NULL) ) {
		cerr << "Missing information to create user." << endl;
		return 1;
	}
	if (mode == MODE_CREATE_USER && quota == 1 && (quotahard == -1 || quotawarn == -1 || quotasoft == -1)) {
		cerr << "Not all user specific quota levels are given." << endl;
		cerr << "Missing information to create user." << endl;
		return 1;
	}

	if (mode == MODE_CREATE_COMPANY &&
		((quota == 1 && quotawarn == -1) ||
		 (ud_quota == 1 && (ud_quotahard == -1 || ud_quotasoft == -1 || ud_quotawarn == -1)))) {
		cerr << "Not all company specific quota levels are given." << endl;
		cerr << "Missing information to create company." << endl;
		return 1;
	}

	if (mode == MODE_UPDATE_USER && isnonactive != -1) {
	    cerr << "You cannot update the nonactive flag for existing users. Delete and re-create the user to change the nonactive flag." << endl;
	    return 1;
    }

	if (mode == MODE_CREATE_STORE && username == NULL) {
		cerr << "Missing information to create store." << endl;
		return 1;
	}

	if (mode == MODE_DELETE_STORE) {
		cerr << "Delete store action is not available anymore. Use --remove-store to remove a store from the database." << endl;
		return 1;
	}

	if (mode == MODE_HOOK_STORE && (storeguid == NULL || username == NULL)) {
		cerr << "Missing information to remove store." << endl;
		return 1;
	}

	if (mode == MODE_REMOVE_STORE && storeguid == NULL) {
		cerr << "Missing information to remove store." << endl;
		return 1;
	}

	if (mode == MODE_UPDATE_USER && password == NULL && passprompt == 0 &&
		emailadr == NULL && fullname == NULL && new_username == NULL && isadmin == -1 &&
		quota == -1 && quotahard == -1 && quotasoft == -1 && quotawarn == -1 &&
		mr_accept == -1 && mr_decline_conflict == -1 && mr_decline_recurring == -1 &&
		sendas_user == NULL) {
		cerr << "Missing information to update user." << endl;
		return 1;
	}

	if (mode == MODE_DELETE_USER && username == NULL) {
		cerr << "Missing information to delete user." << endl;
		return 1;
	}

	if (mode == MODE_CREATE_GROUP && groupname == NULL) {
		cerr << "Missing information to create group." << endl;
		return 1;
	}

	if (mode == MODE_DELETE_GROUP && groupname == NULL) {
		cerr << "Missing information to delete group." << endl;
		return 1;
	}
	
	if (mode == MODE_ADDUSER_GROUP && (groupname == NULL || username == NULL)) {
		cerr << "Missing information to add user to group." << endl;
		return 1;
	}

	if (mode == MODE_DELETEUSER_GROUP && (groupname == NULL || username == NULL)) {
		cerr << "Missing information to remove user from group." << endl;
		return 1;
	}
	
	if (mode == MODE_CREATE_COMPANY && companyname == NULL) {
		cerr << "Missing information to create company" << endl;
		return 1;
	}

    if (mode == MODE_UPDATE_COMPANY &&
		((quota == 1 && quotawarn == -1) ||
		 (ud_quota == 1 && (ud_quotahard == -1 || ud_quotasoft == -1 || ud_quotawarn == -1)))) {
		cerr << "Missing information to update company." << endl;
		return 1;
	}

	if (mode == MODE_DELETE_COMPANY && companyname == NULL) {
		cerr << "Missing information to delete company" << endl;
		return 1;
	}

	if (mode == MODE_ADD_VIEW && set_companyname == NULL) {
		cerr << "Missing information to add remote view privilege" << endl;
		return 1;
	}

	if (mode == MODE_DEL_VIEW && set_companyname == NULL) {
		cerr << "Missing information to delete remote view privilege" << endl;
		return 1;
	}

	if (mode == MODE_ADD_ADMIN && username == NULL) {
		cerr << "Missing information to add remote administrator" << endl;
		return 1;
	}

	if (mode == MODE_DEL_ADMIN && username == NULL) {
		cerr << "Missing information to delete remote administrator" << endl;
		return 1;
	}

	if (mode == MODE_SYSTEM_ADMIN && username == NULL) {
		cerr << "Missing information to set system administrator" << endl;
		return 1;
	}

	if ((mode == MODE_ADD_USERQUOTA_RECIPIENT || mode == MODE_DEL_USERQUOTA_RECIPIENT ||
		 mode == MODE_ADD_COMPANYQUOTA_RECIPIENT || mode == MODE_DEL_COMPANYQUOTA_RECIPIENT) &&
		username == NULL) {
		cerr << "Missing information to edit quota recipients" << endl;
		return 1;
	}

	if (lang && mode != MODE_CREATE_STORE) {
	    cerr << "You can only use the --lang option in combination with --create-store. If you" << endl;
		cerr << "wish to create a store in this language, you must edit the 00createstore" << endl;
		cerr << "script, which is probably in /etc/zarafa/userscripts/createuser.d and" << endl;
		cerr << "specify the --lang option there." << endl;
		return 1;
	}

	// check warnings
	if (new_username != NULL && mode != MODE_UPDATE_USER) {
		cerr << "WARNING: new username '" << new_username << "' will be ignored."  << endl;
	}

	if ((quota == 0 && (quotawarn >= 0 || quotasoft >= 0 || quotahard >= 0)) ||
	    (ud_quota == 0 && (ud_quotawarn >= 0 || ud_quotasoft >= 0 || ud_quotahard >= 0))) {
		cerr << "Disabling quota override, but quota levels are provided." << endl;
		cerr << "By disabling quota overrides the existing values will be reset," << endl;
		cerr << "and these new values will be ignored." << endl;
	}

	if ((quota == -1 && (quotawarn >= 0 || quotasoft >= 0 || quotahard >= 0)) ||
		(ud_quota == -1 && (ud_quotawarn >= 0 || ud_quotasoft >= 0 || ud_quotahard >= 0))) {
		cerr << "Quota levels are provided, but not quota level override" << endl;
		cerr << "Without an explicit quota override value the quota levels will be ignored." << endl;
	}
    
	if (szConfig) {
		if (!sConfig.LoadSettings(szConfig) && bExplicitConfig) {
            cerr << "Unable to open configuration file " << szConfig << endl;
            return 1;
        }			
	}

	if(lang)
        setlocale(LC_MESSAGES, lang);	

	//Init mapi
	hr = MAPIInitialize(NULL);
	if (hr != hrSuccess) {
		cerr << "Unable to initialize" << endl;
		goto exit;
	}
    
	path = GetServerUnixSocket(path);

	hr = HrOpenECAdminSession(&lpSession, path, EC_PROFILE_FLAGS_NO_NOTIFICATIONS,
                              sConfig.GetSetting("sslkey_file", "", NULL), 
                              sConfig.GetSetting("sslkey_pass", "", NULL));
	if(hr != hrSuccess) {
		cerr << "Unable to open Admin session." << endl;
		switch (hr) {
		case MAPI_E_NETWORK_ERROR:
			cerr << "The server is not running, or not accessable through " << path << "." << endl;
			break;
		case MAPI_E_LOGON_FAILED:
		case MAPI_E_NO_ACCESS:
			cerr << "Access was denied on " << path << "." << endl;
			break;
		default:
			cerr << "Unknown cause " << stringify(hr,true) << "." << endl;
			break;
		};
		goto exit;
	}

	hr = HrOpenDefaultStore(lpSession,&lpMsgStore);
	if(hr != hrSuccess) {
		cerr << "Unable to open Admin store." << endl;
		goto exit;
	}

	hr = HrGetOneProp(lpMsgStore, PR_EC_OBJECT, &lpPropValue);
	if(hr != hrSuccess || !lpPropValue || !lpPropValue->Value.lpszA) {
		cerr << "Admin object not found." << endl;
		goto exit;
	}

	lpECMsgStore = (IECUnknown *)lpPropValue->Value.lpszA;
	lpECMsgStore->AddRef();

	MAPIFreeBuffer(lpPropValue);
	lpPropValue = NULL;

	hr = lpECMsgStore->QueryInterface(IID_IECServiceAdmin, (void**)&lpServiceAdmin);
	if(hr != hrSuccess) {
		cerr << "Admin object query error." << endl;
		goto exit;
	}

	// fully logged on, action!

	switch(mode) {
	case MODE_LIST_USERS:
		hr = ForEachCompany(lpServiceAdmin, companyname, ListUsers);
		if (hr != hrSuccess)
			goto exit;
		break;
	case MODE_LIST_ORPHANS:
		hr = list_orphans(lpServiceAdmin);
		if (hr != hrSuccess)
			goto exit;
		break;
		
	case MODE_DETAILS:
		if (detailstype == NULL || stricmp(detailstype, "user") == 0)
			ulType = USEROBJECT_TYPE_USER;
		else if (stricmp(detailstype, "group") == 0)
			ulType = USEROBJECT_TYPE_GROUP;
		else if (stricmp(detailstype, "company") == 0)
			ulType = USEROBJECT_TYPE_COMPANY;
		else {
			hr = MAPI_E_INVALID_TYPE;
			cerr << "Unknown userobject type \"" << detailstype << "\"" << endl;
			goto exit;
		}

		hr = print_details(lpSession, lpECMsgStore, lpServiceAdmin, ulType, username);
		if (hr != hrSuccess)
			goto exit;
		break;

	case MODE_CREATE_PUBLIC:
		if (companyname) {
			hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
			if(hr != hrSuccess) {
				cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
				goto exit;
			}
		} else {
			cbCompanyId = g_cbEveryoneEid;
			MAPIAllocateBuffer(g_cbEveryoneEid, (void**)&lpCompanyId);
			memcpy(lpCompanyId, g_lpEveryoneEid, g_cbEveryoneEid);
		}

		/* The public store is created for a particular company, to do this correctly we will
		 * pass the company id as the group id for the store. */
		hr = lpServiceAdmin->CreateStore(ECSTORE_TYPE_PUBLIC, cbCompanyId, lpCompanyId, &cbStoreId, &lpStoreId, &cbRootId, &lpRootId); 
		if(hr != hrSuccess) {
			cerr << "Unable to create store, " << getMapiCodeString(hr, "public") << endl;
			goto exit;
		} else {
			cout << "Public created." << endl;
		}
		break;

	case MODE_CREATE_USER:
		sECUser.sUserId.cb = g_cbDefaultEid;
		sECUser.sUserId.lpb = g_lpDefaultEid;
		sECUser.lpszUsername = username;
		if(passprompt && isnonactive != 1) {
			sECUser.lpszPassword = get_password();
			if(sECUser.lpszPassword == NULL)
				break;
		} else if (isnonactive == 1)
			sECUser.lpszPassword = NULL;
		else
			sECUser.lpszPassword = password;
		sECUser.lpszFullEmail = emailadr;
		sECUser.lpszFullName = fullname;
		sECUser.ulIsAdmin = (isadmin != -1)?isadmin:0;
		sECUser.ulIsNonActive = (isnonactive != -1) ? isnonactive:0;
		
		hr = lpServiceAdmin->CreateUser(&sECUser, &cbUserId, &lpUserId);
		if(hr != hrSuccess) {
			cerr << "Unable to create user, " << getMapiCodeString(hr) << endl;
			goto exit;
		}

		// set quota data
		if (quota != -1 || quotahard != -1 || quotasoft != -1 || quotawarn != -1) {
			hr = setQuota(lpServiceAdmin, cbUserId, lpUserId, quota, false, quotawarn, quotasoft, quotahard);
			if(hr != hrSuccess)
				goto exit;
		}

		cout << "User created." << endl;
		break;

	case MODE_CREATE_STORE:
		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if(hr != hrSuccess) {
			cerr << "Unable to create store, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->CreateStore(ECSTORE_TYPE_PRIVATE, cbUserId, lpUserId, &cbStoreId, &lpStoreId, &cbRootId, &lpRootId); 
		if(hr != hrSuccess) {
			cerr << "Unable to create store, " << getMapiCodeString(hr, "store") << endl;
			goto exit;
		}
		cout << "User store '" << username << "' created." << endl;
		break;

	case MODE_DELETE_USER:
		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Unable to delete user, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}
		hr = lpServiceAdmin->DeleteUser(cbUserId, lpUserId);
		if(hr != hrSuccess) {
			cerr << "Unable to delete user, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}
		cout << "User deleted." << endl;

		break;

	case MODE_DELETE_STORE:
		// happy compiler
		break;
	case MODE_HOOK_STORE:
		hr = Util::hex2bin(storeguid, sizeof(GUID)*2, &cbGUID, (LPBYTE*)&lpGUID);
		if (hr != hrSuccess) {
			cerr << "Incorrect store guid '" << storeguid << "'" << endl;
			goto exit;
		}

		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Unable to find user, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}		

		hr = lpServiceAdmin->HookStore(cbUserId, lpUserId, lpGUID);
		if (hr != hrSuccess) {
			cerr << "Unable to hook store, " << getMapiCodeString(hr) << endl;
			goto exit;
		}
		cout << "Store hooked." << endl;
		break;
	case MODE_REMOVE_STORE:
		hr = Util::hex2bin(storeguid, sizeof(GUID)*2, &cbGUID, (LPBYTE*)&lpGUID);
		if (hr != hrSuccess) {
			cerr << "Incorrect store guid '" << storeguid << "'" << endl;
			goto exit;
		}

		hr = lpServiceAdmin->RemoveStore(lpGUID);
		if (hr != hrSuccess) {
			cerr << "Unable to remove store, " << getMapiCodeString(hr) << endl;
			goto exit;
		}
		cout << "Store removed." << endl;
		break;

	case MODE_UPDATE_USER:
	    if(new_username) {
    		hr = lpServiceAdmin->ResolveUserName(new_username, &cbUserId, &lpUserId);
	    	if (hr == hrSuccess) {
		    	cerr << "User with name '" << new_username << "' is already present." << endl;
    			hr = MAPI_E_COLLISION;
    			goto exit;
    		}
        }

		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Unable to update user, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}
		
		if(new_username || password || passprompt || emailadr || fullname || isadmin != -1) {
            // copy static info
            sECUser.sUserId.cb = cbUserId;
			sECUser.sUserId.lpb = (unsigned char*)lpUserId;
            // possibly set new values
            sECUser.lpszUsername = new_username;
            if(passprompt) {
                sECUser.lpszPassword = get_password();
                if(sECUser.lpszPassword == NULL)
                    break;
            } else
                sECUser.lpszPassword = password;
            sECUser.lpszFullEmail = emailadr;
            sECUser.lpszFullName = fullname;
            sECUser.ulIsAdmin = isadmin;

            hr = lpServiceAdmin->SetUser(&sECUser);
            if(hr != hrSuccess) {
                cerr << "Unable to update user information, " << getMapiCodeString(hr) << endl;
                goto exit;
            }
        }

		// update quota data
		if (quota != -1 || quotahard != -1 || quotasoft != -1 || quotawarn != -1) {
			hr = setQuota(lpServiceAdmin, cbUserId, lpUserId, quota, false, quotawarn, quotasoft, quotahard, true);
			if (hr != hrSuccess)
				goto exit;
		}
	
		if(mr_accept != -1 || mr_decline_conflict != -1 || mr_decline_recurring != -1)
		{
            hr = lpECMsgStore->QueryInterface(IID_IExchangeManageStore, (void **) &lpIEMS);
            if (hr != hrSuccess) {
                cerr << "Unable to get admin interface." << endl;
                goto exit;
            }
            
            hr = lpIEMS->CreateStoreEntryID("", username, 0, &cbEntryID, &lpEntryID);
            if (hr != hrSuccess) {
                cerr << "Unable to get user store entry id. User has possibly has not store." << endl;
                goto exit;
            }
            
            hr = lpSession->OpenMsgStore(0, cbEntryID, lpEntryID, &IID_IMsgStore, MDB_WRITE, &lpUserStore);
            if (hr != hrSuccess) {
                cerr << "Unable to open user store." << endl;
                goto exit;
            }
            
            hr = GetAutoAcceptSettings(lpUserStore, &bAutoAccept, &bDeclineConflict, &bDeclineRecurring);
            if (hr != hrSuccess) {
                hr = hrSuccess; // ignore and assume 'false' for all values
            }
            
            if(mr_accept != -1)
                bAutoAccept = mr_accept;
            if(mr_decline_conflict != -1)
                bDeclineConflict = mr_decline_conflict;
            if(mr_decline_recurring != -1)
                bDeclineRecurring = mr_decline_recurring;
                
            hr = SetAutoAcceptSettings(lpUserStore, bAutoAccept, bDeclineConflict, bDeclineRecurring);
            if(hr != hrSuccess) {
                cerr << "Unable to set auto-accept settings." << endl;
                goto exit;
            } 
		}

		if (sendas_user) {
			hr = lpServiceAdmin->ResolveUserName(sendas_user, &cbSenderId, &lpSenderId);
			if (hr != hrSuccess) {
				cerr << "Unable to update user, sendas user not available, " << getMapiCodeString(hr, sendas_user) << endl;
				goto exit;
			}

			if (sendas_action == 0) {
				// delete sendas user
				hr = lpServiceAdmin->DelSendAsUser(cbUserId, lpUserId, cbSenderId, lpSenderId);
			} else if (sendas_action == 1) {
				// add sendas user
				hr = lpServiceAdmin->AddSendAsUser(cbUserId, lpUserId, cbSenderId, lpSenderId);
			}
			switch (hr) {
			case MAPI_E_NOT_FOUND:
				// on del, not in list
				cerr << "Unable to update user, sender " << sendas_user << " not found in sendas list" << endl;
				goto exit;
			case MAPI_E_COLLISION:
				// on add, already added ... too bad an insert collision does not return this error :|
				cerr << "Unable to update user, sender " << sendas_user << " already in sendas list" << endl;
				goto exit;
			default:
				if (hr != hrSuccess) {
					cerr << "Unable to update user, unable to update sendas list, " << getMapiCodeString(hr, username) << endl;
					goto exit;
				}
				break;
			};
		}

		cout << "User information updated." << endl;

		break;

	case MODE_CREATE_COMPANY:
		sECCompany.lpszCompanyname = companyname;

		hr = lpServiceAdmin->CreateCompany(&sECCompany, &cbCompanyId, &lpCompanyId);
		if(hr != hrSuccess) {
			cerr << "Unable to create company, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		if (quota != -1 || quotahard != -1 || quotasoft != -1 || quotawarn != -1) {
			hr = setQuota(lpServiceAdmin, cbUserId, lpUserId, quota, false, quotawarn, quotasoft, quotahard);
			if(hr != hrSuccess)
				goto exit;
		}

		if (ud_quota != -1 || ud_quotahard != -1 || ud_quotasoft != -1 || ud_quotawarn != -1) {
			hr = setQuota(lpServiceAdmin, cbUserId, lpUserId, ud_quota, true, ud_quotawarn, ud_quotasoft, ud_quotahard);
			if (hr != hrSuccess)
				goto exit;
		}

		cout << "Company created" << endl;
		break;

	case MODE_UPDATE_COMPANY:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if(hr != hrSuccess) {
			cerr << "Unable to resolve company, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}
	
		if (quota != -1 || quotahard != -1 || quotasoft != -1 || quotawarn != -1) {
			hr = setQuota(lpServiceAdmin, cbCompanyId, lpCompanyId, quota, false, quotawarn, quotasoft, quotahard, true);
			if (hr != hrSuccess)
				goto exit;
		}

		if (ud_quota != -1 || ud_quotahard != -1 || ud_quotasoft != -1 || ud_quotawarn != -1) {
			hr = setQuota(lpServiceAdmin, cbCompanyId, lpCompanyId, ud_quota, true, ud_quotawarn, ud_quotasoft, ud_quotahard);
			if (hr != hrSuccess)
				goto exit;
		}

		break;

	case MODE_DELETE_COMPANY:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if(hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->DeleteCompany(cbCompanyId, lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Unable to delete company, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}
		cout << "Company deleted" << endl;
		break;

	case MODE_LIST_COMPANY:
		hr = lpServiceAdmin->GetCompanyList(&cCompanies, &lpECCompanies);
		if(hr != hrSuccess) {
			cerr << "Unable to list companies, " << getMapiCodeString(hr) << endl;
			goto exit;
		}

		cout << "Company list ("<< cCompanies <<"):" << endl;
		cout << "\t" << "companyname" << "\t\t" << "system administrator" << endl;
		cout << "\t-------------------------------------" << endl;
		for (unsigned int i = 0; i < cCompanies; i++) {
			cout << "\t" << lpECCompanies[i].lpszCompanyname;

			hr = lpServiceAdmin->GetUser(lpECCompanies[i].sAdministrator.cb, (LPENTRYID)lpECCompanies[i].sAdministrator.lpb, &lpECUser);
			if (hr != hrSuccess) {
				cerr << "Unable to get administrator details, " << getMapiCodeString(hr, "administrator") << endl;
				goto exit;
			}
			cout << "\t\t" << lpECUser->lpszUsername << endl;
		}
		break;

	case MODE_CREATE_GROUP:
		sECGroup.lpszGroupname = groupname;	
		sECGroup.lpszFullname = groupname;

		hr = lpServiceAdmin->CreateGroup(&sECGroup, &cbGroupId, &lpGroupId);
		if (hr != hrSuccess) {
			cerr << "Unable to create group, " << getMapiCodeString(hr, groupname) << endl;
			goto exit;
		}
		cout << "Group created." << endl;
		break;

	case MODE_DELETE_GROUP:
		hr = lpServiceAdmin->ResolveGroupName(groupname, &cbGroupId, &lpGroupId);
		if (hr != hrSuccess) {
			cerr << "Unable to delete group, " << getMapiCodeString(hr, groupname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->DeleteGroup(cbGroupId, lpGroupId);
		if (hr != hrSuccess) {
			cerr << "Unable to delete group." << endl;
			goto exit;
		}
		cout << "Group deleted." << endl;
		break;

	case MODE_LIST_GROUP:
		hr = ForEachCompany(lpServiceAdmin, companyname, ListGroups);
		if (hr != hrSuccess)
			goto exit;
		break;

	case MODE_ADDUSER_GROUP:
		hr = lpServiceAdmin->ResolveGroupName(groupname, &cbGroupId, &lpGroupId);
		if (hr != hrSuccess) {
			cerr << "Unable to add user to group, " << getMapiCodeString(hr, groupname) << endl;
			goto exit;
		}
		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Unable to add user to group, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}
		hr = lpServiceAdmin->AddGroupUser(cbGroupId, lpGroupId, cbUserId, lpUserId);
		if (hr != hrSuccess) {
			cerr << "Unable to add user to group." << endl;
			goto exit;
		}
		cout << "User added to group." << endl;
		break;

	case MODE_DELETEUSER_GROUP:
		hr = lpServiceAdmin->ResolveGroupName(groupname, &cbGroupId, &lpGroupId);
		if (hr != hrSuccess) {
			cerr << "Unable to remove user from group, " << getMapiCodeString(hr, groupname) << endl;
			goto exit;
		}
		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Unable to remove user from group, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}
		hr = lpServiceAdmin->DeleteGroupUser(cbGroupId, lpGroupId, cbUserId, lpUserId);
		if (hr != hrSuccess) {
			cerr << "Unable to remove user from group." << endl;
			goto exit;
		}
		cout << "User removed from group." << endl;
		break;

	case MODE_SYNC_USERS:
		hr = ForEachCompany(lpServiceAdmin, companyname, SyncUsers);
		if (hr != hrSuccess)
			goto exit;

		cout << "Users and groups synchronized." << endl;

		break;
	case MODE_ADD_VIEW:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->ResolveCompanyName(set_companyname, &cbSetCompanyId, &lpSetCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, set_companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->AddCompanyToRemoteViewList(cbSetCompanyId, lpSetCompanyId, cbCompanyId, lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to add company to remote-view list" << endl;
			goto exit;
		}
		cout << "Company " << set_companyname << " added to the remote-view list of " << companyname << endl; 
		break;
	case MODE_DEL_VIEW:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->ResolveCompanyName(set_companyname, &cbSetCompanyId, &lpSetCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, set_companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->DelCompanyFromRemoteViewList(cbSetCompanyId, lpSetCompanyId, cbCompanyId, lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to remove company from remote-view list, " << getMapiCodeString(hr) << endl;
			goto exit;
		}
		cout << "Company " << set_companyname << " removed from the remote-view list of " << companyname << endl;
		break;
	case MODE_LIST_VIEW:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetRemoteViewList(cbCompanyId, lpCompanyId, &cCompanies, &lpECCompanies);
		if (hr != hrSuccess) {
			cerr << "Unable to display remote-view list, " << getMapiCodeString(hr) << endl;
			goto exit;
		}

		cout << "remote-view list ("<< cCompanies <<"):" << endl;
		cout << "\t" << "companyname" << "" << endl;
		cout << "\t-------------------------------------" << endl;
		print_companies(cCompanies, lpECCompanies, true);
		break;
	case MODE_ADD_ADMIN:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve user name, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->AddUserToRemoteAdminList(cbUserId, lpUserId, cbCompanyId, lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to add user to remote-admin list, " << getMapiCodeString(hr) << endl;
			goto exit;
		}
		cout << "User " << username << " added to the remote-admin list of " << companyname << endl;
		break;
	case MODE_DEL_ADMIN:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve user name, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->DelUserFromRemoteAdminList(cbUserId, lpUserId, cbCompanyId, lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to delete user from remote-admin list, " << getMapiCodeString(hr) << endl;
			goto exit;
		}
		cout << "User " << username << " removed from the remote-admin list of " << companyname << endl;
		break;
	case MODE_LIST_ADMIN:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetRemoteAdminList(cbCompanyId, lpCompanyId, &cUsers, &lpECUser);
		if (hr != hrSuccess) {
			cerr << "Unable to display remote-admin list, " << getMapiCodeString(hr) << endl;
			goto exit;
		}

		cout << "remote-admin list ("<< cUsers <<"):" << endl;
		cout << "\t" << "username" << "" << endl;
		cout << "\t-------------------------------------" << endl;
		print_users(cUsers, lpECUser, true);
		break;
	case MODE_SYSTEM_ADMIN:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve user name, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}

		sECCompany.sAdministrator.cb = cbUserId;
		sECCompany.sAdministrator.lpb = (unsigned char*)lpUserId;
		sECCompany.lpszCompanyname = companyname;
		sECCompany.sCompanyId.cb = cbCompanyId;
		sECCompany.sCompanyId.lpb = (unsigned char*)lpCompanyId;

		hr = lpServiceAdmin->SetCompany(&sECCompany);
		if (hr != hrSuccess) {
			cerr << "Failed to set company system administrator" << endl;
			goto exit;
		}
		cout << "User " << username << " is set as admin of company " << companyname << endl;
		break;
	case MODE_ADD_USERQUOTA_RECIPIENT:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve user name, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->AddQuotaRecipient(cbCompanyId, lpCompanyId, cbUserId, lpUserId, USEROBJECT_TYPE_USER);
		if (hr != hrSuccess) {
			cerr << "Failed to add recipient to quota list." << endl;
			goto exit;
		}
		cout << "User " << username << " added to user quota recipients list for company " << companyname << endl;
		break;
	case MODE_DEL_USERQUOTA_RECIPIENT:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve user name, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}
	
		hr = lpServiceAdmin->DeleteQuotaRecipient(cbCompanyId, lpCompanyId, cbUserId, lpUserId, USEROBJECT_TYPE_USER);
		if (hr != hrSuccess) {
			cerr << "Failed to remove company from quota list." << endl;
			goto exit;
		}
		cout << "User " << username << " removed from user quota recipients list for company " << companyname << endl;
		break;
	case MODE_LIST_USERQUOTA_RECIPIENT:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		/* HACK: request a user from the specified company, and request the recipients for that user. */
		hr = lpServiceAdmin->GetUserList(cbCompanyId, lpCompanyId, &cUsers, &lpECUser);
		if (hr != hrSuccess || cUsers <= 1) /* First user is always SYSTEM */ {
			cerr << "Failed to get quota recipient list" << endl;
			goto exit;
		}

		cbUserId = lpECUser[1].sUserId.cb;
		MAPIAllocateBuffer(cbUserId, (void**)&lpUserId);
		memcpy(lpUserId, lpECUser[1].sUserId.lpb, cbUserId);
		MAPIFreeBuffer(lpECUser);
		lpECUser = NULL;

		hr = lpServiceAdmin->GetQuotaRecipients(cbUserId, lpUserId, &cUsers, &lpECUser);
		if (hr != hrSuccess) {
			cerr << "Failed to get quota recipient list" << endl;
			goto exit;
		}

		cout << "Recipient list ("<< cUsers-1 <<"):" << endl;
		cout << "\t" << "username" << "\t\t" << "fullname" << endl;
		cout << "\t---------------------------------------------" << endl;
		/* Skip the dummy entry we used to obtain the list */
		print_users(cUsers - 1, &lpECUser[1], true);
		break;
	case MODE_ADD_COMPANYQUOTA_RECIPIENT:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve user name, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}
	
		hr = lpServiceAdmin->AddQuotaRecipient(cbCompanyId, lpCompanyId, cbUserId, lpUserId, USEROBJECT_TYPE_COMPANY);
		if (hr != hrSuccess) {
			cerr << "Failed to add recipient to quota list." << endl;
			goto exit;
		}
		cout << "User " << username << " added to company quota recipients list for company " << companyname << endl;
		break;
	case MODE_DEL_COMPANYQUOTA_RECIPIENT:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

	 	hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve user name, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}
	
		hr = lpServiceAdmin->DeleteQuotaRecipient(cbCompanyId, lpCompanyId, cbUserId, lpUserId, USEROBJECT_TYPE_COMPANY);
		if (hr != hrSuccess) {
			cerr << "Failed to delete recipient to quota list." << endl;
			goto exit;
		}
		cout << "User " << username << " removed from company quota recipients list for company " << companyname << endl;
		break;
	case MODE_LIST_COMPANYQUOTA_RECIPIENT:
		hr = lpServiceAdmin->ResolveCompanyName(companyname, &cbCompanyId, &lpCompanyId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve company name, " << getMapiCodeString(hr, companyname) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetQuotaRecipients(cbCompanyId, lpCompanyId, &cUsers, &lpECUser);
		if (hr != hrSuccess) {
			cerr << "Failed to get quota recipient list." << endl;
			goto exit;
		}

		cout << "Recipient list ("<< cUsers-1 <<"):" << endl;
		cout << "\t" << "username" << "\t\t" << "fullname" << endl;
		cout << "\t---------------------------------------------" << endl;
		/* Skipt first entry, that is the company itself which will not get the mail */
		print_users(cUsers - 1, &lpECUser[1], true);
		break;
	case MODE_LIST_SENDAS:
	 	hr = lpServiceAdmin->ResolveUserName(username, &cbUserId, &lpUserId);
		if (hr != hrSuccess) {
			cerr << "Failed to resolve user name, " << getMapiCodeString(hr, username) << endl;
			goto exit;
		}

		hr = lpServiceAdmin->GetSendAsList(cbUserId, lpUserId, &cSenders, &lpSenders);
		if (hr != hrSuccess) {
			cerr << "Failed to retrieve send-as list for user " << username << endl;
			goto exit;
		}

		cout << "Send-as list ("<< cSenders <<") for user " << username << ":" << endl;
		cout << "\t" << "username" << "\t\t" << "fullname" << endl;
		cout << "\t-------------------------------------" << endl;
		print_users(cSenders, lpSenders, true);
		break;
    case MODE_PURGE_SOFTDELETE:
        hr = lpServiceAdmin->PurgeSoftDelete(ulDays);
        if (hr != hrSuccess) {
            cerr << "Softdelete purge failed" << endl;
            goto exit;
        }
        cout << "Softdelete purge done." << endl;
        break;
	case MODE_INVALID:
	case MODE_HELP:
		// happy compiler
		break;
	};

exit:
	if (lpGUID)
		MAPIFreeBuffer(lpGUID);

	if (lpUserId)
		MAPIFreeBuffer(lpUserId);

	if (lpSenderId)
		MAPIFreeBuffer(lpSenderId);

	if (lpGroupId)
		MAPIFreeBuffer(lpGroupId);

	if (lpCompanyId)
		MAPIFreeBuffer(lpCompanyId);

	if (lpSetCompanyId)
		MAPIFreeBuffer(lpSetCompanyId);

	if (lpEntryID)
	    MAPIFreeBuffer(lpEntryID);
	    
    if (lpUserStore)
        lpUserStore->Release();
        
    if (lpIEMS)
        lpIEMS->Release();
	
	if (lpECUser) 
		MAPIFreeBuffer(lpECUser);

	if (lpECGroups)
		MAPIFreeBuffer(lpECGroups);

	if (lpSenders) 
		MAPIFreeBuffer(lpSenders);

	if (lpMsgStore)
		lpMsgStore->Release();

	if(lpECMsgStore)
		lpECMsgStore->Release();

	if (lpServiceAdmin)
		lpServiceAdmin->Release();

	if (lpSession)
		lpSession->Release();

	if (lpStoreId) 
		MAPIFreeBuffer(lpStoreId);

	if (lpRootId) 
		MAPIFreeBuffer(lpRootId);

	if (lpPropValue)
		MAPIFreeBuffer(lpPropValue);

	if (lpsQuota)
		MAPIFreeBuffer(lpsQuota);

	if (lpsQuotaStatus)
		MAPIFreeBuffer(lpsQuotaStatus);

	MAPIUninitialize();
	if (hr == hrSuccess)
		return 0;
	else
		return 1;
}
