/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

// Mapi includes
#include <mapi.h>
#include <mapix.h>
#include <mapiutil.h>
#include <mapidefs.h>

#include <edkguid.h>
#include <edkmdb.h>

//Zarafa includes
#include "ECDefs.h"
#include "ECABEntryID.h"
#include "IECUnknown.h"
#include "Util.h"

//#include "IECSecurity.h"
#include "ECGuid.h"
#include "ECTags.h"
#include "IECServiceAdmin.h"
#include "CommonUtil.h"
#include "stringutil.h"
#include "mapiext.h"
#include "restrictionutil.h"

// Other
#include "ECMonitorDefs.h"
#include "ECQuotaMonitor.h"

ECQuotaMonitor::ECQuotaMonitor(LPECTHREADMONITOR lpThreadMonitor, LPMAPISESSION lpMAPIAdminSession, LPMDB lpMDBAdmin)
{
	m_lpThreadMonitor = lpThreadMonitor;

	m_lpMAPIAdminSession = lpMAPIAdminSession;
	m_lpMDBAdmin = lpMDBAdmin;
	
	m_ulProcessed = 0;
	m_ulFailed = 0;

	if(lpMAPIAdminSession)
		lpMAPIAdminSession->AddRef();
	if(lpMDBAdmin)
		lpMDBAdmin->AddRef();
}

ECQuotaMonitor::~ECQuotaMonitor()
{
	if(m_lpMDBAdmin)
		m_lpMDBAdmin->Release();

	if(m_lpMAPIAdminSession)
		m_lpMAPIAdminSession->Release();
}

void* ECQuotaMonitor::Create(void* lpVoid)
{
	HRESULT				hr = hrSuccess;
	LPECTHREADMONITOR	lpThreadMonitor = (LPECTHREADMONITOR)lpVoid;
	ECQuotaMonitor*		lpecQuotaMonitor = NULL;

	LPMAPISESSION		lpMAPIAdminSession = NULL;
	LPMDB				lpMDBAdmin = NULL;
	time_t				tmStart = 0;
	time_t				tmEnd = 0;

	bool bMapiInit = false;
	char* lpPath = lpThreadMonitor->lpConfig->GetSetting("server_socket");

	lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_INFO, "Quota monitor starting");

	//Open admin session
	hr = HrOpenECAdminSession(&lpMAPIAdminSession, lpPath, 0,
							  lpThreadMonitor->lpConfig->GetSetting("sslkey_file","",NULL),
							  lpThreadMonitor->lpConfig->GetSetting("sslkey_pass","",NULL));
	if (hr != hrSuccess) {
		lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open a admin session. Error 0x%X", hr);
		goto exit;
	}

	lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_INFO, "Connection to Zarafa server succeeded");

	// Open admin store
	hr = HrOpenDefaultStore(lpMAPIAdminSession, &lpMDBAdmin);
	if (hr != hrSuccess) {
		lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open default store for system account");
		goto exit;
	}

	lpecQuotaMonitor = new ECQuotaMonitor(lpThreadMonitor, lpMAPIAdminSession, lpMDBAdmin);

	// Release session and store (Reference on ECQuotaMonitor)
	if(lpMDBAdmin){ lpMDBAdmin->Release(); lpMDBAdmin = NULL;}
	if(lpMAPIAdminSession){ lpMAPIAdminSession->Release(); lpMAPIAdminSession = NULL;}

	// Check the quota of allstores
	tmStart = GetProcessTime();
	hr = lpecQuotaMonitor->CheckQuota();
	tmEnd = GetProcessTime();

	if(hr != hrSuccess)
		lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Quota monitor failed");
	else
		lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_INFO, "Quota monitor done in %lu seconds. Processed: %u, Failed: %u", tmEnd - tmStart, lpecQuotaMonitor->m_ulProcessed, lpecQuotaMonitor->m_ulFailed);
		
exit:
	if(lpecQuotaMonitor)
		delete lpecQuotaMonitor;

	if(lpMDBAdmin)
		lpMDBAdmin->Release();

	if(lpMAPIAdminSession)
		lpMAPIAdminSession->Release();

	return NULL;
}

HRESULT ECQuotaMonitor::CheckQuota()
{
	HRESULT 			hr = hrSuccess;

	/* Service object */
	IECServiceAdmin		*lpServiceAdmin = NULL;
	LPSPropValue		lpsObject = NULL;
	IExchangeManageStore *lpIEMS = NULL;

	/* Companylist */
	LPECCOMPANY			lpsCompanyList = NULL;
	LPECCOMPANY			lpsCompanyListAlloc = NULL;
	ECCOMPANY			sRootCompany = {{g_cbSystemEid, g_lpSystemEid}, "Default", NULL, {0, NULL}};
    ULONG				cCompanies = 0;

	/* Company store */
	ULONG				cbStoreEntryID = NULL;
	LPENTRYID			lpStoreEntryID = NULL;
	LPMDB				lpMsgStore = NULL;

	/* Quota information */
	LPECQUOTA			lpsQuota = NULL;
	LPECQUOTASTATUS		lpsQuotaStatus = NULL;

	/* Obtain Service object */
	hr = HrGetOneProp(m_lpMDBAdmin, PR_EC_OBJECT, &lpsObject);
	if(hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get internal object, error code: 0x%08X", hr);
		goto exit;
	}

	hr = ((IECUnknown *)lpsObject->Value.lpszA)->QueryInterface(IID_IECServiceAdmin, (void **)&lpServiceAdmin);
	if(hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get service admin, error code: 0x%08X", hr);
		goto exit;
	}

	/* Get companylist */
	hr = lpServiceAdmin->GetCompanyList(&cCompanies, &lpsCompanyListAlloc);
	if (hr == MAPI_E_NO_SUPPORT) {
		lpsCompanyList = &sRootCompany;
		cCompanies = 1;
		hr = hrSuccess;
	} else if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get companylist, error code 0x%08X", hr);
		goto exit;
	} else
		lpsCompanyList = lpsCompanyListAlloc;

	hr = m_lpMDBAdmin->QueryInterface(IID_IExchangeManageStore, (void **)&lpIEMS);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get admin interface, error code 0x%08X", hr);
		goto exit;
	}

	for (ULONG i = 0; i < cCompanies; i++) {
		/* Check company quota for non-default company */
		if (lpsCompanyList[i].sCompanyId.cb != 0 && lpsCompanyList[i].sCompanyId.lpb != NULL) {
			m_ulProcessed++;
		
			hr = lpServiceAdmin->GetQuota(lpsCompanyList[i].sCompanyId.cb, (LPENTRYID)lpsCompanyList[i].sCompanyId.lpb, &lpsQuota);
			if (hr != hrSuccess) {
				m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL,
												"Unable to get quota information for company %s, error code: 0x%08X",
												lpsCompanyList[i].lpszCompanyname, hr);
				hr = hrSuccess;
				m_ulFailed++;
				goto check_stores;
			}

			hr = lpIEMS->CreateStoreEntryID("", lpsCompanyList[i].lpszCompanyname, 0, &cbStoreEntryID, &lpStoreEntryID);
			if (hr != hrSuccess) {
				m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL,
												"Unable to get store entry id for company %s, error code: 0x%08X",
												lpsCompanyList[i].lpszCompanyname, hr);
				hr = hrSuccess;
				m_ulFailed++;
				goto check_stores;
			}

			hr = m_lpMAPIAdminSession->OpenMsgStore(0, cbStoreEntryID, lpStoreEntryID, &IID_IMsgStore, MDB_WRITE, &lpMsgStore);
			if (hr != hrSuccess) {
				m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL,
												"Unable to open store for company %s, error code: 0x%08X",
												lpsCompanyList[i].lpszCompanyname, hr);
				hr = hrSuccess;
				m_ulFailed++;
				goto check_stores;
			}

			hr = Util::HrGetQuotaStatus(lpMsgStore, lpsQuota, &lpsQuotaStatus);
			if (hr != hrSuccess) {
				m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL,
												 "Unable to get quotastatus for company %s, error code: 0x%08X",
												 lpsCompanyList[i].lpszCompanyname, hr);
				hr = hrSuccess;
				m_ulFailed++;
				goto check_stores;
			}

			if (lpsQuotaStatus->quotaStatus != QUOTA_OK) {
				m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL,
												 "Storage size of company %s has exceeded one or more size limits",
												 lpsCompanyList[i].lpszCompanyname);
				Notify(NULL, &lpsCompanyList[i], lpsQuotaStatus);
			}
		}

check_stores:
		/* Whatever the status of the company quota, we should also check the quota of the users */
		CheckCompanyQuota(&lpsCompanyList[i]);

		if (lpsQuotaStatus) {
			MAPIFreeBuffer(lpsQuotaStatus);
			lpsQuotaStatus = NULL;
		}

		if (lpMsgStore) {
			lpMsgStore->Release();
			lpMsgStore = NULL;
		}

		if (lpStoreEntryID) {
			MAPIFreeBuffer(lpStoreEntryID);
			lpStoreEntryID = NULL;
		}

		if (lpsQuota) {
			MAPIFreeBuffer(lpsQuota);
			lpsQuota = NULL;
		}
	}

exit:
	if (lpIEMS)
		lpIEMS->Release();

	if(lpServiceAdmin)
		lpServiceAdmin->Release();

	if(lpsObject)
		MAPIFreeBuffer(lpsObject);

	if(lpsCompanyListAlloc)
		 MAPIFreeBuffer(lpsCompanyListAlloc);

	if (lpsQuotaStatus)
		MAPIFreeBuffer(lpsQuotaStatus);

	if (lpMsgStore) 
		lpMsgStore->Release();

	if (lpStoreEntryID) 
		MAPIFreeBuffer(lpStoreEntryID);

	if (lpsQuota) 
		MAPIFreeBuffer(lpsQuota);

	return hr;
}

HRESULT ECQuotaMonitor::CheckCompanyQuota(LPECCOMPANY lpecCompany)
{
	HRESULT				hr = hrSuccess;
	/* Service object */
	IECServiceAdmin		*lpServiceAdmin = NULL;
	LPSPropValue		lpsObject = NULL;
	/* Userlist */
	LPECUSER			lpsUserList = NULL;
	ULONG				cUsers = 0;
	/* 2nd Server connection */
	LPMAPISESSION		lpSession = NULL;
	LPMDB				lpAdminStore = NULL;

	set<string> setServers;
	set<string>::iterator iServers;
	char *lpszConnection = NULL;
	bool bIsPeer = false;;

	/* Obtain Service object */
	hr = HrGetOneProp(m_lpMDBAdmin, PR_EC_OBJECT, &lpsObject);
	if(hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL,
										 "Unable to get internal object, error code: 0x%08X", hr);
		goto exit;
	}

	hr = ((IECUnknown *)lpsObject->Value.lpszA)->QueryInterface(IID_IECServiceAdmin, (void **)&lpServiceAdmin);
	if(hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL,
										 "Unable to get service admin, error code: 0x%08X", hr);
		goto exit;
	}

	/* Get userlist */
	hr = lpServiceAdmin->GetUserList(lpecCompany->sCompanyId.cb, (LPENTRYID)lpecCompany->sCompanyId.lpb, &cUsers, &lpsUserList);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL,
										 "Unable to get userlist for company %s, error code 0x%08X",
										 lpecCompany->lpszCompanyname, hr);
		goto exit;
	}

	for (ULONG i = 0; i < cUsers; i++) {
		if (lpsUserList[i].lpszServername && lpsUserList[i].lpszServername[0] != '\0')
			setServers.insert(lpsUserList[i].lpszServername);
	}

	if (setServers.empty()) {
		// call server function with current lpMDBAdmin / lpServiceAdmin
		
		hr = CheckServerQuota(cUsers, lpsUserList, lpecCompany, m_lpMDBAdmin);
		if (hr != hrSuccess) {
			m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to resolve servername %s, error code 0x%08X", (*iServers).c_str(), hr);
			goto exit;
		}

	} else {
		for (iServers = setServers.begin(); iServers != setServers.end(); iServers++)
		{
			hr = lpServiceAdmin->ResolvePseudoUrl((char*)string("pseudo://"+ (*iServers)).c_str(), &lpszConnection, &bIsPeer);
			if (hr != hrSuccess) {
				m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to resolve servername %s, error code 0x%08X", (*iServers).c_str(), hr);
				continue;
			}

			m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_INFO, "Connecting to server %s using url %s", iServers->c_str(), lpszConnection);

			// call server function with new lpMDBAdmin / lpServiceAdmin
			if (bIsPeer) {
				// query interface
				hr = m_lpMDBAdmin->QueryInterface(IID_IMsgStore, (void**)&lpAdminStore);
				if (hr != hrSuccess) {
					m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get service interface again, error code 0x%08X", hr);
					continue;
				}
			} else {				
				hr = HrOpenECAdminSession(&lpSession, lpszConnection, 0,
										  m_lpThreadMonitor->lpConfig->GetSetting("sslkey_file","",NULL),
										  m_lpThreadMonitor->lpConfig->GetSetting("sslkey_pass","",NULL));
				if (hr != hrSuccess) {
					m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to connect to server %s, error code 0x%08X", lpszConnection, hr);
					goto next;
				}

				hr = HrOpenDefaultStore(lpSession, &lpAdminStore);
				if (hr != hrSuccess) {
					m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open admin store on server %s, error code 0x%08X", lpszConnection, hr);
					goto next;
				}
				
			}

			hr = CheckServerQuota(cUsers, lpsUserList, lpecCompany, lpAdminStore);
			if (hr != hrSuccess)
				m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to check quota on server %s, error code 0x%08X", lpszConnection, hr);

next:
			if (lpSession)
				lpSession->Release();
			lpSession = NULL;

			if (lpAdminStore)
				lpAdminStore->Release();
			lpAdminStore = NULL;
		}
	}

exit:
	if(lpServiceAdmin)
		lpServiceAdmin->Release();

	if(lpsObject)
		MAPIFreeBuffer(lpsObject);

    if(lpsUserList)
		MAPIFreeBuffer(lpsUserList);

	return hr;
}

HRESULT ECQuotaMonitor::CheckServerQuota(ULONG cUsers, LPECUSER lpsUserList, LPECCOMPANY lpecCompany, LPMDB lpAdminStore)
{
	HRESULT hr = hrSuccess;
	LPSRestriction lpsRestriction = NULL;
	SPropValue sRestrictProp;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRowSet = NULL;
	ECQUOTASTATUS sQuotaStatus;
	bool bSendmail = false;
	ULONG i, u;
	SizedSPropTagArray(7, sCols) = {
		7, {
			PR_EC_USERNAME,
			PR_DISPLAY_NAME,
			PR_MESSAGE_SIZE_EXTENDED,
			PR_QUOTA_WARNING_THRESHOLD,
			PR_QUOTA_SEND_THRESHOLD,
			PR_QUOTA_RECEIVE_THRESHOLD,
			PR_EC_QUOTA_MAIL_TIME,
		}
	};

	hr = lpAdminStore->OpenProperty(PR_EC_STATSTABLE_USERS, &IID_IMAPITable, 0, MAPI_DEFERRED_ERRORS, (LPUNKNOWN*)&lpTable);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open stats table for quota sizes, error 0x%08X", hr);
		goto exit;
	}

	hr = lpTable->SetColumns((LPSPropTagArray)&sCols, MAPI_DEFERRED_ERRORS);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set columns on stats table for quota sizes, error 0x%08X", hr);
		goto exit;
	}

	sRestrictProp.ulPropTag = PR_EC_COMPANY_NAME;
	sRestrictProp.Value.lpszA = lpecCompany->lpszCompanyname;

	CREATE_RESTRICTION(lpsRestriction);
	CREATE_RES_OR(lpsRestriction, lpsRestriction, 2);
	  CREATE_RES_NOT(lpsRestriction, &lpsRestriction->res.resOr.lpRes[0]);
	    DATA_RES_EXIST(lpResendInterval, lpsRestriction->res.resOr.lpRes[0].res.resNot.lpRes[0], PR_EC_COMPANY_NAME);
	  DATA_RES_PROPERTY(lpsRestriction, lpsRestriction->res.resOr.lpRes[1], RELOP_EQ, PR_EC_COMPANY_NAME, &sRestrictProp);

	hr = lpTable->Restrict(lpsRestriction, MAPI_DEFERRED_ERRORS);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to restrict stats table, error 0x%08X", hr);
		goto exit;
	}

	while (TRUE) {
		hr = lpTable->QueryRows(50, 0, &lpRowSet);
		if (hr != hrSuccess) {
			m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to receive stats table data, error 0x%08X", hr);
			goto exit;
		}

		if (lpRowSet->cRows == 0)
			break;

		for (i = 0; i < lpRowSet->cRows; i++) {
			LPSPropValue lpUsername = NULL;
			LPSPropValue lpDisplayname = NULL;
			LPSPropValue lpStoreSize = NULL;
			LPSPropValue lpQuotaWarn = NULL;
			LPSPropValue lpQuotaSoft = NULL;
			LPSPropValue lpQuotaHard = NULL;
			LPSPropValue lpQuotaTime = NULL;

			lpUsername = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_EC_USERNAME);
			lpDisplayname = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_DISPLAY_NAME);
			lpStoreSize = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_MESSAGE_SIZE_EXTENDED);
			lpQuotaWarn = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_QUOTA_WARNING_THRESHOLD);
			lpQuotaSoft = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_QUOTA_SEND_THRESHOLD);
			lpQuotaHard = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_QUOTA_RECEIVE_THRESHOLD);
			lpQuotaTime = PpropFindProp(lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues, PR_EC_QUOTA_MAIL_TIME);

			if (!lpUsername || !lpStoreSize)
				continue;		// don't log error: could be for several valid reasons (contacts, other server, etc)

			if (lpStoreSize->Value.li.QuadPart == 0)
				continue;

			m_ulProcessed++;

			memset(&sQuotaStatus, 0, sizeof(ECQUOTASTATUS));

			sQuotaStatus.llStoreSize = lpStoreSize->Value.li.QuadPart;
			sQuotaStatus.quotaStatus = QUOTA_OK;
			if (lpQuotaHard && lpQuotaHard->Value.li.QuadPart > 0 && lpStoreSize->Value.li.QuadPart > lpQuotaHard->Value.li.QuadPart)
				sQuotaStatus.quotaStatus = QUOTA_HARDLIMIT;
			else if (lpQuotaSoft && lpQuotaSoft->Value.li.QuadPart > 0 && lpStoreSize->Value.li.QuadPart > lpQuotaSoft->Value.li.QuadPart)
				sQuotaStatus.quotaStatus = QUOTA_SOFTLIMIT;
			else if (lpQuotaWarn && lpQuotaWarn->Value.li.QuadPart > 0 && lpStoreSize->Value.li.QuadPart > lpQuotaWarn->Value.li.QuadPart)
				sQuotaStatus.quotaStatus = QUOTA_WARN;

			if (sQuotaStatus.quotaStatus == QUOTA_OK)
				continue;

			m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Mailbox of user %s has exceeded one or more size limits",
											 lpDisplayname ? lpDisplayname->Value.lpszA : lpUsername->Value.lpszA);

			if (lpQuotaTime) {
				CheckQuotaInterval(lpQuotaTime, &bSendmail);
				if (!bSendmail) {
					m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_INFO, "Not sending message since the user has already received a warning in the past time interval");
					continue;
				}
			}

			for (u = 0; u < cUsers; u++) {
				if (strcmp(lpsUserList[u].lpszUsername, lpUsername->Value.lpszA) == 0)
					break;
			}
			if (u == cUsers) {
				m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to find user %s in userlist", lpUsername->Value.lpszA);
				m_ulFailed++;
				continue;
			}
			hr = Notify(&lpsUserList[u], lpecCompany, &sQuotaStatus);
			if (hr != hrSuccess)
				m_ulFailed++;
		}

		if (lpRowSet)
			FreeProws(lpRowSet);
		lpRowSet = NULL;
	}

exit:
	if (lpRowSet)
		FreeProws(lpRowSet);

	if (lpsRestriction)
		MAPIFreeBuffer(lpsRestriction);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT ECQuotaMonitor::CreateMailFromTemplate(TemplateVariables *lpVars, string *lpstrSubject, string *lpstrBody)
{
	HRESULT hr = hrSuccess;
	string strTemplateConfig;
	char *lpszTemplate = NULL;
	FILE *fp = NULL;
	char cBuffer[TEMPLATE_LINE_LENGTH];
	string strLine;
	string strSubject;
	string strBody;
	size_t pos;

	string strVariables[7][2] = {
		{ "${ZARAFA_QUOTA_NAME}", "unknown" },
		{ "${ZARAFA_QUOTA_FULLNAME}" , "unknown" },
		{ "${ZARAFA_QUOTA_COMPANY}", "unknown" },
		{ "${ZARAFA_QUOTA_STORE_SIZE}", "unknown" },
		{ "${ZARAFA_QUOTA_WARN_SIZE}", "unlimited" },
		{ "${ZARAFA_QUOTA_SOFT_SIZE}", "unlimited" },
		{ "${ZARAFA_QUOTA_HARD_SIZE}", "unlimited" },
	};

	enum enumVariables {
		ZARAFA_QUOTA_NAME,
		ZARAFA_QUOTA_FULLNAME,
		ZARAFA_QUOTA_COMPANY,
		ZARAFA_QUOTA_STORE_SIZE,
		ZARAFA_QUOTA_WARN_SIZE,
		ZARAFA_QUOTA_SOFT_SIZE,
		ZARAFA_QUOTA_HARD_SIZE,
		ZARAFA_QUOTA_LAST_ITEM /* KEEP LAST! */
	};

	if (lpVars->ulType == USEROBJECT_TYPE_COMPANY)
		strTemplateConfig = "company";
	else
		strTemplateConfig = "user";

	switch (lpVars->ulStatus) {
	case QUOTA_WARN:
			strTemplateConfig += "quota_warning_template";
		break;
	case QUOTA_SOFTLIMIT:
			strTemplateConfig += "quota_soft_template";
		break;
	case QUOTA_HARDLIMIT:
		strTemplateConfig += "quota_warning_template";
		break;
	case QUOTA_OK:
	default:
		goto exit;
	}

	lpszTemplate = m_lpThreadMonitor->lpConfig->GetSetting(strTemplateConfig.c_str());

	/* Start reading the template mail */
	if((fp = fopen(lpszTemplate, "rt")) == NULL) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open template email: %s", lpszTemplate);
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	while(!feof(fp)) {
		memset(&cBuffer, 0, sizeof(cBuffer));

		if (!fgets(cBuffer, sizeof(cBuffer), fp))
			break;

		strLine = string(cBuffer);

		/* If this is the subject line, don't attach it to the mail */
		if (strLine.compare(0, strlen("Subject:"), "Subject:") == 0)
			strSubject = strLine.substr(strLine.find_first_not_of(" ", strlen("Subject:")));
		else
			strBody += strLine;
	}

	fclose(fp);
	fp = NULL;

	if (!lpVars->strUserName.empty())
		strVariables[ZARAFA_QUOTA_NAME][1] = lpVars->strUserName;
	if (!lpVars->strFullName.empty())
		strVariables[ZARAFA_QUOTA_FULLNAME][1] = lpVars->strFullName;
	if (!lpVars->strCompany.empty())
		strVariables[ZARAFA_QUOTA_COMPANY][1] = lpVars->strCompany;
	if (!lpVars->strStoreSize.empty())
		strVariables[ZARAFA_QUOTA_STORE_SIZE][1] = lpVars->strStoreSize;
	if (!lpVars->strWarnSize.empty())
		strVariables[ZARAFA_QUOTA_WARN_SIZE][1] = lpVars->strWarnSize;
	if (!lpVars->strSoftSize.empty())
		strVariables[ZARAFA_QUOTA_SOFT_SIZE][1] = lpVars->strSoftSize;
	if (!lpVars->strHardSize.empty())
		strVariables[ZARAFA_QUOTA_HARD_SIZE][1] = lpVars->strHardSize;

	for (unsigned int i = 0; i < ZARAFA_QUOTA_LAST_ITEM; i++) {
		pos = 0;
		while ((pos = strSubject.find(strVariables[i][0], pos)) != string::npos) {
			strSubject.replace(pos, strVariables[i][0].size(), strVariables[i][1]);
			pos += strVariables[i][1].size();
		}

		pos = 0;
		while ((pos = strBody.find(strVariables[i][0], pos)) != string::npos) {
			strBody.replace(pos, strVariables[i][0].size(), strVariables[i][1]);
			pos += strVariables[i][1].size();
		}
	}

	/* Clear end-of-line characters from subject */
	pos = strSubject.find('\n');
	if (pos != string::npos)
		strSubject.erase(pos);
	pos = strSubject.find('\r');
	if (pos != string::npos)
		strSubject.erase(pos);

	/* Clear starting blank lines from body */
	while (strBody[0] == '\r' || strBody[0] == '\n')
		strBody.erase(0, 1);

	*lpstrSubject = strSubject;
	*lpstrBody = strBody;

exit:
	return hr;
}

HRESULT ECQuotaMonitor::CreateMessageProperties(LPECUSER lpecToUser, LPECUSER lpecFromUser,
												string strSubject, string strBody,
												ULONG *lpcPropSize, LPSPropValue *lppPropArray)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpPropArray = NULL;
	ULONG cbFromEntryid = 0;
	LPENTRYID lpFromEntryid = NULL;
    ULONG cbToEntryid = 0;
	LPENTRYID lpToEntryid = NULL;
	ULONG cbFromSearchKey = 0;
	LPBYTE lpFromSearchKey = NULL;
	ULONG cbToSearchKey = 0;
	LPBYTE lpToSearchKey = NULL;
	ULONG ulPropArrayMax = 50;
	ULONG ulPropArrayCur = 0;
	FILETIME ft;

	/* We are almost there, we have the mail and the recipients. Now we should create the Message */
	hr = MAPIAllocateBuffer(sizeof(SPropValue)  * ulPropArrayMax, (void**)&lpPropArray);
	if (hr != hrSuccess)
		goto exit;

	hr = ECCreateOneOff(lpecToUser->lpszFullName, "SMTP", lpecToUser->lpszFullEmail,
						MAPI_UNICODE | MAPI_SEND_NO_RICH_INFO, &cbToEntryid, &lpToEntryid);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed creating one-off address: 0x%08X", hr);
		goto exit;
	}

	hr = HrCreateEmailSearchKey("SMTP", lpecToUser->lpszFullEmail, &cbToSearchKey, &lpToSearchKey);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed creating email searchkey: 0x%08X", hr);
		goto exit;
	}

	hr = ECCreateOneOff(lpecFromUser->lpszFullName, "SMTP", lpecFromUser->lpszFullEmail,
						MAPI_UNICODE | MAPI_SEND_NO_RICH_INFO, &cbFromEntryid, &lpFromEntryid);
	if(hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed creating one-off address: 0x%08X", hr);
		goto exit;
	}

	hr = HrCreateEmailSearchKey("SMTP", lpecFromUser->lpszFullEmail, &cbFromSearchKey, &lpFromSearchKey);
	if(hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed creating email searchkey: 0x%08X", hr);
		goto exit;
	}

	/* Messageclass */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_MESSAGE_CLASS;
	lpPropArray[ulPropArrayCur++].Value.lpszA = "IPM.Note.StorageQuotaWarning";

	/* Priority */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_PRIORITY;
	lpPropArray[ulPropArrayCur++].Value.ul = PRIO_URGENT;

	/* Importance */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_IMPORTANCE;
	lpPropArray[ulPropArrayCur++].Value.ul = IMPORTANCE_HIGH;

	/* Set message flags */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_MESSAGE_FLAGS;
	lpPropArray[ulPropArrayCur++].Value.ul = MSGFLAG_UNMODIFIED;

	/* Subject */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_SUBJECT;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)strSubject.c_str();

	/* PR_CONVERSATION_TOPIC */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_CONVERSATION_TOPIC;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)strSubject.c_str();

	/* Body */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_BODY;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)strBody.c_str();

	/* RCVD_REPRESENTING_* properties */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_RCVD_REPRESENTING_ADDRTYPE;
	lpPropArray[ulPropArrayCur++].Value.lpszA = "SMTP";

	lpPropArray[ulPropArrayCur].ulPropTag = PR_RCVD_REPRESENTING_EMAIL_ADDRESS;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)((lpecToUser->lpszFullEmail) ? lpecToUser->lpszFullEmail : "");


	hr = MAPIAllocateMore(cbToEntryid, lpPropArray,
						  (void**)&lpPropArray[ulPropArrayCur].Value.bin.lpb);
	if (hr != hrSuccess)
		goto exit;

	lpPropArray[ulPropArrayCur].ulPropTag = PR_RCVD_REPRESENTING_ENTRYID;
	lpPropArray[ulPropArrayCur].Value.bin.cb = cbToEntryid;
	memcpy(lpPropArray[ulPropArrayCur++].Value.bin.lpb,
		   lpToEntryid, cbToEntryid);

	lpPropArray[ulPropArrayCur].ulPropTag = PR_RCVD_REPRESENTING_NAME;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)((lpecToUser->lpszFullName) ? lpecToUser->lpszFullName : "");

	hr = MAPIAllocateMore(cbToSearchKey, lpPropArray,
						  (void**)&lpPropArray[ulPropArrayCur].Value.bin.lpb);
	if (hr != hrSuccess)
		goto exit;

	lpPropArray[ulPropArrayCur].ulPropTag = PR_RCVD_REPRESENTING_SEARCH_KEY;
	lpPropArray[ulPropArrayCur].Value.bin.cb = cbToSearchKey;
	memcpy(lpPropArray[ulPropArrayCur++].Value.bin.lpb,
		   lpToSearchKey, cbToSearchKey);

	/* RECEIVED_BY_* properties */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_RECEIVED_BY_ADDRTYPE;
	lpPropArray[ulPropArrayCur++].Value.lpszA = "SMTP";

	lpPropArray[ulPropArrayCur].ulPropTag = PR_RECEIVED_BY_EMAIL_ADDRESS;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)((lpecToUser->lpszFullEmail) ? lpecToUser->lpszFullEmail : "");

	hr = MAPIAllocateMore(cbToEntryid, lpPropArray,
						  (void**)&lpPropArray[ulPropArrayCur].Value.bin.lpb);
	if (hr != hrSuccess)
		goto exit;

	lpPropArray[ulPropArrayCur].ulPropTag = PR_RECEIVED_BY_ENTRYID;
	lpPropArray[ulPropArrayCur].Value.bin.cb = cbToEntryid;
	memcpy(lpPropArray[ulPropArrayCur++].Value.bin.lpb,
		   lpToEntryid, cbToEntryid);

	lpPropArray[ulPropArrayCur].ulPropTag = PR_RECEIVED_BY_NAME;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)((lpecToUser->lpszFullName) ? lpecToUser->lpszFullName : "");

	hr = MAPIAllocateMore(cbToSearchKey, lpPropArray,
						  (void**)&lpPropArray[ulPropArrayCur].Value.bin.lpb);
	if (hr != hrSuccess)
		goto exit;

	lpPropArray[ulPropArrayCur].ulPropTag = PR_RECEIVED_BY_SEARCH_KEY;
	lpPropArray[ulPropArrayCur].Value.bin.cb = cbToSearchKey;
	memcpy(lpPropArray[ulPropArrayCur++].Value.bin.lpb,
		   lpToSearchKey, cbToSearchKey);

	/* System user, PR_SENDER* */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENDER_ADDRTYPE;
	lpPropArray[ulPropArrayCur++].Value.lpszA = "SMTP";

	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENDER_EMAIL_ADDRESS;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)((lpecFromUser->lpszFullEmail) ? lpecFromUser->lpszFullEmail : "");

	hr = MAPIAllocateMore(cbFromEntryid, lpPropArray,
						  (void**)&lpPropArray[ulPropArrayCur].Value.bin.lpb);
	if (hr != hrSuccess)
		goto exit;

	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENDER_ENTRYID;
	lpPropArray[ulPropArrayCur].Value.bin.cb = cbFromEntryid;
	memcpy(lpPropArray[ulPropArrayCur++].Value.bin.lpb,
		   lpFromEntryid, cbFromEntryid);

	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENDER_NAME;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)((lpecFromUser->lpszFullName) ? lpecFromUser->lpszFullName : "Zarafa-system");

	hr = MAPIAllocateMore(cbFromSearchKey, lpPropArray,
						  (void**)&lpPropArray[ulPropArrayCur].Value.bin.lpb);
	if (hr != hrSuccess)
		goto exit;

	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENDER_SEARCH_KEY;
	lpPropArray[ulPropArrayCur].Value.bin.cb = cbFromSearchKey;
	memcpy(lpPropArray[ulPropArrayCur++].Value.bin.lpb,
		   lpFromSearchKey, cbFromSearchKey);

	/* System user, PR_SENT_REPRESENTING* */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENT_REPRESENTING_ADDRTYPE;
	lpPropArray[ulPropArrayCur++].Value.lpszA = "SMTP";

	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENT_REPRESENTING_EMAIL_ADDRESS;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)((lpecFromUser->lpszFullEmail) ? lpecFromUser->lpszFullEmail : "");

	hr = MAPIAllocateMore(cbFromEntryid, lpPropArray,
						  (void**)&lpPropArray[ulPropArrayCur].Value.bin.lpb);
	if (hr != hrSuccess)
		goto exit;

	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENT_REPRESENTING_ENTRYID;
	lpPropArray[ulPropArrayCur].Value.bin.cb = cbFromEntryid;
	memcpy(lpPropArray[ulPropArrayCur++].Value.bin.lpb,
		   lpFromEntryid, cbFromEntryid);

	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENT_REPRESENTING_NAME;
	lpPropArray[ulPropArrayCur++].Value.lpszA = (LPSTR)((lpecFromUser->lpszFullName) ? lpecFromUser->lpszFullName : "Zarafa-system");

	hr = MAPIAllocateMore(cbFromSearchKey, lpPropArray,
						  (void**)&lpPropArray[ulPropArrayCur].Value.bin.lpb);
	if (hr != hrSuccess)
		goto exit;

	lpPropArray[ulPropArrayCur].ulPropTag = PR_SENT_REPRESENTING_SEARCH_KEY;
	lpPropArray[ulPropArrayCur].Value.bin.cb = cbFromSearchKey;
	memcpy(lpPropArray[ulPropArrayCur++].Value.bin.lpb,
		   lpFromSearchKey, cbFromSearchKey);

	/* Get the time to add to the message as PR_CLIENT_SUBMIT_TIME */
	GetSystemTimeAsFileTime(&ft);

	/* Submit time */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_CLIENT_SUBMIT_TIME;
	lpPropArray[ulPropArrayCur++].Value.ft = ft;

	/* Delivery time */
	lpPropArray[ulPropArrayCur].ulPropTag = PR_MESSAGE_DELIVERY_TIME;
	lpPropArray[ulPropArrayCur++].Value.ft = ft;

	ASSERT(ulPropArrayCur <= ulPropArrayMax);

	*lppPropArray = lpPropArray;
	*lpcPropSize = ulPropArrayCur;

exit:
	if (lpToEntryid)
		MAPIFreeBuffer(lpToEntryid);

	if (lpFromEntryid)
		MAPIFreeBuffer(lpFromEntryid);

	if (lpToSearchKey)
		MAPIFreeBuffer(lpToSearchKey);

	if(lpFromSearchKey)
		MAPIFreeBuffer(lpFromSearchKey);

	return hr;
}

HRESULT ECQuotaMonitor::CreateRecipientList(ULONG cToUsers, LPECUSER lpToUsers, LPADRLIST *lppAddrList)
{
	HRESULT hr = hrSuccess;
	LPADRLIST lpAddrList = NULL;
	ULONG cbUserEntryid = 0;
	LPENTRYID lpUserEntryid = NULL;
	ULONG cbUserSearchKey = 0;
	LPBYTE lpUserSearchKey = NULL;

	hr = MAPIAllocateBuffer(CbNewADRLIST(cToUsers), (void**)&lpAddrList);
	if(hr != hrSuccess)
		goto exit;

	lpAddrList->cEntries = cToUsers;
	for (ULONG i = 0; i < cToUsers; i++) {
		lpAddrList->aEntries[i].cValues = 7;

		hr = MAPIAllocateBuffer(sizeof(SPropValue) * lpAddrList->aEntries[i].cValues,
						  (void**)&lpAddrList->aEntries[i].rgPropVals);
		if (hr != hrSuccess)
			goto exit;

		hr = ECCreateOneOff(lpToUsers[i].lpszFullName, "SMTP", lpToUsers[i].lpszFullEmail,
							MAPI_UNICODE | MAPI_SEND_NO_RICH_INFO, &cbUserEntryid, &lpUserEntryid);
		if (hr != hrSuccess) {
			m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed creating one-off address: 0x%08X", hr);
			goto exit;
		}

		hr = HrCreateEmailSearchKey("SMTP", lpToUsers[i].lpszFullEmail, &cbUserSearchKey, &lpUserSearchKey);
		if (hr != hrSuccess) {
			m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed creating email searchkey: 0x%08X", hr);
			goto exit;
		}

		lpAddrList->aEntries[i].rgPropVals[0].ulPropTag = PR_ROWID;
		lpAddrList->aEntries[i].rgPropVals[0].Value.l = 0;

		lpAddrList->aEntries[i].rgPropVals[1].ulPropTag = PR_RECIPIENT_TYPE;
		lpAddrList->aEntries[i].rgPropVals[1].Value.l = ((i == 0) ? MAPI_TO : MAPI_CC);

		lpAddrList->aEntries[i].rgPropVals[2].ulPropTag = PR_DISPLAY_NAME;
		lpAddrList->aEntries[i].rgPropVals[2].Value.lpszA = (lpToUsers[i].lpszFullName) ? lpToUsers[i].lpszFullName : lpToUsers[i].lpszUsername;

		lpAddrList->aEntries[i].rgPropVals[3].ulPropTag = PR_EMAIL_ADDRESS;
		lpAddrList->aEntries[i].rgPropVals[3].Value.lpszA = (LPSTR)((lpToUsers[i].lpszFullEmail) ? lpToUsers[i].lpszFullEmail : "");

		lpAddrList->aEntries[i].rgPropVals[4].ulPropTag = PR_ADDRTYPE;
		lpAddrList->aEntries[i].rgPropVals[4].Value.lpszA = "SMTP";

		hr = MAPIAllocateMore(cbUserEntryid, lpAddrList->aEntries[i].rgPropVals,
							  (void**)&lpAddrList->aEntries[i].rgPropVals[5].Value.bin.lpb);
		if (hr != hrSuccess)
			goto exit;

		lpAddrList->aEntries[i].rgPropVals[5].ulPropTag = PR_ENTRYID;
		lpAddrList->aEntries[i].rgPropVals[5].Value.bin.cb = cbUserEntryid;
		memcpy(lpAddrList->aEntries[i].rgPropVals[5].Value.bin.lpb,
			   lpUserEntryid, cbUserEntryid);

		hr = MAPIAllocateMore(cbUserSearchKey, lpAddrList->aEntries[i].rgPropVals,
							  (void**)&lpAddrList->aEntries[i].rgPropVals[6].Value.bin.lpb);
		if (hr != hrSuccess)
			goto exit;
		lpAddrList->aEntries[i].rgPropVals[6].ulPropTag = PR_SEARCH_KEY;
		lpAddrList->aEntries[i].rgPropVals[6].Value.bin.cb = cbUserSearchKey;
		memcpy(lpAddrList->aEntries[i].rgPropVals[6].Value.bin.lpb,
			   lpUserSearchKey, cbUserSearchKey);

		if (lpUserEntryid) {
			MAPIFreeBuffer(lpUserEntryid);
			lpUserEntryid = NULL;
		}

		if (lpUserSearchKey) {
			MAPIFreeBuffer(lpUserSearchKey);
			lpUserSearchKey = NULL;
		}
	}

	*lppAddrList = lpAddrList;

exit:
	if (lpUserEntryid)
		MAPIFreeBuffer(lpUserEntryid);

	if(lpUserSearchKey)
		MAPIFreeBuffer(lpUserSearchKey);

	if (hr != hrSuccess && lpAddrList)
		FreePadrlist(lpAddrList);

	return hr;
}

HRESULT ECQuotaMonitor::SendQuotaWarningMail(IMsgStore* lpMDB, ULONG cPropSize, LPSPropValue lpPropArray, LPADRLIST lpAddrList)
{
	HRESULT hr = hrSuccess;
	IMessage *lpMessage = NULL;
	ULONG cbEntryID = 0;
	LPENTRYID lpEntryID = NULL;
	IMAPIFolder *lpInbox = NULL;
	ULONG ulObjType;

	/* Get the entry id of the inbox */
	hr = lpMDB->GetReceiveFolder("IPM", 0, &cbEntryID, &lpEntryID, NULL);;
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to resolve incoming folder, error code: 0x%08X", hr);
		goto exit;
	}

	/* Open the inbox */
	hr = lpMDB->OpenEntry(cbEntryID, lpEntryID, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN *)&lpInbox);
	if (hr != hrSuccess || ulObjType != MAPI_FOLDER) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open inbox folder, error code: 0x%08X", hr);
		if(ulObjType != MAPI_FOLDER)
			hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	/* Create a new message in the correct folder */
	hr = lpInbox->CreateMessage(NULL, 0, &lpMessage);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_INFO, "Unable to create new message, error code: %08X", hr);
		goto exit;
	}

	hr = lpMessage->SetProps(cPropSize, lpPropArray, NULL);
	if(hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set properties, error code: 0x%08X", hr);
		goto exit;
	}

	hr = lpMessage->ModifyRecipients(MODRECIP_ADD, lpAddrList);
	if (hr != hrSuccess)
		goto exit;

	/* Save Message */
	hr = lpMessage->SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;

	hr = HrNewMailNotification(lpMDB, lpMessage, m_lpThreadMonitor->lpLogger);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMDB->SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;

exit:
	if (lpInbox)
		lpInbox->Release();

	if (lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	if(lpMessage)
		lpMessage->Release();

	return hr;
}

HRESULT ECQuotaMonitor::CreateQuotaWarningMail(TemplateVariables *lpVars,
											   IMsgStore* lpMDB, LPECUSER lpecToUser, LPECUSER lpecFromUser,
											   LPADRLIST lpAddrList)
{
	HRESULT hr = hrSuccess;
	ULONG cPropSize = 0;
	LPSPropValue lpPropArray = NULL;
	string strSubject;
	string strBody;

	hr = CreateMailFromTemplate(lpVars, &strSubject, &strBody);
	if (hr != hrSuccess)
		goto exit;

	hr =  CreateMessageProperties(lpecToUser, lpecFromUser, strSubject, strBody, &cPropSize, &lpPropArray);
	if (hr != hrSuccess)
		goto exit;

	hr = SendQuotaWarningMail(lpMDB, cPropSize, lpPropArray, lpAddrList);
	if (hr != hrSuccess)
		goto exit;

exit:
	if (lpPropArray)
		MAPIFreeBuffer(lpPropArray);

	return hr;
}

HRESULT ECQuotaMonitor::CheckQuotaInterval(LPSPropValue lpsPropTime, bool *lpbSendmail)
{
	HRESULT hr = hrSuccess;
	char *lpResendInterval = NULL;
	ULONG ulResendInterval = 0;
	FILETIME ft;
	FILETIME ftNextRun;

	/* Determine when the last warning mail was send, and if a new one should be send. */
	lpResendInterval = m_lpThreadMonitor->lpConfig->GetSetting("mailquota_resend_interval");
	ulResendInterval = (lpResendInterval && atoui(lpResendInterval) > 0) ? atoui(lpResendInterval) : 1;
	GetSystemTimeAsFileTime(&ft);

	UnixTimeToFileTime(FileTimeToUnixTime(lpsPropTime->Value.ft.dwHighDateTime, lpsPropTime->Value.ft.dwLowDateTime) +
					   (ulResendInterval * 60 * 60 * 24) -(2 * 60), &ftNextRun);

	*lpbSendmail = (ft > ftNextRun);

	return hr;
}

HRESULT ECQuotaMonitor::UpdateQuotaTimestamp(IMsgStore* lpMDB)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropTime;
	FILETIME ft;

	GetSystemTimeAsFileTime(&ft);

	sPropTime.ulPropTag = PR_EC_QUOTA_MAIL_TIME;
	sPropTime.Value.ft = ft;

	hr = HrSetOneProp(lpMDB, &sPropTime);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to add sent time, error code: 0x%08X", hr);
		goto exit;
	}

	hr = lpMDB->SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;

exit:
	return hr;
}

HRESULT ECQuotaMonitor::Notify(LPECUSER lpecUser, LPECCOMPANY lpecCompany, LPECQUOTASTATUS lpecQuotaStatus)
{
	HRESULT hr = hrSuccess;
	IECServiceAdmin *lpServiceAdmin = NULL;
	IExchangeManageStore *lpIEMS = NULL;
	LPMDB lpMDB = NULL;
	LPSPropValue lpsObject = NULL;
	ULONG cbUserStoreEntryID = 0;
	LPENTRYID lpUserStoreEntryID = NULL;
	LPADRLIST lpAddrList = NULL;
	LPECUSER lpecFromUser = NULL;
	ULONG cToUsers = 0;
	LPECUSER lpToUsers = NULL;
	LPECQUOTA lpecQuota = NULL;
	ULONG cbUserId = 0;
	LPENTRYID lpUserId = NULL;
	struct TemplateVariables sVars;
	LPSPropValue lpsPropTime = NULL;
	bool bSendmail = true;

	hr = HrGetOneProp(m_lpMDBAdmin, PR_EC_OBJECT, &lpsObject);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get internal object, error code: 0x%08X", hr);
		goto exit;
	}

	hr = ((IECUnknown *)lpsObject->Value.lpszA)->QueryInterface(IID_IECServiceAdmin, (void **)&lpServiceAdmin);
	if (hr != hrSuccess) {
		m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get service admin, error code: 0x%08X", hr);
		goto exit;
	}

	hr = m_lpMDBAdmin->QueryInterface(IID_IExchangeManageStore, (void **)&lpIEMS);
	if (hr != hrSuccess)
		goto exit;

	hr = lpServiceAdmin->GetUser(lpecCompany->sAdministrator.cb, (LPENTRYID)lpecCompany->sAdministrator.lpb, &lpecFromUser);
	if (hr != hrSuccess)
		goto exit;

	if (lpecUser) {
		cbUserId = lpecUser->sUserId.cb;
		lpUserId = (LPENTRYID)lpecUser->sUserId.lpb;
		sVars.ulType = USEROBJECT_TYPE_USER;
		sVars.strUserName = lpecUser->lpszUsername;
		sVars.strFullName = lpecUser->lpszFullName;
		sVars.strCompany = lpecCompany->lpszCompanyname;
	} else {
		cbUserId = lpecCompany->sCompanyId.cb;
		lpUserId = (LPENTRYID)lpecCompany->sCompanyId.lpb;
		sVars.ulType = USEROBJECT_TYPE_COMPANY;
		sVars.strUserName = lpecCompany->lpszCompanyname;
		sVars.strFullName = lpecCompany->lpszCompanyname;
		sVars.strCompany = lpecCompany->lpszCompanyname;
	}

	hr = lpServiceAdmin->GetQuota(cbUserId, lpUserId, &lpecQuota);
	if (hr != hrSuccess)
		goto exit;

	sVars.ulStatus = lpecQuotaStatus->quotaStatus;
	sVars.strStoreSize = str_storage(lpecQuotaStatus->llStoreSize);
	sVars.strWarnSize = str_storage(lpecQuota->llWarnSize);
	sVars.strSoftSize = str_storage(lpecQuota->llSoftSize);
	sVars.strHardSize = str_storage(lpecQuota->llHardSize);

	hr = lpServiceAdmin->GetQuotaRecipients(cbUserId, lpUserId, &cToUsers, &lpToUsers);
	if (hr != hrSuccess)
		goto exit;

	hr = CreateRecipientList(cToUsers, lpToUsers, &lpAddrList);
	if (hr != hrSuccess)
		goto exit;

	/* Go through all stores to deliver the mail to all recipients.
	 * Note that we will parse the template for each recipient seperately,
	 * this is done to support better language support later on where each user
	 * will get a notification mail in his prefered language. */
	for (ULONG i = 0; i < cToUsers; i++) {
		hr = lpIEMS->CreateStoreEntryID("", lpToUsers[i].lpszUsername, OPENSTORE_HOME_LOGON,
										&cbUserStoreEntryID, &lpUserStoreEntryID);
		if (hr != hrSuccess)
			goto exit;

		hr = m_lpMAPIAdminSession->OpenMsgStore(0, cbUserStoreEntryID, lpUserStoreEntryID,
												NULL, MAPI_BEST_ACCESS, &lpMDB);
		if (hr != hrSuccess)
			goto exit;

		/* We should only check the interval property on the first store,
		 * since that is the company who is over quota.
		 * We only check for company since that wasn't in a table.
		 */
		if (i == 0 && lpecUser == NULL) {
			hr = HrGetOneProp(lpMDB, PR_EC_QUOTA_MAIL_TIME, &lpsPropTime);
			if (hr == hrSuccess)
				hr = CheckQuotaInterval(lpsPropTime, &bSendmail);
			else
				bSendmail = true;
			if (!bSendmail) {
				m_lpThreadMonitor->lpLogger->Log(EC_LOGLEVEL_INFO, "Not sending message since the user has already received a warning in the past time interval");
				goto exit;
            }
		}

		/* When something goes wrong, skip delivery to current user and proceed to the next user
		 * Company quota's shouldn't deliver to the first entry since that is the public store. */
		if (i != 0 || sVars.ulType != USEROBJECT_TYPE_COMPANY)
			CreateQuotaWarningMail(&sVars, lpMDB, &lpToUsers[i], lpecFromUser, lpAddrList);

		if (i == 0) {
			hr = UpdateQuotaTimestamp(lpMDB);
			if (hr != hrSuccess)
				m_ulFailed++;
		}

		if (lpUserStoreEntryID) {
			MAPIFreeBuffer(lpUserStoreEntryID);
			lpUserStoreEntryID = NULL;
		}

		if (lpMDB) {
			lpMDB->Release();
			lpMDB = NULL;
		}
	}

exit:
	if (lpsPropTime)
		MAPIFreeBuffer(lpsPropTime);

	if (lpecFromUser)
		MAPIFreeBuffer(lpecFromUser);

	if (lpToUsers)
		MAPIFreeBuffer(lpToUsers);

	if (lpecQuota)
		MAPIFreeBuffer(lpecQuota);

	if (lpAddrList)
		FreePadrlist(lpAddrList);

	if (lpUserStoreEntryID)
		MAPIFreeBuffer(lpUserStoreEntryID);

	if (lpMDB)
		lpMDB->Release();

	if (lpIEMS)
		lpIEMS->Release();

	if (lpServiceAdmin)
		lpServiceAdmin->Release();

	if(lpsObject)
		MAPIFreeBuffer(lpsObject);

	return hr;
}
