/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "CalDavProto.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//Property mappings.
struct sMymap {
	unsigned int ulPropTag;
	char *name;
} sPropMap[] = {
	{ PR_LOCAL_COMMIT_TIME_MAX, "getctag" },
	{ PR_LAST_MODIFICATION_TIME, "getetag" },
	{ PR_DISPLAY_NAME, "displayname" },
	{ PR_CONTAINER_CLASS, "resourcetype" },
	{ PR_DISPLAY_NAME, "owner" },
	{ PR_ENTRYID, "calendar-data" },
	{ PR_COMMENT, "calendar-description" }
};

// Retrieves the Mapi Property from caldav property
ULONG GetPropmap(char *name)
{
	for(unsigned int i = 0; i < sizeof(sPropMap)/sizeof(sPropMap[0]); i++)
	{
		if (strcmp(name, sPropMap[i].name) == 0)
			return sPropMap[i].ulPropTag;
	}

	return 0;
}

CalDAV::CalDAV(Http *lpRequest, IMAPISession *lpSession, ECLogger *lpLogger, std::string strCharset ) : WebDav(lpRequest, lpSession, lpLogger)
{
	m_lpUsrFld = NULL;
	m_lpRootFld = NULL;
	m_lpDefStore = NULL;
	m_lpAddrBook = NULL;
	m_lpSharedStore = NULL;
	m_lpPubStore = NULL;
	m_lpActiveStore = NULL;
	m_lpImailUser = NULL;
	m_lpNamedProps = NULL;
	m_blFolderAccess = true;
	m_ulFolderFlag = 0;
	m_strCharset = strCharset;

	m_lpConvResponse = new ECIConv("utf-8", "windows-1252");	
	m_lpConvRequest = new ECIConv("windows-1252", m_strCharset);
}

CalDAV::~CalDAV()
{
	if (m_lpNamedProps)
		MAPIFreeBuffer(m_lpNamedProps);

	if (m_lpConvRequest)
		delete m_lpConvRequest;

	if (m_lpConvResponse)
		delete m_lpConvResponse;

	if (m_lpImailUser)
		m_lpImailUser->Release();

	if (m_lpRootFld)
		m_lpRootFld->Release();

	if (m_lpUsrFld)
		m_lpUsrFld->Release();

	if (m_lpAddrBook)
		m_lpAddrBook->Release();

	if (m_lpDefStore)
		m_lpDefStore->Release();

	if (m_lpSharedStore)
		m_lpSharedStore->Release();

	if(m_lpPubStore)
		m_lpPubStore->Release();
}

HRESULT CalDAV::HrHandleCommand(std::string strMethod, std::string strServerTZ)
{
	HRESULT hr = hrSuccess;

	hr = HrGetFolder();
	if (hr != hrSuccess && strMethod.compare("PROPPATCH") && strMethod.compare("MKCALENDAR")) // ignore folder for MKCalendar.
	{
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Folder named '%s' not found, error code : 0x%08X", m_strFldName.c_str(), hr);
		goto exit;
	}

	if(!m_lpUsrFld && strMethod.compare("PROPPATCH") && strMethod.compare("PROPFIND") && strMethod.compare ("MKCALENDAR"))
	{
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Folder named '%s' not found, error code : 0x%08X", m_strFldName.c_str(), hr);
		goto exit;
	}

	if(!strMethod.compare("PROPFIND"))
	{
		hr = HrPropfind();
	}
	else if(!strMethod.compare("REPORT"))
	{
		hr = HrReport();
	}
	else if(!strMethod.compare("PUT"))
	{
		hr = HrPut(strServerTZ);
	}
	else if(!strMethod.compare("DELETE"))
	{
		hr = HrDelete();
	}
	else if(!strMethod.compare("MKCALENDAR"))
	{
		hr = HrMkCalendar();
	}
	else if(!strMethod.compare("PROPPATCH"))
	{
		hr = HrPropPatch();
	}
	else if (!strMethod.compare("POST"))
	{
		hr = HrPost(strServerTZ);
	}
	else
	{
		m_lpRequest->HrResponseHeader(501, "Not Implemented");
	}

exit:
	
	return hr;
}

//Open Folders required for the Request
HRESULT CalDAV::HrGetFolder()
{
	HRESULT hr = hrSuccess;
	std::string strMethod;
	std::string strFldName;	
	std::string strUser;
	std::string strFldOwner;
	std::string strPublic;
	size_t ulFound = 0;
	ULONG ulDepth;
	LPSPropValue lpDefaultProp = NULL;
	LPSPropValue lpFldProp = NULL;
	IMAPIFolder *lpRootCont = NULL;

	m_lpRequest->HrGetMethod(&strMethod);
	m_lpRequest->HrGetFldPath(&strFldName);
	m_lpRequest->HrGetDepth(&ulDepth);
	m_lpRequest->HrGetUser(&strUser);
	m_lpRequest->HrGetFldOwner(&strFldOwner);

	strFldName = U2W(strFldName);

	hr = m_lpSession->OpenAddressBook(0, NULL, 0, &m_lpAddrBook);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening addressbook, error code : 0x%08X", hr);
		goto exit;
	}

	hr = HrOpenDefaultStore(m_lpSession, &m_lpDefStore);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening default store of user %s, error code : 0x%08X", strUser.c_str(), hr);
		goto exit;
	}
	

	m_lpActiveStore = m_lpDefStore;

	if(m_lpRequest->IsPublic())
	{
		hr = HrOpenECPublicStore(m_lpSession, &m_lpPubStore);
		if (hr != hrSuccess)
			goto exit;
				
		m_lpActiveStore = m_lpPubStore;
		hr = OpenSubFolder(m_lpActiveStore, NULL, '/', m_lpLogger, true, false, &m_lpRootFld);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening IPM_SUBTREE Folder of user %s, error code : 0x%08X", strUser.c_str(), hr);
			goto exit;
		}
	}

	// Retrive named properties
	hr = HrLookupNames(m_lpActiveStore, &m_lpNamedProps);
	if (hr != hrSuccess)
		goto exit;

	//PUT & DELETE method have calendar entry uid in the request so remove that
	if(!strMethod.compare("PUT") || !strMethod.compare("DELETE") || !strMethod.compare("GET"))
	{
		ulFound = strFldName.rfind("/");
		if(ulFound == string::npos)
			strFldName = strFldName;
		else
		{
			m_strEntryRef = StripGuid(strFldName);
			strFldName = strFldName.substr( 0, ulFound);
		}

		ulFound = strFldName.rfind(".ics");
		if(ulFound != string::npos)
		{
			m_strEntryRef = StripGuid(strFldName);
			strFldName = "";
		}
	}

	m_strFldName = strFldName;
	if(!strFldOwner.empty() && strFldOwner != strUser && !m_lpRequest->IsPublic())
	{
		hr = HrGetSharedFolder(m_lpSession, m_lpDefStore, strFldOwner, m_strFldName, m_blIsCLMAC, false, m_lpLogger, &m_ulFolderFlag, &m_lpSharedStore, &m_lpUsrFld);
		if(m_lpSharedStore)
		{
			m_lpActiveStore = m_lpSharedStore;
			hr = OpenSubFolder(m_lpActiveStore, NULL, '/', m_lpLogger, false, false, &m_lpRootFld);
			if(hr != hrSuccess)
			{
				hr = hrSuccess;
				m_lpRootFld = NULL;
			}
		}

		if(hr == E_ACCESSDENIED)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Not enough permissions to user %s to access Folder for user %s", strUser.c_str(), strFldOwner.c_str());
			goto exit;
		}
		else if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while accessing shared folder by user %s, error code : 0x%08X", strUser.c_str(), hr);
			goto exit;
		}
		
	}	
	else if(m_blIsCLMAC ) //For Ical Client The folders are opened according to Folder_ID.
	{
		if(!strFldName.empty())
		{
			hr = HrFindFolder(m_lpSession, m_lpActiveStore, m_lpNamedProps, m_lpLogger, m_lpRequest->IsPublic(), strFldName, &m_lpUsrFld);
			if(hr != hrSuccess)
			{
				hr = HrOpenUserFld(m_lpSession, m_lpActiveStore, strFldName, false, m_lpRequest->IsPublic(), m_lpLogger, &m_lpUsrFld);
				if(hr == hrSuccess) // do not exit , need to set the m_lpRootfolder if not set.
					m_ulFolderFlag = SINGLE_FOLDER;
			}
		}
	}
	//For other Clients the Folder name is used to refer to folders.
	else
	{
		if (strMethod.compare("MKCALENDAR"))
		{
			hr = HrOpenUserFld(m_lpSession, m_lpActiveStore, strFldName, false, m_lpRequest->IsPublic(), m_lpLogger, &m_lpUsrFld);
			if(hr != hrSuccess)
			{
				m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening folder of user %s, error code : 0x%08X", strUser.c_str(), hr);
				goto exit;
			}
		}
	}
	
	hr = HrGetUserInfo(m_lpSession, m_lpDefStore, NULL, NULL, &m_lpImailUser);
	if(hr != hrSuccess)
		goto exit;

	if(!m_lpRootFld)
	{
		//Open IPM_SUBTREE
		hr = OpenSubFolder(m_lpDefStore, NULL, '/', m_lpLogger, false, false, &m_lpRootFld);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening IPM_SUBTREE Folder of user %s, error code : 0x%08X", strUser.c_str(), hr);
			goto exit;
		}
	}
	//Set Limited Access to Default Calendar Folder
	//Block Rename & Deletion of this folder.
	if( (!strMethod.compare("PROPPATCH") || !strMethod.compare("DELETE")) && !m_lpRequest->IsPublic())
	{
		int ulCmp = 0;
		ULONG ulType = 0;

		hr = m_lpActiveStore->OpenEntry(0, NULL, NULL, 0, &ulType, (LPUNKNOWN*)&lpRootCont);
		if (hr != hrSuccess || ulType != MAPI_FOLDER)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening root Container of user %s, error code: (0x%08X)", strUser.c_str(), hr);
			goto exit;
		}

		hr = HrGetOneProp(lpRootCont, PR_IPM_APPOINTMENT_ENTRYID, &lpDefaultProp);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving Entry id of Default calendar for user %s, error code: 0x%08X", strUser.c_str(), hr);
			goto exit;
		}

		if(m_lpUsrFld)
			hr = HrGetOneProp(m_lpUsrFld, PR_ENTRYID, &lpFldProp);
		else
		{
			m_blFolderAccess = false;
			goto exit;
		}

		if(hr != hrSuccess)
			goto exit;
		
		
		ulCmp = memcmp(lpDefaultProp->Value.bin.lpb, lpFldProp->Value.bin.lpb, lpFldProp->Value.bin.cb);
		if(!ulCmp)
			m_blFolderAccess = false;
		
		MAPIFreeBuffer(lpFldProp);
		lpFldProp = NULL;

		hr = HrGetOneProp(m_lpUsrFld, PR_SUBFOLDERS, &lpFldProp);
		if(hr != hrSuccess)
			goto exit;

		if(lpFldProp->Value.b == (unsigned short)true && !strMethod.compare("DELETE"))
			m_blFolderAccess = false;
	}

exit:
	if(lpFldProp)
		MAPIFreeBuffer(lpFldProp);

	if(lpDefaultProp)
		MAPIFreeBuffer(lpDefaultProp);

	if(lpRootCont)
		lpRootCont->Release();

	return hr;
}

HRESULT CalDAV::HrPropfind()
{
	HRESULT hr = hrSuccess;

	hr = WebDav::HrPropfind();
	if (hr != hrSuccess)
		goto exit;

exit:
	return hr;
}

// Retrieve data from the server.
HRESULT CalDAV::HrHandlePropfind(WEBDAVREQSTPROPS *lpsDavProp, WEBDAVMULTISTATUS *lpsDavMulStatus)
{
	HRESULT hr = hrSuccess;
	std::string strFldPath;
	ULONG ulDepth = 0;
	
	m_lpRequest->HrGetDepth(&ulDepth);
	m_lpRequest->HrGetFldPath(&strFldPath);

	if(ulDepth == 0 || ( strFldPath == "Inbox" ))
	{
		//Retrieve Properties
		hr = HrHandlePropfindRoot(lpsDavProp, lpsDavMulStatus);
	}
	else if( (strFldPath.empty() || !strFldPath.compare("public") ||( m_ulFolderFlag & SINGLE_FOLDER)) && ulDepth == 1 && m_blIsCLMAC)
	{
		hr = HrListCalendar(m_lpSession, &(lpsDavProp->sProp), m_lpLogger, lpsDavMulStatus);
	}
	else
	{
		//Retrieve the Calendar entries list
		hr = HrListCalEntries(lpsDavProp, lpsDavMulStatus);
	}

	return hr;
}

HRESULT CalDAV::HrHandlePropfindRoot(WEBDAVREQSTPROPS *sDavReqstProps, WEBDAVMULTISTATUS *lpsDavMulStatus)
{
	HRESULT hr = hrSuccess;
	std::string strFldPath;
	std::string strUser;
	std::string strFldOwner;
	std::string strTrail;
	std::string strEmail;
	std::string strCalendarHome;
	std::string strReqPath;
	std::string strUserFullName;
	std::list<WEBDAVPROPERTY>::iterator iter;
	WEBDAVVALUE sWebVal;
	WEBDAVPROPSTAT sPropStat;
	WEBDAVRESPONSE sDavResp;
	WEBDAVPROP sDavPropNF;
	WEBDAVPROP *sDavProp;
	IMAPIProp *lpMapiProp = NULL;
	LPSPropTagArray lpPropTagArr = NULL;
	LPSPropValue lpSpropVal = NULL;
	ULONG cbsize = 0;
	ULONG ulPropCount = 0;
	size_t ulFound = 0;

	sDavProp = &(sDavReqstProps->sProp);
	ulPropCount = sDavProp->lstProps.size();
	sDavPropNF.sPropName = sDavProp->sPropName;

	m_lpRequest->HrGetUser(&strUser);
	m_lpRequest->HrGetFldOwner(&strFldOwner);
	m_lpRequest->HrGetFldPath(&strFldPath);
	m_lpRequest->HrGetReqPath(&strReqPath);

	ulFound = strReqPath.rfind("public");
	if(ulFound == string::npos || ulFound == 0)
		strCalendarHome = "/caldav/" + W2U(strFldOwner.empty()? strUser : strFldOwner) + "/" + strFldPath;
	else
		strCalendarHome = "/caldav/public/" + strFldPath;
	
	if (m_lpRequest->IsPublic() && strFldPath.empty())
		lpMapiProp = m_lpDefStore;
	else if(m_blIsCLMAC && (strFldPath.empty() || strFldPath == "Inbox" || strFldPath == "Outbox") )
		lpMapiProp = m_lpActiveStore;
	else if(m_lpUsrFld)
		lpMapiProp = m_lpUsrFld;
	else
	{
		m_lpLogger->Log( EC_LOGLEVEL_ERROR,"Error Folder named '%s' of user '%s' not found, error code 0x%08X", strFldPath.c_str(), strUser.c_str(), hr);
		goto exit;
	}
	
	hr = HrGetUserInfo(m_lpSession, m_lpDefStore, &strEmail, &strUserFullName, NULL);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving email address of user %s, error code : 0x%08X", strUser.c_str(), hr);
		goto exit;
	}

	// number of properties requested by client.
	cbsize = ulPropCount;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cbsize), (void **) &lpPropTagArr);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot allocate memory");
		goto exit;
	}

	lpPropTagArr->cValues = cbsize;

	// Get corresponding mapi properties.
	iter = sDavProp->lstProps.begin();
	for(int i = 0; iter != sDavProp->lstProps.end(); iter++, i++)
	{
		lpPropTagArr->aulPropTag[i] = GetPropmap((char*)iter->sPropName.strPropname.c_str());
	}

	hr = lpMapiProp->GetProps((LPSPropTagArray)lpPropTagArr, 0, &cbsize, &lpSpropVal);
	if (hr == MAPI_W_ERRORS_RETURNED)
		hr = hrSuccess;		// ignoring some items
	else if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error in GetProps for user %s, error code : 0x%08X", strUser.c_str(), hr);
		goto exit;
	}

	// Add data for each requested property.
	for(ULONG i = 0; i < cbsize; i++)
	{
		WEBDAVVALUE sDavVal;
		WEBDAVPROPERTY sDavProperty;
		sDavProperty = sDavProp->lstProps.front();

		sDavVal.sPropName = sDavProperty.sPropName;
		sDavVal.strValue = SPropValToString(&lpSpropVal[i]);

		if(sDavVal.sPropName.strPropname == "resourcetype" && sDavVal.strValue == "IPF.Appointment")
		{
			HrSetDavPropName(&(sWebVal.sPropName), "collection", "");
			sWebVal.strValue = "";
			sDavProperty.lstValues.push_back(sWebVal);

			HrSetDavPropName(&(sWebVal.sPropName), "calendar", CALDAVNS);
			sWebVal.strValue = "";
			sDavProperty.lstValues.push_back(sWebVal);
		}
		else if (sDavVal.sPropName.strPropname == "resourcetype" && strFldPath == "Inbox")
		{
			HrSetDavPropName(&(sWebVal.sPropName), "collection", "");
			sWebVal.strValue = "";
			sDavProperty.lstValues.push_back(sWebVal);

			HrSetDavPropName(&(sWebVal.sPropName), "schedule-inbox", CALDAVNS);
			sWebVal.strValue = "";
			sDavProperty.lstValues.push_back(sWebVal);
		}
		else if (sDavVal.sPropName.strPropname == "resourcetype" && strFldPath == "Outbox")
		{
			HrSetDavPropName(&(sWebVal.sPropName), "collection", "");
			sWebVal.strValue = "";
			sDavProperty.lstValues.push_back(sWebVal);

			HrSetDavPropName(&(sWebVal.sPropName), "schedule-outbox", CALDAVNS);
			sWebVal.strValue = "";
			sDavProperty.lstValues.push_back(sWebVal);
		}	
		else if(sDavVal.sPropName.strPropname == "owner")
		{
			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			sWebVal.strValue = strCalendarHome;
			sDavProperty.lstValues.push_back(sWebVal);
		}
		else if(sDavVal.sPropName.strPropname == "getctag" || sDavVal.sPropName.strPropname == "getetag")
		{

#ifdef DEBUG
			sDavProperty.strValue = stringify(rand(), false) + sDavVal.strValue;
			// set random to force a reload of this calendar
#else
			sDavProperty.strValue = sDavVal.strValue;
#endif
		}
		else if(sDavVal.sPropName.strPropname == "displayname" )
		{
			// displayname property is used by return both folder name as well as the user name
			// only mac client makes the request for user name
			// the only differenece between PROPFIND requesting for user name
			// and PROPFIND requesting for folder name is the no of properites.
			if (m_blIsCLMAC && (ulPropCount == 10 || ulPropCount == 7))
				sDavProperty.strValue = W2U(strUserFullName);
			else
				sDavProperty.strValue = W2U(sDavVal.strValue);

		}
		else if(sDavVal.sPropName.strPropname == "calendar-home-set" )
		{
			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			sWebVal.strValue = strCalendarHome;
			sDavProperty.lstValues.push_back(sWebVal);
		}	
		else if(sDavVal.sPropName.strPropname == "calendar-user-address-set" )
		{
			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			sWebVal.strValue = "mailto:" + strEmail;
			sDavProperty.lstValues.push_back(sWebVal);
			sWebVal.strValue = strCalendarHome;
			sDavProperty.lstValues.push_back(sWebVal);
		}
		else if(sDavVal.sPropName.strPropname == "acl" || sDavVal.sPropName.strPropname == "current-user-privilege-set")
		{
			sDavProperty.sPropName.strNS = "";
			HrBuildACL(&sDavProperty);
		}
		else if(sDavVal.sPropName.strPropname == "calendar-description")
		{
			sDavProperty.strValue = W2U(sDavVal.strValue);
		}
		else if (sDavVal.sPropName.strPropname == "schedule-inbox-URL")
		{
			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			sWebVal.strValue = ("/caldav/" + W2U(strFldOwner.empty()? strUser : strFldOwner) + "/Inbox");
			sDavProperty.lstValues.push_back(sWebVal);
		}
		else if (sDavVal.sPropName.strPropname == "schedule-outbox-URL")
		{
			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			sWebVal.strValue = ("/caldav/" + W2U(strFldOwner.empty()? strUser : strFldOwner) + "/Outbox");
			sDavProperty.lstValues.push_back(sWebVal);
		}
		else if (sDavVal.sPropName.strPropname == "supported-calendar-component-set")
		{
			HrSetDavPropName(&(sWebVal.sPropName), "comp", "name","VEVENT", CALDAVNS);
			sDavProperty.lstValues.push_back(sWebVal);
		}
		else
		{
			sDavProp->lstProps.pop_front();
			sDavPropNF.lstProps.push_back(sDavProperty);
			continue;
		}
		sDavProp->lstProps.pop_front();
		sDavProp->lstProps.push_back(sDavProperty);
	}

	sPropStat.sProp = *sDavProp;

	HrSetDavPropName(&(sPropStat.sPropName), "propstat", sDavProp->sPropName.strNS);
	HrSetDavPropName(&(sPropStat.sStatus.sPropName), "status", sDavProp->sPropName.strNS);
	sPropStat.sStatus.strValue = "HTTP/1.1 200 OK";

	HrSetDavPropName(&(sDavResp.sPropName), "response", sDavProp->sPropName.strNS);

	hr = m_lpRequest->HrGetReqPath(&strFldPath);
	if (hr != hrSuccess)
		strFldPath = "";

	HrSetDavPropName(&(sDavResp.sHRef.sPropName), "href", sDavProp->sPropName.strNS);
	sDavResp.sHRef.strValue = strFldPath;

	if(!sDavProp->lstProps.empty())
		sDavResp.lstsPropStat.push_back(sPropStat);

	if(!sDavPropNF.lstProps.empty())
	{
		sPropStat.sProp = sDavPropNF;
		sPropStat.sStatus.strValue = "HTTP/1.1 404 Not Found";
		sDavResp.lstsPropStat.push_back(sPropStat);
	}
	HrSetDavPropName(&(lpsDavMulStatus->sPropName), "multistatus", sDavProp->sPropName.strNS);
	lpsDavMulStatus->lstResp.push_back(sDavResp);

exit:

	if(lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);

	if(lpSpropVal)
		MAPIFreeBuffer(lpSpropVal);

	return hr;
}

//REPORT has two parts CALENDAR-QUERY and MULTIGET.
HRESULT CalDAV::HrReport()
{
	HRESULT hr = hrSuccess;

	hr = WebDav::HrReport();

	return hr;
}

//Retrieve data from REPORT Calendar-Query Request.
// FIXME what is the difference between this function and HrHandleReport except for the list of WHICH items to output
HRESULT CalDAV::HrListCalEntries(WEBDAVREQSTPROPS *lpsWebRCalQry, WEBDAVMULTISTATUS *lpsWebMStatus)
{
	HRESULT hr = hrSuccess;
	std::string strConvVal;
	std::string strReqPath;
	std::string strTrail;
	IMAPIFolder *lpUsrFld = NULL;
	IMAPITable *lpTable = NULL;
	LPSRowSet lpRowSet = NULL;
	LPSPropTagArray lpPropTagArr = NULL;
	LPSPropTagArray lpPropTags = NULL;
	MapiToICal *lpMtIcal = NULL;
	ULONG cbsize = 0;
	ULONG ulTagGOID = 0;
	ULONG ulTagTsRef = 0;
	ULONG ulTagPrivate = 0;
	std::list<WEBDAVPROPERTY>::iterator iter;
	WEBDAVPROP sDavProp;
	WEBDAVPROPERTY sDavProperty;
	bool blCensorPrivate = false;
	bool blCensor = false;

	ulTagGOID = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY);
	ulTagTsRef = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTTSREF], PT_STRING8);
	ulTagPrivate = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN);

	if ((m_ulFolderFlag & SHARED_FOLDER) && !HasDelegatePerm(m_lpDefStore, m_lpActiveStore))
		blCensorPrivate = true;
	
	HrSetDavPropName(&(lpsWebMStatus->sPropName), "multistatus", lpsWebRCalQry->sPropName.strNS);
	sDavProp = lpsWebRCalQry->sProp;

	//no support for VTODO events.
	if(!lpsWebRCalQry->sFilter.lstFilters.empty() && lpsWebRCalQry->sFilter.lstFilters.back() == "VTODO")
		goto exit;
	if(m_lpUsrFld)
		lpUsrFld = m_lpUsrFld;
	else
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error user's folder not found, cannot list calendar entries");
		goto exit;
	}

	hr = lpUsrFld->GetContentsTable(0, &lpTable);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error in GetContentsTable, error code : 0x%08X", hr);
		goto exit;
	}

	// +3 to add GlobalObjid, dispidApptTsRef and PR_ENTRYID in SetColumns along with requested data.
	cbsize = (ULONG)sDavProp.lstProps.size() + 4;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cbsize), (void **)&lpPropTagArr);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot allocate memory");
		goto exit;
	}

	lpPropTagArr->cValues = cbsize;

	iter = sDavProp.lstProps.begin();

	// FIXME: get named properties in 1 call, not multiple


	lpPropTagArr->aulPropTag[0] = ulTagTsRef;
	lpPropTagArr->aulPropTag[1] = ulTagGOID;
	lpPropTagArr->aulPropTag[2] = PR_ENTRYID;
	lpPropTagArr->aulPropTag[3] = ulTagPrivate;

	//mapi property mapping for requested properties.
	//FIXME what if the property mapping is not found.
	for (int i = 4; iter != sDavProp.lstProps.end(); iter++, i++)
	{
		sDavProperty = *iter;
		lpPropTagArr->aulPropTag[i] = GetPropmap((char*)sDavProperty.sPropName.strPropname.c_str());

	}

	hr = lpTable->SetColumns((LPSPropTagArray)lpPropTagArr, 0);
	if(hr != hrSuccess)
		goto exit;
	
	CreateMapiToICal(m_lpAddrBook, "utf-8", &lpMtIcal);
	if (!lpMtIcal)
	{
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	while(1)
	{
		std::list<WEBDAVPROPERTY>::iterator iterProperty;

		hr = lpTable->QueryRows(50, 0, &lpRowSet);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving rows of table");
			goto exit;
		}

		if(lpRowSet->cRows == 0)
			break;

		//add data from each requested property.
		for (ULONG ulRowCntr = 0; ulRowCntr < lpRowSet->cRows; ulRowCntr++)
		{
			WEBDAVRESPONSE sWebResp;
			WEBDAVPROPSTAT sWebPStat;
			WEBDAVPROP sWebProp;
			WEBDAVPROPERTY sWebProperyInput;
			blCensor = blCensorPrivate; // reset Censor flag for next message.

			HrSetDavPropName(&(sWebResp.sPropName), "response", lpsWebRCalQry->sPropName.strNS);
			HrSetDavPropName(&(sWebResp.sHRef.sPropName), "href", lpsWebRCalQry->sPropName.strNS);

			hr = m_lpRequest->HrGetReqPath(&strReqPath);
			if(hr != hrSuccess)
				continue;

			strTrail = strReqPath.at(strReqPath.length() - 1);
			if(strTrail.compare("/"))
				strReqPath = strReqPath + "/";

			//For entries created in Evolution client use Timestamp as URL ref
			if(lpRowSet->aRow[ulRowCntr].lpProps[0].ulPropTag == ulTagTsRef && m_blIsCLEVO)
				strConvVal = lpRowSet->aRow[ulRowCntr].lpProps[0].Value.lpszA;
			else if (lpRowSet->aRow[ulRowCntr].lpProps[1].ulPropTag == ulTagGOID)
				strConvVal = SPropValToString((SPropValue *)&(lpRowSet->aRow[ulRowCntr].lpProps[1]));
			else
				strConvVal.clear();

			// On some items, outlook never created the uid, so we need to create one for ical
			if (strConvVal.empty())
			{
				hr = CreateAndGetGuid(lpRowSet->aRow[ulRowCntr].lpProps[2].Value.bin, ulTagGOID, &strConvVal);
				if(hr == E_ACCESSDENIED)
				{
					// use PR_ENTRYID since we couldn't create a new guid for the item
					strConvVal = Util::bin2hex(lpRowSet->aRow[ulRowCntr].lpProps[2].Value.bin.cb, lpRowSet->aRow[ulRowCntr].lpProps[2].Value.bin.lpb);
					hr = hrSuccess;
				}
				else if(hr != hrSuccess)
					continue;
			}

			strConvVal = strConvVal + ".ics";
			strConvVal = strReqPath + strConvVal;

			sWebResp.sHRef.strValue = strConvVal;

			HrSetDavPropName(&(sWebPStat.sPropName), "propstat", lpsWebRCalQry->sPropName.strNS);
			HrSetDavPropName(&(sWebProp.sPropName), "prop", lpsWebRCalQry->sPropName.strNS);

			iterProperty = lpsWebRCalQry->sProp.lstProps.begin();
			sWebProp.lstProps.clear();

			if (lpRowSet->aRow[ulRowCntr].lpProps[3].ulPropTag != ulTagPrivate || !lpRowSet->aRow[ulRowCntr].lpProps[3].Value.b)
				blCensor = false;

			hr = HrConvertRowItemToICal(lpsWebRCalQry->sProp, &lpRowSet->aRow[ulRowCntr], lpMtIcal, blCensor, &sWebProp);
			
			sWebPStat.sProp = sWebProp;
			HrSetDavPropName(&(sWebPStat.sStatus.sPropName), "status", lpsWebRCalQry->sProp.sPropName.strNS);
			if(hr != hrSuccess)
				sWebPStat.sStatus.strValue.assign("HTTP/1.1 500 Internal Server Error");
			else
				sWebPStat.sStatus.strValue.assign("HTTP/1.1 200 OK");

			sWebResp.lstsPropStat.push_back(sWebPStat);
			lpsWebMStatus->lstResp.push_back(sWebResp);
		}

		if(lpRowSet)
		{
			FreeProws(lpRowSet);
			lpRowSet = NULL;
		}
	}

exit:

	if(lpMtIcal)
		delete lpMtIcal;

	if(lpPropTags)
		MAPIFreeBuffer(lpPropTags);

	if(lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);

	if(lpTable)
		lpTable->Release();

	if(lpRowSet)
		FreeProws(lpRowSet);
	
	return hr;
}

HRESULT CalDAV::HrHandleReport(WEBDAVRPTMGET *sWebRMGet, WEBDAVMULTISTATUS *sWebMStatus)
{
	HRESULT hr = hrSuccess;
	std::string strConvVal;
	std::string strReqPath;
	std::string strNsDef;
	IMAPIFolder *lpUsrFld = NULL;
	IMAPITable *lpTable = NULL;
	SRestriction * lpsRoot = NULL;
	LPSPropTagArray lpPropTags = NULL;
	LPSPropTagArray lpPropTagArr = NULL;
	MapiToICal *lpMtIcal = NULL;
	ULONG cbsize = 0;
	ULONG ulPropTag = 0;
	ULONG ulTagPrivate = 0;
	std::list<WEBDAVPROPERTY>::iterator iter;
	WEBDAVPROP sDavProp;
	WEBDAVPROPERTY sDavProperty;	
	bool blCensorPrivate = false;
	bool blCensor = false;

	if ((m_ulFolderFlag & SHARED_FOLDER) && !HasDelegatePerm(m_lpDefStore, m_lpActiveStore))
		blCensorPrivate = true;
	
	//store default Namespace of request
	strNsDef = sWebRMGet->sProp.sPropName.strNS;

	lpUsrFld = m_lpUsrFld;

	hr = lpUsrFld->GetContentsTable(0, &lpTable);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error in GetContentsTable, error code : 0x%08X", hr);
		goto exit;
	}

	sDavProp = sWebRMGet->sProp;

	//Add GUID in Setcolumns.
	cbsize = (ULONG)sDavProp.lstProps.size() + 2;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cbsize), (void **)&lpPropTagArr);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error allocating memory, error code : 0x%08X", hr);
		goto exit;
	}
	lpPropTagArr->cValues = cbsize;

	ulTagPrivate = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN);
	ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY);
	iter = sDavProp.lstProps.begin();
	lpPropTagArr->aulPropTag[0] = ulPropTag;

	lpPropTagArr->aulPropTag[1] = ulTagPrivate;

	for(int i = 2; iter != sDavProp.lstProps.end(); iter++, i++)
	{
		sDavProperty = *iter;
		lpPropTagArr->aulPropTag[i] = GetPropmap((char*)sDavProperty.sPropName.strPropname.c_str());
	}

	hr = lpTable->SetColumns((LPSPropTagArray)lpPropTagArr, 0);
	if(hr != hrSuccess)
		goto exit;

	//{(LPGUID)&PSETID_Meeting, MNID_ID, {dispidGlobalObjectID}};
	CREATE_RESTRICTION(lpsRoot);
	CREATE_RES_OR(lpsRoot, lpsRoot, 3); // max: or guid or raw or entryid
	lpsRoot->res.resOr.lpRes[0].rt = RES_PROPERTY;
	lpsRoot->res.resOr.lpRes[0].res.resProperty.relop = RELOP_EQ;
	lpsRoot->res.resOr.lpRes[0].res.resProperty.ulPropTag = ulPropTag; // (ical) guid
	lpsRoot->res.resOr.lpRes[1].rt = RES_PROPERTY;
	lpsRoot->res.resOr.lpRes[1].res.resProperty.relop = RELOP_EQ;
	lpsRoot->res.resOr.lpRes[1].res.resProperty.ulPropTag = ulPropTag; // raw
	lpsRoot->res.resOr.lpRes[2].rt = RES_PROPERTY;
	lpsRoot->res.resOr.lpRes[2].res.resProperty.relop = RELOP_EQ;
	lpsRoot->res.resOr.lpRes[2].res.resProperty.ulPropTag = PR_ENTRYID; // entryid

	cbsize = (ULONG)sWebRMGet->lstWebVal.size();
	
	CreateMapiToICal(m_lpAddrBook, "utf-8", &lpMtIcal);
	if (!lpMtIcal)
	{
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	for(ULONG i = 0; i < cbsize; i++)
	{
		WEBDAVVALUE sWebDavVal;
		SRowSet *lpValRows = NULL;
		std::string strReqPath;
		std::string strBinGuid;
		std::string strRawGuid;
		SPropValue spvGuid;
		SPropValue spvRaw;
		SPropValue spvEntryId;
		blCensor = blCensorPrivate; // reset bool for next message.

		WEBDAVRESPONSE sWebResponse;
		WEBDAVPROPSTAT sWebPropStat;
		
		HrSetDavPropName(&(sWebResponse.sPropName), "response", strNsDef);
		
		sWebDavVal = sWebRMGet->lstWebVal.front();
		sWebRMGet->lstWebVal.pop_front();

		//assign href;
		m_lpRequest->HrGetReqPath(&strReqPath);
		strReqPath = strReqPath + sWebDavVal.strValue;
		strReqPath = strReqPath + ".ics";
		
		sWebResponse.sHRef = sWebDavVal;
		sWebResponse.sHRef.strValue = W2U(strReqPath);

		if (IsOutlookUid(sWebDavVal.strValue)) {
			Util::hex2bin(sWebDavVal.strValue, &strBinGuid);
			spvGuid.ulPropTag = ulPropTag;
			spvGuid.Value.bin.cb = (ULONG)strBinGuid.size();
			spvGuid.Value.bin.lpb= (LPBYTE)strBinGuid.c_str();

			lpsRoot->res.resOr.cRes = 1;
			lpsRoot->res.resOr.lpRes[0].res.resProperty.lpProp = &spvGuid;
		} else {
			HrMakeBinUidFromICalUid(sWebDavVal.strValue, &strBinGuid);
			spvGuid.ulPropTag = ulPropTag;
			spvGuid.Value.bin.cb = (ULONG)strBinGuid.size();
			spvGuid.Value.bin.lpb= (LPBYTE)strBinGuid.c_str();

			Util::hex2bin(sWebDavVal.strValue, &strRawGuid);
			spvRaw.ulPropTag = ulPropTag; // aka "broken" guids (not conforming the MS specs)
			spvRaw.Value.bin.cb = (ULONG)strRawGuid.size();
			spvRaw.Value.bin.lpb= (LPBYTE)strRawGuid.c_str();

			spvEntryId.ulPropTag = PR_ENTRYID;
			spvEntryId.Value.bin.cb = (ULONG)strRawGuid.size();
			spvEntryId.Value.bin.lpb= (LPBYTE)strRawGuid.c_str();

			lpsRoot->res.resOr.cRes = 3;
			lpsRoot->res.resOr.lpRes[0].res.resProperty.lpProp = &spvGuid;
			lpsRoot->res.resOr.lpRes[1].res.resProperty.lpProp = &spvRaw;
			lpsRoot->res.resOr.lpRes[2].res.resProperty.lpProp = &spvEntryId;
		}
		
		hr = lpTable->FindRow(lpsRoot, BOOKMARK_BEGINNING, 0);
		if (hr != hrSuccess)
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Entry not found (%s), error code : 0x%08X", sWebDavVal.strValue.c_str(), hr);

		// conversion if everthing goes ok, otherwise, add empty item with failed status field
		// we need to return all items requested in the multistatus reply, otherwise sunbird will stop, displaying nothing to the user.

		if (hr == hrSuccess) {
			hr = lpTable->QueryRows(1, TBL_NOADVANCE, &lpValRows); // TODO: what if we get multiple items ?
		
		if(hr != hrSuccess || lpValRows->cRows != 1)
			goto exit;

			if (hr == hrSuccess && (lpValRows->aRow[0].lpProps[1].ulPropTag != ulTagPrivate || lpValRows->aRow[0].lpProps[1].Value.b == FALSE))
				blCensor = false;
		}

		if(hr == hrSuccess)
			hr = HrConvertRowItemToICal(sDavProp, &lpValRows->aRow[0], lpMtIcal, blCensor, &sWebPropStat.sProp);

		if (hr == hrSuccess)
			sWebPropStat.sStatus.strValue.assign("HTTP/1.1 200 OK");
		else
			sWebPropStat.sStatus.strValue.assign("HTTP/1.1 404 Not found "+stringify(hr,true));

		HrSetDavPropName(&sWebPropStat.sPropName, "propstat", strNsDef);
		HrSetDavPropName(&sWebPropStat.sProp.sPropName, "prop", strNsDef);
		HrSetDavPropName(&sWebPropStat.sStatus.sPropName, "status", strNsDef);

		sWebResponse.lstsPropStat.push_back(sWebPropStat);
		sWebMStatus->lstResp.push_back(sWebResponse);

		if(lpValRows)
		{
			FreeProws(lpValRows);
			lpValRows = NULL;
		}
	}

	HrSetDavPropName(&sWebMStatus->sPropName, "multistatus", strNsDef);
	hr = hrSuccess;

exit:
	
	if(lpMtIcal)
		delete lpMtIcal;

	if(lpsRoot)
		FREE_RESTRICTION(lpsRoot);

	if(lpTable)
		lpTable->Release();

	if(lpPropTags)
		MAPIFreeBuffer(lpPropTags);
	
	if(lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);
	
	return hr;
}

HRESULT CalDAV::HrConvertRowItemToICal(WEBDAVPROP &sDavProp, LPSRow lpsRow, MapiToICal *lpMtIcal, bool blCensor, WEBDAVPROP *lpWebProp)
{
	HRESULT hr = hrSuccess;
	std::list<WEBDAVPROPERTY>::iterator iter;
	WEBDAVPROPERTY sProperty;
	LPSPropValue lpFindProp = NULL;
	std::string strIcal;
	LPMESSAGE lpMessage = NULL;
	ULONG ulObjType = 0;
	ULONG ulFlag = 0;

	if (blCensor)
		ulFlag |= M2IC_CENSOR_PRIVATE;

	for (iter = sDavProp.lstProps.begin(); iter != sDavProp.lstProps.end(); iter++)
	{
		sProperty.sPropName = iter->sPropName;
		if(iter->sPropName.strPropname == "calendar-data")
		{
			lpFindProp = PpropFindProp(lpsRow->lpProps, lpsRow->cValues, PR_ENTRYID);
			if (!lpFindProp)
				continue;

			hr = m_lpActiveStore->OpenEntry(lpFindProp->Value.bin.cb, (LPENTRYID)lpFindProp->Value.bin.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN *)&lpMessage);
			if(hr != hrSuccess)
			{
				m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening calendar entry, error code : 0x%08X", hr);
				goto exit;
			}

			hr = lpMtIcal->AddMessage(lpMessage, ulFlag);
			if(hr != hrSuccess)
			{
				m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error converting mapi message to ical, error code : 0x%08X", hr);
				goto exit;
			}

			hr = lpMtIcal->Finalize(0, NULL, &strIcal);
			if(hr != hrSuccess)
			{
				m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error creating ical data, error code : 0x%08X", hr);
				goto exit;
			}

			sProperty.strValue.assign(strIcal);
			lpWebProp->lstProps.push_back(sProperty);

			lpMtIcal->ResetObject();

			if(lpMessage)
			{
				lpMessage->Release();
				lpMessage = NULL;
			}
		}
		else if(iter->sPropName.strPropname == "resourcetype")
		{
			sProperty.strValue = "";
			lpWebProp->lstProps.push_back(sProperty);
		}
		else if(iter->sPropName.strPropname == "getcontenttype")
		{
			sProperty.strValue = "text/calendar";
			lpWebProp->lstProps.push_back(sProperty);
		}
		else if(iter->sPropName.strPropname == "getetag")
		{
			// for etags
			lpFindProp = PpropFindProp(lpsRow->lpProps, lpsRow->cValues, PR_LAST_MODIFICATION_TIME);
			if (lpFindProp)
			{
				sProperty.strValue = SPropValToString(lpFindProp);
				lpWebProp->lstProps.push_back(sProperty);
			}
		}
		else
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error No value found for property %s ", iter->sPropName.strPropname.c_str());
	}

exit:
	if(lpMessage)
		lpMessage->Release();

	return hr;
}

//FIXME Move Code To HrHandleDelete().
//handles Delete folder & delete item
HRESULT CalDAV::HrDelete()
{
	HRESULT hr = hrSuccess;
	std::string strGuid;
	std::string strPath;
	std::string strBinGuid;
	std::string strFldPath;
	size_t ulFound = 0;
	SBinary sbEid = {0,0};
	ULONG ulObjType = 0;
	IMAPIFolder *lpUsrFld = NULL;
	IMAPIFolder *lpRootFld = NULL;
	IMAPIFolder *lpWastBoxFld = NULL;
	LPSPropValue lpProp = NULL;
	IMessage *lpMessage = NULL;
	LPENTRYLIST lpEntryList= NULL;
	bool isFolder = false;
	bool blCensorPrivate = false;

	m_lpRequest->HrGetFldPath(&strPath);
	ulFound = strPath.rfind(".ics");

	if(ulFound == string::npos)
		isFolder = true;

	ulFound = strPath.rfind("/") + 1;
	if(ulFound == string::npos)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Delete Request from client");
		goto exit;
	}

	lpRootFld = m_lpRootFld;
	lpUsrFld = m_lpUsrFld;

	if(!m_blFolderAccess && isFolder)
	{
		hr = MAPI_E_NO_ACCESS;
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Request to delete default calendar denied, error code : 0x%08X", hr);
		goto exit;
	}

	if ((m_ulFolderFlag & SHARED_FOLDER) && !HasDelegatePerm(m_lpDefStore, m_lpActiveStore))
		blCensorPrivate = true;
	
	hr = HrGetOneProp(m_lpDefStore, PR_IPM_WASTEBASKET_ENTRYID, &lpProp);
	if(hr != hrSuccess)
		goto exit;
	
	hr = m_lpDefStore->OpenEntry(lpProp->Value.bin.cb, (LPENTRYID)lpProp->Value.bin.lpb, NULL, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpWastBoxFld);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening \"Deleted items\" folder, error code : 0x%08X", hr);
		goto exit;
	}

	MAPIFreeBuffer(lpProp);
	lpProp= NULL;

	if(isFolder)
	{
		hr = HrGetOneProp(lpUsrFld, PR_ENTRYID, &lpProp);
		if(hr != hrSuccess)
			goto exit;
	}
	else
	{
		//remove trailing chars '.ics' from guid value.
		strGuid = strPath.substr(ulFound, ( strPath.size() - ulFound - 4) );
		if(ulFound == 0)
			strFldPath = ""; //default folder
		else
			strFldPath = strPath.substr(0, ulFound - 1);
	
		//Find Entry With Particular Guid
		hr = HrFindAndGetMessage(strGuid, lpUsrFld,  m_lpNamedProps, &lpMessage);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Entry to be deleted not found : 0x%08X", hr);
			goto exit;
		}

		if (blCensorPrivate && IsPrivate(lpMessage, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN)))
			blCensorPrivate = true;
		else
			blCensorPrivate = false;

		if (blCensorPrivate)
		{
			hr = MAPI_E_NO_ACCESS;
			goto exit;
		}

		hr = HrGetOneProp(lpMessage, PR_ENTRYID, &lpProp);
		if(hr != hrSuccess)
			goto exit;
	}
	
	sbEid = lpProp[0].Value.bin;
	
	//Create Entrylist
	hr = MAPIAllocateBuffer(sizeof(ENTRYLIST), (void**)&lpEntryList);
	if (hr != hrSuccess)
		goto exit;

	lpEntryList->cValues = 1;

	hr = MAPIAllocateMore(sizeof(SBinary), lpEntryList, (void**)&lpEntryList->lpbin);
	if (hr != hrSuccess)
		goto exit;
	
	hr = MAPIAllocateMore(sbEid.cb, lpEntryList, (void**)&lpEntryList->lpbin[0].lpb);
	if (hr != hrSuccess)
		goto exit;

	lpEntryList->lpbin[0].cb = sbEid.cb;
	memcpy((void*)lpEntryList->lpbin[0].lpb, (const void*)sbEid.lpb, sbEid.cb);

	if(isFolder)
	{
		hr = lpRootFld->CopyFolder(sbEid.cb, (LPENTRYID)sbEid.lpb, NULL, lpWastBoxFld, NULL, 0, NULL, MAPI_MOVE);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Deleting Folder error code : 0x%08X", hr);
			goto exit;
		}
	}
	else
	{
		hr = lpUsrFld->CopyMessages(lpEntryList, NULL, lpWastBoxFld, 0, NULL, MAPI_MOVE);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Deleting Entry");
			goto exit;
		}
	}

exit:
	if( hr == MAPI_E_NO_ACCESS )
	{
		m_lpRequest->HrResponseHeader(403, "Forbidden");
		m_lpRequest->HrResponseBody("This folder cannot be deleted");
	}
	else if (hr != hrSuccess)
	{
		m_lpRequest->HrResponseHeader(404, "Not Found");
		m_lpRequest->HrResponseBody("Item to be Deleted not found");
	}
	else
		m_lpRequest->HrResponseHeader(204, "No Content");
	
	if(lpWastBoxFld)
		lpWastBoxFld->Release();

	if(lpMessage)
		lpMessage->Release();

	if(lpProp)
		MAPIFreeBuffer(lpProp);
		
	if(lpEntryList)
		MAPIFreeBuffer(lpEntryList);

	return hr;
}

//PUT is used for both Adding & modifying entry.
HRESULT CalDAV::HrPut(std::string strServerTZ)
{
	HRESULT hr = hrSuccess;
	std::string strGuid;
	std::string strPath;
	std::string strBinGuid;
	std::string strFldPath;
	std::string strIcal;
	std::string strIcalMod;
	std::string strCharSet;
	eIcalType etype = VEVENT;
	SBinary sbUid;
	time_t ttLastModTime = 0;
	size_t ulFound = 0;
	ULONG ulItemCount = 0;
	ULONG ulPropTag = 0;
	IMAPIFolder *lpUsrFld = NULL;
	IMessage *lpMessage = NULL;
	ICalToMapi *lpICalToMapi = NULL;
	bool blCensorPrivate = false;
	
	m_lpRequest->HrGetFldPath(&strPath);
	ulFound = strPath.rfind("/") + 1;

	lpUsrFld = m_lpUsrFld;
	
	if(!lpUsrFld)
	{
		hr = MAPI_E_BAD_VALUE;
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving calendar folders in PUT request, error code : 0x%08X", hr);
		goto exit;
	}
	
	if ((m_ulFolderFlag & SHARED_FOLDER) && !HasDelegatePerm(m_lpDefStore, m_lpActiveStore))
		blCensorPrivate = true;
	

	strGuid = StripGuid(strPath);
	if(ulFound == 0)
		strFldPath = ""; //default Folder
	else
		strFldPath = strPath.substr(0, ulFound - 1);
	
	ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTTSREF], PT_STRING8);	

	//Find the Entry with particular Guid
	hr = HrFindAndGetMessage(strGuid, lpUsrFld, m_lpNamedProps, &lpMessage);
	if(hr != hrSuccess)
	{		
		//new entry.
		hr = lpUsrFld->CreateMessage(NULL, 0, &lpMessage);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error creating new message, error code : 0x%08X", hr);
			goto exit;
		}

		if(IsValueTs(strGuid))
		{
			hr = HrAddProperty(lpMessage, ulPropTag, false, &strGuid);
			if(hr != hrSuccess)
			{
				m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error adding property to new message, error code : 0x%08X", hr);
				goto exit;
			}
		}
		blCensorPrivate = false; // all add of new private items.
	}
	else
	{
		if (blCensorPrivate && IsPrivate(lpMessage, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN)))
			blCensorPrivate = true;
		else
			blCensorPrivate = false;
	}
	
	if (blCensorPrivate)
	{
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	//save Ical data to mapi.
	CreateICalToMapi(m_lpDefStore, m_lpAddrBook, false, &lpICalToMapi);
	
	m_lpRequest->HrGetBody(&strIcal);

	hr = lpICalToMapi->ParseICal(strIcal, m_strCharset, strServerTZ, m_lpImailUser, 0);
	if(hr!=hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Parsing ical data in PUT request, error code : 0x%08X", hr);
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Error Parsing ical data : %s", strIcal.c_str());
		goto exit;
	}

	ulItemCount = lpICalToMapi->GetItemCount();
	
	if(ulItemCount != 1 )
	{
		hr = MAPI_E_NO_MESSAGES;
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "No message in ical data in PUT request, error code : 0x%08X", hr);
		goto exit;
	}
	
	hr = lpICalToMapi->GetItemInfo(0, &etype, &ttLastModTime, &sbUid);
	if(hr != hrSuccess || etype != VEVENT)
	{
		hr = E_ACCESSDENIED;
		goto exit;
	}

	hr = lpICalToMapi->GetItem(0, 0, lpMessage);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error converting ical data to Mapi message in PUT request, error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpMessage->SaveChanges(0);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error saving Mapi message in PUT request, error code : 0x%08X", hr);
		goto exit;
	}

exit:
	if(hr == hrSuccess)
		m_lpRequest->HrResponseHeader(204, "No Content");
	else if( hr == MAPI_E_NOT_FOUND)
		m_lpRequest->HrResponseHeader(404, "Not Found");
	else if(hr == E_ACCESSDENIED)
		m_lpRequest->HrResponseHeader(403, "Forbidden");
	else
		m_lpRequest->HrResponseHeader(400, "Bad Request");
	
	if(lpICalToMapi)
		delete lpICalToMapi;

	if(lpMessage)
		lpMessage->Release();

	return hr;
}

//Creates a new guid in the message and returns it
HRESULT CalDAV::CreateAndGetGuid(SBinary sbEid, ULONG ulPropTag, string *lpstrGuid)
{
	HRESULT hr = hrSuccess;
	string strGuid;
	LPMESSAGE lpMessage = NULL;
	ULONG ulObjType = 0;
	LPSPropValue lpProp = NULL;

	hr = m_lpActiveStore->OpenEntry(sbEid.cb, (LPENTRYID)sbEid.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN*)&lpMessage);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening message to add Guid, error code : 0x%08X", hr);
		goto exit;
	}

	hr = HrCreateGlobalID(ulPropTag, NULL, &lpProp);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error creating Guid, error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpMessage->SetProps(1, lpProp, NULL);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while adding Guid to message, error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpMessage->SaveChanges(0);
	if (hr != hrSuccess)
		goto exit;

	*lpstrGuid = Util::bin2hex(lpProp->Value.bin.cb, lpProp->Value.bin.lpb);

exit:
	if (lpMessage)
		lpMessage->Release();

	if (lpProp)
		MAPIFreeBuffer(lpProp);

	return hr;
}

HRESULT CalDAV::HrMkCalendar()
{
	HRESULT hr = hrSuccess;

	hr = WebDav::HrMkCalendar();
	
	return hr;
}

HRESULT CalDAV::HrHandleMkCal(WEBDAVPROP *lpsDavProp)
{
	HRESULT hr = hrSuccess;
	std::string strFldName;
	std::string strFldComment;
	std::string strUser;
	std::string strFldOwner;
	IMAPIFolder *lpUsrFld = NULL;
	IMAPIFolder *lpRootFld = NULL;
	LPSPropValue lpsPropValSet = NULL;
	size_t ulFound = 0;
	ULONG ulPropTag = 0;

	ulFound = m_strFldName.find("/");

	// block mkcalendar when accessing single folders.
	// single folder mkcalendar folder-name : temp/NEW-FOLDER-ID
	if(!m_lpRootFld || ulFound != string::npos)
	{
		hr = MAPI_E_ACCESS_DENIED;
		goto exit;
	}
	else
		lpRootFld = m_lpRootFld;

	strFldName = U2W(lpsDavProp->lstProps.front().strValue);
	hr = HrOpenUserFld(m_lpSession, m_lpActiveStore, strFldName, false, m_lpRequest->IsPublic(), m_lpLogger, &lpUsrFld);
	if(hr == hrSuccess)
	{
		hr = MAPI_E_COLLISION;
		goto exit;
	}

	m_lpRequest->HrGetUser(&strUser);
	m_lpRequest->HrGetFldOwner(&strFldOwner);
	
	hr = lpRootFld->CreateFolder(FOLDER_GENERIC, (LPSTR)strFldName.c_str(), (LPSTR)strFldComment.c_str(), NULL, 0, &lpUsrFld);
	if( hr != hrSuccess)
		goto exit;
	
	MAPIAllocateBuffer(sizeof(SPropValue), (void**)&lpsPropValSet);
	lpsPropValSet->ulPropTag = PR_CONTAINER_CLASS;
	lpsPropValSet->Value.lpszA = "IPF.Appointment";

	hr = lpUsrFld->SetProps(1, lpsPropValSet, NULL);
	if(hr != hrSuccess)
		goto exit;
	
	m_lpRequest->HrGetFldPath(&strFldName);

	ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_FLDID], PT_STRING8);
	hr = HrAddProperty(lpUsrFld, ulPropTag, true, &strFldName);
	if(hr != hrSuccess)
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot Add named property, error code : 0x%08X", hr);

exit:
	if(lpsPropValSet)
		MAPIFreeBuffer(lpsPropValSet);

	if(lpUsrFld)
		lpUsrFld->Release();
	return hr;
}


HRESULT CalDAV::HrListCalendar(LPMAPISESSION lpSession, WEBDAVPROP *lpsDavProp, ECLogger *lpLogger, WEBDAVMULTISTATUS *lpsMulStatus)
{
	HRESULT hr = hrSuccess;	
	IMAPITable *lpHichyTable = NULL;
	IMAPITable *lpDelHichyTable = NULL;
	IMAPIFolder *lpIPMSubtree = NULL;
	IMAPIFolder *lpWasteBox = NULL;
	IMAPIFolder *lpSingleFld = NULL;
	IMsgStore *lpMsgStore = NULL;
	SRestriction *lpRestrict = NULL;
	LPSPropValue lpSpropWbEID = NULL;
	LPSPropValue lpsPropSingleFld = NULL;
	LPSPropTagArray lpPropTagArr = NULL;
	LPSRowSet lpRowsALL = NULL;
	LPSRowSet lpRowsDeleted = NULL;
	size_t cbsize = 0;
	ULONG ulPropTagFldId = 0;
	ULONG ulObjType = 0;
	ULONG ulCmp = 0;
	ULONG ulDelEntries = 0;
	std::list<WEBDAVPROPERTY>::iterator iter;
	WEBDAVRESPONSE sDavResponse;
	std::string strUser;
	std::string strFldOwner;
	bool blSingleFld = false;

	lpIPMSubtree = m_lpRootFld;
	lpMsgStore = m_lpActiveStore;

	m_lpRequest->HrGetUser(&strUser);
	m_lpRequest->HrGetFldOwner(&strFldOwner);

	if(!m_blIsCLMAC)
	{
		hr = MAPI_E_BAD_VALUE;
		goto exit;
	}

	if(m_ulFolderFlag & DEFAULT_FOLDER || m_ulFolderFlag & SINGLE_FOLDER)
	{
		blSingleFld = true;
		if(m_lpUsrFld)
			lpSingleFld = m_lpUsrFld;
		else
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error User's Folder not found, while accessing single folder");
			goto exit;
		}
	}

	if(!blSingleFld )
	{		
		hr = HrGetSubCalendars(m_lpSession, lpIPMSubtree, NULL, &lpHichyTable);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving subcalendars for IPM_Subtree, error code : 0x%08X", hr);
			goto exit;
		}

		if(m_lpRequest->IsPublic())
			hr = HrGetOneProp(m_lpDefStore, PR_IPM_WASTEBASKET_ENTRYID, &lpSpropWbEID);
		else
			hr = HrGetOneProp(m_lpActiveStore, PR_IPM_WASTEBASKET_ENTRYID, &lpSpropWbEID);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving subcalendars for IPM_Subtree, error code : 0x%08X", hr);
			goto exit;
		}
		
		if(m_lpRequest->IsPublic())
			hr = m_lpDefStore->OpenEntry(lpSpropWbEID->Value.bin.cb, (LPENTRYID)lpSpropWbEID->Value.bin.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN *)&lpWasteBox);
		else
			hr = m_lpActiveStore->OpenEntry(lpSpropWbEID->Value.bin.cb, (LPENTRYID)lpSpropWbEID->Value.bin.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN *)&lpWasteBox);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening deleted items folder, error code : 0x%08X", hr);
			goto exit;
		}
		
		hr = HrGetSubCalendars(m_lpSession, lpWasteBox, NULL, &lpDelHichyTable);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving subcalendars for Deleted items folder, error code : 0x%08X", hr);
			goto exit;
		}
	}

	cbsize = lpsDavProp->lstProps.size() + 2;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray((ULONG)cbsize), (void **)&lpPropTagArr);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot allocate memory");
		goto exit;
	}

	ulPropTagFldId = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_FLDID], PT_STRING8);
	//add PR_ENTRYID & FolderID in setcolumns along with requested data.
	lpPropTagArr->cValues = (ULONG)cbsize;
	lpPropTagArr->aulPropTag[0] = PR_ENTRYID;
	lpPropTagArr->aulPropTag[1] = ulPropTagFldId;

	iter = lpsDavProp->lstProps.begin();
	for(int i = 2; iter != lpsDavProp->lstProps.end(); iter++, i++)	
	{
		lpPropTagArr->aulPropTag[i] = GetPropmap((char*)iter->sPropName.strPropname.c_str());
	}

	if(!blSingleFld)
	{
		hr = lpHichyTable->SetColumns(lpPropTagArr, 0);
		if(hr != hrSuccess)
			goto exit;

		hr = lpDelHichyTable->SetColumns(lpPropTagArr, 0);
		if(hr != hrSuccess)
			goto exit;
	}
	else
	{
		hr = lpSingleFld->GetProps(lpPropTagArr, 0, (ULONG *)&cbsize, &lpsPropSingleFld);
	}

	if (FAILED(hr))
		goto exit;

	hr = hrSuccess;

	HrSetDavPropName(&(lpsMulStatus->sPropName), "multistatus", lpsDavProp->sPropName.strNS);
	while(1)
	{
		ULONG ulEntryCount = 0;
		if(!blSingleFld)
		{
			hr = lpHichyTable->QueryRows(50, 0, &lpRowsALL);
			if(hr != hrSuccess || lpRowsALL->cRows == 0)
				break;
		
			hr = lpDelHichyTable->QueryRows(50, 0, &lpRowsDeleted);
			if(hr != hrSuccess)
				break;
		}
		if(blSingleFld)
			ulEntryCount = 1;
		else
			ulEntryCount = lpRowsALL->cRows;

		for(ULONG i = 0; i < ulEntryCount; i++)
		{
			std::string strFldPath;
			std::string strPath;
			std::string strTrail;
			WEBDAVPROPSTAT sWebPropStat;
			WEBDAVPROP sPropNF;
			WEBDAVPROP sPropOK;
			
			if(!blSingleFld && lpRowsDeleted->cRows != 0 && ulDelEntries != lpRowsDeleted->cRows )
			{
				ulCmp = memcmp((const void *)lpRowsALL->aRow[i].lpProps[0].Value.bin.lpb,
							   (const void *)lpRowsDeleted->aRow[ulDelEntries].lpProps[0].Value.bin.lpb, lpRowsALL->aRow[i].lpProps[0].Value.bin.cb);
				if(ulCmp == 0)
				{
					ulDelEntries++;
					continue;
				}
			}

			HrSetDavPropName(&(sDavResponse.sPropName), "response", lpsDavProp->sPropName.strNS);
			sPropNF.sPropName = lpsDavProp->sPropName;
			sPropOK.sPropName = lpsDavProp->sPropName;

			HrSetDavPropName(&(sWebPropStat.sPropName), "propstat", lpsDavProp->sPropName.strNS);
			m_lpRequest->HrGetReqPath(&strPath);
			strTrail = strPath.at(strPath.length() -1);

			if(blSingleFld && lpsPropSingleFld && lpsPropSingleFld[1].ulPropTag == ulPropTagFldId)
				strFldPath = lpsPropSingleFld[1].Value.lpszA;
			else if (lpRowsALL && lpRowsALL && lpRowsALL->aRow[i].lpProps[1].ulPropTag == ulPropTagFldId)
				strFldPath = lpRowsALL->aRow[i].lpProps[1].Value.lpszA;
			else
				strFldPath.clear();
			
			if(blSingleFld && strFldPath.empty() && lpsPropSingleFld && lpsPropSingleFld[0].ulPropTag == PR_ENTRYID )
				hr = HrAddProperty(m_lpActiveStore, lpsPropSingleFld[0].Value.bin, ulPropTagFldId, true, &strFldPath);
			else if (strFldPath.empty() && lpRowsALL && lpRowsALL->aRow[i].lpProps[0].ulPropTag == PR_ENTRYID )
				hr = HrAddProperty(m_lpActiveStore, lpRowsALL->aRow[i].lpProps[0].Value.bin, ulPropTagFldId, true, &strFldPath);

			if( hr != hrSuccess || strFldPath.empty())
			{
				m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error adding Folder id property, error code : 0x%08X", hr);
				continue;
			}
			if(strFldOwner.empty() && !m_lpRequest->IsPublic())
				strFldPath = W2U(strUser) + "/" + FOLDER_PREFIX + strFldPath + "/";
			else
				strFldPath = FOLDER_PREFIX + strFldPath+ "/";

			strPath = strPath + strFldPath;

			HrSetDavPropName(&(sDavResponse.sHRef.sPropName), "href", lpsDavProp->sPropName.strNS);
			sDavResponse.sHRef.strValue = strPath;

			iter = lpsDavProp->lstProps.begin();
			if(!blSingleFld)
				cbsize = lpRowsALL->aRow[i].cValues;

			for(ULONG j = 2; j < cbsize; j++, iter++)
			{
				WEBDAVPROPERTY sDavProperty;
				sDavProperty = *iter;

				if(sDavProperty.sPropName.strPropname == "resourcetype")
				{
					WEBDAVVALUE sWebVal;

					HrSetDavPropName(&(sWebVal.sPropName), "collection", "");
					sDavProperty.lstValues.push_back(sWebVal);
					HrSetDavPropName(&(sWebVal.sPropName), "calendar", CALDAVNS);
					sDavProperty.lstValues.push_back(sWebVal);
				}
				else if(sDavProperty.sPropName.strPropname == "supported-calendar-component-set")
				{
					WEBDAVVALUE sWebVal;

					HrSetDavPropName(&(sWebVal.sPropName), "comp","name", "VEVENT", CALDAVNS);
					sDavProperty.lstValues.push_back(sWebVal);
				}
				else if(sDavProperty.sPropName.strPropname == "current-user-privilege-set")
				{
					sDavProperty.sPropName.strNS = "";
					HrBuildACL(&sDavProperty);
				}
				else
				{
					std::string strRetVal;
					if(!blSingleFld)
						strRetVal = SPropValToString(&lpRowsALL->aRow[i].lpProps[j]);
					else
						strRetVal = SPropValToString(&lpsPropSingleFld[j]);
					if(!strRetVal.empty())
						sDavProperty.strValue = W2U(strRetVal);
					else
					{
						sPropNF.lstProps.push_back(sDavProperty);
						continue;
					}
				}
				sPropOK.lstProps.push_back(sDavProperty);
			}
			
			HrSetDavPropName(&(sWebPropStat.sStatus.sPropName), "status", lpsDavProp->sPropName.strNS);
			sDavResponse.lstsPropStat.clear();
			if(!sPropOK.lstProps.empty())
			{
				sWebPropStat.sStatus.strValue = "HTTP/1.1 200 OK";
				sWebPropStat.sProp = sPropOK;
				sDavResponse.lstsPropStat.push_back(sWebPropStat);
			}
			if(!sPropNF.lstProps.empty())
			{
				sWebPropStat.sStatus.strValue = "HTTP/1.1 404 Not Found";
				sWebPropStat.sProp = sPropNF;
				sDavResponse.lstsPropStat.push_back(sWebPropStat);
			}
			lpsMulStatus->lstResp.push_back(sDavResponse);

		}

		if(blSingleFld)
			break;

		if(lpRowsALL)
		{
			FreeProws(lpRowsALL);
			lpRowsALL = NULL;
		}

		if(lpRowsDeleted)
		{
			FreeProws(lpRowsDeleted);
			lpRowsDeleted = NULL;
		}
	}

exit:
	if(lpWasteBox)
		lpWasteBox->Release();

	if(lpHichyTable)
		lpHichyTable->Release();
	
	if(lpDelHichyTable)
		lpDelHichyTable->Release();

	if(lpsPropSingleFld)
		MAPIFreeBuffer(lpsPropSingleFld);

	if(lpSpropWbEID)
		MAPIFreeBuffer(lpSpropWbEID);

	if(lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);

	if(lpRowsALL)
		FreeProws(lpRowsALL);
	
	if(lpRowsDeleted)
		FreeProws(lpRowsDeleted);

	if(lpRestrict)
		FREE_RESTRICTION(lpRestrict);

	return hr;
}

HRESULT CalDAV::HrPropPatch()
{
	HRESULT hr = hrSuccess;

	hr = WebDav::HrPropPatch();

	return hr;
}

// function edits name & comments on folder
HRESULT CalDAV::HrHandlePropPatch(WEBDAVPROP *lpsDavProp)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpsPropVal = NULL;
	std::list<WEBDAVPROPERTY>::iterator iter;
	std::string strFldPath;
	std::string strFldName;
	std::string strConvProp;
	IMAPIFolder *lpUsrFld = NULL;
	ULONG ulCntr = 0;
	
	m_lpRequest->HrGetFldPath(&strFldPath);
	
	hr = MAPIAllocateBuffer( (ULONG)lpsDavProp->lstProps.size() * sizeof(SPropValue), (LPVOID*)&lpsPropVal);
	if (hr != hrSuccess)
		goto exit;

	for(iter = lpsDavProp->lstProps.begin(); iter != lpsDavProp->lstProps.end(); iter++)
	{
		if (iter->sPropName.strPropname == "displayname" )
			strFldName = iter->strValue;
		else if (iter->sPropName.strPropname == "calendar-free-busy-set")
			goto exit;
		lpsPropVal[ulCntr].ulPropTag = GetPropmap((char*)iter->sPropName.strPropname.c_str());
		
		if (lpsPropVal[ulCntr].ulPropTag == 0) // GetPropMap may not find Property Tag.
			continue;

		strConvProp = U2W(iter->strValue);
		hr = MAPIAllocateMore(strConvProp.length() + 1, lpsPropVal, (void**)&lpsPropVal[ulCntr].Value.lpszA);
		if (hr != hrSuccess)
			goto exit;
		strncpy(lpsPropVal[ulCntr].Value.lpszA ,(char*)strConvProp.c_str(),strConvProp.length() + 1);

		ulCntr++;
	}
	
	// deny rename of defalt Calendar
	if (!strFldName.empty() && !m_blFolderAccess)
	{
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	if (ulCntr != 0 && m_lpUsrFld)
		hr = m_lpUsrFld->SetProps(ulCntr, lpsPropVal, NULL);

exit:
	if(lpsPropVal)
		MAPIFreeBuffer(lpsPropVal);

	if(lpUsrFld)
		lpUsrFld->Release();

	return hr;
}

HRESULT CalDAV::HrPost(std::string strSrvTz)
{
	HRESULT hr = hrSuccess;
	std::string strIcal;
	ICalToMapi *lpIcalToMapi = NULL;
	LPSPropValue lpsGetPropVal = NULL;
	LPMAPIFOLDER lpOutbox = NULL;
	LPMESSAGE lpNewMsg = NULL;
	LPSPropTagArray lpsPropArray = NULL;
	LPSPropValue lpsSetPropVals = NULL;
	SPropValue spropVal;
	ULONG cValues = 0;
	ULONG ulObjType = 0;
	eIcalType etype = VEVENT;

	hr = m_lpRequest->HrGetBody(&strIcal);
	if(hr != hrSuccess)
		goto exit;

	CreateICalToMapi(m_lpDefStore, m_lpAddrBook, false,&lpIcalToMapi);

	if (!lpIcalToMapi)
	{
		hr = MAPI_E_BAD_VALUE;
		goto exit;
	}

	hr = lpIcalToMapi->ParseICal(strIcal, m_strCharset, strSrvTz, m_lpImailUser, 0);
	if (hr != hrSuccess)
		goto exit;

	if (lpIcalToMapi->GetItemCount() != 1)
	{
		hr = hrSuccess; // skip VFREEBUSY
		goto exit;
	}

	hr = lpIcalToMapi->GetItemInfo( 0, &etype, NULL, NULL);
	if ( hr != hrSuccess || etype != VEVENT)
	{
		hr = hrSuccess; // skip VFREEBUSY
		goto exit;
	}

	cValues = 2;
	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cValues), (void **)&lpsPropArray);
	if (hr != hrSuccess)
		goto exit;

	lpsPropArray->cValues = cValues;
	lpsPropArray->aulPropTag[0] = PR_IPM_OUTBOX_ENTRYID;
	lpsPropArray->aulPropTag[1] = PR_IPM_SENTMAIL_ENTRYID;

	hr = m_lpDefStore->GetProps(lpsPropArray, 0, &cValues, &lpsGetPropVal);
	if (hr != hrSuccess && cValues != 2)
		goto exit;

	hr = m_lpDefStore->OpenEntry(lpsGetPropVal[0].Value.bin.cb, (LPENTRYID) lpsGetPropVal[0].Value.bin.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN*)&lpOutbox);
	if (hr != hrSuccess)
		goto exit;

	hr = lpOutbox->CreateMessage(NULL, MAPI_BEST_ACCESS, &lpNewMsg);
	if (hr != hrSuccess)
		goto exit;

	hr = lpIcalToMapi->GetItem( 0, 0, lpNewMsg);
	if (hr != hrSuccess)
		goto exit;

	cValues = 2;
	hr = MAPIAllocateBuffer(cValues * sizeof(SPropValue), (void**)&lpsSetPropVals);
	if (hr != hrSuccess)
		goto exit;

	lpsSetPropVals[0].ulPropTag = PR_SENTMAIL_ENTRYID;
	lpsSetPropVals[0].Value.bin = lpsGetPropVal[1].Value.bin;

	lpsSetPropVals[1].ulPropTag = PR_DELETE_AFTER_SUBMIT;
	lpsSetPropVals[1].Value.b = false;

	hr = lpNewMsg->SetProps(cValues, lpsSetPropVals, NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpNewMsg->SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;

	hr = lpNewMsg->SubmitMessage(0);
	if (hr != hrSuccess)
		goto exit;

exit:
	
	if(hr == hrSuccess)
		m_lpRequest->HrResponseHeader(200, "Ok");
	else
		m_lpRequest->HrResponseHeader(400, "Bad Request");

	if (lpIcalToMapi)
		delete lpIcalToMapi;

	if (lpNewMsg)
		lpNewMsg->Release();

	if (lpOutbox)
		lpOutbox->Release();

	if (lpsGetPropVal)
		MAPIFreeBuffer(lpsGetPropVal);

	if (lpsSetPropVals)
		MAPIFreeBuffer(lpsSetPropVals);

	return hr;
}

std::string CalDAV::W2U(std::string strWinChar)
{
	return m_lpConvResponse->convert(strWinChar);
}

std::string CalDAV::U2W(std::string strUtfChar)
{
	return m_lpConvRequest->convert(strUtfChar);
}
