/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef _CALDAVPROTO_H_
#define _CALDAVPROTO_H_

#include "WebDav.h"
#include "CalDavUtil.h"
#include <libxml/uri.h>

#include "restrictionutil.h"
#include "mapiext.h"
#include "MAPIToICal.h"
#include "ICalToMAPI.h"
#include "icaluid.h"

class CalDAV : public WebDav
{

public:
	CalDAV(Http *lpRequest, IMAPISession *lpSession, ECLogger *lpLogger, std::string strCharset);
	~CalDAV();
	HRESULT HrHandleCommand(std::string strMethod, std::string strServerTZ);
	
protected:
	IMAPIFolder *m_lpUsrFld;
	IMAPIFolder *m_lpRootFld;
	IMsgStore *m_lpDefStore;
	IMsgStore *m_lpSharedStore;
	IMsgStore *m_lpPubStore;
	IMsgStore *m_lpActiveStore;
	IAddrBook *m_lpAddrBook;
	IMailUser *m_lpImailUser;
	SPropTagArray *m_lpNamedProps;
	std::string m_strEntryRef;
	std::string m_strFldName;
	std::string m_strCharset;
	bool m_blFolderAccess;
	ULONG m_ulFolderFlag;
	ECIConv *m_lpConvResponse;
	ECIConv *m_lpConvRequest;

	HRESULT HrPropfind();
	HRESULT HrReport();
	HRESULT HrMkCalendar();
	HRESULT HrDelete();
	HRESULT HrPut(std::string strServerTZ);
	HRESULT HrPropPatch();
	HRESULT HrPost(std::string strSrvTz);

	HRESULT HrGetFolder();
	HRESULT HrHandlePropfind(WEBDAVREQSTPROPS *sDavProp,WEBDAVMULTISTATUS *lpsDavMulStatus);
	HRESULT HrHandlePropfindRoot(WEBDAVREQSTPROPS *sDavProp, WEBDAVMULTISTATUS *lpsDavMulStatus);	
	HRESULT HrHandleReport();
	HRESULT HrHandleReport(WEBDAVRPTMGET *sWebRMGet, WEBDAVMULTISTATUS *sWebMStatus);
	//Used By both PROPFIND & Report Calendar-query
	HRESULT HrListCalEntries(WEBDAVREQSTPROPS *sWebRCalQry, WEBDAVMULTISTATUS *sWebMStatus);
	HRESULT HrConvertRowItemToICal(WEBDAVPROP &sDavProp, LPSRow lpsRow, MapiToICal *lpMtIcal, bool blCensor, WEBDAVPROP *lpWebProp);
	HRESULT HrHandleMkCal(WEBDAVPROP *lpsDavProp);
	HRESULT HrHandleMkCal(std::string strFldName,std::string strFldComment);
	HRESULT HrHandlePropPatch(WEBDAVPROP *lpsDavProp);

	HRESULT CreateAndGetGuid(SBinary sbEid, ULONG ulPropTag, string *lpstrGuid);
	HRESULT HrListCalendar(LPMAPISESSION lpSession, WEBDAVPROP *lpsDavProp, ECLogger *lpLogger, WEBDAVMULTISTATUS *lpsMulStatus);

	// convert windows-1252 to utf-8
	std::string W2U(std::string);
	// convert utf-8 to windows-1252
	std::string U2W(std::string);
};

#endif
