/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "CalDavUtil.h"
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

HRESULT HrAuthenticate(std::string strUser, std::string strPass, std::string strPath, IMAPISession **lppSession)
{
	HRESULT hr = hrSuccess;	
	
	hr = HrOpenECSession(lppSession, (char *)strUser.c_str(), (char *)strPass.c_str(), (char *)strPath.c_str(), 0, NULL, NULL, NULL);

	return hr;
}

HRESULT HrOpenUserFld(IMAPISession *lpSession, IMsgStore *lpMsgStore, std::string strFolderPath, bool blCreateIfNF, bool blIsPublic, ECLogger *lpLogger, IMAPIFolder **lpUsrFld)
{
	HRESULT hr = hrSuccess;	
	std::string strPublic;	
	char lpszpsep = '/';
	LPMAPIFOLDER  lpSubFolder = NULL;
	SPropValue sPropVal;
	
	if(strFolderPath.empty())
	{
		hr = HrOpenDefaultCalFld(lpMsgStore, lpLogger, &lpSubFolder);
		if (hr != hrSuccess)
		{
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot open Users Default Folder : 0x%08X", hr);
			goto exit;
		}		
	}
	
	else
	{
		hr = OpenSubFolder(lpMsgStore, (char*)strFolderPath.c_str(), lpszpsep, lpLogger, blIsPublic, false, &lpSubFolder);
		if(hr == MAPI_E_NOT_FOUND && blCreateIfNF)
			hr = OpenSubFolder(lpMsgStore, (char*)strFolderPath.c_str(), lpszpsep, lpLogger, blIsPublic, blCreateIfNF, &lpSubFolder);	
		if (hr != hrSuccess)
		{
			lpLogger->Log(EC_LOGLEVEL_DEBUG, "Cannot open Users Folder : 0x%08X", hr);
			goto exit;
		}
	}
	if(blCreateIfNF)
	{
		std::string strClass;
		strClass = "IPF.Appointment";

		sPropVal.ulPropTag = PR_CONTAINER_CLASS;
		sPropVal.Value.lpszA = (LPSTR)strClass.c_str();
		HrSetOneProp(lpSubFolder,&sPropVal);//ignoring if property is not set.		
	}

	*lpUsrFld = lpSubFolder;

exit:
	
	return hr;
}

HRESULT HrOpenDefaultCalFld(IMsgStore *lpMsgStore, ECLogger *lpLogger, LPMAPIFOLDER *lpDefFolder)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpPropIPMSubtree = NULL;
	LPSPropValue lpPropDefFld = NULL;
	LPMAPIFOLDER lpRootFld = NULL;
	LPMAPIFOLDER lpDefaultFolder = NULL;
	ULONG ulType = 0;

	//open Root Container.
	hr = lpMsgStore->OpenEntry(0, NULL, NULL, MAPI_BEST_ACCESS, &ulType, (LPUNKNOWN*)&lpRootFld);
	if (hr != hrSuccess)
	{
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to open root container while opening default calendar, error code : 0x%08X", hr);
		goto exit;
	}

	//retrive Entryid of Default Calendar Folder.
	hr = HrGetOneProp(lpRootFld, PR_IPM_APPOINTMENT_ENTRYID, &lpPropDefFld);
	if (hr != hrSuccess)
	{
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to find PR_IPM_APPOINTMEN_ENTRYID while opening default calendar, error code : 0x%08X", hr);
		goto exit;
	}
	
	hr = lpMsgStore->OpenEntry(lpPropDefFld->Value.bin.cb, (LPENTRYID)lpPropDefFld->Value.bin.lpb, &IID_IMAPIFolder,MAPI_BEST_ACCESS, &ulType, (LPUNKNOWN*)&lpDefaultFolder);
	if (hr != hrSuccess || ulType != MAPI_FOLDER)
	{
		lpLogger->Log(EC_LOGLEVEL_DEBUG, "Unable to open User's Default Calendar , error code : 0x%08X", hr);
		goto exit;
	}

	*lpDefFolder = lpDefaultFolder;
	lpDefaultFolder = NULL;

exit:
	if(lpDefaultFolder)
		lpDefaultFolder->Release();

	if(lpRootFld)
		lpRootFld->Release();

	if(lpPropDefFld)
		MAPIFreeBuffer(lpPropDefFld);
	
	if(lpPropIPMSubtree)
		MAPIFreeBuffer(lpPropIPMSubtree);

	return hr;
}

HRESULT HrAddProperty(IMAPIProp *lpMapiProp, ULONG ulPropTag, bool bFldId, std::string *strProperty)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;
	ULONG ulMsgProps = 0;
	LPSPropValue lpMsgProps = NULL;
	LPSPropValue lpsPropSrch = NULL;
	LPSPropValue lpsPropValEid = NULL;
	GUID sGuid ;

	if(strProperty->empty())
	{
		CoCreateGuid(&sGuid);
		*strProperty = Util::bin2hex(sizeof(GUID), (LPBYTE)&sGuid);
	}
	sPropVal.ulPropTag = ulPropTag;
	sPropVal.Value.lpszA = (LPSTR)strProperty->c_str();
	
	hr = lpMapiProp->GetProps(NULL, 0, &ulMsgProps, &lpMsgProps);
	if (hr != hrSuccess)
		goto exit;
	
	lpsPropSrch = PpropFindProp(lpMsgProps, ulMsgProps, sPropVal.ulPropTag);
	if(lpsPropSrch == NULL)
		hr = HrSetOneProp(lpMapiProp,&sPropVal);
	else
		strProperty->assign(lpsPropSrch->Value.lpszA);
	if (hr == E_ACCESSDENIED && bFldId )
	{
		hr = HrGetOneProp(lpMapiProp,PR_ENTRYID,&lpsPropValEid);
		if(hr != hrSuccess)
			goto exit;
		*strProperty = Util::bin2hex(lpsPropValEid[0].Value.bin.cb,lpsPropValEid[0].Value.bin.lpb);
	}
	if(hr != hrSuccess)
		goto exit;
	
	hr = lpMapiProp->SaveChanges(KEEP_OPEN_READWRITE);

exit:
	if(lpsPropValEid)
		MAPIFreeBuffer(lpsPropValEid);

	if(lpMsgProps)
		MAPIFreeBuffer(lpMsgProps);

	if(lpsPropSrch)
		MAPIFreeBuffer(lpsPropSrch);

	return hr;
}

HRESULT HrAddProperty(IMsgStore *lpMsgStore, SBinary sbEid, ULONG ulPropertyId, bool bIsFldID, std::string *lpstrProperty )
{
	HRESULT hr = hrSuccess;
	IMAPIFolder *lpUsrFld = NULL;
	ULONG ulObjType = 0;
	
	hr = lpMsgStore->OpenEntry(sbEid.cb, (LPENTRYID)sbEid.lpb,NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN *) &lpUsrFld);
	if(hr != hrSuccess)
		goto exit;

	hr = HrAddProperty(lpUsrFld, ulPropertyId, bIsFldID, lpstrProperty);

exit:
	if(lpUsrFld)
		lpUsrFld->Release();

	return hr;
}

HRESULT HrFindFolder(IMAPISession *lpSession, IMsgStore *lpMsgStore, LPSPropTagArray lpNamedProps, ECLogger *lpLogger, bool blIsPublic, std::string strFldId, IMAPIFolder **lppUsrFld)
{
	HRESULT hr = hrSuccess;
	std::string strBinEid;
	SRestriction *lpRestrict = NULL;
	IMAPITable *lpHichyTable = NULL;
	SPropValue sSpropVal ;
	ULONG ulPropTagFldId = 0;
	SPropTagArray *lpPropTagArr = NULL;
	SRowSet *lpRows = NULL;
	SBinary sbEid = {0,0};
	IMAPIFolder *lpUsrFld = NULL;
	IMAPIFolder *lpRootFld = NULL;
	ULONG cbsize = 0;
	ULONG ulObjType = 0;
	size_t ulFound = 0;

	ulFound = strFldId.find(FOLDER_PREFIX);
	if(ulFound != string::npos)
		strFldId = strFldId.substr(ulFound + strlen(FOLDER_PREFIX), strFldId.length() - ulFound);

	//TODO: public 
	hr = OpenSubFolder(lpMsgStore, NULL, '/', lpLogger, blIsPublic, false, &lpRootFld);
	if(hr != hrSuccess)
	{
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot open IPM_SUBTREE Folder");
		goto exit;
	}
	
	hr = lpRootFld->GetHierarchyTable(CONVENIENT_DEPTH,&lpHichyTable);
	if(hr != hrSuccess)
		goto exit;
	//When ENTRY_ID is use For Read Only Calendars assuming the folder id string 
	//not larger than lenght 50
	//FixME: include some Entry-id identifier
	if(strFldId.size()> 50)
	{
		ulPropTagFldId = PR_ENTRYID;
		Util::hex2bin(strFldId, &strBinEid);
		sSpropVal.Value.bin.cb = strBinEid.size();
		sSpropVal.Value.bin.lpb = (LPBYTE)strBinEid.c_str();
	}
	else
	{
		ulPropTagFldId = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_FLDID], PT_STRING8);
		sSpropVal.Value.lpszA = (LPSTR)strFldId.c_str();
	}

	CREATE_RESTRICTION(lpRestrict);
	lpRestrict->rt = RES_PROPERTY;
	lpRestrict->res.resProperty.ulPropTag =  ulPropTagFldId;
	lpRestrict->res.resProperty.relop = RELOP_EQ;
	
	sSpropVal.ulPropTag = ulPropTagFldId;
	
	lpRestrict->res.resProperty.lpProp = &sSpropVal;

	hr = lpHichyTable->Restrict(lpRestrict,TBL_BATCH);
	if(hr != hrSuccess)
		goto exit;

	cbsize = 1;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cbsize), (void **)&lpPropTagArr);
	if(hr != hrSuccess)
	{
		lpLogger->Log(EC_LOGLEVEL_ERROR,"Cannot allocate memory");
		goto exit;
	}

	//add PR_ENTRYID in setcolumns along with requested data.
	lpPropTagArr->cValues = 1;
	lpPropTagArr->aulPropTag[0] = PR_ENTRYID;

	hr = lpHichyTable->SetColumns(lpPropTagArr,0);
	if(hr != hrSuccess)
		goto exit;
	
	hr = lpHichyTable->QueryRows(1,0,&lpRows);
	if (hr != hrSuccess)
		goto exit;

	if (lpRows->cRows != 1) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	sbEid = lpRows->aRow[0].lpProps[0].Value.bin;
	
	hr = lpMsgStore->OpenEntry(sbEid.cb, (LPENTRYID)sbEid.lpb,NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN *) &lpUsrFld);
	if(hr != hrSuccess)
		goto exit;

	*lppUsrFld = lpUsrFld;

exit:
	if(lpRows)
		FreeProws(lpRows);

	if(lpRootFld)
		lpRootFld->Release();

	if(lpHichyTable)
		lpHichyTable->Release();
	
	if(lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);

	if(lpRestrict)
		FREE_RESTRICTION(lpRestrict);

	return hr;
}

HRESULT HrBuildACL(WEBDAVPROPERTY *lpsProperty)
{
	HRESULT hr = hrSuccess;
	int ulDepth = 0 ;
	WEBDAVITEM sDavItem;

	sDavItem.sDavValue.sPropName.strPropname = "privilege";
	sDavItem.ulDepth = ulDepth ;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "all";
	sDavItem.ulDepth = ulDepth + 1;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "privilege";
	sDavItem.ulDepth = ulDepth ;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "read-current-user-privilege-set";
	sDavItem.ulDepth = ulDepth + 1;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "privilege";
	sDavItem.ulDepth = ulDepth ;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "read";
	sDavItem.ulDepth = ulDepth + 1;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "privilege";
	sDavItem.ulDepth = ulDepth ;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "write";
	sDavItem.ulDepth = ulDepth + 1;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "privilege";
	sDavItem.ulDepth = ulDepth;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "write-content";
	sDavItem.ulDepth = ulDepth + 1;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "privilege";
	sDavItem.ulDepth = ulDepth;
	lpsProperty->lstItems.push_back(sDavItem);

	sDavItem.sDavValue.sPropName.strPropname = "write-properties";
	sDavItem.ulDepth = ulDepth + 1;
	lpsProperty->lstItems.push_back(sDavItem);

	return hr;
}

bool IsValueTs(std::string strValue)
{
	bool retVal = false;
	
	if(strValue.length() == 16 && strValue.at(8) == 'T' && strValue.at(15) == 'Z' )
		retVal = true;
		
	return retVal;
}

std::string StripGuid(std::string strInput)
{
	size_t ulFound = -1;
	size_t ulFoundSlash = -1;
	std::string strRetVal;

	ulFoundSlash = strInput.rfind("/");
	if(ulFoundSlash == string::npos)
		ulFoundSlash = 0;
	else
		ulFoundSlash++;

	ulFound = strInput.rfind(".ics");
	if(ulFound != string::npos)
		strRetVal = strInput.substr(ulFoundSlash, (strInput.length() - ulFoundSlash - 4)); //-4 to get rid of .ics

	return strRetVal;
}

HRESULT HrGetUserInfo(IMAPISession *lpSession, IMsgStore *lpDefStore, std::string *strEmailaddress, std::string *strUserName, IMailUser **lppImailUser)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpSProp = NULL;
	IMailUser *lpMailUser = NULL;
	ULONG ulObjType = 0;

	hr = HrGetOneProp(lpDefStore, PR_MAILBOX_OWNER_ENTRYID, &lpSProp);
	if(hr != hrSuccess)
		goto exit;
	
	hr = lpSession->OpenEntry(lpSProp->Value.bin.cb, (LPENTRYID)lpSProp->Value.bin.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN*)&lpMailUser);
	if(hr != hrSuccess)
		goto exit;

	MAPIFreeBuffer(lpSProp);
	lpSProp = NULL;

	hr = HrGetOneProp(lpMailUser, PR_SMTP_ADDRESS, &lpSProp);
	if(hr != hrSuccess)
		goto exit;

	if(strEmailaddress)
		*strEmailaddress = lpSProp->Value.lpszA;

	MAPIFreeBuffer(lpSProp);
	lpSProp = NULL;

	hr = HrGetOneProp(lpMailUser, PR_DISPLAY_NAME, &lpSProp);
	if(hr != hrSuccess)
		goto exit;
	
	if(strUserName)
		*strUserName = lpSProp->Value.lpszA;

	if(lpMailUser && lppImailUser)
	{
		*lppImailUser = lpMailUser;
		lpMailUser = NULL;
	}

exit:
	if(lpSProp)
		MAPIFreeBuffer(lpSProp);

	if(lpMailUser)
		lpMailUser->Release();

	return hr;
}

HRESULT HrGetSubCalendars(IMAPISession *lpSession, IMAPIFolder *lpFolderIn, SBinary *lpsbEid, IMAPITable **lppTable)
{
	HRESULT hr = hrSuccess;
	IMAPIFolder *lpFolder = NULL;
	ULONG ulObjType = 0;
	IMAPITable *lpTable = NULL;
	SPropValue sPropVal;
	SRestriction *lpRestrict = NULL;
	bool FreeFolder = false;
	SizedSSortOrderSet(1,sSortOrder) = {1, 0, 0, { {PR_ENTRYID, TABLE_SORT_ASCEND} } };

	if(!lpFolderIn)
	{
		hr = lpSession->OpenEntry(lpsbEid->cb, (LPENTRYID)lpsbEid->lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN *)&lpFolder);
		if(hr != hrSuccess)
			goto exit;
		FreeFolder = true;
	}
	else
		lpFolder = lpFolderIn;

	hr = lpFolder->GetHierarchyTable(CONVENIENT_DEPTH,&lpTable);
	if(hr != hrSuccess)
		goto exit;
	
	hr = lpTable->SortTable((LPSSortOrderSet)&sSortOrder, 0);
	if(hr != hrSuccess)
		goto exit;

	sPropVal.ulPropTag = PR_CONTAINER_CLASS;
	sPropVal.Value.lpszA = "IPF.Appointment";

	CREATE_RESTRICTION(lpRestrict);
	DATA_RES_PROPERTY(lpRestrict, (*lpRestrict), RELOP_EQ, sPropVal.ulPropTag, &sPropVal);	
	hr = lpTable->Restrict(lpRestrict,TBL_BATCH);
	if(hr != hrSuccess)
		goto exit;

	*lppTable = lpTable;

exit:
	if(lpRestrict)
		FREE_RESTRICTION(lpRestrict);
	
	if(FreeFolder && lpFolder)
		lpFolder->Release();

	return hr;
}

HRESULT HrGetSharedFolder(IMAPISession *lpSession, IMsgStore *lpUsrStore, std::string strUser, std::string strFolderPath, bool blIsClMac,bool blCreateIfNF, ECLogger *lpLogger, ULONG *ulFolderFlag, IMsgStore **lppSharedStore, IMAPIFolder **lppUsrFld)
{
	HRESULT hr = hrSuccess;
	LPEXCHANGEMANAGESTORE lpExchangeManageStore = NULL;
	SBinary sbSharedStoreEid = {0, 0};
	IMsgStore *lpSharedStore = NULL;
	LPSPropValue lpPropIpmEid = NULL;
	LPSPropTagArray lpNamedProps = NULL;
	IMAPIFolder *lpUsrFld = NULL;
	ULONG ulObjType = 0;

	hr = lpUsrStore->QueryInterface(IID_IExchangeManageStore, (void **)&lpExchangeManageStore);
	if(hr != hrSuccess)
	{
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Error in QueryInterface while opening shared folder ,error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpExchangeManageStore->CreateStoreEntryID(NULL,(LPSTR) strUser.c_str(), OPENSTORE_USE_ADMIN_PRIVILEGE | OPENSTORE_TAKE_OWNERSHIP, &sbSharedStoreEid.cb,(LPENTRYID *) &sbSharedStoreEid.lpb);
	if(hr != hrSuccess)
	{
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Error in CreateStoreEntryID while opening shared folder ,error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpSession->OpenMsgStore(0, sbSharedStoreEid.cb, (LPENTRYID)sbSharedStoreEid.lpb, &IID_IMsgStore, MAPI_BEST_ACCESS, &lpSharedStore);
	if(hr != hrSuccess)
	{
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening store of user %s ,error code : 0x%08X", strUser.c_str(), hr);
		goto exit;
	}
	
	if(blIsClMac)
	{	
		hr = HrLookupNames(lpSharedStore, &lpNamedProps);
		if (hr != hrSuccess)
			goto exit;

		hr = HrGetOneProp(lpSharedStore, PR_IPM_SUBTREE_ENTRYID, &lpPropIpmEid);
		if (hr != hrSuccess)
			lpLogger->Log(EC_LOGLEVEL_DEBUG, "Unable to get Entryid of IPM_SUBTREE folder of user %s, error code : 0x%08X", strUser.c_str(), hr);
		else
			hr = lpSharedStore->OpenEntry(lpPropIpmEid[0].Value.bin.cb, (LPENTRYID)lpPropIpmEid[0].Value.bin.lpb, &IID_IMAPIFolder, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN*)&lpUsrFld); 	
		
		//Open root Folder to get all subfolders from this folder
		if(hr == hrSuccess && strFolderPath.empty())
			goto exit;
		//Open Default Calendar if IPM
		else if(strFolderPath.empty())
		{
			hr = HrOpenDefaultCalFld(lpSharedStore, lpLogger, &lpUsrFld);
			if(ulFolderFlag && hr == hrSuccess)
				*ulFolderFlag = *ulFolderFlag | DEFAULT_FOLDER | SHARED_FOLDER;
		}
		else 
		{
			hr = HrFindFolder(lpSession, lpSharedStore, lpNamedProps, lpLogger, false, strFolderPath, &lpUsrFld);
			if(ulFolderFlag && hr == hrSuccess)
				*ulFolderFlag = *ulFolderFlag | SHARED_FOLDER;
		}

		if(hr != hrSuccess && !strFolderPath.empty())
		{
			hr = HrOpenUserFld(lpSession, lpSharedStore, strFolderPath, false, false, lpLogger, &lpUsrFld);
			if(ulFolderFlag && hr == hrSuccess)
				*ulFolderFlag = *ulFolderFlag | SINGLE_FOLDER | SHARED_FOLDER;
		}

		if(hr == E_ACCESSDENIED)
		{
			hr = HrOpenDefaultCalFld(lpSharedStore, lpLogger, &lpUsrFld);
			if(ulFolderFlag && hr == hrSuccess)
				*ulFolderFlag = *ulFolderFlag | DEFAULT_FOLDER | SHARED_FOLDER;
		}
	}
	else if(strFolderPath.empty())
	{
		hr = HrOpenDefaultCalFld(lpSharedStore, lpLogger, &lpUsrFld);
		if(ulFolderFlag && hr == hrSuccess)
			*ulFolderFlag = *ulFolderFlag | DEFAULT_FOLDER | SHARED_FOLDER;
	}
	else
	{
		hr = HrOpenUserFld(lpSession, lpSharedStore, strFolderPath, blCreateIfNF, false, lpLogger, &lpUsrFld);
		if(ulFolderFlag && hr == hrSuccess)
				*ulFolderFlag = *ulFolderFlag | SINGLE_FOLDER | SHARED_FOLDER;
	}
	
exit:

	if(lppUsrFld && lpUsrFld)
	{
		*lppUsrFld = lpUsrFld;
		lpUsrFld = NULL;
	}
	
	if(lppSharedStore && lpSharedStore)
	{
		*lppSharedStore = lpSharedStore;
		lpSharedStore = NULL;
	}
	
	if (lpNamedProps)
		MAPIFreeBuffer(lpNamedProps);

	if(sbSharedStoreEid.lpb)
		MAPIFreeBuffer(sbSharedStoreEid.lpb);

	if(lpSharedStore)
		lpSharedStore->Release();

	if(lpUsrFld)
		lpUsrFld->Release();

	if(lpPropIpmEid)
		MAPIFreeBuffer(lpPropIpmEid);

	if(lpExchangeManageStore)
		lpExchangeManageStore->Release();

	return hr;
}

bool HasDelegatePerm(IMsgStore *lpDefStore, IMsgStore *lpSharedStore)
{
	HRESULT hr = hrSuccess;
	LPMESSAGE lpFbMessage = NULL;
	LPSPropValue lpProp = NULL;
	LPSPropValue lpMailBoxEid = NULL;
	IMAPIContainer *lpRootCont = NULL;
	ULONG ulType = 0;
	ULONG ulPos = 0;
	SBinary sbEid = {0,0};
	bool blFound = false;
	bool blDelegatePerm = false; // no permission

	hr = HrGetOneProp(lpDefStore, PR_MAILBOX_OWNER_ENTRYID, &lpMailBoxEid);
	if (hr != hrSuccess)
		goto exit;

	hr = lpSharedStore->OpenEntry(0, NULL, NULL, 0, &ulType, (LPUNKNOWN*)&lpRootCont);
	if (hr != hrSuccess)
		goto exit;

	hr = HrGetOneProp(lpRootCont, PR_FREEBUSY_ENTRYIDS, &lpProp);
	if (hr != hrSuccess)
		goto exit;

	if (lpProp->Value.MVbin.cValues > 1 && lpProp->Value.MVbin.lpbin[1].cb != 0)
		sbEid = lpProp->Value.MVbin.lpbin[1];
	else
		goto exit;

	hr = lpSharedStore->OpenEntry(sbEid.cb, (LPENTRYID)sbEid.lpb, NULL, MAPI_BEST_ACCESS, &ulType, (LPUNKNOWN*)&lpFbMessage);
	if (hr != hrSuccess)
		goto exit;

	if (lpProp)
	{
		MAPIFreeBuffer(lpProp);
		lpProp = NULL;
	}

	hr = HrGetOneProp(lpFbMessage, PR_SCHDINFO_DELEGATE_ENTRYIDS, &lpProp);
	if (hr != hrSuccess)
		goto exit;

	for (ULONG i = 0; i < lpProp->Value.MVbin.cValues ; i++)
	{
		if (lpProp->Value.MVbin.lpbin[i].cb == lpMailBoxEid->Value.bin.cb
			 && !memcmp((const void *) lpProp->Value.MVbin.lpbin[i].lpb, (const void *)lpMailBoxEid->Value.bin.lpb, lpMailBoxEid->Value.bin.cb))
		{
			blFound = true;
			ulPos = i;
			break;
		}
	}
	
	if (!blFound)
		goto exit;
	
	if (lpProp)
	{
		MAPIFreeBuffer(lpProp);
		lpProp = NULL;
	}

	hr = HrGetOneProp(lpFbMessage, PR_DELEGATE_FLAGS, &lpProp);
	if (hr != hrSuccess)
		goto exit;

	if(lpProp->Value.MVl.cValues >= ulPos && lpProp->Value.MVl.lpl[ulPos])
		blDelegatePerm = true;
	else
		blDelegatePerm = false;

exit:
	
	if (lpMailBoxEid)
		MAPIFreeBuffer(lpMailBoxEid);

	if (lpProp)
		MAPIFreeBuffer(lpProp);

	if (lpFbMessage)
		lpFbMessage->Release();

	if (lpRootCont)
		lpRootCont->Release();

	return blDelegatePerm;
}

bool IsPrivate(LPMESSAGE lpMessage, ULONG ulPropIDPrivate)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpPropPrivate = NULL;
	bool bIsPrivate = false;

	hr = HrGetOneProp(lpMessage, ulPropIDPrivate, &lpPropPrivate);
	if (hr != hrSuccess)
		goto exit;
	
	if(lpPropPrivate->Value.b == true)
		bIsPrivate = true;
	
exit:
	
	if(lpPropPrivate)
		MAPIFreeBuffer(lpPropPrivate);

	return bIsPrivate;
}

std::string SPropValToString(SPropValue * lpSprop)
{
	FILETIME ft;
	time_t tmUnixTime;
	std::string strGuid;

	if (PROP_TYPE(lpSprop->ulPropTag) == PT_SYSTIME)
	{
		std::string strReturn;
		ft = lpSprop->Value.ft;
		FileTimeToUnixTime(ft, &tmUnixTime);
		strReturn = stringify_int64(tmUnixTime, false);
		return strReturn;
	}
	else if (PROP_TYPE(lpSprop->ulPropTag) == PT_STRING8)
	{
		return lpSprop->Value.lpszA;
	}
	//Global UID
	else if (PROP_TYPE(lpSprop->ulPropTag) == PT_BINARY)
	{
		HrGetICalUidFromBinUid(lpSprop->Value.bin, &strGuid);
		return strGuid;
	}	

	return std::string();
}

HRESULT HrFindAndGetMessage(std::string strGuid, IMAPIFolder *lpUsrFld, LPSPropTagArray lpNamedProps, IMessage **lppMessage)
{
	HRESULT hr = hrSuccess;
	std::string strFldPath;
	std::string strTs;
	std::string strBinGuid;
	SBinary sbEid = {0,0};
	SRestriction *lpsRoot = NULL;
	SRowSet *lpValRows = NULL;
	SPropValue sSpropVal = {0};
	SPropValue sSpropValEid = {0};	
	IMAPITable *lpTable = NULL;
	IMessage *lpMessage = NULL;	
	size_t ulFound = 0;	
	ULONG ulPropTag = 0;
	ULONG ulObjType = 0;
	SizedSPropTagArray(1, lpPropTagArr)= {1, {PR_ENTRYID}};
	
	if(IsValueTs(strGuid))
		strBinGuid = strGuid;
	else if (IsOutlookUid(strGuid))
		Util::hex2bin(strGuid, &strBinGuid);
	else
		HrMakeBinUidFromICalUid(strGuid, &strBinGuid);
	
	
	if(!IsValueTs(strGuid))
		ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY);
	else
		ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_APPTTSREF], PT_STRING8);
	
	if(hr != hrSuccess)
		goto exit;

	if(!IsValueTs(strTs))
	{
		sbEid.cb = (ULONG)strBinGuid.size();
		sbEid.lpb = (LPBYTE)strBinGuid.c_str();
		sSpropVal.Value.bin = sbEid;
	}
	else
		sSpropVal.Value.lpszA = (LPSTR)strTs.c_str();

	sSpropVal.ulPropTag = ulPropTag;

	sSpropValEid.Value.bin = sbEid;
	sSpropValEid.ulPropTag = PR_ENTRYID;

	//Restriction to retrieve only one row corresponding to guid.
	CREATE_RESTRICTION(lpsRoot);
	
	
	CREATE_RES_OR(lpsRoot, lpsRoot, 2); // max: or guid or raw or entryid
	lpsRoot->res.resOr.lpRes[0].rt = RES_PROPERTY;
	lpsRoot->res.resOr.lpRes[0].res.resProperty.relop = RELOP_EQ;
	lpsRoot->res.resOr.lpRes[0].res.resProperty.ulPropTag = ulPropTag; // (ical) guid
	lpsRoot->res.resOr.lpRes[0].res.resProperty.lpProp = &sSpropVal;

	lpsRoot->res.resOr.lpRes[1].rt = RES_PROPERTY;
	lpsRoot->res.resOr.lpRes[1].res.resProperty.relop = RELOP_EQ;
	lpsRoot->res.resOr.lpRes[1].res.resProperty.ulPropTag = PR_ENTRYID;
	lpsRoot->res.resOr.lpRes[1].res.resProperty.lpProp = &sSpropValEid;


	hr = lpUsrFld->GetContentsTable(0, &lpTable);
	if(hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns((LPSPropTagArray)&lpPropTagArr, 0);
	if(hr != hrSuccess)
		goto exit;

	hr = lpTable->Restrict(lpsRoot, TBL_BATCH);
	if(hr != hrSuccess)
		goto exit;
	
	hr = lpTable->QueryRows(1, 0, &lpValRows);
	if (hr != hrSuccess)
		goto exit;

	if (lpValRows->cRows != 1)
	{
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	if (PROP_TYPE(lpValRows->aRow[0].lpProps[0].ulPropTag) != PT_BINARY)
	{
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	sbEid = lpValRows->aRow[0].lpProps[0].Value.bin;

	hr = lpUsrFld->OpenEntry(sbEid.cb, (LPENTRYID)sbEid.lpb, NULL, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpMessage);
	if (hr != hrSuccess)
		goto exit;
	
	*lppMessage = lpMessage;
	lpMessage = NULL;

exit:
	if(lpMessage)
		lpMessage->Release();

	if(lpTable)
		lpTable->Release();
	
	if(lpsRoot)
		FREE_RESTRICTION(lpsRoot);

	if(lpValRows)
		FreeProws(lpValRows);
	
	return hr;
}

