/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "WebDav.h"
#include "stringutil.h"
#include "CommonUtil.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// FIXME class variables
WebDav::WebDav(Http *lpRequest, IMAPISession *lpSession,ECLogger *lpLogger)
{	
	std::string strAgent;
	std::string strPrefix;
	
	m_lpRequest = lpRequest;
	m_lpSession = lpSession;
	m_lpLogger  = lpLogger;
	m_lpXmlDoc  = NULL;
	m_blIsCLMAC = false;
	m_blIsCLEVO = false;
	
	m_lpRequest->HrGetUserAgent(&strAgent);
	strPrefix = strAgent.substr(0,strAgent.find("/"));
	if(!strPrefix.compare("DAVKit"))
		m_blIsCLMAC = true;

	if(!strPrefix.compare("Evolution"))
		m_blIsCLEVO = true;
}

WebDav::~WebDav()
{
	if (m_lpXmlDoc)
		xmlFreeDoc(m_lpXmlDoc);
}

HRESULT WebDav::HrParseXml()
{
	HRESULT hr = hrSuccess;
	std::string strBody;

	ASSERT(m_lpXmlDoc == NULL);
	if (m_lpXmlDoc != NULL)
		return hr;

	m_lpRequest->HrGetBody(&strBody);

	m_lpXmlDoc = xmlReadMemory((char *)strBody.c_str(),(int)strBody.length(), "PROVIDE_BASE.xml", NULL,  XML_PARSE_NOBLANKS);
	if (m_lpXmlDoc == NULL)
		hr = MAPI_E_NOT_INITIALIZED;

	strBody.clear();

	return hr;
}

/*
 * propfind xml data example:
 *	<D:propfind xmlns:D="DAV:" xmlns:CS="http://calendarserver.org/ns/">
 *		<D:prop>
 *			<D:resourcetype/>
 *			<D:owner/>
 *			<CS:getctag/>
 *		</D:prop>
 *	</D:propfind>
 */
HRESULT WebDav::HrPropfind()
{
	HRESULT hr = hrSuccess;
		
	WEBDAVPROPSTAT sPropStat;
	WEBDAVRESPONSE sDavResp;
	WEBDAVMULTISTATUS sDavMStatus;
	WEBDAVREQSTPROPS sDavReqsProps;
	WEBDAVPROP sDavPropRet;
	std::string strFldPath;
	std::string strXml;
	xmlNode * lpXmlNode = NULL;

	//libxml parser parses the xml data and returns the DomTree pointer.
	hr = HrParseXml();
	if (hr != hrSuccess)
		goto exit;

	lpXmlNode = xmlDocGetRootElement(m_lpXmlDoc);

	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"propfind"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}
	
	lpXmlNode = lpXmlNode->children;
	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"prop"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}
		
	HrSetDavPropName(&(sDavPropRet.sPropName),lpXmlNode);
	lpXmlNode = lpXmlNode->children;

	while (lpXmlNode)
	{
		WEBDAVPROPERTY sProperty;
		
		HrSetDavPropName(&(sProperty.sPropName),lpXmlNode);
		sDavPropRet.lstProps.push_back(sProperty);
		lpXmlNode = lpXmlNode->next;
	}

	/*
	 *Call to CALDAV::HrHandlePropfind
	 *This function Retrieves data from server and adds it to the structure.
	 */		
	sDavReqsProps.sProp = sDavPropRet;
	hr = HrHandlePropfind(&sDavReqsProps, &sDavMStatus);
	if (hr != hrSuccess)
		goto exit;

	//Convert WEBMULTISTATUS structure to xml data.
	hr = RespStructToXml(&sDavMStatus, &strXml);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Unable to convert reponse to xml: 0x%08X", hr);
		goto exit;
	}
	
exit:
	if(hr == hrSuccess)
	{
		m_lpRequest->HrResponseHeader(207, "Multi-Status");
		m_lpRequest->HrResponseBody(strXml);
	}
	else if (hr == MAPI_E_NOT_FOUND)
		m_lpRequest->HrResponseHeader(404, "Not Found");
	else 
		m_lpRequest->HrResponseHeader(500, "Internal Server Error");

	return hr;
}

HRESULT WebDav::HrHandlePropfind(WEBDAVREQSTPROPS *sProp,WEBDAVMULTISTATUS *lpsDavMulStatus)
{
	return E_NOTIMPL;
}

HRESULT WebDav::HrHandleReport(WEBDAVRPTMGET *sWebRMGet, WEBDAVMULTISTATUS *sWebMStatus)
{
	return E_NOTIMPL;
}

HRESULT WebDav::HrListCalEntries(WEBDAVREQSTPROPS *sWebRCalQry, WEBDAVMULTISTATUS *sWebMStatus)
{
	return E_NOTIMPL;
}

HRESULT WebDav::HrHandlePropPatch(WEBDAVPROP *lpsDavProp)
{
	return E_NOTIMPL;
}

HRESULT WebDav::HrHandleMkCal(WEBDAVPROP *lpsDavProp)
{
	return E_NOTIMPL;
}

HRESULT WebDav::RespStructToXml(WEBDAVMULTISTATUS *sDavMStatus, std::string *strXml)
{
	HRESULT hr = hrSuccess;
	int ulRet = 0;
	std::string strNsPrefix;
	xmlTextWriterPtr xmlWriter = NULL;
    xmlBufferPtr xmlBuff = NULL;
	std::list<WEBDAVRESPONSE>::iterator iterResp;
	std::string strNs;	

	strNsPrefix = "C";
	xmlBuff = xmlBufferCreate();
	std::map<std::string,std::string>::iterator iterMapNS;

	if (xmlBuff == NULL)
	{
		hr = MAPI_E_INSUFFICIENT_MEMORY;
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error allocating memory to xmlBuffer");
		goto exit;
	}

	//Initialize Xml-Writer
	xmlWriter = xmlNewTextWriterMemory(xmlBuff, 0);
	if (xmlWriter == NULL)
	{
		hr = MAPI_E_NOT_INITIALIZED;
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Initializing xmlWriter");
		goto exit;
	}

#ifdef DEBUG
	// let xml use enters and spaces to make it somewhat readable
	xmlTextWriterSetIndent(xmlWriter, 1);
#endif

	//start xml-data i.e <xml version="1.0" encoding="utf-8"?>
	ulRet = xmlTextWriterStartDocument(xmlWriter, NULL, "UTF-8", NULL);
	if (ulRet < 0)
	{
		hr = MAPI_E_CALL_FAILED;
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error writing xml data");
		goto exit;
	}
	
	//<multistatus>
	m_mapNs[CALDAVNS] = "";
	m_mapNs.erase("DAV:");
	xmlTextWriterStartElementNS	(xmlWriter ,NULL ,(const xmlChar *)"multistatus" ,(const xmlChar *)"DAV:" );
	strNsPrefix = "C";

	//write all xmlname spaces in main tag.
	for(iterMapNS = m_mapNs.begin(); iterMapNS != m_mapNs.end(); iterMapNS++)
	{
		std::string strprefix;
		strNs = iterMapNS->first;
		if(strNs.empty())
			continue;
		hr = RegisterNs(strNs,&strNsPrefix);
		m_mapNs[strNs] = strNsPrefix;
		strprefix = "xmlns:" + strNsPrefix;
		
		xmlTextWriterWriteAttribute	(xmlWriter,
									(const xmlChar *) strprefix.c_str(),
									(const xmlChar *) strNs.c_str());
	}
	// <response>
	iterResp = sDavMStatus->lstResp.begin();
	for (int i=0; sDavMStatus->lstResp.end() != iterResp; i++,iterResp++)
	{
		WEBDAVRESPONSE sDavResp;
		sDavResp = *iterResp;
		hr = HrWriteSResponse(xmlWriter, &strNsPrefix, sDavResp);
		if(hr != hrSuccess)
			goto exit;
	}

	//</multistatus>
	ulRet = xmlTextWriterEndElement(xmlWriter);
	//EOF
	ulRet = xmlTextWriterEndDocument(xmlWriter);
	// force write to buffer
	ulRet = xmlTextWriterFlush(xmlWriter);

	strXml->assign((char *)xmlBuff->content, xmlBuff->use);

exit:
	if (xmlWriter)
		xmlFreeTextWriter(xmlWriter);

	if (xmlBuff)
		xmlBufferFree(xmlBuff);

	return hr;
}

/* REPORT request example.
 *
 *	<calendar-query xmlns:D="DAV:" xmlns="urn:ietf:params:xml:ns:caldav">
 *		<D:prop>
 *			<D:getetag/>
 *		</D:prop>
 *		<filter>
 *			<comp-filter name="VCALENDAR">
 *				<comp-filter name="VEVENT"/>
 *			</comp-filter>
 *		</filter>
 *	</calendar-query>
 *
 *******************************  OR *************************************
 *		
 *	<calendar-multiget xmlns:D="DAV:" xmlns="urn:ietf:params:xml:ns:caldav">
 *		<D:prop>
 *			<D:getetag/>
 *			<calendar-data/>
 *		</D:prop>
 *		<D:href>/caldav/user/path-to-calendar/d8516650-b6fc-11dd-92a5-a494cf95cb3a.ics</D:href>
 *	</calendar-multiget>
 */	
HRESULT WebDav::HrReport()
{
	HRESULT hr = hrSuccess;
	xmlNode * lpXmlNode = NULL;
	
	hr = HrParseXml();
	if (hr != hrSuccess)
		goto exit;

	lpXmlNode = xmlDocGetRootElement(m_lpXmlDoc);
	if (!lpXmlNode)
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	if (lpXmlNode->name && !xmlStrcmp(lpXmlNode->name, (const xmlChar *)"calendar-query") )
	{
		//CALENDAR-QUERY
		//Retrieves the list of GUIDs
		hr = HrHandleRptCalQry();

	}
	else if (lpXmlNode->name && !xmlStrcmp(lpXmlNode->name, (const xmlChar *)"calendar-multiget") )
	{
		//MULTIGET
		//Retrieves Ical data for each GUID that client requests
		hr = HrHandleRptMulGet();
	}
	else
		goto exit;

exit:
	return hr;
}

HRESULT WebDav::HrHandleRptCalQry()
{
	HRESULT hr = hrSuccess;
	xmlNode * lpXmlNode = NULL;
	xmlNode * lpXmlChildNode = NULL;
	xmlAttr * lpXmlAttr = NULL;
	WEBDAVREQSTPROPS sReptQuery;
	WEBDAVMULTISTATUS sWebMStatus;
	std::string strXml;

	lpXmlNode = xmlDocGetRootElement(m_lpXmlDoc);
	if (!lpXmlNode)
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	//REPORT CALENDAR-QUERY.
	//xmldata to Structures.
	sReptQuery.sPropName.strPropname.assign((const char*) lpXmlNode->name);
	//HrSetDavPropName(&(sReptQuery.sPropName),lpXmlNode);
	lpXmlNode = lpXmlNode->children;
	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"prop"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	if(lpXmlNode->ns && lpXmlNode->ns->href)
		sReptQuery.sPropName.strNS.assign((const char*) lpXmlNode->ns->href);
	else
		sReptQuery.sPropName.strNS.assign(CALDAVNSDEF);

	HrSetDavPropName(&(sReptQuery.sProp.sPropName),lpXmlNode);
	lpXmlChildNode = lpXmlNode->children;
	while (lpXmlChildNode)
	{
		//properties requested by client.
		WEBDAVPROPERTY sWebProperty;
		
		HrSetDavPropName(&(sWebProperty.sPropName),lpXmlChildNode);
		sReptQuery.sProp.lstProps.push_back(sWebProperty);
		lpXmlChildNode = lpXmlChildNode->next;
	}

	lpXmlNode = lpXmlNode->next;
	if (lpXmlNode && lpXmlNode->children)
	{
		HrSetDavPropName(&(sReptQuery.sFilter.sPropName),lpXmlNode);
		lpXmlNode = lpXmlNode->children;
	}
	else
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	lpXmlAttr = lpXmlNode->properties;
	if (lpXmlAttr && lpXmlAttr->children)
		lpXmlChildNode = lpXmlAttr->children;
	else
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}
	
	if (!lpXmlChildNode || !lpXmlChildNode->content || xmlStrcmp(lpXmlChildNode->content, (const xmlChar *)"VCALENDAR"))
	{
		hr = E_FAIL;
		goto exit;	
	}

	sReptQuery.sFilter.lstFilters.push_back((char *)lpXmlChildNode->content);

	lpXmlNode = lpXmlNode->children;
	if (lpXmlNode && lpXmlNode->properties->children)
	{
		lpXmlAttr = lpXmlNode->properties;
		lpXmlChildNode = lpXmlAttr->children;
	}
	else
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	if (lpXmlChildNode && lpXmlChildNode->content && !xmlStrcmp( lpXmlChildNode->content, (const xmlChar *)"VTODO"))
	{
		//No support for VTODO events currently. (see CalDAV::HrListCalEntries)
		sReptQuery.sFilter.lstFilters.push_back((char *)lpXmlChildNode->content);
	}
	else if (lpXmlChildNode && lpXmlChildNode->content && !xmlStrcmp(lpXmlChildNode->content, (const xmlChar *)"VEVENT"))
	{
		sReptQuery.sFilter.lstFilters.push_back((char *)lpXmlChildNode->content);
	} 
	else
	{
		// unknown/unsupported request
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	//Retrieve Data from the server & return WEBMULTISTATUS structure.
	hr = HrListCalEntries(&sReptQuery, &sWebMStatus);
	if (hr != hrSuccess)
		goto exit;

	//Convert WEBMULTISTATUS structure to xml data.
	hr = RespStructToXml(&sWebMStatus, &strXml);
	if (hr != hrSuccess)
		 goto exit;

	//Respond to the client with xml data.
	m_lpRequest->HrResponseHeader(207, "Multi-Status");
	m_lpRequest->HrResponseBody(strXml);

exit:
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Unable to process report calendar query: 0x%08X", hr);
		m_lpRequest->HrResponseHeader(500, "Internal Server Error");
	}

	return hr;
}

HRESULT WebDav::HrHandleRptMulGet()
{
	HRESULT hr = hrSuccess;
	WEBDAVRPTMGET sRptMGet;
	WEBDAVMULTISTATUS sWebMStatus;
	xmlNode * lpXmlNode = NULL;
	xmlNode * lpXmlChildNode = NULL;
	xmlNode * lpXmlContentNode = NULL;
	std::string strXml;
	
	lpXmlNode = xmlDocGetRootElement(m_lpXmlDoc);
	if (!lpXmlNode)
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	//REPORT Multiget Request.
	// xml data to structures
	HrSetDavPropName(&(sRptMGet.sPropName),lpXmlNode);
	lpXmlNode = lpXmlNode->children;
	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"prop"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	HrSetDavPropName(&(sRptMGet.sProp.sPropName),lpXmlNode);
	lpXmlChildNode = lpXmlNode->children;
	while (lpXmlChildNode)
	{
		//Reqeuested properties of the client.
		WEBDAVPROPERTY sWebProperty;
	
		HrSetDavPropName(&(sWebProperty.sPropName),lpXmlChildNode);
		sRptMGet.sProp.lstProps.push_back(sWebProperty);
		lpXmlChildNode = lpXmlChildNode->next;
	}
	
	if (lpXmlNode->next)
		lpXmlChildNode = lpXmlNode->next;
	else
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	while (lpXmlChildNode)
	{	
		//List of ALL GUIDs whose ical data is requested.
		WEBDAVVALUE sWebVal;
		std::string strGuid;
		size_t found = 0;
		lpXmlContentNode = lpXmlChildNode->children;

		HrSetDavPropName(&(sWebVal.sPropName),lpXmlChildNode);
		strGuid.assign((const char *) lpXmlContentNode->content);

		found = strGuid.rfind("/") + 1;
		if(found == string::npos)
			continue;

		//GUID is of the form "00349543499823933AB332455634C.ics"
		//so remove the last characters ".ics" to get only GUID values
		strGuid = strGuid.substr(found,strGuid.size() - found -4);
		sWebVal.strValue.assign((const char *)strGuid.c_str());

		sRptMGet.lstWebVal.push_back(sWebVal);
		lpXmlChildNode = lpXmlChildNode->next;
	}

	//Retrieve Data from the Server and return WEBMULTISTATUS structure.
	hr = HrHandleReport(&sRptMGet, &sWebMStatus);
	if (hr != hrSuccess)
		goto exit;

	//Convert WEBMULTISTATUS structure to xml data
	hr = RespStructToXml(&sWebMStatus, &strXml);
	if (hr != hrSuccess)
		goto exit;

	m_lpRequest->HrResponseHeader(207, "Multi-Status");
	m_lpRequest->HrResponseBody(strXml);

exit:
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Unable to process report multi-get: 0x%08X", hr);
		m_lpRequest->HrResponseHeader(500, "Internal Server Error");
	}
	
	return hr;
}

HRESULT WebDav::WriteData(xmlTextWriterPtr xmlWriter, WEBDAVVALUE sWebVal, std::string * szNsPrefix)
{
	std::string strNs;
	int ulRet = 0;
	HRESULT hr = hrSuccess;

	strNs = sWebVal.sPropName.strNS;
	if(strNs == "DAV:")
		strNs = "";
	if(strNs.empty())
	{
		ulRet = xmlTextWriterWriteElement(xmlWriter,
										  (const xmlChar *)sWebVal.sPropName.strPropname.c_str(),
										  (const xmlChar *)sWebVal.strValue.c_str());
		goto exit;
	}

	//Retrieve the namespace prefix if present in map.
	hr = GetNs(szNsPrefix, &strNs);
	//if namespace is not present in the map then insert it into map.
	if (hr != hrSuccess)
		hr = RegisterNs(strNs, szNsPrefix);	

	if (hr != hrSuccess)
		goto exit;
	
	/*Write xml none of the form
	 *	<D:href>/caldav/user/calendar/entryGUID.ics</D:href>
	 */
	ulRet =	xmlTextWriterWriteElementNS	(xmlWriter,
										(const xmlChar *)szNsPrefix->c_str(),
										(const xmlChar *)sWebVal.sPropName.strPropname.c_str(),
										(const xmlChar *)(strNs.empty() ? NULL : strNs.c_str()),
										(const xmlChar *)sWebVal.strValue.c_str());
	if (ulRet == -1) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	
exit:
	
	return hr;
}

HRESULT WebDav::WriteNode(xmlTextWriterPtr xmlWriter, WEBDAVPROPNAME sWebPropName, std::string * lpstrNsPrefix)
{
	std::string strNs;
	int ulRet = 0;
	HRESULT hr = hrSuccess;

	strNs = sWebPropName.strNS;
	if(strNs == "DAV:")
		strNs = "";
	if(strNs.empty())
	{
		ulRet = xmlTextWriterStartElement	(xmlWriter,
											(const xmlChar *)sWebPropName.strPropname.c_str());
		goto exit;
	}

	hr = GetNs(lpstrNsPrefix, &strNs);
	if (hr != hrSuccess)
	   hr =	RegisterNs(strNs, lpstrNsPrefix);

	if (hr != hrSuccess)
		goto exit;
	/*Write Xml data of the form
	 * <D:propstat>
	 * the end tag </D:propstat> is written by "xmlTextWriterEndElement(xmlWriter)"
	 */
	ulRet =	xmlTextWriterStartElementNS (xmlWriter,
										(const xmlChar *)lpstrNsPrefix->c_str(),
										(const xmlChar *)sWebPropName.strPropname.c_str(),
										(const xmlChar *)(strNs.empty() ? NULL : strNs.c_str()));

	if (ulRet == -1) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	
	if (!sWebPropName.strPropAttribName.empty())
		ulRet = xmlTextWriterWriteAttribute(xmlWriter, (const xmlChar *)sWebPropName.strPropAttribName.c_str(), (const xmlChar *) sWebPropName.strPropAttribValue.c_str());
	
exit:
	return hr;
}

HRESULT WebDav::RegisterNs(std::string strNs, std::string * lpstrNsPrefix)
{
	(*lpstrNsPrefix)[0]++;
	m_mapNs[strNs] = *lpstrNsPrefix;

	return hrSuccess;
}

HRESULT WebDav::GetNs(std::string * lpstrPrefx, std::string *lpstrNs)
{
	HRESULT hr = hrSuccess;
	map <std::string,std::string>::iterator itMpNs;

	itMpNs = m_mapNs.find(*lpstrNs);
	if (itMpNs != m_mapNs.end())
	{
		lpstrPrefx->assign(itMpNs->second);
		lpstrNs->clear();
	}
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}

HRESULT WebDav::HrWriteSResponse(xmlTextWriterPtr xmlWriter,std::string *lpstrNsPrefix ,WEBDAVRESPONSE sResponse)
{
	HRESULT hr = hrSuccess;
	WEBDAVRESPONSE sWebResp;
	std::string strNsPrefix;
	std::list<WEBDAVPROPSTAT>::iterator iterPropStat;
	int ulRet = 0;

	sWebResp = sResponse;
	//<response>
	hr = WriteNode(xmlWriter, sWebResp.sPropName,lpstrNsPrefix);
	
	//<href>xxxxxxxxxxxxxxxx</href>
	hr = WriteData(xmlWriter,sWebResp.sHRef,lpstrNsPrefix);

	iterPropStat = sWebResp.lstsPropStat.begin();
	while(sWebResp.lstsPropStat.end() != iterPropStat)
	{
		WEBDAVPROPSTAT sDavPropStat;
		sDavPropStat = *iterPropStat;
		HrWriteSPropStat(xmlWriter, lpstrNsPrefix, sDavPropStat);
		iterPropStat++;
	}

	//</response>
	ulRet = xmlTextWriterEndElement(xmlWriter);

	return hr;
}

HRESULT WebDav::HrWriteSPropStat(xmlTextWriterPtr xmlWriter, std::string *lpstrNsPrefix, WEBDAVPROPSTAT lpsPropStat)
{
	HRESULT hr = hrSuccess;
	WEBDAVPROPSTAT sWebPropStat;
	WEBDAVPROP sWebProp;
	int ulRet = 0;
	std::list<WEBDAVPROPERTY>::iterator iterProp;
	
	sWebPropStat = lpsPropStat;
	//<propstat>
	hr = WriteNode(xmlWriter, sWebPropStat.sPropName,lpstrNsPrefix);
	
	sWebProp = sWebPropStat.sProp;
	//<prop>
	hr = WriteNode(xmlWriter, sWebProp.sPropName,lpstrNsPrefix);

	iterProp = sWebProp.lstProps.begin();
	//loop	for properties list
	while (sWebProp.lstProps.end() != iterProp)
	{			
		WEBDAVPROPERTY sWebProperty;
		sWebProperty = *iterProp;
		
		if (!sWebProperty.strValue.empty())
		{
			WEBDAVVALUE sWebVal;
			sWebVal.sPropName = sWebProperty.sPropName;
			sWebVal.strValue = sWebProperty.strValue;
			//<getctag xmlns="xxxxxxxxxxx">xxxxxxxxxxxxxxxxx</getctag>
			WriteData(xmlWriter,sWebVal,lpstrNsPrefix);
		}
		else
		{	//<resourcetype>
			WriteNode(xmlWriter, sWebProperty.sPropName,lpstrNsPrefix);
		}
		
		if(!sWebProperty.lstItems.empty())
			HrWriteItems(xmlWriter, lpstrNsPrefix, &sWebProperty);

        //loop for sub properties
		for (int k = 0; !sWebProperty.lstValues.empty(); k++)
		{
			WEBDAVVALUE sWebVal;
			sWebVal = sWebProperty.lstValues.front();
			//<collection/>
			if (!sWebVal.strValue.empty())
				WriteData(xmlWriter,sWebVal,lpstrNsPrefix);
			else
			{
				WriteNode(xmlWriter, sWebVal.sPropName, lpstrNsPrefix);
				ulRet = xmlTextWriterEndElement(xmlWriter);
			}
			sWebProperty.lstValues.pop_front();
		}
		//end tag if started
		//</resourcetype>
		if (sWebProperty.strValue.empty())
			ulRet = xmlTextWriterEndElement(xmlWriter);

		iterProp++;
	}
	//</prop>
	ulRet = xmlTextWriterEndElement(xmlWriter);
	
	//<status xmlns="xxxxxxx">HTTP/1.1 200 OK</status>
	hr = WriteData(xmlWriter,sWebPropStat.sStatus,lpstrNsPrefix);

	//</propstat>
	ulRet = xmlTextWriterEndElement(xmlWriter);

	return hr;
}

//Write ACL request.
HRESULT WebDav::HrWriteItems(xmlTextWriterPtr xmlWriter, std::string *lpstrNsPrefix,WEBDAVPROPERTY *lpsWebProperty)
{
	HRESULT hr = hrSuccess;
	WEBDAVITEM sDavItem;
	ULONG ulDepthPrev = 0;
	ULONG ulDepthCur = 0;
	bool blFirst = true;

	ulDepthPrev = lpsWebProperty->lstItems.front().ulDepth;
	while(!lpsWebProperty->lstItems.empty())
	{
		sDavItem = lpsWebProperty->lstItems.front();
		ulDepthCur = sDavItem.ulDepth;
		
		if(ulDepthCur >= ulDepthPrev)
		{
			if(ulDepthCur == ulDepthPrev && !blFirst && sDavItem.sDavValue.sPropName.strPropname != "ace")
			{
				xmlTextWriterEndElement(xmlWriter);
			}
			blFirst = false;

			if (!sDavItem.sDavValue.strValue.empty())
			{
				WriteData(xmlWriter,sDavItem.sDavValue,lpstrNsPrefix);
				sDavItem.ulDepth = sDavItem.ulDepth - 1;
			}
			else
			{
				//<resourcetype>
				WriteNode(xmlWriter, sDavItem.sDavValue.sPropName,lpstrNsPrefix);
			}
		}
		else
		{
			for(ULONG i = ulDepthCur ; i <= ulDepthPrev ; i++)
			{
			 xmlTextWriterEndElement(xmlWriter);
			}
			
			if (!sDavItem.sDavValue.strValue.empty())
			{
				WriteData(xmlWriter,sDavItem.sDavValue,lpstrNsPrefix);
				sDavItem.ulDepth = sDavItem.ulDepth - 1;
			}
			else
			{
				//<resourcetype>
				WriteNode(xmlWriter, sDavItem.sDavValue.sPropName,lpstrNsPrefix);
			}
		}
		ulDepthPrev = sDavItem.ulDepth;
		lpsWebProperty->lstItems.pop_front();
	}

	for(ULONG i = 0 ; i <= ulDepthPrev ; i++)
	{
		xmlTextWriterEndElement(xmlWriter);
	}

	return hr;
}

HRESULT WebDav::HrSetDavPropName(WEBDAVPROPNAME *lpsDavPropName,xmlNode *lpXmlNode)
{
	lpsDavPropName->strPropname.assign((const char*)lpXmlNode->name);
	lpsDavPropName->strNS.assign((const char*)lpXmlNode->ns->href);
	m_mapNs[lpsDavPropName->strNS] = "";
	return hrSuccess;
}

HRESULT WebDav::HrSetDavPropName(WEBDAVPROPNAME *lpsDavPropName,std::string strPropName,std::string strNs)
{
	lpsDavPropName->strPropname.assign(strPropName);
	lpsDavPropName->strNS.assign(strNs);
	m_mapNs[lpsDavPropName->strNS] = "";
	return hrSuccess;
}

HRESULT WebDav::HrSetDavPropName(WEBDAVPROPNAME *lpsDavPropName,std::string strPropName, std::string strPropAttribName, std::string strPropAttribValue, std::string strNs)
{
	lpsDavPropName->strPropname.assign(strPropName);
	lpsDavPropName->strNS.assign(strNs);
	lpsDavPropName->strPropAttribName = strPropAttribName;
	lpsDavPropName->strPropAttribValue = strPropAttribValue;
	m_mapNs[lpsDavPropName->strNS] = "";
	return hrSuccess;
}

HRESULT WebDav::HrPropPatch()
{
	HRESULT hr = hrSuccess;	
	WEBDAVPROP sDavProp;
	WEBDAVMULTISTATUS sDavMStatus;
	std::string strFldPath;
	std::string strXml;
	xmlNode * lpXmlNode = NULL;

	//libxml parser parses the xml data and returns the DomTree pointer.
	hr = HrParseXml();
	if (hr != hrSuccess)
		goto exit;

	lpXmlNode = xmlDocGetRootElement(m_lpXmlDoc);
	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"propertyupdate"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	lpXmlNode = lpXmlNode->children;
	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"set"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	lpXmlNode = lpXmlNode->children;
	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"prop"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	if(lpXmlNode->ns && lpXmlNode->ns->href)
		HrSetDavPropName(&(sDavProp.sPropName),"prop",(char *)lpXmlNode->ns->href);
	else
		HrSetDavPropName(&(sDavProp.sPropName),"prop", CALDAVNSDEF);

	lpXmlNode = lpXmlNode->children;
	while(lpXmlNode)
	{
		WEBDAVPROPERTY sProperty;

		if(lpXmlNode->ns && lpXmlNode->ns->href)
			HrSetDavPropName(&(sProperty.sPropName),(char *)lpXmlNode->name,(char*)lpXmlNode->ns->href);
		else
			HrSetDavPropName(&(sProperty.sPropName),(char *)lpXmlNode->name, CALDAVNSDEF);

		if (lpXmlNode->children && lpXmlNode->children->content) {
			sProperty.strValue = (char *)lpXmlNode->children->content;
		}

		sDavProp.lstProps.push_back(sProperty);

		lpXmlNode = lpXmlNode->next;
	}	

	hr = HrHandlePropPatch(&sDavProp);

exit:

	if(hr == MAPI_E_NO_ACCESS)
	{
		m_lpRequest->HrResponseHeader(403,"Forbidden");
		m_lpRequest->HrResponseBody("This Folder Cannot be Modified");
	}
	else if (hr == MAPI_E_COLLISION)
	{
		m_lpRequest->HrResponseHeader(409,"Conflict");
		m_lpRequest->HrResponseBody("A folder same name already exist");
	}
	else if(hr != hrSuccess)
	{
		m_lpRequest->HrResponseHeader(404,"Not Found");
	}
	else
	{
		m_lpRequest->HrResponseHeader(200,"OK");
	}

	return hr;
}

HRESULT WebDav::HrMkCalendar()
{
	HRESULT hr = hrSuccess;
	xmlNode * lpXmlNode = NULL;
	WEBDAVPROP sDavProp;

	hr = HrParseXml();
	if(hr != hrSuccess)
	{
		goto exit;
		m_lpLogger->Log(EC_LOGLEVEL_ERROR,"Parsing Error For MKCALENDAR");
	}
	
	lpXmlNode = xmlDocGetRootElement(m_lpXmlDoc);
	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"mkcalendar"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}
	
	lpXmlNode = lpXmlNode->children;
	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"set"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	lpXmlNode = lpXmlNode->children;
	if (!lpXmlNode || !lpXmlNode->name || xmlStrcmp(lpXmlNode->name, (const xmlChar *)"prop"))
	{
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}
	
	if(lpXmlNode->ns && lpXmlNode->ns->href)
		HrSetDavPropName(&(sDavProp.sPropName),(char*)lpXmlNode->name,(char*)lpXmlNode->ns->href);
	else
		HrSetDavPropName(&(sDavProp.sPropName),(char*)lpXmlNode->name,CALDAVNSDEF);

	lpXmlNode = lpXmlNode->children;
	while(lpXmlNode)
	{
		WEBDAVPROPERTY sProperty;

		if (lpXmlNode->ns && lpXmlNode->ns->href)
			HrSetDavPropName(&(sProperty.sPropName),(char*)lpXmlNode->name,(char*)lpXmlNode->ns->href);
		else
			HrSetDavPropName(&(sProperty.sPropName),(char*)lpXmlNode->name,CALDAVNSDEF);

		if (lpXmlNode->children && lpXmlNode->children->content)
			sProperty.strValue = (char*)lpXmlNode->children->content;
		sDavProp.lstProps.push_back(sProperty);	

		lpXmlNode = lpXmlNode->next;
	}

	hr = HrHandleMkCal(&sDavProp);

exit:

	if(hr == MAPI_E_COLLISION)
	{
		m_lpRequest->HrResponseHeader(409,"CONFLICT");
		m_lpRequest->HrResponseBody("Folder with same name already exists");
	}
	else if(hr == E_ACCESSDENIED || hr == MAPI_E_ACCESS_DENIED)
		m_lpRequest->HrResponseHeader(403 ,"Forbidden");
	else if(hr == hrSuccess)
		m_lpRequest->HrResponseHeader(201,"Created");
	else
		m_lpRequest->HrResponseHeader(500,"Bad Request");

	return hr;
}
