/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <string>
#include "str1252.h"

#include <mapi.h>
#include <mapidefs.h>
#include <mapiutil.h>
#include <mapitags.h>
#include <mapicode.h>
#include <errno.h>
#include <iconv.h>

#include "ECLogger.h"
#include "CommonUtil.h"
#include "ECTags.h"
#include "ECGuid.h"
#include "Util.h"
#include "stringutil.h"

#include "mapiext.h"

#include "edkguid.h"
#include "edkmdb.h"
#include "IECUnknown.h"
#include "EMSAbTag.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define PROFILEPREFIX		"ec-adm-"

// HACK: prototypes may differ depending on the compiler and/or system (the
// second parameter may or may not be 'const'). This redeclaration is a hack
// to have a common prototype "iconv_cast".
class ICONV_HACK
{
public:
	ICONV_HACK(const char** ptr) : m_ptr(ptr) { }

	// the compiler will choose the right operator
	operator const char**() { return m_ptr; }
	operator char**() { return const_cast <char**>(m_ptr); }

private:
	const char** m_ptr;
};


bool operator ==(SBinary left, SBinary right)
{
	return (left.cb == right.cb && memcmp(left.lpb, right.lpb, left.cb) == 0);
}


char* GetServerUnixSocket(char* szPreferred) {
	char *env = getenv("ZARAFA_SOCKET");
	if (env && env[0] != '\0')
		return env;
	else if (szPreferred && szPreferred[0] != '\0')
		return szPreferred;
	else
		return (char*)CLIENT_ADMIN_SOCKET;
}


// FIXME: replace it to the right place
#define	pbGlobalProfileSectionGuid	"\x13\xDB\xB0\xC8\xAA\x05\x10\x1A\x9B\xB0\x00\xAA\x00\x2F\xC4\x5A"

HRESULT CreateProfileTemp(char *username, char *password, char *path, char* szProfName, ULONG ulProfileFlags,
						  char *sslkey_file, char *sslkey_password) {
	HRESULT hr = hrSuccess;
	LPPROFADMIN	lpProfAdmin = NULL;
	LPSERVICEADMIN lpServiceAdmin = NULL;
	LPSPropValue lpServiceUID = NULL;
	SPropValue sProps[7];	// server, username, password and profile -name and -flags, optional sslkey file with sslkey password
	LPSPropValue lpServiceName = NULL;

	LPMAPITABLE	lpTable = NULL;
	LPSRowSet	lpRows = NULL;
	int i;


//-- create profile
	hr = MAPIAdminProfiles(0, &lpProfAdmin);
	if (hr != hrSuccess)
		goto exit;

	lpProfAdmin->DeleteProfile(szProfName, 0);
	hr = lpProfAdmin->CreateProfile(szProfName,"",0,0);
	if (hr != hrSuccess)
		goto exit;

	hr = lpProfAdmin->AdminServices(szProfName,"",0,0, &lpServiceAdmin);
	if (hr != hrSuccess)
		goto exit;
	
	hr = lpServiceAdmin->CreateMsgService("ZARAFA6","",0,0);
	if (hr != hrSuccess)
		goto exit;

	// Strangely we now have to get the SERVICE_UID for the service we just added from
	// the table. (see MSDN help page of CreateMsgService at the bottom of the page)
	hr = lpServiceAdmin->GetMsgServiceTable(0, &lpTable);

	if(hr != hrSuccess) {
		goto exit;
	}

	// Find the correct row
	while(TRUE) {
		hr = lpTable->QueryRows(1, 0, &lpRows);
		
		if(hr != hrSuccess)
			goto exit;
			
		if(lpRows->cRows != 1)
			break;
	
		lpServiceName = PpropFindProp(lpRows->aRow[0].lpProps, lpRows->aRow[0].cValues, PR_SERVICE_NAME);
		
		if(lpServiceName && strcmp(lpServiceName->Value.lpszA, "ZARAFA6") == 0)
			break;
			
		FreeProws(lpRows);
		lpRows = NULL;
			
	}
	
	if(lpRows->cRows != 1) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	// Get the PR_SERVICE_UID from the row
	lpServiceUID = PpropFindProp(lpRows->aRow[0].lpProps, lpRows->aRow[0].cValues, PR_SERVICE_UID);

	if(!lpServiceUID) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	i = 0;
	sProps[i].ulPropTag = PR_EC_PATH;
	sProps[i].Value.lpszA = path;
	i++;

	sProps[i].ulPropTag = PR_EC_USERNAME;
	sProps[i].Value.lpszA = username;
	i++;

	sProps[i].ulPropTag = PR_EC_USERPASSWORD;
	sProps[i].Value.lpszA = password;
	i++;

	sProps[i].ulPropTag = PR_EC_FLAGS;
	sProps[i].Value.ul = ulProfileFlags;
	i++;

	sProps[i].ulPropTag = PR_PROFILE_NAME;
	sProps[i].Value.lpszA = szProfName;
	i++;

	if (sslkey_file) {
		// always add ssl keys info as we might be redirected to an ssl connection
		sProps[i].ulPropTag = PR_EC_SSLKEY_FILE;
		sProps[i].Value.lpszA = sslkey_file;
		i++;

		if (sslkey_password) {
			sProps[i].ulPropTag = PR_EC_SSLKEY_PASS;
			sProps[i].Value.lpszA = sslkey_password;
			i++;
		}
	}

	hr = lpServiceAdmin->ConfigureMsgService((MAPIUID *)lpServiceUID->Value.bin.lpb, 0, 0, i, sProps);
	if (hr != hrSuccess)
		goto exit;

exit:
	if (lpRows)
		FreeProws(lpRows);
	
	if (lpTable)
		lpTable->Release();
		
	if (lpProfAdmin)
		lpProfAdmin->Release();

	if (lpServiceAdmin)
		lpServiceAdmin->Release();

	return hr;
}

HRESULT DeleteProfileTemp(char *szProfName)
{
	LPPROFADMIN	lpProfAdmin = NULL;
	HRESULT hr = hrSuccess;

	// Get the MAPI Profile administration object
	hr = MAPIAdminProfiles(0, &lpProfAdmin);

	if(hr != hrSuccess)
		goto exit;

	lpProfAdmin->DeleteProfile(szProfName,0);

exit:
	if(lpProfAdmin)
		lpProfAdmin->Release();

	return hr;
}

HRESULT HrOpenECAdminSession(IMAPISession **lppSession, char *szPath, ULONG ulProfileFlags, char *sslkey_file, char *sslkey_password)
{
	return HrOpenECSession(lppSession, ZARAFA_SYSTEM_USER, ZARAFA_SYSTEM_USER, szPath, ulProfileFlags, sslkey_file, sslkey_password);
}

HRESULT HrOpenECSession(IMAPISession **lppSession, char *szUsername, char *szPassword, char *szPath, ULONG ulProfileFlags,
						char *sslkey_file, char *sslkey_password, char *profname)
{
	HRESULT		hr = hrSuccess;
	ULONG		ulProfNum = 0;
	char		*szProfName = new char[strlen(PROFILEPREFIX)+10+1];
	IMAPISession *lpMAPISession = NULL;

	ulProfNum = rand_mt();

	snprintf(szProfName, strlen(PROFILEPREFIX)+10+1, "%s%010u", PROFILEPREFIX, ulProfNum);

	if(!profname)
		profname = szProfName;

	if(szPath != NULL) {

		if(sslkey_file != NULL)
		{
			FILE *ssltest;
			ssltest = fopen(sslkey_file, "r");
			if (!ssltest) {
				// do not pass sslkey if the file does not exists
				// otherwise normal connections do not work either
				sslkey_file = NULL;
				sslkey_password = NULL;
			} else {
				// TODO: test password of certificate
				fclose(ssltest);
			}
		}

		hr = CreateProfileTemp(szUsername, szPassword, szPath, profname, ulProfileFlags, sslkey_file, sslkey_password); 
	} else {
		// these connections cannot be ssl, so no keys needed
		hr = CreateProfileTemp(szUsername, szPassword, GetServerUnixSocket(), profname, ulProfileFlags, NULL, NULL);
	}

	if (hr != hrSuccess)
		goto exit;

	// Log on the the profile
	hr = MAPILogonEx(0, profname, "", MAPI_EXTENDED | MAPI_NEW_SESSION | MAPI_NO_MAIL, &lpMAPISession);

	if(hr != hrSuccess)
		goto exit;

	*lppSession = lpMAPISession;

exit:
	// always try to delete the temporary profile
	DeleteProfileTemp(profname);
		
	if (szProfName)
		delete [] szProfName;

	return hr;
}

HRESULT HrSearchECStoreEntryId(IMAPISession *lpMAPISession, BOOL bPublic, ULONG *lpcbEntryID, LPENTRYID *lppEntryID)
{
	HRESULT			hr = hrSuccess;
	LPSRowSet		lpRows = NULL;
	IMAPITable		*lpStoreTable = NULL;
	LPSPropValue	lpStoreProp = NULL;
	LPSPropValue	lpEntryIDProp = NULL;

	// Get the default store by searching through the message store table and finding the
	// store with PR_MDB_PROVIDER set to the zarafa public store GUID

	hr = lpMAPISession->GetMsgStoresTable(0, &lpStoreTable);
	if(hr != hrSuccess)
		goto exit;

	while(TRUE) {
			hr = lpStoreTable->QueryRows(1, 0, &lpRows);

			if (hr != hrSuccess || lpRows->cRows != 1) {
				hr = MAPI_E_NOT_FOUND;
				break;
			}

			if (bPublic) {
				lpStoreProp = PpropFindProp(lpRows->aRow[0].lpProps,lpRows->aRow[0].cValues, PR_MDB_PROVIDER);
				if (lpStoreProp != NULL && memcmp(lpStoreProp->Value.bin.lpb, &ZARAFA_STORE_PUBLIC_GUID, sizeof(MAPIUID)) == 0 )
					break;
			} else {
				lpStoreProp = PpropFindProp(lpRows->aRow[0].lpProps,lpRows->aRow[0].cValues, PR_RESOURCE_FLAGS);
				if (lpStoreProp != NULL && lpStoreProp->Value.ul & STATUS_DEFAULT_STORE)
					break;
			}

			FreeProws(lpRows);
			lpRows = NULL;
	}

	if (hr != hrSuccess)
		goto exit;

	lpEntryIDProp = PpropFindProp(lpRows->aRow[0].lpProps, lpRows->aRow[0].cValues, PR_ENTRYID);
	if (lpEntryIDProp == NULL) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	// copy entryid so we continue in the same code piece in windows/linux
	hr = Util::HrCopyEntryId(lpEntryIDProp->Value.bin.cb, (LPENTRYID)lpEntryIDProp->Value.bin.lpb, lpcbEntryID, lppEntryID);
	if (hr != hrSuccess)
		goto exit;

exit:
	if (lpRows)
		FreeProws(lpRows);

	if (lpStoreTable)
		lpStoreTable->Release();

	return hr;
}

HRESULT HrOpenDefaultStore(IMAPISession *lpMAPISession, IMsgStore **lppMsgStore) {
	return HrOpenDefaultStore(lpMAPISession, MDB_WRITE | MDB_NO_DIALOG | MDB_NO_MAIL | MDB_TEMPORARY, lppMsgStore);
}

HRESULT HrOpenDefaultStoreOffline(IMAPISession *lpMAPISession, IMsgStore **lppMsgStore)
{
	HRESULT	hr = hrSuccess;
	IMsgStore *lpMsgStore = NULL;
	IMsgStore *lpProxedMsgStore = NULL;

	hr = HrOpenDefaultStore(lpMAPISession, MDB_WRITE | MDB_NO_DIALOG | MDB_NO_MAIL | MDB_TEMPORARY, &lpMsgStore);
	if(hr != hrSuccess)
		goto exit;

	hr = GetProxyStoreObject(lpMsgStore, &lpProxedMsgStore);
	if (hr != hrSuccess)
		goto exit;
	
	hr = lpProxedMsgStore->QueryInterface(IID_ECMsgStoreOffline, (void**)lppMsgStore);
	if(hr != hrSuccess)
		goto exit;

exit:
	if (lpProxedMsgStore)
		lpProxedMsgStore->Release();

	if (lpMsgStore)
		lpMsgStore->Release();

	return hr;
}

HRESULT HrOpenDefaultStoreOnline(IMAPISession *lpMAPISession, IMsgStore **lppMsgStore)
{
	HRESULT	hr = hrSuccess;
	IMsgStore *lpMsgStore = NULL;
	IMsgStore *lpProxedMsgStore = NULL;

	hr = HrOpenDefaultStore(lpMAPISession, MDB_WRITE | MDB_NO_DIALOG | MDB_NO_MAIL | MDB_TEMPORARY, &lpMsgStore);
	if(hr != hrSuccess)
		goto exit;

	hr = GetProxyStoreObject(lpMsgStore, &lpProxedMsgStore);
	if (hr != hrSuccess)
		goto exit;
	
	hr = lpProxedMsgStore->QueryInterface(IID_ECMsgStoreOnline, (void**)lppMsgStore);
	if(hr != hrSuccess)
		goto exit;

exit:
	if (lpProxedMsgStore)
		lpProxedMsgStore->Release();

	if (lpMsgStore)
		lpMsgStore->Release();

	return hr;
}

HRESULT HrOpenStoreOnline(IMAPISession *lpMAPISession, ULONG cbEntryID, LPENTRYID lpEntryID, IMsgStore **lppMsgStore)
{
	HRESULT	hr = hrSuccess;
	IMsgStore *lpMsgStore = NULL;
	IMsgStore *lpProxedMsgStore = NULL;

	if (lpMAPISession == NULL || lppMsgStore == NULL || lpEntryID == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpMAPISession->OpenMsgStore(0, cbEntryID, lpEntryID, &IID_IMsgStore, MDB_WRITE | MDB_NO_DIALOG | MDB_NO_MAIL | MDB_TEMPORARY, &lpMsgStore);
	if (hr != hrSuccess)
		goto exit;

	hr = GetProxyStoreObject(lpMsgStore, &lpProxedMsgStore);
	if (hr != hrSuccess)
		goto exit;
	
	hr = lpProxedMsgStore->QueryInterface(IID_ECMsgStoreOnline, (void**)lppMsgStore);
	if (hr != hrSuccess)
		goto exit;

exit:
	if (lpProxedMsgStore)
		lpProxedMsgStore->Release();

	if (lpMsgStore)
		lpMsgStore->Release();

	return hr;
}

HRESULT GetProxyStoreObject(IMsgStore *lpMsgStore, IMsgStore **lppMsgStore)
{
	HRESULT	hr = hrSuccess;
	IProxyStoreObject *lpProxyStoreObject = NULL;
	IECUnknown* lpECMsgStore = NULL;
	LPSPropValue lpPropValue = NULL;

	if (lpMsgStore == NULL || lppMsgStore == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if ( lpMsgStore->QueryInterface(IID_IProxyStoreObject, (void**)&lpProxyStoreObject) == hrSuccess) {

		hr = lpProxyStoreObject->UnwrapNoRef((LPVOID*)lppMsgStore);
		if (hr != hrSuccess)
			goto exit;

		(*lppMsgStore)->AddRef();

	} else if(HrGetOneProp(lpMsgStore, PR_EC_OBJECT, &lpPropValue) == hrSuccess) {

		lpECMsgStore = (IECUnknown *)lpPropValue->Value.lpszA;
		if (lpECMsgStore == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		hr = lpECMsgStore->QueryInterface(IID_IMsgStore, (void**)lppMsgStore);
	} else {
		// Possible object already wrapped, gives the orignale object back
		(*lppMsgStore) = lpMsgStore;
		(*lppMsgStore)->AddRef();
	}

exit:
	if (lpPropValue)
		MAPIFreeBuffer(lpPropValue);

	if (lpProxyStoreObject)
		lpProxyStoreObject->Release();

	return hr;
}

HRESULT HrOpenDefaultStore(IMAPISession *lpMAPISession, ULONG ulFlags, IMsgStore **lppMsgStore) {
	HRESULT			hr = hrSuccess;
	IMsgStore		*lpMsgStore = NULL;
	ULONG			cbEntryID = 0;
	LPENTRYID		lpEntryID = NULL;

	hr = HrSearchECStoreEntryId(lpMAPISession, FALSE, &cbEntryID, &lpEntryID);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMAPISession->OpenMsgStore(0, cbEntryID, lpEntryID, &IID_IMsgStore, ulFlags, &lpMsgStore);

	if(hr != hrSuccess)
		goto exit;

	*lppMsgStore = lpMsgStore;

exit:
	if(lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	return hr;
}

HRESULT HrOpenECPublicStore(IMAPISession *lpMAPISession, IMsgStore **lppMsgStore){
	return HrOpenECPublicStore(lpMAPISession, MDB_WRITE | MDB_NO_DIALOG | MDB_NO_MAIL | MDB_TEMPORARY, lppMsgStore);
}

HRESULT HrOpenECPublicStoreOnline(IMAPISession *lpMAPISession, IMsgStore **lppMsgStore)
{
	HRESULT	hr = hrSuccess;
	IMsgStore *lpMsgStore = NULL;
	IMsgStore *lpProxedMsgStore = NULL;

	hr = HrOpenECPublicStore(lpMAPISession, MDB_WRITE | MDB_NO_DIALOG | MDB_NO_MAIL | MDB_TEMPORARY, &lpMsgStore);
	if(hr != hrSuccess)
		goto exit;

	hr = GetProxyStoreObject(lpMsgStore, &lpProxedMsgStore);
	if (hr != hrSuccess)
		goto exit;
	
	hr = lpProxedMsgStore->QueryInterface(IID_ECMsgStoreOnline, (void**)lppMsgStore);
	if(hr != hrSuccess)
		goto exit;

exit:
	if (lpProxedMsgStore)
		lpProxedMsgStore->Release();

	if (lpMsgStore)
		lpMsgStore->Release();

	return hr;
}

HRESULT HrOpenECPublicStore(IMAPISession *lpMAPISession, ULONG ulFlags, IMsgStore **lppMsgStore)
{
	HRESULT			hr = hrSuccess;
	IMsgStore		*lpMsgStore = NULL;
	ULONG			cbEntryID = 0;
	LPENTRYID		lpEntryID = NULL;

	hr = HrSearchECStoreEntryId(lpMAPISession, TRUE, &cbEntryID, &lpEntryID);
	if(hr != hrSuccess)
		goto exit;

	hr = lpMAPISession->OpenMsgStore(0, cbEntryID, lpEntryID, &IID_IMsgStore, ulFlags, &lpMsgStore);
	if(hr != hrSuccess)
		goto exit;

	*lppMsgStore = lpMsgStore;

exit:
	if(lpEntryID)
		MAPIFreeBuffer(lpEntryID);
		
	return hr;
}

HRESULT HrGetECProviderAdmin(LPMAPISESSION lpSession, LPPROVIDERADMIN *lppProviderAdmin)
{
	HRESULT			hr = hrSuccess;
	LPSERVICEADMIN	lpMsgServiceAdmin = NULL;
	LPMAPITABLE		lpServiceTable = NULL;
	SRestriction	sRestrict;
	SPropValue		sPropRestrict;
	LPSRowSet		lpsRowSet = NULL;
	LPSPropValue	lpProviderUID = NULL;

	// Get the service admin
	hr = lpSession->AdminServices(0, &lpMsgServiceAdmin);
	if(hr != hrSuccess)
		goto exit;

	//Getdefault profile
	hr = lpMsgServiceAdmin->GetMsgServiceTable(0, &lpServiceTable);
	if(hr != hrSuccess)
		goto exit;
	
	// restrict the table
	sPropRestrict.ulPropTag = PR_SERVICE_NAME;
	sPropRestrict.Value.lpszA = "ZARAFA6";
	sRestrict.res.resContent.ulFuzzyLevel = FL_FULLSTRING;
	sRestrict.res.resContent.ulPropTag = PR_SERVICE_NAME;
	sRestrict.res.resContent.lpProp = &sPropRestrict;
	sRestrict.rt = RES_CONTENT;

	hr = lpServiceTable->Restrict(&sRestrict,0);
	if(hr != hrSuccess)
		goto exit;
	
	//Seek to the end
	hr = lpServiceTable->SeekRow(BOOKMARK_END, -1, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = lpServiceTable->QueryRows(1, 0, &lpsRowSet);
	if(hr != hrSuccess || lpsRowSet == NULL || lpsRowSet->cRows != 1)
	{
		if(hr == hrSuccess) hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	// Get the Service UID
	lpProviderUID = PpropFindProp(lpsRowSet->aRow[0].lpProps, lpsRowSet->aRow[0].cValues, PR_SERVICE_UID);
	if(lpProviderUID == NULL){
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	// Get a admin provider pointer
	hr = lpMsgServiceAdmin->AdminProviders((MAPIUID *)lpProviderUID->Value.bin.lpb, 0, lppProviderAdmin);
	if(hr != hrSuccess)
		goto exit;


exit:
	if(lpServiceTable)
		lpServiceTable->Release();

	if(lpMsgServiceAdmin)
		lpMsgServiceAdmin->Release();

	if(lpsRowSet)
		FreeProws(lpsRowSet);

	return hr;
}

HRESULT HrRemoveECMailBox(LPMAPISESSION lpSession, LPMAPIUID lpsProviderUID)
{
	HRESULT			hr = hrSuccess;
	LPPROVIDERADMIN lpProviderAdmin = NULL;

	hr = HrGetECProviderAdmin(lpSession, &lpProviderAdmin);
	if(hr != hrSuccess)
		goto exit;

	hr = HrRemoveECMailBox(lpProviderAdmin, lpsProviderUID);
	if(hr != hrSuccess)
		goto exit;

exit:
	if(lpProviderAdmin)
		lpProviderAdmin->Release();

	return hr;
}

HRESULT HrRemoveECMailBox(LPPROVIDERADMIN lpProviderAdmin, LPMAPIUID lpsProviderUID)
{
	HRESULT			hr = hrSuccess;
	
	LPPROFSECT		lpGlobalProfSect = NULL;

	LPSPropTagArray	lpsPropTagArray = NULL;
	LPSPropValue	lpGlobalProps = NULL;	
	LPSPropValue	lpNewProp = NULL;
	ULONG			cValues = 0;
	ULONG			cSize = 0;
	unsigned int	i = 0;

	//Open global profile, add the store.(for show list, delete etc)
	hr = lpProviderAdmin->OpenProfileSection((LPMAPIUID)pbGlobalProfileSectionGuid, NULL, MAPI_MODIFY , &lpGlobalProfSect);
	if(hr != hrSuccess)
		goto exit;

	// Allocate array of Proptags 
	cValues = 1;
	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cValues), (void **)&lpsPropTagArray);
	if(hr != hrSuccess)
		goto exit;

	//Fill array
	lpsPropTagArray->aulPropTag[0] = PR_STORE_PROVIDERS;
	lpsPropTagArray->cValues = cValues;

	// The the prop value PR_STORE_PROVIDERS
	hr = lpGlobalProfSect->GetProps(lpsPropTagArray, 0, &cValues, &lpGlobalProps);
	if(hr == hrSuccess && lpGlobalProps->Value.bin.cb >= sizeof(MAPIUID)) 
	{
		hr = MAPIAllocateBuffer(sizeof(SPropValue), (void**)&lpNewProp);
		if(hr != hrSuccess)
			goto exit;

		cSize = lpGlobalProps->Value.bin.cb - sizeof(MAPIUID);

		hr = MAPIAllocateMore(cSize, lpNewProp, (void**)&lpNewProp->Value.bin.lpb);
		if(hr != hrSuccess)
			goto exit;

		lpNewProp->Value.bin.cb	= 0;
		lpNewProp->ulPropTag	= PR_STORE_PROVIDERS;

		for(i=0; i < lpGlobalProps->Value.bin.cb / sizeof(MAPIUID); i++)
		{
			if(memcmp(lpGlobalProps->Value.bin.lpb+(sizeof(MAPIUID) * i), lpsProviderUID, sizeof(MAPIUID)) != 0)
			{
				memcpy(lpNewProp->Value.bin.lpb+lpNewProp->Value.bin.cb, lpGlobalProps->Value.bin.lpb+(sizeof(MAPIUID) * i), sizeof(MAPIUID));
				lpNewProp->Value.bin.cb += sizeof(MAPIUID);
			}
		}

		hr = lpGlobalProfSect->SetProps(1, lpNewProp, NULL);
		if(hr != hrSuccess)
			goto exit;

		hr = lpGlobalProfSect->SaveChanges(0);
		if(hr != hrSuccess)
			goto exit;	
	}

	
	if(lpGlobalProfSect){ lpGlobalProfSect->Release(); lpGlobalProfSect = NULL; }

	//Remove Store
	hr = lpProviderAdmin->DeleteProvider(lpsProviderUID);
	//FIXME: unknown error 0x80070005 by delete (HACK)
	hr = hrSuccess;
	if(hr != hrSuccess)
		goto exit;

exit:

	if(lpGlobalProfSect)
		lpGlobalProfSect->Release();

	if(lpsPropTagArray)
		MAPIFreeBuffer(lpsPropTagArray);

	if(lpGlobalProps)
		MAPIFreeBuffer(lpGlobalProps);

	if(lpNewProp)
		MAPIFreeBuffer(lpNewProp);

	return hr;
}

HRESULT HrAddECMailBox(LPMAPISESSION lpSession, char * lpszUserName)
{
	HRESULT			hr = hrSuccess;
	LPPROVIDERADMIN lpProviderAdmin = NULL;

	hr = HrGetECProviderAdmin(lpSession, &lpProviderAdmin);
	if(hr != hrSuccess)
		goto exit;

	hr = HrAddECMailBox(lpProviderAdmin, lpszUserName);
	if(hr != hrSuccess)
		goto exit;

exit:
	if(lpProviderAdmin)
		lpProviderAdmin->Release();
	
	return hr;
}

HRESULT HrAddECMailBox(LPPROVIDERADMIN lpProviderAdmin, char * lpszUserName)
{
	HRESULT			hr = hrSuccess;
	MAPIUID			sNewProfileUID;
	ULONG			csNewMapiUID = 0;
	ULONG			cValues = 0;
	int				i = 0;

	LPSPropTagArray	lpsPropTagArray = NULL;
	LPPROFSECT		lpGlobalProfSect = NULL;
	LPSPropValue	lpProps = NULL;
	LPSPropValue	lpGlobalProps = NULL;	
	LPSPropValue	lpNewProp = NULL;
	

	cValues = 3;
	MAPIAllocateBuffer( sizeof(SPropValue) * cValues, (void**)&lpProps);

	memset(lpProps, 0, sizeof(SPropValue) * cValues);

	i = 0;

	lpProps[i].ulPropTag = PR_EC_USERNAME;
	lpProps[i++].Value.lpszA = lpszUserName;

	// Create the profile, now the profile is shown in outlook
	hr = lpProviderAdmin->CreateProvider((TCHAR*)"ZARAFA6_MSMDB_Delegate", i, lpProps, 0, 0, &sNewProfileUID);
	if(hr != hrSuccess)
		goto exit;

	if(lpProps){ MAPIFreeBuffer(lpProps); lpProps = NULL; }

	//Open global profile, add the store.(for show list, delete etc)
	hr = lpProviderAdmin->OpenProfileSection((LPMAPIUID)pbGlobalProfileSectionGuid, NULL, MAPI_MODIFY , &lpGlobalProfSect);
	if(hr != hrSuccess)
		goto exit;

	// Allocate array of Proptags 
	cValues = 1;
	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cValues), (void **)&lpsPropTagArray);
	if(hr != hrSuccess)
		goto exit;

	//Fill array
	lpsPropTagArray->aulPropTag[0] = PR_STORE_PROVIDERS;
	lpsPropTagArray->cValues = cValues;

	// The the prop value PR_STORE_PROVIDERS
	hr = lpGlobalProfSect->GetProps(lpsPropTagArray, 0, &cValues, &lpGlobalProps);
	if(HR_FAILED(hr))
		goto exit;

	hr = hrSuccess;

	if(lpGlobalProps->ulPropTag != PR_STORE_PROVIDERS)
		lpGlobalProps->Value.bin.cb = 0;

	// The new size of stores provider uid
	csNewMapiUID = lpGlobalProps->Value.bin.cb + sizeof(MAPIUID); //lpNewProfileUID

	hr = MAPIAllocateBuffer(sizeof(SPropValue), (void**)&lpNewProp);
	if(hr != hrSuccess)
		goto exit;

	hr = MAPIAllocateMore(csNewMapiUID, lpNewProp, (void**)&lpNewProp->Value.bin.lpb);
	if(hr != hrSuccess)
		goto exit;

	lpNewProp->Value.bin.cb = csNewMapiUID;
	lpNewProp->ulPropTag = PR_STORE_PROVIDERS;

	if(lpGlobalProps->Value.bin.cb > 0)
		memcpy(lpNewProp->Value.bin.lpb, lpGlobalProps->Value.bin.lpb, lpGlobalProps->Value.bin.cb);

	memcpy(lpNewProp->Value.bin.lpb+lpGlobalProps->Value.bin.cb, &sNewProfileUID, sizeof(MAPIUID));

	hr = lpGlobalProfSect->SetProps(1, lpNewProp, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = lpGlobalProfSect->SaveChanges(0);
	if(hr != hrSuccess)
		goto exit;

exit:
	if(lpsPropTagArray)
		MAPIFreeBuffer(lpsPropTagArray);

	if(lpsPropTagArray)
		MAPIFreeBuffer(lpsPropTagArray);

	if(lpGlobalProfSect)
		lpGlobalProfSect->Release();

	if(lpProps)
		MAPIFreeBuffer(lpProps);

	if(lpGlobalProps)
		MAPIFreeBuffer(lpGlobalProps);

	if(lpNewProp)
		MAPIFreeBuffer(lpNewProp);

	return hr;
}

HRESULT ECCreateOneOff(char* lpszName, char* lpszAdrType, char* lpszAddress, ULONG ulFlags, ULONG* lpcbEntryID, LPENTRYID* lppEntryID)
{
	HRESULT		hr = hrSuccess;
	MAPIUID		muidOneOff = {MAPI_ONE_OFF_UID};
	ULONG		cbEntryID = 0;
	LPBYTE		lpEntryID = NULL;
	
	WCHAR		wszDisplayName[256];
	WCHAR		wszAdrType[256];
	WCHAR		wszAddress[256];
	unsigned int ulSizeDisplayname = 0;
	unsigned int ulSizeAdrType = 0;
	unsigned int ulSizeAddress = 0;
	unsigned int offset = 0;
	ULONG		ulFlagsOneOff = 0;

	//Check Flags
	if( (ulFlags &~(MAPI_UNICODE | MAPI_SEND_NO_RICH_INFO)) != 0 ||
		lpszAdrType == NULL || lpszAddress == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Calculate object size
	cbEntryID = 4; // abFlags
	cbEntryID+= sizeof(MAPIUID);						// muid
	cbEntryID+= 2;										// Version
	cbEntryID+= 2;										// Flags	
	
	// DisplayName
	if(lpszName) ulSizeDisplayname = strlen(lpszName)+1; 
	else ulSizeDisplayname = 1;
		
	ulSizeAdrType = strlen(lpszAdrType)+1;// AddressType	
	ulSizeAddress = strlen(lpszAddress)+1;// EmailAddress

	// Check for unicode
	if(ulFlags & MAPI_UNICODE)
		cbEntryID+=(ulSizeDisplayname + ulSizeAdrType + ulSizeAddress)*sizeof(WCHAR);
	else
		cbEntryID+=(ulSizeDisplayname + ulSizeAdrType + ulSizeAddress);
	
	// Allocate memory block
	hr = MAPIAllocateBuffer(cbEntryID, (void**)&lpEntryID);
	if(hr != hrSuccess)
		goto exit;

	memset(lpEntryID, 0, cbEntryID);

	offset+= 4;// abFlags all ready set on 00 00
		
	memcpy(lpEntryID + offset, (LPBYTE)&muidOneOff, sizeof(MAPIUID));// Muid
	offset+= sizeof(MAPIUID);
	
	offset+= 2;// Version all ready set on 00

	ulFlagsOneOff = (((ulFlags & MAPI_UNICODE)?MAPI_ONE_OFF_UNICODE:0) | ((ulFlags & MAPI_SEND_NO_RICH_INFO)?MAPI_ONE_OFF_NO_RICH_INFO:0));

	memcpy(lpEntryID+offset, &ulFlagsOneOff, 2);// Flags: used only first 16bits
	offset+= 2;

	if(ulFlags & MAPI_UNICODE){
		memset(&wszDisplayName, 0, sizeof(wszDisplayName));
		memset(&wszAdrType, 0, sizeof(wszAdrType));
		memset(&wszAddress, 0, sizeof(wszAddress));

		if(lpszName)
			mbstounicode(wszDisplayName, lpszName, sizeof(wszDisplayName)/sizeof(wszDisplayName[0]) );
		mbstounicode(wszAdrType, lpszAdrType, sizeof(wszAdrType)/sizeof(wszAdrType[0]) );
		mbstounicode(wszAddress, lpszAddress, sizeof(wszAdrType)/sizeof(wszAddress[0]) );

		memcpy(lpEntryID+offset, wszDisplayName, (unicodelen(wszDisplayName)+1)*sizeof(WCHAR));
		offset+=(unicodelen(wszDisplayName)+1)*sizeof(WCHAR);
		memcpy(lpEntryID+offset, wszAdrType, (unicodelen(wszAdrType)+1)*sizeof(WCHAR));
		offset+=(unicodelen(wszAdrType)+1)*sizeof(WCHAR);
		memcpy(lpEntryID+offset, wszAddress, (unicodelen(wszAddress)+1)*sizeof(WCHAR));
		offset+=(unicodelen(wszAddress)+1)*sizeof(WCHAR);
	}else {
		if(lpszName) memcpy(lpEntryID+offset, lpszName, ulSizeDisplayname);
		offset+=ulSizeDisplayname;
		
		memcpy(lpEntryID+offset, lpszAdrType, ulSizeAdrType);
		offset+=ulSizeAdrType;
		
		memcpy(lpEntryID+offset, lpszAddress, ulSizeAddress);		
		offset+=ulSizeAddress;
	}

	*lpcbEntryID= cbEntryID;
	*lppEntryID	= (LPENTRYID)lpEntryID;

exit:
	return hr;

}

HRESULT ECParseOneOff(LPENTRYID lpEntryID, ULONG cbEntryID, std::string &strName, std::string &strType, std::string &strAddress)
{
	HRESULT hr = hrSuccess;
	MAPIUID		muidOneOff = {MAPI_ONE_OFF_UID};
	char *lpBuffer = (char *)lpEntryID;
	char *szDisplayName = NULL;
	char *szType = NULL;
	char *szAddress = NULL;
	unsigned short ulFlags;
	size_t		size = 0;

	if (lpEntryID == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(*(unsigned int*)lpBuffer != 0) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	lpBuffer += 4;

	if(memcmp(&muidOneOff, lpBuffer, sizeof(MAPIUID)) != 0) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	lpBuffer += sizeof(MAPIUID);

	if(*(unsigned short *)lpBuffer != 0) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	lpBuffer += 2;

	memcpy(&ulFlags, lpBuffer, sizeof(ulFlags));

	lpBuffer += 2;

	if(ulFlags & MAPI_ONE_OFF_UNICODE) {
		// FIXME check len

		/* DARGH!  Unicode conversion coming up!
		   The functions wcstombs() e.a. return the converted bytes WITHOUT the trailing 0.
		   Since in Linux wchar_t is 32-bit by POSIX standard, we cannot use this function, because in
		   Windows it's 16-bit. So we have these functions wrapped by unicodetombs() e.a.
		   The Linux wrapper uses iconv() function to convert the 16 to 8 bit string, but iconv() returns
		   the converted bytes. Including the trailing 0, if you asked for the conversion of it.

		   So, we convert the string WITHOUT the 0. Now Windows and Linux do the same, converting the string,
		   and we terminate it here ourselfs.

		   This is somewhat dangerous, because if you use this function without knowing this, you WILL get
		   different results on Linux than on Windows.
		   Proposed solution is to use iconv() on Windows as well.
		*/

		size = unicodetombs(NULL, (const WCHAR *)lpBuffer, 0);
		if(size == (size_t)-1) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
			
		szDisplayName = new char [size+1];
		unicodetombs(szDisplayName, (WCHAR *)lpBuffer, size+1);
		szDisplayName[size] = '\0';
		lpBuffer +=  (unicodelen((WCHAR *)lpBuffer)+1) * sizeof(WCHAR);

		size = unicodetombs(NULL, (const WCHAR *)lpBuffer, 0);
		if(size == (size_t)-1) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		szType = new char [size+1];
		unicodetombs(szType, (WCHAR *)lpBuffer, size+1);
		szType[size] = '\0';
		lpBuffer +=  (unicodelen((WCHAR *)lpBuffer)+1) * sizeof(WCHAR);

		size = unicodetombs(NULL, (const WCHAR *)lpBuffer, 0);
		if(size == (size_t)-1) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		
		szAddress = new char [size+1];
		unicodetombs(szAddress, (WCHAR *)lpBuffer, size+1);
		szAddress[size] = '\0';
		lpBuffer +=  (unicodelen((WCHAR *)lpBuffer)+1) * sizeof(WCHAR);
		
	} else {
		szDisplayName = new char [strlen(lpBuffer)+1];
		strcpy(szDisplayName, lpBuffer);
		lpBuffer += strlen(lpBuffer)+1;

		szType = new char [strlen(lpBuffer)+1];
		strcpy(szType, lpBuffer);
		lpBuffer += strlen(lpBuffer)+1;

		szAddress = new char[strlen(lpBuffer)+1];
		strcpy(szAddress, lpBuffer);
		lpBuffer += strlen(lpBuffer)+1;
	}

	strName = szDisplayName;
	strType = szType;
	strAddress = szAddress;

exit:
	if(szDisplayName)
		delete [] szDisplayName;

	if(szType)
		delete [] szType;

	if(szAddress)
		delete [] szAddress;

	return hr;
}

#define BUFSIZE 4096
// Convert to/from wstring type from a specific charset (ie to/from UCS-2)
HRESULT HrConvertToWString(const std::string &strInput, const std::string &strCharset, std::basic_string<WCHAR> &wstrOutput)
{
	HRESULT hr = hrSuccess;
	iconv_t cd = (iconv_t)(-1);
	WCHAR buf[BUFSIZE];
	const char *lpSrc = NULL;
	char *lpDst = NULL;
	size_t cbSrc = 0;
	size_t cbDst = 0;
	size_t err;
	
	cd = iconv_open("UCS-2LE", strCharset.c_str());
	if (cd == (iconv_t)(-1)) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	lpSrc = strInput.c_str();
	cbSrc = strInput.length();
	
	wstrOutput.resize(0);
	
	while(cbSrc) {
		lpDst = (char *)buf;
		cbDst = sizeof(buf);
		err = iconv(cd, ICONV_HACK(&lpSrc), &cbSrc, &lpDst, &cbDst);
		
		if (err == (size_t)(-1) && cbDst == sizeof(buf)) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		
		// buf now contains converted chars, append them to output
		wstrOutput.append(buf, (sizeof(buf)-cbDst)/sizeof(WCHAR));
	}

    // Finalize (needed for stateful conversion)	
	lpDst = (char *)buf;
	cbDst = sizeof(buf);
	err = iconv(cd, NULL, NULL, &lpDst, &cbDst);
    wstrOutput.append(buf, (sizeof(buf)-cbDst)/sizeof(WCHAR));
	
exit:
	if (cd != (iconv_t)(-1))
		iconv_close(cd);

	return hr;
}

HRESULT HrConvertFromWString(const std::basic_string<WCHAR> wstrInput, std::string strCharset, std::string strOutput)
{
	HRESULT hr = hrSuccess;
	iconv_t cd = (iconv_t)(-1);
	const char *lpSrc;
	size_t cbSrc;
	char *lpDst;
	size_t cbDst;
	char buf[BUFSIZE];
	size_t err;
	
	cd = iconv_open(strCharset.c_str(), "UCS-2LE");
	if (cd == (iconv_t)(-1)) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	lpSrc = (char *)wstrInput.c_str();
	cbSrc = wstrInput.length() * sizeof(WCHAR);
	
	strOutput = "";
	
	while(cbSrc) {
		lpDst = buf;
		cbDst = sizeof(buf);
		
		err = iconv(cd, ICONV_HACK(&lpSrc), &cbSrc, &lpDst, &cbDst);
		if (err == (size_t)(-1) && cbDst == sizeof(buf)) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		
		// Append converted bytes
		strOutput.append(buf, sizeof(buf)-cbDst);
	}	
	
exit:
	if (cd != (iconv_t)(-1))
		iconv_close(cd);

	return hr;
}

// Convert WCHAR input to CHAR output, escaping unicode chars to &#XXXX;
HRESULT HrEscapeHTML(std::basic_string<WCHAR> &strInput, std::string &strOutput)
{
	HRESULT hr = hrSuccess;
	unsigned int i = 0;
	WCHAR c = 0;
	char buffer[16];
	
	strOutput.resize(0);
	
	for(i=0;i<strInput.size();i++) {
		c = strInput[i];
		if(c < 128) {
			strOutput.append(1,(char)c);
		} else {
			strOutput.append("&#");
			snprintf(buffer,sizeof(buffer),"%d", c);
			strOutput.append(buffer);
			strOutput.append(";");
		}
	}
	
	return hr;
}

HRESULT HrToQuotedPrintable1252(std::string &input, std::string *output, bool header)
{
	HRESULT hr = hrSuccess;
	ULONG i;
	string tmp;
	bool qp = false;
	const char digits[] = "0123456789ABCDEF";

	// only email headers have this prefix
	if (header)
		tmp = "=?windows-1252?Q?";

	for  (i = 0; i < input.size(); i++) {
		if ((unsigned char)input[i] > 127) {
			tmp.push_back('=');
			tmp.push_back(digits[((unsigned char)input[i]>>4)]);
			tmp.push_back(digits[((unsigned char)input[i]&0x0F)]);
			qp = true;
		} else {
			switch ((unsigned char)input[i]) {
			case ' ':
				if (header)
					tmp.push_back('_'); // only email headers need this, don't set qp marker if only spaces are found
				else
					tmp.push_back(input[i]); // leave email body unaffected
				break;
			case '\r':
			case '\n':
				// no idea how a user would enter a \r\n in the subject, but it's do-able ofcource ;)
				if (header) {
					tmp.push_back('=');
					tmp.push_back(digits[((unsigned char)input[i]>>4)]);
					tmp.push_back(digits[((unsigned char)input[i]&0x0F)]);
					qp = true;
				} else
					tmp.push_back(input[i]); // leave email body unaffected
				break;
			case '\t':
			case ',':
			case ';':
			case ':':
			case '_':
			case '@':
			case '(':
			case ')':
			case '<':
			case '>':
			case '[':
			case ']':
			case '?':
			case '=':
				tmp.push_back('=');
				tmp.push_back(digits[((unsigned char)input[i]>>4)]);
				tmp.push_back(digits[((unsigned char)input[i]&0x0F)]);
				qp = true;
				break;
			default:
				tmp.push_back(input[i]);
			};
		}
	}

	if (header)
		tmp += "?=";

	if (qp)
		*output = tmp;
	else
		*output = input;           // simple string was good enough

	return hr;
}

/////////////////////////////////////////////////////////
// Newmail notify
//
HRESULT HrNewMailNotification(IMsgStore* lpMDB, IMessage* lpMessage, ECLogger* lpLogger)
{
	HRESULT hr = hrSuccess;

	// Newmail notify
	ULONG			cNewMailValues = 0;
	LPSPropValue	lpNewMailPropArray = NULL;
	NOTIFICATION	sNotification;

	// Get notify properties
	hr = lpMessage->GetProps((LPSPropTagArray)&sPropNewMailColumns, 0, &cNewMailValues, &lpNewMailPropArray);
	if (hr != hrSuccess) {
		if(lpLogger)
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to notify newmail, error code: 0x%08X", hr);
		goto exit;
	}

	// Notification type
	sNotification.ulEventType = fnevNewMail;
	
	// PR_ENTRYID
	sNotification.info.newmail.cbEntryID = lpNewMailPropArray[NEWMAIL_ENTRYID].Value.bin.cb;
	sNotification.info.newmail.lpEntryID = (LPENTRYID)lpNewMailPropArray[NEWMAIL_ENTRYID].Value.bin.lpb;
	
	// PR_PARENT_ENTRYID
	sNotification.info.newmail.cbParentID = lpNewMailPropArray[NEWMAIL_PARENT_ENTRYID].Value.bin.cb;
	sNotification.info.newmail.lpParentID = (LPENTRYID)lpNewMailPropArray[NEWMAIL_PARENT_ENTRYID].Value.bin.lpb;

	// flags if unicode
	sNotification.info.newmail.ulFlags = 0;

	// PR_MESSAGE_CLASS
	sNotification.info.newmail.lpszMessageClass = lpNewMailPropArray[NEWMAIL_MESSAGE_CLASS].Value.lpszA;

	// PR_MESSAGE_FLAGS
	sNotification.info.newmail.ulMessageFlags = lpNewMailPropArray[NEWMAIL_MESSAGE_FLAGS].Value.ul;

	if (lpMDB->NotifyNewMail(&sNotification) != hrSuccess) {
		if(lpLogger)
			lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to send 'New Mail' notification, error code: 0x%08X",hr);
		// since this is a warning .. email should be correctly made
		// fallthrough to rules and vacation msg
	}

exit:
	// Newmail notify
	if(lpNewMailPropArray)
		MAPIFreeBuffer(lpNewMailPropArray);

	return hr;
}

///////////////////////////////////////////
// Create Search key for recipients
//
HRESULT HrCreateEmailSearchKey(char* lpszEmailType, char* lpszEmail, ULONG* cb, LPBYTE* lppByte)
{
	HRESULT	hr = hrSuccess;
	LPBYTE	lpByte = NULL;
	ULONG	size;
	ULONG	sizeEmailType;
	ULONG	sizeEmail;

	size = 2; // : and \0
	sizeEmailType = (lpszEmailType)?strlen(lpszEmailType) : 0;
	sizeEmail = (lpszEmail)?strlen(lpszEmail) : 0;

	size = sizeEmailType + sizeEmail + 2; // : and \0
	
	hr = MAPIAllocateBuffer(size, (void**)&lpByte);
	if(hr != hrSuccess)
		goto exit;

	memcpy(lpByte, lpszEmailType, sizeEmailType);
	*(lpByte + sizeEmailType) = ':';
	memcpy(lpByte + sizeEmailType + 1, lpszEmail, sizeEmail);
	*(lpByte + size - 1) = 0;

	strupr((char*)lpByte);

	*lppByte = lpByte;
	*cb = size;
exit:
	if(hr != hrSuccess && lpByte != NULL)
		MAPIFreeBuffer(lpByte);

	return hr;
}


HRESULT HrGetAddress(IMAPISession *lpSession, LPSPropValue lpProps, ULONG cValues, ULONG ulPropTagEntryID, ULONG ulPropTagName, ULONG ulPropTagType, ULONG ulPropTagEmailAddress, std::string &strName, std::string &strType, std::string &strEmailAddress)
{
	HRESULT hr = hrSuccess;
	LPADRBOOK lpAdrBook = NULL;

	if (!lpSession || !lpProps) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	// If entryid is invalid, there is no need in opening the addressbook,
	// though we still call HrGetAddress with lpAdrBook
	if (PpropFindProp(lpProps, cValues, ulPropTagEntryID)){

		// First, try through the entryid
		lpSession->OpenAddressBook(0,NULL,AB_NO_DIALOG,&lpAdrBook);
		// fallthrough .. don't mind if no Addressbook could not be created (probably never happens)
	}
	 
	hr = HrGetAddress(lpAdrBook, lpProps, cValues, ulPropTagEntryID, ulPropTagName, ulPropTagType, ulPropTagEmailAddress, strName, strType, strEmailAddress);

exit:
	if(lpAdrBook)
		lpAdrBook->Release();

	return hr;
}

HRESULT HrGetAddress(IMAPISession *lpSession, IMessage *lpMessage, ULONG ulPropTagEntryID, ULONG ulPropTagName, ULONG ulPropTagType, ULONG ulPropTagEmailAddress, std::string &strName, std::string &strType, std::string &strEmailAddress)
{
	HRESULT hr = hrSuccess;
	SizedSPropTagArray(4, sptaProps) = { 4, { ulPropTagEntryID, ulPropTagName, ulPropTagType, ulPropTagEmailAddress } };
	ULONG cValues = 0;
	LPSPropValue lpProps = NULL;

	if (!lpSession || !lpMessage) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpMessage->GetProps((LPSPropTagArray)&sptaProps, 0, &cValues, &lpProps);
	if(FAILED(hr))
		goto exit;

	hr = HrGetAddress(lpSession, lpProps, cValues, ulPropTagEntryID, ulPropTagName, ulPropTagType, ulPropTagEmailAddress, strName, strType, strEmailAddress);

exit:
	if(lpProps)
		MAPIFreeBuffer(lpProps);

	return hr;
}

HRESULT HrGetAddress(LPADRBOOK lpAdrBook, IMessage *lpMessage, ULONG ulPropTagEntryID, ULONG ulPropTagName, ULONG ulPropTagType, ULONG ulPropTagEmailAddress, std::string &strName, std::string &strType, std::string &strEmailAddress)
{
	HRESULT hr = hrSuccess;
	SizedSPropTagArray(4, sptaProps) = { 4, { ulPropTagEntryID, ulPropTagName, ulPropTagType, ulPropTagEmailAddress } };
	ULONG cValues = 0;
	LPSPropValue lpProps = NULL;

	if (!lpAdrBook || !lpMessage) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpMessage->GetProps((LPSPropTagArray)&sptaProps, 0, &cValues, &lpProps);
	if(FAILED(hr))
		goto exit;

	hr = HrGetAddress(lpAdrBook, lpProps, cValues, ulPropTagEntryID, ulPropTagName, ulPropTagType, ulPropTagEmailAddress, strName, strType, strEmailAddress);

exit:
	if(lpProps)
		MAPIFreeBuffer(lpProps);

	return hr;
}

/*
 * Attempts to get the SMTP email address for the addressbook entity in strResolve. May pass EMS_AB_ADDRESS_LOOKUP in ulFlags
 * for exact match only
 */
HRESULT HrResolveToSMTP(LPADRBOOK lpAdrBook, std::string strResolve, unsigned int ulFlags, std::string &strSMTPAddress)
{
    HRESULT hr = hrSuccess;
    LPADRLIST lpAdrList = NULL;
    LPSPropValue lpEntryID = NULL;
    ULONG ulType = 0;
    IMAPIProp *lpMailUser = NULL;
    LPSPropValue lpSMTPAddress = NULL;
    
    hr = MAPIAllocateBuffer(sizeof(ADRLIST), (void **)&lpAdrList);
    if(hr != hrSuccess)
        goto exit;
    
    lpAdrList->cEntries = 1;
    lpAdrList->aEntries[0].cValues = 1;

    hr = MAPIAllocateBuffer(sizeof(SPropValue), (void **)&lpAdrList->aEntries[0].rgPropVals);
    if(hr != hrSuccess)
        goto exit;
        
    lpAdrList->aEntries[0].rgPropVals[0].ulPropTag = PR_DISPLAY_NAME;
    lpAdrList->aEntries[0].rgPropVals[0].Value.lpszA = (char *)strResolve.c_str();
    
    hr = lpAdrBook->ResolveName(0, ulFlags, NULL, lpAdrList);
    if(hr != hrSuccess)
        goto exit;
        
    if(lpAdrList->cEntries != 1) {
        hr = MAPI_E_NOT_FOUND;
        goto exit;
    }
    
    lpEntryID = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_ENTRYID);
    if(!lpEntryID) {
        hr = MAPI_E_NOT_FOUND;
        goto exit;
    }
    
    hr = lpAdrBook->OpenEntry(lpEntryID->Value.bin.cb, (LPENTRYID)lpEntryID->Value.bin.lpb, &IID_IMAPIProp, 0, &ulType, (LPUNKNOWN *)&lpMailUser);
    if(hr != hrSuccess) {
        goto exit;
    }
    
    hr = HrGetOneProp(lpMailUser, PR_SMTP_ADDRESS, &lpSMTPAddress);
    if(hr != hrSuccess) {
        goto exit;
    }
    
    strSMTPAddress = lpSMTPAddress->Value.lpszA;
    
exit:
    if(lpAdrList)
        FreePadrlist(lpAdrList);
        
    if(lpSMTPAddress)
        MAPIFreeBuffer(lpSMTPAddress);
        
    if(lpMailUser)
        lpMailUser->Release();
        
    return hr;
}

HRESULT HrGetAddress(LPADRBOOK lpAdrBook, LPSPropValue lpProps, ULONG cValues, ULONG ulPropTagEntryID, ULONG ulPropTagName, ULONG ulPropTagType, ULONG ulPropTagEmailAddress, std::string &strName, std::string &strType, std::string &strEmailAddress)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpEntryID = NULL;
	LPSPropValue lpName = NULL;
	LPSPropValue lpType = NULL;
	LPSPropValue lpAddress = NULL;
	std::string strSMTPAddress;

	strName = "";
	strType = "";
	strEmailAddress = "";

	if (lpProps && cValues) {
		lpEntryID	= PpropFindProp(lpProps, cValues, ulPropTagEntryID);
		lpName		= PpropFindProp(lpProps, cValues, ulPropTagName);
		lpType		= PpropFindProp(lpProps, cValues, ulPropTagType);
		lpAddress	= PpropFindProp(lpProps, cValues, ulPropTagEmailAddress);
		if (lpEntryID && PROP_TYPE(lpEntryID->ulPropTag) != PT_BINARY) lpEntryID = NULL;
		if (lpName && PROP_TYPE(lpName->ulPropTag) != PT_STRING8) lpName = NULL;
		if (lpType && PROP_TYPE(lpType->ulPropTag) != PT_STRING8) lpType = NULL;
		if (lpAddress && PROP_TYPE(lpAddress->ulPropTag) != PT_STRING8) lpAddress = NULL;
	}
	if (lpEntryID == NULL || lpAdrBook == NULL || HrGetAddress(lpAdrBook, (LPENTRYID)lpEntryID->Value.bin.lpb, lpEntryID->Value.bin.cb, strName, strType, strEmailAddress) != hrSuccess) {
        // EntryID failed, try fallback
        if (lpName)
            strName = lpName->Value.lpszA;
        if (lpType)
            strType = lpType->Value.lpszA;
        if (lpAddress)
            strEmailAddress = lpAddress->Value.lpszA;
    }
    		
    // If we don't have an SMTP address yet, try to resolve the item to get the SMTP address
    if (lpAdrBook && lpType && lpAddress && stricmp(strType.c_str(), "SMTP") != 0) {
        if (HrResolveToSMTP(lpAdrBook, strEmailAddress, EMS_AB_ADDRESS_LOOKUP, strSMTPAddress) == hrSuccess)
            strEmailAddress = strSMTPAddress;
    }

exit:
	return hr;
}


HRESULT HrGetAddress(LPADRBOOK lpAdrBook, LPENTRYID lpEntryID, ULONG cbEntryID, std::string &strName, std::string &strType, std::string &strEmailAddress)
{
	HRESULT hr = hrSuccess;
	IMailUser	*lpMailUser = NULL;
	ULONG		ulType = 0;
	ULONG		cMailUserValues = 0;
	LPSPropValue lpMailUserProps = NULL;
	SizedSPropTagArray(4, sptaAddressProps) = { 4, { PR_DISPLAY_NAME, PR_ADDRTYPE, PR_EMAIL_ADDRESS, PR_SMTP_ADDRESS } };

	if (!lpAdrBook || !lpEntryID) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpAdrBook->OpenEntry(cbEntryID, lpEntryID, &IID_IMailUser, 0, &ulType, (IUnknown **)&lpMailUser);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMailUser->GetProps((LPSPropTagArray)&sptaAddressProps, 0, &cMailUserValues, &lpMailUserProps);

	if (FAILED(hr))
		goto exit;

	if(lpMailUserProps[0].ulPropTag == PR_DISPLAY_NAME)
		strName = lpMailUserProps[0].Value.lpszA;
	if(lpMailUserProps[1].ulPropTag == PR_ADDRTYPE)
		strType = lpMailUserProps[1].Value.lpszA;

	if(lpMailUserProps[3].ulPropTag == PR_SMTP_ADDRESS)
		strEmailAddress = lpMailUserProps[3].Value.lpszA;
	else if(lpMailUserProps[2].ulPropTag == PR_EMAIL_ADDRESS)
		strEmailAddress = lpMailUserProps[2].Value.lpszA;

	hr = hrSuccess;

exit:
	if(lpMailUser)
		lpMailUser->Release();

	if(lpMailUserProps)
		MAPIFreeBuffer(lpMailUserProps);

	return hr;
}

HRESULT DoSentMail(IMAPISession *lpSession, ULONG ulFlags, LPMESSAGE lpMessage) {
	HRESULT			hr = hrSuccess;
	LPMDB			lpMDB = NULL;
	LPMAPIFOLDER 	lpFolder = NULL;
	ENTRYLIST		sMsgList;
	SBinary			sEntryID;
	LPSPropValue	lpPropValue = NULL;
	ULONG			cValues = 0;
	ULONG			ulType = 0;
	
	enum esPropDoSentMail{ DSM_ENTRYID, DSM_PARENT_ENTRYID, DSM_SENTMAIL_ENTRYID, DSM_DELETE_AFTER_SUBMIT, DSM_STORE_ENTRYID};
	SizedSPropTagArray(5, sPropDoSentMail) = {5, {PR_ENTRYID, PR_PARENT_ENTRYID, PR_SENTMAIL_ENTRYID, PR_DELETE_AFTER_SUBMIT, PR_STORE_ENTRYID} };
    
	// Check incomming parameter
	if(lpMessage == NULL) {
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}

	// Get Sentmail properties
	hr = lpMessage->GetProps((LPSPropTagArray)&sPropDoSentMail, 0, &cValues, &lpPropValue);
	if(FAILED(hr) || 
		(lpPropValue[DSM_SENTMAIL_ENTRYID].ulPropTag != PR_SENTMAIL_ENTRYID && 
		lpPropValue[DSM_DELETE_AFTER_SUBMIT].ulPropTag != PR_DELETE_AFTER_SUBMIT)
	  )
	{
		// Ignore error, leave the mail where it is
		hr = hrSuccess;
		lpMessage->Release();
		goto exit;
	}else if(lpPropValue[DSM_ENTRYID].ulPropTag != PR_ENTRYID || 
			 lpPropValue[DSM_PARENT_ENTRYID].ulPropTag != PR_PARENT_ENTRYID ||
			 lpPropValue[DSM_STORE_ENTRYID].ulPropTag != PR_STORE_ENTRYID)
	{
		// Those properties always needed
		hr = MAPI_E_NOT_FOUND;
		lpMessage->Release();
		goto exit;
	}

	lpMessage->Release(); // Yes, we release the message for the caller

	hr = lpSession->OpenMsgStore(0, lpPropValue[DSM_STORE_ENTRYID].Value.bin.cb, (LPENTRYID)lpPropValue[DSM_STORE_ENTRYID].Value.bin.lpb, NULL, MDB_WRITE | MDB_NO_DIALOG | MDB_NO_MAIL |MDB_TEMPORARY, &lpMDB);
	if(hr != hrSuccess)
		goto exit;

	sEntryID.cb = lpPropValue[DSM_ENTRYID].Value.bin.cb;
	sEntryID.lpb = lpPropValue[DSM_ENTRYID].Value.bin.lpb;
	sMsgList.cValues = 1;
	sMsgList.lpbin = &sEntryID;

	// Handle PR_SENTMAIL_ENTRYID
	if(lpPropValue[DSM_SENTMAIL_ENTRYID].ulPropTag == PR_SENTMAIL_ENTRYID)
	{
		//Open Sentmail Folder
		hr = lpMDB->OpenEntry(lpPropValue[DSM_SENTMAIL_ENTRYID].Value.bin.cb, (LPENTRYID)lpPropValue[DSM_SENTMAIL_ENTRYID].Value.bin.lpb, NULL, MAPI_MODIFY, &ulType, (IUnknown **)&lpFolder);
		if(hr != hrSuccess)
			goto exit;

		// Move Message
		hr = lpFolder->CopyMessages(&sMsgList, &IID_IMAPIFolder, lpFolder, 0, NULL, MESSAGE_MOVE);
	}

	// Handle PR_DELETE_AFTER_SUBMIT
	if(lpPropValue[DSM_DELETE_AFTER_SUBMIT].ulPropTag == PR_DELETE_AFTER_SUBMIT && lpPropValue[DSM_DELETE_AFTER_SUBMIT].Value.b == true)
	{
		if(lpFolder == NULL)
		{
			// Open parent folder of the sent message
			hr = lpMDB->OpenEntry(lpPropValue[DSM_PARENT_ENTRYID].Value.bin.cb, (LPENTRYID)lpPropValue[DSM_PARENT_ENTRYID].Value.bin.lpb, NULL, MAPI_MODIFY, &ulType, (IUnknown **)&lpFolder);
			if(hr != hrSuccess)
				goto exit;
		}

		// Delete Message
		hr = lpFolder->DeleteMessages(&sMsgList, 0, NULL, 0); 
	}

exit:
	if(lpFolder)
		lpFolder->Release();

	if(lpMDB)
		lpMDB->Release();

	if(lpPropValue)
		MAPIFreeBuffer(lpPropValue);

	return hr;
}

// This is a class that implements IMAPIProp's GetProps(), and nothing else. Its data
// is retrieved from the passed lpProps/cValues property array
class ECRowWrapper : public IMAPIProp {
public:
	ECRowWrapper(LPSPropValue lpProps, ULONG cValues) : m_cValues(cValues), m_lpProps(lpProps) {};
	~ECRowWrapper() {};

	ULONG __stdcall AddRef() { return 1; }; // no ref. counting
	ULONG __stdcall Release() { return 1; };
	HRESULT __stdcall QueryInterface(const IID &iid, LPVOID *lpvoid) { return MAPI_E_INTERFACE_NOT_SUPPORTED; };

	HRESULT __stdcall GetLastError(HRESULT hResult, ULONG ulFlags, LPMAPIERROR FAR * lppMAPIError) { return MAPI_E_NOT_FOUND; }
	HRESULT __stdcall SaveChanges(ULONG ulFlags) { return MAPI_E_NO_SUPPORT; }
	HRESULT __stdcall GetProps(LPSPropTagArray lpTags, ULONG ulFlags, ULONG *lpcValues, LPSPropValue *lppProps) { 
		HRESULT hr = hrSuccess;
		ULONG i = 0;
		LPSPropValue lpProps = NULL;
		LPSPropValue lpFind = NULL;
		SPropValue sError;

		MAPIAllocateBuffer(sizeof(SPropValue) * lpTags->cValues, (void **) &lpProps);

		for(i=0;i<lpTags->cValues;i++) {
			lpFind = PpropFindProp(m_lpProps, m_cValues, lpTags->aulPropTag[i]);
			if(lpFind)
				Util::HrCopyProperty(&lpProps[i], lpFind, lpProps);
			else {
				sError.ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpTags->aulPropTag[i]));
				sError.Value.err = MAPI_E_NOT_FOUND;

				Util::HrCopyProperty(&lpProps[i], &sError, lpProps);
				hr = MAPI_W_ERRORS_RETURNED;
			}
		}

		*lppProps = lpProps;
		*lpcValues = lpTags->cValues;

		return hr;
	};
	HRESULT __stdcall GetPropList(ULONG ulFlags, LPSPropTagArray *lppTags) { return MAPI_E_NO_SUPPORT; }
	HRESULT __stdcall OpenProperty(ULONG ulPropTag, LPCIID lpiid, ULONG ulInterfaceOptions, ULONG ulFlags, LPUNKNOWN *lppUnk){ return MAPI_E_NO_SUPPORT; }
	HRESULT __stdcall SetProps(ULONG cValues, LPSPropValue lpProps, LPSPropProblemArray *lppProblems) { return MAPI_E_NO_SUPPORT; }
	HRESULT __stdcall DeleteProps(LPSPropTagArray lpToDelete, LPSPropProblemArray *lppProblems) { return MAPI_E_NO_SUPPORT; }
	HRESULT __stdcall CopyTo(ULONG ciidExclude, LPCIID rgiidExclude, LPSPropTagArray lpExcludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpInterface, LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray FAR * lppProblems) { return MAPI_E_NO_SUPPORT; }
	HRESULT __stdcall CopyProps(LPSPropTagArray lpIncludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpInterface, LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray FAR * lppProblems) { return MAPI_E_NO_SUPPORT; }
	HRESULT __stdcall GetNamesFromIDs( LPSPropTagArray FAR * lppPropTags, LPGUID lpPropSetGuid, ULONG ulFlags, ULONG FAR * lpcPropNames, LPMAPINAMEID FAR * FAR * lpppPropNames) { return MAPI_E_NO_SUPPORT; }
	HRESULT __stdcall GetIDsFromNames( ULONG cPropNames, LPMAPINAMEID FAR * lppPropNames, ULONG ulFlags, LPSPropTagArray FAR * lppPropTags) { return MAPI_E_NO_SUPPORT; }
private:
	ULONG			m_cValues;
	LPSPropValue	m_lpProps;
};

HRESULT TestRelop(ULONG relop, int result, bool* fMatch) {
	HRESULT hr = hrSuccess;
	switch (relop) {
	case RELOP_LT:
		*fMatch = result < 0;
		break;
	case RELOP_LE:
		*fMatch = result <= 0;
		break;
	case RELOP_GT:
		*fMatch = result > 0;
		break;
	case RELOP_GE:
		*fMatch = result >= 0;
		break;
	case RELOP_EQ:
		*fMatch = result == 0;
		break;
	case RELOP_NE:
		*fMatch = result != 0;
		break;
	case RELOP_RE:
	default:
		*fMatch = false;
		hr = MAPI_E_TOO_COMPLEX;
		break;
	};
	return hr;
}

#define RESTRICT_MAX_RECURSE_LEVEL 16
HRESULT GetRestrictTagsRecursive(LPSRestriction lpRestriction, std::list<unsigned int> *lpList, ULONG ulLevel)
{
	HRESULT hr = hrSuccess;
	ULONG i;

	if(ulLevel > RESTRICT_MAX_RECURSE_LEVEL)
		return MAPI_E_TOO_COMPLEX;

	switch(lpRestriction->rt) {
		case RES_AND:
			for(i=0;i<lpRestriction->res.resAnd.cRes;i++) {
				hr = GetRestrictTagsRecursive(&lpRestriction->res.resAnd.lpRes[i], lpList, ulLevel+1);
				if(hr != hrSuccess)
					goto exit;
			}
			break;
		case RES_OR:
			for(i=0;i<lpRestriction->res.resOr.cRes;i++) {
				hr = GetRestrictTagsRecursive(&lpRestriction->res.resOr.lpRes[i], lpList, ulLevel+1);
				if(hr != hrSuccess)
					goto exit;
			}
			break;
		case RES_NOT:
			hr = GetRestrictTagsRecursive(lpRestriction->res.resNot.lpRes, lpList, ulLevel+1);
			break;
		case RES_CONTENT:
			lpList->push_back(lpRestriction->res.resContent.ulPropTag);
			lpList->push_back(lpRestriction->res.resContent.lpProp->ulPropTag);
			break;
		case RES_PROPERTY:
			lpList->push_back(lpRestriction->res.resProperty.ulPropTag);
			lpList->push_back(lpRestriction->res.resProperty.lpProp->ulPropTag);
			break;
		case RES_COMPAREPROPS:
			lpList->push_back(lpRestriction->res.resCompareProps.ulPropTag1);
			lpList->push_back(lpRestriction->res.resCompareProps.ulPropTag2);
			break;
		case RES_BITMASK:
			lpList->push_back(lpRestriction->res.resBitMask.ulPropTag);
			break;
		case RES_SIZE:
			lpList->push_back(lpRestriction->res.resSize.ulPropTag);
			break;
		case RES_EXIST:
			lpList->push_back(lpRestriction->res.resExist.ulPropTag);
			break;
		case RES_SUBRESTRICTION:
			lpList->push_back(lpRestriction->res.resSub.ulSubObject);
			break;
		case RES_COMMENT:
			hr = GetRestrictTagsRecursive(lpRestriction->res.resComment.lpRes, lpList, ulLevel+1);
			break;
	}

exit:
	return hr;
}

HRESULT GetRestrictTags(LPSRestriction lpRestriction, LPSPropTagArray *lppTags)
{
	HRESULT hr = hrSuccess;
	std::list<unsigned int> lstTags;
	std::list<unsigned int>::iterator iterTags;
	ULONG n = 0;

	LPSPropTagArray lpTags = NULL;

	hr = GetRestrictTagsRecursive(lpRestriction, &lstTags, 0);
	if(hr != hrSuccess)
		goto exit;

	MAPIAllocateBuffer(CbNewSPropTagArray(lstTags.size()), (void **) &lpTags);
	lpTags->cValues = lstTags.size();

	lstTags.sort();
	lstTags.unique();

	for(iterTags = lstTags.begin(); iterTags != lstTags.end() && n < lpTags->cValues; iterTags++) {
		lpTags->aulPropTag[n] = *iterTags;
		n++;
	}
	
	lpTags->cValues = n;

	*lppTags = lpTags;

exit:
	return hr;
}

HRESULT TestRestriction(LPSRestriction lpCondition, ULONG cValues, LPSPropValue lpPropVals, ULONG ulLevel) {
	HRESULT hr = hrSuccess;
	ECRowWrapper *lpRowWrapper = new ECRowWrapper(lpPropVals, cValues);

	hr = TestRestriction(lpCondition, (IMAPIProp *)lpRowWrapper, ulLevel);

	if(lpRowWrapper)
		delete lpRowWrapper;

	return hr;
}

HRESULT TestRestriction(LPSRestriction lpCondition, IMAPIProp *lpMessage, ULONG ulLevel) {
	HRESULT hr = hrSuccess;
	ULONG c;
	bool fMatch = false;
	LPSPropValue lpProp = NULL;
	LPSPropValue lpProp2 = NULL;
	ULONG ulPropType;
	int result;
	unsigned int ulSize;
	IMAPITable *lpTable = NULL;
	LPSPropTagArray lpTags = NULL;
	LPSRowSet lpRowSet = NULL;
	ECRowWrapper *lpRowWrapper = NULL;

	if (ulLevel > RESTRICT_MAX_RECURSE_LEVEL)
		return MAPI_E_TOO_COMPLEX;

	if (!lpCondition)
		return MAPI_E_INVALID_PARAMETER;

	switch (lpCondition->rt) {
	// loops
	case RES_AND:
		for (c = 0; c < lpCondition->res.resAnd.cRes; c++) {
			hr = TestRestriction(&lpCondition->res.resAnd.lpRes[c], lpMessage, ulLevel+1);
			if (hr != hrSuccess) {
				fMatch = false;
				break;
			}
			fMatch = true;
		}
		break;
	case RES_OR:
		for (c = 0; c < lpCondition->res.resAnd.cRes; c++) {
			hr = TestRestriction(&lpCondition->res.resOr.lpRes[c], lpMessage, ulLevel+1);
			if (hr == hrSuccess) {
				fMatch = true;
				break;
			} else if (hr == MAPI_E_TOO_COMPLEX)
				break;
		}
		break;
	case RES_NOT:
		hr = TestRestriction(lpCondition->res.resNot.lpRes, lpMessage, ulLevel+1);
		if (hr != MAPI_E_TOO_COMPLEX) {
			if (hr == hrSuccess) {
				hr = MAPI_E_NOT_FOUND;
				fMatch = false;
			} else {
				hr = hrSuccess;
				fMatch = true;
			}
		}
		break;

	// Prop compares
	case RES_CONTENT:

// FIXME: support PT_MV_STRING8 and PT_MV_BINARY

		// fuzzy string compare
		if (PROP_TYPE(lpCondition->res.resContent.ulPropTag) != PT_STRING8 &&
			PROP_TYPE(lpCondition->res.resContent.ulPropTag) != PT_BINARY) {
			hr = MAPI_E_TOO_COMPLEX;
			break;
		}
		ulPropType = PROP_TYPE(lpCondition->res.resContent.ulPropTag);
		hr = HrGetOneProp(lpMessage, lpCondition->res.resContent.ulPropTag, &lpProp);
		if (hr == hrSuccess) {
			char *lpSearchString, *lpSearchData;
			unsigned int ulSearchStringSize, ulSearchDataSize;
			if(PROP_TYPE(lpCondition->res.resContent.ulPropTag) == PT_STRING8)	{
				lpSearchString = lpCondition->res.resContent.lpProp->Value.lpszA;
				lpSearchData = lpProp->Value.lpszA;
				ulSearchStringSize = lpSearchString?strlen(lpSearchString):0;
				ulSearchDataSize = lpSearchData?strlen(lpSearchData):0;
			} else {
				// PT_BINARY
				lpSearchString = (char*)lpCondition->res.resContent.lpProp->Value.bin.lpb;
				lpSearchData = (char*)lpProp->Value.bin.lpb;
				ulSearchStringSize = lpCondition->res.resContent.lpProp->Value.bin.cb;
				ulSearchDataSize = lpProp->Value.bin.cb;
			}

			switch(lpCondition->res.resContent.ulFuzzyLevel & 0xFFFF) {
			case FL_FULLSTRING:
				if(ulSearchDataSize == ulSearchStringSize) {
					if(ulPropType == PT_STRING8 && lpCondition->res.resContent.ulFuzzyLevel & FL_IGNORECASE) {
						if(stricmp_str1252(lpSearchData, lpSearchString) == 0) {
							fMatch = true;
							break;
						}
					} else { // Compare PT_STRING8 and PT_BINARY
						if(memcmp(lpSearchData, lpSearchString, ulSearchDataSize) == 0) {
							fMatch = true;
							break;
						}
					}
				}
				break;
			case FL_PREFIX:
				if(ulSearchDataSize >= ulSearchStringSize) {
					if(ulPropType == PT_STRING8 && lpCondition->res.resContent.ulFuzzyLevel & FL_IGNORECASE) {
						if(strnicmp_str1252(lpSearchData, lpSearchString, ulSearchStringSize) == 0) {
							fMatch = true;
							break;
						}
					} else { // Compare PT_STRING8 and PT_BINARY
						if(memcmp(lpSearchData, lpSearchString, ulSearchStringSize) == 0) {
							fMatch = true;
							break;
						}
					}
				}
				break;
			case FL_SUBSTRING:
				if(ulPropType == PT_STRING8 && lpCondition->res.resContent.ulFuzzyLevel & FL_IGNORECASE) {
					if(strcasestr_str1252(lpSearchData, lpSearchString) != NULL) {
						fMatch = true;
						break;
					}
				} else { // Compare PT_STRING8 and PT_BINARY
					if(memsubstr(lpSearchData, ulSearchDataSize, lpSearchString, ulSearchStringSize) == 0) {
						fMatch = true;
						break;
					}
				}
				break;
			};
		}
		break;
	case RES_PROPERTY:
		if(PROP_TYPE(lpCondition->res.resProperty.ulPropTag) != PROP_TYPE(lpCondition->res.resProperty.lpProp->ulPropTag)) {
			// cannot compare two different types
			hr = MAPI_E_TOO_COMPLEX;
			break;
		}
		hr = HrGetOneProp(lpMessage, lpCondition->res.resProperty.ulPropTag, &lpProp);
		if (hr != hrSuccess)
			break;
		Util::CompareProp(lpCondition->res.resProperty.lpProp, lpProp, &result);
		hr = TestRelop(lpCondition->res.resProperty.relop, result, &fMatch);
		break;
	case RES_COMPAREPROPS:
		if(PROP_TYPE(lpCondition->res.resCompareProps.ulPropTag1) != PROP_TYPE(lpCondition->res.resCompareProps.ulPropTag2)) {
			// cannot compare two different types
			hr = MAPI_E_TOO_COMPLEX;
			break;
		}
		hr = HrGetOneProp(lpMessage, lpCondition->res.resCompareProps.ulPropTag1, &lpProp);
		if (hr != hrSuccess) {
			break;
		}
		hr = HrGetOneProp(lpMessage, lpCondition->res.resCompareProps.ulPropTag2, &lpProp2);
		if (hr != hrSuccess) {
			break;
		}
		Util::CompareProp(lpProp, lpProp2, &result);
		hr = TestRelop(lpCondition->res.resProperty.relop, result, &fMatch);
		break;
	case RES_BITMASK:
		if (PROP_TYPE(lpCondition->res.resBitMask.ulPropTag) != PT_LONG) {
			hr = MAPI_E_TOO_COMPLEX;
			break;
		}
		hr = HrGetOneProp(lpMessage, lpCondition->res.resBitMask.ulPropTag, &lpProp);
		if (hr != hrSuccess) {
			break;
		}
		fMatch = (lpProp->Value.ul & lpCondition->res.resBitMask.ulMask) == 0;
		if (lpCondition->res.resBitMask.relBMR == BMR_NEZ)
			fMatch = !fMatch;
		break;
	case RES_SIZE:
		hr = HrGetOneProp(lpMessage, lpCondition->res.resSize.ulPropTag, &lpProp);
		if (hr != hrSuccess) {
			break;
		}
		ulSize = Util::PropSize(lpProp);
		result = ulSize - lpCondition->res.resSize.cb;
		hr = TestRelop(lpCondition->res.resSize.relop, result, &fMatch);
		break;
	case RES_EXIST:
		hr = HrGetOneProp(lpMessage, lpCondition->res.resExist.ulPropTag, &lpProp);
		if (hr != hrSuccess) {
			break;
		}
		fMatch = true;
		break;
	case RES_SUBRESTRICTION:
		// A subrestriction is basically an OR restriction over all the rows in a specific
		// table. We currently support the attachment table (PR_MESSAGE_ATTACHMENTS) and the 
		// recipient table (PR_MESSAGE_RECIPIENTS) here.
		hr = lpMessage->OpenProperty(lpCondition->res.resSub.ulSubObject, &IID_IMAPITable, 0, 0, (LPUNKNOWN *)&lpTable);
		if(hr != hrSuccess) {
			hr = MAPI_E_TOO_COMPLEX;
			goto exit;
		}
		// Get a list of properties we may be needing
		hr = GetRestrictTags(lpCondition->res.resSub.lpRes, &lpTags);
		if(hr != hrSuccess)
			goto exit;

		hr = lpTable->SetColumns(lpTags, 0);
		if(hr != hrSuccess)
			goto exit;

		while(1) {
			hr = lpTable->QueryRows(1, 0, &lpRowSet);
			if(hr != hrSuccess)
				goto exit;

			if(lpRowSet->cRows != 1)
				break;

			// Wrap the row into an IMAPIProp compatible object so we can recursively call
			// this function (which obviously itself doesn't support RES_SUBRESTRICTION as 
			// there aren't any subobjects under the subobjects .. unless we count
			// messages in PR_ATTACH_DATA_OBJ under attachments... Well we don't support
			// that in any case ...)

			hr = TestRestriction(lpCondition->res.resSub.lpRes, lpRowSet->aRow[0].cValues, lpRowSet->aRow[0].lpProps, ulLevel+1);
			if(hr == hrSuccess) {
				fMatch = true;
				break;
			}

			FreeProws(lpRowSet);
			lpRowSet = NULL;

			delete lpRowWrapper;
			lpRowWrapper = NULL;
		}
		break;

	case RES_COMMENT:
		hr = TestRestriction(lpCondition->res.resComment.lpRes, lpMessage, ulLevel+1);
		if(hr == hrSuccess)
			fMatch = true;
		else
			fMatch = false;
		break;

	default:
		break;
	};

exit:
	if (lpRowWrapper)
		delete lpRowWrapper;
	if (lpRowSet)
		FreeProws(lpRowSet);
	if (lpTags)
		MAPIFreeBuffer(lpTags);
	if (lpTable)
		lpTable->Release();
	if (lpProp)
		MAPIFreeBuffer(lpProp);
	if (lpProp2)
		MAPIFreeBuffer(lpProp2);

	if (fMatch)
		return hrSuccess;
	else if (hr == hrSuccess)
		return MAPI_E_NOT_FOUND;
	return hr;
}

HRESULT GetClientVersion(unsigned int* ulVersion)
{
	HRESULT hr = hrSuccess;
	*ulVersion = 0;

	return hr;
}

HRESULT FindFolder(LPMAPITABLE lpTable, const char *folder, LPSPropValue *lppFolderProp) {
	HRESULT hr;
	LPSRowSet		lpRowSet = NULL;
	ULONG nValues;
	SizedSPropTagArray(2, sptaName) = { 2, { PR_DISPLAY_NAME, PR_ENTRYID } };

	hr = lpTable->SetColumns((LPSPropTagArray)&sptaName, 0);
	if (hr != hrSuccess)
		goto exit;

	while (TRUE) {
		hr = lpTable->QueryRows(1, 0, &lpRowSet);
		if (hr != hrSuccess)
			break;

		if(lpRowSet->cRows == 0) {
			hr = MAPI_E_NOT_FOUND;
			break;
		}

		if (stricmp(lpRowSet->aRow[0].lpProps[0].Value.lpszA, folder) == 0) {
			// found the folder
			hr = Util::HrCopyPropertyArray(&lpRowSet->aRow[0].lpProps[1], 1, lppFolderProp, &nValues);
			break;
		}

		FreeProws(lpRowSet);
		lpRowSet = NULL;
	}

exit:
	if (lpRowSet)
		FreeProws(lpRowSet);

	return hr;
}

HRESULT OpenSubFolder(LPMDB lpMDB, char *folder, char psep, ECLogger *lpLogger, bool blIsPublic, bool bCreateFolder, LPMAPIFOLDER *lppSubFolder) {
	HRESULT hr;
	ECLogger *lpNullLogger = new ECLogger_Null();
	LPSPropValue	lpPropIPMSubtree = NULL;
	LPMAPITABLE		lpTable = NULL;
	ULONG			ulObjType;
	LPSPropValue	lpPropFolder = NULL;
	LPMAPIFOLDER	lpFoundFolder = NULL;
	LPMAPIFOLDER	lpNewFolder = NULL;
	char *ptr = NULL;

	if (lpLogger == NULL)
		lpLogger = lpNullLogger;
	if(blIsPublic)
	{
		hr = HrGetOneProp(lpMDB, PR_IPM_PUBLIC_FOLDERS_ENTRYID, &lpPropIPMSubtree);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to find PR_IPM_PUBLIC_FOLDERS_ENTRYID object, error code: 0x%08X", hr);
			goto exit;
		}
	}
	else
	{
		hr = HrGetOneProp(lpMDB, PR_IPM_SUBTREE_ENTRYID, &lpPropIPMSubtree);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to find IPM_SUBTREE object, error code: 0x%08X", hr);
			goto exit;
		}
	}

	hr = lpMDB->OpenEntry(lpPropIPMSubtree->Value.bin.cb, (LPENTRYID)lpPropIPMSubtree->Value.bin.lpb,
						  &IID_IMAPIFolder, 0, &ulObjType, (LPUNKNOWN*)&lpFoundFolder);
	if (hr != hrSuccess || ulObjType != MAPI_FOLDER) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open IPM_SUBTREE object, error code: 0x%08X", hr);
		goto exit;
	}
	
	if(!folder)
		goto next;

	// deliver in a subfolder from the store
	do {
		string subfld;

		ptr = strchr(folder, psep);
		if (ptr)
			subfld = string(folder, ptr-folder);
		else
			subfld = string(folder);
		folder = ptr?ptr+1:NULL;

		hr = lpFoundFolder->GetHierarchyTable(0, &lpTable);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to view folder, error code: 0x%08X", hr);
			goto exit;
		}

		hr = FindFolder(lpTable, subfld.c_str(), &lpPropFolder);
		if (hr == MAPI_E_NOT_FOUND && bCreateFolder) {
			hr = lpFoundFolder->CreateFolder(FOLDER_GENERIC, (LPTSTR)subfld.c_str(), "Auto-created by Zarafa", &IID_IMAPIFolder, OPEN_IF_EXISTS, &lpNewFolder);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to create folder '%s', error code: 0x%08X", subfld.c_str(), hr);
				goto exit;
			}
		} else if (hr != hrSuccess)
			goto exit;

		// not needed anymore
		lpFoundFolder->Release();
		lpFoundFolder = NULL;
		lpTable->Release();
		lpTable = NULL;
		if (lpNewFolder) {
			lpFoundFolder = lpNewFolder;
			lpNewFolder = NULL;
		} else {
			hr = lpMDB->OpenEntry(lpPropFolder->Value.bin.cb, (LPENTRYID)lpPropFolder->Value.bin.lpb,
								  &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpFoundFolder);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open folder '%s', error code: 0x%08X", subfld.c_str(), hr);
				goto exit;
			}
		}
	} while (ptr);
next:
	if (lpFoundFolder) {
		lpFoundFolder->AddRef();
		*lppSubFolder = lpFoundFolder;
	}

exit:
	delete lpNullLogger;

	if (lpPropFolder)
		MAPIFreeBuffer(lpPropFolder);

	if (lpPropIPMSubtree)
		MAPIFreeBuffer(lpPropIPMSubtree);

	if (lpFoundFolder)
		lpFoundFolder->Release();

	if (lpTable)
		lpTable->Release();

	return hr;
}

/*
 * NAMED PROPERTY util functions (used with PROPMAP_* macro's)
 */

ECPropMapEntry::ECPropMapEntry(GUID guid, ULONG ulId) { 
    m_sMAPINameId.ulKind = MNID_ID; 
    m_sGuid = guid;  
    m_sMAPINameId.lpguid = &m_sGuid; 
    m_sMAPINameId.Kind.lID = ulId; 
}
    
ECPropMapEntry::ECPropMapEntry(GUID guid, char *strId) { 
    m_sMAPINameId.ulKind = MNID_ID; 
    m_sGuid = guid;  
    m_sMAPINameId.lpguid = &m_sGuid; 
    m_sMAPINameId.Kind.lpwstrName = new WCHAR[strlen(strId)+1];
    mbstounicode(m_sMAPINameId.Kind.lpwstrName, strId, strlen(strId)+1);
}
    
ECPropMapEntry::ECPropMapEntry(const ECPropMapEntry &other) { 
    m_sMAPINameId.ulKind = other.m_sMAPINameId.ulKind;
    m_sGuid = other.m_sGuid;
    m_sMAPINameId.lpguid = &m_sGuid;
    if(other.m_sMAPINameId.ulKind == MNID_ID) {
        m_sMAPINameId.Kind.lID = other.m_sMAPINameId.Kind.lID;
    } else {
        m_sMAPINameId.Kind.lpwstrName = new WCHAR[unicodelen( other.m_sMAPINameId.Kind.lpwstrName )+1];
        unicodecpy(m_sMAPINameId.Kind.lpwstrName, other.m_sMAPINameId.Kind.lpwstrName);
    }
}

ECPropMapEntry::~ECPropMapEntry()
{
    if(m_sMAPINameId.ulKind == MNID_STRING && m_sMAPINameId.Kind.lpwstrName)
        delete [] m_sMAPINameId.Kind.lpwstrName;
}
    
MAPINAMEID* ECPropMapEntry::GetMAPINameId() { 
    return &m_sMAPINameId; 
}

ECPropMap::ECPropMap() { 
}
    
void ECPropMap::AddProp(ULONG *lpId, ULONG ulType, ECPropMapEntry entry) {
    // Add reference to proptag for later Resolve();
    lstNames.push_back(entry);
    lstVars.push_back(lpId);
    lstTypes.push_back(ulType);
}
    
HRESULT ECPropMap::Resolve(IMAPIProp *lpMAPIProp) {
    HRESULT hr = hrSuccess;
    std::list<ECPropMapEntry>::iterator i;
    std::list<ULONG *>::iterator j;
    std::list<ULONG>::iterator k;
    int n = 0;
    LPSPropTagArray lpPropTags = NULL;
    
    // Do GetIDsFromNames() and store result in correct places
    MAPINAMEID **lppNames = new MAPINAMEID *[lstNames.size()];
    
    for(i=lstNames.begin(); i != lstNames.end(); i++) {
        lppNames[n++] = i->GetMAPINameId();
    }
    
    hr = lpMAPIProp->GetIDsFromNames(n, lppNames, MAPI_CREATE, &lpPropTags);
    if(hr != hrSuccess)
        goto exit;
    
    n = 0;
    k = lstTypes.begin();
    for(j=lstVars.begin(); j != lstVars.end(); j++, k++) {
        *(*j) = CHANGE_PROP_TYPE(lpPropTags->aulPropTag[n++], *k);
    }
    
exit:
    if(lpPropTags)
        MAPIFreeBuffer(lpPropTags);
    if(lppNames)
        delete [] lppNames;
        
    return hr;        
}
