/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "ECIConv.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// HACK: prototypes may differ depending on the compiler and/or system (the
// second parameter may or may not be 'const'). This redeclaration is a hack
// to have a common prototype "iconv_cast".
class iconv_HACK
{
public:
	iconv_HACK(const char** ptr) : m_ptr(ptr) { }

	// the compiler will choose the right operator
	operator const char**() { return m_ptr; }
	operator char**() { return const_cast <char**>(m_ptr); }

private:
	const char** m_ptr;
};

ECIConv::ECIConv(const std::string &strToCharset, const std::string &strFromCharset) {
    this->m_cd = iconv_open(strToCharset.c_str(), strFromCharset.c_str());
}

ECIConv::~ECIConv()
{
    if(m_cd != (iconv_t)(-1))
        iconv_close(m_cd);
}

bool ECIConv::canConvert() {
	return (m_cd != (iconv_t)(-1));
}

std::string ECIConv::convert(const std::string &strinput)
{
    std::string stroutput;
    const char *lpSrc = NULL;
    char *lpDst = NULL;
    size_t cbSrc = 0;
    size_t cbDst = 0;
    size_t err = 0;
    char buf[8192];
    
    // Return unconverted string if charset is unsupported
    if(m_cd == (iconv_t)(-1)) {
        return strinput;
    }
    
    lpSrc = strinput.c_str();
    cbSrc = strinput.length();
    
    while(cbSrc) {
        lpDst = (char *)buf;
        cbDst = sizeof(buf);
        err = iconv(m_cd, iconv_HACK(&lpSrc), &cbSrc, &lpDst, &cbDst);
        
        if(err == (size_t)(-1) && cbDst == sizeof(buf)) {
            return stroutput;
        }
        
        stroutput.append(buf, sizeof(buf)-cbDst);
    }

	// for statefull encodings, we need to flush the iconv state
	lpDst = (char *)buf;
	cbDst = sizeof(buf);
	iconv(m_cd, NULL, NULL, &lpDst, &cbDst);
	stroutput.append(buf, sizeof(buf)-cbDst);
    
    return stroutput;
}
