/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <mapidefs.h>
#include <mapiutil.h>
#include <mapispi.h>

#include <string>
#include <stack>
#include <set>

#include "edkmdb.h"

#include "Util.h"
#include "ECIConv.h"
#include "CommonUtil.h"
#include "stringutil.h"

#include "IECSingleInstance.h"
#include "ECGuid.h"

#include "rtfutil.h"

#include "mapiext.h"

#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#ifndef NO_GETTEXT
#include <libintl.h>
#define _(string) dcgettext ("zarafa", string, LC_MESSAGES)
#else
#define _(string) string
#endif

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define BLOCKSIZE	65536

struct _HTMLEntity {
	char *s;
	int c;
} HTMLEntity[] = {
{"quot"	,34},
{"amp"	,38},
{"lt"	,60},
{"gt"	,62},
{"nbsp"	,160},
{"iexcl"	,161},
{"curren"	,164},
{"cent"	,162},
{"pound"	,163},
{"yen"	,165},
{"brvbar"	,166},
{"sect"	,167},
{"uml"	,168},
{"copy"	,169},
{"ordf"	,170},
{"laquo"	,171},
{"not"	,172},
{"shy"	,173},
{"reg"	,174},
{"trade"	,8482},
{"macr"	,175},
{"deg"	,176},
{"plusmn"	,177},
{"sup2"	,178},
{"sup3"	,179},
{"acute"	,180},
{"micro"	,181},
{"para"	,182},
{"middot"	,183},
{"cedil"	,184},
{"sup1"	,185},
{"ordm"	,186},
{"raquo"	,187},
{"frac14"	,188},
{"frac12"	,189},
{"frac34"	,190},
{"iquest"	,191},
{"times"	,215},
{"divide"	,247},
{"Agrave"	,192},
{"Aacute"	,193},
{"Acirc"	,194},
{"Atilde"	,195},
{"Auml"	,196},
{"Aring"	,197},
{"AElig"	,198},
{"Ccedil"	,199},
{"Egrave"	,200},
{"Eacute"	,201},
{"Ecirc"	,202},
{"Euml"	,203},
{"Igrave"	,204},
{"Iacute"	,205},
{"Icirc"	,206},
{"Iuml"	,207},
{"ETH"	,208},
{"Ntilde"	,209},
{"Ograve"	,210},
{"Oacute"	,211},
{"Ocirc"	,212},
{"Otilde"	,213},
{"Ouml"	,214},
{"Oslash"	,216},
{"Ugrave"	,217},
{"Uacute"	,218},
{"Ucirc"	,219},
{"Uuml"	,220},
{"Yacute"	,221},
{"THORN"	,222},
{"szlig"	,223},
{"agrave"	,224},
{"aacute"	,225},
{"acirc"	,226},
{"atilde"	,227},
{"auml"	,228},
{"aring"	,229},
{"aelig"	,230},
{"ccedil"	,231},
{"egrave"	,232},
{"eacute"	,233},
{"ecirc"	,234},
{"euml"	,235},
{"igrave"	,236},
{"iacute"	,237},
{"icirc"	,238},
{"iuml"	,239},
{"eth"	,240},
{"ntilde"	,241},
{"ograve"	,242},
{"oacute"	,243},
{"ocirc"	,244},
{"otilde"	,245},
{"ouml"	,246},
{"oslash"	,248},
{"ugrave"	,249},
{"uacute"	,250},
{"ucirc"	,251},
{"uuml"	,252},
{"yacute"	,253},
{"thorn"	,254},
{"yuml"	,255},
{"euro", 128} // Euro is 0x80 in windows-1252, which is what HTMLToText converts to
};

class PropTagCompare
{
public:
	PropTagCompare() {}
	bool operator()(ULONG lhs, ULONG rhs) const { 
		if (PROP_TYPE(lhs) == PT_UNSPECIFIED || PROP_TYPE(rhs) == PT_UNSPECIFIED)
			return PROP_ID(lhs) < PROP_ID(rhs); 
		return lhs < rhs;
	}
};

typedef std::set<ULONG,PropTagCompare> PropTagSet;

std::string Util::HtmlEntityFromChar(unsigned char c){
	std::string html;
	bool bHTML;
	unsigned int j=0;
	
	switch(c) {
		case 0:
			break;
		case '\r':
			break;
		case '\n':
			html = "<br>\n";
			break;
		case '\t':
			html = "&nbsp;&nbsp;&nbsp; ";
			break;
		case ' ':
			html = "&nbsp;";
			break;
		default:
			bHTML = false;

			for(j=0;j<sizeof(HTMLEntity)/sizeof(HTMLEntity[0]);j++) {
				if(HTMLEntity[j].c == c) {
					html = "&";
					html += HTMLEntity[j].s;
					html += ";";
					bHTML = true;
					break;
				}
			}

			if(bHTML == false) {
				html.assign(1, c);
			}

			break;
	}
	return html;
}

/**
 * Validate HTML entity
 *
 * Valid:
 *   &{#100 | #x64 | amp};test
 */
bool Util::validateHtmlEntity(const std::string &strEntity)
{
	if(strEntity.size() < 3 || strEntity[0] != '&')
		return false;
	
	size_t pos = strEntity.find(';');

	if (pos == std::string::npos || pos < 3)
		return false;
	
	std::string str;

	if (strEntity[1] == '#') {
		str = strEntity.substr(2, pos-2);

		if(str[0] == 'x') {
			if (strtoul(str.c_str()+1, NULL, 16) != 0)
				return true;
		} else {
			if(atoui((char*)str.c_str()) != 0)
				return true;
		}

	} else {
		str = strEntity.substr(1, pos-2);
		for(int i=0; i<sizeof(HTMLEntity)/sizeof(HTMLEntity[0]);i++) {
			if(strcmp(HTMLEntity[i].s, str.c_str()) == 0)
				return true;
		}
	}

	return false;
}

char Util::CharFromHtmlEntity(const std::string &strEntity) {
    if(strEntity[0] == '#') {
        // We have a unicode number, use iconv to get the windows-1252 char
        unsigned int ulCode;
        std::string strUnicode;
        ECIConv conv("windows-1252", "UCS-4LE");
        
		if(strEntity.size() > 2 && strEntity[1] == 'x')
			ulCode = strtoul(strEntity.substr(2).c_str(), NULL, 16);
		else
			ulCode = atoui((char *)strEntity.substr(1).c_str());

        strUnicode.append(1,(ulCode & 0xff));
        strUnicode.append(1,(ulCode >> 8) & 0xff);
        strUnicode.append(1,(ulCode >> 16) & 0xff);
        strUnicode.append(1,(ulCode >> 24) & 0xff);
        return conv.convert(strUnicode)[0];
    } else {
        for(int i=0; i<sizeof(HTMLEntity)/sizeof(HTMLEntity[0]);i++) {
            if(strcmp(HTMLEntity[i].s, strEntity.c_str()) == 0)
                return HTMLEntity[i].c;
        }
    }
    
    return '?';
}

std::string Util::GetServerNameFromPath(const char *szPath) {
	std::string path = szPath;
	int pos, pos2;

	pos = path.find("://");
	pos2 = path.find(":", pos+3);

	if(pos2 == -1)
		pos2 = path.size();

	if(pos != -1 && pos2 != -1){
		pos+=3;
		path = path.substr(pos, pos2-pos);
	} 

	return path;
}

std::string Util::GetServerTypeFromPath(const char *szPath) {
	std::string path = szPath;
	int pos;

	pos = path.find(":");

	if(pos != -1){
		path = path.substr(0, pos);
	} else {
		path = "";
	}

	return path;
}

std::string Util::GetServerPortFromPath(const char *szPath) {
	std::string path = szPath;
	int pos,pos2;

	if(path.find("http://") == string::npos && path.find("https://") == string::npos)
		return "";

	pos = path.rfind(":");
	pos2 = path.rfind("/zarafa");

	if(pos != -1 && pos2 != -1){
		pos+=1;
		path = path.substr(pos, pos2-pos);
	}else
		path = "";

	return path;
}

std::string Util::ServerNamePortToURL(const char *lpszType, const char *lpszServerName, const char *lpszServerPort) {
	std::string strURL;

	strURL = (std::string)lpszType + "://"+(std::string)lpszServerName;

	if(strlen(lpszServerPort) > 0)
		strURL += (std::string)":"+(std::string)lpszServerPort;

	if(strnicmp(lpszType,"http", 4) == 0)
		strURL += (std::string)"/zarafa";

	return strURL;
}

// Add or replaces a prop value in a an SPropValue array
HRESULT Util::HrAddToPropertyArray(LPSPropValue lpSrc, ULONG cValues, LPSPropValue lpToAdd, LPSPropValue *lppDest, ULONG *cDestValues)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpDest = NULL;
	LPSPropValue lpFind = NULL;
	unsigned int i = 0;
	unsigned int n = 0;

	MAPIAllocateBuffer(sizeof(SPropValue) * (cValues + 1), (void **)&lpDest);

	for(i=0;i<cValues;i++) {
		hr = HrCopyProperty(&lpDest[n], &lpSrc[i], lpDest);

		if(hr == hrSuccess)
			n++;

		hr = hrSuccess;
	}

	lpFind = PpropFindProp(lpDest, n, lpToAdd->ulPropTag);

	if(lpFind) {
		hr = HrCopyProperty(lpFind, lpToAdd, lpDest);
	} else {

		hr = HrCopyProperty(&lpDest[n], lpToAdd, lpDest);

		n++;
	}

	if(hr != hrSuccess)
		goto exit;

	*lppDest = lpDest;
	*cDestValues = n;
exit:
	return hr;
}

// Merges to proptag arrays, lpAdds properties may overwrite properties from lpSrc
HRESULT	Util::HrMergePropertyArrays(LPSPropValue lpSrc, ULONG cValues, LPSPropValue lpAdds, ULONG cAddValues, LPSPropValue *lppDest, ULONG *cDestValues)
{
	HRESULT hr = hrSuccess;
	map<ULONG, LPSPropValue> mapPropSource;
	map<ULONG, LPSPropValue>::iterator iterPropSource;
	ULONG i = 0;
	LPSPropValue lpProps = NULL;

	for (i = 0; i < cValues; i ++) {
		mapPropSource[lpSrc[i].ulPropTag] = &lpSrc[i];
	}

	for (i = 0; i < cAddValues; i ++) {
		mapPropSource[lpAdds[i].ulPropTag] = &lpAdds[i];
	}

	MAPIAllocateBuffer(sizeof(SPropValue)*mapPropSource.size(), (void**)&lpProps);

	for (i = 0, iterPropSource = mapPropSource.begin(); iterPropSource != mapPropSource.end(); iterPropSource++, i++) {
		hr = Util::HrCopyProperty(&lpProps[i], iterPropSource->second, lpProps);
		if (hr != hrSuccess)
			goto exit;
	}

	*cDestValues = i;
	*lppDest = lpProps;
	lpProps = NULL;

exit:
	if (lpProps)
		MAPIFreeBuffer(lpProps);

	return hr;
}

// Copies a whole array of properties, but leaves the external data where it is (ie binary, string data is not copied)
HRESULT Util::HrCopyPropertyArrayByRef(LPSPropValue lpSrc, ULONG cValues, LPSPropValue *lppDest, ULONG *cDestValues, bool bExcludeErrors)
{
    HRESULT hr = hrSuccess;
    LPSPropValue lpDest = NULL;
    unsigned int i = 0;
    unsigned int n = 0;
    
    MAPIAllocateBuffer(sizeof(SPropValue) * cValues, (void **)&lpDest);
    
    for(i=0;i<cValues;i++) {
        if(!bExcludeErrors || PROP_TYPE(lpSrc[i].ulPropTag) != PT_ERROR) {
            hr = HrCopyPropertyByRef(&lpDest[n], &lpSrc[i]);
        
            if(hr == hrSuccess)
                n++;
            
            hr = hrSuccess;
        }
    }
    
    *lppDest = lpDest;
	*cDestValues = n;
    
    return hr;
}

// Copies whole array of properties
HRESULT Util::HrCopyPropertyArray(LPSPropValue lpSrc, ULONG cValues, LPSPropValue *lppDest, ULONG *cDestValues, bool bExcludeErrors)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpDest = NULL;
	unsigned int i = 0;
	unsigned int n = 0;

	MAPIAllocateBuffer(sizeof(SPropValue) * cValues, (void **)&lpDest);

	for(i=0;i<cValues;i++) {
	    if(!bExcludeErrors || PROP_TYPE(lpSrc[i].ulPropTag) != PT_ERROR) {
    		hr = HrCopyProperty(&lpDest[n], &lpSrc[i], lpDest);

	    	if(hr == hrSuccess)
		    	n++;

    		hr = hrSuccess;
        }
	}

	*lppDest = lpDest;
	*cDestValues = n;

	return hr;
	// FIXME potential incomplete copy possible, should free data (?)

}

HRESULT Util::HrCopyPropertyArray(LPSPropValue lpSrc, ULONG cValues, LPSPropValue lpDest, void *lpBase)
{
	HRESULT hr = hrSuccess;
	unsigned int i;

	for(i=0;i<cValues;i++) {
		hr = HrCopyProperty(&lpDest[i], &lpSrc[i], lpBase);

		if(hr != hrSuccess)
			goto exit;
	}

exit:
	return hr;
}

// Copies one property to somewhere else, but doesn't copy external data (ie binary or string data)
HRESULT Util::HrCopyPropertyByRef(LPSPropValue lpDest, LPSPropValue lpSrc)
{
    // Just a simple memcpy !
    memcpy(lpDest, lpSrc, sizeof(SPropValue));
    
    return hrSuccess;
}

// Copies one property to somewhere else, alloc'ing space if required
HRESULT Util::HrCopyProperty(LPSPropValue lpDest, LPSPropValue lpSrc, void *lpBase, ALLOCATEMORE * lpfAllocMore)
{
	HRESULT hr = hrSuccess;

	if(lpfAllocMore == NULL)
		lpfAllocMore = MAPIAllocateMore;

	switch(PROP_TYPE(lpSrc->ulPropTag)) {	
	case PT_I2:
		lpDest->Value.i = lpSrc->Value.i;
		break;
	case PT_LONG:
		lpDest->Value.ul = lpSrc->Value.ul;
		break;
	case PT_BOOLEAN:
		lpDest->Value.b = lpSrc->Value.b;
		break;
	case PT_R4:
		lpDest->Value.flt = lpSrc->Value.flt;
		break;
	case PT_DOUBLE:
		lpDest->Value.dbl = lpSrc->Value.dbl;
		break;
	case PT_APPTIME:
		lpDest->Value.at = lpSrc->Value.at;
		break;
	case PT_CURRENCY:
		lpDest->Value.cur = lpSrc->Value.cur;
		break;
	case PT_SYSTIME:
		lpDest->Value.ft = lpSrc->Value.ft;
		break;
	case PT_I8:
		lpDest->Value.li = lpSrc->Value.li;
		break;
	case PT_UNICODE:
		if(lpSrc->Value.lpszW == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpfAllocMore(unicodelen(lpSrc->Value.lpszW)*sizeof(WCHAR)+sizeof(WCHAR), lpBase, (void**)&lpDest->Value.lpszW);
		unicodecpy(lpDest->Value.lpszW, lpSrc->Value.lpszW);
		break;
	case PT_STRING8:
		if(lpSrc->Value.lpszA == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpfAllocMore(strlen(lpSrc->Value.lpszA)*sizeof(TCHAR)+sizeof(TCHAR), lpBase, (void**)&lpDest->Value.lpszA);
		lstrcpy(lpDest->Value.lpszA, lpSrc->Value.lpszA);
		break;
	case PT_BINARY:
		if(lpSrc->Value.bin.cb > 0)
			lpfAllocMore(lpSrc->Value.bin.cb, lpBase, (void **) &lpDest->Value.bin.lpb);
		
		lpDest->Value.bin.cb = lpSrc->Value.bin.cb;
		
		if(lpSrc->Value.bin.cb > 0)
			memcpy(lpDest->Value.bin.lpb, lpSrc->Value.bin.lpb, lpSrc->Value.bin.cb);
		else
			lpDest->Value.bin.lpb = NULL;

		break;
	case PT_CLSID:
		lpfAllocMore(sizeof(GUID), lpBase, (void **)&lpDest->Value.lpguid);
		memcpy(lpDest->Value.lpguid, lpSrc->Value.lpguid, sizeof(GUID));
		break;
	case PT_ERROR:
		lpDest->Value.err = lpSrc->Value.err;
		break;
	case PT_SRESTRICTION:
		if(lpSrc->Value.lpszA == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		// NOTE: we place the object pointer in lpszA to make sure it's on the same offset as Value.x on 32bit as 64bit machines
		lpfAllocMore(sizeof(SRestriction), lpBase, (void **)&lpDest->Value.lpszA);
		hr = Util::HrCopySRestriction((LPSRestriction)lpDest->Value.lpszA, (LPSRestriction)lpSrc->Value.lpszA, lpBase);
		break;
	case PT_ACTIONS:
		if(lpSrc->Value.lpszA == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		// NOTE: we place the object pointer in lpszA to make sure it's on the same offset as Value.x on 32bit as 64bit machines
		lpfAllocMore(sizeof(ACTIONS), lpBase, (void **)&lpDest->Value.lpszA);
		hr = Util::HrCopyActions((ACTIONS *)lpDest->Value.lpszA, (ACTIONS *)lpSrc->Value.lpszA, lpBase);
		break;
	case PT_NULL:
		break;
	case PT_OBJECT:
		lpDest->Value.x = 0;
		break;
	// MV properties
	case PT_MV_I2:
		lpfAllocMore(sizeof(short int) * lpSrc->Value.MVi.cValues, lpBase, (void **)&lpDest->Value.MVi.lpi);
		memcpy(lpDest->Value.MVi.lpi, lpSrc->Value.MVi.lpi, sizeof(short int) * lpSrc->Value.MVi.cValues);
		lpDest->Value.MVi.cValues = lpSrc->Value.MVi.cValues;
		break;
	case PT_MV_LONG:
		lpfAllocMore(sizeof(LONG) * lpSrc->Value.MVl.cValues, lpBase, (void **)&lpDest->Value.MVl.lpl);
		memcpy(lpDest->Value.MVl.lpl, lpSrc->Value.MVl.lpl, sizeof(LONG) * lpSrc->Value.MVl.cValues);
		lpDest->Value.MVl.cValues = lpSrc->Value.MVl.cValues;
		break;
	case PT_MV_FLOAT:
		lpfAllocMore(sizeof(float) * lpSrc->Value.MVflt.cValues, lpBase, (void **)&lpDest->Value.MVflt.lpflt);
		memcpy(lpDest->Value.MVflt.lpflt, lpSrc->Value.MVflt.lpflt, sizeof(float) * lpSrc->Value.MVflt.cValues);
		lpDest->Value.MVflt.cValues = lpSrc->Value.MVflt.cValues;
		break;
	case PT_MV_DOUBLE:
	case PT_MV_APPTIME:
		lpfAllocMore(sizeof(double) * lpSrc->Value.MVdbl.cValues, lpBase, (void **)&lpDest->Value.MVdbl.lpdbl);
		memcpy(lpDest->Value.MVdbl.lpdbl, lpSrc->Value.MVdbl.lpdbl, sizeof(double) * lpSrc->Value.MVdbl.cValues);
		lpDest->Value.MVdbl.cValues = lpSrc->Value.MVdbl.cValues;
		break;
	case PT_MV_I8:
		lpfAllocMore(sizeof(LONGLONG) * lpSrc->Value.MVli.cValues, lpBase, (void **)&lpDest->Value.MVli.lpli);
		memcpy(lpDest->Value.MVli.lpli, lpSrc->Value.MVli.lpli, sizeof(LONGLONG) * lpSrc->Value.MVli.cValues);
		lpDest->Value.MVli.cValues = lpSrc->Value.MVli.cValues;
		break;
	case PT_MV_CURRENCY:
		lpfAllocMore(sizeof(CURRENCY) * lpSrc->Value.MVcur.cValues, lpBase, (void **)&lpDest->Value.MVcur.lpcur);
		memcpy(lpDest->Value.MVcur.lpcur, lpSrc->Value.MVcur.lpcur, sizeof(CURRENCY) * lpSrc->Value.MVcur.cValues);
		lpDest->Value.MVcur.cValues = lpSrc->Value.MVcur.cValues;
		break;
	case PT_MV_SYSTIME:
		lpfAllocMore(sizeof(FILETIME) * lpSrc->Value.MVft.cValues, lpBase, (void **)&lpDest->Value.MVft.lpft);
		memcpy(lpDest->Value.MVft.lpft, lpSrc->Value.MVft.lpft, sizeof(FILETIME) * lpSrc->Value.MVft.cValues);
		lpDest->Value.MVft.cValues = lpSrc->Value.MVft.cValues;
		break;
	case PT_MV_STRING8:
		lpfAllocMore(sizeof(LPSTR *) * lpSrc->Value.MVszA.cValues, lpBase, (void **)&lpDest->Value.MVszA.lppszA);
		for (ULONG i = 0; i < lpSrc->Value.MVszA.cValues; i++) {
			int datalength = strlen(lpSrc->Value.MVszA.lppszA[i]) + 1;
			lpfAllocMore(datalength, lpBase, (void **)&lpDest->Value.MVszA.lppszA[i]);
			memcpy(lpDest->Value.MVszA.lppszA[i], lpSrc->Value.MVszA.lppszA[i], datalength);
		}
		lpDest->Value.MVszA.cValues = lpSrc->Value.MVszA.cValues;
		break;
	case PT_MV_UNICODE:
		lpfAllocMore(sizeof(LPWSTR *) * lpSrc->Value.MVszW.cValues, lpBase, (void **)&lpDest->Value.MVszW.lppszW);
		for (ULONG i = 0; i < lpSrc->Value.MVszW.cValues; i++) {
			lpfAllocMore(unicodelen(lpSrc->Value.MVszW.lppszW[i]) * sizeof(WCHAR) + sizeof(WCHAR), lpBase, (void**)&lpDest->Value.MVszW.lppszW[i]);
			unicodecpy(lpDest->Value.MVszW.lppszW[i], lpSrc->Value.MVszW.lppszW[i]);
		}
		lpDest->Value.MVszW.cValues = lpSrc->Value.MVszW.cValues;
		break;
	case PT_MV_BINARY:
		lpfAllocMore(sizeof(SBinary) * lpSrc->Value.MVbin.cValues, lpBase, (void **)&lpDest->Value.MVbin.lpbin);
		for (ULONG i = 0; i < lpSrc->Value.MVbin.cValues; i++) {
			lpfAllocMore(lpSrc->Value.MVbin.lpbin[i].cb, lpBase, (void **)&lpDest->Value.MVbin.lpbin[i].lpb);
			memcpy(lpDest->Value.MVbin.lpbin[i].lpb, lpSrc->Value.MVbin.lpbin[i].lpb, lpSrc->Value.MVbin.lpbin[i].cb);
			lpDest->Value.MVbin.lpbin[i].cb = lpSrc->Value.MVbin.lpbin[i].cb;
		}
		lpDest->Value.MVbin.cValues = lpSrc->Value.MVbin.cValues;
		break;
	case PT_MV_CLSID:
		lpfAllocMore(sizeof(GUID) * lpSrc->Value.MVguid.cValues, lpBase, (void **)&lpDest->Value.MVguid.lpguid);
		memcpy(lpDest->Value.MVguid.lpguid, lpSrc->Value.MVguid.lpguid, sizeof(GUID) * lpSrc->Value.MVguid.cValues);
		lpDest->Value.MVguid.cValues = lpSrc->Value.MVguid.cValues;
		break;

	default:
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	lpDest->ulPropTag = lpSrc->ulPropTag;

exit:
	return hr;
}

HRESULT Util::HrCopySRestriction(LPSRestriction *lppDest, LPSRestriction lpSrc)
{
	HRESULT hr = hrSuccess;
	LPSRestriction lpDest = NULL;

	MAPIAllocateBuffer(sizeof(SRestriction), (void **)&lpDest);

	hr = HrCopySRestriction(lpDest, lpSrc, lpDest);

	if(hr != hrSuccess)
		goto exit;

	*lppDest = lpDest;
exit:
	return hr;
}

HRESULT	Util::HrCopySRestriction(LPSRestriction lpDest, LPSRestriction lpSrc, void *lpBase)
{
	HRESULT hr = hrSuccess;
	unsigned int i = 0;

	if (!lpDest || !lpSrc || !lpBase) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	lpDest->rt = lpSrc->rt;

	switch(lpSrc->rt) {
		case RES_AND:
			lpDest->res.resAnd.cRes = lpSrc->res.resAnd.cRes;
			MAPIAllocateMore(sizeof(SRestriction) * lpSrc->res.resAnd.cRes, lpBase, (void **)&lpDest->res.resAnd.lpRes);

			for(i=0;i<lpSrc->res.resAnd.cRes;i++) {
				hr = HrCopySRestriction(&lpDest->res.resAnd.lpRes[i], &lpSrc->res.resAnd.lpRes[i], lpBase);
				if(hr != hrSuccess)
					goto exit;
			}
			break;
		case RES_OR:
			lpDest->res.resOr.cRes = lpSrc->res.resOr.cRes;
			MAPIAllocateMore(sizeof(SRestriction) * lpSrc->res.resOr.cRes, lpBase, (void **)&lpDest->res.resOr.lpRes);

			for(i=0;i<lpSrc->res.resOr.cRes;i++) {
				hr = HrCopySRestriction(&lpDest->res.resOr.lpRes[i], &lpSrc->res.resOr.lpRes[i], lpBase);
				if(hr != hrSuccess)
					goto exit;
			}
			break;
		case RES_NOT:
			MAPIAllocateMore(sizeof(SRestriction), lpBase, (void **) &lpDest->res.resNot.lpRes);
			hr = HrCopySRestriction(lpDest->res.resNot.lpRes, lpSrc->res.resNot.lpRes, lpBase);
			
			break;
		case RES_CONTENT:
			lpDest->res.resContent.ulFuzzyLevel = lpSrc->res.resContent.ulFuzzyLevel;
			lpDest->res.resContent.ulPropTag = lpSrc->res.resContent.ulPropTag;
			MAPIAllocateMore(sizeof(SPropValue), lpBase, (void **) &lpDest->res.resContent.lpProp);
			hr = HrCopyProperty(lpDest->res.resContent.lpProp, lpSrc->res.resContent.lpProp, lpBase);
			break;
		case RES_PROPERTY:
			lpDest->res.resProperty.relop = lpSrc->res.resProperty.relop;
			lpDest->res.resProperty.ulPropTag = lpSrc->res.resProperty.ulPropTag;
			MAPIAllocateMore(sizeof(SPropValue), lpBase, (void **) &lpDest->res.resProperty.lpProp);
			hr = HrCopyProperty(lpDest->res.resProperty.lpProp, lpSrc->res.resProperty.lpProp, lpBase);
			break;
		case RES_COMPAREPROPS:	 
			lpDest->res.resCompareProps.relop = lpSrc->res.resCompareProps.relop;
			lpDest->res.resCompareProps.ulPropTag1 = lpSrc->res.resCompareProps.ulPropTag1;
			lpDest->res.resCompareProps.ulPropTag2 = lpSrc->res.resCompareProps.ulPropTag2;
			break;
		case RES_BITMASK:			 
			lpDest->res.resBitMask.relBMR = lpSrc->res.resBitMask.relBMR;
			lpDest->res.resBitMask.ulMask = lpSrc->res.resBitMask.ulMask;
			lpDest->res.resBitMask.ulPropTag = lpSrc->res.resBitMask.ulPropTag;
			break;
		case RES_SIZE:			
			lpDest->res.resSize.cb = lpSrc->res.resSize.cb;
			lpDest->res.resSize.relop = lpSrc->res.resSize.relop;
			lpDest->res.resSize.ulPropTag = lpSrc->res.resSize.ulPropTag;
			break;
		case RES_EXIST:			
			lpDest->res.resExist.ulPropTag = lpSrc->res.resExist.ulPropTag;
			break;
		case RES_SUBRESTRICTION:	
			lpDest->res.resSub.ulSubObject = lpSrc->res.resSub.ulSubObject;
			MAPIAllocateMore(sizeof(SRestriction), lpBase, (void **)&lpDest->res.resSub.lpRes);
			hr = HrCopySRestriction(lpDest->res.resSub.lpRes, lpSrc->res.resSub.lpRes, lpBase);
			break;
		case RES_COMMENT: // What a weird restriction type
			lpDest->res.resComment.cValues	= lpSrc->res.resComment.cValues;
			lpDest->res.resComment.lpRes	= NULL;

			if(lpSrc->res.resComment.cValues > 0)
			{
				MAPIAllocateMore(sizeof(SPropValue) * lpSrc->res.resComment.cValues, lpBase, (void **) &lpDest->res.resComment.lpProp);

				hr = HrCopyPropertyArray(lpSrc->res.resComment.lpProp, lpSrc->res.resComment.cValues, lpDest->res.resComment.lpProp, lpBase);
				if(hr != hrSuccess)
					goto exit;
			}

			if(lpSrc->res.resComment.lpRes) {
				MAPIAllocateMore(sizeof(SRestriction), lpBase, (void **) &lpDest->res.resComment.lpRes);
				hr = HrCopySRestriction(lpDest->res.resComment.lpRes, lpSrc->res.resComment.lpRes, lpBase);
			}

			break;
	}

exit:
	return hr;
}

HRESULT	Util::HrCopyActions(ACTIONS * lpDest, ACTIONS * lpSrc, void *lpBase)
{
	HRESULT hr = hrSuccess;
	unsigned int i;

	lpDest->cActions = lpSrc->cActions;
	lpDest->ulVersion = lpSrc->ulVersion;
	MAPIAllocateMore(sizeof(ACTION) * lpSrc->cActions, lpBase, (void **)&lpDest->lpAction);

	memset(lpDest->lpAction, 0, sizeof(ACTION) * lpSrc->cActions);

	for(i=0;i<lpSrc->cActions;i++) {
		hr = HrCopyAction(&lpDest->lpAction[i], &lpSrc->lpAction[i], lpBase);
		if(hr != hrSuccess)
			goto exit;
	}

exit:
	return hr;
}

HRESULT	Util::HrCopyAction(ACTION * lpDest, ACTION * lpSrc, void *lpBase)
{
	HRESULT hr = hrSuccess;

	lpDest->acttype = lpSrc->acttype;
	lpDest->ulActionFlavor = lpSrc->ulActionFlavor;
	lpDest->lpRes = NULL; // also unused
	lpDest->lpPropTagArray = NULL; // unused according to edkmdb.h
	lpDest->ulFlags = lpSrc->ulFlags;

	switch(lpSrc->acttype) {
		case OP_MOVE:
		case OP_COPY:
			lpDest->actMoveCopy.cbStoreEntryId = lpSrc->actMoveCopy.cbStoreEntryId;
			MAPIAllocateMore(lpSrc->actMoveCopy.cbStoreEntryId, lpBase, (void **) &lpDest->actMoveCopy.lpStoreEntryId);
			memcpy(lpDest->actMoveCopy.lpStoreEntryId, lpSrc->actMoveCopy.lpStoreEntryId, lpSrc->actMoveCopy.cbStoreEntryId);

			lpDest->actMoveCopy.cbFldEntryId = lpSrc->actMoveCopy.cbFldEntryId;
			MAPIAllocateMore(lpSrc->actMoveCopy.cbFldEntryId, lpBase, (void **) &lpDest->actMoveCopy.lpFldEntryId);
			memcpy(lpDest->actMoveCopy.lpFldEntryId, lpSrc->actMoveCopy.lpFldEntryId, lpSrc->actMoveCopy.cbFldEntryId);

			break;
		case OP_REPLY:
		case OP_OOF_REPLY:
			lpDest->actReply.cbEntryId = lpSrc->actReply.cbEntryId;
			MAPIAllocateMore(lpSrc->actReply.cbEntryId, lpBase, (void **) &lpDest->actReply.lpEntryId);
			memcpy(lpDest->actReply.lpEntryId, lpSrc->actReply.lpEntryId, lpSrc->actReply.cbEntryId);

			lpDest->actReply.guidReplyTemplate = lpSrc->actReply.guidReplyTemplate;
			break;
		case OP_DEFER_ACTION:
			lpDest->actDeferAction.cbData = lpSrc->actDeferAction.cbData;
			MAPIAllocateMore(lpSrc->actDeferAction.cbData, lpBase, (void **)&lpDest->actDeferAction.pbData);
			memcpy(lpDest->actDeferAction.pbData, lpSrc->actDeferAction.pbData, lpSrc->actDeferAction.cbData);

			break;
		case OP_BOUNCE:
			lpDest->scBounceCode = lpSrc->scBounceCode;
			break;
		case OP_FORWARD:
		case OP_DELEGATE:
			MAPIAllocateMore(CbNewADRLIST(lpSrc->lpadrlist->cEntries), lpBase, (void **)&lpDest->lpadrlist);
			hr = HrCopySRowSet((LPSRowSet)lpDest->lpadrlist, (LPSRowSet)lpSrc->lpadrlist, lpBase);
			break;
		case OP_TAG:
			hr = HrCopyProperty(&lpDest->propTag, &lpSrc->propTag, lpBase);
			break;
		case OP_DELETE:
		case OP_MARK_AS_READ:
			 break;
		default:
			break;
	}

	return hr;
}

HRESULT Util::HrCopySRowSet(LPSRowSet lpDest, LPSRowSet lpSrc, void *lpBase)
{
	HRESULT hr = hrSuccess;
	unsigned int i;

	lpDest->cRows = 0;
	for(i=0;i<lpSrc->cRows;i++) {
		hr = HrCopySRow(&lpDest->aRow[i], &lpSrc->aRow[i], lpBase);
		if(hr != hrSuccess)
			goto exit;

		lpDest->cRows++;
	}

exit:
	return hr;
}

HRESULT	Util::HrCopySRow(LPSRow lpDest, LPSRow lpSrc, void *lpBase)
{
	HRESULT hr = hrSuccess;

	lpDest->cValues = lpSrc->cValues;
	// According to MSDN, rows in an SRowSet should use seperate
	// MAPIAllocate() calls (not MAPIAllocateMore) so that rows can
	// be freed individually.
	// Make sure to free your RowSet with FreeProws(), which frees the
	// rows individually.

	// However, when you have a rowset within a rowset (eg. lpadrlist in OP_FORWARD and OP_DELEGATE rules)
	// these need to be allocated to the original row, and not separate

	if (lpBase)
		MAPIAllocateMore(sizeof(SPropValue) * lpSrc->cValues, lpBase, (void **) &lpDest->lpProps);
	else
		MAPIAllocateBuffer(sizeof(SPropValue) * lpSrc->cValues, (void **) &lpDest->lpProps);
    
	hr = HrCopyPropertyArray(lpSrc->lpProps, lpSrc->cValues, lpDest->lpProps, lpBase ? lpBase : lpDest->lpProps);

	return hr;
}

HRESULT Util::HrCopyBinary(ULONG ulSize, LPBYTE lpSrc, ULONG *lpulDestSize, LPBYTE *lppDest)
{
	HRESULT hr = hrSuccess;
	LPBYTE lpDest = NULL;

	if (ulSize == 0) {
		*lpulDestSize = 0;
		*lppDest = NULL;
		goto exit;
	}

	hr = MAPIAllocateBuffer(ulSize, (void **) &lpDest);
	if(hr != hrSuccess)
		goto exit;

	memcpy(lpDest, lpSrc, ulSize);

	*lppDest = lpDest;
	*lpulDestSize = ulSize;

exit:
	if(hr != hrSuccess && lpDest)
		MAPIFreeBuffer(lpDest);

	return hr;
}

HRESULT	Util::HrCopyEntryId(ULONG ulSize, LPENTRYID lpSrc, ULONG *lpulDestSize, LPENTRYID* lppDest)
{
	return HrCopyBinary(ulSize, (LPBYTE)lpSrc, lpulDestSize, (LPBYTE *)lppDest);
}

HRESULT Util::CompareProp(LPSPropValue lpProp1, LPSPropValue lpProp2, int* lpCompareResult)
{
	HRESULT	hr = hrSuccess;
	int		nCompareResult = 0;
	unsigned int		i;

	if(lpProp1 == NULL || lpProp2 == NULL || lpCompareResult == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(PROP_TYPE(lpProp1->ulPropTag) != PROP_TYPE(lpProp2->ulPropTag)) {
		// Treat this as equal
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	switch(PROP_TYPE(lpProp1->ulPropTag)) {
	case PT_I2:
		nCompareResult = lpProp1->Value.i - lpProp2->Value.i;
		break;
	case PT_LONG:
		nCompareResult = lpProp1->Value.ul - lpProp2->Value.ul;
		break;
	case PT_R4:
		if(lpProp1->Value.flt == lpProp2->Value.flt)
			nCompareResult = 0;
		else if(lpProp1->Value.flt < lpProp2->Value.flt)
			nCompareResult = -1;
		else
			nCompareResult = 1;
		break;
	case PT_BOOLEAN:
		nCompareResult = lpProp1->Value.b - lpProp2->Value.b;
		break;
	case PT_DOUBLE:
	case PT_APPTIME:
		if(lpProp1->Value.dbl == lpProp2->Value.dbl)
			nCompareResult = 0;
		else if(lpProp1->Value.dbl < lpProp2->Value.dbl)
			nCompareResult = -1;
		else
			nCompareResult = 1;
		break;
	case PT_I8:
		nCompareResult = (int)(lpProp1->Value.li.QuadPart - lpProp2->Value.li.QuadPart);
		break;
	// FIXME: support unicode, this is broken right now
	case PT_UNICODE:
	case PT_STRING8:
		if (lpProp1->Value.lpszA && lpProp2->Value.lpszA)
			if(lpProp2->ulPropTag == PR_ANR)
				nCompareResult = stristr(lpProp1->Value.lpszA, lpProp2->Value.lpszA) != NULL;
			else
				nCompareResult = stricmp(lpProp1->Value.lpszA, lpProp2->Value.lpszA);
		else
			nCompareResult = lpProp1->Value.lpszA != lpProp2->Value.lpszA;
		break;
	case PT_SYSTIME:
	case PT_CURRENCY:
		if(lpProp1->Value.cur.Hi == lpProp2->Value.cur.Hi)
			nCompareResult = lpProp1->Value.cur.Lo - lpProp2->Value.cur.Lo;
		else
			nCompareResult = lpProp1->Value.cur.Hi - lpProp2->Value.cur.Hi;
		break;
	case PT_BINARY:
		if (lpProp1->Value.bin.lpb && lpProp2->Value.bin.lpb &&
			lpProp1->Value.bin.cb && lpProp2->Value.bin.cb &&
			lpProp1->Value.bin.cb == lpProp2->Value.bin.cb)
			nCompareResult = memcmp(lpProp1->Value.bin.lpb, lpProp2->Value.bin.lpb, lpProp1->Value.bin.cb);
		else
			nCompareResult = lpProp1->Value.bin.cb - lpProp2->Value.bin.cb;
		break;
	case PT_CLSID:
		nCompareResult = memcmp(lpProp1->Value.lpguid, lpProp2->Value.lpguid, sizeof(GUID));
		break;
		
	case PT_MV_I2:
		if (lpProp1->Value.MVi.cValues == lpProp2->Value.MVi.cValues) {
			for(i=0; i < lpProp1->Value.MVi.cValues; i++) {
				nCompareResult = lpProp1->Value.MVi.lpi[i] - lpProp2->Value.MVi.lpi[i];
				if(nCompareResult != 0)
					break;
			}
		} else
			nCompareResult = lpProp1->Value.MVi.cValues - lpProp2->Value.MVi.cValues;
		break;
	case PT_MV_LONG:
		if (lpProp1->Value.MVl.cValues == lpProp2->Value.MVl.cValues) {
			for(i=0; i < lpProp1->Value.MVl.cValues; i++) {
				nCompareResult = lpProp1->Value.MVl.lpl[i] - lpProp2->Value.MVl.lpl[i];
				if(nCompareResult != 0)
					break;
			}
		} else
			nCompareResult = lpProp1->Value.MVl.cValues - lpProp2->Value.MVl.cValues;
		break;
	case PT_MV_R4:
		if (lpProp1->Value.MVflt.cValues == lpProp2->Value.MVflt.cValues) {
			for(i=0; i < lpProp1->Value.MVflt.cValues; i++) {
				nCompareResult = lpProp1->Value.MVflt.lpflt[i] - lpProp2->Value.MVflt.lpflt[i];
				if(nCompareResult != 0)
					break;
			}
		} else
			nCompareResult = lpProp1->Value.MVflt.cValues - lpProp2->Value.MVflt.cValues;
		break;
	case PT_MV_DOUBLE:
	case PT_MV_APPTIME:
		if (lpProp1->Value.MVdbl.cValues == lpProp2->Value.MVdbl.cValues) {
			for(i=0; i < lpProp1->Value.MVdbl.cValues; i++) {
				nCompareResult = lpProp1->Value.MVdbl.lpdbl[i] - lpProp2->Value.MVdbl.lpdbl[i];
				if(nCompareResult != 0)
					break;
			}
		} else
			nCompareResult = lpProp1->Value.MVdbl.cValues - lpProp2->Value.MVdbl.cValues;
		break;
	case PT_MV_I8:
		if (lpProp1->Value.MVli.cValues == lpProp2->Value.MVli.cValues) {
			for(i=0; i < lpProp1->Value.MVli.cValues; i++) {
				nCompareResult = (int)(lpProp1->Value.MVli.lpli[i].QuadPart - lpProp2->Value.MVli.lpli[i].QuadPart);
				if(nCompareResult != 0)
					break;
			}
		} else
			nCompareResult = lpProp1->Value.MVli.cValues - lpProp2->Value.MVli.cValues;
		break;
	case PT_MV_SYSTIME:
	case PT_MV_CURRENCY:
		if (lpProp1->Value.MVcur.cValues == lpProp2->Value.MVcur.cValues) {
			for(i=0; i < lpProp1->Value.MVcur.cValues; i++) {
				if(lpProp1->Value.MVcur.lpcur[i].Hi == lpProp2->Value.MVcur.lpcur[i].Hi)
					nCompareResult = lpProp1->Value.MVcur.lpcur[i].Lo - lpProp2->Value.MVcur.lpcur[i].Lo;
				else
					nCompareResult = lpProp1->Value.MVcur.lpcur[i].Hi - lpProp2->Value.MVcur.lpcur[i].Hi;

				if(nCompareResult != 0)
					break;
			}
		} else
			nCompareResult = lpProp1->Value.MVcur.cValues == lpProp2->Value.MVcur.cValues;
		break;
	case PT_MV_CLSID:
		if (lpProp1->Value.MVguid.cValues == lpProp2->Value.MVguid.cValues) {
			nCompareResult = memcmp(lpProp1->Value.MVguid.lpguid, lpProp2->Value.MVguid.lpguid, sizeof(GUID)*lpProp1->Value.MVguid.cValues);
		} else {
			nCompareResult = lpProp1->Value.MVguid.cValues - lpProp2->Value.MVguid.cValues;
		}
		break;
	case PT_MV_BINARY:
		if (lpProp1->Value.MVbin.cValues == lpProp2->Value.MVbin.cValues) {
			for(i=0; i < lpProp1->Value.MVbin.cValues; i++) {
				if(lpProp1->Value.MVbin.lpbin[i].lpb && lpProp2->Value.MVbin.lpbin[i].lpb &&
				   lpProp1->Value.MVbin.lpbin[i].cb == lpProp2->Value.MVbin.lpbin[i].cb)
					nCompareResult = memcmp(lpProp1->Value.MVbin.lpbin[i].lpb, lpProp2->Value.MVbin.lpbin[i].lpb, lpProp1->Value.MVbin.lpbin[i].cb);
				else
					nCompareResult = lpProp1->Value.MVbin.lpbin[i].cb - lpProp2->Value.MVbin.lpbin[i].cb;

				if(nCompareResult != 0)
					break;
			}
		} else
			nCompareResult = lpProp1->Value.MVbin.cValues - lpProp2->Value.MVbin.cValues;
		break;
	// FIXME: support unicode, this is broken right now
	case PT_MV_STRING8:
	case PT_MV_UNICODE:
		if (lpProp1->Value.MVszA.cValues == lpProp2->Value.MVszA.cValues) {
			for(i=0; i < lpProp1->Value.MVszA.cValues; i++) {
				if (lpProp1->Value.MVszA.lppszA[i] && lpProp2->Value.MVszA.lppszA[i])
					nCompareResult = stricmp(lpProp1->Value.MVszA.lppszA[i], lpProp2->Value.MVszA.lppszA[i]);
				else
					nCompareResult = lpProp1->Value.MVszA.lppszA[i] != lpProp2->Value.MVszA.lppszA[i];

				if(nCompareResult != 0)
					break;
			}
		} else
			nCompareResult = lpProp1->Value.MVszA.cValues - lpProp2->Value.MVszA.cValues;
		break;
	default:
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
		break;
	}

	*lpCompareResult = nCompareResult;

exit:
	return hr;
}

unsigned int Util::PropSize(LPSPropValue lpProp)
{
	unsigned int ulSize, i;

	if(lpProp == NULL)
		return 0;

	switch(PROP_TYPE(lpProp->ulPropTag)) {
	case PT_I2:
		return 2;
	case PT_BOOLEAN:
	case PT_R4:
	case PT_LONG:
		return 4;
	case PT_APPTIME:
	case PT_DOUBLE:
	case PT_I8:
		return 8;
	case PT_UNICODE:
		return lpProp->Value.lpszW ? unicodelen(lpProp->Value.lpszW) : 0;
	case PT_STRING8:
		return lpProp->Value.lpszA ? strlen(lpProp->Value.lpszA) : 0;
	case PT_SYSTIME:
	case PT_CURRENCY:
		return 8;
	case PT_BINARY:
		return lpProp->Value.bin.cb;
	case PT_CLSID:
		return sizeof(GUID);
	case PT_MV_I2:
		return 2 * lpProp->Value.MVi.cValues;
	case PT_MV_R4:
		return 4 * lpProp->Value.MVflt.cValues;
	case PT_MV_LONG:
		return 4 * lpProp->Value.MVl.cValues;
	case PT_MV_APPTIME:
	case PT_MV_DOUBLE:
		return 8 * lpProp->Value.MVdbl.cValues;
	case PT_MV_I8:
		return 8 * lpProp->Value.MVli.cValues;
	case PT_MV_UNICODE:
		ulSize = 0;
		for(i=0; i < lpProp->Value.MVszW.cValues; i++)
			ulSize += (lpProp->Value.MVszW.lppszW[i]) ? unicodelen(lpProp->Value.MVszW.lppszW[i]) : 0;
		return ulSize;
	case PT_MV_STRING8:
		ulSize = 0;
		for(i=0; i < lpProp->Value.MVszA.cValues; i++)
			ulSize += (lpProp->Value.MVszA.lppszA[i]) ? strlen(lpProp->Value.MVszA.lppszA[i]) : 0;
		return ulSize;
	case PT_MV_SYSTIME:
	case PT_MV_CURRENCY:
		return 8 * lpProp->Value.MVcur.cValues;	
	case PT_MV_BINARY:
		ulSize = 0;
		for(i=0; i < lpProp->Value.MVbin.cValues; i++)
			ulSize+= lpProp->Value.MVbin.lpbin[i].cb;
		return ulSize;
	case PT_MV_CLSID:
		return sizeof(GUID) * lpProp->Value.MVguid.cValues;
	default:
		return 0;
	}
}

#define BUFSIZE 65536

HRESULT Util::HrTextToHtml(IStream *text, IStream *html)
{
	ULONG cRead;
	std::string strHtml;
	TBYTE c[BUFSIZE];
	LPTSTR header = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n" \
					"<HTML>\n" \
					"<HEAD>\n" \
					"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=windows-1252\">\n" \
					"<META NAME=\"Generator\" CONTENT=\"Zarafa HTML builder 1.0\">\n" \
					"<TITLE></TITLE>\n" \
					"</HEAD>\n" \
					"<BODY>\n" \
					"<!-- Converted from text/plain format -->\n" \
					"\n" \
					"<P><FONT STYLE=\"font-family: courier\" SIZE=2>\n";
					
	LPTSTR footer = "</FONT>\n" \
					"</P>\n" \
					"\n" \
					"</BODY>" \
					"</HTML>";
	ULONG i = 0;

	html->Write(header,strlen(header),NULL);
	while(1) {
		text->Read(c, BUFSIZE, &cRead);

		if(cRead == 0)
			break;
	
		for(i=0;i<cRead;i++) {
			if(c[i] == ' '){
				if((i+1) < cRead && c[i+1] == ' ')
					html->Write("&nbsp;", 6, NULL);
				else
					html->Write(" ", 1,NULL);
			}else{
				strHtml = HtmlEntityFromChar(c[i]);
				html->Write(strHtml.c_str(), strHtml.size(), NULL);
			}
		}
	}

	html->Write(footer,strlen(footer),NULL);

	return hrSuccess;
}

HRESULT Util::HrTextToRtf(IStream *text, IStream *rtf)
{
	ULONG cRead;
	TBYTE c[BUFSIZE];
	LPTSTR header = "{\\rtf1\\ansi\\ansicpg1252\\fromtext \\deff0{\\fonttbl\n" \
					"{\\f0\\fswiss Arial;}\n" \
					"{\\f1\\fmodern Courier New;}\n" \
					"{\\f2\\fnil\\fcharset2 Symbol;}\n" \
					"{\\f3\\fmodern\\fcharset0 Courier New;}}\n" \
					"{\\colortbl\\red0\\green0\\blue0;\\red0\\green0\\blue255;}\n" \
					"\\uc1\\pard\\plain\\deftab360 \\f0\\fs20 ";
					
	LPTSTR footer = "}";
	ULONG i = 0;

	rtf->Write(header,strlen(header),NULL);

	while(1) {
		text->Read(c, BUFSIZE, &cRead);

		if(cRead == 0)
			break;

		for(i=0;i<cRead;i++) {

			switch(c[i]) {
			case 0:
				break;
			case '\r':
				break;
			case '\n':
				rtf->Write("\\par\n",5,NULL);
				break;

			case '\\':
				rtf->Write("\\\\",2,NULL);
				break;

			case '{':
				rtf->Write("\\{",2,NULL);
				break;

			case '}':
				rtf->Write("\\}",2,NULL);
				break;

			case 9:
				rtf->Write("\\tab ",5,NULL);
				break;

			case 12:
				rtf->Write("\\page\n",6,NULL);
				break;
			default:
				if(c[i] < ' ' || c[i] > 127) {
					char hex[16];
					snprintf(hex, 16, "\\'%X", c[i]);
					rtf->Write(hex,strlen(hex),NULL);

				} else {
					rtf->Write(&c[i],1,NULL);
				}
			}
		}
	}

	rtf->Write(footer,strlen(footer),NULL);

	return hrSuccess;
}


LONG Util::HrFindPropInArray(LPSPropTagArray lpPropTags, ULONG ulPropTag)
{
	unsigned int i = 0;

	for(i=0 ; i < lpPropTags->cValues; i++) {
		if(lpPropTags->aulPropTag[i] == ulPropTag)
			break;
		if(PROP_TYPE(ulPropTag) == PT_UNSPECIFIED && PROP_ID(lpPropTags->aulPropTag[i]) == PROP_ID(ulPropTag))
			break;
	}

	if(i != lpPropTags->cValues)
		return i;

	return -1;
}

enum eListMode { lmDefinition, lmOrdered, lmUnordered };
struct ListInfo {
	eListMode mode;
	unsigned count;
};
typedef std::stack<ListInfo> ListInfoStack;

HRESULT Util::HrHTMLToText(char *lpBuffer)
{
	HRESULT hr = hrSuccess;
	unsigned int i = 0;
	bool fCommentMode = false;
	bool fTagMode = false;
	bool fScriptMode = false;
	bool fHeadMode = false;
	bool fSpace = true;
	bool fNewline = false;
	bool fFirstCol = true;
	bool fAllowNewline = false;
	
	ListInfo listInfo;
	ListInfoStack listInfoStack;

	// As lpDst always trails behind lpSrc, we can do an in-place update
	char *lpSrc = lpBuffer;
	char *lpDst = lpBuffer;
	char out = 0;
	std::string entity;

	// We 'convert' from HTML to text by doing a rather simple stripping
	// of tags, and conversion of some strings
	while(*lpSrc != 0) {
		out = 0;

		if(*lpSrc == '\n' || *lpSrc == '\r') { // replace newlines with spaces
			lpSrc++;
			out = ' ';
		} else if (*lpSrc == '\t') { // remove tabs
			lpSrc++;
		} else if(strnicmp(lpSrc, "<--", 3) == 0) {
			fCommentMode = true;
			lpSrc += 3;
		} else if(strnicmp(lpSrc, "<!--", 4) == 0) {
			fCommentMode = true;
			lpSrc += 4;
		} else if(strnicmp(lpSrc, "-->", 3) == 0) {
			fCommentMode = false;
			lpSrc += 3;
		} else if(strnicmp(lpSrc, "</DIV", 5) == 0) {
			if (fAllowNewline) {
				*lpDst++ = '\r';
				*lpDst++ = '\n';
				fNewline = true; 
			}
			lpSrc += 5;
			fSpace = true;
		} else if(strnicmp(lpSrc, "<BR", 3) == 0) {
			if (fAllowNewline) {
				*lpDst++ = '\r';
				*lpDst++ = '\n';
				fNewline = true; 
			}
			fSpace = true;
			lpSrc += 3;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "<P", 2) == 0) {
			if (fAllowNewline) {
				*lpDst++ = '\r';
				*lpDst++ = '\n';
				fNewline = true; 
			}
			fSpace = true;
			lpSrc += 2;
			fTagMode = true;
		}  else if(strnicmp(lpSrc, "</P", 3) == 0) {
			if (fAllowNewline) {
				*lpDst++ = '\r';
				*lpDst++ = '\n';
				fNewline = true; 
			}
			fSpace = true;
			lpSrc += 3;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "<SCRIPT", 7) == 0) {
			lpSrc += 7;
			fScriptMode = true;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "</SCRIPT", 8) == 0) {
			lpSrc += 8;
			fScriptMode = false;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "<STYLE", 6) == 0) {
			lpSrc += 6;
			fScriptMode = true;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "</STYLE", 7) == 0) {
			lpSrc += 7;
			fScriptMode = false;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "<SPAN", 5) == 0) {
			lpSrc += 5;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "</SPAN", 6) == 0) {
			lpSrc += 6;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "<HEAD", 5) == 0) {
			lpSrc += 5;
			fHeadMode = true;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "</HEAD", 6) == 0) {
			lpSrc += 6;
			fHeadMode = false;
			fTagMode = true;
		} else if(strnicmp(lpSrc, "<OL", 3) == 0) {
			listInfo.mode = lmOrdered;
			listInfo.count = 1;
			listInfoStack.push(listInfo);
			fTagMode = true;
			lpSrc += 3;
		} else if(strnicmp(lpSrc, "<UL", 3) == 0) {
			listInfo.mode = lmUnordered;
			listInfo.count = 1;
			listInfoStack.push(listInfo);
			fTagMode = true;
			lpSrc += 3;
		} else if(strnicmp(lpSrc, "<DL", 3) == 0) {
			listInfo.mode = lmDefinition;
			listInfo.count = 1;
			listInfoStack.push(listInfo);
			fTagMode = true;
			lpSrc += 3;
		} else if(strnicmp(lpSrc, "</OL", 4) == 0) {
			if (!listInfoStack.empty())
				listInfoStack.pop();
			fTagMode = false;
			lpSrc += 4;
		} else if(strnicmp(lpSrc, "</UL", 4) == 0) {
			if (!listInfoStack.empty())
				listInfoStack.pop();
			fTagMode = false;
			lpSrc += 4;
		} else if(strnicmp(lpSrc, "</DL", 4) == 0) {
			if (!listInfoStack.empty())
				listInfoStack.pop();
			fTagMode = false;
			lpSrc += 4;
		} else if(strnicmp(lpSrc, "<LI", 3) == 0) {
			if (fAllowNewline && !fNewline) {
				*lpDst++ = '\r';
				*lpDst++ = '\n';
				fNewline = true;
			}
			if (!listInfoStack.empty()) {
				for (int i = 0; i < listInfoStack.size() - 1; ++i)
					*lpDst++ = '\t';
					
				if (listInfoStack.top().mode == lmOrdered) {
					std::string strCount = stringify(listInfoStack.top().count++);
					for (int i = 0; i < strCount.size(); ++i)
						*lpDst++ = strCount.at(i);
					*lpDst++ = '.';
				} else 
					*lpDst++ = '*';
				
				*lpDst++ = '\t';
			}
			fTagMode = true;
			lpSrc += 3;
		} else if(strnicmp(lpSrc, "<DT", 3) == 0) {
			if (fAllowNewline && !fNewline) {
				*lpDst++ = '\r';
				*lpDst++ = '\n';
				fNewline = true;
			}
			if (!listInfoStack.empty()) {
				for (int i = 0; i < listInfoStack.size() - 1; ++i)
					*lpDst++ = '\t';
			}
			fTagMode = true;
			lpSrc += 3;
		} else if(strnicmp(lpSrc, "<DD", 3) == 0) {
			if (fAllowNewline && !fNewline) {
				*lpDst++ = '\r';
				*lpDst++ = '\n';
				fNewline = true;
			}
			if (!listInfoStack.empty()) {
				for (int i = 0; i < listInfoStack.size(); ++i)
					*lpDst++ = '\t';
			}
			fTagMode = true;
			lpSrc += 3;
		} else if (strnicmp(lpSrc, "<TR", 3) == 0) {
			if (fAllowNewline && !fNewline) {
				*lpDst++ = '\r';
				*lpDst++ = '\n';
				fNewline = true;
			}
			fFirstCol = true;
			fTagMode = true;
			lpSrc += 3;
		} else if (strnicmp(lpSrc, "<TD", 3) == 0) {
			if (!fFirstCol)
				*lpDst++ = '\t';
			fFirstCol = false;
			fTagMode = true;
			lpSrc += 3;
		} else if(*lpSrc == '<') {
			fTagMode = true;
			lpSrc++;
		} else if(*lpSrc == '>') {
			fTagMode = false;
			lpSrc++;
		} else if(*lpSrc == '&') {
			lpSrc++;
			if(*lpSrc == '#') {
				entity = "";
				lpSrc++;
				while(isdigit(*lpSrc)) {
					entity += *lpSrc;
					lpSrc++;
				}
				if(*lpSrc == ';')
					lpSrc++;
				
                ECIConv conv("windows-1252", "UCS-4LE");
                unsigned int ulUnicode = atoi(entity.c_str());
                std::string strUnicode;
                std::string strWindows1252;
                strUnicode.append(1,(ulUnicode & 0xff));
                strUnicode.append(1,(ulUnicode >> 8) & 0xff);
                strUnicode.append(1,(ulUnicode >> 16) & 0xff);
                strUnicode.append(1,(ulUnicode >> 24) & 0xff);
                
                // strUnicode is now a byte-stream with the unicode char in it as UCS-4 (32bit) UNICODE
                strWindows1252 = conv.convert(strUnicode);
                
                // strWindows1252 is now the windows-1252 representation of that Unicode number, max 1 char
                if(!strWindows1252.empty())
                    out = strWindows1252.c_str()[0];
                else
                    out = ' ';
                    
			} else {
				entity = "";
				while(isalpha(*lpSrc)) {
					entity += *lpSrc;
					lpSrc++;
				}
				if(*lpSrc == ';')
					lpSrc++;
				
				for(i=0;i<sizeof(HTMLEntity)/sizeof(HTMLEntity[0]);i++) {
					if(strcmp(entity.c_str(), HTMLEntity[i].s) == 0) {
						out = HTMLEntity[i].c;
						break;
					} else {
						out = ' '; // unknown entity, assume space
					}
				}
			}
					
		} else {
			out = *lpSrc;
			lpSrc++;
		}
		
		// Blank out non-text 
		if(fTagMode || fCommentMode || fScriptMode || fHeadMode) 
			out = 0;
		
		if(out == ' ' || out == 0) {
			if(out == ' ' && !fSpace) {
				fSpace = true;
				*lpDst++ = ' ';
			}
		} else {
			fSpace = false;
			fAllowNewline = true;
			fNewline = false;
			*lpDst++ = out;
		}
	}

	*lpDst = 0;

	return hr;
}

char* Util::HrMAPIErrorToText(HRESULT hr)
{
	char* lpszError = NULL;

	switch(hr)
	{
	case MAPI_E_END_OF_SESSION:
		lpszError = _("End of Session");
		break;
	case MAPI_E_NETWORK_ERROR:
		lpszError = _("Connection lost");
		break;
	case MAPI_E_NO_ACCESS:
		lpszError = _("Access denied");
		break;
	case MAPI_E_FOLDER_CYCLE:
		lpszError = _("Unable to move or copy folders. Can't copy folder. A top-level can't be copied to one of its subfolders. Or, you may not have appropriate permissions for the folder. To check your permissions for the folder, right-click the folder, and then click Properties on the shortcut menu.");
		break;
	case MAPI_E_STORE_FULL:
		lpszError = _("The message store has reached its maximum size. To reduce the amount of data in this message store, select some items that you no longer need, and permanently (SHIFT + DEL) delete them.");
		break;
	case MAPI_E_USER_CANCEL:
		lpszError = _("The user canceled the operation, typically by clicking the Cancel button in a dialog box.");
		break;
	case MAPI_E_LOGON_FAILED:
		lpszError = _("A logon session could not be established.");
		break;
	case MAPI_E_COLLISION:
		lpszError = _("The name of the folder being moved or copied is the same as that of a subfolder in the destination folder. The message store provider requires that folder names be unique. The operation stops without completing.");
		break;
	case MAPI_W_PARTIAL_COMPLETION:
		lpszError = _("The operation succeeded, but not all entries were successfully processed, copied, deleted or moved");// emptied
		break;
	case MAPI_E_UNCONFIGURED:
		lpszError = _("The provider does not have enough information to complete the logon. Or, the service provider has not been configured.");
		break;
	case MAPI_E_FAILONEPROVIDER:
		lpszError = _("One of the providers cannot log on, but this error should not disable the other services.");
		break;
	default:
		lpszError = _("Access denied");
		break;
	}

	return lpszError;
}

bool Util::ValidatePropTagArray(LPSPropTagArray lpPropTagArray)
{
	bool bResult = false;
	unsigned int i;

	if(lpPropTagArray == NULL){
		bResult = true;
		goto exit;
	}

	for(i=0; i < lpPropTagArray->cValues; i++)
	{
		switch (PROP_TYPE(lpPropTagArray->aulPropTag[i]))
		{
			case PT_UNSPECIFIED:
			case PT_NULL:
			case PT_I2:
			case PT_I4:
			case PT_R4:
			case PT_R8:
			case PT_BOOLEAN:
			case PT_CURRENCY:
			case PT_APPTIME:
			case PT_SYSTIME:
			case PT_I8:
			case PT_STRING8:
			case PT_BINARY:
			case PT_UNICODE:
			case PT_CLSID:
			case PT_OBJECT:
			case PT_MV_I2:
			case PT_MV_LONG:
			case PT_MV_R4:
			case PT_MV_DOUBLE:
			case PT_MV_CURRENCY:
			case PT_MV_APPTIME:
			case PT_MV_SYSTIME:
			case PT_MV_BINARY:
			case PT_MV_STRING8:
			case PT_MV_UNICODE:
			case PT_MV_CLSID:
			case PT_MV_I8:
			case PT_ERROR:
				bResult = true;
				break;
			default:
				bResult = false;
				goto exit;
				break;
		}
	}

exit:
	return bResult;
}

HRESULT Util::HrHtmlToText(IStream *html, IStream *text)
{
	char *lpBuffer = NULL, bufTmp[4096];
	STATSTG statstg;
	ULONG	ulRead = 0;
	ULONG	ulWrite = 0;
	HRESULT hr = hrSuccess;

	hr = html->Stat(&statstg, 0);
	if(hr != hrSuccess)
		goto exit;

	lpBuffer = (char*)malloc(statstg.cbSize.QuadPart+1);

	while(1) {
		hr = html->Read(&bufTmp, 4096, &ulRead);
		if(hr != hrSuccess)
			goto exit;

		if(ulRead == 0)
			break;

		memcpy(lpBuffer+ulWrite, bufTmp, ulRead);
		ulWrite+=ulRead;
	}

	lpBuffer[statstg.cbSize.QuadPart] = 0;

	hr = HrHTMLToText(lpBuffer);
	if(hr != hrSuccess)
		goto exit;

	hr = text->Write(lpBuffer, strlen(lpBuffer)+1, NULL);

exit:
	if(lpBuffer)
		free((void*)lpBuffer);

	return hr;
}

/**
 * This converts from HTML to RTF by doing to following:
 *
 * Always escape { and } to \{ and \}
 * Always escape \r\n to \par
 * All HTML tags are converted from, say <BODY onclick=bla> to \r\n{\htmltagX <BODY onclick=bla>}
 * Each tag with text content gets an extra {\htmltag64} to suppress generated <P>'s in the final HTML output
 * Some tags output \htmlrtf \par \htmlrtf0 so that the plaintext version of the RTF has newlines in the right places
 * Some effort is done so that data between <STYLE> tags is output as a single entity
 * <!-- and --> tags are supported and output as a single htmltagX entity
 *
 * This gives an RTF stream that converts back to the original HTML when viewed in OL, but also preserves plaintext content
 * when all HTML content is removed.
 */
HRESULT Util::HrHtmlToRtf(std::string &strHTML, std::string &strRTF, unsigned int ulCodepage)
{
    HRESULT hr = hrSuccess;
    
	int tag = 0, type = 0;
	stack<unsigned int> stackTag;
	size_t pos = 0;
	bool inTag = false;
	int ulCommentMode = 0;		// 0=no comment, 1=just starting top-level comment, 2=inside comment level 1, 3=inside comment level 2, etc
	int ulStyleMode = 0;
	int ulParMode = 0;
	bool bFirstText = true;
	bool bPlainCRLF = false;

	// Apparently, windows doesn't really like UTF-8 in RTF. So if we find that, just convert it back to windows-1252. 
	if(ulCodepage == 65001)
	{
		ECIConv ic("UCS-2LE", "UTF-8");
		ucs2string wstrHTML;
		std::string strHTMLUCS2 = ic.convert(strHTML);
		
		wstrHTML.assign((WCHAR *)strHTMLUCS2.c_str(), strHTMLUCS2.size()/2);

		hr = HrEscapeHTML(wstrHTML, strHTML);
		if(hr != hrSuccess)
			goto exit;

		// strHTML is now a 7-bit clean HTML stream. We now set the codepage to 1252 but it doesn't really matter since there
		// are no chars > 127 in the stream now
		ulCodepage = 1252;

		// FIXME we could improve this by only doing the conversion to wstrHTML, and then have the code below only escape non-1252
		// chars by outputting an RTF \uXXXX sequence, or an HTML &#XXXXX; sequence. This is what outlook seems to do.
	}
	
	strRTF = "{\\rtf1\\ansi\\ansicpg" + stringify(ulCodepage) + "\\fromhtml1 \\deff0{\\fonttbl\r\n"
	        "{\\f0\\fswiss\\fcharset0 Arial;}\r\n"
            "{\\f1\\fmodern Courier New;}\r\n"
            "{\\f2\\fnil\\fcharset2 Symbol;}\r\n"
            "{\\f3\\fmodern\\fcharset0 Courier New;}\r\n"
            "{\\f4\\fswiss\\fcharset0 Arial;}\r\n"
            "{\\f5\\fswiss Tahoma;}\r\n"
            "{\\f6\\fswiss\\fcharset0 Times New Roman;}}\r\n"
            "{\\colortbl\\red0\\green0\\blue0;\\red0\\green0\\blue255;\\red0\\green0\\blue255;}\r\n"
            "\\uc1\\pard\\plain\\deftab360 \\f0\\fs24 ";

	stackTag.push(RTF_OUTHTML);
	// Convert strHTML to strRTF
    	
	// We 'convert' from HTML to text by doing a rather simple stripping
	// of tags, and conversion of some strings
	while(pos < strHTML.size()) {
		type = RTF_TAG_TYPE_UNK;

		// Remember if this tag should output a CRLF
        if(strHTML[pos] == '<')
            bPlainCRLF = false;
            		
        // Process important tags first
		if(StrCaseCompare(strHTML, "<HTML", pos)) {
		    type = RTF_TAG_TYPE_HTML;
		} else if(StrCaseCompare(strHTML, "</HTML", pos)) {
		    type = RTF_TAG_TYPE_HTML;
		} else if(StrCaseCompare(strHTML, "<HEAD", pos)) {
		    type = RTF_TAG_TYPE_HEAD;
		} else if(StrCaseCompare(strHTML, "</HEAD", pos)) {
		    type = RTF_TAG_TYPE_HEAD;
		} else if(StrCaseCompare(strHTML, "<BODY", pos)) {
		    type = RTF_TAG_TYPE_BODY;
		} else if(StrCaseCompare(strHTML, "</BODY", pos)) {
		    type = RTF_TAG_TYPE_BODY;
		} else if(StrCaseCompare(strHTML, "<P", pos)) {
		    type = RTF_TAG_TYPE_P;
		    bPlainCRLF = true;
		} else if(StrCaseCompare(strHTML, "</P", pos)) {
		    type = RTF_TAG_TYPE_P;
		    bPlainCRLF = true;
		} else if(StrCaseCompare(strHTML, "<DIV", pos)) {
		    type = RTF_TAG_TYPE_ENDP;
		} else if(StrCaseCompare(strHTML, "</DIV", pos)) {
		    type = RTF_TAG_TYPE_ENDP;
		    bPlainCRLF = true;
		} else if(StrCaseCompare(strHTML, "<SPAN", pos)) {
		    type = RTF_TAG_TYPE_STARTP;
		} else if(StrCaseCompare(strHTML, "</SPAN", pos)) {
		    type = RTF_TAG_TYPE_STARTP;
		} else if(StrCaseCompare(strHTML, "<A", pos)) {
		    type = RTF_TAG_TYPE_STARTP;
		} else if(StrCaseCompare(strHTML, "</A", pos)) {
		    type = RTF_TAG_TYPE_STARTP;
		} else if(StrCaseCompare(strHTML, "<BR", pos)) {
		    type = RTF_TAG_TYPE_BR;
		    bPlainCRLF = true;
		} else if(StrCaseCompare(strHTML, "<PRE", pos)) {
		    type = RTF_TAG_TYPE_PRE;
		} else if(StrCaseCompare(strHTML, "</PRE", pos)) {
		    type = RTF_TAG_TYPE_PRE;
		} else if(StrCaseCompare(strHTML, "<FONT", pos)) {
		    type = RTF_TAG_TYPE_FONT;
		} else if(StrCaseCompare(strHTML, "</FONT", pos)) {
		    type = RTF_TAG_TYPE_FONT;
		} else if(StrCaseCompare(strHTML, "<META", pos)) {
		    type = RTF_TAG_TYPE_HEADER;
		} else if(StrCaseCompare(strHTML, "<LINK", pos)) {
		    type = RTF_TAG_TYPE_HEADER;
		} else if(StrCaseCompare(strHTML, "<H", pos) && isdigit(strHTML[pos+2]) ) {
		    type = RTF_TAG_TYPE_HEADER;
		} else if(StrCaseCompare(strHTML, "</H", pos) && isdigit(strHTML[pos+3]) ) {
		    type = RTF_TAG_TYPE_HEADER;
		} else if(StrCaseCompare(strHTML, "<TITLE", pos)) {
		    type = RTF_TAG_TYPE_TITLE;
		} else if(StrCaseCompare(strHTML, "</TITLE", pos)) {
		    type = RTF_TAG_TYPE_TITLE;
		} else if(StrCaseCompare(strHTML, "<PLAIN", pos)) {
		    type = RTF_TAG_TYPE_FONT;
		} else if(StrCaseCompare(strHTML, "</PLAIN", pos)) {
		    type = RTF_TAG_TYPE_FONT;
        }
        
        if(StrCaseCompare(strHTML, "</", pos)) {
            type |= RTF_FLAG_CLOSE;
        }

        // Set correct state flag if closing tag (RTF_IN*)
        if(type & RTF_FLAG_CLOSE) {
            switch(type & 0xF0) {
                case RTF_TAG_TYPE_HEAD:
                    if(!stackTag.empty() && stackTag.top() == RTF_INHEAD)
                        stackTag.pop();
                    break;
                case RTF_TAG_TYPE_BODY:
                    if(!stackTag.empty() && stackTag.top() == RTF_INBODY)
                        stackTag.pop();
                    break;
                case RTF_TAG_TYPE_HTML:
                    if(!stackTag.empty() && stackTag.top() == RTF_INHTML)
                        stackTag.pop();
                    break;
                default:
                    break;
            }
        }
        
        // Process special tag input
        if(strHTML[pos] == '<' && !inTag) {
            if(StrCaseCompare(strHTML, "<!--", pos))
                ulCommentMode++;
            
            if(ulCommentMode == 0) {
                if(StrCaseCompare(strHTML, "<STYLE", pos))
                    ulStyleMode = 1;
                else if(StrCaseCompare(strHTML, "</STYLE", pos)) {
                    if(ulStyleMode == 3) {
                        // Close the style content tag
                        strRTF += "}";
                    }
                    ulStyleMode = 0;
                } else if(StrCaseCompare(strHTML, "<DIV", pos) || StrCaseCompare(strHTML, "<P", pos)) {
                    ulParMode = 1;
                } else if(StrCaseCompare(strHTML, "</DIV", pos) || StrCaseCompare(strHTML, "</P", pos)) {
                    ulParMode = 0;
                }
            }
            
            if(ulCommentMode < 2 && ulStyleMode < 2) {
                strRTF += "\r\n{\\*\\htmltag" + stringify((ulParMode == 2 ? RTF_FLAG_INPAR : 0) | tag | type | stackTag.top()) + " ";
                inTag = true;
                bFirstText = true;
                
                if(ulCommentMode) {
                    // Inside comment now
                    ulCommentMode++;
                }
            }
        }
        
        // Do actual output
        if(strHTML[pos] == '\r') {
            // Ingore \r
        } else if(strHTML[pos] == '\n') {
            if(inTag || ulCommentMode || ulStyleMode)
                strRTF += "\\par ";
            else
                strRTF += "\r\n{\\*\\htmltag" + stringify((ulParMode == 2 ? RTF_FLAG_INPAR : 0) | stackTag.top()) + " \\par }";
        } else if(strHTML[pos] == '\t') {
            if(inTag || ulCommentMode || ulStyleMode)
                strRTF += "\\tab ";
            else
                strRTF += "\r\n{\\*\\htmltag" + stringify((ulParMode == 2 ? RTF_FLAG_INPAR : 0) | stackTag.top()) + " \\tab }";
        } else if(strHTML[pos] == '{') {
            strRTF += "\\{";
        } else if(strHTML[pos] == '}') {
            strRTF += "\\}";
        } else if(strHTML[pos] == '\\') {
            strRTF += "\\\\";
		} else if((unsigned char)strHTML[pos] > 127) {
			char hex[3];
			// Only output htmltag if we're already in a htmltag (which is the case for inTag=1 and ulStyleMode and ulCommentMode)
			if(!ulCommentMode && !inTag && !ulStyleMode)
    			strRTF += "{\\*\\htmltag ";
			snprintf(hex, 3, "%02x", (unsigned int)(unsigned char)strHTML[pos]);
			strRTF += "\\'";
			strRTF += hex;
			if(!ulCommentMode && !inTag && !ulStyleMode)
    			strRTF += "}";
        } else if(StrCaseCompare(strHTML, "&nbsp;", pos)) {
            if(inTag || ulCommentMode || ulStyleMode)
                strRTF += "&nbsp;";
            else
                strRTF += "\r\n{\\*\\htmltag64}{\\*\\htmltag" + stringify((ulParMode == 2 ? RTF_FLAG_INPAR : 0) | RTF_TAG_TYPE_STARTP | stackTag.top()) + " &nbsp;}";
                
            pos+=5;
        } else if(!inTag && !ulCommentMode && !ulStyleMode && strHTML[pos] == '&' && validateHtmlEntity(strHTML.substr(pos, 10).c_str()) ) {
            size_t semicolon = strHTML.find(";", pos);
            if(semicolon != std::string::npos) {
                std::string strEntity = strHTML.substr(pos+1, semicolon-pos-1);
                char c = CharFromHtmlEntity(strEntity);
                std::string strChar;
                
                if(c < 128 && c >= 32)
                    strChar = c;
                else {
                    char hex[16];
                    sprintf(hex, "%02X", (unsigned char)c);
                    strChar = (std::string)"\\'" + hex;
                }
                
                strRTF += (std::string)"{\\htmlrtf " + strChar + "}{\\*\\htmltag" + stringify((ulParMode == 2 ? RTF_FLAG_INPAR : 0) | RTF_TAG_TYPE_STARTP | stackTag.top()) + "&" + strEntity + ";}";
                pos += strEntity.size() + 2;
                continue;
            }
        } else {
            if(!inTag && bFirstText) {
                bFirstText = false;
            }
            strRTF += strHTML[pos];
        }

        // Do post-processing output
        if(strHTML[pos] == '>' && (inTag || ulCommentMode)) {
            if(!ulCommentMode && ulStyleMode < 2)
                strRTF += "}";
                
            if(pos > 2 && StrCaseCompare(strHTML, "-->", pos-2) && ulCommentMode) {
                ulCommentMode--;
                
                if(ulCommentMode == 1) {
                    ulCommentMode = 0;
                    strRTF += "}";
                }
            }

            if(pos > 6 && StrCaseCompare(strHTML, "/STYLE>", pos-6) && ulStyleMode) {
                ulStyleMode = 0;
                strRTF += "}";
            }

            if(ulStyleMode == 1)
                ulStyleMode++;
                
            if(ulParMode == 1)
                ulParMode++;
                
            if(ulStyleMode == 2) {
                // Output the style content as a tag
                ulStyleMode = 3;
                strRTF += "\r\n{\\*\\htmltag" + stringify(RTF_TAG_TYPE_UNK | stackTag.top()) + " ";
            } 
            
            if(!ulStyleMode && !ulCommentMode) {
                // Normal text must have \*\htmltag64 to suppress <p> in the final html output
                strRTF += "{\\*\\htmltag64}";
            }            
            
            inTag = false;

            if(bPlainCRLF && !ulCommentMode && !ulStyleMode) {
                // Add a plaintext newline if needed, but only for non-style and non-comment parts
                strRTF += "\\htmlrtf \\par \\htmlrtf0 ";
            }
        }
        
        
        // Next char
        pos++;

        // Set correct state flag (RTF_IN*)
        if(!(type & RTF_FLAG_CLOSE)) {
            switch(type & 0xF0) {
                case RTF_TAG_TYPE_HTML:
                    stackTag.push(RTF_INHTML);
                    break;
                case RTF_TAG_TYPE_BODY:
                    stackTag.push(RTF_INBODY);
                    break;
                case RTF_TAG_TYPE_HEAD:
                    stackTag.push(RTF_INHEAD);
                    break;
                default:
                    break;
            }
        }
	}
	
	strRTF +="}\r\n";

exit:
	return hr;
}

HRESULT	Util::HrHtmlToRtf(IStream *html, IStream *rtf, unsigned int ulCodepage)
{
	HRESULT hr = hrSuccess;
	std::string strHTML;
	std::string strRTF;
	char bufTmp[4096];
	ULONG	ulRead = 0;
	
	while(1) {
		hr = html->Read(&bufTmp, 4096, &ulRead);
		if(hr != hrSuccess)
			goto exit;

		if(ulRead == 0)
			break;

		strHTML.append(bufTmp, ulRead);
	}
	
	HrHtmlToRtf(strHTML, strRTF, ulCodepage);


	hr = rtf->Write(strRTF.c_str(), strRTF.size(), NULL);
exit:
	return hr;
}

HRESULT Util::HrConvertUnicodeToString8(ULONG cValues, LPSPropValue lpPropArrayIn, LPSPropValue* lppPropArrayOut)
{
	HRESULT			hr = hrSuccess;
	LPSPropValue	lpPropArrayOut = NULL;
	ULONG			ulSize = 0;

	if(lpPropArrayIn == NULL || lppPropArrayOut == NULL) {
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}

	hr = MAPIAllocateBuffer(sizeof(SPropValue) * cValues, (void**)&lpPropArrayOut);
	if(hr != hrSuccess)
		goto exit;

	for(ULONG i =0; i < cValues; i++)
	{
		if( PROP_TYPE(lpPropArrayIn[i].ulPropTag) == PT_UNICODE) {
			
			lpPropArrayOut[i].ulPropTag = PROP_TAG(PT_STRING8, PROP_ID(lpPropArrayIn[i].ulPropTag));
			ulSize = unicodelen(lpPropArrayIn[i].Value.lpszW) +1;

			hr = MAPIAllocateMore(ulSize, lpPropArrayOut, (void**)&lpPropArrayOut[i].Value.lpszA);
			if(hr != hrSuccess)
				goto exit;

			if(unicodetombs(lpPropArrayOut[i].Value.lpszA, lpPropArrayIn[i].Value.lpszW, ulSize) == -1) {
				hr = MAPI_E_BAD_CHARWIDTH;
				goto exit;
			}

		} else if( PROP_TYPE(lpPropArrayIn[i].ulPropTag) == PT_MV_UNICODE) {
			lpPropArrayOut[i].ulPropTag = PROP_TAG(PT_MV_STRING8, PROP_ID(lpPropArrayIn[i].ulPropTag));

			hr = MAPIAllocateMore(lpPropArrayIn[i].Value.MVszW.cValues * sizeof(char *), lpPropArrayOut, (void**)&lpPropArrayOut[i].Value.MVszA.lppszA);
			if(hr != hrSuccess)
				goto exit;

			lpPropArrayOut[i].Value.MVszA.cValues = lpPropArrayIn[i].Value.MVszW.cValues;

			for(ULONG j=0; j < lpPropArrayIn[i].Value.MVszW.cValues; j++)
			{
				lpPropArrayOut[i].Value.MVszA.lppszA[j] = NULL;

				if(lpPropArrayIn[i].Value.MVszW.lppszW[j] == NULL)
					continue;

				ulSize = unicodelen(lpPropArrayIn[i].Value.MVszW.lppszW[j]) +1;

				hr = MAPIAllocateMore(ulSize, lpPropArrayOut, (void**)&lpPropArrayOut[i].Value.MVszA.lppszA[j]);
				if(hr != hrSuccess)
					goto exit;

				if(unicodetombs(lpPropArrayOut[i].Value.MVszA.lppszA[j], lpPropArrayIn[i].Value.MVszW.lppszW[j], ulSize) == -1) {
					hr = MAPI_E_BAD_CHARWIDTH;
					goto exit;
				}
			}

		} else {
			lpPropArrayOut[i] = lpPropArrayIn[i];
		}
	}

	*lppPropArrayOut = lpPropArrayOut;
exit:
	if(hr != hrSuccess && lpPropArrayOut)
		MAPIFreeBuffer(lpPropArrayOut);

	return hr;
}

std::string Util::bin2hex(ULONG inLength, LPBYTE input) {
	const char digits[] = "0123456789ABCDEF";
	std::string buffer;
	ULONG i;

	if (input) {
		for (i = 0; i < inLength; i++) {
			buffer += digits[input[i]>>4];
			buffer += digits[input[i]&0x0F];
		}
	}

	return buffer;
}

HRESULT Util::bin2hex(ULONG inLength, LPBYTE input, char **output, void *parent) {
	const char digits[] = "0123456789ABCDEF";
	char *buffer = NULL;
	HRESULT hr = hrSuccess;
	ULONG i, j;

	if (parent)
		hr = MAPIAllocateMore(inLength*2+1, parent, (void**)&buffer);
	else
		hr = MAPIAllocateBuffer(inLength*2+1, (void**)&buffer);
	if (hr != hrSuccess)
		goto exit;

	for (i = 0, j = 0; i < inLength; i++) {
		buffer[j++] = digits[input[i]>>4];
		buffer[j++] = digits[input[i]&0x0F];
	}

	buffer[j] = '\0';
	*output = buffer;
exit:
	
	return hr;
}

unsigned char Util::x2b(char c) {
	// expects sensible input
	if (c >= '0' && c <= '9')
		return c - '0';
	else if (c >= 'a')
		return c - 'a' + 10;
	return c - 'A' + 10;
}

HRESULT Util::hex2bin(char *input, size_t len, ULONG *outLength, LPBYTE *output, void *parent) {
	HRESULT hr = hrSuccess;
	LPBYTE buffer = NULL;
	ULONG i, j;

	if (len % 2 != 0) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if (parent)
		hr = MAPIAllocateMore(len/2+1, parent, (void**)&buffer);
	else
		hr = MAPIAllocateBuffer(len/2+1, (void**)&buffer);
	if (hr != hrSuccess)
		goto exit;

	for (i = 0, j = 0; i < len; j++) {
		buffer[j] = x2b(input[i++]) << 4;
		buffer[j] |= x2b(input[i++]);
	}
	buffer[j] = '\0';

	*outLength = len/2;
	*output = buffer;

exit:
	return hr;
}

HRESULT Util::hex2bin(const std::string &input, std::string *output) {
	HRESULT hr = hrSuccess;
	std::string buffer;
	ULONG i;

	if (input.length() % 2 != 0) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	for (i = 0; i < input.length(); ) {
		unsigned char c;
		c = x2b(input[i++]) << 4;
		c |= x2b(input[i++]);
		buffer += c;
	}

	*output = buffer;

exit:
	return hr;
}

HRESULT Util::FileLFtoCRLF(FILE *fin, FILE** fout)
{
	HRESULT hr = hrSuccess;
	char	bufferin[BLOCKSIZE / 2];
	char	bufferout[BLOCKSIZE];
	size_t	sizebufferout, readsize;
	FILE*	fTmp = NULL;

	if(fin == NULL || fout == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	fTmp = tmpfile();
	if(fTmp == NULL) {
		perror("Unable to create tmp file");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	while (!feof(fin)) {
		readsize = fread(bufferin, sizeof(char), BLOCKSIZE / 2, fin);
		if (ferror(fin)) {
			perror("Read error");//FIXME: What an error?, what now?
			hr = MAPI_E_CORRUPT_DATA;
			break;
		}

		Util::BufferLFtoCRLF(readsize, bufferin, bufferout, &sizebufferout);

		if (fwrite(bufferout, sizeof(char), sizebufferout, fTmp) != sizebufferout) {
			perror("Write error");//FIXME: What an error?, what now?
			hr = MAPI_E_CORRUPT_DATA;
			break;
		}
	}

	*fout = fTmp;

exit:
	return hr;
}

void Util::BufferLFtoCRLF(size_t size, const char *input, char *output, size_t *outsize) {
	size_t j = 0;
	for (size_t i = 0; i < size; i++) {
		if (input[i] == '\r') {
			if ((i+1) < size && input[i+1] == '\n') {
				output[j++] = '\r';
				output[j++] = '\n';
				i++;
			} else {
				output[j++] = '\r';
				output[j++] = '\n';
			}
		} else if (input[i] == '\n') {
			output[j++] = '\r';
			output[j++] = '\n';
		} else {
			output[j++] = input[i];
		}
	}
	output[j] = '\0';
	*outsize = j;
}

bool Util::StrCaseCompare(const std::string &strInput, const char *lpFind, size_t pos) {
	size_t findlen = strlen(lpFind);

	if(pos + findlen > strInput.size())
		return false;

	return (0 == strnicmp(strInput.c_str() + pos, lpFind, findlen));
}

static inline int mmapsize(unsigned int size)
{
	/* align to page boundary (4k) */
	return (((size + 1) >> 12) + 1) << 12;
}

HRESULT Util::HrMapFileToString(FILE *f, std::string *lpstrBuffer, int *lpSize)
{
	HRESULT hr = hrSuccess;
	char *lpBuffer = NULL;
	int ulBufferSize = 0;
	bool immap = false;

	hr = Util::HrMapFileToBuffer(f, &lpBuffer, &ulBufferSize, &immap); // what if message was half read?
	if (hr != hrSuccess || !lpBuffer)
		goto exit;

	if (lpstrBuffer)
		*lpstrBuffer = string(lpBuffer, ulBufferSize);
	if (lpSize)
		*lpSize = ulBufferSize;

exit:
	if (lpBuffer)
		Util::HrUnmapFileBuffer(lpBuffer, ulBufferSize, immap);

	return hr;
}

HRESULT Util::HrMapFileToBuffer(FILE *f, char **lppBuffer, int *lpSize, bool *lpImmap)
{
	HRESULT hr = hrSuccess;
	char *lpBuffer = NULL;
	int offset = 0;
	long ulBufferSize = BLOCKSIZE;
	long ulReadsize;
	struct stat stat;
	int fd = fileno(f);

	*lpImmap = false;

	/* Try mmap first */
	if (fstat(fd, &stat) != 0) {
		perror("Stat failed");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	/* auto-zero-terminate because mmap zeroes bytes after the file */
	lpBuffer = (char *)mmap(0, mmapsize(stat.st_size), PROT_READ, MAP_PRIVATE, fd, 0);
	if (lpBuffer != MAP_FAILED) {
		*lpImmap = true;
		*lppBuffer = lpBuffer;
		*lpSize = stat.st_size;
		goto exit;
	}

	/* mmap failed (probably reading from STDIN as a stream), just read the file into memory, and return that */
	lpBuffer = (char*)malloc(BLOCKSIZE); // will be deleted as soon as possible
	while (!feof(f)) {
		ulReadsize = fread(lpBuffer+offset, sizeof(char), BLOCKSIZE, f);
		if (ferror(f)) {
			perror("Read error");
			break;
		}

		offset += ulReadsize;
		if (offset + BLOCKSIZE > ulBufferSize) {    // Next read could cross buffer boundary, realloc
			lpBuffer = (char*)realloc(lpBuffer, offset + BLOCKSIZE);
			ulBufferSize += BLOCKSIZE;
		}
	}

	/* Nothing was read */
    if (offset == 0) {
		*lppBuffer = NULL;
		*lpSize = 0;
	} else {
		/* Add terminate character */
		lpBuffer[offset] = 0;

		*lppBuffer = lpBuffer;
		*lpSize = offset;
	}

exit:
	return hr;
}

void Util::HrUnmapFileBuffer(char *lpBuffer, int ulSize, bool bImmap)
{
	if (bImmap)
		munmap(lpBuffer, mmapsize(ulSize));
	else
		free(lpBuffer);
}

HRESULT Util::FindInterface(LPCIID lpIID, ULONG ulIIDs, LPCIID lpIIDs) {
	HRESULT hr = MAPI_E_NOT_FOUND;
	ULONG i;

	if (!lpIIDs || !lpIID)
		goto exit;

	for (i = 0; i < ulIIDs; i++) {
		if (*lpIID == lpIIDs[i]) {
			hr = hrSuccess;
			break;
		}
	}

exit:
	return hr;
}

HRESULT Util::CopyStream(LPSTREAM lpSrc, LPSTREAM lpDest) {
	HRESULT hr;
	ULARGE_INTEGER liRead = {{0}}, liWritten = {{0}};
	STATSTG stStatus;

	hr = lpSrc->Stat(&stStatus, 0);
	if (FAILED(hr))
		goto exit;

	hr = lpSrc->CopyTo(lpDest, stStatus.cbSize, &liRead, &liWritten);
	if (FAILED(hr))
		goto exit;

	if (liRead.QuadPart != liWritten.QuadPart) {
		hr = MAPI_W_PARTIAL_COMPLETION;
		goto exit;
	}

	hr = lpDest->Commit(0);

exit:
	return hr;
}

HRESULT Util::CopyRecipients(LPMESSAGE lpSrc, LPMESSAGE lpDest) {
	HRESULT hr;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	LPSPropTagArray lpTableColumns = NULL;
	ULONG ulRows = 0;

	hr = lpSrc->GetRecipientTable(0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryColumns(TBL_ALL_COLUMNS, &lpTableColumns);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns(lpTableColumns, 0);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->GetRowCount(0, &ulRows);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryRows(ulRows, 0, &lpRows);
	if (hr != hrSuccess)
		goto exit;

	// LPADRLIST and LPSRowSet are binary compatible \o/
	hr = lpDest->ModifyRecipients(MODRECIP_ADD, (LPADRLIST)lpRows);
	if (hr != hrSuccess)
		goto exit;

exit:
	if (lpTableColumns)
		MAPIFreeBuffer(lpTableColumns);

	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT Util::CopyAttachmentInstance(LPATTACH lpSrcAttach, LPATTACH lpDstAttach)
{
	HRESULT hr = hrSuccess;
	IECSingleInstance *lpSrcInstance = NULL;
	IECSingleInstance *lpDstInstance = NULL;
	ULONG cbInstanceID = 0;
	LPENTRYID lpInstanceID = NULL;

	/* 
	 * We are always going to return hrSuccess, if for some reason we can't copy the single instance,
	 * we always have the real data as fallback.
	 */
	if (lpSrcAttach->QueryInterface(IID_IECSingleInstance, (LPVOID *)&lpSrcInstance) != hrSuccess)
		goto exit;

	if (lpDstAttach->QueryInterface(IID_IECSingleInstance, (LPVOID *)&lpDstInstance) != hrSuccess)
		goto exit;

	/*
	 * Transfer instance Id, if this succeeds we're in luck and we might not
	 * have to send the attachment to the server. Note that while SetSingleInstanceId()
	 * might succeed now, the attachment might be deleted on the server between
	 * SetSingleInstanceId() and SaveChanges(). In that case SaveChanges will fail
	 * and we will have to resend the attachment data.
	 */
	if (lpSrcInstance->GetSingleInstanceId(&cbInstanceID, &lpInstanceID) != hrSuccess)
		goto exit;

	if (lpDstInstance->SetSingleInstanceId(cbInstanceID, lpInstanceID) != hrSuccess)
		goto exit;

exit:
	if (lpSrcInstance)
		lpSrcInstance->Release();
	if (lpDstInstance)
		lpDstInstance->Release();
	if (lpInstanceID)
		MAPIFreeBuffer(lpInstanceID);

	return hr;
}

HRESULT Util::CopyAttachmentProps(LPATTACH lpSrcAttach, LPATTACH lpDstAttach, LPSPropTagArray lpExcludeProps)
{
	HRESULT hr = hrSuccess;

	hr = Util::DoCopyTo(&IID_IAttachment, lpSrcAttach, 0, NULL, lpExcludeProps, 0, NULL, &IID_IAttachment, lpDstAttach, 0, NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpDstAttach->SaveChanges(0);
	if (hr != hrSuccess)
		goto exit;

exit:
	return hr;
}

HRESULT Util::CopyAttachments(LPMESSAGE lpSrc, LPMESSAGE lpDest) {
	HRESULT hr;
	bool bPartial = false;

	// table
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	LPSPropTagArray lpTableColumns = NULL;
	ULONG ulRows = 0;

	// attachments
	LPSPropValue lpAttachNum = NULL;
	LPSPropValue lpHasAttach = NULL;
	ULONG ulAttachNr = 0;
	LPATTACH lpSrcAttach = NULL;
	LPATTACH lpDestAttach = NULL;

	hr = HrGetOneProp(lpSrc, PR_HASATTACH, &lpHasAttach);
	if (hr != hrSuccess) {
		hr = hrSuccess;
		goto exit;
	}
	if (lpHasAttach->Value.b == FALSE)
		goto exit;

	hr = lpSrc->GetAttachmentTable(0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryColumns(TBL_ALL_COLUMNS, &lpTableColumns);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns(lpTableColumns, 0);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->GetRowCount(0, &ulRows);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryRows(ulRows, 0, &lpRows);
	if (hr != hrSuccess)
		goto exit;

	for (ULONG i=0; i < lpRows->cRows; i++) {
		lpAttachNum = PpropFindProp(lpRows->aRow[i].lpProps, lpRows->aRow[i].cValues, PR_ATTACH_NUM);
		if (!lpAttachNum) {
			bPartial = true;
			goto next_attach;
		}

		hr = lpSrc->OpenAttach(lpAttachNum->Value.ul, NULL, 0, &lpSrcAttach);
		if (hr != hrSuccess) {
			bPartial = true;
			goto next_attach;
		}

		hr = lpDest->CreateAttach(NULL, 0, &ulAttachNr, &lpDestAttach);
		if (hr != hrSuccess) {
			bPartial = true;
			goto next_attach;
		}

		/*
		 * Try making a single instance copy (without sending the attachment data to server).
		 */
		hr = CopyAttachmentInstance(lpSrcAttach, lpDestAttach);
		if (hr != hrSuccess) {
			bPartial = true;
			goto next_attach;
		}

		hr = CopyAttachmentProps(lpSrcAttach, lpDestAttach);
		if (hr != hrSuccess) {
			bPartial = true;
			goto next_attach;
		}

next_attach:
		if (lpSrcAttach) {
			lpSrcAttach->Release();
			lpSrcAttach = NULL;
		}

		if (lpDestAttach) {
			lpDestAttach->Release();
			lpDestAttach = NULL;
		}
	}

	if (bPartial)
		hr = MAPI_W_PARTIAL_COMPLETION;

exit:
	if (lpHasAttach)
		MAPIFreeBuffer(lpHasAttach);

	if (lpTableColumns)
		MAPIFreeBuffer(lpTableColumns);

	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT Util::CopyHierarchy(LPMAPIFOLDER lpSrc, LPMAPIFOLDER lpDest, ULONG ulFlags, ULONG ulUIParam, LPMAPIPROGRESS lpProgress) {
	HRESULT hr;
	bool bPartial = false;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRowSet = NULL;
	SizedSPropTagArray(2, sptaName) = { 2, { PR_DISPLAY_NAME, PR_ENTRYID } };
	LPMAPIFOLDER lpSrcFolder = NULL, lpDestFolder = NULL;
	ULONG ulObj;

	hr = lpSrc->GetHierarchyTable(0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns((LPSPropTagArray)&sptaName, 0);
	if (hr != hrSuccess)
		goto exit;

	while (true) {
		hr = lpTable->QueryRows(1, 0, &lpRowSet);
		if (hr != hrSuccess)
			goto exit;

		if (lpRowSet->cRows == 0)
			break;

		hr = lpSrc->OpenEntry(lpRowSet->aRow[0].lpProps[1].Value.bin.cb, (LPENTRYID)lpRowSet->aRow[0].lpProps[1].Value.bin.lpb, &IID_IMAPIFolder, 0, &ulObj, (LPUNKNOWN*)&lpSrcFolder);
		if (hr != hrSuccess) {
			bPartial = true;
			goto next_folder;
		}

		hr = lpDest->CreateFolder(FOLDER_GENERIC, lpRowSet->aRow[0].lpProps[0].Value.lpszA, NULL, &IID_IMAPIFolder, ulFlags&MAPI_NOREPLACE?0:OPEN_IF_EXISTS, &lpDestFolder);
		if (hr != hrSuccess) {
			bPartial = true;
			goto next_folder;
		}

		hr = Util::DoCopyTo(&IID_IMAPIFolder, lpSrcFolder, 0, NULL, NULL, ulUIParam, lpProgress, &IID_IMAPIFolder, lpDestFolder, ulFlags, NULL);
		if (FAILED(hr))
			goto exit;
		else if (hr != hrSuccess) {
			bPartial = true;
			goto next_folder;
		}

		if (ulFlags & MAPI_MOVE)
			lpSrc->DeleteFolder(lpRowSet->aRow[0].lpProps[1].Value.bin.cb, (LPENTRYID)lpRowSet->aRow[0].lpProps[1].Value.bin.lpb, 0, NULL, 0);

next_folder:
		if (lpRowSet) {
			FreeProws(lpRowSet);
			lpRowSet = NULL;
		}

		if (lpSrcFolder) {
			lpSrcFolder->Release();
			lpSrcFolder = NULL;
		}

		if (lpDestFolder) {
			lpDestFolder->Release();
			lpDestFolder = NULL;
		}
	}

	if (bPartial)
		hr = MAPI_W_PARTIAL_COMPLETION;

exit:
	if (lpRowSet)
		FreeProws(lpRowSet);

	if (lpSrcFolder)
		lpSrcFolder->Release();

	if (lpDestFolder)
		lpDestFolder->Release();

	if (lpTable)
		lpTable->Release();

	return hr;
}

#define MAX_ROWS 50
HRESULT Util::CopyContents(ULONG ulWhat, LPMAPIFOLDER lpSrc, LPMAPIFOLDER lpDest, ULONG ulFlags, ULONG ulUIParam, LPMAPIPROGRESS lpProgress) {
	HRESULT hr;
	bool bPartial = false;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRowSet = NULL;
	SizedSPropTagArray(1, sptaEntryID) = { 1, { PR_ENTRYID } };
	ULONG ulObj;
	LPMESSAGE lpSrcMessage = NULL, lpDestMessage = NULL;
	LPENTRYLIST lpDeleteEntries = NULL;


	hr = lpSrc->GetContentsTable(ulWhat, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns((LPSPropTagArray)&sptaEntryID, 0);
	if (hr != hrSuccess)
		goto exit;

	hr = MAPIAllocateBuffer(sizeof(ENTRYLIST), (void**)&lpDeleteEntries);
	if (hr != hrSuccess)
		goto exit;

	hr = MAPIAllocateMore(sizeof(SBinary)*MAX_ROWS, lpDeleteEntries, (void**)&lpDeleteEntries->lpbin);
	if (hr != hrSuccess)
		goto exit;

	while (true) {
		hr = lpTable->QueryRows(MAX_ROWS, 0, &lpRowSet);
		if (hr != hrSuccess)
			goto exit;

		if (lpRowSet->cRows == 0)
			break;

		lpDeleteEntries->cValues = 0;

		for (ULONG i = 0; i < lpRowSet->cRows; i++) {
			hr = lpSrc->OpenEntry(lpRowSet->aRow[i].lpProps[0].Value.bin.cb, (LPENTRYID)lpRowSet->aRow[i].lpProps[0].Value.bin.lpb, &IID_IMessage, 0, &ulObj, (LPUNKNOWN*)&lpSrcMessage);
			if (hr != hrSuccess) {
				bPartial = true;
				goto next_item;
			}

			hr = lpDest->CreateMessage(&IID_IMessage, ulWhat | MAPI_MODIFY, &lpDestMessage);
			if (hr != hrSuccess) {
				bPartial = true;
				goto next_item;
			}

			hr = Util::DoCopyTo(&IID_IMessage, lpSrcMessage, 0, NULL, NULL, ulUIParam, lpProgress, &IID_IMessage, lpDestMessage, ulFlags, NULL);
			if (FAILED(hr))
				goto exit;
			else if (hr != hrSuccess) {
				bPartial = true;
				goto next_item;
			}

			hr = lpDestMessage->SaveChanges(0);
			if (hr != hrSuccess) {
				bPartial = true;
			} else if (ulFlags & MAPI_MOVE) {
				lpDeleteEntries->lpbin[lpDeleteEntries->cValues].cb = lpRowSet->aRow[i].lpProps[0].Value.bin.cb;
				lpDeleteEntries->lpbin[lpDeleteEntries->cValues].lpb = lpRowSet->aRow[i].lpProps[0].Value.bin.lpb;
				lpDeleteEntries->cValues++;
			}
next_item:
			if (lpDestMessage) {
				lpDestMessage->Release();
				lpDestMessage = NULL;
			}

			if (lpSrcMessage) {
				lpSrcMessage->Release();
				lpSrcMessage = NULL;
			}
		}

		if ((ulFlags & MAPI_MOVE) && lpDeleteEntries->cValues > 0) {
			if (lpSrc->DeleteMessages(lpDeleteEntries, 0, NULL, 0) != hrSuccess)
				bPartial = true;
		}

		if (lpRowSet) {
			FreeProws(lpRowSet);
			lpRowSet = NULL;
		}
	}

	if (bPartial)
		hr = MAPI_W_PARTIAL_COMPLETION;

exit:
	if (lpDeleteEntries)
		MAPIFreeBuffer(lpDeleteEntries);

	if (lpDestMessage)
		lpDestMessage->Release();

	if (lpSrcMessage)
		lpSrcMessage->Release();

	if (lpRowSet)
		FreeProws(lpRowSet);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT Util::TryOpenProperty(ULONG ulPropType, ULONG ulSrcPropTag, LPMAPIPROP lpPropSrc, ULONG ulDestPropTag, LPMAPIPROP lpPropDest, LPSTREAM *lppSrcStream, LPSTREAM *lppDestStream) {
	HRESULT hr;
	LPSTREAM lpSrc = NULL, lpDest = NULL;

	hr = lpPropSrc->OpenProperty(PROP_TAG(ulPropType, PROP_ID(ulSrcPropTag)), &IID_IStream, 0, 0, (LPUNKNOWN*)&lpSrc);
	if (hr != hrSuccess)
		goto exit;

	// some mapi functions/providers doesn't implement STGM_TRANSACTED, retry againg without this flag
	hr = lpPropDest->OpenProperty(PROP_TAG(ulPropType, PROP_ID(ulDestPropTag)), &IID_IStream, STGM_WRITE | STGM_TRANSACTED, MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN*)&lpDest);
	if (hr != hrSuccess) {
		hr = lpPropDest->OpenProperty(PROP_TAG(ulPropType, PROP_ID(ulDestPropTag)), &IID_IStream, STGM_WRITE, MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN*)&lpDest);
	}
	if (hr != hrSuccess)
		goto exit;

	*lppSrcStream = lpSrc;
	*lppDestStream = lpDest;

exit:
	if (hr != hrSuccess) {
		if (lpSrc)
			lpSrc->Release();

		if (lpDest)
			lpDest->Release();
	}

	return hr;
}

HRESULT Util::AddProblemToArray(LPSPropProblem lpProblem, LPSPropProblemArray *lppProblems) {
	HRESULT hr = hrSuccess;
	LPSPropProblemArray lpNewProblems = NULL;
	LPSPropProblemArray lpOrigProblems = *lppProblems;

	if (!lpOrigProblems) {
		hr = MAPIAllocateBuffer(CbNewSPropProblemArray(1), (void**)&lpNewProblems);
		if (hr != hrSuccess)
			goto exit;

		lpNewProblems->cProblem = 1;
	} else {
		hr = MAPIAllocateBuffer(CbNewSPropProblemArray(lpOrigProblems->cProblem+1), (void**)&lpNewProblems);
		if (hr != hrSuccess)
			goto exit;

		lpNewProblems->cProblem = lpOrigProblems->cProblem +1;
		memcpy(lpNewProblems->aProblem, lpOrigProblems->aProblem, sizeof(SPropProblem) * lpOrigProblems->cProblem);
		MAPIFreeBuffer(lpOrigProblems);
	}

	memcpy(&lpNewProblems->aProblem[lpNewProblems->cProblem -1], lpProblem, sizeof(SPropProblem));

	*lppProblems = lpNewProblems;

exit:
	return hr;
}

HRESULT Util::DoCopyTo(LPCIID lpSrcInterface, LPVOID lpSrcObj, ULONG ciidExclude, LPCIID rgiidExclude,
					   LPSPropTagArray lpExcludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpDestInterface,
					   LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray * lppProblems)
{
	HRESULT hr = hrSuccess;
	LPUNKNOWN lpUnkSrc = (LPUNKNOWN)lpSrcObj, lpUnkDest = (LPUNKNOWN)lpDestObj;
	bool bPartial = false;
	// Properties that can never be copied (if you do this wrong, copying a message to a PST will fail)
	SizedSPropTagArray(21, sExtraExcludes) = { 18, { PR_STORE_ENTRYID, PR_STORE_RECORD_KEY, PR_STORE_SUPPORT_MASK, PR_MAPPING_SIGNATURE, PR_MDB_PROVIDER, PR_ACCESS_LEVEL, PR_RECORD_KEY, PR_HASATTACH, PR_NORMALIZED_SUBJECT, PR_MESSAGE_SIZE, PR_DISPLAY_TO, PR_DISPLAY_CC, PR_DISPLAY_BCC, PR_ACCESS, PR_SUBJECT_PREFIX, PR_OBJECT_TYPE, PR_ENTRYID, PR_PARENT_ENTRYID, PR_NULL, PR_NULL, PR_NULL }};

	LPMAPIPROP lpPropSrc = NULL, lpPropDest = NULL;
	LPSPropTagArray lpSPropTagArray = NULL;
	LPSPropValue lpProps = NULL;

	if (!lpSrcInterface || !lpSrcObj || !lpDestInterface || !lpDestObj) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// source is "usually" the same as dest .. so we don't check (as ms mapi doesn't either)

	hr = FindInterface(lpSrcInterface, ciidExclude, rgiidExclude);
	if (hr == hrSuccess) {
		hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
		goto exit;
	}

	hr = FindInterface(lpDestInterface, ciidExclude, rgiidExclude);
	if (hr == hrSuccess) {
		hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
		goto exit;
	}

	// first test IID_IStream .. the rest is IID_IMAPIProp compatible
	if (*lpSrcInterface == IID_IStream) {
		hr = FindInterface(&IID_IStream, ciidExclude, rgiidExclude);
		if (hr == hrSuccess) {
			hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
			goto exit;
		}
		if (*lpDestInterface != IID_IStream) {
			hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
			goto exit;
		}
		hr = CopyStream((LPSTREAM)lpSrcObj, (LPSTREAM)lpDestObj);
		goto exit;
	}


	hr = FindInterface(&IID_IMAPIProp, ciidExclude, rgiidExclude);
	if (hr == hrSuccess) {
		hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
		goto exit;
	}

	// end sanity checks


	// check message, folder, attach, recipients, stream, mapitable, ... ?

	if (*lpSrcInterface == IID_IMAPIFolder) {

		// MS MAPI does not perform this check
		if (*lpDestInterface != IID_IMAPIFolder) {
			// on store, create folder and still go ?
			hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
			goto exit;
		}

		if (!lpExcludeProps || Util::HrFindPropInArray(lpExcludeProps, PR_CONTAINER_CONTENTS) == -1) {
			sExtraExcludes.aulPropTag[sExtraExcludes.cValues++] = PR_CONTAINER_CONTENTS;
			hr = CopyContents(0, (LPMAPIFOLDER)lpSrcObj, (LPMAPIFOLDER)lpDestObj, ulFlags, ulUIParam, lpProgress);
			if (hr != hrSuccess)
				bPartial = true;
		}

		if (!lpExcludeProps || Util::HrFindPropInArray(lpExcludeProps, PR_FOLDER_ASSOCIATED_CONTENTS) == -1) {
			sExtraExcludes.aulPropTag[sExtraExcludes.cValues++] = PR_FOLDER_ASSOCIATED_CONTENTS;
			hr = CopyContents(MAPI_ASSOCIATED, (LPMAPIFOLDER)lpSrcObj, (LPMAPIFOLDER)lpDestObj, ulFlags, ulUIParam, lpProgress);
			if (hr != hrSuccess)
				bPartial = true;
		}

		if (!lpExcludeProps || Util::HrFindPropInArray(lpExcludeProps, PR_CONTAINER_HIERARCHY) == -1) {
			// add to lpExcludeProps so CopyProps ignores them
			sExtraExcludes.aulPropTag[sExtraExcludes.cValues++] = PR_CONTAINER_HIERARCHY;
			hr = CopyHierarchy((LPMAPIFOLDER)lpSrcObj, (LPMAPIFOLDER)lpDestObj, ulFlags, ulUIParam, lpProgress);
			if (hr != hrSuccess)
				bPartial = true;
		}

	} else if (*lpSrcInterface == IID_IMessage) {
		// recipients & attachments
		// this is done in CopyProps ()
	} else if (*lpSrcInterface == IID_IAttachment) {
		// data stream
		// this is done in CopyProps ()
	} else if (*lpSrcInterface == IID_IMAPIContainer || *lpSrcInterface == IID_IMAPIProp) {
		// props only
		// this is done in CopyProps ()
	} else if (*lpSrcInterface == IID_IMailUser || *lpSrcInterface == IID_IDistList) {
		// in one if() ?
		// this is done in CopyProps () ???
		// what else besides props ???
	} else {
		// stores, ... ?
		hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
		goto exit;
	}

	// we have a IMAPIProp compatible interface here, and we don't want to crash

	hr = lpUnkSrc->QueryInterface(IID_IMAPIProp, (void**)&lpPropSrc);
	if (hr != hrSuccess)
		goto exit;

	hr = lpUnkDest->QueryInterface(IID_IMAPIProp, (void**)&lpPropDest);
	if (hr != hrSuccess)
		goto exit;

	hr = lpPropSrc->GetPropList(0, &lpSPropTagArray);
	if (FAILED(hr))
		goto exit;

	// filter excludes
	if (lpExcludeProps || sExtraExcludes.cValues != 0) {
		for (ULONG i = 0; i < lpSPropTagArray->cValues; i++) {
			if (lpExcludeProps && Util::HrFindPropInArray(lpExcludeProps, lpSPropTagArray->aulPropTag[i]) != -1)
				lpSPropTagArray->aulPropTag[i] = PR_NULL;
			else if (Util::HrFindPropInArray((LPSPropTagArray)&sExtraExcludes, lpSPropTagArray->aulPropTag[i]) != -1)
				lpSPropTagArray->aulPropTag[i] = PR_NULL;
		}
	}

	// this is input for CopyProps
	hr = Util::DoCopyProps(lpSrcInterface, lpSrcObj, lpSPropTagArray, ulUIParam, lpProgress, lpDestInterface, lpDestObj, 0, lppProblems);
	if (hr != hrSuccess)
		goto exit;

	// TODO: mapi move, delete OPEN message ???

exit:
	// Partial warning when data was copied.
	if (bPartial)
		hr = MAPI_W_PARTIAL_COMPLETION;

	if (lpProps)
		MAPIFreeBuffer(lpProps);

	if (lpSPropTagArray)
		MAPIFreeBuffer(lpSPropTagArray);

	if (lpPropSrc)
		lpPropSrc->Release();

	if (lpPropDest)
		lpPropDest->Release();

	return hr;
}

HRESULT Util::DoCopyProps(LPCIID lpSrcInterface, LPVOID lpSrcObj, LPSPropTagArray lpIncludeProps, ULONG ulUIParam,
						  LPMAPIPROGRESS lpProgress, LPCIID lpDestInterface, LPVOID lpDestObj, ULONG ulFlags,
						  LPSPropProblemArray * lppProblems)
{
	HRESULT hr = hrSuccess;
	LPUNKNOWN lpUnkSrc = (LPUNKNOWN)lpSrcObj, lpUnkDest = (LPUNKNOWN)lpDestObj;
	bool bPartial = false;

	LPMAPIPROP lpSrcProp = NULL, lpDestProp = NULL;
	LPSPropValue lpProps = NULL;
	ULONG cValues = 0;
	LPSPropTagArray lpsDestPropArray = NULL;
	LPSPropProblemArray lpProblems = NULL;

	// named props
	ULONG cNames = 0;
	LPSPropTagArray lpsSrcNameTagArray = NULL, lpsDestNameTagArray = NULL;
	LPMAPINAMEID *lppNames = NULL;
	LPSPropTagArray lpsDestTagArray = NULL;

	// attachments
	LPSPropValue lpAttachMethod = NULL;
	LPSTREAM lpSrcStream = NULL, lpDestStream = NULL;
	LPMESSAGE lpSrcMessage = NULL, lpDestMessage = NULL;

	LONG ulIdRTF = Util::HrFindPropInArray(lpIncludeProps, PR_RTF_COMPRESSED);
	LONG ulIdHTML = Util::HrFindPropInArray(lpIncludeProps, PR_HTML);
	LONG ulIdBODY = Util::HrFindPropInArray(lpIncludeProps, PR_BODY);

	if (!lpSrcInterface || !lpDestInterface || !lpSrcObj || !lpDestObj || !lpIncludeProps) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	if (*lpSrcInterface != *lpDestInterface) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// q-i src and dest to check if IID_IMAPIProp is present
	// exit if not! (TODO: error??)
	hr = lpUnkSrc->QueryInterface(IID_IMAPIProp, (void**)&lpSrcProp);
	if (hr != hrSuccess)
		goto exit;

	hr = lpUnkDest->QueryInterface(IID_IMAPIProp, (void**)&lpDestProp);
	if (hr != hrSuccess)
		goto exit;


	if (ulFlags & MAPI_NOREPLACE) {
		hr = lpDestProp->GetPropList(0, &lpsDestPropArray);
		if (hr != hrSuccess)
			goto exit;

		for (ULONG i = 0; i < lpIncludeProps->cValues; i++) {
			if (Util::HrFindPropInArray(lpsDestPropArray, lpIncludeProps->aulPropTag[i]) != -1) {
				// hr = MAPI_E_COLLISION;
				// goto exit;
				// MSDN says collision, MS MAPI ignores these properties.
				lpIncludeProps->aulPropTag[i] = PR_NULL;
			}
		}
	}

	// Use only one body text property, RTF, HTML or BODY
	ulIdRTF = Util::HrFindPropInArray(lpIncludeProps, PR_RTF_COMPRESSED);
	ulIdHTML = Util::HrFindPropInArray(lpIncludeProps, PR_HTML);
	ulIdBODY = Util::HrFindPropInArray(lpIncludeProps, PR_BODY);

	if (ulIdRTF != -1) {
		if(ulIdHTML != -1)
			lpIncludeProps->aulPropTag[ulIdHTML] = PR_NULL;
		if(ulIdBODY != -1)
			lpIncludeProps->aulPropTag[ulIdBODY] = PR_NULL;
	} else if (ulIdHTML != -1) {
		if(ulIdBODY != -1)
			lpIncludeProps->aulPropTag[ulIdBODY] = PR_NULL;
	}

	for (ULONG i = 0; i < lpIncludeProps->cValues; i++) {
		bool isProblem = false;

		// TODO: ?
		// for all PT_OBJECT properties on IMAPIProp, MS MAPI tries:
		// IID_IMessage, IID_IStreamDocfile, IID_IStorage

		if (PROP_TYPE(lpIncludeProps->aulPropTag[i]) == PT_OBJECT ||
			PROP_ID(lpIncludeProps->aulPropTag[i]) == PROP_ID(PR_ATTACH_DATA_BIN)) {
			// if IMessage: PR_MESSAGE_RECIPIENTS, PR_MESSAGE_ATTACHMENTS
			if (*lpSrcInterface == IID_IMessage) {

				if (lpIncludeProps->aulPropTag[i] == PR_MESSAGE_RECIPIENTS) {
					// TODO: add ulFlags, and check for MAPI_NOREPLACE
					hr = Util::CopyRecipients((LPMESSAGE)lpSrcObj, (LPMESSAGE)lpDestObj);
				} else if (lpIncludeProps->aulPropTag[i] == PR_MESSAGE_ATTACHMENTS) {
					// TODO: add ulFlags, and check for MAPI_NOREPLACE
					hr = Util::CopyAttachments((LPMESSAGE)lpSrcObj, (LPMESSAGE)lpDestObj);
				} else {
					hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
				}
				if (hr != hrSuccess) {
					isProblem = true;
					goto next_include_check;
				}

			} else if (*lpSrcInterface == IID_IMAPIFolder) {
				// MS MAPI skips these in CopyProps(), for unknown reasons
				if (lpIncludeProps->aulPropTag[i] == PR_CONTAINER_CONTENTS ||
					lpIncludeProps->aulPropTag[i] == PR_CONTAINER_HIERARCHY ||
					lpIncludeProps->aulPropTag[i] == PR_FOLDER_ASSOCIATED_CONTENTS) {
					lpIncludeProps->aulPropTag[i] = PR_NULL;
				} else {
					isProblem = true;
				}
			} else if (*lpSrcInterface == IID_IAttachment) {
				// In attachments, IID_IMessage can be present!  for PR_ATTACH_DATA_OBJ
				// find method and copy this PT_OBJECT
				hr = HrGetOneProp(lpSrcProp, PR_ATTACH_METHOD, &lpAttachMethod);
				if (hr != hrSuccess) {
					isProblem = true;
					goto next_include_check;
				}

				switch(lpAttachMethod->Value.ul) {
				case ATTACH_BY_VALUE:
				case ATTACH_OLE:
					// stream

					// Not being able to open the source message is not an error: it may just not be there
					if(((LPATTACH)lpSrcObj)->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, STGM_TRANSACTED, 0, (LPUNKNOWN *)&lpSrcStream) == hrSuccess) {
						// While dragging and dropping, Outlook 2007 (atleast) returns an internal MAPI object to CopyTo as destination
						// The internal MAPI object is unable to make a stream STGM_TRANSACTED, so we retry the action without that flag
						// to get the stream without the transaction feature.
						hr = ((LPATTACH)lpDestObj)->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, STGM_WRITE | STGM_TRANSACTED, MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN *)&lpDestStream);
						if (hr == E_FAIL)
							hr = ((LPATTACH)lpDestObj)->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, STGM_WRITE, MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN *)&lpDestStream);
						if (hr != hrSuccess) {
							isProblem = true;
							goto next_include_check;
						}

						hr = Util::CopyStream(lpSrcStream, lpDestStream);
						if (hr != hrSuccess) {
							isProblem = true;
							goto next_include_check;
						}
					}

					break;
				case ATTACH_EMBEDDED_MSG:
					// message

					if(((LPATTACH)lpSrcObj)->OpenProperty(PR_ATTACH_DATA_OBJ, &IID_IMessage, 0, 0, (LPUNKNOWN *)&lpSrcMessage) == hrSuccess) {
						// Not being able to open the source message is not an error: it may just not be there
						hr = ((LPATTACH)lpDestObj)->OpenProperty(PR_ATTACH_DATA_OBJ, &IID_IMessage, 0, MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN *)&lpDestMessage);
						if (hr != hrSuccess) {
							isProblem = true;
							goto next_include_check;
						}

						hr = Util::DoCopyTo(&IID_IMessage, lpSrcMessage, 0, NULL, NULL, ulUIParam, lpProgress, &IID_IMessage, lpDestMessage, 0, NULL);
						if (hr != hrSuccess) {
							isProblem = true;
							goto next_include_check;
						}

						hr = lpDestMessage->SaveChanges(0);
						if (hr != hrSuccess) {
							isProblem = true;
							goto next_include_check;
						}
					}

					break;
				default:
					// OLE objects?
					isProblem = true;
					break;
				};
			} else {
				isProblem = true;
			}

			// TODO: try the 3 MSMAPI interfaces (message, stream, storage) if unhandled?

next_include_check:

			if (isProblem) {
				SPropProblem sProblem;

				bPartial = true;

				sProblem.ulIndex = i;
				sProblem.ulPropTag = lpIncludeProps->aulPropTag[i];
				sProblem.scode = MAPI_E_INTERFACE_NOT_SUPPORTED; // hr?

				hr = AddProblemToArray(&sProblem, &lpProblems);
				if (hr != hrSuccess)
					goto exit;
			}

			if (lpAttachMethod) {
				MAPIFreeBuffer(lpAttachMethod);
				lpAttachMethod = NULL;
			}

			if (lpDestStream) {
				lpDestStream->Release();
				lpDestStream = NULL;
			}

			if (lpSrcStream) {
				lpSrcStream->Release();
				lpSrcStream = NULL;
			}

			if (lpSrcMessage) {
				lpSrcMessage->Release();
				lpSrcMessage = NULL;
			}

			if (lpDestMessage) {
				lpDestMessage->Release();
				lpDestMessage = NULL;
			}

			// skip this prop for the final SetProps()
			lpIncludeProps->aulPropTag[i] = PR_NULL;
		}
	}


	hr = lpSrcProp->GetProps(lpIncludeProps, 0, &cValues, &lpProps);
	if (FAILED(hr))
		goto exit;

	// make map for destination property tags, because named id's may differ in src and dst
	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cValues), (void**)&lpsDestTagArray);
	if (hr != hrSuccess)
		goto exit;

	// get named props
	for (ULONG i = 0; i < cValues; i++) {
		lpsDestTagArray->aulPropTag[i] = lpProps[i].ulPropTag;

		if (PROP_ID(lpProps[i].ulPropTag) >= 0x8000)
			cNames++;
	}

	if (cNames) {
		hr = MAPIAllocateBuffer(CbNewSPropTagArray(cNames), (void**)&lpsSrcNameTagArray);
		if (hr != hrSuccess)
			goto exit;

		lpsSrcNameTagArray->cValues = cNames;
		cNames = 0;
		for (ULONG i = 0; i < cValues; i++) {
			if (PROP_ID(lpProps[i].ulPropTag) >= 0x8000)
				lpsSrcNameTagArray->aulPropTag[cNames++] = lpProps[i].ulPropTag;
		}

		// ignore warnings on unknown named properties, but don't copy those either (see PT_ERROR below)
		hr = lpSrcProp->GetNamesFromIDs(&lpsSrcNameTagArray, NULL, 0, &cNames, &lppNames);
		if (FAILED(hr))
			goto exit;

		hr = lpDestProp->GetIDsFromNames(cNames, lppNames, MAPI_CREATE, &lpsDestNameTagArray);
		if (FAILED(hr))
			goto exit;

		// make new lookup map for lpProps[] -> lpsDestNameTag[]
		for (ULONG i=0, j=0; i < cValues && j < cNames; i++) {
			if (PROP_ID(lpProps[i].ulPropTag) == PROP_ID(lpsSrcNameTagArray->aulPropTag[j])) {
				if (PROP_TYPE(lpsDestNameTagArray->aulPropTag[j]) != PT_ERROR) {
					// replace with new proptag, so we can open the correct property
					lpsDestTagArray->aulPropTag[i] = PROP_TAG(PROP_TYPE(lpProps[i].ulPropTag), PROP_ID(lpsDestNameTagArray->aulPropTag[j]));
				} else {
					// leave on PT_ERROR, so we don't copy the property
					lpsDestTagArray->aulPropTag[i] = PROP_TAG(PT_ERROR, PROP_ID(lpsDestNameTagArray->aulPropTag[j]));
					// don't even return a warning because although not all data could be copied
				}
				j++;
			}
		}
	}


	// find all MAPI_E_NOT_ENOUGH_MEMORY errors
	for (ULONG i = 0; i < cValues; i++) {
		if (PROP_TYPE(lpProps[i].ulPropTag) == PT_ERROR && lpProps[i].Value.err == MAPI_E_NOT_ENOUGH_MEMORY) {
			hr = Util::TryOpenProperty(PT_STRING8, lpProps[i].ulPropTag, lpSrcProp, lpsDestTagArray->aulPropTag[i], lpDestProp, &lpSrcStream, &lpDestStream);
			if (hr != hrSuccess)
				hr = Util::TryOpenProperty(PT_BINARY, lpProps[i].ulPropTag, lpSrcProp, lpsDestTagArray->aulPropTag[i], lpDestProp, &lpSrcStream, &lpDestStream);
			if (hr != hrSuccess)
				hr = Util::TryOpenProperty(PT_UNICODE, lpProps[i].ulPropTag, lpSrcProp, lpsDestTagArray->aulPropTag[i], lpDestProp, &lpSrcStream, &lpDestStream);

			if (hr != hrSuccess) {
				// TODO: check, partial or problemarray?
				bPartial = true;
				goto next_stream_prop;
			}

			hr = Util::CopyStream(lpSrcStream, lpDestStream);
			if (hr != hrSuccess)
				bPartial = true;

next_stream_prop:
			if (lpSrcStream) {
				lpSrcStream->Release();
				lpSrcStream = NULL;
			}

			if (lpDestStream) {
				lpDestStream->Release();
				lpDestStream = NULL;
			}
		}
	}

	// set destination proptags in orignal properties
	for (ULONG i=0; i < cValues; i++) {
		lpProps[i].ulPropTag = lpsDestTagArray->aulPropTag[i];
	}

	hr = lpDestProp->SetProps(cValues, lpProps, NULL);
	if (FAILED(hr))
		goto exit;

	// TODO: test how this should work on CopyProps() !!
	if (ulFlags & MAPI_MOVE) {
		// TODO: add problem array
		hr = lpSrcProp->DeleteProps(lpIncludeProps, NULL);
		if (FAILED(hr))
			goto exit;
	}

exit:
	if (bPartial)
		hr = MAPI_W_PARTIAL_COMPLETION;

	if (hr != hrSuccess) {
		// may not return a problem set when we have an warning/error code in hr
		if (lpProblems)
			MAPIFreeBuffer(lpProblems);
	} else {
		if (lppProblems)
			*lppProblems = lpProblems;
		else if (lpProblems)
			MAPIFreeBuffer(lpProblems);
	}

	if (lppNames)
		MAPIFreeBuffer(lppNames);

	if (lpsSrcNameTagArray)
		MAPIFreeBuffer(lpsSrcNameTagArray);

	if (lpsDestNameTagArray)
		MAPIFreeBuffer(lpsDestNameTagArray);

	if (lpsDestTagArray)
		MAPIFreeBuffer(lpsDestTagArray);

	if (lpSrcMessage)
		lpSrcMessage->Release();

	if (lpDestMessage)
		lpDestMessage->Release();

	if (lpSrcStream)
		lpSrcStream->Release();

	if (lpDestStream)
		lpDestStream->Release();

	if (lpAttachMethod)
		MAPIFreeBuffer(lpAttachMethod);

	if (lpSrcProp)
		lpSrcProp->Release();

	if (lpDestProp)
		lpDestProp->Release();

	if (lpProps)
		MAPIFreeBuffer(lpProps);

	if (lpsDestPropArray)
		MAPIFreeBuffer(lpsDestPropArray);

	return hr;
}

HRESULT Util::HrGetQuotaStatus(IMsgStore *lpMsgStore, LPECQUOTA lpsQuota, LPECQUOTASTATUS *lppsQuotaStatus)
{
	HRESULT			hr = hrSuccess;
	LPECQUOTASTATUS	lpsQuotaStatus = NULL;
	LPSPropValue 	lpProps = NULL;
    SizedSPropTagArray(1, sptaProps) = {1, {PR_MESSAGE_SIZE_EXTENDED}};
    ULONG 			cValues = 0;
	
	if (lpMsgStore == NULL || lppsQuotaStatus == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	hr = lpMsgStore->GetProps((LPSPropTagArray)&sptaProps, 0, &cValues, &lpProps);
	if (hr != hrSuccess)
		goto exit;
		
	if (cValues != 1 || lpProps[0].ulPropTag != PR_MESSAGE_SIZE_EXTENDED) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	hr = MAPIAllocateBuffer(sizeof *lpsQuotaStatus, (void**)&lpsQuotaStatus);
	if (hr != hrSuccess)
		goto exit;
	memset(lpsQuotaStatus, 0, sizeof *lpsQuotaStatus);
	
	lpsQuotaStatus->llStoreSize = lpProps[0].Value.li.QuadPart;
	lpsQuotaStatus->quotaStatus = QUOTA_OK;
	if (lpsQuota && lpsQuotaStatus->llStoreSize > 0) {
		if (lpsQuota->llHardSize > 0 && lpsQuotaStatus->llStoreSize > lpsQuota->llHardSize)
			lpsQuotaStatus->quotaStatus = QUOTA_HARDLIMIT;
		else if (lpsQuota->llSoftSize > 0 && lpsQuotaStatus->llStoreSize > lpsQuota->llSoftSize)
			lpsQuotaStatus->quotaStatus = QUOTA_SOFTLIMIT;
		else if (lpsQuota->llWarnSize > 0 && lpsQuotaStatus->llStoreSize > lpsQuota->llWarnSize)
			lpsQuotaStatus->quotaStatus = QUOTA_WARN;
	}
	
	*lppsQuotaStatus = lpsQuotaStatus;
	lpsQuotaStatus = NULL;
	
exit:
	if (lpsQuotaStatus)
		MAPIFreeBuffer(lpsQuotaStatus);
		
	if (lpProps)
		MAPIFreeBuffer(lpProps);

	return hr;
}

HRESULT Util::HrDeleteResidualProps(LPMESSAGE lpDestMsg, LPMESSAGE lpSourceMsg, LPSPropTagArray lpsValidProps)
{
	HRESULT			hr = hrSuccess;
	LPSPropTagArray	lpsPropArray = NULL;
	LPSPropTagArray	lpsNamedPropArray = NULL;
	LPSPropTagArray	lpsMappedPropArray = NULL;
	ULONG			cPropNames = 0;
	LPMAPINAMEID	*lppPropNames = NULL;
	PropTagSet		sPropTagSet;

	if (lpDestMsg == NULL || lpSourceMsg == NULL || lpsValidProps == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpDestMsg->GetPropList(0, &lpsPropArray);
	if (hr != hrSuccess || lpsPropArray->cValues == 0)
		goto exit;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray(lpsValidProps->cValues), (void**)&lpsNamedPropArray);
	if (hr != hrSuccess)
		goto exit;
	memset(lpsNamedPropArray, 0, CbNewSPropTagArray(lpsValidProps->cValues));

	for (unsigned i = 0; i < lpsValidProps->cValues; ++i)
		if (PROP_ID(lpsValidProps->aulPropTag[i]) >= 0x8000)
			lpsNamedPropArray->aulPropTag[lpsNamedPropArray->cValues++] = lpsValidProps->aulPropTag[i];

	if (lpsNamedPropArray->cValues > 0) {
		hr = lpSourceMsg->GetNamesFromIDs(&lpsNamedPropArray, NULL, 0, &cPropNames, &lppPropNames);
		if (FAILED(hr))
			goto exit;

		hr = lpDestMsg->GetIDsFromNames(cPropNames, lppPropNames, MAPI_CREATE, &lpsMappedPropArray);
		if (FAILED(hr))
			goto exit;
	}
	hr = hrSuccess;

	// Add the PropTags the message currently has
	for (unsigned i = 0; i < lpsPropArray->cValues; ++i)
		sPropTagSet.insert(lpsPropArray->aulPropTag[i]);

	// Remove the regular properties we want to keep
	for (unsigned i = 0; i < lpsValidProps->cValues; ++i)
		if (PROP_ID(lpsValidProps->aulPropTag[i]) < 0x8000)
			sPropTagSet.erase(lpsValidProps->aulPropTag[i]);

	// Remove the mapped named properties we want to keep. Filter failed named properties, so they will be removed
	for (unsigned i = 0; lpsMappedPropArray != NULL && i < lpsMappedPropArray->cValues; ++i)
		if (PROP_TYPE(lpsMappedPropArray->aulPropTag[i]) != PT_ERROR)
			sPropTagSet.erase(lpsMappedPropArray->aulPropTag[i]);

	if (sPropTagSet.empty())
		goto exit;

	// Reuse lpsPropArray to hold the properties we're going to delete
	ASSERT(lpsPropArray->cValues >= sPropTagSet.size());
	memset(lpsPropArray->aulPropTag, 0, lpsPropArray->cValues * sizeof *lpsPropArray->aulPropTag);
	lpsPropArray->cValues = 0;

	for (PropTagSet::const_iterator it = sPropTagSet.begin(); it != sPropTagSet.end(); ++it)
		lpsPropArray->aulPropTag[lpsPropArray->cValues++] = *it;

	hr = lpDestMsg->DeleteProps(lpsPropArray, NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpDestMsg->SaveChanges(KEEP_OPEN_READWRITE);

exit:
	if (lpsMappedPropArray)
		MAPIFreeBuffer(lpsMappedPropArray);

	if (lppPropNames)
		MAPIFreeBuffer(lppPropNames);

	if (lpsNamedPropArray)
		MAPIFreeBuffer(lpsNamedPropArray);

	if (lpsPropArray)
		MAPIFreeBuffer(lpsPropArray);

	return hr;
}
