/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "str1252.h"
#include "string.h"
#include "stringutil.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#endif

// Lookup table from windows1252-char to sort key. Generated from 'ECtools/strxlat 3'
unsigned char windows1252_rev[] = {
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F, 
    0x20, 0x21, 0x22, 0x23, 0x70, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 
    0x73, 0x75, 0x79, 0x7B, 0x7E, 0x7F, 0x80, 0x81, 0x82, 0x83, 0x2F, 0x30, 0x31, 0x32, 0x33, 0x34, 
    0x35, 0x87, 0x99, 0x9B, 0x9F, 0xA3, 0xAD, 0xAF, 0xB1, 0xB3, 0xBD, 0xBF, 0xC1, 0xC3, 0xC5, 0xC9, 
    0xD8, 0xDA, 0xDC, 0xDF, 0xE3, 0xE5, 0xF0, 0xF2, 0xF5, 0xF7, 0xFC, 0x36, 0x37, 0x38, 0x39, 0x3A, 
    0x3B, 0x85, 0x98, 0x9A, 0x9E, 0xA2, 0xAC, 0xAE, 0xB0, 0xB2, 0xBC, 0xBE, 0xC0, 0xC2, 0xC4, 0xC8, 
    0xD7, 0xD9, 0xDB, 0xDE, 0xE2, 0xE4, 0xEF, 0xF1, 0xF3, 0xF6, 0xFB, 0x3C, 0x3D, 0x3E, 0x3F, 0x40, 
    0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x96, 0x74, 0x49, 0x84, 0xDD, 0x4A, 0x4B, 0x4C, 
    0x4D, 0x4E, 0x4F, 0x50, 0x51, 0x52, 0x53, 0x54, 0xED, 0x55, 0x86, 0x56, 0xE0, 0x57, 0x58, 0xF4, 
    0x59, 0x5A, 0x6F, 0x71, 0x6E, 0x72, 0x5B, 0x5C, 0x5D, 0x5E, 0x88, 0x5F, 0x60, 0x61, 0x62, 0x63, 
    0x64, 0x65, 0x7A, 0x7C, 0x66, 0xFF, 0x67, 0x68, 0x69, 0x76, 0xCA, 0x6A, 0x78, 0x77, 0x7D, 0x6B, 
    0x8C, 0x8A, 0x8E, 0x94, 0x92, 0x90, 0x97, 0x9D, 0xA7, 0xA5, 0xA9, 0xAB, 0xB7, 0xB5, 0xB9, 0xBB, 
    0xA1, 0xC7, 0xCE, 0xCC, 0xD0, 0xD4, 0xD2, 0x6C, 0xD6, 0xE9, 0xE7, 0xEB, 0xEE, 0xF9, 0xFE, 0xE1, 
    0x8B, 0x89, 0x8D, 0x93, 0x91, 0x8F, 0x95, 0x9C, 0xA6, 0xA4, 0xA8, 0xAA, 0xB6, 0xB4, 0xB8, 0xBA, 
    0xA0, 0xC6, 0xCD, 0xCB, 0xCF, 0xD3, 0xD1, 0x6D, 0xD5, 0xE8, 0xE6, 0xEA, 0xEC, 0xF8, 0xFD, 0xFA, 
};

// This is the inverse of the above:

// 00, 01, 02, 03, 04, 05, 06, 07, 08, 09, 0A, 0B, 0C, 0D, 0E, 0F, 
// 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 1A, 1B, 1C, 1D, 1E, 1F, 
//  , !, ", #, %, &, ', (, ), *, +, ,, -, ., /, :, 
// ;, <, =, >, ?, @, [, \, ], ^, _, `, {, |, }, ~, 
// , , , , , , , , , , , , , , , , 
// , , , , , , , , , , , , , , , , 
// , , , , , , , , , , , , , , , , 
// $, , , 0, , 1, , , , 2, , 3, , , 4, 5, 
// 6, 7, 8, 9, , a, , A, , , , , , , , , 
// , , , , , , , , b, B, c, C, , , d, D, 
// , , e, E, , , , , , , , , f, F, g, G, 
// h, H, i, I, , , , , , , , , j, J, k, K, 
// l, L, m, M, n, N, , , o, O, , , , , , , 
// , , , , , , , p, P, q, Q, r, R, , s, S, 
// , , t, T, u, U, , , , , , , , , , v, 
// V, w, W, x, , X, y, Y, , , , z, Z, , , , 


unsigned char windows1252_upper[] = {
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F, 
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F, 
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F, 0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F, 
    0x60, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F, 0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F, 
    0x80, 0x81, 0x82, 0x00, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F, 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x8A, 0x9B, 0x8C, 0x9D, 0x8E, 0x9F, 
    0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF, 0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0x00, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF, 
    0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF, 0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF, 
    0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF, 0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xF7, 0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0x9F, 
};

// This is the above table:
// 00 , 01 , 02 , 03 , 04 , 05 , 06 , 07 , 08 , 09 , 0A , 0B , 0C , 0D , 0E , 0F , 10 , 11 , 12 , 13 , 14 , 15 , 16 , 17 , 18 , 19 , 1A , 1B , 1C , 1D , 1E , 1F , 
//   , ! , " , # , $ , % , & , ' , ( , ) , * , + , , , - , . , / , 0 , 1 , 2 , 3 , 4 , 5 , 6 , 7 , 8 , 9 , : , ; , < , = , > , ? , 
// @ , A , B , C , D , E , F , G , H , I , J , K , L , M , N , O , P , Q , R , S , T , U , V , W , X , Y , Z , [ , \ , ] , ^ , _ , 
// ` , A , B , C , D , E , F , G , H , I , J , K , L , M , N , O , P , Q , R , S , T , U , V , W , X , Y , Z , { , | , } , ~ ,  , 
//  ,  ,  , 00 ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  , 
//  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  , 00 ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  , 
//  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  , 
//  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  , 

unsigned char windows1252_lower[] = {
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F, 
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F, 
    0x40, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F, 
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F, 
    0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x9A, 0x8B, 0x9C, 0x8D, 0x9E, 0x8F, 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0xFF, 
    0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF, 0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF, 
    0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF, 0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xD7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xDF, 
    0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF, 0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF, 
};

// This is the above table:
// 00 , 01 , 02 , 03 , 04 , 05 , 06 , 07 , 08 , 09 , 0A , 0B , 0C , 0D , 0E , 0F , 10 , 11 , 12 , 13 , 14 , 15 , 16 , 17 , 18 , 19 , 1A , 1B , 1C , 1D , 1E , 1F , 
//   , ! , " , # , $ , % , & , ' , ( , ) , * , + , , , - , . , / , 0 , 1 , 2 , 3 , 4 , 5 , 6 , 7 , 8 , 9 , : , ; , < , = , > , ? , 
// @ , a , b , c , d , e , f , g , h , i , j , k , l , m , n , o , p , q , r , s , t , u , v , w , x , y , z , [ , \ , ] , ^ , _ , 
// ` , a , b , c , d , e , f , g , h , i , j , k , l , m , n , o , p , q , r , s , t , u , v , w , x , y , z , { , | , } , ~ ,  , 
//  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  , 
//  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  , 
//  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  , 
//  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  ,  , 


void memxfrm_str1252(char *str, int len) {
    int i;
    
    for(i=0;i<len;i++) {
        str[i] = windows1252_rev[(unsigned char)str[i]];
    }
}

void strxfrm_str1252(char *str) {
    while(*str) {
        *str = windows1252_rev[(unsigned char)*str];
        str++;
    }
}

void strupr_str1252(char *str) {
    while(*str) {
        *str = windows1252_upper[(unsigned char)*str];
        str++;
    }
}

void strlwr_str1252(char *str) {
    while(*str) {
        *str = windows1252_lower[(unsigned char)*str];
        str++;
    }
}

int stricmp_str1252(char *str1, char *str2) {
    return strcasecmp_str1252(str1, str2);
}

int strncmp_str1252(char *str1, char *str2, int n) {
    return strncasecmp_str1252(str1, str2, n);
}

int strcasecmp_str1252(char *str1, char *str2) {
    return strncasecmp_str1252(str1, str2, 0x7fffffff);
}

int strnicmp_str1252(char *str1, char *str2, int n) {
    return strncasecmp_str1252(str1, str2, n);
}

// First, upcase both strings, then convert to 'sortkey', and THEN compare. This makes sure
// that it will return TRUE when strings are equal while ignoring case, and ordering will
// be same as when using strxfrm_str1252

int strncasecmp_str1252(char *str1, char *str2, int n) {
    while(*str1 && *str2 && n) {
        int c1 = windows1252_rev[windows1252_upper[(unsigned char)*str1]];
        int c2 = windows1252_rev[windows1252_upper[(unsigned char)*str2]];
        
        if(c1 < c2) return -1;
        if(c1 > c2) return 1;
        
        str1++;
        str2++;
        n--;
    }
    
    if(n == 0 || (*str1 == 0 && *str2 == 0))
        return 0;
        
    if(*str1 == 0)
        return -1;

    return 1;
}

int strcmp_str1252(char *str1, char *str2) {
    while(*str1 && *str2) {
        int c1 = windows1252_rev[(unsigned char)*str1];
        int c2 = windows1252_rev[(unsigned char)*str2];
        
        if(c1 < c2) return -1;
        if(c1 > c2) return 1;
        
        str1++;
        str2++;
    }
    
    if(*str1 == 0 && *str2 == 0)
        return 0;
        
    if(*str1 == 0)
        return -1;
    
    return 1;
}

int memcmp_str1252(char *str1, unsigned int len1, char *str2, unsigned int len2) {
    while(len1 && len2) {
        int c1 = windows1252_rev[(unsigned char)*str1];
        int c2 = windows1252_rev[(unsigned char)*str2];
        
        if(c1 < c2) return -1;
        if(c1 > c2) return 1;
        
        str1++;
        len1--;
        str2++;
        len2--;
    }
    
    if(len1 == 0 && len2 == 0)
        return 0;
        
    if(len1 == 0)
        return -1;
    
    return 1;
}

int memcasecmp_str1252(char *str1, unsigned int len1, char *str2, unsigned int len2) {
    while(len1 && len2) {
        int c1 = windows1252_rev[windows1252_upper[(unsigned char)*str1]];
        int c2 = windows1252_rev[windows1252_upper[(unsigned char)*str2]];
        
        if(c1 < c2) return -1;
        if(c1 > c2) return 1;
        
        str1++;
        len1--;
        str2++;
        len2--;
    }
    
    if(len1 == 0 && len2 == 0)
        return 0;
        
    if(len1 == 0)
        return -1;

    return 1;
}

char *stristr_str1252(char *haystack, char *needle) {
    return strcasestr_str1252(haystack, needle);
}

char *strcasestr_str1252(char *haystack, char *needle) {
    char *needlepos = needle;
    char *needlestart = haystack;
    
    while(*haystack) {
        if(windows1252_upper[(unsigned char)*haystack] == windows1252_upper[(unsigned char)*needlepos]) {
            needlepos++;

            if(*needlepos == 0)
                return needlestart;
        } else {
            haystack = needlestart++;
            needlepos = needle;
        }

        haystack++;
    }
    
    return 0;
}

char *strstr_str1252(char *haystack, char *needle) {
    return strstr(haystack, needle);
}

int memsubstr_str1252(const void *haystack, size_t haystacklen, const void *needle, size_t needlelen) {
    return memsubstr(haystack, haystacklen, needle, needlelen);
}
