/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IMAP_H
#define IMAP_H

#define IMAP_HIERARCHY_DELIMITER "/"
#include <string>
#include <vector>
#include <list>
#include <set>

#include <ECIConv.h>
#include "ECChannel.h"

using namespace std;

#define ROWS_PER_REQUEST 20

#define PUBLIC_FOLDERS_NAME "Public folders"

enum IMAP_Command {
	IMAP_CMD_CAPABILITY, IMAP_CMD_NOOP, IMAP_CMD_LOGOUT, IMAP_CMD_STARTTLS, IMAP_CMD_AUTHENTICATE,
	IMAP_CMD_LOGIN, IMAP_CMD_SELECT, IMAP_CMD_EXAMINE, IMAP_CMD_CREATE, IMAP_CMD_DELETE, IMAP_CMD_RENAME,
	IMAP_CMD_SUBSCRIBE, IMAP_CMD_UNSUBSCRIBE, IMAP_CMD_LIST, IMAP_CMD_LSUB, IMAP_CMD_STATUS, IMAP_CMD_APPEND,
	IMAP_CMD_CHECK, IMAP_CMD_CLOSE, IMAP_CMD_EXPUNGE, IMAP_CMD_SEARCH, IMAP_CMD_FETCH, IMAP_CMD_STORE,
	IMAP_CMD_COPY, IMAP_CMD_UID, IMAP_CMD_XAOL_MOVE, IMAP_CMD_IDLE
};

LONG __stdcall IMAPIdleAdviseCallback(void *lpContext, ULONG cNotif, LPNOTIFICATION lpNotif);

class BinaryArray {
public:
    BinaryArray() { lpb = NULL; cb = 0; }
    BinaryArray(BYTE *lpData, ULONG cbData) { if(cbData == 0) { cb = 0; lpb = NULL; } else { lpb = new BYTE[cbData]; memcpy(lpb, lpData, cbData); cb = cbData; } }
    BinaryArray(const BinaryArray &old)     { if(old.cb == 0) { cb = 0; lpb = NULL; } else { cb = old.cb; lpb = new BYTE[cb]; memcpy(lpb, old.lpb, cb); } }
    BinaryArray(SBinary &bin) { if(bin.cb == 0) { cb = 0; lpb = NULL; } else { lpb = new BYTE[bin.cb]; memcpy(lpb, bin.lpb, bin.cb); cb = bin.cb; } }
    ~BinaryArray() { if(lpb) delete [] lpb; }

    bool operator == (const BinaryArray &b) { if(b.cb == 0 && this->cb == 0) return true; if(b.cb != this->cb) return false; else return memcmp(lpb, b.lpb, cb) == 0; }

    void operator = (const BinaryArray &b) { BYTE *lpbPrev = lpb; if(b.cb == 0) {cb=0; lpb = NULL; } else { cb=b.cb; lpb = new BYTE[cb]; memcpy(lpb, b.lpb, cb); if(lpbPrev) delete [] lpbPrev; } }
    
    BYTE *lpb;
    ULONG cb;
};


// FLAGS: \Seen \Answered \Flagged \Deleted \Draft \Recent
class IMAP {
public:
	IMAP(ECChannel *lpChan, char *szServerPath, ECLogger *lpLog, ECConfig *lpConf, bool bMailFoldersOnly, unsigned int tc, bool bPublicFolder = false);
	~IMAP();
	bool isLoggedIn();
	bool isIdle();

	HRESULT HrSplitInput(string &strInput, vector<string> &lstWords);
	HRESULT HrGetCommand(string &strCommand, IMAP_Command &eCommand);

	// All IMAP4rev1 commands
	HRESULT HrCmdCapability(string strTag);
	HRESULT HrCmdNoop(string strTag);
	HRESULT HrCmdLogout(string strTag);
	HRESULT HrCmdStarttls(string strTag);
	HRESULT HrCmdAuthenticate(string strTag, string strAuthMethod);
	HRESULT HrCmdLogin(string strTag, string strUser, string strPass);
	HRESULT HrCmdSelect(string strTag, string strFolder);
	HRESULT HrCmdExamine(string strTag, string strFolder);
	HRESULT HrCmdCreate(string strTag, string strFolder);
	HRESULT HrCmdDelete(string strTag, string strFolder);
	HRESULT HrCmdRename(string strTag, string strExistingFolder, string strNewFolder);
	HRESULT HrCmdSubscribe(string strTag, string strFolder, bool bSubscribe);
	HRESULT HrCmdList(string strTag, string strReferenceFolder, string strFolder, bool bSubscribedOnly);
	HRESULT HrCmdStatus(string strTag, string strFolder, string strStatusData);
	HRESULT HrCmdAppend(string strTag, string strFolder, string strData, string strFlags="", string strTime="");
	HRESULT HrCmdCheck(string strTag);
	HRESULT HrCmdClose(string strTag);
	HRESULT HrCmdExpunge(string strTag);
	HRESULT HrCmdSearch(string strTag, vector<string> &lstSearchCriteria, bool bUidMode);
	HRESULT HrCmdFetch(string strTag, string strSeqSet, string strMsgDataItemNames, bool bUidMode);
	HRESULT HrCmdStore(string strTag, string strSeqSet, string strMsgDataItemName, string strMsgDataItemValue, bool bUidMode);
	HRESULT HrCmdCopy(string strTag, string strSeqSet, string strFolder, bool bUidMode);
	HRESULT HrCmdUidXaolMove(string strTag, string strSeqSet, string strFolder);
	HRESULT HrCmdIdle(string strTag);
	HRESULT HrDone(bool bSendResponse);

	//untagged responses
	HRESULT HrResponse(string strResponse); // *
	HRESULT HrResponseContinuation(string strResponse); // +  

	//tagged responses = result
	HRESULT HrResultOk(string strTag, string strResponse); // OK
	HRESULT HrResultNo(string strTag, string strResponse); // NO
	HRESULT HrResultBad(string strTag, string strResponse); // BAD

private:
	bool bOnlyMailFolders;
	bool bShowPublicFolder;

	// All data per folder for the folderlist
	struct SFolder {
		BinaryArray sEntryID;	// EntryID of folder
		string strFolderName;	// Folder name
		bool bActive;			// Subscribed folder
		bool bMailFolder;		// E-mail type folder
		bool bSpecialFolder;	// 'special' folder (eg inbox)
		bool bHasSubfolders;	// Has child folders
		list<SFolder>::iterator lpParentFolder;
	};

	// All data to be mapped per mail in the current folder
	// Used class to be able to use sort
	class SMail {
    public:
        BinaryArray sEntryID;		// EntryID of message
        BinaryArray sInstanceKey;	// Instance key of message
		ULONG ulUid;				// PR_EC_IMAP_UID of message
		bool bRecent;				// \Recent flag
		std::string strFlags;		// String of all flags, including \Recent

		bool operator < (SMail sMail) const {
			return ulUid < sMail.ulUid;
		}
		bool operator == (unsigned int ulUid) const {
		    return this->ulUid == ulUid;
		}
	};

	IMAPISession *lpSession;
	IAddrBook *lpAddrBook;

	// list containing the tree of al folders in the store
	list<SFolder> lstFolders;

	// point to the current folder in lstFolders
	list<SFolder>::iterator lpCurrentFolder;

	// true if folder is opened with examine
	bool bCurrentFolderReadOnly;

	// vector of mails in the current folder. The index is used for mail number.
	vector<SMail> lstFolderMailEIDs;
	IMsgStore *lpStore;
	IMsgStore *lpPublicStore;

	// special folder entryids (not able to move/delete inbox and such ...)
	list<BinaryArray> lstSpecialEntryIDs;

	// path for the imap server
	string strPath;

	// Message cache
	string m_strCache;
	unsigned m_ulCacheUID;

	ECChannel *lpChannel;
	ECLogger *lpLogger;
	ECConfig *lpConfig;
	unsigned int m_thread_counter;

	bool m_bIdleMode;
	IMAPIAdviseSink *m_lpIdleAdviseSink;
	ULONG m_ulIdleAdviseConnection;
	string m_strIdleTag;
	IMAPITable *m_lpIdleTable;
	pthread_mutex_t m_mIdleLock;
	unsigned int ulLastUid;

	delivery_options dopt;

	/* subscribed folders */
	vector<BinaryArray> m_vSubscriptions;
	bool IsSubscribed(BinaryArray sEntryId);
	HRESULT HrGetSubscribedList();
	HRESULT HrSetSubscribedList();

	HRESULT HrMakeSpecialsList();
	bool isSpecialFolder(ULONG cb, LPBYTE lpEntryID);

	HRESULT HrRefreshFolders();
	HRESULT HrRefreshFolderMails(bool bInitialLoad, bool bResetRecent, bool bShowUID, unsigned int *lpulUnseen);

	list<SFolder>::iterator findFolder(string strFolder);
	HRESULT HrSelectFolder(string strFolder);

	HRESULT HrGetSubTree(SBinary &sEntryID, string strFolderName, list<SFolder>::iterator lpParentFolder, bool bSubfolders = true, bool bIsEmailFolder = true);
	HRESULT HrGetFolderPath(list<SFolder>::iterator lpFolder, string &strPath);
	HRESULT HrGetDataItems(string strMsgDataItemNames, vector<string> &lstDataItems);
	HRESULT HrSemicolonToComma(string &strData);

	// fetch calls an other fetch depending on the data items requested
	HRESULT HrPropertyFetch(list<ULONG> &lstMails, vector<string> &lstDataItems);
	HRESULT HrPropertyFetchRow(LPSPropValue lpProps, ULONG cValues, string &strResponse, ULONG ulMailnr, vector<string> &lstDataItems);

	HRESULT HrGetMessageEnvelope(string &strResponse, LPMESSAGE lpMessage);
	HRESULT HrGetMessageFlags(string &strResponse, LPMESSAGE lpMessage, bool bRecent);
	HRESULT HrGetMessageInterDate(string &strResponse, LPMESSAGE lpMessage);

	HRESULT HrGetMessagePart(string &strMessagePart, string &strMessage, string strPartName);
	HRESULT HrParseSeqSet(string strSeqSet, list<ULONG> &lstMails, bool bUID);
	HRESULT HrStore(list<ULONG> &lstMails, string strMsgDataItemName, string strMsgDataItemValue);
	HRESULT HrCopy(list<ULONG> &lstMails, string strFolder, bool bMove=false);
	HRESULT HrSearch(vector<string> &lstSearchCriteria, ULONG &ulStartCriteria, list<ULONG> &lstMailnr, ECIConv *iconv);

	string GetHeaderValue(const string &strMessage, const string &strHeader, const string &strDefault);
	HRESULT HrGetBodyStructure(bool bExtended, string &strBodyStructure, const string& strMessage);

	HRESULT HrGetEmailAddress(LPSPropValue lpPropValues, ULONG ulAddrType, ULONG ulEntryID, ULONG ulName, ULONG ulEmail, string strHeaderName, string *strHeaders);

	// Make the string uppercase
	void ToUpper(string &strString);
	bool CaseCompare(string strA, string strB);

	// used for Date in header
	string FileTimeToRFC822String(FILETIME sFileTime);

	// IMAP4rev1 date format: 01-Jan-2000 00:00 +0000
	string FileTimeToString(FILETIME sFiletime);
	FILETIME StringToFileTime(string strTime, bool bDateOnly = false);
	// add 24 hour to the time to be able to check if a time is on a date
	FILETIME AddDay(FILETIME sFileTime);

	// used for making an unique id with PR_RECORD_KEY
	ULONG BinaryToLong(LPSBinary lpBinary);

	// encodes an 8 bit string into a quoted-printable string
	string ToQP(char *string);

	// escape (quote) a string for a specific charset and encoding
	string EscapeString(char *string, char *charset = "windows-1252", char *encoding = "quoted-printable");

	// escape (quote) a string for a quoted-text (between "")
	string EscapeStringQT(string str);

	// Folder names are in a *modified* utf-7 form. See RFC2060, chapter 5.1.3
	string MAPI2IMAPCharset(string input);
	string IMAP2MAPICharset(string input);
	
	// Match a folder path
	bool MatchFolderPath(string strFolder, string strPattern);

	// Various conversion functions
	string PropsToFlags(LPSPropValue lpProps, unsigned int cValues, bool bRecent);
	string PropsToEmailAddress(LPSPropValue lpProps, unsigned int cValues, ULONG ulEmailPropTag, ULONG ulNamePropTag);

	void FixupBoundary(string &strHeaders);
	
	void HrParseHeaders(string &strHeaders, list<pair<string, string> > &lstHeaders);
	void HrGetSubString(string &strOutput, string &strInput, string strBegin, string strEnd);
	void HrTokenize(set<string> &setTokens, string &strInput);
	void HrTokenize(list<string> &lstTokens, string &strInput);

	friend LONG __stdcall IMAPIdleAdviseCallback(void *lpContext, ULONG cNotif, LPNOTIFICATION lpNotif);
};

#define IMAP_ARG_MAX	65536
#define IMAP_RESP_MAX	65536

#endif
