/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

// Damn windows header defines max which break C++ header files
#undef max

#include <iostream>

#include "ECVMIMEUtils.h"
#include "MAPISMTPTransport.h"
#include "CommonUtil.h"

#include <mapi.h>
#include <mapitags.h>
#include <mapidefs.h>
#include <mapiutil.h>
#include <mapix.h>
#include "mapiext.h"

ECVMIMESender::ECVMIMESender(ECLogger *newlpLogger, std::string strSMTPHost) : ECSender(newlpLogger, strSMTPHost) {
	lpSess = NULL;
	lpTransport = NULL;
}

ECVMIMESender::~ECVMIMESender() {
	if (lpSess)
		delete lpSess;
	if (lpTransport)
		delete lpTransport;
}

#ifdef DEBUG
HRESULT ECVMIMESender::printMail(vmime::message *lpVMMessage) {
	HRESULT hr = hrSuccess;
	string mail;

	try {
		// WARNING: BCC is not filtered here!
		mail = lpVMMessage->generate();
		cout << endl << endl << "----------" << endl << endl <<
			mail << endl << endl << "----------" << endl;
	}
	catch (...) {
		cerr << "bah, exception.\n";
		hr = MAPI_E_CALL_FAILED;
	}
	return hr;
}
#endif

// copied from vmime/messaging/transport.cpp
static void extractMailboxes(vmime::mailboxList& recipients, const vmime::addressList& list)
{
	for (int i = 0 ; i < list.getAddressCount() ; ++i)
	{
		vmime::mailbox* mbox = dynamic_cast <vmime::mailbox*>(list.getAddressAt(i)->clone());

		if (mbox != NULL)
			recipients.appendMailbox(mbox);
	}
}

HRESULT ECVMIMESender::HrMakeRecipientsList(LPADRBOOK lpAdrBook, LPMESSAGE lpMessage, vmime::message *lpVMMessage, vmime::mailboxList& recipients)
{
	HRESULT hr = hrSuccess;
	SRestriction sRestriction;
	SPropValue sRestrictProp;
	LPMAPITABLE lpRTable = NULL;
	LPSRowSet lpRowSet = NULL;
	std::string strName, strEmail, strType;

	sRestriction.rt = RES_PROPERTY;
	sRestriction.res.resProperty.relop = RELOP_EQ;
	sRestriction.res.resProperty.ulPropTag = PR_RECIPIENT_TYPE;
	sRestriction.res.resProperty.lpProp = &sRestrictProp;

	sRestrictProp.ulPropTag = PR_RECIPIENT_TYPE;
	sRestrictProp.Value.ul = MAPI_P1;

 
	hr = lpMessage->GetRecipientTable(0, &lpRTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpRTable->Restrict(&sRestriction, TBL_BATCH);
	if (hr != hrSuccess)
		goto exit;

	hr = lpRTable->QueryRows(-1, 0, &lpRowSet);
	if (hr != hrSuccess)
		goto exit;

	if (lpRowSet->cRows == 0) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	for (ULONG i = 0; i < lpRowSet->cRows; i++) {
		if (HrGetAddress(lpAdrBook, lpRowSet->aRow[i].lpProps, lpRowSet->aRow[i].cValues,
						 PR_ENTRYID, PR_DISPLAY_NAME, PR_ADDRTYPE, PR_EMAIL_ADDRESS,
						 strName, strType, strEmail) == hrSuccess) {

			recipients.appendMailbox(new vmime::mailbox(strEmail));
		}
	}
	
exit:
	if (lpRowSet)
		FreeProws(lpRowSet);

	if (lpRTable)
		lpRTable->Release();

	return hr;
}


HRESULT ECVMIMESender::sendMail(LPADRBOOK lpAdrBook, LPMESSAGE lpMessage, vmime::message *lpVMMessage, char *lpszRawSMTP) {
	HRESULT hr = hrSuccess;
	vmime::messaging::smtp::MAPISMTPTransport *lpMSTransport = NULL;
	vmime::mailbox expeditor;
	vmime::mailboxList recipients;

	if (!lpMessage || !lpVMMessage) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// recover from prev. error?
	if (lpTransport) {
		delete lpTransport;
		lpTransport = NULL;
	}
	if (lpSess) {
		delete lpSess;
		lpSess = NULL;
	}
	smtpresult = 0;
	error.clear();

	try {
		// Session initialization (global properties)
		lpSess = new vmime::messaging::session();

		// set the server address and port, plus type of service by use of url
		// and get our special mapismtp mailer
		vmime::utility::url url("mapismtp://"+smtphost);
		lpTransport = lpSess->getTransport(url, NULL);
		lpMSTransport = dynamic_cast<vmime::messaging::smtp::MAPISMTPTransport*>(lpTransport);

		// get expeditor for 'mail from:' smtp command
		try	{
			const vmime::mailboxField& from = dynamic_cast <const vmime::mailboxField&>
				(*lpVMMessage->getHeader()->findField(vmime::fields::FROM));
			expeditor = from.getValue();
		}
		catch (vmime::exceptions::no_such_field&) {
			throw vmime::exceptions::no_expeditor();
		}

		hr = HrMakeRecipientsList(lpAdrBook, lpMessage, lpVMMessage, recipients);
		if (hr != hrSuccess) {
			// no MAPI_P1 recipients, so use the normal recipients set in the email
			try {
				const vmime::addressListField& to = dynamic_cast <const vmime::addressListField&>
					(*lpVMMessage->getHeader()->findField(vmime::fields::TO));
				extractMailboxes(recipients, to.getValue());
			}
			catch (vmime::exceptions::no_such_field&) { }

			try	{
				const vmime::addressListField& cc = dynamic_cast <const vmime::addressListField&>
					(*lpVMMessage->getHeader()->findField(vmime::fields::CC));
				extractMailboxes(recipients, cc.getValue());
			}
			catch (vmime::exceptions::no_such_field&) { }

			try	{
				const vmime::addressListField& bcc = dynamic_cast <const vmime::addressListField&>
					(*lpVMMessage->getHeader()->findField(vmime::fields::BCC));
				extractMailboxes(recipients, bcc.getValue());
			}
			catch (vmime::exceptions::no_such_field&) { }
		}
		hr = hrSuccess;
		// now we either have the MAPI_P1 list, or the actual recipient list.

        // Remove BCC headers from the message we're about to send
        try {   
			vmime::headerField* bcc = lpVMMessage->getHeader()->findField(vmime::fields::BCC);
			lpVMMessage->getHeader()->removeField(bcc);
        }
        catch (vmime::exceptions::no_such_field&) { }

		// Generate the message, "stream" it and delegate the sending
		// to the generic send() function.
		std::ostringstream oss;
		vmime::utility::outputStreamAdapter ossAdapter(oss);

		if (lpszRawSMTP) {
			// get headers from normal vmime message
			lpVMMessage->getHeader()->generate(ossAdapter);

			// mark end of headers and append raw smtp data
			oss << lpszRawSMTP;
		} else {
			// complete vmime message
			lpVMMessage->generate(ossAdapter);
		}

		const string& str(oss.str());
		vmime::utility::inputStreamStringAdapter isAdapter(str);

		if (lpMSTransport)
			lpMSTransport->setLogger(lpLogger);

		// send the email already!
		try {
			lpTransport->connect();
		} catch (vmime::exceptions::connection_error &e) {
			// special error, smtp server not respoding, so try later again
			hr = MAPI_W_NO_SERVICE;
			goto exit;
		}
		lpTransport->send(expeditor, recipients, isAdapter, str.length(), NULL);
		lpTransport->disconnect();

		if (lpMSTransport && lpMSTransport->getRecipientErrorCount() > 0) {
			// hr value returned to spooler
			hr = MAPI_W_PARTIAL_COMPLETION;
			lstFailedRecipients = lpMSTransport->getRecipientErrorList();
			error = "Unable to reach all recipients";
			smtpresult = 250;	// ok value
			lpLogger->Log(EC_LOGLEVEL_ERROR, "SMTP Error: Not all recipients could be reached. User will be notified.");
		}
	}
	catch (vmime::exceptions::command_error& e) {
		if (lpMSTransport)
			lstFailedRecipients = lpMSTransport->getRecipientErrorList();
		lpLogger->Log(EC_LOGLEVEL_ERROR, "SMTP: %s Response: %s", e.what().c_str(), e.response().c_str());
		smtpresult = atoi(e.response().substr(0, e.response().find_first_of(" ")).c_str());
		error = e.response();
		// message should be cancelled, unsendable, test by smtp result code.
		hr = MAPI_W_CANCEL_MESSAGE;
	} 
	catch (vmime::exception& e) {
		// connection_greeting_error, ...?
		lpLogger->Log(EC_LOGLEVEL_ERROR, e.what());
		error = e.what();
		hr = MAPI_E_NETWORK_ERROR;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "%s",e.what());
		error = e.what();
		hr = MAPI_E_NETWORK_ERROR;
	}
	if (lpTransport) {
		delete lpTransport;
		lpTransport = NULL;
	}
	if (lpSess) {
		delete lpSess;
		lpSess = NULL;
	}

exit:
	return hr;
}


// ---
// c style util functions
// ---

// need this for attachments
// CHECK: is this correct? How about spaces? need to make longfi~.jpg kind of name?
// should always be 8.3 format?
void getShortFilename(std::string& strLongIn, std::string& strShortOut) {
	int	iNewNameLength = 8;
	int iDotPos = strLongIn.find_last_of(".");

	if (iDotPos > 0 && iDotPos < iNewNameLength)
		iNewNameLength = iDotPos;
	
	strShortOut.append(strLongIn.substr(0, iNewNameLength));
	// take iDotPos + 4 ( include the dot ) and not strlength to be sure to cutoff [.mpeg and .jpeg]'ish stuff..
	if (iDotPos > 0)
		strShortOut.append(strLongIn.substr(iDotPos, iDotPos+4));
}

// for debugging purposes
void componentToString(vmime::component& comp, std::string& strOut) {
	vmime::utility::outputStreamStringAdapter os(strOut);
	comp.generate(os);
}

// for debugging purposes
void printComponent(vmime::component& comp) {
	std::string out;

	vmime::utility::outputStreamStringAdapter os(out);
	comp.generate(os);
	std::cout << out.c_str() << std::endl;
}
