/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

// Damn windows header defines max which break C++ header files
#undef max

#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <sstream>

// vmime
#include "vmime/vmime.hpp"
#include "vmime/textPartFactory.hpp"
#include "vmime/platforms/posix/posixHandler.hpp"

// mapi
#include <mapi.h>
#include <mapiutil.h>
#include <mapiext.h>
#include <edkmdb.h>
#include <mapiguidext.h>

#include "tnef.h"

// inetmapi
#include "ECMapiUtils.h"
#include "MAPIToVMIME.h"
#include "VMIMEToMAPI.h"
#include "outputStreamMAPIAdapter.h"
#include "inputStreamMAPIAdapter.h"
#include "mapiAttachment.h"
#include "mapiTextPart.h"
#include "rtfutil.h"
#include "CommonUtil.h"
#include "stringutil.h"
#include "mapicontact.h"
#include "Util.h"
#include "ECLogger.h"
#include "codepage.h"
#include "ecversion.h"

// icalmapi
#include "MAPIToICal.h"

// others

#define BASETMPDIR "/tmp/"

#define SIZETOREAD 4096



/* These are the properties that we DON'T send through TNEF. Reasons can be:
 *
 * - Privacy (PR_MDB_PROVIDER, etc)
 * - Sent through rfc2822 (PR_SUBJECT, PR_BODY, etc)
 * - Not part of an outgoing message (PR_TRANSPORT_MESSAGE_HEADERS)
 */

SizedSPropTagArray(1, sptaRTFInclude) = {1, {PR_RTF_COMPRESSED} };

SizedSPropTagArray(53, sptaExclude) = {
    53, 	{
        PR_BODY,
        PR_RTF_COMPRESSED,
        PR_RTF_IN_SYNC,
        PR_ACCESS,
        PR_ACCESS_LEVEL,
        PR_CHANGE_KEY,
        PR_CLIENT_SUBMIT_TIME,
        PR_CREATION_TIME,
        PR_DELETE_AFTER_SUBMIT,
        PR_DISPLAY_BCC,
        PR_DISPLAY_CC,
        PR_DISPLAY_TO,
        PR_ENTRYID,
        PR_HASATTACH,
        PR_LAST_MODIFICATION_TIME,
        PR_MAPPING_SIGNATURE,
        PR_MDB_PROVIDER,
        PR_MESSAGE_DELIVERY_TIME,
        PR_MESSAGE_FLAGS,
        PR_MESSAGE_SIZE,
        PR_NORMALIZED_SUBJECT,
        PR_OBJECT_TYPE,
        PR_PARENT_ENTRYID,
        PR_PARENT_SOURCE_KEY,
        PR_PREDECESSOR_CHANGE_LIST,
        PR_RECORD_KEY,
        PR_RTF_SYNC_BODY_COUNT,
        PR_RTF_SYNC_BODY_CRC,
        PR_RTF_SYNC_BODY_TAG,
        PR_RTF_SYNC_PREFIX_COUNT,
        PR_RTF_SYNC_TRAILING_COUNT,
        PR_SEARCH_KEY,
        PR_SENDER_ADDRTYPE,
        PR_SENDER_EMAIL_ADDRESS,
        PR_SENDER_ENTRYID,
        PR_SENDER_NAME,
        PR_SENDER_SEARCH_KEY,
        PR_SENTMAIL_ENTRYID,
        PR_SENT_REPRESENTING_ADDRTYPE,
        PR_SENT_REPRESENTING_EMAIL_ADDRESS,
        PR_SENT_REPRESENTING_ENTRYID,
        PR_SENT_REPRESENTING_NAME,
        PR_SENT_REPRESENTING_SEARCH_KEY,
        PR_SOURCE_KEY,
        PR_STORE_ENTRYID,
        PR_STORE_RECORD_KEY,
        PR_STORE_SUPPORT_MASK,
        PR_SUBJECT,
        PR_SUBJECT_PREFIX,
        PR_SUBMIT_FLAGS,
        PR_TRANSPORT_MESSAGE_HEADERS,
		PR_LAST_MODIFIER_NAME, 		// IMPORTANT: exchange will drop the message if this is included
		PR_LAST_MODIFIER_ENTRYID,   // IMPORTANT: exchange will drop the message if this is included
    }
};

// These are named properties that are mapped to fixed values in Zarafa, so
// we can use the ID values instead of using GetIDsFromNames every times
#define PR_EC_USE_TNEF		PROP_TAG(PT_BOOLEAN, 0x8222)
#define PR_EC_SEND_AS_ICAL 	PROP_TAG(PT_BOOLEAN, 0x8000)
#define PR_EC_OUTLOOK_VERSION PROP_TAG(PT_STRING8, 0x81F4)

MAPIToVMIME::MAPIToVMIME()
{
	srand((unsigned)time(NULL));
	lpLogger = new ECLogger_File(EC_LOGLEVEL_ANY, 1, "-");	// TODO: remove fixed defaults
	delete_logger = true;
	m_lpAdrBook = NULL;

	// need to have a unique indentifier in the mediaType
	vmime::textPartFactory::getInstance()->registerType<vmime::mapiTextPart>(vmime::mediaType(vmime::mediaTypes::TEXT, "mapi"));
}

MAPIToVMIME::MAPIToVMIME(IMAPISession *lpSession, IAddrBook *lpAddrBook, ECLogger *newlogger, sending_options sopt)
{
	srand((unsigned)time(NULL));
	lpLogger = newlogger;
	if (!lpLogger) {
		lpLogger = new ECLogger_File(EC_LOGLEVEL_ANY, 1, "-");	// TODO: remove fixed defaults
		delete_logger = true;
	} else {
		delete_logger = false;
	}
	
	this->sopt = sopt;
	if (lpSession && !lpAddrBook) {
		lpSession->OpenAddressBook(0, NULL, AB_NO_DIALOG, &m_lpAdrBook);
		// ignore error
	} else {
		lpAddrBook->QueryInterface(IID_IAddrBook, (void**)&m_lpAdrBook);
	}

	m_lpSession = lpSession;

	// need to have a unique indentifier in the mediaType
	vmime::textPartFactory::getInstance()->registerType<vmime::mapiTextPart>(vmime::mediaType(vmime::mediaTypes::TEXT, "mapi"));
}

MAPIToVMIME::~MAPIToVMIME()
{
	if (delete_logger)
		delete lpLogger;

	if (m_lpAdrBook)
		m_lpAdrBook->Release();
}

HRESULT MAPIToVMIME::processRecipients(IMessage* lpMessage, vmime::messageBuilder* VMMessageBuilder, vmime::charset charset) {
	HRESULT			hr					= hrSuccess;
	vmime::mailbox	*lpVMMailbox		= NULL;
	LPMAPITABLE		lpRecipientTable	= NULL;
	LPSRowSet		pRows				= NULL;
	LPSPropValue	pPropRecipType		= NULL;
	LPSPropValue	pPropDispl			= NULL;
	LPSPropValue	pPropAType			= NULL;
	LPSPropValue	pPropEAddr			= NULL;
	bool			fToFound			= false;
	bool			hasRecips			= false;
	
	SizedSPropTagArray(6, sPropRecipColumns) = {6, { PR_ENTRYID, PR_EMAIL_ADDRESS, PR_DISPLAY_NAME, PR_RECIPIENT_TYPE, PR_SMTP_ADDRESS, PR_ADDRTYPE} };

	hr = lpMessage->GetRecipientTable(MAPI_DEFERRED_ERRORS, &lpRecipientTable);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to open recipient table. Error: 0x%08X", hr);
		goto exit;
	}

	hr = lpRecipientTable->SetColumns((LPSPropTagArray)&sPropRecipColumns, 0);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to set columns on recipient table. Error: 0x%08X", hr);
		goto exit;
	}

	hr = HrQueryAllRows(lpRecipientTable, NULL, NULL, NULL, 0, &pRows);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to read recipient table. Error: 0x%08X", hr);
		goto exit;
	}

	try {
		for (ULONG i = 0; i < pRows->cRows; i++) {
			pPropRecipType = PpropFindProp(pRows->aRow[i].lpProps, pRows->aRow[i].cValues, PR_RECIPIENT_TYPE);

			if(pPropRecipType == NULL) {
				// getMailBox properties
				pPropDispl = PpropFindProp(pRows->aRow[i].lpProps, pRows->aRow[i].cValues, PR_DISPLAY_NAME);
				pPropAType = PpropFindProp(pRows->aRow[i].lpProps, pRows->aRow[i].cValues, PR_ADDRTYPE);
				pPropEAddr = PpropFindProp(pRows->aRow[i].lpProps, pRows->aRow[i].cValues, PR_EMAIL_ADDRESS);
				lpLogger->Log(EC_LOGLEVEL_FATAL, "No recipient type set for recipient. DisplayName: %s, AddrType: %s, Email: %s",
							  pPropDispl?pPropDispl->Value.lpszA:"(none)",
							  pPropAType?pPropAType->Value.lpszA:"(none)",
							  pPropEAddr?pPropEAddr->Value.lpszA:"(none)");
				continue;
			}

			// ATTN: lpVMMailbox will be added to the VMMessageBuilder. The VMMessageBuilder will delete() it
			hr = getMailBox(&pRows->aRow[i], &lpVMMailbox, MAPI_CHARSET);
			if (hr != hrSuccess)
				goto exit;			// Logging done in getMailBox

			switch (pPropRecipType->Value.ul) {
			case MAPI_TO:
				VMMessageBuilder->getRecipients().appendAddress(lpVMMailbox);
				fToFound = true;
				hasRecips = true;
				break;
			case MAPI_CC:
				VMMessageBuilder->getCopyRecipients().appendAddress(lpVMMailbox);
				hasRecips = true;
				break;
			case MAPI_BCC:
				VMMessageBuilder->getBlindCopyRecipients().appendAddress(lpVMMailbox);
				hasRecips = true;
				break;
			default:
				// we're not giving the object to vmime, so remove it ourselves
				delete lpVMMailbox;
				break;
			}
		}
		
		if(!fToFound) {
			if (!hasRecips && sopt.no_recipients_workaround == false) {
				// spooler will exit here, gateway will continue with the workaround
				lpLogger->Log(EC_LOGLEVEL_FATAL, "No valid recipients found.");
				hr = MAPI_E_NOT_FOUND;
				goto exit;
			}

			// No recipients were in the 'To' .. add 'undisclosed recipients'
			vmime::mailboxGroup *undisclosed = new vmime::mailboxGroup(vmime::text("undisclosed-recipients"));
			VMMessageBuilder->getRecipients().appendAddress(undisclosed);
		}
			
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on recipients: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on recipients");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (lpRecipientTable)
		lpRecipientTable->Release();

	if (pRows)
		FreeProws(pRows);

	return hr;
}

HRESULT MAPIToVMIME::handleSingleAttachment(IMessage* lpMessage, LPSRow lpRow, vmime::messageBuilder* lpVMMessageBuilder) {
	HRESULT			hr					= hrSuccess;
	IStream*		lpStream			= NULL; 
	LPATTACH		lpAttach			= NULL;
	LPSPropValue	pPropAttachNum		= NULL;
	LPSPropValue	pPropAttachType		= NULL;
	LPSPropValue	lpContentId 		= NULL;
	LPSPropValue	lpContentLocation	= NULL;
	LPSPropValue	lpHidden			= NULL;
	LPSPropValue	lpFilename			= NULL;
	ULONG			ulAttachmentNum		= 0;
	IMessage*		lpAttachedMessage	= NULL;
	inputStreamMAPIAdapter *inputDataStream = NULL;
	mapiAttachment *lpMapiAttach 		= NULL;
	vmime::defaultAttachment* lpVMDefAtt = NULL;
	std::string		strContentId;
	std::string		strContentLocation;
	bool			bHidden = false;
	sending_options sopt_keep;
	LPSPropValue	lpAMClass = NULL;
	LPSPropValue	lpAMAttach = NULL;
	LPSPropValue	lpMIMETag = NULL;
	char *			szFilename = NULL;  // just a reference, don't free
	vmime::mediaType szMIMEType;
	MapiToICal		*mapiical = NULL;

	std::string		strBoundary;
	bool			bSendBinary = true;
	
	pPropAttachNum = PpropFindProp(lpRow->lpProps, lpRow->cValues, PR_ATTACH_NUM);
	if (pPropAttachNum == NULL) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Attachment in table not correct, no attachment number present.");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	ulAttachmentNum = pPropAttachNum->Value.ul;

	// check PR_ATTACH_METHOD to determine Attachment or email
	pPropAttachType	= PpropFindProp(lpRow->lpProps, lpRow->cValues, PR_ATTACH_METHOD); 
	if (pPropAttachType == NULL) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Attachment in table not correct, no attachment method present for attachment %d.", ulAttachmentNum);
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	if (pPropAttachType->Value.ul == ATTACH_EMBEDDED_MSG) {
		vmime::message* lpVMNewMess = NULL;
		std::string strBuff;
		vmime::utility::outputStreamStringAdapter mout(strBuff);

		hr = lpMessage->OpenAttach(ulAttachmentNum, NULL, MAPI_BEST_ACCESS, &lpAttach);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open message attachment %d. Error: 0x%08X", ulAttachmentNum, hr);
			goto exit;
		}

		hr = lpAttach->OpenProperty(PR_ATTACH_DATA_OBJ, &IID_IMessage, 0, MAPI_DEFERRED_ERRORS, (LPUNKNOWN *)&lpAttachedMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open data of message attachment %d. Error: 0x%08X", ulAttachmentNum, hr);
			goto exit;
		}

		// Check whether we're sending a calender object
		// if so, we do not need to create a message-in-message object, but just attach the ics file.
		hr = HrGetOneProp(lpAttachedMessage, PR_MESSAGE_CLASS, &lpAMClass);
		if (hr == hrSuccess && strncmp(lpAMClass->Value.lpszA, "IPM.Note", 8) != 0 &&
			// currently no task support for ical
			strncmp(lpAMClass->Value.lpszA, "IPM.Task", 8) != 0) {
			// attach ical without submessage
			// still tnef when attachments are in the submessage
			hr = HrGetOneProp(lpAttachedMessage, PR_HASATTACH, &lpAMAttach);
			if (hr == hrSuccess && lpAMAttach->Value.b == false) {
				string ical;
				mapiAttachment *lpVMAttach = NULL;

				CreateMapiToICal(m_lpAdrBook, "utf-8", &mapiical); // charset...

				hr = mapiical->AddMessage(lpAttachedMessage, 0);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create ical object");
				} else {
					hr = mapiical->Finalize(0, NULL, &ical);
					if (hr != hrSuccess) {
						lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create ical object");
					} else {
						lpVMAttach = new mapiAttachment(vmime::stringContentHandler(ical),
														vmime::encoding("8bit"),
														vmime::mediaType("application/ics"),
														"", vmime::word("calendar.ics"));
						// ics files are in utf-8
						lpVMAttach->addCharset(vmime::charset(vmime::charsets::UTF_8));
						lpVMMessageBuilder->appendAttachment(lpVMAttach);
						// and we're done
						goto exit;
					}
				}
			}
		}

		// sub objects do not necessarily need to have valid recipients, so disable the test
		sopt_keep = sopt;
		sopt.no_recipients_workaround = true;
		sopt.msg_in_msg = true;
		sopt.add_received_date = false;
		
		if(sopt.alternate_boundary) {
		    strBoundary = sopt.alternate_boundary;
		    strBoundary += "_sub_";
		    strBoundary += stringify(ulAttachmentNum);
		    sopt.alternate_boundary = (char *)strBoundary.c_str();
		}

		// recursive processing of embedded message as a new e-mail
		hr = convertMAPIToVMIME(lpAttachedMessage, &lpVMNewMess);
		if (hr != hrSuccess) {
			// Logging has been done by convertMAPIToVMIME()
			goto exit;
		}
		sopt = sopt_keep;

		lpVMNewMess->generate(mout);
		lpVMDefAtt = new vmime::defaultAttachment(vmime::stringContentHandler(strBuff),
												  vmime::encoding("8bit"),
												  vmime::mediaType("message/rfc822"));
		// append it (copies pointer, not data)
		lpVMMessageBuilder->appendAttachment(lpVMDefAtt);

		delete lpVMNewMess;

	} else if (pPropAttachType->Value.ul == ATTACH_BY_VALUE) {
		hr = lpMessage->OpenAttach(ulAttachmentNum, NULL, MAPI_BEST_ACCESS, &lpAttach);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open attachment %d. Error: 0x%08X", ulAttachmentNum, hr);
			goto exit;
		}
		
		if (HrGetOneProp(lpAttach, PR_ATTACH_CONTENT_ID, &lpContentId) == hrSuccess)
			strContentId = lpContentId->Value.lpszA;

		if (HrGetOneProp(lpAttach, PR_ATTACH_CONTENT_LOCATION, &lpContentLocation) == hrSuccess)
			strContentLocation = lpContentLocation->Value.lpszA;
			
		if (HrGetOneProp(lpAttach, PR_ATTACHMENT_HIDDEN, &lpHidden) == hrSuccess)
			bHidden = lpHidden->Value.b;

        // Optimize: if we're only outputting headers, we needn't bother with attachment data
        if(sopt.headers_only)
            CreateStreamOnHGlobal(NULL, true, &lpStream);
        else {
            hr = lpAttach->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, 0, MAPI_DEFERRED_ERRORS, (LPUNKNOWN *)&lpStream);
            if (hr != hrSuccess) {
                lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open data of attachment %d. Error: 0x%08X", ulAttachmentNum, hr);
                goto exit;
            }
        }
        	
		try {
			inputDataStream = new inputStreamMAPIAdapter(lpStream);	// add ref to lpStream
			
			
			// Set filename
			szFilename = "data.bin";
			
			if (HrGetOneProp(lpAttach, PR_ATTACH_LONG_FILENAME, &lpFilename) == hrSuccess) {
                szFilename = lpFilename->Value.lpszA;
            } else {
				if (HrGetOneProp(lpAttach, PR_ATTACH_FILENAME, &lpFilename) == hrSuccess) {
                    szFilename = lpFilename->Value.lpszA;
				}
            }
            
            // Set MIME type
            parseMimeTypeFromFilename(szFilename, &szMIMEType, &bSendBinary);
            
            if (HrGetOneProp(lpAttach, PR_ATTACH_MIME_TAG, &lpMIMETag) == hrSuccess) {
                szMIMEType = lpMIMETag->Value.lpszA;
            }

			// inline attachments: Only make attachments inline if they are hidden and have a content-id or content-location.
			if ((!strContentId.empty() || !strContentLocation.empty()) && bHidden &&
				lpVMMessageBuilder->getTextPart()->getType() == vmime::mediaType(vmime::mediaTypes::TEXT, vmime::mediaTypes::TEXT_HTML))
			{
				// add inline object to html part
				vmime::mapiTextPart& textPart = dynamic_cast<vmime::mapiTextPart&>(*lpVMMessageBuilder->getTextPart());
				textPart.addObject(vmime::streamContentHandler(inputDataStream, 0, true), vmime::encoding("base64"), szMIMEType, strContentId, szFilename, strContentLocation);
			} else {
				// add as normal attachment
				lpMapiAttach = new mapiAttachment(vmime::streamContentHandler(inputDataStream, 0, true),
												  vmime::encoding("base64"), szMIMEType,
												  strContentId, vmime::word(szFilename, MAPI_CHARSET));
				// add to message (copies pointer, not data)
				lpVMMessageBuilder->appendAttachment(lpMapiAttach);
			}

		}
		catch (vmime::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (std::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on attachment: %s", e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (...) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on attachment");
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
	} else if (pPropAttachType->Value.ul == ATTACH_OLE) {
	    // Ignore ATTACH_OLE attachments, they are handled in handleTNEF()
	} else {
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Attachment %d contains invalid method %d. Not adding attachment to e-mail.", ulAttachmentNum, pPropAttachType->Value.ul);
	}

exit:
	// ATTN: lpMapiAttach are linked in the VMMessageBuilder. The VMMessageBuilder will delete() it.
	if (mapiical)
		delete mapiical;

	if (lpMIMETag)
        MAPIFreeBuffer(lpMIMETag);

	if (lpAMClass)
		MAPIFreeBuffer(lpAMClass);

	if (lpAMAttach)
		MAPIFreeBuffer(lpAMAttach);

	if (lpFilename) 
		MAPIFreeBuffer(lpFilename);

	if (lpContentId) 
		MAPIFreeBuffer(lpContentId);
		
	if (lpHidden)
		MAPIFreeBuffer(lpHidden);

	if (lpStream)
		lpStream->Release();

	if (lpAttach)
		lpAttach->Release();

	if (lpAttachedMessage)
		lpAttachedMessage->Release();

	return hr;
}

// TODO: FIXME with something nifty with /etc/mime.types
HRESULT MAPIToVMIME::parseMimeTypeFromFilename(std::string strFilename, vmime::mediaType *lpMT, bool *lpbSendBinary) {
	std::string strExt;
	std::string strMedType;
	bool bSendBinary = true;

	// to lowercase
	transform(strFilename.begin(), strFilename.end(), strFilename.begin(), ::tolower);
	strExt.assign(strFilename.substr(strFilename.find_last_of(".")+1, strFilename.length()));

	// application
	if (strExt == "bin" || strExt == "exe") {
		strMedType = "application/octet-stream";
	} else if (strExt == "ai" || strExt == "eps" || strExt == "ps") {
		strMedType = "application/postscript";
	} else if (strExt == "pdf") {
		strMedType = "application/pdf";
	} else if (strExt == "rtf") {
		strMedType = "application/rtf";
	} else if (strExt == "zip") {
		strMedType = "application/zip";
	} else if (strExt == "doc" || strExt == "dot") {
		strMedType = "application/msword";
	} else if (strExt == "mdb") {
		strMedType = "application/x-msaccess";
	} else if (strExt == "xla" || strExt == "xls" || strExt == "xlt" || strExt == "xlw") {
		strMedType = "application/vnd.ms-excel";
	} else if (strExt == "pot" || strExt == "ppt" || strExt == "pps") {
		strMedType = "application/vnd.ms-powerpoint";
	} else if (strExt == "mpp") {
		strMedType = "application/vnd.ms-project";
	} else if (strExt == "edi") {
		strMedType = "application/edifact";
		bSendBinary = false;
	}
	
	// audio

	else if (strExt == "ua") {
		strMedType = "audio/basic";
	} else if (strExt == "wav") {
		strMedType = "audio/x-wav";
	} else if (strExt == "mid") {
		strMedType = "audio/x-midi";
	}
	
	// image

	else if (strExt == "gif") {
		strMedType = "image/gif";
	} else if (strExt == "jpg" || strExt == "jpe" || strExt == "jpeg") {
		strMedType = "image/jpeg";
	} else if (strExt == "png") {
		strMedType = "image/png";
	} else if (strExt == "bmp") {
		strMedType = "image/x-ms-bmp";
	} else if (strExt == "tiff") {
		strMedType = "image/tiff";
	} else if (strExt == "xbm") {
		strMedType = "image/xbm";
	}
	
	// video

	else if (strExt == "mpg" || strExt == "mpe" || strExt == "mpeg") {
		strMedType = "video/mpeg";
	} else if (strExt == "qt" || strExt == "mov") {
		strMedType = "video/quicktime";
	} else if (strExt == "avi") {
		strMedType = "video/x-msvideo";
	}


	else {
		strMedType = "application/octet-stream";
	}

	*lpMT = vmime::mediaType(strMedType);
	*lpbSendBinary = bSendBinary;

	return hrSuccess;
}

HRESULT MAPIToVMIME::handleAttachments(IMessage* lpMessage, vmime::messageBuilder *lpVMMessageBuilder) {
	HRESULT		hr					= hrSuccess;
	LPSRowSet	pRows				= NULL;
	LPMAPITABLE	lpAttachmentTable	= NULL;

	// get attachment table
	hr = lpMessage->GetAttachmentTable(0, &lpAttachmentTable);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to open attachment table. Error: 0x%08X", hr);
		goto exit;
	}

	hr = HrQueryAllRows(lpAttachmentTable, NULL, NULL, NULL, 0, &pRows);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to fetch rows of attachment table. Error: 0x%08X", hr);
		goto exit;
	}
	
	// iterate through attachments and hand them off...
	for (ULONG i=0; i < pRows->cRows; i++) {
		hr = handleSingleAttachment(lpMessage, &pRows->aRow[i], lpVMMessageBuilder);
		// ignoring error... if one attachment fails we still want the others .. logging was done in handleSingleAttachment()
		hr = hrSuccess;
	} 

exit:
	if(pRows)
		FreeProws(pRows);

	if (lpAttachmentTable)
		lpAttachmentTable->Release();

	return hr;
}

HRESULT MAPIToVMIME::setBoundaries(vmime::header *lpVMHeader, vmime::body *lpVMBody, const std::string &boundary) {
    int i = 0;
    
    if(lpVMHeader->ContentType().getValue().getType() == "multipart") {
        // This is a multipart, so set the boundary for this part
        lpVMHeader->ContentType().setBoundary(boundary);
        
        // Set boundaries on children also
        for(i=0;i<lpVMBody->getPartCount();i++) {
            std::ostringstream os;

            vmime::bodyPart* VMBodyPart = lpVMBody->getPartAt(i);

            os << boundary << "_" << i;
            
            setBoundaries(VMBodyPart->getHeader(), VMBodyPart->getBody(), os.str());
        }
    }
    
    return hrSuccess;
}

HRESULT MAPIToVMIME::BuildNoteMessage(IMessage *lpMessage, bool bSkipContent, vmime::charset& charset, vmime::message **lppVMMessage)
{
	HRESULT					hr					= hrSuccess;
	LPSPropValue			lpDeliveryDate		= NULL;
	LPSPropValue			lpTransportHeaders	= NULL;
	vmime::messageBuilder*	lpVMMessageBuilder	= NULL;
	vmime::message*			lpVMMessage			= NULL;	

	// construct the message..
	try {
		lpVMMessageBuilder = new vmime::messageBuilder();
		// messageBuilder is body and simple headers only (to/from/cc/subject/...)
		hr = fillVMIMEMail(lpMessage, bSkipContent, lpVMMessageBuilder, charset);
		if (hr != hrSuccess)
			goto exit;			// Logging has been done in fillVMIMEMail()
		
		lpVMMessage = lpVMMessageBuilder->construct();
		// message builder has set Date header to now(). this will be overwritten.

		vmime::header* lpVMHeader = lpVMMessage->getHeader();

		hr = handleExtraHeaders(lpMessage, charset, lpVMHeader);
		if (hr!= hrSuccess)
			goto exit;

		hr = handleSenderInfo(lpMessage, lpVMHeader);
		if (hr!= hrSuccess)
			goto exit;
		
		// add reply to email, ignore errors
		hr = handleReplyTo(lpMessage, lpVMHeader);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to set reply-to address");
			hr = hrSuccess;
		}

		// Set consistent boundary (optional)
		if(sopt.alternate_boundary) {
		    setBoundaries(lpVMMessage->getHeader(), lpVMMessage->getBody(), sopt.alternate_boundary);
		}

		HrGetOneProp(lpMessage, PR_MESSAGE_DELIVERY_TIME, &lpDeliveryDate);

		// If we're sending a msg-in-msg, use the original date of that message
		if (sopt.msg_in_msg && lpDeliveryDate) {
			lpVMHeader->Date().setValue(FiletimeTovmimeDatetime(lpDeliveryDate->Value.ft));
		}
		
		// Generate 'Received:' headers if available (basically a copy of the headers in
		// PR_TRANSPORT_MESSAGE_HEADERS		
		HrGetOneProp(lpMessage, PR_TRANSPORT_MESSAGE_HEADERS, &lpTransportHeaders);

		if(lpTransportHeaders) {
			try {
				int j=0;
				vmime::header headers;
				std::string strHeaders = lpTransportHeaders->Value.lpszA;
				strHeaders += "\r\n\r\n";
				headers.parse(strHeaders);
				
				for(int i=0; i<headers.getFieldCount(); i++) {
					vmime::headerField *field = headers.getFieldAt(i);
					
					if(stricmp(field->getName().c_str(), vmime::fields::RECEIVED) == 0 ||
					   stricmp(field->getName().c_str(), vmime::fields::RETURN_PATH) == 0)
					{
						// Insert in same order at start of headers
						lpVMHeader->insertFieldBefore(j++, field->clone());
					}
				}
			} catch(...) { 
				// If we can't parse, just ignore
			}
		}

		// POP3 wants the delivery date as received header to correctly show in outlook
		if (sopt.add_received_date && lpDeliveryDate) {
			vmime::headerFieldFactory* hff = vmime::headerFieldFactory::getInstance();
			vmime::relayField *rf = dynamic_cast<vmime::relayField*>(hff->create(vmime::fields::RECEIVED));
			rf->getValue().setDate(FiletimeTovmimeDatetime(lpDeliveryDate->Value.ft));
			// set received header at the start
			lpVMHeader->insertFieldBefore(0, rf);
		}

		// no hr checking
		handleXHeaders(lpMessage, lpVMHeader);
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
		hr = MAPI_E_CALL_FAILED; // set real error
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on note message: %s", e.what());
		hr = MAPI_E_CALL_FAILED; // set real error
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception on note message");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	*lppVMMessage = lpVMMessage;

exit:
	if (hr != hrSuccess && lpVMMessage)
		delete lpVMMessage;

	if (lpTransportHeaders)
		MAPIFreeBuffer(lpTransportHeaders);
		
	if (lpDeliveryDate)
		MAPIFreeBuffer(lpDeliveryDate);

	if (lpVMMessageBuilder)
		delete(lpVMMessageBuilder);

	return hr;
}

HRESULT MAPIToVMIME::BuildMDNMessage(IMessage *lpMessage, vmime::charset& charset, vmime::message **lppVMMessage)
{
	HRESULT				hr = hrSuccess;
	vmime::message*		lpVMMessage = NULL;
	IStream*			lpBodyStream = NULL;
	LPSPropValue		lpiNetMsgId = NULL;
	LPSPropValue		lpMsgClass = NULL;
	LPSPropValue		lpSubject = NULL;
	LPMAPITABLE			lpRecipientTable	= NULL;
	LPSRowSet			pRows				= NULL;
	
	vmime::mailbox		expeditor; // From
	string				strMDNText;
	vmime::disposition	dispo;
	string				reportingUA; //empty
	std::vector<string>	reportingUAProducts; //empty
	vmime::message		msgOriginal;
	vmime::mailbox*		lpRecipientbox = NULL;
	string				strActionMode;
	std::string			strOut,	strOutConverted;

	// sender information
	std::string strEmailAdd, strName, strType;

	if (lpMessage == NULL || m_lpAdrBook == NULL || lppVMMessage == NULL) {
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}

	try {
		// Create original vmime message
		if (HrGetOneProp(lpMessage, PR_INTERNET_MESSAGE_ID, &lpiNetMsgId) == hrSuccess)
		{
			msgOriginal.getHeader()->OriginalMessageId().setValue(lpiNetMsgId->Value.lpszA);
			msgOriginal.getHeader()->MessageId().setValue(lpiNetMsgId->Value.lpszA);
		}
		
		// Create Recipient
		hr = lpMessage->GetRecipientTable(MAPI_DEFERRED_ERRORS, &lpRecipientTable);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to open MDN recipient table. Error: 0x%08X", hr);
			goto exit;
		}

		hr = HrQueryAllRows(lpRecipientTable, NULL, NULL, NULL, 0, &pRows);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to read MDN recipient table. Error: 0x%08X", hr);
			goto exit;
		}

		if (pRows->cRows == 0) {
			if (sopt.no_recipients_workaround == false) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "No MDN recipient found");
				hr = MAPI_E_NOT_FOUND;
				goto exit;
			} else {
				// no recipient, but need to continue ... is this correct??
				lpRecipientbox = new vmime::mailbox(string("undisclosed-recipients"));
			}
		}else {

			hr = getMailBox(&pRows->aRow[0], &lpRecipientbox, MAPI_CHARSET);
			if (hr != hrSuccess)
				goto exit;			// Logging done in getMailBox
		}

		vmime::mdn::sendableMDNInfos mdnInfos(&msgOriginal, *lpRecipientbox);
		
		//FIXME: Get the ActionMode and sending mode from the property 0x0081001E.
		//		And the type in property 0x0080001E.

		// Vmime is broken, The RFC 3798 says: "The action type can be 'manual-action' or 'automatic-action'"
		// but VMime can set only 'manual' or 'automatic' so should be add the string '-action'
		strActionMode = vmime::dispositionActionModes::MANUAL;
		strActionMode+= "-action";

		dispo.setActionMode(strActionMode);
		dispo.setSendingMode(vmime::dispositionSendingModes::SENT_MANUALLY);

		if (HrGetOneProp(lpMessage, PR_MESSAGE_CLASS, &lpMsgClass) != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "MDN message has no class.");
			hr = MAPI_E_CORRUPT_DATA;
			goto exit;
		}

		if(stricmp(lpMsgClass->Value.lpszA, "REPORT.IPM.Note.IPNNRN") == 0)
			dispo.setType(vmime::dispositionTypes::DELETED);
		else // if(stricmp(lpMsgClass->Value.lpszA, "REPORT.IPM.Note.IPNRN") == 0)
			dispo.setType(vmime::dispositionTypes::DISPLAYED);		
	
		strMDNText = "";// Default Empty body

		hr = lpMessage->OpenProperty(PR_BODY, &IID_IStream, 0, 0, (IUnknown**)&lpBodyStream);
		if(hr == hrSuccess)
		{

			hr = readStream(lpBodyStream, strMDNText);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to read MDN message body.");
				goto exit;
			}

		}

		// Store owner, actual sender
		hr = HrGetAddress(m_lpAdrBook, lpMessage, PR_SENDER_ENTRYID, PR_SENDER_NAME, PR_SENDER_ADDRTYPE, PR_SENDER_EMAIL_ADDRESS, strName, strType, strEmailAdd);
		if(hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get MDN sender information. Error: 0x%08X", hr);
		    goto exit;
		}

		expeditor.setEmail(strEmailAdd);
		if(!strName.empty())
			expeditor.setName(vmime::text(vmime::word(strName, MAPI_CHARSET)));
		else
			expeditor.setName(vmime::text(vmime::word(strEmailAdd, MAPI_CHARSET)));

		//Create MDN message
		lpVMMessage = vmime::mdn::MDNHelper::buildMDN(mdnInfos, strMDNText, charset, expeditor, dispo, reportingUA, reportingUAProducts);
		if (!lpVMMessage) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to build MDN message");
			hr = MAPI_E_CORRUPT_DATA;
			goto exit;
		}

		// vmime is bad, so write subject
		if (HrGetOneProp(lpMessage, PR_SUBJECT, &lpSubject) == hrSuccess) {
			removeEnters(lpSubject->Value.lpszA);
			
			strOut = lpSubject->Value.lpszA;
			vmime::charset::convert(strOut, strOutConverted, MAPI_CHARSET, charset);

			lpVMMessage->getHeader()->Subject().getValue() = vmime::text(vmime::word(strOutConverted, charset));

		}

	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on MDN message: %s", e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception on MDN message");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	*lppVMMessage = lpVMMessage;

exit:
	if(lpBodyStream)
		lpBodyStream->Release();

	if(lpRecipientbox)
		delete lpRecipientbox;

	if(lpiNetMsgId)
		MAPIFreeBuffer(lpiNetMsgId);
	
	if(lpMsgClass)
		MAPIFreeBuffer(lpMsgClass);

	if(lpRecipientTable)
		lpRecipientTable->Release();

	if(lpSubject)
		MAPIFreeBuffer(lpSubject);

	if(pRows)
		FreeProws(pRows);

	return hr;
}

HRESULT MAPIToVMIME::convertMAPIToVMIME(IMessage *lpMessage,  vmime::message **lppVMMessage, char **lppszRawSMTP) {
	HRESULT					hr					= hrSuccess;
	LPSPropValue			lpInternetCPID		= NULL;
	LPSPropValue			lpMsgClass			= NULL;
	vmime::message*			lpVMMessage			= NULL;
	vmime::charset			charset;
	char *					lpszCharset			= NULL;
	char *					lpszRawSMTP			= NULL;
	LPMAPITABLE				lpAttachmentTable	= NULL;
	LPSRowSet				lpRows				= NULL;
	LPSPropValue			lpPropAttach		= NULL;
	LPATTACH				lpAttach			= NULL;
	LPSTREAM				lpStream			= NULL;
	STATSTG					sStreamStat;
	SizedSPropTagArray(2, sPropAttachColumns) = {2, { PR_ATTACH_NUM, PR_ATTACH_MIME_TAG} };

	// Get the outgoing charset we want to be using
	if (HrGetOneProp(lpMessage, PR_INTERNET_CPID, &lpInternetCPID) == hrSuccess &&
		HrGetCharsetByCP(lpInternetCPID->Value.ul, &lpszCharset) == hrSuccess) {
		charset = vmime::charset(lpszCharset);
	} else {
		// default to windows-1252 if not set
		charset = vmime::charset("windows-1252");
	}

	if (HrGetOneProp(lpMessage, PR_MESSAGE_CLASS, &lpMsgClass) != hrSuccess) {
		MAPIAllocateBuffer(sizeof(SPropValue), (void**)&lpMsgClass);
		lpMsgClass->ulPropTag = PR_MESSAGE_CLASS;
		lpMsgClass->Value.lpszA = "IPM.Note";
	}

	if ((stricmp(lpMsgClass->Value.lpszA, "REPORT.IPM.Note.IPNNRN") == 0 ||
		stricmp(lpMsgClass->Value.lpszA, "REPORT.IPM.Note.IPNRN") == 0)
		)
	{
		// Create a read receipt message
		hr = BuildMDNMessage(lpMessage, charset, &lpVMMessage);
		if(hr != hrSuccess)
			goto exit;		
	} else if ((lppszRawSMTP && stricmp(lpMsgClass->Value.lpszA, "IPM.Note.SMIME.MultiPartSigned") == 0) ||
			   (stricmp(lpMsgClass->Value.lpszA, "IPM.Note.SMIME") == 0)){
		// - find attachment, and convert to char, and place in lpszRawSMTP
		// - normal convert the message, but only from/to headers and such .. nothing else

		hr = lpMessage->GetAttachmentTable(0, &lpAttachmentTable);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not get attachment table of signed attachment. Error: 0x%08X", hr);
			goto exit;
		}

		// set columns to get pr attach mime tag and pr attach num only.
		hr = lpAttachmentTable->SetColumns((LPSPropTagArray)&sPropAttachColumns, 0);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could set table contents of attachment table of signed attachment. Error: 0x%08X", hr);
			goto exit;
		}

		hr = HrQueryAllRows(lpAttachmentTable, NULL, NULL, NULL, 0, &lpRows);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not get table contents of attachment table of signed attachment. Error: 0x%08X", hr);
			goto exit;
		}

		if (lpRows->cRows != 1)
			goto normal;

		lpPropAttach = PpropFindProp(lpRows->aRow[0].lpProps, lpRows->aRow[0].cValues, PR_ATTACH_MIME_TAG);
		if (!lpPropAttach)
			goto normal;

		lpPropAttach = PpropFindProp(lpRows->aRow[0].lpProps, lpRows->aRow[0].cValues, PR_ATTACH_NUM);
		if (!lpPropAttach)
			goto normal;

		hr = lpMessage->OpenAttach(lpPropAttach->Value.ul, NULL, MAPI_BEST_ACCESS, &lpAttach);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open signed attachment. Error: 0x%08X", hr);
			goto exit;
		}

		hr = lpAttach->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, 0, MAPI_DEFERRED_ERRORS, (LPUNKNOWN *)&lpStream);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open data of signed attachment. Error: 0x%08X", hr);
			goto exit;
		}

		hr = lpStream->Stat(&sStreamStat, 0);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not find size of signed attachment. Error: 0x%08X", hr);
			goto exit;
		}

		lpszRawSMTP = new char[(ULONG)sStreamStat.cbSize.QuadPart+1];

		hr = lpStream->Read(lpszRawSMTP, (ULONG)sStreamStat.cbSize.QuadPart, NULL);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not read data of signed attachment. Error: 0x%08X", hr);
			goto exit;
		}
		lpszRawSMTP[(ULONG)sStreamStat.cbSize.QuadPart] = '\0';

		// build the message, but without the bodies and attachments
		hr = BuildNoteMessage(lpMessage, true, charset, &lpVMMessage);
		if (hr != hrSuccess)
			goto exit;

		// remove excess headers
		lpVMMessage->getHeader()->removeField(lpVMMessage->getHeader()->findField(vmime::fields::CONTENT_TYPE));
		lpVMMessage->getHeader()->removeField(lpVMMessage->getHeader()->findField(vmime::fields::CONTENT_TRANSFER_ENCODING));

		if (stricmp(lpMsgClass->Value.lpszA, "IPM.Note.SMIME") != 0) {
			// signed mail, return actual smpt data
			lpVMMessage->getHeader()->removeField(lpVMMessage->getHeader()->findField(vmime::fields::MIME_VERSION));
			*lppszRawSMTP = lpszRawSMTP;
		} else {
			// encoded mail, set data as only mail body
			LPMAPINAMEID lpNameID = NULL;
			unsigned int vlen = 0;
			LPSPropTagArray lpPropTags = NULL;
			LPSPropValue lpPropContentType = NULL;
			char *lpszContentType = NULL;

			hr = MAPIAllocateBuffer(sizeof(MAPINAMEID), (void**)&lpNameID);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, "Not enough memory. Error: 0x%08X", hr);
				goto exit;
			}

			lpNameID->lpguid = (GUID*)&PS_INTERNET_HEADERS;
			lpNameID->ulKind = MNID_STRING;

			// calculate from normal string the new unicode string len
			vlen = mbstounicode(NULL, "Content-Type", 0) + 1;

			hr = MAPIAllocateMore(vlen * sizeof(WCHAR), lpNameID, (void**)&lpNameID->Kind.lpwstrName);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, "Not enough memory. Error: 0x%08X", hr);
				goto exit;
			}

			// Convert from string to unicode
			mbstounicode(lpNameID->Kind.lpwstrName, "Content-Type", vlen);

			hr = lpMessage->GetIDsFromNames(1, &lpNameID, MAPI_CREATE, &lpPropTags);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to read encrypted mail properties. Error: 0x%08X", hr);
				goto exit;
			}

			// Free Memory
			if(lpNameID) {
				MAPIFreeBuffer(lpNameID);
				lpNameID = NULL;
			}

			if (HrGetOneProp(lpMessage, CHANGE_PROP_TYPE(lpPropTags->aulPropTag[0], PT_STRING8), &lpPropContentType) == hrSuccess) {
				lpszContentType = lpPropContentType->Value.lpszA;
			} else {
				// default, or exit?
				lpszContentType = "application/x-pkcs7-mime;smime-type=enveloped-data;name=smime.p7m";
			}

			lpVMMessage->getHeader()->ContentType().parse(lpszContentType);

			if (lpPropContentType)
				MAPIFreeBuffer(lpPropContentType);

			// copy via string so we can set the size of the string since it's binary
			vmime::string inString(lpszRawSMTP, (size_t)sStreamStat.cbSize.QuadPart);
			lpVMMessage->getBody()->setContents(vmime::stringContentHandler(inString));

			// vmime now encodes the body too, so I don't have to
			lpVMMessage->getHeader()->appendField(vmime::headerFieldFactory::getInstance()->create(vmime::fields::CONTENT_TRANSFER_ENCODING, "base64"));
		}

	} else {
normal:
		// Create default
		hr = BuildNoteMessage(lpMessage, false, charset, &lpVMMessage);
		if (hr != hrSuccess)
			goto exit;
	}

	*lppVMMessage = lpVMMessage;

exit:
	if (lpStream)
		lpStream->Release();

	if (lpAttach)
		lpAttach->Release();

	if (lpRows)
		FreeProws(lpRows);

	if (lpAttachmentTable)
		lpAttachmentTable->Release();

	if (lpInternetCPID)
		MAPIFreeBuffer(lpInternetCPID);

	if(lpMsgClass)
		MAPIFreeBuffer(lpMsgClass);

	return hr;
}


HRESULT MAPIToVMIME::fillVMIMEMail(IMessage *lpMessage, bool bSkipContent, vmime::messageBuilder* lpVMMessageBuilder, vmime::charset charset) {
	std::string		strOut,	strOutConverted;
	HRESULT			hr				= hrSuccess;
	LPSRowSet		prows			= NULL;
	LPSPropValue	lpSubject		= NULL;
	BOOL			bRealRTF = FALSE;

	try {
		if (HrGetOneProp(lpMessage, PR_SUBJECT, &lpSubject) == hrSuccess) {
			removeEnters(lpSubject->Value.lpszA);
			strOut = lpSubject->Value.lpszA;
		}
		else
			strOut = "";

		vmime::charset::convert(strOut, strOutConverted, MAPI_CHARSET, charset);

		// we can't ignore any errors.. the sender should know if an email is not sent correctly
		lpVMMessageBuilder->setSubject(vmime::text(vmime::word(strOutConverted, charset)));

		// handle recipients
		hr = processRecipients(lpMessage, lpVMMessageBuilder, charset);
		if (hr != hrSuccess)
			goto exit;			// Logging has been done in processRecipients()

		if (!bSkipContent) {
			// handle text
			hr = handleTextparts(lpMessage, lpVMMessageBuilder, charset, &bRealRTF);
			if (hr != hrSuccess)
				goto exit;

			// handle attachments
			hr = handleAttachments(lpMessage, lpVMMessageBuilder);
			if (hr != hrSuccess)
				goto exit;

			// check for TNEF information, and add winmail.dat if needed
			hr = handleTNEF(lpMessage, lpVMMessageBuilder, bRealRTF);
			if (hr != hrSuccess)
				goto exit;
		}

		/*
		  To convert a VMMessageBuilder to a VMMessage object, an 'expeditor' needs to be set in vmime
		  later on, this from will be overwritten .. so maybe replace this with a dummy value?
		 */

		std::string strName, strType, strEmAdd;
		hr = HrGetAddress(m_lpAdrBook, lpMessage, PR_SENDER_ENTRYID, PR_SENDER_NAME, PR_SENDER_ADDRTYPE, PR_SENDER_EMAIL_ADDRESS, strName, strType, strEmAdd);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get sender information. Error: 0x%08X", hr);
		    goto exit;
		}

		if (strName != "") {
			lpVMMessageBuilder->setExpeditor(vmime::mailbox(vmime::text(vmime::word(strName, MAPI_CHARSET)), strEmAdd));
		} else {
			lpVMMessageBuilder->setExpeditor(vmime::mailbox(strEmAdd));
		}
		// sender and reply-to is set elsewhere because it can only be done on a vmime::message object...
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on fill message: %s", e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception  on fill message");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (lpSubject)
		MAPIFreeBuffer(lpSubject);

	if (prows)
		FreeProws(prows);

	return hr;
}


// make vmime::mailbox from row
HRESULT MAPIToVMIME::getMailBox(LPSRow lpRow, vmime::mailbox** lppVMMailbox, vmime::charset charset) {
	HRESULT			hr				= hrSuccess;
	vmime::mailbox	*lpVMMailbox	= NULL;
	std::string	strName, strEmail, strType;

	hr = HrGetAddress(m_lpAdrBook, lpRow->lpProps, lpRow->cValues, PR_ENTRYID, PR_DISPLAY_NAME, PR_ADDRTYPE, PR_EMAIL_ADDRESS, strName, strType, strEmail);
	if(hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to create mailbox. Error: %08X", hr);
		goto exit;
	}

	if (strName == "" || strName == strEmail) {
		lpVMMailbox = new vmime::mailbox(strEmail);
	} else {
		lpVMMailbox = new vmime::mailbox(vmime::text(vmime::word(strName, MAPI_CHARSET)), strEmail);
	}

	*lppVMMailbox = lpVMMailbox;

exit:
	return hr;
}


// put textpart into PR_RTF_COMPRESSED
HRESULT MAPIToVMIME::handleTextparts(IMessage* lpMessage, vmime::messageBuilder* lpVMMessageBuilder, vmime::charset charset, BOOL* lpbRealRTF) {
	std::string strHTMLOut, strRtf;
	std::string strBody, strBodyConverted;
	HRESULT		hr						= hrSuccess;
	IStream*	lpCompressedRTFStream	= NULL; 
	IStream*	lpUncompressedRTFStream	= NULL; 
	char		*lpRBuff				= NULL;
	IStream*	lpBody					= NULL;
	IStream*	lpHTMLStream			= NULL;
	vmime::charset HTMLcharset;

	// set the encoder of plaintext body part
	vmime::encoding bodyEncoding("quoted-printable");
	// will skip enters in q-p encoding, "fixes" plaintext mails to be more plain text
	vmime::encoder *bodyEncoder = bodyEncoding.getEncoder();
	bodyEncoder->getProperties().setProperty("text", true);

	lpRBuff = new char[SIZETOREAD];

	if(lpbRealRTF) *lpbRealRTF = FALSE;

	// grabbing rtf
	hr = lpMessage->OpenProperty(PR_RTF_COMPRESSED, &IID_IStream, 0, 0, (LPUNKNOWN *)&lpCompressedRTFStream);
	if(hr == hrSuccess) {

		hr = WrapCompressedRTFStream(lpCompressedRTFStream, 0, &lpUncompressedRTFStream);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to create RTF-text stream. Error: 0x%08X", hr);
			goto exit;
		}

		hr = readStream(lpUncompressedRTFStream, strRtf);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to read RTF-text stream. Error: 0x%08X", hr);
			goto exit;
		}

		if(isrtfhtml(strRtf.c_str(), strRtf.size()) )
		{
			// Open html
			hr = lpMessage->OpenProperty(PR_HTML, &IID_IStream, 0, 0, (LPUNKNOWN *)&lpHTMLStream);
			if (hr == hrSuccess) {
				hr = readStream(lpHTMLStream, strHTMLOut);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to read HTML-text stream. Error: 0x%08X", hr);
					goto exit;
				}

				// strHTMLout is now escaped us-ascii HTML, this is what we will be sending
				// Or, if something failed, the HTML is now empty
				lpHTMLStream->Release(); lpHTMLStream = NULL;
			} else {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open HTML-text stream. Error: 0x%08X", hr);
				// continue with plaintext
			}
		} else if(isrtftext(strRtf.c_str(), strRtf.size()) )	{
			//Do nothing, only plain/text
		} else {
			// Real rtf
			if(lpbRealRTF) *lpbRealRTF = TRUE;
		}
	} else {
		lpLogger->Log(EC_LOGLEVEL_INFO, "Unable to open rtf-text stream. Error: 0x%08X", hr);
		hr = hrSuccess;
	}
	
	try {
		hr = lpMessage->OpenProperty(PR_BODY, &IID_IStream, 0, 0, (IUnknown**)&lpBody);
		if (hr == hrSuccess) {
			hr = readStream(lpBody, strBody);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to read plain-text stream. Error: 0x%08X", hr);
				goto exit;
			}
			lpBody->Release(); lpBody = NULL;
		} else {
			lpLogger->Log(EC_LOGLEVEL_INFO, "Unable to open plain-text stream. Error: 0x%08X", hr);
			hr = hrSuccess;
		}

		// Convert body to correct charset		
		vmime::charset::convert(strBody, strBodyConverted, MAPI_CHARSET, charset);

		// always use our textpart class
		lpVMMessageBuilder->constructTextPart(vmime::mediaType(vmime::mediaTypes::TEXT, "mapi"));

		// If HTML, create HTML and plaintext parts
		if (!strHTMLOut.empty()) {
		 	vmime::mapiTextPart& textPart = dynamic_cast<vmime::mapiTextPart&>(*lpVMMessageBuilder->getTextPart());
			textPart.setText(vmime::stringContentHandler(strHTMLOut));
			textPart.setCharset(charset);
			if (strBodyConverted != "")
				textPart.setPlainText(vmime::stringContentHandler(strBodyConverted));
		} else { // else just plaintext
			if (strBodyConverted != "") {
				// make sure we give vmime CRLF data, so SMTP servers (qmail) won't complain on the forced plaintext
				char *crlfconv = NULL;
				size_t outsize = 0;
				crlfconv = new char[strBodyConverted.length()*2];
				Util::BufferLFtoCRLF(strBodyConverted.length(), strBodyConverted.c_str(), crlfconv, &outsize);
				strBodyConverted.assign(crlfconv, outsize);
				delete [] crlfconv;

				// encode to q-p ourselves
				vmime::utility::inputStreamStringAdapter in(strBodyConverted);
				vmime::string outString;
				vmime::utility::outputStreamStringAdapter out(outString);

				bodyEncoder->encode(in, out);

				lpVMMessageBuilder->getTextPart()->setCharset(charset);
				dynamic_cast<vmime::mapiTextPart*>(lpVMMessageBuilder->getTextPart())->setPlainText(vmime::stringContentHandler(outString, bodyEncoding));
			}
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on text part: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception on text part");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (hr != hrSuccess)
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Could not parse mail body");

	if (lpHTMLStream)
		lpHTMLStream->Release();

	if (lpBody)
		lpBody->Release();
		
	if (lpRBuff)
		delete [] lpRBuff;

	if (lpUncompressedRTFStream)
		lpUncompressedRTFStream->Release();
		
	if (lpCompressedRTFStream)
		lpCompressedRTFStream->Release(); 

	if (bodyEncoder)
		delete bodyEncoder;

	return hr;
}

HRESULT MAPIToVMIME::handleXHeaders(IMessage *lpMessage, vmime::header *lpVMHeader) {
	HRESULT hr;
	ULONG i;
	ULONG cValues;
	LPSPropValue lpPropArray = NULL;
	LPSPropTagArray lpsPropTagArray = NULL;
	ULONG cNames;
	LPMAPINAMEID *lppNames = NULL;
	map<ULONG, ULONG> namemap;
	vmime::headerFieldFactory* hff = vmime::headerFieldFactory::getInstance();

	// get named props on message

	hr = lpMessage->GetProps(NULL, 0, &cValues, &lpPropArray);
	if (FAILED(hr))
		goto exit;

	cNames = 0;
	for (i = 0; i < cValues; i++) {
		if (PROP_ID(lpPropArray[i].ulPropTag) >= 0x8000)
			cNames++;
	}

	MAPIAllocateBuffer(CbNewSPropTagArray(cNames), (void **)&lpsPropTagArray);
	lpsPropTagArray->cValues = cNames;

	cNames = 0;
	for (i = 0; i < cValues; i++) {
		if (PROP_ID(lpPropArray[i].ulPropTag) >= 0x8000) {
			namemap[cNames] = i;
			lpsPropTagArray->aulPropTag[cNames++] = lpPropArray[i].ulPropTag;
		}
	}
	
	hr = lpMessage->GetNamesFromIDs(&lpsPropTagArray, NULL, 0, &cNames, &lppNames);
	if (FAILED(hr))
		goto exit;

	for (i = 0; i < cNames; i++) {
		if (lppNames[i] && lppNames[i]->ulKind == MNID_STRING && lppNames[i]->Kind.lpwstrName &&
			PROP_TYPE(lpPropArray[namemap[i]].ulPropTag) == PT_STRING8)
		{
			char *str;
			int l = unicodetombs(NULL, lppNames[i]->Kind.lpwstrName, 0) +1;
			str = new char[l];
			unicodetombs(str, lppNames[i]->Kind.lpwstrName, l);

			if ((str[0] == 'X' || str[0] == 'x') && str[1] == '-') {
				if (str[0] == 'x')
					capitalize(str);

				try {
					lpVMHeader->findField(str);
				}
				catch (vmime::exceptions::no_such_field) {
					lpVMHeader->appendField(hff->create(str, lpPropArray[namemap[i]].Value.lpszA));
				}
			}
			delete [] str;
		}
	}

exit:
	if (lpPropArray)
		MAPIFreeBuffer(lpPropArray);

	if (lpsPropTagArray)
		MAPIFreeBuffer(lpsPropTagArray);

	if (lppNames)
		MAPIFreeBuffer(lppNames);

	return hr;
}

HRESULT MAPIToVMIME::handleExtraHeaders(IMessage *lpMessage, vmime::charset& charset, vmime::header *lpVMHeader) {
	HRESULT hr = hrSuccess;
	LPSPropValue			lpMessageId			= NULL;
	LPSPropValue			lpImportance		= NULL;
	LPSPropValue			lpPriority			= NULL;
	LPSPropValue			lpConversationIndex	= NULL;
	LPSPropValue			lpConversationTopic = NULL;
	LPSPropValue			lpNormSubject		= NULL;
	LPSPropValue			lpSensitivity		= NULL;
	LPSPropValue			lpExpiryTime		= NULL;

	// Conversation headers. New Message-Id header is set just before sending.		
	if(HrGetOneProp(lpMessage, PR_IN_REPLY_TO_ID, &lpMessageId) == hrSuccess) {
		// reply
		vmime::messageId* mid = new vmime::messageId(lpMessageId->Value.lpszA);
		lpVMHeader->InReplyTo().getValue().appendMessageId(mid);
		// vmime will delete the mid
		MAPIFreeBuffer(lpMessageId);
	}
	lpMessageId = NULL;

	// Outlook never adds this property
	if(HrGetOneProp(lpMessage, PR_INTERNET_REFERENCES, &lpMessageId) == hrSuccess) {
		// reply
		vmime::messageId* mid = new vmime::messageId(lpMessageId->Value.lpszA);
		lpVMHeader->References().getValue().appendMessageId(mid);
		// vmime will delete the mid
		MAPIFreeBuffer(lpMessageId);
	}
	lpMessageId = NULL;

	// priority settings
	char* priomap[3] = { "5 (Lowest)", "3 (Normal)", "1 (Highest)" }; // 2 and 4 cannot be set from outlook
	vmime::headerFieldFactory* hff = vmime::headerFieldFactory::getInstance();
	if (HrGetOneProp(lpMessage, PR_IMPORTANCE, &lpImportance) == hrSuccess) {
		lpVMHeader->appendField(hff->create("X-Priority", priomap[(lpImportance->Value.ul)&3])); // IMPORTANCE_* = 0..2
	} else if (HrGetOneProp(lpMessage, PR_PRIORITY, &lpPriority) == hrSuccess) {
		lpVMHeader->appendField(hff->create("X-Priority", priomap[(lpPriority->Value.ul+1)&3])); // PRIO_* = -1..1
	}

	// When adding a X-Priority, spamassassin may add a severe punishment because no User-Agent header
	// or X-Mailer header is present. So we set the X-Mailer header :)
	lpVMHeader->appendField(hff->create("X-Mailer", "Zarafa "PROJECT_VERSION_DOT_STR"-"PROJECT_SVN_REV_STR));

	// PR_CONVERSATION_INDEX
	if (HrGetOneProp(lpMessage, PR_CONVERSATION_INDEX, &lpConversationIndex) == hrSuccess) {
		vmime::string inString;
		inString.assign((const char*)lpConversationIndex->Value.bin.lpb, lpConversationIndex->Value.bin.cb);

		vmime::encoder* enc = vmime::encoderFactory::getInstance()->create("base64");
		vmime::utility::inputStreamStringAdapter in(inString);
		vmime::string outString;
		vmime::utility::outputStreamStringAdapter out(outString);

		enc->encode(in, out);
		delete enc;

		lpVMHeader->appendField(hff->create("Thread-Index", outString));
	}

	// PR_CONVERSATION_TOPIC is always the original started topic
	if (HrGetOneProp(lpMessage, PR_CONVERSATION_TOPIC, &lpConversationTopic) == hrSuccess) {
		if (HrGetOneProp(lpMessage, PR_NORMALIZED_SUBJECT, &lpNormSubject) != hrSuccess ||
			strcmp(lpNormSubject->Value.lpszA, lpConversationTopic->Value.lpszA) != 0)
			{
				removeEnters(lpConversationTopic->Value.lpszA);
				lpVMHeader->appendField(hff->create("Thread-Topic",
													vmime::text(lpConversationTopic->Value.lpszA, charset).generate()));
			}
	}

	if (HrGetOneProp(lpMessage, PR_SENSITIVITY, &lpSensitivity) == hrSuccess) {
		char *strSens;
		switch (lpSensitivity->Value.ul) {
		case SENSITIVITY_PERSONAL:
			strSens = "Personal";
			break;
		case SENSITIVITY_PRIVATE:
			strSens = "Private";
			break;
		case SENSITIVITY_COMPANY_CONFIDENTIAL:
			strSens = "Company-Confidential";
			break;
		case SENSITIVITY_NONE:
		default:
			strSens = NULL;
			break;
		};
		if (strSens)
			lpVMHeader->appendField(hff->create("Sensitivity", strSens));
	}

	if (HrGetOneProp(lpMessage, PR_EXPIRY_TIME, &lpExpiryTime) == hrSuccess) {
		lpVMHeader->appendField(hff->create("Expiry-Time", FiletimeTovmimeDatetime(lpExpiryTime->Value.ft).generate()));
	}
	
	

//exit:
	if (lpExpiryTime)
		MAPIFreeBuffer(lpExpiryTime);

	if (lpImportance)
		MAPIFreeBuffer(lpImportance);

	if (lpPriority)
		MAPIFreeBuffer(lpPriority);

	if (lpSensitivity)
		MAPIFreeBuffer(lpSensitivity);

	if (lpConversationIndex)
		MAPIFreeBuffer(lpConversationIndex);

	if(lpConversationTopic)
		MAPIFreeBuffer(lpConversationTopic);

	if(lpNormSubject)
		MAPIFreeBuffer(lpNormSubject);

	return hr;
}

HRESULT MAPIToVMIME::handleContactEntryID(ULONG cValues, LPSPropValue lpProps, string &strName, string &strType, string &strEmail)
{
	HRESULT hr = hrSuccess;
	LPCONTAB_ENTRYID lpContabEntryID = NULL;
	GUID* guid = NULL;
	ULONG ulObjType;
	LPSPropTagArray lpNameTags = NULL;
	LPSPropValue lpNamedProps = NULL;
	LPMAILUSER lpContact = NULL;
	LPMAPINAMEID *lppNames = NULL;
	ULONG i;
	ULONG ulNames = 5;
	MAPINAMEID mnNamedProps[5] = {
		// offset 0, every offset < 3 is + 0x10
		{(LPGUID)&PSETID_Address, MNID_ID, {0x8080}}, // display name
		{(LPGUID)&PSETID_Address, MNID_ID, {0x8082}}, // address type
		{(LPGUID)&PSETID_Address, MNID_ID, {0x8083}}, // email address
		{(LPGUID)&PSETID_Address, MNID_ID, {0x8084}}, // original display name (unused)
		{(LPGUID)&PSETID_Address, MNID_ID, {0x8085}}, // real entryid
	};

	if (PROP_TYPE(lpProps[0].ulPropTag) != PT_BINARY) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	lpContabEntryID = (LPCONTAB_ENTRYID)lpProps[0].Value.bin.lpb;
	if (lpContabEntryID == NULL) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	guid = (GUID*)&lpContabEntryID->muid;
	if (sizeof(CONTAB_ENTRYID) > lpProps[0].Value.bin.cb || *guid != PSETID_CONTACT_FOLDER_RECIPIENT || lpContabEntryID->email_offset > 2) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = m_lpSession->OpenEntry(lpContabEntryID->cbeid, (LPENTRYID)lpContabEntryID->abeid, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpContact);
	if (hr != hrSuccess)
		goto exit;

	// add offset to get correct named properties
	for (i = 0; i < ulNames; i++)
		mnNamedProps[i].Kind.lID += (lpContabEntryID->email_offset * 0x10);

	hr = MAPIAllocateBuffer(sizeof(LPMAPINAMEID) * (ulNames), (void**)&lppNames);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "No memory for named ids from contact");
		goto exit;
	}

	for (i = 0; i < ulNames; i++)
		lppNames[i] = &mnNamedProps[i];

	hr = lpContact->GetIDsFromNames(ulNames, lppNames, MAPI_CREATE, &lpNameTags);
	if (FAILED(hr))
		goto exit;

	lpNameTags->aulPropTag[0] = CHANGE_PROP_TYPE(lpNameTags->aulPropTag[0], PT_STRING8);
	lpNameTags->aulPropTag[1] = CHANGE_PROP_TYPE(lpNameTags->aulPropTag[1], PT_STRING8);
	lpNameTags->aulPropTag[2] = CHANGE_PROP_TYPE(lpNameTags->aulPropTag[2], PT_STRING8);
	lpNameTags->aulPropTag[3] = CHANGE_PROP_TYPE(lpNameTags->aulPropTag[3], PT_STRING8);
	lpNameTags->aulPropTag[4] = CHANGE_PROP_TYPE(lpNameTags->aulPropTag[4], PT_BINARY);

	hr = lpContact->GetProps(lpNameTags, 0, &ulNames, &lpNamedProps);
	if (FAILED(hr))
		goto exit;

	hr = HrGetAddress(m_lpAdrBook, lpNamedProps, ulNames,
					  lpNameTags->aulPropTag[4], lpNameTags->aulPropTag[0], lpNameTags->aulPropTag[1], lpNameTags->aulPropTag[2],
					  strName, strType, strEmail);

exit:
	if (lpNamedProps)
		MAPIFreeBuffer(lpNamedProps);

	if (lpNameTags)
		MAPIFreeBuffer(lpNameTags);

	if (lpContact)
		lpContact->Release();

	if (lppNames)
		MAPIFreeBuffer(lppNames);

	return hr;
}

HRESULT MAPIToVMIME::handleSenderInfo(IMessage *lpMessage, vmime::header *lpVMHeader) {
	HRESULT hr = hrSuccess;
	ULONG cValues;
	LPSPropValue lpProps = NULL;
	LPSPropTagArray lpPropTags = NULL;
	LPSPropValue lpReadReceipt = NULL;

	// sender information
	std::string strEmail, strName, strType;
	std::string strResEmail, strResName, strResType;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray(4), (void**)&lpPropTags);
	if (hr != hrSuccess)
		goto exit;

	lpPropTags->cValues = 4;
	lpPropTags->aulPropTag[0] = PR_SENDER_ENTRYID;
	lpPropTags->aulPropTag[1] = PR_SENDER_NAME;
	lpPropTags->aulPropTag[2] = PR_SENDER_ADDRTYPE;
	lpPropTags->aulPropTag[3] = PR_SENDER_EMAIL_ADDRESS;

	hr = lpMessage->GetProps(lpPropTags, 0, &cValues, &lpProps);
	if (FAILED(hr))
		goto exit;

	hr = handleContactEntryID(cValues, lpProps, strName, strType, strEmail);
	if (hr != hrSuccess) {
		// Store owner, actual sender
		hr = HrGetAddress(m_lpAdrBook, lpProps, cValues, PR_SENDER_ENTRYID, PR_SENDER_NAME, PR_SENDER_ADDRTYPE, PR_SENDER_EMAIL_ADDRESS, strName, strType, strEmail);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get sender information. Error: 0x%08X", hr);
			goto exit;
		}
	}

	MAPIFreeBuffer(lpProps);
	lpProps = NULL;

	// -- sender
	lpPropTags->cValues = 4;
	lpPropTags->aulPropTag[0] = PR_SENT_REPRESENTING_ENTRYID;
	lpPropTags->aulPropTag[1] = PR_SENT_REPRESENTING_NAME;
	lpPropTags->aulPropTag[2] = PR_SENT_REPRESENTING_ADDRTYPE;
	lpPropTags->aulPropTag[3] = PR_SENT_REPRESENTING_EMAIL_ADDRESS;

	hr = lpMessage->GetProps(lpPropTags, 0, &cValues, &lpProps);
	if (FAILED(hr))
		goto exit;

	hr = handleContactEntryID(cValues, lpProps, strName, strType, strEmail);
	if (hr != hrSuccess) {
		hr = HrGetAddress(m_lpAdrBook, lpProps, cValues, PR_SENT_REPRESENTING_ENTRYID, PR_SENT_REPRESENTING_NAME, PR_SENT_REPRESENTING_ADDRTYPE, PR_SENT_REPRESENTING_EMAIL_ADDRESS, strResName, strResType, strResEmail);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to get representing information. Error: 0x%08X", hr);
			// ignore error, since we still have enough information to send, maybe not just under the correct name
			hr = hrSuccess;
		}
	}

	// Set representing as from address, when possible
	// Ignore PR_SENT_REPRESENTING if the email adress is the same as the PR_SENDER email address
	if (strResEmail != "" && strResEmail != strEmail) {
		if (strResName == "" || strResName == strResEmail) 
			lpVMHeader->From().setValue(vmime::mailbox(strResEmail));
		else
			lpVMHeader->From().setValue(vmime::mailbox(vmime::text(vmime::word(strResName, MAPI_CHARSET)), strResEmail));

		// spooler checked if this is allowed
		if(strResEmail != strEmail) {
			// Set store owner as sender
			if (strName == "" || strName == strEmail) 
				lpVMHeader->Sender().setValue(vmime::mailbox(strEmail));
			else
				lpVMHeader->Sender().setValue(vmime::mailbox(vmime::text(vmime::word(strName, MAPI_CHARSET)), strEmail));
		}
	} else {
		// Set store owner as from, sender does not need to be set
		if (strName == "" || strName == strEmail) 
			lpVMHeader->From().setValue(vmime::mailbox(strEmail));
		else
			lpVMHeader->From().setValue(vmime::mailbox(vmime::text(vmime::word(strName, MAPI_CHARSET)), strEmail));
	}

	// read receipt request
	if ((HrGetOneProp(lpMessage, PR_READ_RECEIPT_REQUESTED, &lpReadReceipt) == hrSuccess) && (lpReadReceipt->Value.b == TRUE)) {
		vmime::mailboxList mbl;
		if (strResEmail != "" && strResEmail != strEmail) {
			// use user added from address
			if (strResName == "" || strName == strResEmail)
				mbl.appendMailbox(new vmime::mailbox(strResEmail));
			else
				mbl.appendMailbox(new vmime::mailbox(vmime::text(vmime::word(strResName, MAPI_CHARSET)), strResEmail));
		} else {
			if (strName == "" || strName == strEmail)
				mbl.appendMailbox(new vmime::mailbox(strEmail));
			else
				mbl.appendMailbox(new vmime::mailbox(vmime::text(vmime::word(strName, MAPI_CHARSET)), strEmail));
		}
		lpVMHeader->DispositionNotificationTo().setValue(mbl);
	}


exit:
	if (lpPropTags)
		MAPIFreeBuffer(lpPropTags);
		
	if (lpProps)
		MAPIFreeBuffer(lpProps);

	if (lpReadReceipt)
		MAPIFreeBuffer(lpReadReceipt);

	return hr;
}

HRESULT MAPIToVMIME::handleReplyTo(IMessage *lpMessage, vmime::header *lpVMHeader) {
	HRESULT			hr = hrSuccess;
	LPSPropValue	lpReplyTo = NULL;
	FLATENTRYLIST	*lpEntryList = NULL;
	FLATENTRY		*lpEntry = NULL;
	LPCONTAB_ENTRYID lpContabEntryID = NULL;
	GUID*			guid = NULL;
	ULONG			ulObjType;
	LPMAILUSER		lpContact = NULL;
	string			strName, strType, strEmail;

	// "Email1DisplayName","Email1AddressType","Email1Address","Email1EntryID"
	ULONG lpulNamesIDs[] = {0x8080, 0x8082, 0x8083, 0x8085,
							0x8090, 0x8092, 0x8093, 0x8095,
							0x80A0, 0x80A2, 0x80A3, 0x80A5};
	ULONG cNames, i, offset;
	LPMAPINAMEID lpNames = NULL;
	LPMAPINAMEID *lppNames = NULL;
	LPSPropTagArray lpNameTagArray = NULL;
	LPSPropValue lpAddressProps = NULL;


	if (HrGetOneProp(lpMessage, PR_REPLY_RECIPIENT_ENTRIES, &lpReplyTo) != hrSuccess)
		goto exit;

	if (lpReplyTo->Value.bin.cb == 0)
		goto exit;


	lpEntryList = (FLATENTRYLIST *)lpReplyTo->Value.bin.lpb;

	if (lpEntryList->cEntries == 0)
		goto exit;

	lpEntry = (FLATENTRY *)&lpEntryList->abEntries;

	hr = HrGetAddress(m_lpAdrBook, (LPENTRYID)lpEntry->abEntry, lpEntry->cb, strName, strType, strEmail);
	if (hr != hrSuccess) {
		if (!m_lpSession) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		// user selected a contact (or distrolist ?) in the reply-to
		lpContabEntryID = (LPCONTAB_ENTRYID)lpEntry->abEntry;
		guid = (GUID*)&lpContabEntryID->muid;

		if (sizeof(CONTAB_ENTRYID) > lpEntry->cb || *guid != PSETID_CONTACT_FOLDER_RECIPIENT || lpContabEntryID->email_offset > 2)
			goto exit;

		hr = m_lpSession->OpenEntry(lpContabEntryID->cbeid, (LPENTRYID)lpContabEntryID->abeid, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpContact);
		if (hr != hrSuccess)
			goto exit;

		cNames = (sizeof(lpulNamesIDs)/sizeof(ULONG));

		hr = MAPIAllocateBuffer(sizeof(MAPINAMEID) * cNames, (void**)&lpNames);
		if (hr != hrSuccess)
			goto exit;

		hr = MAPIAllocateBuffer(sizeof(LPMAPINAMEID) * cNames, (void**)&lppNames);
		if (hr != hrSuccess)
			goto exit;

		for (i = 0; i < cNames; i++) {
			lpNames[i].lpguid = (GUID*)&PSETID_Address;
			lpNames[i].ulKind = MNID_ID;
			lpNames[i].Kind.lID = lpulNamesIDs[i];
			lppNames[i] = &lpNames[i];
		}

		hr = lpContact->GetIDsFromNames(cNames, lppNames, 0, &lpNameTagArray);
		if (FAILED(hr))
			goto exit;

		hr = lpContact->GetProps(lpNameTagArray, 0, &cNames, &lpAddressProps);
		if (FAILED(hr))
			goto exit;

		offset = lpContabEntryID->email_offset * 4; // 4 props per email address

		if (PROP_TYPE(lpAddressProps[offset+0].ulPropTag) == PT_ERROR || PROP_TYPE(lpAddressProps[offset+1].ulPropTag) == PT_ERROR ||
			PROP_TYPE(lpAddressProps[offset+2].ulPropTag) == PT_ERROR || PROP_TYPE(lpAddressProps[offset+3].ulPropTag) == PT_ERROR)
			goto exit;

		if (strcmp(lpAddressProps[offset+1].Value.lpszA, "SMTP") == 0) {
			strName = lpAddressProps[offset+0].Value.lpszA;
			strEmail = lpAddressProps[offset+2].Value.lpszA;
		} else if (strcmp(lpAddressProps[offset+1].Value.lpszA, "ZARAFA") == 0) {
			hr = HrGetAddress(m_lpAdrBook, (LPENTRYID)lpAddressProps[offset+2].Value.bin.lpb, lpAddressProps[offset+2].Value.bin.cb, strName, strType, strEmail);
			if (hr != hrSuccess)
				goto exit;
		}
	}

	// vmime-0.7.1 can only set 1 email address in the ReplyTo field.
	if (strName != "" && strName != strEmail)
		lpVMHeader->ReplyTo().setValue(vmime::mailbox(vmime::text(vmime::word(strName, MAPI_CHARSET)), strEmail));
	else
		lpVMHeader->ReplyTo().setValue(vmime::mailbox(strEmail));

	hr = hrSuccess;

exit:
	if (lpNames)
		MAPIFreeBuffer(lpNames);

	if (lppNames)
		MAPIFreeBuffer(lppNames);

	if (lpNameTagArray)
		MAPIFreeBuffer(lpNameTagArray);

	if (lpAddressProps)
		MAPIFreeBuffer(lpAddressProps);

	if (lpContact)
		lpContact->Release();

	if (lpReplyTo)
		MAPIFreeBuffer(lpReplyTo);

	return hr;
}

HRESULT MAPIToVMIME::handleTNEF(IMessage* lpMessage, vmime::messageBuilder* lpVMMessageBuilder, BOOL bRealRTF) {
	HRESULT			hr				= hrSuccess;
	LPSPropValue	lpUseTNEF		= NULL;
	LPSPropValue	lpSendAsICal	= NULL;
	LPSPropValue	lpOutlookVersion = NULL;
	LPSPropValue	lpMessageClass	= NULL;
	LPSPropValue	lpDelegateRule	= NULL;
	LPSPropValue	lpApptCounterPropose = NULL;
	IStream 		*lpStream = NULL;
	MapiToICal		*mapiical = NULL;
	
	LPMAPINAMEID	lpNames			= NULL;	
	LPSPropTagArray lpNameTagArray	= NULL;
	ULONG			cNames		= 0;
	ULONG			ulNamesIDs	= 0;
	bool			blForceTnef = sopt.force_tnef; // force sending TNEF when requested
	
	std::list<ULONG> lstOLEAttach; // list of OLE attachments that must be sent via TNEF
	std::list<ULONG>::iterator iterAttach;
	IMAPITable *	lpAttachTable = NULL;
	LPSRowSet		lpAttachRows = NULL;
	SizedSPropTagArray(2, sptaAttachProps) = {2, {PR_ATTACH_METHOD, PR_ATTACH_NUM }};
	SizedSPropTagArray(5, sptaOLEAttachProps) = {5, {PR_ATTACH_FILENAME, PR_ATTACH_LONG_FILENAME, PR_ATTACH_DATA_OBJ, PR_ATTACH_CONTENT_ID, PR_ATTACHMENT_HIDDEN}};
	
	try {
	    // Find all ATTACH_OLE attachments and put them in lstOLEAttach
	    hr = lpMessage->GetAttachmentTable(0, &lpAttachTable);
	    if(hr != hrSuccess)
	        goto exit;
	    
        hr = HrQueryAllRows(lpAttachTable, (LPSPropTagArray)&sptaAttachProps, NULL, NULL, -1, &lpAttachRows);
        if(hr != hrSuccess)
            goto exit;
            
        for(unsigned int i=0; i < lpAttachRows->cRows; i++) {
            if(lpAttachRows->aRow[i].lpProps[0].ulPropTag == PR_ATTACH_METHOD && 
                lpAttachRows->aRow[i].lpProps[1].ulPropTag == PR_ATTACH_NUM &&
                lpAttachRows->aRow[i].lpProps[0].Value.ul == ATTACH_OLE)
            {
                lstOLEAttach.push_back(lpAttachRows->aRow[i].lpProps[1].Value.ul);
            }
        }
	
        // Start processing TNEF properties
		if(HrGetOneProp(lpMessage, PR_EC_USE_TNEF, &lpUseTNEF) != hrSuccess) {
			lpUseTNEF = NULL;
		}

		if(HrGetOneProp(lpMessage, PR_EC_SEND_AS_ICAL, &lpSendAsICal) != hrSuccess) {
			lpSendAsICal = NULL;
		}
		
		if(HrGetOneProp(lpMessage, PR_EC_OUTLOOK_VERSION, &lpOutlookVersion) != hrSuccess) {
			lpOutlookVersion = NULL;
		} 
		
		if(HrGetOneProp(lpMessage, PR_MESSAGE_CLASS, &lpMessageClass) != hrSuccess) {
			lpMessageClass = NULL;
		}

		if(HrGetOneProp(lpMessage, PR_DELEGATED_BY_RULE, &lpDelegateRule) != hrSuccess) {
			lpDelegateRule = NULL;
		}

		ulNamesIDs = 0x8257; //dispidApptCounterProposal

		cNames = 1;

		hr = MAPIAllocateBuffer(sizeof(MAPINAMEID) * cNames, (void**)&lpNames);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to reserve memory for named properties");
			goto exit;
		}

		lpNames[0].lpguid = (GUID*)&PSETID_Appointment;
		lpNames[0].ulKind = MNID_ID;
		lpNames[0].Kind.lID = ulNamesIDs;
		
		hr = lpMessage->GetIDsFromNames(cNames, &lpNames, 0, &lpNameTagArray);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get named properties information");
			goto exit;
		}
		lpNameTagArray->aulPropTag[0] = CHANGE_PROP_TYPE(lpNameTagArray->aulPropTag[0], PT_BOOLEAN);
		
		hr = lpMessage->GetProps(lpNameTagArray, 0, &cNames, &lpApptCounterPropose);
		if (FAILED(hr)) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get named properties values");
			goto exit;
		}
		hr = hrSuccess;

		// no propose new time support for ical
		if(lpApptCounterPropose && lpApptCounterPropose->ulPropTag == lpNameTagArray->aulPropTag[0] && lpApptCounterPropose->Value.b)
			blForceTnef = true;

		// currently no task support for ical
		if (lpMessageClass && strncmp("IPM.Task", lpMessageClass->Value.lpszA, 8) == 0)
			blForceTnef = true;

		// delegation of meeting requests need to be in tnef too because of special properties
		if (lpDelegateRule && lpDelegateRule->Value.b == TRUE)
			blForceTnef = true;

        /*
         * Outlook 2000 always sets PR_EC_SEND_AS_ICAL to FALSE, because the
         * iCal option is somehow missing from the options property sheet, and 
         * PR_EC_USE_TNEF is never set. So for outlook 2000 (9.0), we check the
         * _existence_ of PR_EC_SEND_AS_ICAL as a hint to use TNEF.
         */
		if ((lpUseTNEF && lpUseTNEF->Value.b) || (lpSendAsICal && lpSendAsICal->Value.b) || 
			(lpSendAsICal && lpOutlookVersion && strcmp(lpOutlookVersion->Value.lpszA,"9.0") == 0) ||
			(lpMessageClass && (strncmp("IPM.Note", lpMessageClass->Value.lpszA, 8) != 0) ) || 
			bRealRTF == TRUE)
		{
		    // Send either TNEF or iCal data
		    
			vmime::attachment *lpTNEFAtt;
			inputStreamMAPIAdapter *inputDataStream = NULL;

			/* 
			 * Send TNEF information for this message if we really need to, or otherwise iCal
			 */
			
			if (lstOLEAttach.size() == 0 && !blForceTnef && lpMessageClass && (strncmp("IPM.Note", lpMessageClass->Value.lpszA, 8) != 0)) {
				// iCAL
				string ical, method;

				lpLogger->Log(EC_LOGLEVEL_INFO, "Adding ICS attachment for extra information");

				CreateMapiToICal(m_lpAdrBook, "utf-8", &mapiical);

				hr = mapiical->AddMessage(lpMessage, 0);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create ical object, sending as TNEF");
					goto tnef_anyway;
				}

				hr = mapiical->Finalize(0, &method, &ical);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create ical object, sending as TNEF");
					goto tnef_anyway;
				}

				vmime::mapiTextPart *lpvmMapiText = dynamic_cast<vmime::mapiTextPart*>(lpVMMessageBuilder->getTextPart());
				lpvmMapiText->setOtherText(vmime::stringContentHandler(ical));
				lpvmMapiText->setOtherContentType(vmime::mediaType(vmime::mediaTypes::TEXT, "calendar"));
				lpvmMapiText->setOtherContentEncoding(vmime::encoding(vmime::encodingTypes::EIGHT_BIT));
				lpvmMapiText->setOtherMethod(method);
				lpvmMapiText->setOtherCharset(vmime::charset("utf-8"));

			} else {
				lpLogger->Log(EC_LOGLEVEL_INFO, "Adding TNEF attachment for extra information");
tnef_anyway:
				hr = CreateStreamOnHGlobal(NULL, TRUE, &lpStream);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to create stream for TNEF attachment. Error 0x%08X", hr);
					goto exit;
				}
			 
				ECTNEF tnef(TNEF_ENCODE, lpMessage, lpStream);
			
				// Encode the properties now
				hr = tnef.AddProps(TNEF_PROP_EXCLUDE, (LPSPropTagArray)&sptaExclude);
				if(hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to exclude properties from TNEF object");
					goto exit;
				}
			
				if (bRealRTF == TRUE) {
					hr = tnef.AddProps(TNEF_PROP_INCLUDE, (LPSPropTagArray)&sptaRTFInclude);
					if(hr != hrSuccess) {
						lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to include properties to TNEF object");
						goto exit;
					}
				}
				
				// Add all OLE attachments
				for(iterAttach = lstOLEAttach.begin(); iterAttach != lstOLEAttach.end(); iterAttach++) {
				    tnef.FinishComponent(0x00002000, *iterAttach, (LPSPropTagArray)&sptaOLEAttachProps);
				}
		
				// Write the stream
				hr = tnef.Finish();

				inputDataStream =  new inputStreamMAPIAdapter(lpStream);
			
				// Now, add the stream as an attachment to the message, filename winmail.dat 
				// and MIME type 'application/ms-tnef', no content-id
				lpTNEFAtt = new mapiAttachment(vmime::streamContentHandler(inputDataStream, 0 ,true),
											   vmime::encoding("base64"), vmime::mediaType("application/ms-tnef"), "",
											   vmime::word("winmail.dat"));
											  
				// add to message (copies pointer, not data)
				lpVMMessageBuilder->appendAttachment(lpTNEFAtt); 
			}
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on fill message: %s", e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception on fill message");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
    if (lpAttachRows)
        FreeProws(lpAttachRows);
        
    if (lpAttachTable)
        lpAttachTable->Release();
    
	if (mapiical)
		delete mapiical;

	if (lpStream)
		lpStream->Release();

	if (lpNames)
		MAPIFreeBuffer(lpNames);

	if (lpNameTagArray)
		MAPIFreeBuffer(lpNameTagArray);

	if(lpApptCounterPropose)
		MAPIFreeBuffer(lpApptCounterPropose);

	if (lpMessageClass)
		MAPIFreeBuffer(lpMessageClass);

	if (lpOutlookVersion)
		MAPIFreeBuffer(lpOutlookVersion);

	if (lpUseTNEF)
		MAPIFreeBuffer(lpUseTNEF);

	if (lpSendAsICal)
		MAPIFreeBuffer(lpSendAsICal);

	return hr;
}

// converts x-mail-sender -> X-Mail-Sender
void MAPIToVMIME::capitalize(char *s) {
	char *p;

	s[0] = toupper(s[0]);		// x to X
	p = s;
	while ((p = strchr(p, '-'))) { // capitalize every char after a -
		p++;
		if (*p != '\0')
			*p = toupper(*p);
	}
}

// makes spaces from enters, avoid spaces in Subject
void MAPIToVMIME::removeEnters(char *s) {
	char *p = s;

	while (*p) {
		if (*p == '\r' || *p == '\n') *p = ' ';
		p++;
	}
}
