/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include "stringutil.h"

// Damn windows header defines max which break C++ header files
#undef max

#include <string>
#include <fstream>
#include <iostream>
#include <stdlib.h>

// vmime
#include "vmime/vmime.hpp"
#include "vmime/platforms/posix/posixHandler.hpp"

// mapi
#include <mapix.h>
#include <mapiutil.h>
#include <mapiext.h>
#include "edkmdb.h"

#include "CommonUtil.h"
// inetmapi
#include "inetmapi.h"
#include "VMIMEToMAPI.h"
#include "MAPIToVMIME.h"
#include "ECVMIMEUtils.h"
#include "ECMapiUtils.h"


// sets only once to init library
static const vmime::platformDependant::handler *vphInited = NULL;

ECSender::ECSender(ECLogger *newlpLogger, const std::string &strSMTPHost) {
	lpLogger = newlpLogger;
	if (!lpLogger) {
		lpLogger = new ECLogger_Null();
		delete_logger = true;
	} else {
		delete_logger = false;
	}
	smtpresult = 0;
	smtphost = strSMTPHost;
}

ECSender::~ECSender() {
	if (delete_logger)
		delete lpLogger;
}

int ECSender::getSMTPResult() {
	return smtpresult;
}

const char* ECSender::getErrorString() {
	return error.c_str();
}

void ECSender::setError(const std::string &newError) {
	error = newError;
}

const unsigned int ECSender::getRecipientErrorCount() const
{
	return lstFailedRecipients.size();
}

const unsigned int ECSender::getRecipientErrorSMTPCode(unsigned int offset) const
{
	if (offset >= lstFailedRecipients.size())
		return 0;
	return lstFailedRecipients[offset].ulSMTPcode;
}

const string ECSender::getRecipientErrorText(unsigned int offset) const
{
	if (offset >= lstFailedRecipients.size())
		return string();
	return lstFailedRecipients[offset].strSMTPResponse;
}

const string ECSender::getRecipientErrorDisplayName(unsigned int offset) const
{
	if (offset >= lstFailedRecipients.size())
		return string();
	return lstFailedRecipients[offset].strRecipName;
}

const string ECSender::getRecipientErrorEmailAddress(unsigned int offset) const
{
	if (offset >= lstFailedRecipients.size())
		return string();
	return lstFailedRecipients[offset].strRecipEmail;
}


INETMAPI_API ECSender* CreateSender(ECLogger *lpLogger, const std::string &smtp) {
	return new ECVMIMESender(lpLogger, smtp);
}

// parse rfc822 input, and set props in lpMessage
INETMAPI_API HRESULT IMToMAPI(IMAPISession *lpSession, IMsgStore *lpMsgStore, IAddrBook *lpAddrBook, IMessage *lpMessage, const string &input, delivery_options dopt, ECLogger *lpLogger)
{
	HRESULT hr = hrSuccess;
	VMIMEToMAPI *VMToM = NULL;
	ULONG cbEntryId = 0;
	LPENTRYID lpEntryId = NULL;
	IMailUser *lpMailUser = NULL;
	ULONG ulObjType = 0;

	hr = lpSession->QueryIdentity(&cbEntryId, &lpEntryId);
	if (hr != hrSuccess)
		goto exit;
		
	hr = lpAddrBook->OpenEntry(cbEntryId, lpEntryId, NULL, 0 , &ulObjType, (LPUNKNOWN*)&lpMailUser);
	if (hr != hrSuccess)
		goto exit;
	
	VMToM = new VMIMEToMAPI(lpAddrBook, lpLogger, lpMailUser, dopt);

	if (!vphInited) {
		try {
			vphInited = vmime::platformDependant::getHandler();
		}
		catch (vmime::exceptions::no_platform_dependant_handler) {
			vmime::platformDependant::setHandler<vmime::platforms::posix::posixHandler>();
		};
	}

	// fill mapi object from buffer
	hr = VMToM->convertVMIMEToMAPI(input, lpMessage);

exit:
	delete VMToM;
	
	if(lpMailUser)
		lpMailUser->Release();

	if (lpEntryId)
		MAPIFreeBuffer(lpEntryId);

	return hr;
}

// Read properties from lpMessage object and fill a buffer with internet rfc822 format message
INETMAPI_API HRESULT IMToINet(IMAPISession *lpSession, IAddrBook *lpAddrBook, IMessage *lpMessage, char** lppbuf, sending_options sopt, ECLogger *lpLogger)
{
    HRESULT hr = hrSuccess;
    std::ostringstream oss;
    char *lpszData = NULL;
    
    hr = IMToINet(lpSession, lpAddrBook, lpMessage, oss, sopt, lpLogger);
    
    if(hr != hrSuccess)
        goto exit;
        
    lpszData = new char[oss.str().size()+1];
    strcpy(lpszData, oss.str().c_str());

    *lppbuf = lpszData;
exit:    
    return hr;
}

INETMAPI_API HRESULT IMToINet(IMAPISession *lpSession, IAddrBook *lpAddrBook, IMessage *lpMessage, std::ostream &os, sending_options sopt, ECLogger *lpLogger)
{
	HRESULT			hr			= hrSuccess;
	LPSPropValue	lpTime		= NULL;
	LPSPropValue	lpMessageId	= NULL;
	MAPIToVMIME*	mToVM		= new MAPIToVMIME(lpSession, lpAddrBook, lpLogger, sopt);
	vmime::message	*lpVMMessage = NULL;
	char			*lpszRawSMTP = NULL;
	vmime::utility::outputStreamAdapter adapter(os);

	if (!vphInited) {
		try {
			vphInited = vmime::platformDependant::getHandler();
		}
		catch (vmime::exceptions::no_platform_dependant_handler) {
			vmime::platformDependant::setHandler<vmime::platforms::posix::posixHandler>();
		};
	}

	hr = mToVM->convertMAPIToVMIME(lpMessage, &lpVMMessage, &lpszRawSMTP);
	if (hr != hrSuccess)
		goto exit;

	try {
		// vmime messageBuilder has set Date header to now(), so we overwrite it.
		if (HrGetOneProp(lpMessage, PR_CLIENT_SUBMIT_TIME, &lpTime) == hrSuccess) {
			lpVMMessage->getHeader()->Date().setValue(FiletimeTovmimeDatetime(lpTime->Value.ft));
		}
		// else, try PR_MESSAGE_DELIVERY_TIME, maybe other timestamps?

		if (HrGetOneProp(lpMessage, PR_INTERNET_MESSAGE_ID, &lpMessageId) == hrSuccess) {
			lpVMMessage->getHeader()->MessageId().setValue(lpMessageId->Value.lpszA);
		}

		if (lpszRawSMTP) {
			// create message from raw data
			lpVMMessage->getHeader()->generate(adapter);
			os << lpszRawSMTP;
		} else {
			// generate message and put into buffer.
			if(sopt.headers_only)
			    lpVMMessage->getHeader()->generate(adapter);
            else
    			lpVMMessage->generate(adapter);
		}
	}
	catch (vmime::exception& e) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	catch (std::exception& e) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
exit:
	if (lpszRawSMTP)
		delete [] lpszRawSMTP;

	if (lpTime)
		MAPIFreeBuffer(lpTime);

	if (lpMessageId)
		MAPIFreeBuffer(lpMessageId);

	if (lpVMMessage)
		delete lpVMMessage;
		
	delete mToVM;

	return hr;
}


// Read properties from lpMessage object and to internet rfc2822 format message
// then send it using the provided ECSender object
INETMAPI_API HRESULT IMToINet(IMAPISession *lpSession, IAddrBook *lpAddrBook, IMessage *lpMessage, ECSender *mailer_base, sending_options sopt, ECLogger *lpLogger)
{
	HRESULT			hr				= hrSuccess;
	MAPIToVMIME		*mToVM			= new MAPIToVMIME(lpSession, lpAddrBook, lpLogger, sopt);
	vmime::message	*lpVMMessage 	= NULL;
	ECVMIMESender	*mailer			= dynamic_cast<ECVMIMESender*>(mailer_base);
	char			*lpszRawSMTP	= NULL;

	if (!vphInited) {
		try {
			vphInited = vmime::platformDependant::getHandler();
		}
		catch (vmime::exceptions::no_platform_dependant_handler) {
			vmime::platformDependant::setHandler<vmime::platforms::posix::posixHandler>();
		};
	}

	hr = mToVM->convertMAPIToVMIME(lpMessage, &lpVMMessage, &lpszRawSMTP);

	if (hr != hrSuccess) {
		mailer->setError("Conversion error: "+stringify(hr,true));
		goto exit;
	}

	try {
		// vmime messageBuilder has set Date header to now()
		lpVMMessage->getHeader()->MessageId().setValue(vmime::messageId::generateId());
	}
	catch (vmime::exception& e) {
	    mailer->setError(e.what());
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	catch (std::exception& e) {
	    mailer->setError(e.what());
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	catch (...) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = mailer->sendMail(lpAddrBook, lpMessage, lpVMMessage, lpszRawSMTP);

exit:
	if (lpszRawSMTP)
		delete [] lpszRawSMTP;

	if (lpVMMessage)
		delete lpVMMessage;		// allocated by vmime::messageBuilder

	delete mToVM;

	return hr;
}
