/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "vconverter.h"
#include "valarm.h"
#include "icalrecurrence.h"
#include <mapi.h>
#include <mapiutil.h>
#include "mapiext.h"
#include "restrictionutil.h"
#include "CommonUtil.h"
#include "Util.h"
#include "icaluid.h"
#include "nameids.h"
#include "stringutil.h"
#include "time.h"

HRESULT HrCopyString(ECIConv *lpIConv, void *base, const char* lpszSrc, char** lppszDst)
{
	HRESULT hr = hrSuccess;
	char* lpszDst = NULL;
	std::string strText;

	if(!lpszSrc)
		strText = "";
	else if (lpIConv)
		strText = lpIConv->convert(lpszSrc);
	else
		strText = lpszSrc;

	hr = MAPIAllocateMore(strText.length()+1, base, (void**)&lpszDst);
	if (hr != hrSuccess)
		goto exit;

	strncpy(lpszDst, (char*)strText.c_str(), strText.length()+1);

	*lppszDst = lpszDst;

exit:
	return hr;
}

VConverter::VConverter(LPADRBOOK lpAdrBook, timezone_map *mapTimeZones, LPSPropTagArray lpNamedProps, ECIConv *lpIConv, bool blCensor, bool bNoRecipients, IMailUser *lpMailUser)
{
	m_lpAdrBook = lpAdrBook;
	m_mapTimeZones = mapTimeZones;
	m_iCurrentTimeZone = m_mapTimeZones->end();
	m_lpNamedProps = lpNamedProps;
	m_lpIConv = lpIConv;
	m_lpMailUser = lpMailUser;
	m_blCensorPrivate = blCensor;
	m_bNoRecipients = bNoRecipients;
	m_ulUserStatus = 0;
}

VConverter::~VConverter()
{
}

HRESULT VConverter::HrICal2MAPI(icalcomponent *lpEventRoot /* in */, icalcomponent *lpEvent /* in */, icalitem *lpPrevItem /* in */, icalitem **lppRet /* out */)
{
	HRESULT hr = hrSuccess;
	icalitem *lpIcalItem = NULL;
	icalproperty_method icMethod;
	icalproperty *lpicLastModified = NULL;
	icaltimetype icLastModifed;


	// we might be updating for exceptions
	if (*lppRet != NULL && lpPrevItem != NULL && lpPrevItem == *lppRet) {
		hr = HrAddException(lpEventRoot, lpEvent, lpPrevItem);
		if (hr == hrSuccess)
			goto exit;
	}

	lpIcalItem = new icalitem;
	MAPIAllocateBuffer(sizeof(void*), &lpIcalItem->base);
	lpIcalItem->lpRecurrence = NULL;

	// ---------------------------

	icMethod = icalcomponent_get_method(lpEventRoot);

	lpicLastModified = icalcomponent_get_first_property(lpEvent, ICAL_LASTMODIFIED_PROPERTY);
	if (lpicLastModified)
		icLastModifed = icalproperty_get_lastmodified(lpicLastModified);
	else
		icLastModifed = icaltime_null_time();
	// according to the RFC, LASTMODIFIED is always in UTC
	lpIcalItem->tLastModified = icaltime_as_timet(icLastModifed);

	// also sets strUid in icalitem struct
	hr = HrAddUids(lpEvent, lpIcalItem);
	if (hr != hrSuccess)
		goto exit;

	hr = HrAddMessageClass(icMethod, lpEvent, lpIcalItem, false, &lpIcalItem->lstMsgProps);
	if (hr != hrSuccess)
		goto exit;

	hr = HrAddStaticProps(icMethod, lpIcalItem);
	if (hr != hrSuccess)
		goto exit;

	hr = HrAddSimpleHeaders(lpEvent, lpIcalItem); // subject, location, ...
	if (hr != hrSuccess)
		goto exit;

	
	hr = HrAddXHeaders(lpEvent, lpIcalItem);
	if (hr != hrSuccess)
		goto exit;

	hr = HrAddCategories(lpEvent, lpIcalItem);
	if (hr != hrSuccess)
		goto exit;

	if (icMethod == ICAL_METHOD_REPLY)
		hr = HrAddReplyRecipients(lpEvent, lpIcalItem);
	else						// CANCEL, REQUEST, PUBLISH
		hr = HrAddRecipients(lpEvent, lpIcalItem, &lpIcalItem->lstMsgProps, &lpIcalItem->lstRecips);
	if (hr != hrSuccess)
		goto exit;
	
	hr = HrResolveUser(lpIcalItem, &(lpIcalItem->lstRecips));
	if (hr != hrSuccess)
		goto exit;

	hr = HrAddBusyStatus(lpEvent, icMethod, lpIcalItem);
	if (hr != hrSuccess)
		goto exit;
	
	// Important: m_iCurrentTimeZone will be set by this function, because of the possible recurrence lateron
	hr = HrAddTimes(lpEventRoot, lpEvent, lpIcalItem);
	if (hr != hrSuccess)
		goto exit;

	// Set reminder / alarm
	hr = HrAddReminder(lpEventRoot, lpEvent, lpIcalItem);
	if (hr != hrSuccess)
		goto exit;

	// Set recurrence.
	hr = HrAddRecurrence(lpEventRoot, lpEvent, lpIcalItem);
	if (hr != hrSuccess)
		goto exit;

	*lppRet = lpIcalItem;
	lpIcalItem = NULL;
	
exit:
	if (lpIcalItem) {
		MAPIFreeBuffer(lpIcalItem->base);
		delete lpIcalItem;
	}

	return hr;
}


HRESULT VConverter::HrGetUID(icalcomponent *lpEvent, std::string *strUid)
{
	HRESULT hr = hrSuccess;
	icalproperty *icProp = NULL;
	const char *uid = NULL;

	icProp = icalcomponent_get_first_property(lpEvent, ICAL_UID_PROPERTY);
	if (!icProp) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	uid = icalproperty_get_uid(icProp);

	if (!uid || strcmp(uid,"") == 0) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	*strUid = uid;

exit:
	return hr;
}

/* UID contents according to [MS-OXCICAL].pdf

   UID = EncodedGlobalId / ThirdPartyGlobalId

   EncodedGlobalId    = Header GlobalIdData
   ThirdPartyGlobalId = 1*UTF8-octets    ; Assuming UTF-8 is the encoding

   Header = ByteArrayID InstanceDate CreationDateTime Padding DataSize

   ByteArrayID        = "040000008200E00074C5B7101A82E008"
   InstanceDate       = InstanceYear InstanceMonth InstanceDay
   InstanceYear       = 4*4HEXDIGIT      ; UInt16
   InstanceMonth      = 2*2HEXDIGIT      ; UInt8
   InstanceDay        = 2*2HEXDIGIT      ; UInt8
   CreationDateTime   = FileTime
   FileTime           = 16*16HEXDIGIT    ; UInt6
   Padding            = 16*16HEXDIGIT    ; "0000000000000000" recommended
   DataSize           = 8*8HEXDIGIT      ; UInt32 little-endian
   GlobalIdData       = 2*HEXDIGIT4
*/
HRESULT VConverter::HrMakeBinaryUID(std::string strUid, void *base, SPropValue *lpPropValue)
{
	SPropValue sPropValue;
	std::string strBinUid;
	std::string strByteArrayID = "040000008200E00074C5B7101A82E008";

	// Check whether this is a default Outlook UID
	// Exchange example: UID:040000008200E00074C5B7101A82E008 00000000 305D0F2A9A06C901 0000000000000000 10000000 7F64D28AE2DCC64C88F849733F5FBD1D
	// GMail example:    UID:rblkvqecgurvb0all6rjb3d1j8@google.com
	// Sunbird example: UID:1090c3de-36b2-4352-a155-a1436bc806b8
	if (strUid.compare(0, strByteArrayID.length(), strByteArrayID) == 0) {
		// EncodedGlobalId
		Util::hex2bin(strUid, &strBinUid);
	} else {
		// ThirdPartyGlobalId
		HrMakeBinUidFromICalUid(strUid, &strBinUid);
	}

	// Caller sets .ulPropTag
	sPropValue.Value.bin.cb = strBinUid.size();
	MAPIAllocateMore(sPropValue.Value.bin.cb, base, (void**)&sPropValue.Value.bin.lpb);
	memcpy(sPropValue.Value.bin.lpb, strBinUid.data(), sPropValue.Value.bin.cb);

	// set return value
	*lpPropValue = sPropValue;

	return hrSuccess;
}
bool VConverter::bIsUserLoggedIn(std::string strUser)

{
	HRESULT hr = hrSuccess;
	LPSPropValue lpUsrProp = NULL;
	bool blRetVal = false;
	
	if(m_lpMailUser)
		hr = HrGetOneProp(m_lpMailUser, PR_SMTP_ADDRESS, &lpUsrProp);
	else
		hr = MAPI_E_CALL_FAILED;

	if (hr != hrSuccess)
		goto exit;
	if(!strncmp(lpUsrProp->Value.lpszA, strUser.c_str() , strUser.length()))
		blRetVal = true;
	
exit:
	if (lpUsrProp)
		MAPIFreeBuffer(lpUsrProp);

	return blRetVal;
}

HRESULT VConverter::HrResolveUser(icalitem *lpIcalItem , std::list<icalrecip> *lplstIcalRecip)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpUsrEidProp = NULL; 
	LPSPropValue lpMappedProp = NULL;
	LPADRLIST lpAdrList	= NULL;	
	IABContainer *lpAddrFolder = NULL;
	FlagList *lpFlagList = NULL;
	std::list<icalrecip>::iterator iIcalRecip;
	icalrecip icalRecipient;
	ULONG ulRecpCnt = 0;
	ULONG ulRetn = 0;
	ULONG ulObjType = 0;
	ULONG cbEID = 0;
	LPENTRYID lpEID = NULL;

	if (lplstIcalRecip->empty())
		goto exit;
	
	// ignore error
	 if(m_lpMailUser)
		HrGetOneProp(m_lpMailUser, PR_ENTRYID, &lpUsrEidProp);
	
	 hr = m_lpAdrBook->OpenEntry(0,NULL,0,0,&ulObjType,(LPUNKNOWN*)&lpAddrFolder);
	if(hr != hrSuccess)
		goto exit;

	ulRecpCnt = lplstIcalRecip->size();

	hr = MAPIAllocateBuffer(CbNewFlagList(ulRecpCnt), (void **) &lpFlagList);
	if (hr != hrSuccess)
		goto exit;

	lpFlagList->cFlags = ulRecpCnt;
	
	hr = MAPIAllocateBuffer(CbNewSRowSet(ulRecpCnt), (void **) &lpAdrList);
	if (hr != hrSuccess)
		goto exit;

	lpAdrList->cEntries = ulRecpCnt;

	for (iIcalRecip = lplstIcalRecip->begin(), ulRecpCnt = 0; iIcalRecip != lplstIcalRecip->end(); iIcalRecip++, ulRecpCnt++) {
		lpAdrList->aEntries[ulRecpCnt].cValues = 1;

		hr = MAPIAllocateBuffer(sizeof(SPropValue), (void **) &lpAdrList->aEntries[ulRecpCnt].rgPropVals);
		if (hr != hrSuccess)
			goto exit;

		lpAdrList->aEntries[ulRecpCnt].rgPropVals[0].ulPropTag = PR_DISPLAY_NAME;
		lpAdrList->aEntries[ulRecpCnt].rgPropVals[0].Value.lpszA = (char *)iIcalRecip->strEmail.c_str();
		lpFlagList->ulFlag[ulRecpCnt] = MAPI_UNRESOLVED;
	}

	hr = lpAddrFolder->ResolveNames(NULL, 0, lpAdrList, lpFlagList);
	if (hr != hrSuccess)
		goto exit;

	//reset the recepients with mapped names
	for(icalRecipient = lplstIcalRecip->front(), ulRecpCnt = 0 ; ulRecpCnt < lplstIcalRecip->size(); ulRecpCnt++)
	{
		if (lpFlagList->ulFlag[ulRecpCnt] == MAPI_RESOLVED)
		{
			lpMappedProp = PpropFindProp(lpAdrList->aEntries[ulRecpCnt].rgPropVals, lpAdrList->aEntries[ulRecpCnt].cValues, PR_DISPLAY_NAME);
			if (lpMappedProp)
				icalRecipient.strName = lpMappedProp->Value.lpszA;
		}
		
		//save the logged in user's satus , used in setting FB status  
		lpMappedProp = PpropFindProp(lpAdrList->aEntries[ulRecpCnt].rgPropVals, lpAdrList->aEntries[ulRecpCnt].cValues, PR_ENTRYID);
		if (lpMappedProp && lpUsrEidProp)
			hr = m_lpAdrBook->CompareEntryIDs(lpUsrEidProp->Value.bin.cb, (LPENTRYID)lpUsrEidProp->Value.bin.lpb, lpMappedProp->Value.bin.cb, (LPENTRYID)lpMappedProp->Value.bin.lpb , 0 , &ulRetn);
		if (hr == hrSuccess && ulRetn == true)
			m_ulUserStatus = icalRecipient.ulTrackStatus;

		//Create EntryID by using mapped names, ical data might not have names.
		hr = ECCreateOneOff((LPSTR)icalRecipient.strName.c_str(), "SMTP", (LPSTR)icalRecipient.strEmail.c_str(), 0, &cbEID, &lpEID);
		if (hr == hrSuccess) {
			// realloc on lpIcalItem
			hr = MAPIAllocateMore(cbEID, lpIcalItem->base, (void**)&icalRecipient.lpEntryID);
			if (hr != hrSuccess)
				goto exit;

			icalRecipient.cbEntryID = cbEID;
			memcpy(icalRecipient.lpEntryID, lpEID, cbEID);
			
			MAPIFreeBuffer(lpEID);
			lpEID = NULL;
		}
		
		lplstIcalRecip->push_back(icalRecipient);
		lplstIcalRecip->pop_front();
		icalRecipient = lplstIcalRecip->front();

	}

exit:
	if (lpUsrEidProp)
		MAPIFreeBuffer(lpUsrEidProp);

	if (lpFlagList)
		MAPIFreeBuffer(lpFlagList);

	if (lpAdrList)
		FreeProws((LPSRowSet)lpAdrList);

	if (lpAddrFolder)
		lpAddrFolder->Release();

	return hr;
}

HRESULT VConverter::HrCompareUids(icalitem *lpIcalItem, icalcomponent *lpicEvent)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpPropVal = NULL;
	std::string strUid;
	int res;

	hr = HrGetUID(lpicEvent, &strUid);
	if (hr != hrSuccess)
		goto exit;

	MAPIAllocateBuffer(sizeof(SPropValue), (void**)&lpPropVal);

	hr = HrMakeBinaryUID(strUid, lpPropVal, lpPropVal);
	if (hr != hrSuccess)
		goto exit;

	lpPropVal->ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY);

	hr = Util::CompareProp(lpPropVal, &lpIcalItem->sBinGuid, &res);
	if (!(hr == hrSuccess && res == 0))
		hr = MAPI_E_BAD_VALUE;

exit:
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	return hr;
}

HRESULT VConverter::HrAddUids(icalcomponent *lpicEvent, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropValue;
	std::string strUid;

	// Get Unique ID of ical item, or create new
	hr = HrGetUID(lpicEvent, &strUid);
	if (hr != hrSuccess)
		hr = HrGenerateUid(&strUid);
	if (hr != hrSuccess)
		goto exit;

	hr = HrMakeBinaryUID(strUid, lpIcalItem->base, &sPropValue);
	if (hr != hrSuccess)
		goto exit;

	// set as dispidMeetingRecurrenceId ...
	sPropValue.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY);
	lpIcalItem->lstMsgProps.push_back(sPropValue);
	
	// replace date in GUID, the dispidGlobalObjectID should be same for exception and reccurnece message.
	// used for exceptions in outlook
	if(IsOutlookUid(strUid))
		strUid.replace(32, 8, "00000000");

	hr = HrMakeBinaryUID(strUid, lpIcalItem->base, &sPropValue);
	if (hr != hrSuccess)
		goto exit;

	// set as dispidGlobalObjectID...
	sPropValue.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RID], PT_BINARY);
	lpIcalItem->lstMsgProps.push_back(sPropValue);
	// save the strUid to lookup for occurrences
	lpIcalItem->sBinGuid = sPropValue;

exit:
	return hr;
}

HRESULT VConverter::HrAddStaticProps(icalproperty_method icMethod, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_SIDEEFFECT], PT_LONG);
	if(icMethod == ICAL_METHOD_REPLY || icMethod == ICAL_METHOD_REQUEST)
		sPropVal.Value.ul = 0x1c61; // seOpenToDelete | seOpenToCopy | seOpenToMove | seCannotUndoDelete | seCannotUndoCopy | seCannotUndoMove
	else
		sPropVal.Value.ul = 0x171;	// seOpenToDelete | seCoerceToInbox | seOpenToCopy | seOpenToMove | seOpenForCtxMenu 
	
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_SENDASICAL], PT_BOOLEAN);
	sPropVal.Value.b = 1;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// Needed for deleting an occurrence of a recurring item in outlook
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_COMMONASSIGN], PT_LONG);
	sPropVal.Value.ul = 0;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	return hr;
}

HRESULT VConverter::HrAddSimpleHeaders(icalcomponent *lpicEvent, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;
	icalproperty *lpicProp = NULL;
	int lPriority;
	int lClass = 0;
	std::string strClass;

	// Set subject / SUMMARY
	sPropVal.ulPropTag = PR_SUBJECT;
	HrCopyString(m_lpIConv, lpIcalItem->base, icalcomponent_get_summary(lpicEvent), &sPropVal.Value.lpszA);
	if (sPropVal.Value.lpszA == NULL)
		sPropVal.Value.lpszA = "";
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// Set body / DESCRIPTION
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_DESCRIPTION_PROPERTY);
	if (lpicProp){
		sPropVal.ulPropTag = PR_BODY;
		HrCopyString(m_lpIConv, lpIcalItem->base, icalproperty_get_description(lpicProp), &sPropVal.Value.lpszA);
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	} else {
		lpIcalItem->lstDelPropTags.push_back(PR_BODY);
	}

	// Set location / LOCATION
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_LOCATION_PROPERTY);
	if (lpicProp) {
		HrCopyString(m_lpIConv, lpIcalItem->base, icalproperty_get_location(lpicProp), &sPropVal.Value.lpszA);

		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_LOCATION], PT_STRING8);
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MEETINGLOCATION], PT_STRING8);
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	} else {
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_LOCATION], PT_STRING8));
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MEETINGLOCATION], PT_STRING8));
	}

	// Set importance and priority / PRIORITY
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_PRIORITY_PROPERTY);
	if (lpicProp) {
		lPriority = icalproperty_get_priority(lpicProp);
		// TODO: test input and output!
		if (lPriority == 0) {
		} else if (lPriority < 5) {
			lPriority = 1;
		} else if (lPriority > 5) {
			lPriority = -1;
		} else {
			lPriority = 0;
		}
		
		sPropVal.ulPropTag = PR_IMPORTANCE;
		sPropVal.Value.ul = lPriority + 1;
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		sPropVal.ulPropTag = PR_PRIORITY;
		sPropVal.Value.l = lPriority;
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	} else {
		lpIcalItem->lstDelPropTags.push_back(PR_IMPORTANCE);
		lpIcalItem->lstDelPropTags.push_back(PR_PRIORITY);
	}
	
	// Private
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_CLASS_PROPERTY);
	if (lpicProp){
		lClass = icalproperty_get_class(lpicProp);
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN);
		sPropVal.Value.b = (lClass == ICAL_CLASS_PRIVATE);
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	}
	// Sensitivity, from same class property
	sPropVal.ulPropTag = PR_SENSITIVITY;
	if (lClass == ICAL_CLASS_PRIVATE)
		sPropVal.Value.ul = 2; // Private
	else if (lClass == ICAL_CLASS_CONFIDENTIAL)
		sPropVal.Value.ul = 3; // CompanyConfidential
	else
		sPropVal.Value.ul = 0; // Public
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	return hr;
}

HRESULT VConverter::HrAddBusyStatus(icalcomponent *lpicEvent, icalproperty_method icMethod, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;
	icalproperty* lpicProp = NULL;
	std::list<icalrecip>::iterator iIcalRecip;	
	ULONG ulXMStatus = -1; // as 0-3 states some status

	// default: busy
	// 0: free
	// 1: tentative
	// 2: busy
	// 3: oof
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_BUSYSTATUS], PT_LONG);
   	sPropVal.Value.ul = 2;
	
	// caldav clients only uses the TRASP property to set FreeBusy
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_TRANSP_PROPERTY);
	if (lpicProp) {
		switch (icalproperty_get_transp(lpicProp)) {
		case ICAL_TRANSP_TRANSPARENT: // free
		case ICAL_TRANSP_TRANSPARENTNOCONFLICT:
			sPropVal.Value.ul = 0;
			break;
		case ICAL_TRANSP_X:
		case ICAL_TRANSP_OPAQUE: // busy
		case ICAL_TRANSP_OPAQUENOCONFLICT:
		case ICAL_TRANSP_NONE:
        	sPropVal.Value.ul = 2;
			break;
		}
	}
	
	// Only process for Meeting Req from dagent
	if(m_bNoRecipients)
	{
	    // Meeting requests always have a BusyStatus of 1 (tentative), since this is the status of
	    // the meeting which will be placed in your calendar when it has been processed but not accepted
	    // The busy status of meeting responses is less important but seems to be 2 (Busy) in Outlook.
		if (icMethod == ICAL_METHOD_REQUEST)
			sPropVal.Value.ul = 1;
	}
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// X-MICROSOFT-CDO-BUSYSTATUS is used to set IntendedFreeBusyStatus
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_X_PROPERTY);
	while (lpicProp) {
		// X-MICROSOFT-CDO-BUSYSTATUS:FREE
		if (strcmp(icalproperty_get_x_name(lpicProp), "X-MICROSOFT-CDO-BUSYSTATUS") == 0) {
			const char *lpVal = icalproperty_get_x(lpicProp);
			if (strcmp(lpVal, "FREE") == 0)
				ulXMStatus = 0;
			else if (strcmp(lpVal, "TENTATIVE") == 0)
				ulXMStatus = 1;
			else if(strcmp(lpVal, "BUSY") == 0)
				ulXMStatus = 2;
			else if (strcmp(lpVal, "OOF") == 0)
				ulXMStatus = 3;
			 
		}
		lpicProp = icalcomponent_get_next_property(lpicEvent, ICAL_X_PROPERTY);
	}
	
	// only set for REQUEST
	if (ulXMStatus != -1 && icMethod == ICAL_METHOD_REQUEST)
	{
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_INTENDEDBUSYSTATUS], PT_LONG);
		sPropVal.Value.ul = ulXMStatus;
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	}

	return hr;
}

HRESULT VConverter::HrAddXHeaders(icalcomponent *lpicEvent, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;
	icalproperty* lpicProp = NULL;
	icalvalue *lpicValue = NULL;	
	time_t ttCritcalChange = 0;
	ULONG ulApptSeqNo = 0;
	ULONG ulMozGen = 0;
	ULONG ulSequence = 0;
	ULONG ulMaxCounter = 0;
	bool bOwnerCritical = false;
	bool bAttendeeCritical = false;
	bool bOwnerApptID = false;
	bool bMozGen = false;
	
	// TODO: maybe save/restore headers to get "original" ical again?
	
	// add X-MICROSOFT-CDO & X-MOZ properties
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_X_PROPERTY);	
	while (lpicProp) {
		if (strcmp(icalproperty_get_x_name(lpicProp), "X-MICROSOFT-CDO-ATTENDEE-CRITICAL-CHANGE") == 0){

			lpicValue = icalvalue_new_from_string(ICAL_DATETIME_VALUE, icalproperty_get_x(lpicProp));
			ttCritcalChange = icaltime_as_timet_with_zone(icalvalue_get_datetime(lpicValue), NULL); // no timezone
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_ATTENDEECRITICALCHANGE], PT_SYSTIME);
			UnixTimeToFileTime(ttCritcalChange, &sPropVal.Value.ft);
			lpIcalItem->lstMsgProps.push_back(sPropVal);
			bAttendeeCritical = true;

			if (lpicValue)
				icalvalue_free(lpicValue);

		}else if (strcmp(icalproperty_get_x_name(lpicProp), "X-MICROSOFT-CDO-OWNER-CRITICAL-CHANGE") == 0){
			
			lpicValue = icalvalue_new_from_string(ICAL_DATETIME_VALUE, icalproperty_get_x(lpicProp));
			ttCritcalChange = icaltime_as_timet_with_zone(icalvalue_get_datetime(lpicValue), NULL); // no timezone
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_OWNERCRITICALCHANGE], PT_SYSTIME);
			UnixTimeToFileTime(ttCritcalChange, &sPropVal.Value.ft);
			lpIcalItem->lstMsgProps.push_back(sPropVal);
			bOwnerCritical = true;

			if (lpicValue)
				icalvalue_free(lpicValue);

		}else if (strcmp(icalproperty_get_x_name(lpicProp), "X-MICROSOFT-CDO-OWNERAPPTID") == 0){
			
			lpicValue = icalvalue_new_from_string(ICAL_INTEGER_VALUE, icalproperty_get_x(lpicProp));
			sPropVal.ulPropTag = PR_OWNER_APPT_ID;
			sPropVal.Value.ul = icalvalue_get_integer(lpicValue);
			lpIcalItem->lstMsgProps.push_back(sPropVal);
			bOwnerApptID = true;

			if (lpicValue)
				icalvalue_free(lpicValue);

		}else if (strcmp(icalproperty_get_x_name(lpicProp), "X-MICROSOFT-CDO-APPT-SEQUENCE") == 0){
			
			lpicValue = icalvalue_new_from_string(ICAL_INTEGER_VALUE, icalproperty_get_x(lpicProp));
			ulApptSeqNo = icalvalue_get_integer(lpicValue) ;

			if (lpicValue)
				icalvalue_free(lpicValue);

		}else if (strncmp(icalproperty_get_x_name(lpicProp), "X-MOZ-GENERATION", strlen("X-MOZ-GENERATION")) == 0) {

			lpicValue = icalvalue_new_from_string(ICAL_INTEGER_VALUE, icalproperty_get_x(lpicProp));
			ulMozGen = icalvalue_get_integer(lpicValue);
			bMozGen = true;

			if (lpicValue)
				icalvalue_free(lpicValue);

		}else if (strncmp(icalproperty_get_x_name(lpicProp), "X-MOZ-SEND-INVITATIONS", strlen("X-MOZ-SEND-INVITATIONS")) == 0) {

			lpicValue =  icalvalue_new_from_string(ICAL_X_VALUE, icalproperty_get_x(lpicProp));
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZSENDINVITE], PT_BOOLEAN);
			sPropVal.Value.b = (strncmp(icalvalue_get_x(lpicValue), "TRUE", strlen("TRUE"))? 0 : 1);
			lpIcalItem->lstMsgProps.push_back(sPropVal);
			
			if (lpicValue)
				icalvalue_free(lpicValue);
		}

		lpicProp = icalcomponent_get_next_property(lpicEvent, ICAL_X_PROPERTY);
	}
	
	ulSequence = icalcomponent_get_sequence(lpicEvent);
	ulMaxCounter = (ulApptSeqNo > ulMozGen ? (ulApptSeqNo > ulSequence?ulApptSeqNo : ulSequence ): (ulMozGen > ulSequence ? ulMozGen: ulSequence ) );

	// ApptSequenceNo
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTSEQNR], PT_LONG);
	sPropVal.Value.ul = ulMaxCounter;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	if (!bAttendeeCritical) {
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_ATTENDEECRITICALCHANGE], PT_SYSTIME);
		UnixTimeToFileTime(time(NULL), &sPropVal.Value.ft);
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	}
	
	if (!bOwnerCritical) {
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_OWNERCRITICALCHANGE], PT_SYSTIME);
		UnixTimeToFileTime(time(NULL), &sPropVal.Value.ft);
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	}

	if (!bOwnerApptID) {
		sPropVal.ulPropTag = PR_OWNER_APPT_ID;
		sPropVal.Value.ul = -1;
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	}

	if (bMozGen) {
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZGEN], PT_LONG);
		sPropVal.Value.ul = ulMaxCounter;
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	}

	return hr;
}

HRESULT VConverter::HrAddCategories(icalcomponent *lpicEvent, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;
	icalproperty *lpicProp = NULL;
	const char* lpszCategories = NULL;
	std::vector<std::string> vCategories;
	std::vector<std::string>::iterator iCats;
	int i;

	// Set keywords / CATEGORIES
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_CATEGORIES_PROPERTY);
	if (!lpicProp) {
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_KEYWORDS], PT_MV_STRING8));
		goto exit;
	}

	while (lpicProp != NULL && (lpszCategories = icalproperty_get_categories(lpicProp)) != NULL) {
		vCategories.push_back(lpszCategories);
		lpicProp = icalcomponent_get_next_property(lpicEvent, ICAL_CATEGORIES_PROPERTY);
	}

	hr = MAPIAllocateMore(vCategories.size() * sizeof(LPSTR), lpIcalItem->base, (void**)&sPropVal.Value.MVszA.lppszA);
	if (hr != hrSuccess)
		goto exit;

	for (i = 0, iCats = vCategories.begin(); iCats != vCategories.end(); iCats++, i++) {
		int length = iCats->length() + 1;
		hr = MAPIAllocateMore(length, lpIcalItem->base, (void **) &sPropVal.Value.MVszA.lppszA[i]);
		if (hr != hrSuccess)
			goto exit;

		memcpy((void*)sPropVal.Value.MVszA.lppszA[i], iCats->c_str(), length);
	}

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_KEYWORDS], PT_MV_STRING8);
	sPropVal.Value.MVszA.cValues = vCategories.size();
	lpIcalItem->lstMsgProps.push_back(sPropVal);

exit:
	return hr;
}

HRESULT VConverter::HrAddRecipients(icalcomponent *lpicEvent, icalitem *lpIcalItem,std::list<SPropValue> *lplstMsgProps, std::list<icalrecip> *lplstIcalRecip)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;
	std::string strEmail, strName, strSearchKey;
	icalproperty *lpicProp = NULL;
	icalparameter *lpicParam = NULL;
	icalrecip icrAttendee;
	ULONG cbEntryID;
	LPENTRYID lpEntryID = NULL;
	LPSPropValue lpsPropVal = NULL;

	// exchange organizer CN="John van der Kamp":MAILTO:john@exchange.ztest.nl
	// google organizer: CN=Malcolm Yates:mailto:mdy@canonical.com
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_ORGANIZER_PROPERTY);
	if (lpicProp) {
		strEmail = icalproperty_get_organizer(lpicProp);
		if (stricmp(strEmail.substr(0,7).c_str(), "mailto:") == 0)
			strEmail = strEmail.substr(7);

		lpicParam = icalproperty_get_first_parameter(lpicProp, ICAL_CN_PARAMETER);
		if (lpicParam != NULL)
			strName = icalparameter_get_cn(lpicParam);
		else
			strName = strEmail; // set email as name OL does not display organiser name if not set.

		if (bIsUserLoggedIn(strEmail)) {
			hr = HrGetOneProp(m_lpMailUser, PR_ENTRYID, &lpsPropVal);
			if (hr != hrSuccess)
				goto exit;

			// Warning: watch out for nice free code at the end
			cbEntryID = lpsPropVal->Value.bin.cb;
			lpEntryID = (LPENTRYID)lpsPropVal->Value.bin.lpb;
		} else {
			hr = ECCreateOneOff((LPSTR)m_lpIConv->convert(strName).c_str(), "SMTP", (LPSTR)m_lpIConv->convert(strEmail).c_str(), 0, &cbEntryID, &lpEntryID);
			if (hr != hrSuccess)
				goto exit;
		}

		// add the organiser to the recipient list
		MAPIAllocateMore(cbEntryID, lpIcalItem->base, (void**)&icrAttendee.lpEntryID);
		
		memcpy(icrAttendee.lpEntryID, lpEntryID, cbEntryID);
		icrAttendee.cbEntryID = cbEntryID;

		icrAttendee.strEmail = strEmail;
		icrAttendee.strName = strName;
		icrAttendee.ulRecipientType = MAPI_ORIG;
		icrAttendee.ulTrackStatus = 0;
	
		lplstIcalRecip->push_back(icrAttendee);

		// The DAgent does not want these properties from ical, since it writes them itself
		if (!m_bNoRecipients) {

			strSearchKey = "SMTP:"+strEmail;
			transform(strSearchKey.begin(), strSearchKey.end(), strSearchKey.begin(), ::toupper);

			sPropVal.ulPropTag = PR_SENDER_ADDRTYPE;
			HrCopyString(NULL, lpIcalItem->base, "SMTP", &sPropVal.Value.lpszA);
			lplstMsgProps->push_back(sPropVal);

			sPropVal.ulPropTag = PR_SENT_REPRESENTING_ADDRTYPE;
			lplstMsgProps->push_back(sPropVal);

			sPropVal.ulPropTag = PR_SENDER_EMAIL_ADDRESS;
			HrCopyString(m_lpIConv, lpIcalItem->base, strEmail.c_str(), &sPropVal.Value.lpszA);
			lplstMsgProps->push_back(sPropVal);

			sPropVal.ulPropTag = PR_SENT_REPRESENTING_EMAIL_ADDRESS;
			lplstMsgProps->push_back(sPropVal);

			sPropVal.ulPropTag = PR_SENDER_NAME;
			HrCopyString(m_lpIConv, lpIcalItem->base, strName.c_str(), &sPropVal.Value.lpszA);
			lplstMsgProps->push_back(sPropVal);

			sPropVal.ulPropTag = PR_SENT_REPRESENTING_NAME;
			lplstMsgProps->push_back(sPropVal);

			sPropVal.ulPropTag = PR_SENDER_SEARCH_KEY;
			HrCopyString(m_lpIConv, lpIcalItem->base, strSearchKey.c_str(), (char**)&sPropVal.Value.bin.lpb);
			sPropVal.Value.bin.cb = strSearchKey.length() + 1;
			lplstMsgProps->push_back(sPropVal);

			sPropVal.ulPropTag = PR_SENT_REPRESENTING_SEARCH_KEY;
			lplstMsgProps->push_back(sPropVal);

			// re-allocate memory to list with lpIcalItem
			MAPIAllocateMore(cbEntryID, lpIcalItem->base, (void**)&sPropVal.Value.bin.lpb);
			sPropVal.Value.bin.cb = cbEntryID;
			memcpy(sPropVal.Value.bin.lpb, lpEntryID, cbEntryID);

			sPropVal.ulPropTag = PR_SENDER_ENTRYID;
			lplstMsgProps->push_back(sPropVal);

			sPropVal.ulPropTag = PR_SENT_REPRESENTING_ENTRYID;
			lplstMsgProps->push_back(sPropVal);
		}
	}

	for (lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_ATTENDEE_PROPERTY);
		 lpicProp != NULL;
		 lpicProp = icalcomponent_get_next_property(lpicEvent, ICAL_ATTENDEE_PROPERTY))
	{
		icrAttendee.strEmail = icalproperty_get_attendee(lpicProp);
		if (stricmp(icrAttendee.strEmail.substr(0,7).c_str(), "mailto:") == 0) {
			icrAttendee.strEmail = icrAttendee.strEmail.substr(7);
		}

		// TODO: Add organiser details if requeried.
		if(icrAttendee.strEmail == strEmail)// remove organiser from attendee list.
			continue;
		
		icrAttendee.strEmail = m_lpIConv->convert(icrAttendee.strEmail);

		lpicParam = icalproperty_get_first_parameter(lpicProp, ICAL_CN_PARAMETER);
		if (lpicParam)
			icrAttendee.strName = m_lpIConv->convert(icalparameter_get_cn(lpicParam));
		else
			icrAttendee.strName = icrAttendee.strEmail;

		lpicParam = icalproperty_get_first_parameter(lpicProp, ICAL_ROLE_PARAMETER);
		if (!lpicParam) {
			icrAttendee.ulRecipientType = MAPI_TO;
		} else {
			switch (icalparameter_get_role(lpicParam)) {
			case ICAL_ROLE_OPTPARTICIPANT:
				icrAttendee.ulRecipientType = MAPI_CC;
				break;
			case ICAL_ROLE_NONPARTICIPANT:
				icrAttendee.ulRecipientType = MAPI_BCC;
				break;
			case ICAL_ROLE_REQPARTICIPANT:
			default:
				icrAttendee.ulRecipientType = MAPI_TO;
				break;
			}
		}

		lpicParam = icalproperty_get_first_parameter(lpicProp, ICAL_PARTSTAT_PARAMETER);
		if (lpicParam) {
			switch (icalparameter_get_partstat(lpicParam)) {
			case ICAL_PARTSTAT_TENTATIVE:
				icrAttendee.ulTrackStatus = 2;
				break;
			case ICAL_PARTSTAT_ACCEPTED:
				icrAttendee.ulTrackStatus = 3;
				break;
			case ICAL_PARTSTAT_DECLINED:
				icrAttendee.ulTrackStatus = 4;
				break;
			case ICAL_PARTSTAT_NEEDSACTION:
			default:
				icrAttendee.ulTrackStatus = 0;
				break;
			}
		}
		
		lplstIcalRecip->push_back(icrAttendee);
	}

exit:
	// because of the shortcut, it's either, not both
	if (lpsPropVal)
		MAPIFreeBuffer(lpsPropVal);
	else if (lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	return hr;
}

HRESULT VConverter::HrAddReplyRecipients(icalcomponent *lpicEvent, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;
	string strEmail, strName, strSearchKey;
	icalproperty *lpicProp = NULL;
	icalparameter *lpicParam = NULL;
	icalrecip icrAttendee;
	ULONG cbEntryID;
	LPENTRYID lpEntryID = NULL;

	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_ORGANIZER_PROPERTY);
	if (lpicProp) {
		icrAttendee.strEmail = icalproperty_get_organizer(lpicProp);
		if (stricmp(icrAttendee.strEmail.substr(0,7).c_str(), "mailto:") == 0)
			icrAttendee.strEmail = icrAttendee.strEmail.substr(7);

		lpicParam = icalproperty_get_first_parameter(lpicProp, ICAL_CN_PARAMETER);
		if (lpicParam != NULL)
			icrAttendee.strName = icalparameter_get_cn(lpicParam);

		icrAttendee.ulRecipientType = MAPI_TO;
		lpIcalItem->lstRecips.push_back(icrAttendee);
	}

	// The DAgent does not want these properties from ical, since it writes them itself
	if (!m_bNoRecipients) {
		// TODO: what if >1 attendee ?!?

		//PR_SENDER = ATTENDEE
		lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_ATTENDEE_PROPERTY);
		if (lpicProp) {
			strEmail = icalproperty_get_attendee(lpicProp);
			if (stricmp(strEmail.substr(0,7).c_str(), "mailto:") == 0) {
				strEmail = strEmail.substr(7);
			}

			lpicParam = icalproperty_get_first_parameter(lpicProp, ICAL_CN_PARAMETER);
			if (lpicParam)
				strName = icalparameter_get_cn(lpicParam);

			strSearchKey = "SMTP:"+strEmail;
			transform(strSearchKey.begin(), strSearchKey.end(), strSearchKey.begin(), ::toupper);
		}

		sPropVal.ulPropTag = PR_SENDER_ADDRTYPE;
		HrCopyString(NULL, lpIcalItem->base, "SMTP", &sPropVal.Value.lpszA);
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		sPropVal.ulPropTag = PR_SENT_REPRESENTING_ADDRTYPE;
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		sPropVal.ulPropTag = PR_SENDER_EMAIL_ADDRESS;
		HrCopyString(NULL, lpIcalItem->base, (LPSTR)strEmail.c_str(), &sPropVal.Value.lpszA);
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		sPropVal.ulPropTag = PR_SENT_REPRESENTING_EMAIL_ADDRESS;
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		sPropVal.ulPropTag = PR_SENDER_NAME;
		HrCopyString(m_lpIConv, lpIcalItem->base, (LPSTR)strName.c_str(), &sPropVal.Value.lpszA);
		lpIcalItem->lstMsgProps.push_back(sPropVal);
		
		sPropVal.ulPropTag = PR_SENT_REPRESENTING_NAME;
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		sPropVal.ulPropTag = PR_SENDER_SEARCH_KEY;
		HrCopyString(m_lpIConv, lpIcalItem->base, strSearchKey.c_str(), (char**)&sPropVal.Value.bin.lpb);
		sPropVal.Value.bin.cb = strSearchKey.length() + 1;
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		sPropVal.ulPropTag = PR_SENT_REPRESENTING_SEARCH_KEY;
		lpIcalItem->lstMsgProps.push_back(sPropVal);


		hr = ECCreateOneOff((LPSTR)strName.c_str(), "SMTP", (LPSTR)strEmail.c_str(), 0, &cbEntryID, &lpEntryID);
		if (hr != hrSuccess)
			goto exit;

		// re-allocate memory to list with lpIcalItem
		MAPIAllocateMore(cbEntryID, lpIcalItem->base, (void**)&sPropVal.Value.bin.lpb);
		sPropVal.Value.bin.cb = cbEntryID;
		memcpy(sPropVal.Value.bin.lpb, lpEntryID, cbEntryID);

		sPropVal.ulPropTag = PR_SENDER_ENTRYID;
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		sPropVal.ulPropTag = PR_SENT_REPRESENTING_ENTRYID;
		lpIcalItem->lstMsgProps.push_back(sPropVal);
	}

exit:
	if (lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	return hr;
}

HRESULT VConverter::HrAddReminder(icalcomponent *lpicEventRoot, icalcomponent *lpicEvent, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;
	SPropValue sPropMozAck;
	icalcomponent *lpicAlarm = NULL;
	LONG ulRemindBefore = 0;
	time_t ttReminderTime = 0;
	time_t ttReminderNext = 0;
	time_t ttMozLastAck = 0;
	time_t ttMozLastAckMax = 0;
	bool bReminderSet = false;
	bool bHasMozAck = false;
	icalproperty* lpicDTStartProp = NULL;
	icalproperty* lpicProp = NULL;
	icalvalue *lpicValue = NULL;
	std::string strSuffix;

	lpicAlarm = icalcomponent_get_first_component(lpicEvent, ICAL_VALARM_COMPONENT);
	if (lpicAlarm == NULL) {
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERSET], PT_BOOLEAN);
		sPropVal.Value.b = false;
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERTIME], PT_SYSTIME));
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERNEXTTIME], PT_SYSTIME));
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERMINUTESBEFORESTART], PT_LONG));

		goto exit; // No alarms found, so we can safely exit here.
	}

	hr = HrParseVAlarm(lpicAlarm, &ulRemindBefore, &ttReminderTime, &bReminderSet);
	if (hr != hrSuccess) {
		// just skip the reminder
		hr = hrSuccess;
		goto exit;
	}

	// Handle Sunbird's dismiss/snooze, see: https://wiki.mozilla.org/Calendar:Feature_Implementations:Alarms
	// X-MOZ-SNOOZE-TIME-1231250400000000:20090107T132846Z
	// X-MOZ-LASTACK:20090107T132846Z
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_X_PROPERTY);
	while (lpicProp) {
		if (strcmp(icalproperty_get_x_name(lpicProp), "X-MOZ-LASTACK") == 0){
			
			lpicValue = icalvalue_new_from_string(ICAL_DATETIME_VALUE, icalproperty_get_x(lpicProp));
			ttMozLastAck = icaltime_as_timet_with_zone(icalvalue_get_datetime(lpicValue), NULL);
			if(ttMozLastAck > ttMozLastAckMax)//save max of X-MOZ-LAST-ACK if present twice.
				ttMozLastAckMax = ttMozLastAck;
			icalvalue_free(lpicValue);
			bHasMozAck = true;
		}
		else if (strncmp(icalproperty_get_x_name(lpicProp), "X-MOZ-SNOOZE-TIME", strlen("X-MOZ-SNOOZE-TIME")) == 0) {
			// x properties always return a char* as value :(
			lpicValue = icalvalue_new_from_string(ICAL_DATETIME_VALUE, icalproperty_get_x(lpicProp));
			ttReminderNext = icaltime_as_timet_with_zone(icalvalue_get_datetime(lpicValue), NULL); // no timezone			
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERNEXTTIME], PT_SYSTIME);
			UnixTimeToFileTime(ttReminderNext, &sPropVal.Value.ft);
			lpIcalItem->lstMsgProps.push_back(sPropVal);
			
			// X-MOZ-SNOOZE-TIME-1231250400000000
			strSuffix = icalproperty_get_x_name(lpicProp);
			if(strSuffix != "X-MOZ-SNOOZE-TIME")
			{
				sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZ_SNOOZE_SUFFIX], PT_SYSTIME);
				strSuffix = strSuffix.substr(strlen("X-MOZ-SNOOZE-TIME")+ 1, 10); //ignoring trailing 6 zeros for hh:mm:ss
				UnixTimeToFileTime(atoi(strSuffix.c_str()), &sPropVal.Value.ft);
				lpIcalItem->lstMsgProps.push_back(sPropVal);
			}
			icalvalue_free(lpicValue);
		}

		lpicProp = icalcomponent_get_next_property(lpicEvent, ICAL_X_PROPERTY);
	}

	if(ttReminderNext == 0)//delete the next-reminder time if X-MOZ-SNOOZE-TIME is absent.
	{
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERNEXTTIME], PT_SYSTIME));
	}
	
	if (bHasMozAck) { // save X-MOZ-LAST-ACK if found in request.
		sPropMozAck.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZLASTACK], PT_SYSTIME);
		UnixTimeToFileTime(ttMozLastAckMax, &sPropMozAck.Value.ft);		
		lpIcalItem->lstMsgProps.push_back(sPropMozAck);
	}
	else { //delete X-MOZ-LAST-ACK if not found in request.
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZLASTACK], PT_SYSTIME));		
	}

	// reminderset
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERSET], PT_BOOLEAN);
	sPropVal.Value.b = bReminderSet;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// remindbefore
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERMINUTESBEFORESTART], PT_LONG);
	sPropVal.Value.ul = ulRemindBefore;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// remindertime
	if (ttReminderTime == 0) {
		// get starttime from item
		// DTSTART must be available
		lpicDTStartProp = icalcomponent_get_first_property(lpicEvent, ICAL_DTSTART_PROPERTY);
		if (!lpicDTStartProp) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		ttReminderTime = ICalTimeTypeToUTC(lpicEventRoot, lpicDTStartProp);
	}
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERTIME], PT_SYSTIME);
	UnixTimeToFileTime(ttReminderTime, &sPropVal.Value.ft);
	lpIcalItem->lstMsgProps.push_back(sPropVal);

exit:
	return hr;
}

HRESULT VConverter::HrAddRecurrence(icalcomponent *lpicEventRoot, icalcomponent *lpicEvent, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	ICalRecurrence icRecClass;
	icalproperty *lpicProp = NULL;
	SPropValue spSpropVal = {0};

	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_RRULE_PROPERTY);
	if (lpicProp == NULL) {

		// set isRecurring to false , property required by BlackBerry.
		spSpropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRING], PT_BOOLEAN);
		spSpropVal.Value.b = false;
		lpIcalItem->lstMsgProps.push_back(spSpropVal);

		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRENCESTATE], PT_BINARY));
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRENCEPATTERN], PT_STRING8));
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRENCE_START], PT_SYSTIME));
		lpIcalItem->lstDelPropTags.push_back(CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRENCE_END], PT_SYSTIME));

		lpIcalItem->lpRecurrence = NULL;

		// remove all exception attachments from existing message, done in ICal2Mapi.cpp
		goto exit;
	}

	if (m_iCurrentTimeZone == m_mapTimeZones->end()) {
		// if we have an RRULE, we must have a timezone
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	hr = icRecClass.HrParseICalRecurrenceRule(m_iCurrentTimeZone->second, lpicEventRoot, lpicEvent, m_lpNamedProps, lpIcalItem, &lpIcalItem->lpRecurrence);
	if (hr != hrSuccess)
		goto exit;

	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_X_PROPERTY);
	while (lpicProp) {
		if (strcmp(icalproperty_get_x_name(lpicProp), "X-ZARAFA-REC-PATTERN") == 0){
			spSpropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRENCEPATTERN], PT_STRING8);
			HrCopyString(m_lpIConv, lpIcalItem->base, icalproperty_get_x(lpicProp), &spSpropVal.Value.lpszA);
			lpIcalItem->lstMsgProps.push_back(spSpropVal);
		}
		lpicProp = icalcomponent_get_next_property(lpicEvent, ICAL_X_PROPERTY);
	}
exit:
	return hr;
}

HRESULT VConverter::HrAddException(icalcomponent *lpEventRoot, icalcomponent *lpEvent, icalitem *lpPrevItem)
{
	HRESULT hr = hrSuccess;
	ICalRecurrence cRec;
	icalitem::exception ex;
	icalproperty_method icMethod = ICAL_METHOD_NONE;

	hr = HrCompareUids(lpPrevItem, lpEvent);
	if (hr != hrSuccess)
		goto exit;

	if (!lpPrevItem->lpRecurrence) {
		// can't add exceptions if the previous item did not have an RRULE
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	icMethod = icalcomponent_get_method(lpEventRoot);

	// it's the same item, handle exception
	hr = cRec.HrMakeMAPIException(lpEventRoot, lpEvent, lpPrevItem, m_lpNamedProps, m_lpIConv, &ex);
	if (hr != hrSuccess)
		goto exit;
	

	hr = HrAddRecipients(lpEvent, lpPrevItem, &ex.lstMsgProps, &ex.lstRecips);
	if (hr != hrSuccess)
		goto exit;
	
	hr = HrResolveUser(lpPrevItem, &ex.lstRecips);
	if (hr != hrSuccess)
		goto exit;

	hr = HrAddMessageClass(icMethod, lpEvent, lpPrevItem, true, &ex.lstMsgProps);
	if (hr != hrSuccess)
		goto exit;

	lpPrevItem->lstExceptionAttachments.push_back(ex);

exit:
	return hr;
}



HRESULT VConverter::HrFindTimezone(ULONG ulProps, LPSPropValue lpProps, std::string *lpstrTZid, TIMEZONE_STRUCT *lpTZinfo, icaltimezone **lppicTZinfo)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpPropTimeZoneString = NULL;
	LPSPropValue lpPropTimeZoneStruct = NULL;
	string strTZid;
	string::size_type pos;
	TIMEZONE_STRUCT ttTZinfo = {0};
	icaltimezone *lpicTZinfo = NULL;
	icalcomponent *lpicComp = NULL;
	size_t ulPos = 0;

	// Retrieve timezone. If available (outlook fills this in for recurring items), place it in lpMapTimeZones
	lpPropTimeZoneString = PpropFindProp(lpProps, ulProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_TIMEZONE], PT_STRING8));
	if (lpPropTimeZoneString == NULL) {
		// use all dates/times as UTC
		strTZid = "(GMT+0000)";
		// UTC is not placed in the map, and not placed in ical, so we're done here
		goto done;
	}
	strTZid = lpPropTimeZoneString->Value.lpszA;

	if (strTZid[0] == '(') {
		// shorten string so the timezone id is not so long, and hopefully matches more for the same timezone

		// from: (GMT+01:00) Amsterdam, Berlijn, Bern, Rome, Stockholm, Wenen
		// from: (GMT+01.00) Sarajevo/Warsaw/Zagreb
		// to: (GMT+01:00) (note: the dot is not converted .. should we?)

		pos = strTZid.rfind(')');
		strTZid = strTZid.substr(0, pos+1);
	}
	ulPos = strTZid.find('+');

	// check if strTZid is in map
	m_iCurrentTimeZone = m_mapTimeZones->find(strTZid);
	if (m_iCurrentTimeZone != m_mapTimeZones->end()) {
		// already used this timezone before
		ttTZinfo = m_iCurrentTimeZone->second;
	} else {
		lpPropTimeZoneStruct = PpropFindProp(lpProps, ulProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_TIMEZONEDATA], PT_BINARY));
		if (lpPropTimeZoneStruct) {
			ttTZinfo = *(TIMEZONE_STRUCT*)lpPropTimeZoneStruct->Value.bin.lpb;
			(*m_mapTimeZones)[strTZid] = ttTZinfo;
			// keep timezone pointer for recurrence
			m_iCurrentTimeZone = m_mapTimeZones->find(strTZid);
		} else if(ulPos != string::npos){
			strTZid = "(GMT+0000)"; // identify GMT+XXX timezones
			goto done;
		}
		else {
			strTZid = "(GMT-0000)"; // identify GMT-XXX timezones
			goto done;
		}
	}

	// construct ical version for icaltime_from_timet_with_zone()
	hr = HrCreateVTimeZone(strTZid, ttTZinfo, &lpicComp);
	if (hr == hrSuccess) {
		lpicTZinfo = icaltimezone_new();
		if (icaltimezone_set_component(lpicTZinfo, lpicComp) == 0) {
			icalcomponent_free(lpicComp);
			icaltimezone_free(lpicTZinfo, true);
		}
	}
	hr = hrSuccess;

done:
	*lpstrTZid = strTZid;
	*lpTZinfo = ttTZinfo;
	*lppicTZinfo = lpicTZinfo;

	return hr;
}

HRESULT VConverter::HrSetTimeProperty(time_t tStamp, bool bDateOnly, icaltimezone *lpicTZinfo, std::string strTZid, icalproperty_kind icalkind, icalcomponent *lpicEvent)
{
	HRESULT hr = hrSuccess;
	icaltimetype ittStamp;
	icalproperty *lpicProp = NULL;

	if(bDateOnly && strTZid == "(GMT+0000)") // add a day for GMT+XXXX timezones
		tStamp += 86400;

	if (lpicTZinfo)
		ittStamp = icaltime_from_timet_with_zone(tStamp, bDateOnly, lpicTZinfo);
	else
		ittStamp = icaltime_from_timet_with_zone(tStamp, bDateOnly, icaltimezone_get_utc_timezone());

	lpicProp = icalproperty_new(icalkind);
	if (!lpicProp) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	icalproperty_set_value(lpicProp, icalvalue_new_datetime(ittStamp));

	if (lpicTZinfo)
		icalproperty_add_parameter(lpicProp, icalparameter_new_from_value_string(ICAL_TZID_PARAMETER, strTZid.c_str()));
	icalcomponent_add_property(lpicEvent, lpicProp);

exit:
	return hr;
}

HRESULT VConverter::HrSetOrganizerAndAttendees(LPMESSAGE lpParentMsg, LPMESSAGE lpMessage, ULONG ulProps, LPSPropValue lpProps, icalproperty_method *lpicMethod, icalcomponent *lpicEvent)
{
	HRESULT hr = hrSuccess;
	icalproperty_method icMethod = ICAL_METHOD_NONE;
	string strSenderName, strSenderType, strSenderEmailAddr;
	string strReceiverName, strReceiverType, strReceiverEmailAddr;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	LPSPropValue lpPropVal = NULL;
	LPSPropValue lpSpropVal = NULL;
	icalproperty *lpicProp = NULL;
	icalparameter *lpicParam = NULL;
	string strBuf;
	ULONG ulMeetingStatus = 0;
	
	//Remove Organiser & Attendees of Root event for exception.
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_ORGANIZER_PROPERTY);
	if (lpicProp) {
		icalcomponent_remove_property(lpicEvent, lpicProp);
		icalproperty_free(lpicProp);
	}
	
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_ATTENDEE_PROPERTY);
	while (lpicProp) {
		if (lpicProp) {
			icalcomponent_remove_property(lpicEvent, lpicProp);
			icalproperty_free(lpicProp);
		}
		lpicProp = icalcomponent_get_next_property(lpicEvent, ICAL_X_PROPERTY);
	}

	// Request mail address from addressbook to get the actual email address
	// use the parent message, OL does not set PR_SENDER_ENTRYID in exception message.
	hr = HrGetAddress(m_lpAdrBook, lpParentMsg,
					  PR_SENDER_ENTRYID, PR_SENDER_NAME, PR_SENDER_ADDRTYPE, PR_SENDER_EMAIL_ADDRESS,
					  strSenderName, strSenderType, strSenderEmailAddr);
	if (hr != hrSuccess)
		goto exit;

	// get class to find method and type for attendees and organizer
	lpPropVal = PpropFindProp(lpProps, ulProps, PR_MESSAGE_CLASS);
	if (lpPropVal == NULL) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	strBuf = lpPropVal->Value.lpszA;

	// Set attendee info
	if (strBuf.compare(0, string("IPM.Schedule.Meeting.Resp.").length(), string("IPM.Schedule.Meeting.Resp.")) == 0)
	{
		// responding to a meeting request:
		// the to should only be the organizer of this meeting
		icMethod = ICAL_METHOD_REPLY;

		// gmail always sets CONFIRMED, exchange fills in the correct value ... doesn't seem to matter .. for now
		icalcomponent_add_property(lpicEvent, icalproperty_new_status(ICAL_STATUS_CONFIRMED));

		if (strBuf.rfind("Pos") != string::npos) {
			lpicParam = icalparameter_new_partstat(ICAL_PARTSTAT_ACCEPTED);
		} else if (strBuf.rfind("Neg") != string::npos) {
			lpicParam = icalparameter_new_partstat(ICAL_PARTSTAT_DECLINED);
		} else if (strBuf.rfind("Tent") != string::npos) {
			lpicParam = icalparameter_new_partstat(ICAL_PARTSTAT_TENTATIVE);
		} else {
			// shouldn't happen, but better than having no lpicParam pointer
			lpicParam = icalparameter_new_partstat(ICAL_PARTSTAT_ACCEPTED);
		}

		// I am the only attendee that is replying
		strBuf = "mailto:" + strSenderEmailAddr;
		lpicProp = icalproperty_new_attendee(m_lpIConv->convert(strBuf).c_str());
		icalproperty_add_parameter(lpicProp, lpicParam);

		if (!strSenderName.empty())
			icalproperty_add_parameter(lpicProp, icalparameter_new_cn(m_lpIConv->convert(strSenderName).c_str()));
		icalcomponent_add_property(lpicEvent, lpicProp);

		// Organizer should be the only MAPI_TO entry
		hr = lpMessage->GetRecipientTable(0, &lpTable);
		if (hr != hrSuccess)
			goto exit;

		hr = lpTable->QueryRows(-1, 0, &lpRows);
		if (hr != hrSuccess)
			goto exit;
		
		// The response should only be sent to the organizer (restrict on MAPI_TO ? ...)
		if (lpRows->cRows != 1) {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		// TODO: use correct index number?
		hr = HrGetAddress(m_lpAdrBook, lpRows->aRow[0].lpProps, lpRows->aRow[0].cValues,
						  PR_ENTRYID, PR_DISPLAY_NAME, PR_ADDRTYPE, PR_EMAIL_ADDRESS,
						  strReceiverName, strReceiverType, strReceiverEmailAddr);
		if (hr != hrSuccess)
			goto exit;

		strBuf = "mailto:" + strReceiverEmailAddr;
		lpicProp = icalproperty_new_organizer(m_lpIConv->convert(strBuf).c_str());

		if (!strReceiverName.empty()) {
			lpicParam = icalparameter_new_cn(m_lpIConv->convert(strReceiverName).c_str());
			icalproperty_add_parameter(lpicProp, lpicParam);
		}

		icalcomponent_add_property(lpicEvent, lpicProp);
	}
	else
	{
		// strBuf == "IPM.Schedule.Meeting.Request", "IPM.Schedule.Meeting.Canceled" or ....?
		// strBuf == "IPM.Appointment": normal calender item

		// If we're dealing with a meeting, preset status to 1. PROP_MEETINGSTATUS may not be set
		if (strBuf.compare(0, string("IPM.Schedule.Meeting").length(), string("IPM.Schedule.Meeting")) == 0)
			ulMeetingStatus = 1;

		// a normal calendar item has meeting status == 0, all other types != 0
		lpPropVal = PpropFindProp(lpProps, ulProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MEETINGSTATUS], PT_LONG));
		if (lpPropVal)
			ulMeetingStatus = lpPropVal->Value.ul;
		else {
			// if MeetingStatus flag is not set in exception message, retrive it from parent message.
			if (HrGetOneProp(lpParentMsg, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MEETINGSTATUS], PT_LONG), &lpSpropVal) == hrSuccess)
				ulMeetingStatus = lpSpropVal->Value.ul;
		}

		// meeting bit enabled
		if (ulMeetingStatus & 1) {
			if (ulMeetingStatus & 4) {
				icalcomponent_add_property(lpicEvent, icalproperty_new_status(ICAL_STATUS_CANCELLED));
				icMethod = ICAL_METHOD_CANCEL;
			} else {
				icalcomponent_add_property(lpicEvent, icalproperty_new_status(ICAL_STATUS_CONFIRMED));
				icMethod = ICAL_METHOD_REQUEST;
			}

			// meeting action, add all attendees, request reply when needed
			hr = HrSetICalAttendees(lpMessage, strSenderEmailAddr, lpicEvent);
			if (hr != hrSuccess)
				goto exit;

			//Set this property to force thunderbird to send invitations mails.
			lpPropVal = PpropFindProp (lpProps, ulProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZSENDINVITE], PT_BOOLEAN));
			if (lpPropVal && !lpPropVal->Value.b) 
				lpicProp = icalproperty_new_x("FALSE");
			else
				lpicProp = icalproperty_new_x("TRUE");
	
			icalproperty_set_x_name(lpicProp, "X-MOZ-SEND-INVITATIONS"); 
			icalcomponent_add_property(lpicEvent, lpicProp);
			
			// I am the Organizer
			strBuf = "mailto:" + strSenderEmailAddr;
			lpicProp = icalproperty_new_organizer(m_lpIConv->convert(strBuf).c_str());

			if (!strSenderName.empty())
				icalproperty_add_parameter(lpicProp, icalparameter_new_cn(m_lpIConv->convert(strSenderName).c_str()));

			lpPropVal = PpropFindProp(lpProps, ulProps, PR_SENT_REPRESENTING_ENTRYID);
			if (lpPropVal)
			{
				// skip error check
				if (HrGetAddress(m_lpAdrBook, (LPENTRYID)lpPropVal->Value.bin.lpb, lpPropVal->Value.bin.cb, strSenderName, strSenderType, strSenderEmailAddr) == hrSuccess)
					icalproperty_add_parameter(lpicProp, icalparameter_new_sentby(strSenderEmailAddr.c_str()));				
			}
			
			icalcomponent_add_property(lpicEvent, lpicProp);
		} else {
			// normal calendar item
			icMethod = ICAL_METHOD_PUBLISH;
		}
	}

	*lpicMethod = icMethod;

exit:
	if (lpSpropVal)
		MAPIFreeBuffer (lpSpropVal);

	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT VConverter::HrSetTimeProperties(LPSPropValue lpMsgProps, ULONG ulMsgProps, icaltimezone *lpicTZinfo, std::string strTZid, icalcomponent *lpEvent)
{
	HRESULT hr = hrSuccess;
	icalproperty *lpProp = NULL;
	LPSPropValue lpPropVal = NULL;
	icaltimetype ittICalTime;

	// Set creation time / CREATED
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, PR_CREATION_TIME);
	if (lpPropVal) {
		ittICalTime = icaltime_from_timet(FileTimeToUnixTime(lpPropVal->Value.ft.dwHighDateTime, lpPropVal->Value.ft.dwLowDateTime), 0);
		ittICalTime.is_utc = 1;

		lpProp = icalproperty_new_created(ittICalTime);
		icalcomponent_add_property(lpEvent, lpProp);
	}

	// Set modification time / LAST-MODIFIED + DTSTAMP
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, PR_LAST_MODIFICATION_TIME);
	if (lpPropVal) {
		ittICalTime = icaltime_from_timet(FileTimeToUnixTime(lpPropVal->Value.ft.dwHighDateTime, lpPropVal->Value.ft.dwLowDateTime), 0);
		ittICalTime.is_utc = 1;

		lpProp = icalproperty_new_lastmodified(ittICalTime);
		icalcomponent_add_property(lpEvent,lpProp);

		lpProp = icalproperty_new_dtstamp(ittICalTime);
		icalcomponent_add_property(lpEvent,lpProp);
	}

	return hr;
}

HRESULT VConverter::HrSetICalAttendees(LPMESSAGE lpMessage, std::string strOrganizer, icalcomponent *lpicEvent)
{
	HRESULT hr = hrSuccess;
	icalproperty *lpProp = NULL;
	icalparameter *lpParam = NULL;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	LPSPropValue lpPropVal = NULL;
	ULONG ulCount = 0;
	string strBuf;
	string strName, strType, strEmailAddress;
	SizedSPropTagArray(7, lpRecipProps) = {7, {PR_ENTRYID, PR_DISPLAY_NAME, PR_ADDRTYPE, PR_EMAIL_ADDRESS,
											   PR_RECIPIENT_FLAGS, PR_RECIPIENT_TYPE, PR_RECIPIENT_TRACKSTATUS}
	};

	hr = lpMessage->GetRecipientTable(0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns((LPSPropTagArray)&lpRecipProps, 0);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryRows(-1, 0, &lpRows);
	if (hr != hrSuccess)
		goto exit;
	
	// Set all recipients into icalcomponent lpicEvent
	for (ulCount = 0; ulCount < lpRows->cRows; ulCount++) {
		// ZARAFA types go correct because of addressbook, (slow?, should use PR_SMTP_ADDRESS?)
		// SMTP types go correct because of PR_EMAIL_ADDRESS 
		hr = HrGetAddress(m_lpAdrBook, lpRows->aRow[ulCount].lpProps, lpRows->aRow[ulCount].cValues,
						  PR_ENTRYID, PR_DISPLAY_NAME, PR_ADDRTYPE, PR_EMAIL_ADDRESS,
						  strName, strType, strEmailAddress);

		// skip the organiser if present in the recipient table.
		if (hr != hrSuccess || strEmailAddress == strOrganizer)
			continue;

		// flags set to 3 is organizer, so skip that entry
		lpPropVal = PpropFindProp(lpRows->aRow[ulCount].lpProps, lpRows->aRow[ulCount].cValues, PR_RECIPIENT_FLAGS);
		if (lpPropVal != NULL && lpPropVal->Value.ul == 3)
			continue;

		lpPropVal = PpropFindProp(lpRows->aRow[ulCount].lpProps, lpRows->aRow[ulCount].cValues, PR_RECIPIENT_TYPE);
		if (lpPropVal == NULL)
			continue;

		switch (lpPropVal->Value.ul) {
		case MAPI_TO:
			lpParam = icalparameter_new_role(ICAL_ROLE_REQPARTICIPANT);
			break;
		case MAPI_CC:
			lpParam = icalparameter_new_role(ICAL_ROLE_OPTPARTICIPANT);
			break;
		case MAPI_BCC:
			lpParam = icalparameter_new_role(ICAL_ROLE_NONPARTICIPANT);
			break;
		default:
			continue;
		}

		strBuf = "mailto:" + strEmailAddress;

		lpProp = icalproperty_new_attendee(m_lpIConv->convert(strBuf).c_str());
		icalproperty_add_parameter(lpProp, lpParam);

		lpPropVal = PpropFindProp(lpRows->aRow[ulCount].lpProps, lpRows->aRow[ulCount].cValues, PR_RECIPIENT_TRACKSTATUS);
		if (lpPropVal != NULL) {
			if (lpPropVal->Value.ul == 2)
				icalproperty_add_parameter(lpProp, icalparameter_new_partstat(ICAL_PARTSTAT_TENTATIVE));
			else if (lpPropVal->Value.ul == 3)
				icalproperty_add_parameter(lpProp, icalparameter_new_partstat(ICAL_PARTSTAT_ACCEPTED));
			else if (lpPropVal->Value.ul == 4)
				icalproperty_add_parameter(lpProp, icalparameter_new_partstat(ICAL_PARTSTAT_DECLINED));					
			else {
				icalproperty_add_parameter(lpProp, icalparameter_new_partstat(ICAL_PARTSTAT_NEEDSACTION));
				icalproperty_add_parameter(lpProp, icalparameter_new_rsvp(ICAL_RSVP_TRUE));
			}
		} else {
			// make sure clients are requested to send a reply on meeting requests
			icalproperty_add_parameter(lpProp, icalparameter_new_partstat(ICAL_PARTSTAT_NEEDSACTION));
			icalproperty_add_parameter(lpProp, icalparameter_new_rsvp(ICAL_RSVP_TRUE));
		}

		if (!strName.empty())
			icalproperty_add_parameter(lpProp, icalparameter_new_cn(m_lpIConv->convert(strName).c_str()));

		icalcomponent_add_property(lpicEvent, lpProp);
	}

exit:
	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT VConverter::HrSetBusyStatus(LPMESSAGE lpMessage, ULONG ulBusyStatus, icalcomponent *lpicEvent)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpSpropVal = NULL;
	icalproperty *lpicProp = NULL;
	
	// set the TRANSP property
	if (ulBusyStatus == 0) {
		lpicProp = icalproperty_new_transp(ICAL_TRANSP_TRANSPARENT);
	} else {
		lpicProp = icalproperty_new_transp(ICAL_TRANSP_OPAQUE);
	}
	icalcomponent_add_property(lpicEvent, lpicProp);
	
	// set the X-MICROSOFT-CDO-BUSYSTATUS property
	hr = HrGetOneProp(lpMessage, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_INTENDEDBUSYSTATUS], PT_LONG),&lpSpropVal);
	if(hr == hrSuccess && lpSpropVal->Value.ul != (ULONG)-1)
		ulBusyStatus = lpSpropVal->Value.ul;

	switch (ulBusyStatus) {
		case 0:
			lpicProp = icalproperty_new_x("FREE");
			break;
		case 1:
			lpicProp = icalproperty_new_x("TENTATIVE");
			break;
		case 2:
			lpicProp = icalproperty_new_x("BUSY");
			break;
		case 3:
			lpicProp = icalproperty_new_x("OOF");
			break;
	}
	
	icalproperty_set_x_name(lpicProp, "X-MICROSOFT-CDO-BUSYSTATUS"); 
	icalcomponent_add_property(lpicEvent, lpicProp);

	if(lpSpropVal)
		MAPIFreeBuffer(lpSpropVal);
	return hrSuccess;
}

HRESULT VConverter::HrSetXHeaders(ULONG ulMsgProps, LPSPropValue lpMsgProps, icalcomponent *lpEvent)
{
	LPSPropValue lpPropVal = NULL;
	icaltimetype icCriticalChange;
	icalvalue *lpicValue = NULL;
	icalproperty *lpProp = NULL;
	time_t ttCriticalChange = 0;
	ULONG ulApptSeqNo = 0;
	ULONG ulOwnerApptID = 0;
	char *lpszTemp = NULL;

	// set X-MICROSOFT-CDO & X-MOZ properties 
	// X-MICROSOFT-CDO-OWNER-CRITICAL-CHANGE
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_OWNERCRITICALCHANGE], PT_SYSTIME));
	if (lpPropVal) {
		FileTimeToUnixTime(lpPropVal->Value.ft, &ttCriticalChange);
	}else {
		ttCriticalChange = time(NULL);
	}

	icCriticalChange = icaltime_from_timet_with_zone(ttCriticalChange, false, icaltimezone_get_utc_timezone());
	lpicValue = icalvalue_new_datetime(icCriticalChange);
	lpszTemp = icalvalue_as_ical_string_r(lpicValue);
	lpProp = icalproperty_new_x(lpszTemp);
	icalmemory_free_buffer(lpszTemp);
	icalproperty_set_x_name(lpProp, "X-MICROSOFT-CDO-OWNER-CRITICAL-CHANGE"); 
	icalcomponent_add_property(lpEvent, lpProp);
	icalvalue_free(lpicValue);

	// X-MICROSOFT-CDO-ATTENDEE-CRITICAL-CHANGE
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_ATTENDEECRITICALCHANGE], PT_SYSTIME));
	if (lpPropVal) {
		FileTimeToUnixTime(lpPropVal->Value.ft, &ttCriticalChange);
	}else {
		ttCriticalChange = time(NULL);
	}

	icCriticalChange = icaltime_from_timet_with_zone(ttCriticalChange, false, icaltimezone_get_utc_timezone());
	lpicValue = icalvalue_new_datetime(icCriticalChange);
	lpszTemp = icalvalue_as_ical_string_r(lpicValue);
	lpProp = icalproperty_new_x(lpszTemp);
	icalmemory_free_buffer(lpszTemp);
	
	icalproperty_set_x_name(lpProp, "X-MICROSOFT-CDO-ATTENDEE-CRITICAL-CHANGE"); 
	icalcomponent_add_property(lpEvent, lpProp);
	icalvalue_free(lpicValue);
	
	// X-MICROSOFT-CDO-APPT-SEQUENCE
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTSEQNR], PT_LONG));
	if (lpPropVal) {
		ulApptSeqNo = lpPropVal->Value.ul;
	}else {
		ulApptSeqNo = 0;
	}
	lpicValue = icalvalue_new_integer(ulApptSeqNo);
	lpszTemp = icalvalue_as_ical_string_r(lpicValue);
	lpProp = icalproperty_new_x(lpszTemp);
	icalmemory_free_buffer(lpszTemp);	
	icalproperty_set_x_name(lpProp, "X-MICROSOFT-CDO-APPT-SEQUENCE"); 
	icalcomponent_add_property(lpEvent, lpProp);
	icalvalue_free(lpicValue);

	// X-MICROSOFT-CDO-OWNERAPPTID
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, PR_OWNER_APPT_ID);
	if (lpPropVal) {
		ulOwnerApptID = lpPropVal->Value.ul;
	}else {
		ulOwnerApptID = -1;
	}
	lpicValue = icalvalue_new_integer(ulOwnerApptID);
	lpszTemp = icalvalue_as_ical_string_r(lpicValue);
	lpProp = icalproperty_new_x(lpszTemp);
	icalmemory_free_buffer(lpszTemp);	
	icalproperty_set_x_name(lpProp, "X-MICROSOFT-CDO-OWNERAPPTID"); 
	icalcomponent_add_property(lpEvent, lpProp);
	icalvalue_free(lpicValue);

	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps,  CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZGEN], PT_LONG));
	if (lpPropVal)
	{
		LONG ulXmozGen = 0;
		icalvalue *lpicValue = NULL;
		ulXmozGen = lpPropVal->Value.ul;
		lpicValue = icalvalue_new_integer(ulXmozGen);

		lpszTemp = icalvalue_as_ical_string_r(lpicValue);
		lpProp = icalproperty_new_x(lpszTemp);
		icalmemory_free_buffer(lpszTemp);		
		icalproperty_set_x_name(lpProp, "X-MOZ-GENERATION"); 
		icalcomponent_add_property(lpEvent, lpProp);		
		icalvalue_free(lpicValue);
	}

	return hrSuccess;
}

HRESULT VConverter::HrSetVAlarm(ULONG ulProps, LPSPropValue lpProps, icalcomponent *lpicEvent)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpPropVal = NULL;
	icalcomponent *lpAlarm = NULL;
	icalproperty *lpicProp = NULL;
	time_t ttSnooze = 0;
	time_t ttSnoozeSuffix = 0;
	bool blxmozgen = false;
	bool blisItemReccr = false;
	char *lpszTemp = NULL;
	
	// find bool, skip if error or false
	lpPropVal = PpropFindProp(lpProps, ulProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERSET], PT_BOOLEAN));
	if (!lpPropVal || lpPropVal->Value.b == FALSE)
		goto exit;

	lpPropVal = PpropFindProp(lpProps, ulProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERMINUTESBEFORESTART], PT_LONG));
	if (!lpPropVal) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = HrParseReminder(lpPropVal->Value.l, &lpAlarm);
	if (hr != hrSuccess)
		goto exit;

	icalcomponent_add_component(lpicEvent, lpAlarm);

	lpPropVal = PpropFindProp(lpProps, ulProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRING], PT_BOOLEAN));
	if(lpPropVal && lpPropVal->Value.b == true)
		blisItemReccr = true;

	// retrieve the suffix time for property X-MOZ-SNOOZE-TIME
	lpPropVal = PpropFindProp(lpProps, ulProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZ_SNOOZE_SUFFIX], PT_SYSTIME));
	if(lpPropVal)
		FileTimeToUnixTime(lpPropVal->Value.ft, &ttSnoozeSuffix);	
	// check latest snooze time
	lpPropVal = PpropFindProp(lpProps, ulProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERNEXTTIME], PT_SYSTIME));
	if (lpPropVal) {
		icaltimetype icSnooze;
		icalvalue *lpicValue = NULL;
		std::string strSnoozeTime = "X-MOZ-SNOOZE-TIME";

		// set timestamp in name for recurring items i.e. X-MOZ-SNOOZE-TIME-1231250400000000:20090107T132846Z
		if (ttSnoozeSuffix != 0 && blisItemReccr)
			strSnoozeTime += "-" + stringify(ttSnoozeSuffix) + "000000";

		FileTimeToUnixTime(lpPropVal->Value.ft, &ttSnooze);
		icSnooze = icaltime_from_timet_with_zone(ttSnooze, false, icaltimezone_get_utc_timezone());
		lpicValue = icalvalue_new_datetime(icSnooze);

		lpszTemp = icalvalue_as_ical_string_r(lpicValue);
		lpicProp = icalproperty_new_x(lpszTemp);
		icalmemory_free_buffer(lpszTemp);
		icalproperty_set_x_name(lpicProp, strSnoozeTime.c_str()); 

		icalcomponent_add_property(lpicEvent, lpicProp);
		icalvalue_free(lpicValue);
	}

	lpPropVal = PpropFindProp(lpProps, ulProps,  CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZGEN], PT_LONG));
	if (lpPropVal)
		blxmozgen = true;

	// send X-MOZ-LASTACK
	lpPropVal = PpropFindProp(lpProps, ulProps,  CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MOZLASTACK], PT_SYSTIME));
	if (lpPropVal)
	{
		time_t ttLastAckTime = 0;
		icaltimetype icModTime;
		icalvalue *lpicValue = NULL;
		
		FileTimeToUnixTime(lpPropVal->Value.ft, &ttLastAckTime);
		//do not send X-MOZ-LASTACK if reminder older than last ack time
		if(ttLastAckTime > ttSnooze && !blxmozgen)
			goto exit;
		icModTime = icaltime_from_timet_with_zone(ttLastAckTime, false, icaltimezone_get_utc_timezone());
		lpicValue = icalvalue_new_datetime(icModTime);

		lpszTemp = icalvalue_as_ical_string_r(lpicValue);
		lpicProp = icalproperty_new_x(lpszTemp);
		icalmemory_free_buffer(lpszTemp);
		icalproperty_set_x_name(lpicProp, "X-MOZ-LASTACK"); 
		icalcomponent_add_property(lpicEvent, lpicProp);
		icalvalue_free(lpicValue);
	}

exit:
	return hr;
}

HRESULT VConverter::HrSetBody(LPMESSAGE lpMessage, icalproperty **lppicProp)
{
	HRESULT hr = hrSuccess;
	LPSTREAM lpStream = NULL;
	STATSTG sStreamStat;
	char *lpBody = NULL;

	hr = lpMessage->OpenProperty(PR_BODY, &IID_IStream, 0, MAPI_DEFERRED_ERRORS, (LPUNKNOWN*)&lpStream);
	if (hr != hrSuccess)
		goto exit;

	hr = lpStream->Stat(&sStreamStat, 0);
	if (hr != hrSuccess)
		goto exit;

	if (sStreamStat.cbSize.LowPart == 0) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	lpBody = new char[sStreamStat.cbSize.LowPart +1];

	hr = lpStream->Read(lpBody, sStreamStat.cbSize.LowPart, NULL);
	if (hr != hrSuccess)
		goto exit;

	lpBody[sStreamStat.cbSize.LowPart] = '\0';

	*lppicProp = icalproperty_new_description(m_lpIConv->convert(lpBody).c_str());

exit:
	if (lpStream)
		lpStream->Release();

	if (lpBody)
		delete [] lpBody;

	return hr;
}

HRESULT VConverter::HrSetItemSpecifics(ULONG ulProps, LPSPropValue lpProps, icalcomponent *lpicEvent)
{
	return hrSuccess;
}

HRESULT VConverter::HrSetRecurrence(LPMESSAGE lpMessage, icalcomponent *lpicEvent, icaltimezone *lpicTZinfo, std::string strTZid, std::list<icalcomponent*> *lpEventList)
{
	HRESULT hr = hrSuccess;
	ULONG ulRecurrenceStateTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRENCESTATE], PT_BINARY);
	bool bIsAllDay = false;
	bool bIsAllDayException = false;

	LPSPropValue lpPropVal = NULL;
	recurrence cRecurrence;
	LPSTREAM lpStream = NULL;
	STATSTG sStreamStat;
	char *lpRecurrenceData = NULL;

	ICalRecurrence cICalRecurrence;
	LPMESSAGE lpException = NULL;
	
	icalcomponent *lpicException = NULL;
	icalcomponent *lpicComp = NULL;
	icalproperty *lpicProp = NULL;
	ULONG ulModCount = 0;
	ULONG ulModifications = 0;
	time_t tNewTime = 0;
	time_t tExceptionStart = 0;
	std::list<icalcomponent*> lstExceptions;

	//set X-ZARAFA-REC-PATTERN property.
	hr = HrGetOneProp(lpMessage, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRENCEPATTERN], PT_STRING8), &lpPropVal);
	if (hr == hrSuccess)
	{		
		lpicProp = icalproperty_new_x(m_lpIConv->convert(lpPropVal->Value.lpszA).c_str());
		icalproperty_set_x_name(lpicProp, "X-ZARAFA-REC-PATTERN"); 
		icalcomponent_add_property(lpicEvent, lpicProp);
	}

	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	hr = HrGetOneProp(lpMessage, ulRecurrenceStateTag, &lpPropVal);
	if (hr == hrSuccess)
	{
		hr = cRecurrence.HrLoadRecurrenceState((char*)lpPropVal->Value.bin.lpb, lpPropVal->Value.bin.cb);
	} else if (hr == MAPI_E_NOT_FOUND) {
		// When exception is created in MR, the IsRecurring is set - true by OL
		// but Recurring state is not set in MR.
		hr = hrSuccess;
		goto exit;
	} else {
		// open property and read full blob
		hr = lpMessage->OpenProperty(ulRecurrenceStateTag, &IID_IStream, 0, MAPI_DEFERRED_ERRORS, (LPUNKNOWN*)&lpStream);
		if (hr != hrSuccess)
			goto exit;

		hr = lpStream->Stat(&sStreamStat, 0);
		if (hr != hrSuccess)
			goto exit;

		lpRecurrenceData = new char[sStreamStat.cbSize.LowPart];

		hr = lpStream->Read(lpRecurrenceData, sStreamStat.cbSize.LowPart, NULL);
		if (hr != hrSuccess)
			goto exit;

		hr = cRecurrence.HrLoadRecurrenceState(lpRecurrenceData, sStreamStat.cbSize.LowPart);
	}
	if (FAILED(hr))
		goto exit;
	hr = hrSuccess;

	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	hr = HrGetOneProp(lpMessage, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_ALLDAYEVENT], PT_BOOLEAN), &lpPropVal);
	if (hr == hrSuccess)
		bIsAllDay = lpPropVal->Value.b;

	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	ASSERT(m_iCurrentTimeZone != m_mapTimeZones->end());

	// now that we have the recurrence state class, we can create rrules in lpicEvent
	hr = cICalRecurrence.HrCreateICalRecurrence(m_iCurrentTimeZone->second, bIsAllDay, &cRecurrence, lpicEvent);
	if (hr != hrSuccess)
		goto exit;

	// all modifications create new event item:
	// RECURRENCE-ID: contains local timezone timestamp of item that is changed
	// other: CREATED, LAST-MODIFIED, DTSTAMP, UID (copy from original)
	// and then exception properties are replaced
	ulModCount = cRecurrence.getModifiedCount();
	for (ULONG i = 0; i < ulModCount; i++) {
		
		LPSPropValue lpMsgProps = NULL;
		ULONG ulMsgProps = 0;
		icalproperty_method icMethod = ICAL_METHOD_NONE;

		ulModifications = cRecurrence.getModifiedFlags(i);

		bIsAllDayException = bIsAllDay;

		hr = cICalRecurrence.HrMakeICalException(lpicEvent, &lpicException);
		if (hr != hrSuccess)
			continue;

		tExceptionStart = tNewTime = cRecurrence.getModifiedStartDateTime(i);
		
		hr = HrGetExceptionMessage(lpMessage, tExceptionStart, &lpException);
		if (hr != hrSuccess)
		{
			hr = hrSuccess;
			continue;
		}
		
		hr = lpException->GetProps(NULL, 0, &ulMsgProps, &lpMsgProps);
		if (FAILED(hr))
			goto exit;
		
		hr = HrSetOrganizerAndAttendees(lpMessage, lpException, ulMsgProps, lpMsgProps, &icMethod, lpicException);
		if (hr != hrSuccess)
			goto exit;

		if (ulModifications & ARO_SUBTYPE)
		{
			hr = HrGetOneProp(lpException, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_ALLDAYEVENT], PT_BOOLEAN),&lpPropVal);
			if (hr == hrSuccess)
				bIsAllDayException = lpPropVal->Value.b;

			if(lpPropVal)
			{
				MAPIFreeBuffer(lpPropVal);
				lpPropVal = NULL;
			}
		}
		
		// 1. get new StartDateTime and EndDateTime from exception and make DTSTART and DTEND in sTimeZone
		tNewTime = LocalToUTC(tNewTime, m_iCurrentTimeZone->second);
		hr = HrSetTimeProperty(tNewTime, bIsAllDayException, lpicTZinfo, strTZid, ICAL_DTSTART_PROPERTY, lpicException);
		if (hr != hrSuccess)
			continue;

		tNewTime = LocalToUTC(cRecurrence.getModifiedEndDateTime(i), m_iCurrentTimeZone->second);
		hr = HrSetTimeProperty(tNewTime, bIsAllDayException, lpicTZinfo, strTZid, ICAL_DTEND_PROPERTY, lpicException);
		if (hr != hrSuccess)
			continue;

		hr = HrGetOneProp(lpException, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRINGBASE], PT_SYSTIME), &lpPropVal);
		if (hr == hrSuccess)
		{
			tNewTime = FileTimeToUnixTime(lpPropVal->Value.ft.dwHighDateTime, lpPropVal->Value.ft.dwLowDateTime);
			hr = HrSetTimeProperty(tNewTime, bIsAllDay, lpicTZinfo, strTZid, ICAL_RECURRENCEID_PROPERTY, lpicException);
			if (hr != hrSuccess)
				continue;

			if (lpPropVal)
			{
				MAPIFreeBuffer(lpPropVal);
				lpPropVal = NULL;
			}
		}

		// 2. for each (useful?) bit in ulOverrideFlags, set property
		if (ulModifications & ARO_SUBJECT) {
			// find the previous value, and remove it
			lpicProp = icalcomponent_get_first_property(lpicException, ICAL_SUMMARY_PROPERTY);
			if (lpicProp) {
				icalcomponent_remove_property(lpicException, lpicProp);
				icalproperty_free(lpicProp);
			}

			icalcomponent_add_property(lpicException, icalproperty_new_summary (m_lpIConv->convert (cRecurrence.getModifiedSubject (i)).c_str()));
		}

		if (ulModifications & ARO_MEETINGTYPE) {
			// make this in invite, cancel, ... ?
		}

		if (ulModifications & ARO_REMINDERDELTA && !(ulModifications & ARO_REMINDERSET)) {
			HrUpdateReminderTime(lpicException, cRecurrence.getModifiedReminderDelta(i));
		}

		if (ulModifications & ARO_REMINDERSET) {
			// Outlook is nasty!
			// If you make an exception reminder enable with the DEFAULT 15 minutes alarm,
			// the value is NOT saved in the exception attachment.
			// That's why I don't open the attachment if the value isn't going to be present anyway.
			// Also, it (the default) is present in the main item ... not very logical.
			if (ulModifications & ARO_REMINDERDELTA)
				hr = HrGetOneProp(lpException, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REMINDERMINUTESBEFORESTART], PT_LONG), &lpPropVal);
			hr = hrSuccess;

			// add new valarm
			// although a previous valarm should not be here, the webaccess always says it's been changed, so we remove the old one too
			lpicComp = icalcomponent_get_first_component(lpicException, ICAL_VALARM_COMPONENT);
			if (lpicComp) {
				icalcomponent_remove_component(lpicException, lpicComp);
				icalcomponent_free(lpicComp);
			}
			lpicComp = NULL;

			if (cRecurrence.getModifiedReminder(i) != 0) {
				hr = HrParseReminder(lpPropVal ? lpPropVal->Value.l : 15, &lpicComp);
				if (hr == hrSuccess)
					icalcomponent_add_component(lpicException, lpicComp);
			}
		}

		if (ulModifications & ARO_LOCATION) {
			lpicProp = icalcomponent_get_first_property(lpicException, ICAL_LOCATION_PROPERTY);
			if (lpicProp) {
				icalcomponent_remove_property(lpicException, lpicProp);
				icalproperty_free(lpicProp);
			}
			icalcomponent_add_property(lpicException, icalproperty_new_location (m_lpIConv->convert (cRecurrence.getModifiedLocation (i)).c_str() ));
		}

		if (ulModifications & ARO_BUSYSTATUS) {
			// new X-MICROSOFT-CDO-BUSYSTATUS and TRANSP
			lpicProp = icalcomponent_get_first_property(lpicException, ICAL_TRANSP_PROPERTY);
			if (lpicProp) {
				icalcomponent_remove_property(lpicException, lpicProp);
				icalproperty_free(lpicProp);
			}

			lpicProp = icalcomponent_get_first_property(lpicException, ICAL_X_PROPERTY);
			while (lpicProp && (strcmp(icalproperty_get_x_name(lpicProp), "X-MICROSOFT-CDO-BUSYSTATUS") != 0))
				lpicProp = icalcomponent_get_next_property(lpicException, ICAL_X_PROPERTY);
			if (lpicProp) {
				icalcomponent_remove_property(lpicException, lpicProp);
				icalproperty_free(lpicProp);
			}

			HrSetBusyStatus(lpException, cRecurrence.getModifiedBusyStatus(i), lpicException);
		}

		if (ulModifications & ARO_ATTACHMENT) {
			// ..?
		}

		if (ulModifications & ARO_APPTCOLOR) {
			// should never happen, according to the specs
		}

		if (ulModifications & ARO_EXCEPTIONAL_BODY) {
			lpicProp = icalcomponent_get_first_property(lpicException, ICAL_DESCRIPTION_PROPERTY);
			if (lpicProp) {
				icalcomponent_remove_property(lpicException, lpicProp);
				icalproperty_free(lpicProp);
			}
			lpicProp = NULL;

			if (HrSetBody(lpException, &lpicProp) == hrSuccess)
				icalcomponent_add_property(lpicException, lpicProp);
		}

		lstExceptions.push_back(lpicException);
		lpicException = NULL;

		if (lpPropVal)
			MAPIFreeBuffer(lpPropVal);
		lpPropVal = NULL;

		if (lpException)
			lpException->Release();
		lpException = NULL;
	}	

	*lpEventList = lstExceptions;

exit:
	if (lpicException)
		icalcomponent_free(lpicException);

	if (lpException)
		lpException->Release();

	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	if (lpRecurrenceData)
		delete [] lpRecurrenceData;

	if (lpStream)
		lpStream->Release();

	return hr;
}

HRESULT VConverter::HrUpdateReminderTime(icalcomponent *lpicEvent, LONG lReminder)
{
	HRESULT hr = hrSuccess;
	icalcomponent *lpicAlarm = NULL;
	icalproperty *lpicProp = NULL;
	icaltriggertype sittTrigger;

	lpicAlarm = icalcomponent_get_first_component(lpicEvent, ICAL_VALARM_COMPONENT);
	if (lpicAlarm == NULL) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	memset((void *) &sittTrigger, 0, sizeof(icaltriggertype));
	sittTrigger.duration = icaldurationtype_from_int(-1 * lReminder * 60); // set seconds

	lpicProp = icalcomponent_get_first_property(lpicAlarm, ICAL_TRIGGER_PROPERTY);
	if (lpicProp) {
		icalcomponent_remove_property(lpicAlarm, lpicProp);
		icalproperty_free(lpicProp);
	}
	icalcomponent_add_property(lpicAlarm, icalproperty_new_trigger(sittTrigger));

exit:
	return hr;
}

HRESULT VConverter::HrGetExceptionMessage(LPMESSAGE lpMessage, time_t tStart, LPMESSAGE *lppMessage)
{
	HRESULT hr = hrSuccess;
	LPMAPITABLE lpAttachTable = NULL;
	LPSRestriction lpAttachRestrict = NULL;
	LPSRowSet lpRows = NULL;
	LPSPropValue lpPropVal = NULL;
	LPATTACH lpAttach = NULL;
	LPMESSAGE lpAttachedMessage = NULL;
	SPropValue sStart = {0};
	SPropValue sMethod = {0};

	sStart.ulPropTag = PR_EXCEPTION_STARTTIME;
	UnixTimeToFileTime(tStart, &sStart.Value.ft);
	sMethod.ulPropTag = PR_ATTACH_METHOD;
	sMethod.Value.ul = ATTACH_EMBEDDED_MSG;

	hr = lpMessage->GetAttachmentTable(0, &lpAttachTable);
	if (hr != hrSuccess)
		goto exit;

	// restrict to only exception attachments
	CREATE_RESTRICTION(lpAttachRestrict);
	CREATE_RES_AND(lpAttachRestrict, lpAttachRestrict, 4);
	DATA_RES_EXIST(lpAttachRestrict, lpAttachRestrict->res.resAnd.lpRes[0], sStart.ulPropTag);
	DATA_RES_PROPERTY(lpAttachRestrict, lpAttachRestrict->res.resAnd.lpRes[1], RELOP_EQ, sStart.ulPropTag, &sStart);
	DATA_RES_EXIST(lpAttachRestrict, lpAttachRestrict->res.resAnd.lpRes[2], sMethod.ulPropTag);
	DATA_RES_PROPERTY(lpAttachRestrict, lpAttachRestrict->res.resAnd.lpRes[3], RELOP_EQ, sMethod.ulPropTag, &sMethod);

	hr = lpAttachTable->Restrict(lpAttachRestrict, 0);
	if (hr != hrSuccess)
		goto exit;

	// should result in 1 attachment
	hr = lpAttachTable->QueryRows(-1, 0, &lpRows);
	if (hr != hrSuccess)
		goto exit;

	ASSERT(lpRows->cRows == 1);

	if (lpRows->cRows == 0) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	lpPropVal = PpropFindProp(lpRows->aRow[0].lpProps, lpRows->aRow[0].cValues, PR_ATTACH_NUM);
	if (lpPropVal == NULL) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = lpMessage->OpenAttach(lpPropVal->Value.ul, NULL, 0, &lpAttach);
	if (hr != hrSuccess)
		goto exit;

	hr = lpAttach->OpenProperty(PR_ATTACH_DATA_OBJ, &IID_IMessage, 0, 0, (LPUNKNOWN *)&lpAttachedMessage);
	if (hr != hrSuccess)
		goto exit;

	*lppMessage = lpAttachedMessage;

exit:
	if (lpAttach)
		lpAttach->Release();

	if (lpRows)
		FreeProws(lpRows);

	if (lpAttachRestrict)
		MAPIFreeBuffer(lpAttachRestrict);

	if (lpAttachTable)
		lpAttachTable->Release();

	return hr;
}



HRESULT VConverter::HrMAPI2ICal(LPMESSAGE lpMessage, icalproperty_method *lpicMethod, std::list<icalcomponent*> *lpEventList)
{
	HRESULT hr = hrSuccess;
	std::list<icalcomponent*> lstEvents;
	icalproperty_method icMainMethod = ICAL_METHOD_NONE;
	icalcomponent* lpicEvent = NULL;
	LPSPropValue lpPropVal = NULL;
	LPSPropValue lpPropVal2 = NULL;
	icaltimezone *lpicTZinfo = NULL;
	std::string strTZid;

	// handle toplevel
	hr = HrMAPI2ICal(lpMessage, &icMainMethod, &lpicTZinfo, &strTZid, &lpicEvent);
	if (hr != hrSuccess)
		goto exit;

	// if recurring, add recurrence. We have to check two props since CDO only sets the second, while Outlook only sets the first :S
	if ((HrGetOneProp(lpMessage, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RECURRING], PT_BOOLEAN), &lpPropVal) == hrSuccess &&
		lpPropVal->Value.b == TRUE) || 
		(HrGetOneProp(lpMessage, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_ISRECURRING], PT_BOOLEAN), &lpPropVal2) == hrSuccess &&
		lpPropVal2->Value.b == TRUE))
	{
		hr = HrSetRecurrence(lpMessage, lpicEvent, lpicTZinfo, strTZid, &lstEvents);
		if (hr != hrSuccess)
			goto exit;
	}

	// push the main event in the front, before all exceptions
	lstEvents.push_front(lpicEvent);
	lpicEvent = NULL;

	// end
	*lpicMethod = icMainMethod;
	*lpEventList = lstEvents;

exit:
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	if (lpPropVal2)
		MAPIFreeBuffer(lpPropVal2);

	if (lpicEvent)
		icalcomponent_free(lpicEvent);

	if (lpicTZinfo)
		icaltimezone_free(lpicTZinfo, true);
	
	return hr;
}

HRESULT VConverter::HrMAPI2ICal(LPMESSAGE lpMessage, icalproperty_method *lpicMethod, icaltimezone **lppicTZinfo, std::string *lpstrTZid, icalcomponent *lpEvent)
{
	HRESULT hr = hrSuccess;
	icalproperty_method icMethod = ICAL_METHOD_NONE;
	icalproperty *lpProp = NULL;
	LPSPropValue lpPropVal = NULL;
	LPSPropValue lpMsgProps = NULL;
	ULONG ulMsgProps = 0;
	TIMEZONE_STRUCT ttTZinfo;
	icaltimezone *lpicTZinfo = NULL;
	ULONG ulCount = 0;
	std::string strTZid;
	std::string strBuf;
	std::string strUid;

	hr = lpMessage->GetProps(NULL, 0, &ulMsgProps, &lpMsgProps);
	if (FAILED(hr))
		goto exit;

	hr = HrFindTimezone(ulMsgProps, lpMsgProps, &strTZid, &ttTZinfo, &lpicTZinfo);
	if (hr != hrSuccess)
		goto exit;
	// non-UTC timezones are in the placed map, and converted using HrCreateVTimeZone() in MAPIToICal.cpp

	if(!m_blCensorPrivate) {
		// not a exception, so parent message is the message itself
		hr = HrSetOrganizerAndAttendees(lpMessage, lpMessage, ulMsgProps, lpMsgProps, &icMethod, lpEvent);
		if (hr != hrSuccess)
			goto exit;
	}

	// Set show_time_as / TRANSP
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_BUSYSTATUS], PT_LONG));
	if (!m_blCensorPrivate && lpPropVal)
		HrSetBusyStatus(lpMessage, lpPropVal->Value.ul, lpEvent);


	hr = HrSetTimeProperties(lpMsgProps, ulMsgProps, lpicTZinfo, strTZid, lpEvent);
	if (hr != hrSuccess)
		goto exit;

	// Set subject / SUMMARY
	if(m_blCensorPrivate) {
		lpProp = icalproperty_new_summary("Private Appointment");
		icalcomponent_add_property(lpEvent, lpProp);
	}
	else {
		lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, PR_SUBJECT);
		if (lpPropVal && lpPropVal->Value.lpszA[0] != '\0') {
			lpProp = icalproperty_new_summary(m_lpIConv->convert(lpPropVal->Value.lpszA).c_str());
			icalcomponent_add_property(lpEvent, lpProp);
		}
	}

	// Set location / LOCATION
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_LOCATION], PT_STRING8));
	if (!m_blCensorPrivate && lpPropVal && lpPropVal->Value.lpszA[0] != '\0') {
		lpProp = icalproperty_new_location(m_lpIConv->convert(lpPropVal->Value.lpszA).c_str());
		icalcomponent_add_property(lpEvent, lpProp);
	}

	// Set body / DESCRIPTION
	if(!m_blCensorPrivate) {
		lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, PR_BODY);
		if (lpPropVal && lpPropVal->Value.lpszA[0] != '\0') {
			lpProp = icalproperty_new_description(m_lpIConv->convert(lpPropVal->Value.lpszA).c_str());
		} else {
			hr = HrSetBody(lpMessage, &lpProp);
		}
		if (hr == hrSuccess)
			icalcomponent_add_property(lpEvent, lpProp);
		hr = hrSuccess;
	}

	// Set priority - use PR_IMPORTANCE or PR_PRIORITY
	lpPropVal = PpropFindProp (lpMsgProps, ulMsgProps, PR_IMPORTANCE);
	if (!m_blCensorPrivate && lpPropVal) {
		lpProp = icalproperty_new_priority(5 - ((lpPropVal->Value.l - 1) * 4));
		icalcomponent_add_property(lpEvent, lpProp);
	} else {
		lpPropVal = PpropFindProp (lpMsgProps, ulMsgProps, PR_PRIORITY);
		if (!m_blCensorPrivate && lpPropVal && lpPropVal->Value.l != 0) {
			lpProp = icalproperty_new_priority(5 - (lpPropVal->Value.l * 4));
			icalcomponent_add_property(lpEvent, lpProp);
		}
	}

	// Set keywords / CATEGORIES
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_KEYWORDS], PT_MV_STRING8));
	if (lpPropVal && lpPropVal->Value.MVszA.cValues > 0) {
		// The categories need to be comma-seperated
		strBuf = "";
		for (ulCount = 0; ulCount < lpPropVal->Value.MVszA.cValues; ulCount++) {
			strBuf += ",";
			strBuf += lpPropVal->Value.MVszA.lppszA[ulCount];
		}

		if (!strBuf.empty()) {
			strBuf = strBuf.substr(1); // strip 1st ,
			lpProp = icalproperty_new_categories(m_lpIConv->convert(strBuf).c_str());
			icalcomponent_add_property(lpEvent, lpProp);
		}
	}

	// Set url
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_NETSHOWURL], PT_STRING8));
	if (lpPropVal && lpPropVal->Value.lpszA[0] != '\0') {
		if (strlen(lpPropVal->Value.lpszA)) {
			lpProp = icalproperty_new_url(m_lpIConv->convert(lpPropVal->Value.lpszA).c_str());
			icalcomponent_add_property(lpEvent,lpProp);
		}
	}

	// Set contacts
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_CONTACTS], PT_MV_STRING8));
	if (lpPropVal) {
		for (ulCount = 0; ulCount < lpPropVal->Value.MVszA.cValues; ulCount++) {
			lpProp = icalproperty_new_contact(m_lpIConv->convert(lpPropVal->Value.MVszA.lppszA[ulCount]).c_str());
			icalcomponent_add_property(lpEvent,lpProp);
		}
	}

	// Set sensivity / CLASS
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, PR_SENSITIVITY);
	if (lpPropVal) {
		switch (lpPropVal->Value.ul) {
		case 1: //Personal
		case 2: //Private
			lpProp = icalproperty_new_class(ICAL_CLASS_PRIVATE);
			break;
		case 3: //CompanyConfidential
			lpProp = icalproperty_new_class(ICAL_CLASS_CONFIDENTIAL);
			break;
		default:
			lpProp = icalproperty_new_class(ICAL_CLASS_PUBLIC);
			break;
		}
		icalcomponent_add_property(lpEvent, lpProp);
	}

	// Set and/or create UID
	strBuf.clear();
	// Global Object ID
	//   In Microsoft Office Outlook 2003 Service Pack 1 (SP1) and earlier versions, the Global Object ID is generated when an organizer first sends a meeting request.
	//   Earlier versions of Outlook do not generate a Global Object ID for unsent meetings or for appointments that have no recipients.
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY));
	if (lpPropVal == NULL)
		lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RID], PT_BINARY));

	// If lpPropVal is 0, the global object id and recurrence id haven't been found.
	// The iCal UID is saved into both the global object id and recurrence id.
	if (lpPropVal == NULL) {
		SPropValue propUid;

		hr = HrGenerateUid(&strUid);
		if (hr != hrSuccess)
			goto exit;

		hr = HrMakeBinaryUID(strUid, lpMsgProps, &propUid); // base is lpMsgProps, which will be freed later
		
		// Set global object id and recurrence id.
		// ignore write errors, not really required that these properties are saved
		propUid.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY);
		HrSetOneProp(lpMessage, &propUid);
		
		propUid.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RID], PT_BINARY);
		HrSetOneProp(lpMessage, &propUid);

		// if we cannot write the message, use PR_ENTRYID to have the same uid every time we return the item
		// otherwise, ignore the error
		hr = lpMessage->SaveChanges(KEEP_OPEN_READWRITE);
		if (hr == E_ACCESSDENIED) {
			lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, PR_ENTRYID);
			if (lpPropVal)
				strUid = Util::bin2hex(lpPropVal->Value.bin.cb,lpPropVal->Value.bin.lpb);
		}
		hr = hrSuccess;
	} else {
		HrGetICalUidFromBinUid(lpPropVal->Value.bin, &strUid);
	}

	lpProp = icalproperty_new_uid(strUid.c_str());
	icalcomponent_add_property(lpEvent,lpProp);

	hr = HrSetItemSpecifics(ulMsgProps, lpMsgProps, lpEvent);
	if (hr != hrSuccess)
		goto exit;

	//Sequence
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTSEQNR], PT_LONG));
	if(lpPropVal)
	{
		lpProp = icalproperty_new_sequence(lpPropVal->Value.ul);
		icalcomponent_add_property(lpEvent, lpProp);
	}

	// Set alarm / VALARM (if alarm is found in lpMessage)
	if(!m_blCensorPrivate)	{
		hr = HrSetVAlarm(ulMsgProps, lpMsgProps, lpEvent);
		if (hr != hrSuccess)
			goto exit;
	}

	// Set X-PRoperites.
	hr = HrSetXHeaders(ulMsgProps, lpMsgProps, lpEvent);
	if (hr != hrSuccess)
		goto exit;
	
	// set return values
	if (lpicMethod)
		*lpicMethod = icMethod;

	if (lppicTZinfo)
		*lppicTZinfo = lpicTZinfo;

	if (lpstrTZid)
		*lpstrTZid = strTZid;

exit:
	if (lpMsgProps)
		MAPIFreeBuffer(lpMsgProps);

	return hr;
}
