/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "vevent.h"
#include <mapiutil.h>
#include "mapiext.h"
#include "CommonUtil.h"
#include "nameids.h"
#include "icaluid.h"
#include "stringutil.h"

VEventConverter::VEventConverter(LPADRBOOK lpAdrBook, timezone_map *mapTimeZones, LPSPropTagArray lpNamedProps, ECIConv *lpIConv, bool blCensor, bool bNoRecipients, IMailUser *lpMailUser)
	: VConverter(lpAdrBook, mapTimeZones, lpNamedProps, lpIConv, blCensor, bNoRecipients, lpMailUser)
{
}

VEventConverter::~VEventConverter()
{
}

HRESULT VEventConverter::HrICal2MAPI(icalcomponent *lpEventRoot, icalcomponent *lpEvent, icalitem *lpPrevItem, icalitem **lppRet)
{
	HRESULT hr = hrSuccess;

	hr = VConverter::HrICal2MAPI(lpEventRoot, lpEvent, lpPrevItem, lppRet);
	if (hr != hrSuccess)
		goto exit;

	(*lppRet)->eType = VEVENT;
	
exit:
	return hr;
}
// TODO: need to rename function
HRESULT VEventConverter::HrAddMessageClass(icalproperty_method icMethod, icalcomponent *lpicEvent, icalitem *lpIcalItem, bool bisException, std::list<SPropValue> *lstMsgProps)
{
	HRESULT hr = hrSuccess;
	icalproperty *icProp = NULL;
	icalparameter *icParam = NULL;
	SPropValue sPropVal;
	bool bMeeting = true;
	bool bMeetingOrganised = false;
	std::string strEmail;
	time_t tNow = 0;
	icProp = icalcomponent_get_first_property(lpicEvent, ICAL_ORGANIZER_PROPERTY);
	if (icProp) 
	{
		strEmail = icalproperty_get_organizer(icProp);
		
		if (stricmp(strEmail.substr(0,7).c_str(), "mailto:") == 0)
			strEmail = strEmail.substr(7);
		
		if(bIsUserLoggedIn(strEmail))
		{
			bMeetingOrganised = true;
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REQUESTSENT], PT_BOOLEAN);
			sPropVal.Value.b = true;
			lstMsgProps->push_back(sPropVal);
		}

	}

	switch (icMethod) {
	case ICAL_METHOD_REQUEST:
		sPropVal.ulPropTag = PR_RESPONSE_REQUESTED;
		sPropVal.Value.b = true;
		lstMsgProps->push_back(sPropVal);

		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RESPONSESTATUS], PT_LONG);
		sPropVal.Value.ul = 5; // olResponseNotResponded
		lstMsgProps->push_back(sPropVal);

		HrCopyString(NULL, lpIcalItem->base, "IPM.Schedule.Meeting.Request", &sPropVal.Value.lpszA);
		break;

	case ICAL_METHOD_REPLY:
		// A reply message must have only one attendee, the attendee replying
		if (icalcomponent_count_properties(lpicEvent, ICAL_ATTENDEE_PROPERTY) != 1) {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		icProp = icalcomponent_get_first_property(lpicEvent, ICAL_ATTENDEE_PROPERTY);
		if (!icProp) {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		icParam = icalproperty_get_first_parameter(icProp, ICAL_PARTSTAT_PARAMETER);
		if (!icParam) {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		switch (icalparameter_get_partstat(icParam)) {
		case ICAL_PARTSTAT_ACCEPTED:
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RESPONSESTATUS], PT_LONG);
			sPropVal.Value.ul = 3; // olResponseAccepted
			lstMsgProps->push_back(sPropVal);

			HrCopyString(NULL, lpIcalItem->base, "IPM.Schedule.Meeting.Resp.Pos", &sPropVal.Value.lpszA);
			break;

		case ICAL_PARTSTAT_DECLINED:
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RESPONSESTATUS], PT_LONG);
			sPropVal.Value.ul = 4; // olResponseDeclined
			lstMsgProps->push_back(sPropVal);

			HrCopyString(NULL, lpIcalItem->base, "IPM.Schedule.Meeting.Resp.Neg", &sPropVal.Value.lpszA);
			break;

		case ICAL_PARTSTAT_TENTATIVE:
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RESPONSESTATUS], PT_LONG);
			sPropVal.Value.ul = 2; // olResponseTentative
			lstMsgProps->push_back(sPropVal);

			HrCopyString(NULL, lpIcalItem->base, "IPM.Schedule.Meeting.Resp.Tent", &sPropVal.Value.lpszA);
			break;

		default:
			hr = MAPI_E_TYPE_NO_SUPPORT;
			goto exit;
		}
		break;

	case ICAL_METHOD_CANCEL:
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RESPONSESTATUS], PT_LONG);
		sPropVal.Value.ul = 5; // olResponseNotResponded
		lstMsgProps->push_back(sPropVal);

		HrCopyString(NULL, lpIcalItem->base, "IPM.Schedule.Meeting.Canceled", &sPropVal.Value.lpszA);
		break;

	case ICAL_METHOD_PUBLISH:
	default:
		
		// set ResponseStatus to 0 fix for BlackBerry.
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_RESPONSESTATUS], PT_LONG);
		if (m_ulUserStatus != 0)
			sPropVal.Value.ul = m_ulUserStatus;
		else if (bMeetingOrganised)
			sPropVal.Value.ul = 1;
		else
			sPropVal.Value.ul = 0;
		lstMsgProps->push_back(sPropVal);
		
		// time(NULL) returns UTC time as libical sets application to UTC time.
		tNow = time(NULL);
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTREPLYTIME], PT_SYSTIME);		
		UnixTimeToFileTime(tNow, &sPropVal.Value.ft);
		lstMsgProps->push_back(sPropVal);

		// Publish is used when mixed events are in the vcalender
		// we should determine on other properties if this is a meeting request related item
		HrCopyString(NULL, lpIcalItem->base, "IPM.Appointment", &sPropVal.Value.lpszA);

		// if we don't have attendees, skip the meeting request props
		if (icalcomponent_get_first_property(lpicEvent, ICAL_ATTENDEE_PROPERTY) == NULL)
			bMeeting = false;

		break;
	}

	if (!bisException) {
		sPropVal.ulPropTag = PR_MESSAGE_CLASS;
		lstMsgProps->push_back(sPropVal);
	}

	if (icMethod == ICAL_METHOD_REPLY || icMethod == ICAL_METHOD_CANCEL || icMethod == ICAL_METHOD_REQUEST)
	{
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_REQUESTSENT], PT_BOOLEAN); 
		sPropVal.Value.b = true;
		lstMsgProps->push_back(sPropVal);
	}

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_MEETINGSTATUS], PT_LONG);
	if (bMeeting) {
		// Set meeting status flags
		sPropVal.Value.ul = 1 | 2; // this-is-a-meeting-object flag + received flag
		if (icMethod == ICAL_METHOD_CANCEL)
			sPropVal.Value.ul |= 4; // cancelled flag
		else if(bMeetingOrganised)
			sPropVal.Value.ul = 1;
	} else {
		sPropVal.Value.ul = 0; // this-is-a-meeting-object flag off
	}
	lstMsgProps->push_back(sPropVal);

	sPropVal.ulPropTag = PR_ICON_INDEX;
	if (bMeeting)
		sPropVal.Value.ul = 1026;
	else
		sPropVal.Value.ul = 1024;
	// 1024: normal calendar item
	// 1025: recurring item
	// 1026: meeting request
	// 1027: recurring meeting request
	lstMsgProps->push_back(sPropVal);

exit:
	return hr;
}

HRESULT VEventConverter::HrAddTimes(icalcomponent *lpicEventRoot, icalcomponent *lpicEvent, icalitem *lpIcalItem)
{
	HRESULT hr = hrSuccess;
	SPropValue sPropVal;
	bool bIsAllday = false;
	bool bIsTZEmpty = false;
	icalparameter* lpicTZParam = NULL;
	icalproperty* lpicDTStartProp = NULL;
	time_t timeDTStart = 0;
	icalproperty* lpicDTEndProp = NULL;
	time_t timeDTEnd = 0;
	icalproperty* lpicDurationProp = NULL;
	icalproperty* lpicProp = NULL;
	const char *lpszTZID = NULL;
	std::string strTZ;

	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_RRULE_PROPERTY);
	// DTSTART must be available
	lpicDTStartProp = icalcomponent_get_first_property(lpicEvent, ICAL_DTSTART_PROPERTY);
	if (!lpicDTStartProp) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Take the timezone from DTSTART and set that as the item timezone
	lpicTZParam = icalproperty_get_first_parameter(lpicDTStartProp, ICAL_TZID_PARAMETER);
	// All day recurring items may not have timezone data.
	if (lpicTZParam || lpicProp) 
	{
		sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_TIMEZONE], PT_STRING8);
		if(lpicTZParam) {
			strTZ = strUnEscapeHex(icalparameter_get_tzid(lpicTZParam));
			lpszTZID = strTZ.c_str();
        }
		else 
		{
			bIsTZEmpty = true;
			lpszTZID = (m_mapTimeZones->begin()->first).c_str();
		}

		HrCopyString(NULL, lpIcalItem->base, lpszTZID, &sPropVal.Value.lpszA);
		lpIcalItem->lstMsgProps.push_back(sPropVal);

		// keep found timezone also as current timezone. will be used in recurrence
		m_iCurrentTimeZone = m_mapTimeZones->find(lpszTZID);
		if (m_iCurrentTimeZone != m_mapTimeZones->end()) {
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_TIMEZONEDATA], PT_BINARY);
			sPropVal.Value.bin.cb = sizeof(TIMEZONE_STRUCT);

			hr = MAPIAllocateMore(sizeof(TIMEZONE_STRUCT), lpIcalItem->base, (void**)&sPropVal.Value.bin.lpb);
			if (hr != hrSuccess)
				goto exit;
			memcpy(sPropVal.Value.bin.lpb, &m_iCurrentTimeZone->second, sizeof(TIMEZONE_STRUCT));
			lpIcalItem->lstMsgProps.push_back(sPropVal);
		} else {
			//.. huh? did find a timezone id, but not the actual timezone?? FAIL!
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
	}

	// when DTSTART is only a date, the item is an allday event
	if (icalproperty_get_dtstart(lpicDTStartProp).is_date == 1)
		bIsAllday = true;
	// Note: we do not set bIsAllDay to true when (END-START)%24h == 0
	// If the user forced it's ical client not to set this to 'true', it really wants an item that is a multiple of 24h, but specify the times too.

	// get timezone of DTSTART
	if (bIsAllday)
		timeDTStart = ICalTimeTypeToLocal(lpicEventRoot, lpicDTStartProp);
	else
		timeDTStart = ICalTimeTypeToUTC(lpicEventRoot, lpicDTStartProp);

	// Set 0x820D / ApptStartWhole
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTSTARTWHOLE], PT_SYSTIME);
	UnixTimeToFileTime(timeDTStart, &sPropVal.Value.ft);
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// Set 0x8516 / CommonStart
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_COMMONSTART], PT_SYSTIME);
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// Set PR_START_DATE
	sPropVal.ulPropTag = PR_START_DATE;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// Set 0x8215 / AllDayEvent
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_ALLDAYEVENT], PT_BOOLEAN);
	sPropVal.Value.b = bIsAllday;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// Set endtime / DTEND
	lpicDTEndProp = icalcomponent_get_first_property(lpicEvent, ICAL_DTEND_PROPERTY);
	if (lpicDTEndProp) {

		if (bIsAllday)
			timeDTEnd = ICalTimeTypeToLocal(lpicEventRoot, lpicDTEndProp);
		else
			timeDTEnd = ICalTimeTypeToUTC(lpicEventRoot, lpicDTEndProp);
		
		UnixTimeToFileTime(timeDTEnd, &sPropVal.Value.ft);

		// TODO: calc duration in minutes, set in PROP_APPTDURATION
	} else {
		// When DTEND is not in the ical, is should be a recurring message, which never ends!

		// use duration for "end"
		lpicDurationProp = icalcomponent_get_first_property(lpicEvent, ICAL_DURATION_PROPERTY);
		if (!lpicDurationProp) {
			// and then we get here when it never ends??
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		icaldurationtype dur = icalproperty_get_duration(lpicDurationProp);
		int sec = (dur.is_neg?-1:1)*(dur.weeks*7*24*60*60 + dur.days*24*60*60 + dur.hours*60*60 + dur.minutes*60 + dur.seconds);
		// why is negative compensated?!
		// switch start <-> end if negative?
		UnixTimeToFileTime(timeDTEnd + sec, &sPropVal.Value.ft);
	}
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTENDWHOLE], PT_SYSTIME);
	lpIcalItem->lstMsgProps.push_back(sPropVal);
	
	// Set 0x8517 / CommonEnd
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_COMMONEND], PT_SYSTIME);
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// Set PR_END_DATE
	sPropVal.ulPropTag = PR_END_DATE;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// Set duration
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTDURATION], PT_LONG);
	// Convert from seconds to minutes.
	sPropVal.Value.ul = icaldurationtype_as_int(icalcomponent_get_duration(lpicEvent)) / 60; // gort when not existing????
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	


	goto exit;
// TODO: add flags not to add these props ??
//       or maybe use a exclude filter in ICalToMAPI::GetItem()

	// Set submit time / DTSTAMP
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_DTSTAMP_PROPERTY);
	if (lpicProp)
		UnixTimeToFileTime(ICalTimeTypeToUTC(lpicEventRoot, lpicProp), &sPropVal.Value.ft);
	else
		UnixTimeToFileTime(time(NULL), &sPropVal.Value.ft);

	sPropVal.ulPropTag = PR_CLIENT_SUBMIT_TIME;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// Set delivery time / DTSTAMP
	sPropVal.ulPropTag = PR_MESSAGE_DELIVERY_TIME;
	UnixTimeToFileTime(time(NULL), &sPropVal.Value.ft);
	lpIcalItem->lstMsgProps.push_back(sPropVal);

exit:
	return hr;
}


HRESULT VEventConverter::HrMAPI2ICal(LPMESSAGE lpMessage, icalproperty_method *lpicMethod, icaltimezone **lppicTZinfo, std::string *lpstrTZid, icalcomponent **lppEvent)
{
	HRESULT hr = hrSuccess;
	icalcomponent *lpEvent = NULL;

	lpEvent = icalcomponent_new(ICAL_VEVENT_COMPONENT);

	hr = VConverter::HrMAPI2ICal(lpMessage, lpicMethod, lppicTZinfo, lpstrTZid, lpEvent);
	if (hr != hrSuccess)
		goto exit;

	if (lppEvent)
		*lppEvent = lpEvent;
	lpEvent = NULL;

exit:
	if (lpEvent)
		icalcomponent_free(lpEvent);

	return hr;
}

HRESULT VEventConverter::HrSetTimeProperties(LPSPropValue lpMsgProps, ULONG ulMsgProps, icaltimezone *lpicTZinfo, std::string strTZid, icalcomponent *lpEvent)
{
	HRESULT hr = hrSuccess;
	bool bIsAllDay = false;
	LPSPropValue lpPropVal = NULL;


	hr = VConverter::HrSetTimeProperties(lpMsgProps, ulMsgProps, lpicTZinfo, strTZid, lpEvent);
	if (hr != hrSuccess)
		goto exit;

	// vevent extra

	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_ALLDAYEVENT], PT_BOOLEAN));
	if (lpPropVal)
		bIsAllDay = lpPropVal->Value.b;
	// Note: If bIsAllDay == true, the item is an allday event "in the timezone it was created in" (and the user selected the allday event option)
	// In another timezone, Outlook will display the item as a 24h event with times (and the allday event option disabled)
	// However, in ICal, you cannot specify a timezone for a date, so ICal will show this as an allday event in every timezone your client is in.

	// Set start time / DTSTART
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTSTARTWHOLE], PT_SYSTIME));
	if (lpPropVal != NULL) {
		time_t ttTime = FileTimeToUnixTime(lpPropVal->Value.ft.dwHighDateTime, lpPropVal->Value.ft.dwLowDateTime);

		hr = HrSetTimeProperty(ttTime, bIsAllDay, lpicTZinfo, strTZid, ICAL_DTSTART_PROPERTY, lpEvent);
		if (hr != hrSuccess)
			goto exit;
	}

	// Set end time / DTEND
	lpPropVal = PpropFindProp(lpMsgProps, ulMsgProps, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTENDWHOLE], PT_SYSTIME));
	if (lpPropVal) {
		time_t ttTime = FileTimeToUnixTime(lpPropVal->Value.ft.dwHighDateTime, lpPropVal->Value.ft.dwLowDateTime);

		hr = HrSetTimeProperty(ttTime, bIsAllDay, lpicTZinfo, strTZid, ICAL_DTEND_PROPERTY, lpEvent);
		if (hr != hrSuccess)
			goto exit;
	}
	// else duration ?

exit:
	return hr;
}
