/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <mapiguid.h>
#include <mapicode.h>
#include <mapiutil.h>

#include "ECAttach.h"

#include "guid.h"
#include "ECGuid.h"

#include "ECDebug.h"
#include "ECDebugClient.h"

#include "ZarafaUtil.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

ECAttach::ECAttach(ECMsgStore *lpMsgStore, ULONG ulObjType, BOOL fModify, ULONG ulAttachNum, ECMAPIProp *lpRoot) : ECMAPIProp(lpMsgStore, ulObjType, fModify, lpRoot, "IAttach")
{
	this->ulAttachNum = ulAttachNum;

	this->HrAddPropHandlers(PR_ATTACH_DATA_OBJ,	GetPropHandler,	DefaultSetPropComputed,	(void*) this, TRUE,  FALSE);
	this->HrAddPropHandlers(PR_ATTACH_DATA_BIN,	GetPropHandler,	DefaultSetPropSetReal,	(void*) this, TRUE,  FALSE);
	this->HrAddPropHandlers(PR_ATTACH_SIZE,		DefaultGetProp,	DefaultSetPropComputed,	(void*) this, FALSE, FALSE);
	this->HrAddPropHandlers(PR_ATTACH_NUM,		GetPropHandler,	DefaultSetPropComputed,	(void*) this, FALSE, FALSE);
	this->HrAddPropHandlers(PR_ENTRYID,			GetPropHandler,	DefaultSetPropComputed,	(void*) this, FALSE, FALSE);
}

ECAttach::~ECAttach()
{
}

HRESULT ECAttach::Create(ECMsgStore *lpMsgStore, ULONG ulObjType, BOOL fModify, ULONG ulAttachNum, ECMAPIProp *lpRoot, ECAttach **lppAttach)
{
	HRESULT hr = hrSuccess;
	ECAttach *lpAttach = NULL;

	lpAttach = new ECAttach(lpMsgStore, ulObjType, fModify, ulAttachNum, lpRoot);

	hr = lpAttach->QueryInterface(IID_ECAttach, (void **)lppAttach);

	if(hr != hrSuccess)
		delete lpAttach;

	return hr;
}

HRESULT ECAttach::QueryInterface(REFIID iid, void **lpvoid)
{
	if(iid == IID_ECAttach || iid == IID_ECMAPIProp || iid == IID_ECUnknown) {
		AddRef();
		*lpvoid = this;
		return hrSuccess;
	}

	if(iid == IID_IAttachment || iid == IID_IMAPIProp || iid == IID_IUnknown) {
		AddRef();
		*lpvoid = &this->m_xAttach;
		return hrSuccess;
	}

	if(iid == IID_IECSingleInstance) {
		AddRef();
		*lpvoid = &this->m_xECSingleInstance;
		return hrSuccess;
	}

	return MAPI_E_INTERFACE_NOT_SUPPORTED;
}

HRESULT ECAttach::OpenProperty(ULONG ulPropTag, LPCIID lpiid, ULONG ulInterfaceOptions, ULONG ulFlags, LPUNKNOWN FAR * lppUnk)
{
	HRESULT			hr = hrSuccess;
	ECMessage*		lpMessage = NULL;
	IECPropStorage*	lpParentStorage = NULL;

	SPropValue		sPropValue[3];
	LPSPropValue	lpPropAttachType = NULL;
	LPMAPIUID		lpMapiUID = NULL;
	ULONG			ulAttachType = 0;
	BOOL			fNew = FALSE;
	ULONG			ulObjId = 0;

	if (lpiid == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Get the attachement method
	if(HrGetOneProp(&m_xAttach, PR_ATTACH_METHOD, &lpPropAttachType) == hrSuccess)
	{
		ulAttachType = lpPropAttachType->Value.ul;

		ECFreeBuffer(lpPropAttachType); lpPropAttachType = NULL;
	} else {
		// The client is creating a new attachment, which may be embedded. Fix for the next if check
		if ((ulFlags & MAPI_CREATE) && PROP_ID(ulPropTag) == PROP_ID(PR_ATTACH_DATA_OBJ) && *lpiid == IID_IMessage)
			ulAttachType = ATTACH_EMBEDDED_MSG;
	}

	if(ulAttachType == ATTACH_EMBEDDED_MSG && (PROP_ID(ulPropTag) == PROP_ID(PR_ATTACH_DATA_OBJ) && *lpiid == IID_IMessage)) {
		// Client is opening an IMessage submessage

		if (!m_sMapiObject->lstChildren->empty()) {
			fNew = FALSE;			// Create the submessage object from my sSavedObject data
			ulObjId = (*m_sMapiObject->lstChildren->begin())->ulObjId;
		} else {
			if(!fModify || !(ulFlags & MAPI_CREATE)) {
				hr = MAPI_E_NO_ACCESS;
				goto exit;
			}

			fNew = TRUE;			// new message in message
			ulObjId = 0;
		}

		hr = ECMessage::Create(this->GetMsgStore(), fNew, ulFlags & MAPI_MODIFY, 0, TRUE, m_lpRoot, &lpMessage);
		if(hr != hrSuccess)
			goto exit;

		// Client side unique ID is 0. Attachment can only have 1 submessage
		hr = this->GetMsgStore()->lpTransport->HrOpenParentStorage(this, 0, ulObjId, this->lpStorage->GetServerStorage(), &lpParentStorage);
		if(hr != hrSuccess)
			goto exit;

		hr = lpMessage->HrSetPropStorage(lpParentStorage, !fNew);
		if(hr != hrSuccess)
			goto exit;

		if (fNew) {
			// Load an empty property set
			hr = lpMessage->HrLoadEmptyProps();

			if(hr != hrSuccess)
				goto exit;

			//Set defaults
			// Same as ECMAPIFolder::CreateMessage
			ECAllocateBuffer(sizeof(MAPIUID), (void **) &lpMapiUID);

			hr = this->GetMsgStore()->lpSupport->NewUID(lpMapiUID);
			if(hr != hrSuccess)
				goto exit;

			sPropValue[0].ulPropTag = PR_MESSAGE_FLAGS;
			sPropValue[0].Value.l = MSGFLAG_UNSENT | MSGFLAG_READ;

			sPropValue[1].ulPropTag = PR_MESSAGE_CLASS;
			sPropValue[1].Value.lpszA = "IPM";
			
			sPropValue[2].ulPropTag = PR_SEARCH_KEY;
			sPropValue[2].Value.bin.cb = sizeof(MAPIUID);
			sPropValue[2].Value.bin.lpb = (LPBYTE)lpMapiUID;

			lpMessage->SetProps(3, sPropValue, NULL);
		}

		hr = lpMessage->QueryInterface(IID_IMessage, (void **)lppUnk);

		AddChild(lpMessage);

	} else {
		if(PROP_ID(ulPropTag) == PROP_ID(PR_ATTACH_DATA_OBJ))
			ulPropTag = PROP_TAG(PT_BINARY,PROP_ID(PR_ATTACH_DATA_OBJ));

		if(ulAttachType == ATTACH_OLE && (*lpiid != IID_IStorage && *lpiid != IID_IStream) ) {
			hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
			goto exit;
		}

		hr = ECMAPIProp::OpenProperty(ulPropTag, lpiid, ulInterfaceOptions, ulFlags, lppUnk);
	}

exit:
	if(lpParentStorage)
		lpParentStorage->Release();

	if(lpMessage)
		lpMessage->Release();

	if(lpMapiUID)
		ECFreeBuffer(lpMapiUID);

	return hr;
}

HRESULT	ECAttach::GetPropHandler(ULONG ulPropTag, void *lpProvider, ULONG ulFlags, LPSPropValue lpsPropValue, void *lpParam, void *lpBase)
{
	HRESULT hr = hrSuccess;
	ECAttach *lpAttach = (ECAttach*)lpParam;

	SPropTagArray sPropArray;
	ULONG cValues = 0;
	LPSPropValue lpProps = NULL;

	switch(ulPropTag) {
	case PR_ATTACH_DATA_OBJ:
		sPropArray.cValues = 1;
		sPropArray.aulPropTag[0] = PR_ATTACH_METHOD;
		hr = lpAttach->GetProps(&sPropArray, 0, &cValues, &lpProps);
		if(hr == hrSuccess && cValues == 1 && lpProps[0].ulPropTag == PR_ATTACH_METHOD && (lpProps[0].Value.ul == ATTACH_EMBEDDED_MSG || lpProps[0].Value.ul == ATTACH_OLE) )
		{
			lpsPropValue->ulPropTag = PR_ATTACH_DATA_OBJ;
			lpsPropValue->Value.x = 1;
		}else
			hr = MAPI_E_NOT_FOUND;
	
		break;
	case PR_ATTACH_DATA_BIN:
		sPropArray.cValues = 1;
		sPropArray.aulPropTag[0] = PR_ATTACH_METHOD;
		hr = lpAttach->GetProps(&sPropArray, 0, &cValues, &lpProps);
		if(lpProps[0].Value.ul == ATTACH_OLE) {
			hr = MAPI_E_NOT_FOUND;
		}else {
			// 8k limit
			hr = lpAttach->HrGetRealProp(PR_ATTACH_DATA_BIN, ulFlags, lpBase, lpsPropValue, 8192);
		}
		break;
	case PR_ATTACH_NUM:
		lpsPropValue->ulPropTag = PR_ATTACH_NUM;
		lpsPropValue->Value.ul = lpAttach->ulAttachNum;
		break;
	case PR_ENTRYID:// ignore property
	default:
		hr = MAPI_E_NOT_FOUND;
	}

	if(lpProps){ ECFreeBuffer(lpProps); lpProps = NULL; }

	return hr;
}

// Use the support object to do the copying
HRESULT ECAttach::CopyTo(ULONG ciidExclude, LPCIID rgiidExclude, LPSPropTagArray lpExcludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpInterface, LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray FAR * lppProblems)
{
	HRESULT hr = hrSuccess;
	
	hr = Util::DoCopyTo(&IID_IAttachment, &this->m_xAttach, ciidExclude, rgiidExclude, lpExcludeProps, ulUIParam, lpProgress, lpInterface, lpDestObj, ulFlags, lppProblems);

	return hr;
}

HRESULT ECAttach::SaveChanges(ULONG ulFlags)
{
	HRESULT hr = hrSuccess;
	ECPropertyEntryIterator iterProps;
	ULONG ulPropTag = 0;

	if (lpStorage == NULL) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	if (!fModify) {
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	HrSIEntryIDToID(m_sMapiObject->cbInstanceID, (LPBYTE)m_sMapiObject->lpInstanceID, NULL, NULL, (unsigned int *)&ulPropTag);

	if (ulPropTag) {
		// Search if one of the single-instanced properties has been changed,
		// if that has happened delete the single-instance id.
		for (iterProps = lstProps->begin(); iterProps != lstProps->end(); iterProps++) {
			if (PROP_ID(iterProps->GetPropTag()) != ulPropTag)
				continue;

			if (iterProps->FIsDirty() && !iterProps->FIsDeleted())
				SetSingleInstanceId(0, NULL);

			break;
		}
	}

	hr = ECMAPIProp::SaveChanges(ulFlags);

exit:
	return hr;
}

HRESULT ECAttach::HrSaveChild(ULONG ulFlags, MAPIOBJECT *lpsMapiObject)
{
	HRESULT hr = hrSuccess;
	ECMapiObjects::iterator iterSObj;

	if (!m_sMapiObject) {
		ASSERT(m_sMapiObject != NULL);
		AllocNewMapiObject(0, 0, MAPI_MESSAGE, &m_sMapiObject);
	}

	if (lpsMapiObject->ulObjType != MAPI_MESSAGE) {
		// can only save messages in an attachment
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}

	// attachments can only have 1 sub-message
	iterSObj = m_sMapiObject->lstChildren->begin();
	if (iterSObj != m_sMapiObject->lstChildren->end()) {
		FreeMapiObject(*iterSObj);
		m_sMapiObject->lstChildren->erase(iterSObj);
	}

	m_sMapiObject->lstChildren->push_back(new MAPIOBJECT(lpsMapiObject));

exit:
	return hr;
}

HRESULT ECAttach::GetSingleInstanceId(ULONG *lpcbInstanceID, LPSIEID *lppInstanceID)
{
	HRESULT hr = hrSuccess;

	if (!m_sMapiObject) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	if (!lpcbInstanceID || !lppInstanceID) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = Util::HrCopyEntryId(m_sMapiObject->cbInstanceID, (LPENTRYID)m_sMapiObject->lpInstanceID,
							 lpcbInstanceID, (LPENTRYID *)lppInstanceID);
	if (hr != hrSuccess)
		goto exit;

exit:
	return hr;
}

HRESULT ECAttach::SetSingleInstanceId(ULONG cbInstanceID, LPSIEID lpInstanceID)
{
	HRESULT hr = hrSuccess;

	if (!m_sMapiObject) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	if (m_sMapiObject->lpInstanceID)
		ECFreeBuffer(m_sMapiObject->lpInstanceID);

	m_sMapiObject->lpInstanceID = NULL;
	m_sMapiObject->cbInstanceID = 0;
	m_sMapiObject->bChangedInstance = false;

	hr = Util::HrCopyEntryId(cbInstanceID, (LPENTRYID)lpInstanceID,
							 &m_sMapiObject->cbInstanceID, (LPENTRYID *)&m_sMapiObject->lpInstanceID);
	if (hr != hrSuccess)
		goto exit;

	m_sMapiObject->bChangedInstance = true;

exit:
	return hr;
}

// Proxy routines for IAttach
HRESULT __stdcall ECAttach::xAttach::QueryInterface(REFIID iid, void ** ppvObject)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::QueryInterface", "%s", DBGGUIDToString(iid).c_str());
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->QueryInterface(iid, ppvObject);
	TRACE_MAPI(TRACE_RETURN, "IAttach::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

ULONG __stdcall ECAttach::xAttach::AddRef()
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::AddRef", "");
	METHOD_PROLOGUE_(ECAttach , Attach);
	return pThis->AddRef();
}

ULONG __stdcall ECAttach::xAttach::Release()
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::Release", "");
	METHOD_PROLOGUE_(ECAttach , Attach);
	return pThis->Release();
}

HRESULT __stdcall ECAttach::xAttach::GetLastError(HRESULT hError, ULONG ulFlags,
    LPMAPIERROR * lppMapiError)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::GetLastError", "hError=%s", GetMAPIErrorDescription(hError).c_str());
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->GetLastError(hError, ulFlags, lppMapiError);
	TRACE_MAPI(TRACE_RETURN, "IAttach::GetLastError", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::SaveChanges(ULONG ulFlags)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::SaveChanges", "");
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->SaveChanges(ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IAttach::SaveChanges", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::GetProps(LPSPropTagArray lpPropTagArray, ULONG ulFlags, ULONG FAR * lpcValues, LPSPropValue FAR * lppPropArray)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::GetProps", "%s", PropNameFromPropTagArray(lpPropTagArray).c_str());
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->GetProps(lpPropTagArray, ulFlags, lpcValues, lppPropArray);
	TRACE_MAPI(TRACE_RETURN, "IAttach::GetProps", "%s\n %s", GetMAPIErrorDescription(hr).c_str(), PropNameFromPropArray(*lpcValues, *lppPropArray).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::GetPropList(ULONG ulFlags, LPSPropTagArray FAR * lppPropTagArray)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::GetPropList", "flags=%d",ulFlags);
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->GetPropList(ulFlags, lppPropTagArray);
	TRACE_MAPI(TRACE_RETURN, "IAttach::GetPropList", "%s proplist=%s", GetMAPIErrorDescription(hr).c_str(), (lppPropTagArray)?PropNameFromPropTagArray(*lppPropTagArray).c_str(): "NULL");
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::OpenProperty(ULONG ulPropTag, LPCIID lpiid, ULONG ulInterfaceOptions, ULONG ulFlags, LPUNKNOWN FAR * lppUnk)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::OpenProperty",  "proptag=%s, flags=0x%08X, lpiid=%s, InterfaceOptions=0x%08X", PropNameFromPropTag(ulPropTag).c_str(), ulFlags, (lpiid)?DBGGUIDToString(*lpiid).c_str():"NULL", ulInterfaceOptions);
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->OpenProperty(ulPropTag, lpiid, ulInterfaceOptions, ulFlags, lppUnk);
	TRACE_MAPI(TRACE_RETURN, "IAttach::OpenProperty", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::SetProps(ULONG cValues, LPSPropValue lpPropArray, LPSPropProblemArray FAR * lppProblems)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::SetProps", "%s", PropNameFromPropArray(cValues, lpPropArray).c_str());
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->SetProps(cValues, lpPropArray, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IAttach::SetProps", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::DeleteProps(LPSPropTagArray lpPropTagArray, LPSPropProblemArray FAR * lppProblems)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::DeleteProps", "%s", PropNameFromPropTagArray(lpPropTagArray).c_str());
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->DeleteProps(lpPropTagArray, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IAttach::DeleteProps", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::CopyTo(ULONG ciidExclude, LPCIID rgiidExclude, LPSPropTagArray lpExcludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpInterface, LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray FAR * lppProblems)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::CopyTo", "ciidExclude=%d, lpExcludeProps=%s, lpDestInterface=%s, ulFlags=0x%X", ciidExclude, PropNameFromPropTagArray(lpExcludeProps).c_str(), (lpInterface)?DBGGUIDToString(*lpInterface).c_str():"NULL", ulFlags);
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->CopyTo(ciidExclude, rgiidExclude, lpExcludeProps, ulUIParam, lpProgress, lpInterface, lpDestObj, ulFlags, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IAttach::CopyTo", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::CopyProps(LPSPropTagArray lpIncludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpInterface, LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray FAR * lppProblems)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::CopyProps", "");
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->CopyProps(lpIncludeProps, ulUIParam, lpProgress, lpInterface, lpDestObj, ulFlags, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IAttach::CopyProps", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::GetNamesFromIDs(LPSPropTagArray * pptaga,
    LPGUID lpguid, ULONG ulFlags, ULONG * pcNames, LPMAPINAMEID ** pppNames)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::GetNamesFromIDs", "%s", PropNameFromPropTagArray(*pptaga).c_str());
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->GetNamesFromIDs(pptaga, lpguid, ulFlags, pcNames, pppNames);
	TRACE_MAPI(TRACE_RETURN, "IAttach::GetNamesFromIDs", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xAttach::GetIDsFromNames(ULONG cNames,
    LPMAPINAMEID * ppNames, ULONG ulFlags, LPSPropTagArray * pptaga)
{
	TRACE_MAPI(TRACE_ENTRY, "IAttach::GetIDsFromNames", "");
	METHOD_PROLOGUE_(ECAttach , Attach);
	HRESULT hr = pThis->GetIDsFromNames(cNames, ppNames, ulFlags, pptaga);
	TRACE_MAPI(TRACE_RETURN, "IAttach::GetIDsFromNames", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

// Proxy routines for IECSingleInstance
HRESULT __stdcall ECAttach::xECSingleInstance::QueryInterface(REFIID iid, void ** ppvObject)
{
	TRACE_MAPI(TRACE_ENTRY, "IECSingleInstance::QueryInterface", "%s", DBGGUIDToString(iid).c_str());
	METHOD_PROLOGUE_(ECAttach , ECSingleInstance);
	HRESULT hr = pThis->QueryInterface(iid, ppvObject);
	TRACE_MAPI(TRACE_RETURN, "IECSingleInstance::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

ULONG __stdcall ECAttach::xECSingleInstance::AddRef()
{
	TRACE_MAPI(TRACE_ENTRY, "IECSingleInstance::AddRef", "");
	METHOD_PROLOGUE_(ECAttach , ECSingleInstance);
	return pThis->AddRef();
}

ULONG __stdcall ECAttach::xECSingleInstance::Release()
{
	TRACE_MAPI(TRACE_ENTRY, "IECSingleInstance::Release", "");
	METHOD_PROLOGUE_(ECAttach , ECSingleInstance);
	return pThis->Release();
}

HRESULT __stdcall ECAttach::xECSingleInstance::GetSingleInstanceId(ULONG *lpcbInstanceID, LPENTRYID *lppInstanceID)
{
	TRACE_MAPI(TRACE_ENTRY, "IECSingleInstance::GetSingleInstanceId", "");
	METHOD_PROLOGUE_(ECAttach , ECSingleInstance);
	HRESULT hr = pThis->GetSingleInstanceId(lpcbInstanceID, (LPSIEID *)lppInstanceID);
	TRACE_MAPI(TRACE_RETURN, "IECSingleInstance::GetSingleInstanceId", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT __stdcall ECAttach::xECSingleInstance::SetSingleInstanceId(ULONG cbInstanceID, LPENTRYID lpInstanceID)
{
	TRACE_MAPI(TRACE_ENTRY, "IECSingleInstance::SetSingleInstanceId", "");
	METHOD_PROLOGUE_(ECAttach , ECSingleInstance);
	HRESULT hr = pThis->SetSingleInstanceId(cbInstanceID, (LPSIEID)lpInstanceID);
	TRACE_MAPI(TRACE_RETURN, "IECSingleInstance::SetSingleInstanceId", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

