/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "ECExchangeExportChanges.h"

#include <set>

#include "Util.h"
#include "guid.h"
#include "edkguid.h"
#include "ECGuid.h"
#include "mapiguid.h"
#include "mapiext.h"

#include <mapiutil.h>
#include "ZarafaICS.h"
#include "ECDebug.h"
#include "ECDebugClient.h"
#include "Mem.h"
#include "ECMessage.h"
#include "stringutil.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

ECExchangeExportChanges::ECExchangeExportChanges(ECMAPIFolder *lpFolder, unsigned int ulSyncType){
	m_lpLogger = new ECLogger_Null();
	m_lpFolder = lpFolder;
	m_ulSyncType = ulSyncType;

	m_bConfiged = false;
	m_lpStream = NULL;
	m_lpImportContents = NULL;
	m_lpImportHierarchy = NULL;
	m_ulFlags = 0;
	m_ulSyncId = 0;
	m_ulChangeId = 0;
	m_ulStep = 0;
	m_ulBufferSize = 0;
	m_ulChanges = 0;
	m_lpChanges = NULL;
	m_lpRestrict = NULL;
	m_lpChangePropTagArray = NULL;
	m_ulMaxChangeId = 0;

	m_lpFolder->AddRef();
}

ECExchangeExportChanges::~ECExchangeExportChanges(){
	delete m_lpLogger;
	
	if(m_lpChanges)
		MAPIFreeBuffer(m_lpChanges);

	if(m_lpChangePropTagArray)
		MAPIFreeBuffer(m_lpChangePropTagArray);

	if(m_lpFolder)
		m_lpFolder->Release();

	if(m_lpStream)
		m_lpStream->Release();
	
	if(m_lpImportContents)
		m_lpImportContents->Release();

	if(m_lpImportHierarchy)
		m_lpImportHierarchy->Release();
		
	if(m_lpRestrict)
		MAPIFreeBuffer(m_lpRestrict);
}

HRESULT ECExchangeExportChanges::Create(ECMAPIFolder *lpFolder, unsigned int ulSyncType, LPEXCHANGEEXPORTCHANGES* lppExchangeExportChanges){
	HRESULT hr = hrSuccess;
	ECExchangeExportChanges *lpEEC = NULL;
	
	if(!lpFolder || (ulSyncType != ICS_SYNC_CONTENTS && ulSyncType != ICS_SYNC_HIERARCHY)){
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	lpEEC = new ECExchangeExportChanges(lpFolder, ulSyncType);

	hr = lpEEC->QueryInterface(IID_IExchangeExportChanges, (void **)lppExchangeExportChanges);

exit:
	return hr;
}

HRESULT	ECExchangeExportChanges::QueryInterface(REFIID refiid, void **lpvoid)
{

	if(refiid == IID_ECExchangeExportChanges || refiid == IID_ECUnknown) {
		*lpvoid = this;
		AddRef();
		return hrSuccess;
	}
	
	if(refiid == IID_IExchangeExportChanges || refiid == IID_IUnknown) {
		*lpvoid = &this->m_xExchangeExportChanges;
		AddRef();
		return hrSuccess;
	}
	
	if(refiid == IID_IECExportChanges) {
		*lpvoid = &this->m_xECExportChanges;
		AddRef();
		return hrSuccess;
	}

	return MAPI_E_INTERFACE_NOT_SUPPORTED;
}

HRESULT ECExchangeExportChanges::GetLastError(HRESULT hResult, ULONG ulFlags, LPMAPIERROR *lppMAPIError){
	HRESULT		hr = hrSuccess;
	LPMAPIERROR	lpMapiError = NULL;
	const char* lpszCompName = "Zarafa Outlook Sharing";
	char*		lpszErrorMsg = NULL;
	
	if(ulFlags& MAPI_UNICODE) {
		hr = MAPI_E_BAD_CHARWIDTH;
		goto exit;
	}

	//FIXME: give synchronization errors messages
	lpszErrorMsg = Util::HrMAPIErrorToText((hResult == hrSuccess)?MAPI_E_NO_ACCESS : hResult);

	hr = ECAllocateBuffer(sizeof(MAPIERROR),(void **)&lpMapiError);
	if(hr != hrSuccess)
		goto exit;

	hr = ECAllocateMore(strlen(lpszCompName)+1, lpMapiError, (void**)&lpMapiError->lpszComponent);
	if(hr != hrSuccess)
		goto exit;

	strcpy(lpMapiError->lpszComponent, lpszCompName);

	hr = ECAllocateMore(strlen(lpszErrorMsg)+1, lpMapiError, (void**)&lpMapiError->lpszError);
	if(hr != hrSuccess)
		goto exit;

	strcpy(lpMapiError->lpszError, lpszErrorMsg);

	lpMapiError->ulContext		= 0;
	lpMapiError->ulLowLevelError= 0;
	lpMapiError->ulVersion		= 0;

	*lppMAPIError = lpMapiError;

exit:
	if( hr != hrSuccess && lpMapiError)
		ECFreeBuffer(lpMapiError);

	return hr;
}

HRESULT ECExchangeExportChanges::Config(LPSTREAM lpStream, ULONG ulFlags, LPUNKNOWN lpCollector, LPSRestriction lpRestriction, LPSPropTagArray lpIncludeProps, LPSPropTagArray lpExcludeProps, ULONG ulBufferSize){
	HRESULT		hr = hrSuccess;

	STATSTG		stat;
	ULONG		ulSyncId = 0;
	ULONG		ulChangeId = 0;
	ULONG		ulStep = 0;
	ULONG		ulCount = 0;
	//PR_PARENT_SOURCE_KEY, PR_SOURCE_KEY, PR_CHANGE_KEY, PR_PREDECESSOR_CHANGE_LIST, PR_DISPLAY_NAME, PR_LAST_MODIFICATION_TIME
	ULONG		ulDefaultContentProps[] = {PR_SOURCE_KEY, PR_LAST_MODIFICATION_TIME, PR_PARENT_SOURCE_KEY, PR_CHANGE_KEY, PR_MESSAGE_FLAGS, PR_PREDECESSOR_CHANGE_LIST, PR_CREATION_TIME};
	ULONG		ulDefaultHierarchyProps[] = {PR_SOURCE_KEY, PR_LAST_MODIFICATION_TIME, PR_PARENT_SOURCE_KEY, PR_CHANGE_KEY, PR_PREDECESSOR_CHANGE_LIST, PR_DISPLAY_NAME, PR_FOLDER_TYPE};
	ULONG 		* ulDefaultProps = ulDefaultContentProps;
	ULONG		ulChangeCount = 0;
	ULONG		ulProcessedChangeId = 0;
	bool		bFound = false;
	LPSPropValue lpPropSourceKey = NULL;
	LPSPropValue lpPropDisplayName = NULL;
	unsigned int i = 0;
	PROCESSEDCHANGESSET::iterator iterProcessedChange;
	std::list<ICSCHANGE>::iterator iterDeleteChange;
	ULONG		ulSourceKeySize = 0;
	char *		lpData = NULL;
	SBinary		sbNull = {0};
	bool		bForceImplicitStateUpdate = false;

	LARGE_INTEGER liPos = {{0, 0}};

	if(m_bConfiged){
		hr = MAPI_E_UNCONFIGURED;
		goto exit;
	}
	
	if(lpRestriction) {
		hr = Util::HrCopySRestriction(&m_lpRestrict, lpRestriction);
		if(hr != hrSuccess)
	    	goto exit;
	} else {
		m_lpRestrict = NULL;
	}
	
	m_ulFlags = ulFlags;
	
	if(! (ulFlags & SYNC_CATCHUP)) {
		if(lpCollector == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		
		// We don't need the importer when doing SYNC_CATCHUP
		if(this->m_ulSyncType == ICS_SYNC_CONTENTS){
			ulDefaultProps = ulDefaultContentProps;
			hr = lpCollector->QueryInterface(IID_IExchangeImportContentsChanges, (LPVOID*) &m_lpImportContents);
		}else if(m_ulSyncType == ICS_SYNC_HIERARCHY){
			ulDefaultProps = ulDefaultHierarchyProps;
			hr = lpCollector->QueryInterface(IID_IExchangeImportHierarchyChanges, (LPVOID*) &m_lpImportHierarchy);
		}else{
			hr = MAPI_E_INVALID_PARAMETER;
		}
		if(hr != hrSuccess)
			goto exit;
	}
	
	if(lpStream == NULL){
		LARGE_INTEGER lint = {{ 0, 0 }};
		ULONG tmp[2] = { 0, 0 };
		ULONG ulSize = 0;

		hr = CreateStreamOnHGlobal(GlobalAlloc(GPTR, sizeof(tmp)), true, &lpStream);
		if (hr != hrSuccess)
			goto exit;

		lpStream->Seek(lint, STREAM_SEEK_SET, NULL);
		lpStream->Write(tmp, sizeof(tmp), &ulSize);
	}

	m_lpStream = lpStream;
	m_lpStream->AddRef();


	//include PR_MESSAGE_SIZE or MAPI PR_ENTRYID
	if(lpIncludeProps){
		hr = MAPIAllocateBuffer(CbNewSPropTagArray(lpIncludeProps->cValues + 7), (LPVOID *) &m_lpChangePropTagArray);
		if(hr != hrSuccess)
			goto exit;
		m_lpChangePropTagArray->cValues = 0;
		for(ulStep = 0; ulStep < lpIncludeProps->cValues; ulStep++){
			m_lpChangePropTagArray->aulPropTag[m_lpChangePropTagArray->cValues++] = lpIncludeProps->aulPropTag[ulStep];
		}
	}else if(!lpExcludeProps && m_ulSyncType == ICS_SYNC_HIERARCHY){
		m_lpChangePropTagArray = NULL;
	}else{
		hr = MAPIAllocateBuffer(CbNewSPropTagArray(8), (LPVOID *) &m_lpChangePropTagArray);
		if(hr != hrSuccess)
			goto exit;
		m_lpChangePropTagArray->cValues = 0;
	}

	if(m_lpChangePropTagArray){
		for(ulCount = 0; ulCount < 7; ulCount++){
			bFound = false;
			if(lpExcludeProps){
				for(ulStep = 0; !bFound && ulStep < lpExcludeProps->cValues; ulStep++){
					if(lpExcludeProps->aulPropTag[ulStep] == ulDefaultProps[ulCount]){
						bFound = true;
					}
				}
			}
			if(!bFound){
				m_lpChangePropTagArray->aulPropTag[m_lpChangePropTagArray->cValues++] = ulDefaultProps[ulCount];
			}
		}
	}


	hr = m_lpStream->Stat(&stat, STATFLAG_NONAME);
	if(hr != hrSuccess)
		goto exit;

	if(stat.cbSize.HighPart == 0 && stat.cbSize.LowPart == 0) {
		ulSyncId = 0;
		ulChangeId = 0;
	} else {

		if(stat.cbSize.HighPart != 0 || stat.cbSize.LowPart < 8){
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		
		hr = m_lpStream->Seek(liPos, STREAM_SEEK_SET, NULL);
		if(hr != hrSuccess)
			goto exit;

		hr = m_lpStream->Read(&ulSyncId, 4, NULL);
		if(hr != hrSuccess)
			goto exit;

		hr = m_lpStream->Read(&ulChangeId, 4, NULL);
		if(hr != hrSuccess)
			goto exit;
			
		// Following the sync ID and the change ID is the list of changes that were already processed for
		// this sync ID / change ID combination. This allows us partial processing of items retrieved from 
		// the server.
		if(m_lpStream->Read(&ulChangeCount, 4, NULL) == hrSuccess) {
			// The stream contains a list of already processed items, read them
			
			for(i=0; i < ulChangeCount;i++) {
				hr = m_lpStream->Read(&ulProcessedChangeId, 4, NULL);
				if(hr != hrSuccess)
					goto exit; // Not the amount of expected bytes are there
				
				hr = m_lpStream->Read(&ulSourceKeySize, 4, NULL);
				if(hr != hrSuccess)
					goto exit;
					
				if(ulSourceKeySize > 1024) {
					// Stupidly large source key, the stream must be bad.
					hr = MAPI_E_INVALID_PARAMETER;
					goto exit;
				}
					
				lpData = new char[ulSourceKeySize];
					
				hr = m_lpStream->Read(lpData, ulSourceKeySize, NULL);
				if(hr != hrSuccess)
					goto exit;
					
				m_setProcessedChanges.insert(std::pair<unsigned int, std::string>(ulProcessedChangeId, std::string(lpData, ulSourceKeySize)));
				
				delete []lpData;
				lpData =  NULL;
			}
		}
	}

	if (m_lpLogger->Log(EC_LOGLEVEL_DEBUG))
		HrGetOneProp(&m_lpFolder->m_xMAPIFolder, PR_DISPLAY_NAME, &lpPropDisplayName);// ignore error, used for logging

	if(ulSyncId == 0) {
		// Ignore any trailing garbage in the stream
		hr = HrGetOneProp(&m_lpFolder->m_xMAPIFolder, PR_SOURCE_KEY, &lpPropSourceKey);
		if(hr != hrSuccess)
			goto exit;

		// Register our sync with the server, get a sync ID
		hr = m_lpFolder->GetMsgStore()->lpTransport->HrSetSyncStatus(lpPropSourceKey->Value.bin , 0, 0, m_ulSyncType, 0, &ulSyncId);
		if(hr != hrSuccess)
			goto exit;

		bForceImplicitStateUpdate = true;
	}

	if(m_lpChanges) {
		MAPIFreeBuffer(m_lpChanges);
		m_lpChanges = NULL;
	}

	hr = m_lpFolder->GetMsgStore()->lpTransport->HrGetChanges(lpPropSourceKey ? lpPropSourceKey->Value.bin : sbNull, ulSyncId, ulChangeId, m_ulSyncType, ulFlags, m_lpRestrict, &m_ulMaxChangeId, &m_ulChanges, &m_lpChanges);
	if(hr != hrSuccess)
		goto exit;

	m_ulSyncId = ulSyncId;
	m_ulChangeId = ulChangeId;

	m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "%s folder: %s changes: %u syncid: %u changeid: %u",(m_lpFolder->GetMsgStore()->IsOfflineStore() ? "offline" : "online"), (lpPropDisplayName)?lpPropDisplayName->Value.lpszA:"<Unknown>", m_ulChanges, m_ulSyncId, m_ulChangeId);
	
	for(ulStep = 0; ulStep < m_ulChanges; ulStep++){
		// We split the changes into message changes, flag changes, and deletes. This allows us to get a good count of items.
		// It is OK to use push_back here because the data the struct points to is available in m_lpChanges.
		
		// If we have already processed the change, completely ignore it.
		iterProcessedChange = m_setProcessedChanges.find(std::pair<unsigned int, std::string>(m_lpChanges[ulStep].ulChangeId, std::string((char *)m_lpChanges[ulStep].sSourceKey.lpb, m_lpChanges[ulStep].sSourceKey.cb)));
		if(iterProcessedChange != m_setProcessedChanges.end())
		    continue;
		if((m_lpChanges[ulStep].ulChangeType & ICS_ACTION_MASK) == ICS_CHANGE || (m_lpChanges[ulStep].ulChangeType & ICS_ACTION_MASK) == ICS_NEW){
			//delete + change = change
			iterDeleteChange = m_lstSoftDelete.begin();
			while(iterDeleteChange != m_lstSoftDelete.end()){
				if(iterDeleteChange->sSourceKey.cb == m_lpChanges[ulStep].sSourceKey.cb && memcmp(iterDeleteChange->sSourceKey.lpb, m_lpChanges[ulStep].sSourceKey.lpb, iterDeleteChange->sSourceKey.cb) == 0){
					iterDeleteChange = m_lstSoftDelete.erase(iterDeleteChange);
				}else{
					iterDeleteChange++;
				}
			}

			iterDeleteChange = m_lstHardDelete.begin();
			while(iterDeleteChange != m_lstHardDelete.end()){
				if(iterDeleteChange->sSourceKey.cb == m_lpChanges[ulStep].sSourceKey.cb && memcmp(iterDeleteChange->sSourceKey.lpb, m_lpChanges[ulStep].sSourceKey.lpb, iterDeleteChange->sSourceKey.cb) == 0){
					iterDeleteChange = m_lstHardDelete.erase(iterDeleteChange);
				}else{
					iterDeleteChange++;
				}
			}
			m_lstChange.push_back(m_lpChanges[ulStep]);
		}
		if((m_lpChanges[ulStep].ulChangeType & ICS_ACTION_MASK) == ICS_FLAG)
			m_lstFlag.push_back(m_lpChanges[ulStep]);
		if((m_lpChanges[ulStep].ulChangeType & ICS_ACTION_MASK) == ICS_HARD_DELETE)
			m_lstHardDelete.push_back(m_lpChanges[ulStep]);
		if((m_lpChanges[ulStep].ulChangeType & ICS_ACTION_MASK) == ICS_SOFT_DELETE)
			m_lstSoftDelete.push_back(m_lpChanges[ulStep]);
	}

	if(ulBufferSize != 0){
		m_ulBufferSize = ulBufferSize;
	}else{
		m_ulBufferSize = 10;
	}

	m_bConfiged = true;

	if (bForceImplicitStateUpdate) {
		ASSERT(m_ulChangeId == 0);
		UpdateState(NULL);
	}

exit:
	if(lpData)
		delete []lpData;
	
	if(lpPropDisplayName)
		MAPIFreeBuffer(lpPropDisplayName);

	if(lpPropSourceKey)
		ECFreeBuffer(lpPropSourceKey);
		
	return hr;
}

HRESULT ECExchangeExportChanges::Synchronize(ULONG FAR * lpulSteps, ULONG FAR * lpulProgress){
	HRESULT			hr = hrSuccess;
	PROCESSEDCHANGESSET::iterator iterProcessedChange;
	LPSPropValue lpPropSourceKey = NULL;
	
	if(!m_bConfiged){
		hr = MAPI_E_UNCONFIGURED;
		goto exit;
	}

	if(m_ulFlags & SYNC_CATCHUP){
		m_ulChangeId = m_ulMaxChangeId > m_ulChangeId ? m_ulMaxChangeId : m_ulChangeId;
		hr = UpdateStream(m_lpStream);
		goto exit;
	}

	if(m_ulSyncType == ICS_SYNC_CONTENTS){
		hr = ExportMessageChanges();
		if(hr == SYNC_W_PROGRESS)
			goto progress;
			
		if(hr != hrSuccess)
			goto exit;

		hr = ExportMessageDeletes();
		if(hr != hrSuccess)
			goto exit;

		hr = ExportMessageFlags();
		if(hr != hrSuccess)
			goto exit;

	}else if(m_ulSyncType == ICS_SYNC_HIERARCHY){
		hr = ExportFolderChanges();
		if(hr == SYNC_W_PROGRESS)
			goto progress;
		if(hr != hrSuccess)
			goto exit;

		hr = ExportFolderDeletes();
		if(hr != hrSuccess)
			goto exit;

	}else{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = UpdateStream(m_lpStream);
	if(hr != hrSuccess)
		goto exit;
		
	if(! (m_ulFlags & SYNC_CATCHUP)) {
		if(m_ulSyncType == ICS_SYNC_CONTENTS){
			hr = m_lpImportContents->UpdateState(NULL);
		}else{
			hr = m_lpImportHierarchy->UpdateState(NULL);
		}
		if(hr != hrSuccess)
			goto exit;
	}

progress:
	if(hr == hrSuccess) {
		// The sync has completed. This means we can move to the next change ID on the server, and clear our
		// processed change list. If we can't save the new state the the server, we just keep the previous
		// change ID and the (large) change list. This allows us always to have a state, even when we can't
		// communicate with the server.
		hr = HrGetOneProp(&m_lpFolder->m_xMAPIFolder, PR_SOURCE_KEY, &lpPropSourceKey);
		if(hr != hrSuccess)
			goto exit;

		if(m_lpFolder->GetMsgStore()->lpTransport->HrSetSyncStatus(lpPropSourceKey->Value.bin , m_ulSyncId, m_ulMaxChangeId, m_ulSyncType, 0, &m_ulSyncId) == hrSuccess) {
			m_ulChangeId = m_ulMaxChangeId;
			m_setProcessedChanges.clear();

			if(m_ulChanges)
				m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "folder changes synchronized");
			}
		}

	*lpulSteps = m_lstChange.size();
	*lpulProgress = m_ulStep;

exit:
	if(lpPropSourceKey)
		MAPIFreeBuffer(lpPropSourceKey);
		
	return hr;
}

HRESULT ECExchangeExportChanges::UpdateState(LPSTREAM lpStream){
	HRESULT hr = hrSuccess;
	
	if(!m_bConfiged){
		hr = MAPI_E_UNCONFIGURED;
		goto exit;
	}
	
	if(lpStream == NULL) {
		lpStream = m_lpStream;
	}

	hr = UpdateStream(lpStream);
	if(hr != hrSuccess)
		goto exit;


exit:
	return hr;
}

HRESULT ECExchangeExportChanges::GetChangeCount(ULONG *lpcChanges) {
	HRESULT hr = hrSuccess;
	ULONG cChanges = 0;
	
	if(!m_bConfiged){
		hr = MAPI_E_UNCONFIGURED;
		goto exit;
	}

	// Changes in flags or deletions are only one step all together
	if(!m_lstHardDelete.empty() || !m_lstSoftDelete.empty() || !m_lstFlag.empty())
	    cChanges++;
	cChanges += m_lstChange.size();
	
	*lpcChanges = cChanges;
exit:
	return hr;
}

ULONG ECExchangeExportChanges::xExchangeExportChanges::AddRef(){
	TRACE_MAPI(TRACE_ENTRY, "IExchangeExportChanges::AddRef", "");
	METHOD_PROLOGUE_(ECExchangeExportChanges, ExchangeExportChanges);
	return pThis->AddRef();
}

ULONG ECExchangeExportChanges::xExchangeExportChanges::Release()
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeExportChanges::Release", "");
	METHOD_PROLOGUE_(ECExchangeExportChanges, ExchangeExportChanges);
	return pThis->Release();
}

HRESULT ECExchangeExportChanges::xExchangeExportChanges::QueryInterface(REFIID refiid, void **pInterface)
{
	HRESULT hr;
	TRACE_MAPI(TRACE_ENTRY, "IExchangeExportChanges::QueryInterface", "%s", DBGGUIDToString(refiid).c_str());
	METHOD_PROLOGUE_(ECExchangeExportChanges, ExchangeExportChanges);
	hr = pThis->QueryInterface(refiid, pInterface);
	TRACE_MAPI(TRACE_RETURN, "IExchangeExportChanges::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECExchangeExportChanges::xExchangeExportChanges::GetLastError(HRESULT hError, ULONG ulFlags, LPMAPIERROR * lppMapiError)
{
	HRESULT hr;
	TRACE_MAPI(TRACE_ENTRY, "IExchangeExportChanges::GetLastError", "");
	METHOD_PROLOGUE_(ECExchangeExportChanges, ExchangeExportChanges);
	hr = pThis->GetLastError(hError, ulFlags, lppMapiError);
	TRACE_MAPI(TRACE_RETURN, "IExchangeExportChanges::GetLastError", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECExchangeExportChanges::xExchangeExportChanges::Config(LPSTREAM lpStream, ULONG ulFlags, LPUNKNOWN lpCollector, LPSRestriction lpRestriction, LPSPropTagArray lpIncludeProps, LPSPropTagArray lpExcludeProps, ULONG ulBufferSize){
	HRESULT hr;
	TRACE_MAPI(TRACE_ENTRY, "IExchangeExportChanges::Config", "lpStream=%08x, ulFlags=%d, lpCollector=0x%08x, lpRestriction = %08x, lpIncludeProps = %08x, lpExcludeProps = %08x, ulBufferSize = %d", lpStream, ulFlags, lpCollector, lpRestriction, lpIncludeProps, lpExcludeProps, ulBufferSize);
	METHOD_PROLOGUE_(ECExchangeExportChanges, ExchangeExportChanges);
	hr =  pThis->Config(lpStream, ulFlags, lpCollector, lpRestriction, lpIncludeProps, lpExcludeProps, ulBufferSize);
	TRACE_MAPI(TRACE_RETURN, "IExchangeExportChanges::Config", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECExchangeExportChanges::xExchangeExportChanges::Synchronize(ULONG FAR * pulSteps, ULONG FAR * pulProgress){
	HRESULT hr;
	TRACE_MAPI(TRACE_ENTRY, "IExchangeExportChanges::Synchronize", "");
	METHOD_PROLOGUE_(ECExchangeExportChanges, ExchangeExportChanges);
	hr = pThis->Synchronize(pulSteps, pulProgress);
	TRACE_MAPI(TRACE_RETURN, "IExchangeExportChanges::Synchronize", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECExchangeExportChanges::xExchangeExportChanges::UpdateState(LPSTREAM lpStream){
	HRESULT hr;
	TRACE_MAPI(TRACE_ENTRY, "IExchangeExportChanges::UpdateState", "");
	METHOD_PROLOGUE_(ECExchangeExportChanges, ExchangeExportChanges);
	hr = pThis->UpdateState(lpStream);
	TRACE_MAPI(TRACE_RETURN, "IExchangeExportChanges::UpdateState", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

ULONG ECExchangeExportChanges::xECExportChanges::AddRef() {
	TRACE_MAPI(TRACE_ENTRY, "IECExportChanges::AddRef", "");
	METHOD_PROLOGUE_(ECExchangeExportChanges, ECExportChanges);
	return pThis->AddRef();
}

ULONG ECExchangeExportChanges::xECExportChanges::Release() {
	TRACE_MAPI(TRACE_ENTRY, "IECExportChanges::Release", "");
	METHOD_PROLOGUE_(ECExchangeExportChanges, ECExportChanges);
	return pThis->Release();
}

HRESULT ECExchangeExportChanges::xECExportChanges::QueryInterface(REFIID iid, void **lpvoid) {
	TRACE_MAPI(TRACE_ENTRY, "IECExportChanges::QueryInterface", "");
	METHOD_PROLOGUE_(ECExchangeExportChanges, ECExportChanges);
	return pThis->QueryInterface(iid, lpvoid);
}

HRESULT ECExchangeExportChanges::xECExportChanges::GetChangeCount(ULONG *lpcChanges) {
	TRACE_MAPI(TRACE_ENTRY, "IECExportChanges::GetChangeCount", "");
	METHOD_PROLOGUE_(ECExchangeExportChanges, ECExportChanges);
	return pThis->GetChangeCount(lpcChanges);
}

HRESULT ECExchangeExportChanges::ExportMessageChanges(){
	HRESULT			hr = hrSuccess;
	
	LPMESSAGE		lpSourceMessage = NULL;
	LPATTACH		lpSourceAttach = NULL;

	LPMESSAGE		lpDestMessage = NULL;
	LPATTACH		lpDestAttach = NULL;

	LPMAPITABLE		lpTable = NULL;
	LPSRowSet		lpRows = NULL;

	LPSPropValue	lpPropArray = NULL;
	LPSPropValue	lpPropVal = NULL;

	LPSPropTagArray lpPropTagArray = NULL;

	ULONG			ulObjType;
	ULONG			ulCount;
	ULONG			ulFlags;
	ULONG			ulSteps = 0;
	
	ULONG			cbEntryID = 0;
	LPENTRYID		lpEntryID = NULL;

	SizedSPropTagArray(5, sptMessageExcludes) = { 5, {
		PR_MESSAGE_SIZE, 
		PR_MESSAGE_RECIPIENTS, 
		PR_MESSAGE_ATTACHMENTS, 
		PR_ATTACH_SIZE, 
		PR_PARENT_SOURCE_KEY
	} };
	SizedSPropTagArray(1, sptAttach) = { 1, {PR_ATTACH_NUM} };

	while(m_ulStep < m_lstChange.size() && (m_ulBufferSize == 0 || ulSteps < m_ulBufferSize)){
		hr = m_lpFolder->GetMsgStore()->EntryIDFromSourceKey(
			m_lstChange.at(m_ulStep).sParentSourceKey.cb, 
			m_lstChange.at(m_ulStep).sParentSourceKey.lpb,
			m_lstChange.at(m_ulStep).sSourceKey.cb,
			m_lstChange.at(m_ulStep).sSourceKey.lpb,
			&cbEntryID, &lpEntryID);
		if(hr == MAPI_E_NOT_FOUND){
			hr = hrSuccess;
			goto next;
		}
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "change sourcekey: %s", Util::bin2hex(m_lstChange.at(m_ulStep).sSourceKey.cb, m_lstChange.at(m_ulStep).sSourceKey.lpb).c_str());
		
		if(hr != hrSuccess)
			goto exit;
		
		hr = m_lpFolder->OpenEntry(cbEntryID, lpEntryID, &IID_IMessage, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*) &lpSourceMessage);
		if(hr == MAPI_E_NOT_FOUND){
			hr = hrSuccess;
			goto next;
		}
		if(hr != hrSuccess)
			goto exit;

		hr = lpSourceMessage->GetProps((LPSPropTagArray) m_lpChangePropTagArray, 0, &ulCount, &lpPropArray);
		if(hr == MAPI_W_ERRORS_RETURNED){
			hr = hrSuccess;
		}else if(hr != hrSuccess)
			goto exit;

		ulFlags = 0;
		lpPropVal = PpropFindProp(lpPropArray, ulCount, PR_MESSAGE_FLAGS);
		if(lpPropVal != NULL && (lpPropVal->Value.ul & MSGFLAG_ASSOCIATED)){
			ulFlags |= SYNC_ASSOCIATED;
		}
		if((m_lstChange.at(m_ulStep).ulChangeType & ICS_ACTION_MASK) == ICS_NEW){
			ulFlags |= SYNC_NEW_MESSAGE;
		}

		hr = m_lpImportContents->ImportMessageChange(ulCount, lpPropArray, ulFlags, &lpDestMessage);
		if (hr == SYNC_E_IGNORE) {
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "ignored change");
			// Mark this change as processed
			
			hr = hrSuccess;
			goto next;
		}else if(hr == SYNC_E_OBJECT_DELETED){
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "ignored change for deleted item");

			// Mark this change as processed
			hr = hrSuccess;
			goto next;
		}else if(hr == SYNC_E_INVALID_PARAMETER){
			//exchange doesn't like our input sometimes
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "ignored change parameter");
			hr = hrSuccess;
			goto next;
		// TODO handle SYNC_E_OBJECT_DELETED, SYNC_E_CONFLICT, SYNC_E_NO_PARENT, SYNC_E_INCEST, SYNC_E_UNSYNCHRONIZED
		}else if(hr != hrSuccess){
			//m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "change error: %s", stringify(hr, true).c_str());
			goto exit;
		} 

		if(lpDestMessage == NULL) {
			// Import succeeded, but we have not message to export to. Treat this the same as
			// SYNC_E_IGNORE. This is required for the BES ICS exporter
			goto next;
		}

		hr = lpSourceMessage->CopyTo(0, NULL, (LPSPropTagArray)&sptMessageExcludes, 0, NULL, &IID_IMessage, lpDestMessage, 0, NULL);
		if(hr != hrSuccess)
			goto exit;

		hr = lpSourceMessage->GetRecipientTable(0, &lpTable);
		if(hr !=  hrSuccess)
			goto exit;

		hr = lpTable->QueryColumns(TBL_ALL_COLUMNS, &lpPropTagArray);
		if (hr != hrSuccess)
			goto exit;

		hr = lpTable->SetColumns(lpPropTagArray, 0);
		if (hr != hrSuccess)
			goto exit;

		hr = lpTable->QueryRows(0xFFFF, 0, &lpRows);
		if(hr !=  hrSuccess)
			goto exit;
		
		//FIXME: named property in the recipienttable ?

		hr = lpDestMessage->ModifyRecipients(0, (LPADRLIST)lpRows);
		if(hr !=  hrSuccess)
			hr = hrSuccess;
			//goto exit;

		if(lpRows){
			FreeProws(lpRows);
			lpRows = NULL;
		}
		
		if(lpTable){
			lpTable->Release();
			lpTable = NULL;
		}

		//delete every attachment
		hr = lpDestMessage->GetAttachmentTable(0, &lpTable);
		if(hr !=  hrSuccess)
			goto exit;

		hr = lpTable->SetColumns((LPSPropTagArray)&sptAttach, 0);
		if(hr !=  hrSuccess)
			goto exit;

		hr = lpTable->QueryRows(0xFFFF, 0, &lpRows);
		if(hr !=  hrSuccess)
			goto exit;

		for(ulCount = 0; ulCount < lpRows->cRows; ulCount++){
			hr = lpDestMessage->DeleteAttach(lpRows->aRow[ulCount].lpProps[0].Value.ul, 0, NULL, 0);
			if(hr != hrSuccess)
				goto exit;
		}

		if(lpRows){
			FreeProws(lpRows);
			lpRows = NULL;
		}
		
		if(lpTable){
			lpTable->Release();
			lpTable = NULL;
		}

		//add every attachment
		hr = lpSourceMessage->GetAttachmentTable(0, &lpTable);
		if(hr !=  hrSuccess)
			goto exit;

		hr = lpTable->SetColumns((LPSPropTagArray)&sptAttach, 0);
		if(hr !=  hrSuccess)
			goto exit;

		hr = lpTable->QueryRows(0xFFFF, 0, &lpRows);
		if(hr !=  hrSuccess)
			goto exit;

		for(ulCount = 0; ulCount < lpRows->cRows; ulCount++){
			hr = lpSourceMessage->OpenAttach(lpRows->aRow[ulCount].lpProps[0].Value.ul, &IID_IAttachment, 0, &lpSourceAttach);
			if(hr !=  hrSuccess)
				goto exit;

			hr = lpDestMessage->CreateAttach(&IID_IAttachment, 0, &ulObjType, &lpDestAttach);
			if(hr !=  hrSuccess)
				goto exit;

			hr = lpSourceAttach->CopyTo(0, NULL, (LPSPropTagArray)&sptAttach, 0, NULL, &IID_IAttachment, lpDestAttach, 0, NULL);
			if(hr !=  hrSuccess)
				goto exit;
			
			hr = lpDestAttach->SaveChanges(0);
			if(hr !=  hrSuccess)
				goto exit;

			if(lpSourceAttach){
				lpSourceAttach->Release();
				lpSourceAttach = NULL;
			}

			if(lpDestAttach){
				lpDestAttach->Release();
				lpDestAttach = NULL;
			}
		}
		
		if(lpRows){
			FreeProws(lpRows);
			lpRows = NULL;
		}
		
		if(lpTable){
			lpTable->Release();
			lpTable = NULL;
		}

		if(lpPropTagArray){
			MAPIFreeBuffer(lpPropTagArray);
			lpPropTagArray = NULL;
		}

		hr = lpSourceMessage->GetPropList(0, &lpPropTagArray);
		if (hr != hrSuccess)
			goto exit;

		hr = Util::HrDeleteResidualProps(lpDestMessage, lpSourceMessage, lpPropTagArray);
		if (hr != hrSuccess)
			goto exit;

		hr = lpDestMessage->SaveChanges(0);
		if(hr != hrSuccess)
			goto exit;


next:
		// Mark this change as processed, even if we skipped it due to SYNC_E_IGNORE or because the item was deleted on the source server
	    
		m_setProcessedChanges.insert(std::pair<unsigned int, std::string>(m_lstChange.at(m_ulStep).ulChangeId, std::string((char *)m_lstChange.at(m_ulStep).sSourceKey.lpb, m_lstChange.at(m_ulStep).sSourceKey.cb)));
	    
		if(lpEntryID) {
			MAPIFreeBuffer(lpEntryID);
			lpEntryID = NULL;
		}
		
		if(lpRows){
			FreeProws(lpRows);
			lpRows = NULL;
		}
		
		if(lpTable){
			lpTable->Release();
			lpTable = NULL;
		}

		if(lpPropArray){
			MAPIFreeBuffer(lpPropArray);
			lpPropArray = NULL;
		}

		if(lpSourceAttach){
			lpSourceAttach->Release();
			lpSourceAttach = NULL;
		}

		if(lpDestAttach){
			lpDestAttach->Release();
			lpDestAttach = NULL;
		}
		
		if(lpSourceMessage){
			lpSourceMessage->Release();
			lpSourceMessage = NULL;
		}
		
		if(lpDestMessage){
			lpDestMessage->Release();
			lpDestMessage = NULL;
		}

		if(lpPropTagArray){
			MAPIFreeBuffer(lpPropTagArray);
			lpPropTagArray = NULL;
		}

		m_ulStep++;
		ulSteps++;
	}

	if(m_ulStep < m_lstChange.size())
		hr = SYNC_W_PROGRESS;
exit:
	if(hr != hrSuccess && hr != SYNC_W_PROGRESS)
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "change error: %s", stringify(hr, true).c_str());

	if(lpEntryID)
		MAPIFreeBuffer(lpEntryID);
		
	if(lpRows)
		FreeProws(lpRows);
	
	if(lpTable)
		lpTable->Release();

	if(lpDestAttach)
		lpDestAttach->Release();

	if(lpDestMessage)
		lpDestMessage->Release();

	if(lpSourceAttach)
		lpSourceAttach->Release();

	if(lpSourceMessage)
		lpSourceMessage->Release();

	if(lpPropArray)
		MAPIFreeBuffer(lpPropArray);

	if(lpPropTagArray)
			MAPIFreeBuffer(lpPropTagArray);

	return hr;
}

HRESULT ECExchangeExportChanges::ExportMessageFlags(){
	HRESULT			hr = hrSuccess;
	LPREADSTATE		lpReadState = NULL;
	ULONG			ulCount;
	std::list<ICSCHANGE>::iterator lpChange;
	
	if(m_lstFlag.empty())
		goto exit;

	MAPIAllocateBuffer(sizeof(READSTATE) * m_lstFlag.size(), (LPVOID *)&lpReadState);

	ulCount = 0;
	for(lpChange = m_lstFlag.begin(); lpChange != m_lstFlag.end(); lpChange++){

		MAPIAllocateMore(lpChange->sSourceKey.cb, lpReadState, (LPVOID *)&lpReadState[ulCount].pbSourceKey);
		lpReadState[ulCount].cbSourceKey = lpChange->sSourceKey.cb;
		memcpy(lpReadState[ulCount].pbSourceKey, lpChange->sSourceKey.lpb, lpChange->sSourceKey.cb );
		lpReadState[ulCount].ulFlags = lpChange->ulFlags;

		ulCount++;
	}

	if(ulCount > 0){
		hr = m_lpImportContents->ImportPerUserReadStateChange(ulCount, lpReadState);
		if (hr == SYNC_E_IGNORE){
			hr = hrSuccess;
		}
		if(hr != hrSuccess)
			goto exit;

		// Mark the flag changes as processed
		for(lpChange = m_lstFlag.begin(); lpChange != m_lstFlag.end(); lpChange++){
			m_setProcessedChanges.insert(std::pair<unsigned int, std::string>(lpChange->ulChangeId, std::string((char *)lpChange->sSourceKey.lpb, lpChange->sSourceKey.cb)));
		}
	}

exit:
	if (hr != hrSuccess)
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to sync message flags, 0x%08X", hr);

	if(lpReadState)
		MAPIFreeBuffer(lpReadState);

	return hr;
}

HRESULT ECExchangeExportChanges::ExportMessageDeletes(){
	HRESULT			hr = hrSuccess;
	
	LPENTRYLIST		lpEntryList = NULL;

	if(!m_lstSoftDelete.empty()){
		hr = ChangesToEntrylist(&m_lstSoftDelete, &lpEntryList);
		if(hr != hrSuccess)
			goto exit;

		hr = m_lpImportContents->ImportMessageDeletion(SYNC_SOFT_DELETE, lpEntryList);
		if (hr == SYNC_E_IGNORE){
			hr = hrSuccess;	
		}
		if(hr != hrSuccess)
			goto exit;
	}
	
	if(lpEntryList){
		MAPIFreeBuffer(lpEntryList);
		lpEntryList = NULL;
	}

	if(!m_lstHardDelete.empty()){
		hr = ChangesToEntrylist(&m_lstHardDelete, &lpEntryList);
		if(hr != hrSuccess)
			goto exit;

		hr = m_lpImportContents->ImportMessageDeletion(0, lpEntryList);
		if (hr == SYNC_E_IGNORE){
			hr = hrSuccess;	
		}
		if(hr != hrSuccess)
			goto exit;
	}
	
exit:
	if(lpEntryList)
		MAPIFreeBuffer(lpEntryList);
	return hr;
}

HRESULT ECExchangeExportChanges::ExportFolderChanges(){
	HRESULT			hr = hrSuccess;

	LPMAPIFOLDER	lpFolder = NULL;
	LPSPropValue	lpPropArray = NULL;
	LPSPropValue	lpPropSK = NULL;
	LPSPropValue	lpPropVal = NULL;
	LPSPropValue	lpPropFolderType = NULL;

	ULONG			ulObjType = 0;
	ULONG			ulCount = 0;
	ULONG			ulSteps = 0;
	
	ULONG			cbEntryID = 0;
	LPENTRYID		lpEntryID = NULL;
	LPSTREAM		lpStream = NULL;
	STATSTG			sStat;
	ULONG			ulProp = 0;
	ULONG			ulRead = 0;

	SizedSPropTagArray(1, sptSK) = { 1, {PR_SOURCE_KEY} };
	
	//PR_LAST_MODIFICATION_TIME missing in documentation
	//The array must contain at least the PR_PARENT_SOURCE_KEY, PR_SOURCE_KEY, PR_CHANGE_KEY, PR_PREDECESSOR_CHANGE_LIST, and MAPI PR_DISPLAY_NAME properties.
	//SizedSPropTagArray(6, sptFolder) = { 6, {PR_PARENT_SOURCE_KEY, PR_SOURCE_KEY, PR_CHANGE_KEY, PR_PREDECESSOR_CHANGE_LIST, PR_DISPLAY_NAME, PR_LAST_MODIFICATION_TIME} };

	hr = m_lpFolder->GetProps((LPSPropTagArray)&sptSK, 0, &ulCount, &lpPropSK);
	if(hr != hrSuccess)
		goto exit;

	if(!lpPropSK){
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	while(m_ulStep < m_lstChange.size() && (m_ulBufferSize == 0 || ulSteps < m_ulBufferSize)){

		hr = m_lpFolder->GetMsgStore()->EntryIDFromSourceKey(
			m_lstChange.at(m_ulStep).sSourceKey.cb, 
			m_lstChange.at(m_ulStep).sSourceKey.lpb,
			0, NULL,
			&cbEntryID, &lpEntryID);

		if(hr != hrSuccess){
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "change sourcekey not found");
			hr = hrSuccess;
			goto next;
		}
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "change sourcekey: %s", Util::bin2hex(m_lstChange.at(m_ulStep).sSourceKey.cb, m_lstChange.at(m_ulStep).sSourceKey.lpb).c_str());
		
		hr = m_lpFolder->OpenEntry(cbEntryID, lpEntryID, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*) &lpFolder);
		if(hr != hrSuccess){
			hr = hrSuccess;
			goto next;
		}

		hr = lpFolder->GetProps(m_lpChangePropTagArray /*NULL*/, 0, &ulCount, &lpPropArray);
		if(hr == MAPI_W_ERRORS_RETURNED){
			hr = hrSuccess;
		}else if(hr != hrSuccess)
			goto exit;

		// if NULL is given for include props in config function, we don't support large properties for now
		if(m_lpChangePropTagArray){
			for(ulProp = 0; ulProp < ulCount; ulProp++){
				if (PROP_TYPE(lpPropArray[ulProp].ulPropTag) != PT_ERROR || lpPropArray[ulProp].Value.err != MAPI_E_NOT_ENOUGH_MEMORY)
					goto nextprop;

				hr = lpFolder->OpenProperty(m_lpChangePropTagArray->aulPropTag[ulProp], &IID_IStream, 0, 0, (IUnknown **)&lpStream);
				if(hr != hrSuccess)
					goto nextprop;

				hr = lpStream->Stat(&sStat, 0);
				if(hr != hrSuccess)
					goto nextprop;

				switch(PROP_TYPE(m_lpChangePropTagArray->aulPropTag[ulProp])){
					case PT_STRING8:
						hr = MAPIAllocateMore(sStat.cbSize.QuadPart, lpPropArray, (LPVOID*)&lpPropArray[ulProp].Value.lpszA);
						if(hr != hrSuccess)
							goto nextprop;

						hr = lpStream->Read(lpPropArray[ulProp].Value.lpszA, sStat.cbSize.QuadPart, &ulRead);
						if(hr != hrSuccess)
							goto nextprop;
						lpPropArray[ulProp].ulPropTag = m_lpChangePropTagArray->aulPropTag[ulProp];

						break;
					case PT_BINARY:
						lpPropArray[ulProp].Value.bin.cb = 0;
						hr = MAPIAllocateMore(sStat.cbSize.QuadPart, lpPropArray, (LPVOID*)&lpPropArray[ulProp].Value.bin.lpb);
						if(hr != hrSuccess)
							goto nextprop;

						hr = lpStream->Read(lpPropArray[ulProp].Value.bin.lpb, sStat.cbSize.QuadPart, &ulRead);
						if(hr != hrSuccess)
							goto nextprop;

						lpPropArray[ulProp].ulPropTag = m_lpChangePropTagArray->aulPropTag[ulProp];
						lpPropArray[ulProp].Value.bin.cb = sStat.cbSize.QuadPart;

						break;
					default:
						break;
				}
nextprop:
				if(lpStream){
					lpStream->Release();
					lpStream = NULL;
				}
			}
		}

		lpPropFolderType = PpropFindProp(lpPropArray, ulCount, PR_FOLDER_TYPE);
		if(lpPropFolderType && lpPropFolderType->Value.ul == FOLDER_SEARCH){
			//skipping search folder
			goto next;
		}
		
		lpPropVal = PpropFindProp(lpPropArray, ulCount, PR_SOURCE_KEY);
		if(!lpPropVal){
			goto next;
		}

		//for folders directly under m_lpFolder PR_PARENT_SOURCE_KEY must be NULL
		//this protects against recursive problems during syncing when the PR_PARENT_SOURCE_KEY
		//equals the sourcekey of the folder which we are already syncing.
		//If the exporter sends an empty PR_PARENT_SOURCE_KEY to the importer the importer can
		//assume the parent is the folder which it is syncing.

		lpPropVal = PpropFindProp(lpPropArray, ulCount, PR_PARENT_SOURCE_KEY);
		if(lpPropVal && lpPropSK->Value.bin.cb == lpPropVal->Value.bin.cb && memcmp(lpPropVal->Value.bin.lpb, lpPropSK->Value.bin.lpb, lpPropSK->Value.bin.cb)==0){
			lpPropVal->Value.bin.cb = 0;
		}

		hr = m_lpImportHierarchy->ImportFolderChange(ulCount, lpPropArray);
		if (hr == SYNC_E_IGNORE){
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "change ignored");
			hr = hrSuccess;
			goto next;
		}else if (hr == MAPI_E_INVALID_PARAMETER){
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "change invalid parameter");
			hr = hrSuccess;
			goto next;
		}else if(hr == MAPI_E_NOT_FOUND){
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "change not found");
			hr = hrSuccess;
			goto next;
		}else if(hr != hrSuccess){
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "change error: %s", stringify(hr, true).c_str());
			goto exit;
		}
next:
		// Mark this change as processed
		m_setProcessedChanges.insert(std::pair<unsigned int, std::string>(m_lstChange.at(m_ulStep).ulChangeId, std::string((char *)m_lstChange.at(m_ulStep).sSourceKey.lpb, m_lstChange.at(m_ulStep).sSourceKey.cb)));

		if(lpFolder){
			lpFolder->Release();
			lpFolder = NULL;
		}

		if(lpPropArray){
			MAPIFreeBuffer(lpPropArray);
			lpPropArray = NULL;
		}

		if(lpEntryID) {
			MAPIFreeBuffer(lpEntryID);
			lpEntryID = NULL;
		}
		
		ulSteps++;
		m_ulStep++;
	}
	
	if(m_ulStep < m_lstChange.size())
		hr = SYNC_W_PROGRESS;

exit:
	if(lpStream)
		lpStream->Release();

	if(lpEntryID)
		MAPIFreeBuffer(lpEntryID);
		
	if(lpFolder)
		lpFolder->Release();

	if(lpPropArray)
		MAPIFreeBuffer(lpPropArray);

	return hr;
}

HRESULT ECExchangeExportChanges::ExportFolderDeletes(){
	HRESULT			hr = hrSuccess;
	
	LPENTRYLIST		lpEntryList = NULL;

	if(!m_lstSoftDelete.empty()){
		hr = ChangesToEntrylist(&m_lstSoftDelete, &lpEntryList);
		if(hr != hrSuccess)
			goto exit;

		hr = m_lpImportHierarchy->ImportFolderDeletion(SYNC_SOFT_DELETE, lpEntryList);
		if (hr == SYNC_E_IGNORE){
			hr = hrSuccess;	
		}
		if(hr != hrSuccess)
			goto exit;
	}
	
	if(lpEntryList){
		MAPIFreeBuffer(lpEntryList);
		lpEntryList = NULL;
	}

	if(!m_lstHardDelete.empty()){
		hr = ChangesToEntrylist(&m_lstHardDelete, &lpEntryList);
		if(hr != hrSuccess)
			goto exit;

		hr = m_lpImportHierarchy->ImportFolderDeletion(0, lpEntryList);
		if (hr == SYNC_E_IGNORE){
			hr = hrSuccess;	
		}
		if(hr != hrSuccess)
			goto exit;
	}
	
exit:
	if(lpEntryList)
		MAPIFreeBuffer(lpEntryList);

	return hr;
}

//write in the stream 4 bytes syncid, 4 bytes changeid, 4 bytes changecount, {4 bytes changeid, 4 bytes sourcekeysize, sourcekey}
HRESULT ECExchangeExportChanges::UpdateStream(LPSTREAM lpStream){
	HRESULT hr;
	LARGE_INTEGER liPos = {{0, 0}};
	ULARGE_INTEGER liZero = {{0, 0}};
	ULONG ulSize;
	ULONG ulChangeCount = 0;
	ULONG ulChangeId = 0;
	ULONG ulSourceKeySize = 0;
	PROCESSEDCHANGESSET::iterator iterProcessedChange;

	hr = lpStream->SetSize(liZero);
	if(hr != hrSuccess)
		goto exit;

	hr = lpStream->Seek(liPos, STREAM_SEEK_SET, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = lpStream->Write(&m_ulSyncId, 4, &ulSize);
	if(hr != hrSuccess)
		goto exit;

	if(m_ulSyncId == 0){
		m_ulChangeId = 0;
	}

	hr = lpStream->Write(&m_ulChangeId, 4, &ulSize);
	if(hr != hrSuccess)
		goto exit;

	if(!m_setProcessedChanges.empty()) {
		ulChangeCount = m_setProcessedChanges.size();
		
		hr = lpStream->Write(&ulChangeCount, 4, &ulSize);
		if(hr != hrSuccess)
			goto exit;
		
		for(iterProcessedChange = m_setProcessedChanges.begin(); iterProcessedChange != m_setProcessedChanges.end(); iterProcessedChange++) {
			ulChangeId = iterProcessedChange->first;
			hr = lpStream->Write(&ulChangeId, 4, &ulSize);
			if(hr != hrSuccess)
				goto exit;
			
			ulSourceKeySize = iterProcessedChange->second.size();
			
			hr = lpStream->Write(&ulSourceKeySize, 4, &ulSize);
			if(hr != hrSuccess)
				goto exit;
				
			hr = lpStream->Write(iterProcessedChange->second.c_str(), iterProcessedChange->second.size(), &ulSize);
			if(hr != hrSuccess)
				goto exit;
		}
	}

	// Seek back to the beginning after we've finished
	lpStream->Seek(liPos, 0, NULL);

exit:
	return hr;
}

//convert (delete) changes to entrylist for message and folder deletion.
HRESULT ECExchangeExportChanges::ChangesToEntrylist(std::list<ICSCHANGE> * lpLstChanges, LPENTRYLIST * lppEntryList){
	HRESULT 		hr = hrSuccess;
	LPENTRYLIST		lpEntryList = NULL;
	ULONG			ulCount = 0;
	std::list<ICSCHANGE>::iterator lpChange;
	
	MAPIAllocateBuffer(sizeof(ENTRYLIST), (LPVOID *)&lpEntryList);
	lpEntryList->cValues = lpLstChanges->size();
	if(lpEntryList->cValues > 0){
		MAPIAllocateMore(sizeof(SBinary) * lpEntryList->cValues, lpEntryList, (LPVOID *)&lpEntryList->lpbin);
	}else{
		lpEntryList->lpbin = NULL;
	}
	ulCount = 0;
	for(lpChange = lpLstChanges->begin(); lpChange != lpLstChanges->end(); lpChange++){
		m_setProcessedChanges.insert(std::pair<unsigned int, std::string>(lpChange->ulChangeId, std::string((char *)lpChange->sSourceKey.lpb, lpChange->sSourceKey.cb)));

		lpEntryList->lpbin[ulCount].cb = lpChange->sSourceKey.cb;
		MAPIAllocateMore(lpChange->sSourceKey.cb, lpEntryList, (void **)&lpEntryList->lpbin[ulCount].lpb);
		memcpy(lpEntryList->lpbin[ulCount].lpb, lpChange->sSourceKey.lpb, lpChange->sSourceKey.cb);
		ulCount++;
	}
	
	lpEntryList->cValues = ulCount;

	*lppEntryList = lpEntryList;

	if(hr != hrSuccess && lpEntryList)
		MAPIFreeBuffer(lpEntryList);

	return hr;
}
