/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "ECExchangeImportContentsChanges.h"

#include "Util.h"
#include "guid.h"
#include "edkguid.h"
#include "ECGuid.h"
#include "mapiguid.h"
#include "ECMessage.h"
#include "ECDebug.h"
#include "ECDebugClient.h"
#include "ZarafaICS.h"
#include "mapiext.h"

#include <mapiutil.h>

#ifndef NO_GETTEXT
#include <libintl.h>
#define _(string) dcgettext ("zarafa", string, LC_MESSAGES)
#else
#define _(string) string
#endif

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

ECExchangeImportContentsChanges::ECExchangeImportContentsChanges(ECMAPIFolder *lpFolder){
	m_lpFolder = lpFolder;
	m_lpStream = NULL;
	m_lpFolder->AddRef();
}

ECExchangeImportContentsChanges::~ECExchangeImportContentsChanges(){
	m_lpFolder->Release();
}

HRESULT ECExchangeImportContentsChanges::Create(ECMAPIFolder *lpFolder, LPEXCHANGEIMPORTCONTENTSCHANGES* lppExchangeImportContentsChanges){
	HRESULT hr = hrSuccess;
	ECExchangeImportContentsChanges *lpEICC = NULL;

	if(!lpFolder)
		return MAPI_E_INVALID_PARAMETER;

	lpEICC = new ECExchangeImportContentsChanges(lpFolder);

	hr = lpEICC->QueryInterface(IID_IExchangeImportContentsChanges, (void **)lppExchangeImportContentsChanges);

	return hr;

}

HRESULT	ECExchangeImportContentsChanges::QueryInterface(REFIID refiid, void **lpvoid)
{

	if(refiid == IID_ECExchangeImportContentsChanges || refiid == IID_ECUnknown) {
		*lpvoid = this;
		AddRef();
		return hrSuccess;
	}

	
	if(refiid == IID_IExchangeImportContentsChanges || refiid == IID_IUnknown) {
		*lpvoid = &this->m_xExchangeImportContentsChanges;
		AddRef();
		return hrSuccess;
	}

	return MAPI_E_INTERFACE_NOT_SUPPORTED;
}

HRESULT ECExchangeImportContentsChanges::GetLastError(HRESULT hResult, ULONG ulFlags, LPMAPIERROR *lppMAPIError){
	HRESULT		hr = hrSuccess;
	LPMAPIERROR	lpMapiError = NULL;
	
	MAPIAllocateBuffer(sizeof(MAPIERROR),(void **)&lpMapiError);

	lpMapiError->lpszComponent	= "Zarafa Outlook Sharing";
	//FIXME: give synchronization errors messages
	lpMapiError->lpszError		= Util::HrMAPIErrorToText(hResult);
	lpMapiError->ulContext		= 0;
	lpMapiError->ulLowLevelError= 0;
	lpMapiError->ulVersion		= 0;

	*lppMAPIError = lpMapiError;

	return hr;
}

HRESULT ECExchangeImportContentsChanges::Config(LPSTREAM lpStream, ULONG ulFlags){
	HRESULT hr = hrSuccess;
	LARGE_INTEGER zero = {{0,0}};
	ULONG ulLen = 0;
	LPSPropValue lpPropSourceKey = NULL;
	
	m_lpStream = lpStream;

	if(lpStream == NULL) {
		m_ulSyncId = 0;
		m_ulChangeId = 0;
	} else {
		hr = lpStream->Seek(zero, STREAM_SEEK_SET, NULL);
		if(hr != hrSuccess)
			goto exit;
		
		hr = lpStream->Read(&m_ulSyncId, 4, &ulLen);
		if(hr != hrSuccess)
			goto exit;
			
		if(ulLen != 4) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		
		hr = lpStream->Read(&m_ulChangeId, 4, &ulLen);
		if(hr != hrSuccess)
			goto exit;
			
		if(ulLen != 4) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
		
		hr = HrGetOneProp(&m_lpFolder->m_xMAPIFolder, PR_SOURCE_KEY, &lpPropSourceKey);
		if(hr != hrSuccess)
			goto exit;
		
		// The user specified the special sync key '0000000000000000', get a sync key from the server.
		if(m_ulSyncId == 0) {
			hr = m_lpFolder->GetMsgStore()->lpTransport->HrSetSyncStatus(lpPropSourceKey->Value.bin, m_ulSyncId, m_ulChangeId, ICS_SYNC_CONTENTS, 0, &m_ulSyncId);
			if(hr != hrSuccess)
				goto exit;
		}
		
		// The sync key we got from the server can be used to retrieve all items in the database now when given to IEEC->Config(). At the same time, any
		// items written to this importer will send the sync ID to the server so that any items written here will not be returned by the exporter,
		// preventing local looping of items.
	}		
		
	m_ulFlags = ulFlags;
exit:	
	if(lpPropSourceKey)
		MAPIFreeBuffer(lpPropSourceKey);
		
	return hrSuccess;
}

HRESULT ECExchangeImportContentsChanges::UpdateState(LPSTREAM lpStream){
	HRESULT hr = hrSuccess;
	LARGE_INTEGER zero = {{0,0}};
	ULONG ulLen = 0;

	if(lpStream == NULL) {
		if(m_lpStream == NULL){
			goto exit;
		}
		lpStream = m_lpStream;
	}

	if(m_ulSyncId == 0)
		goto exit; // config() called with NULL stream, so we'll ignore the UpdateState()
	
	hr = lpStream->Seek(zero, STREAM_SEEK_SET, NULL);
	if(hr != hrSuccess)
		goto exit;
		
	hr = lpStream->Write(&m_ulSyncId, 4, &ulLen);
	if(hr != hrSuccess)
		goto exit;
		
	hr = lpStream->Write(&m_ulChangeId, 4, &ulLen);
	if(hr != hrSuccess)
		goto exit;
	 
exit: 
	return hr;
}


HRESULT ECExchangeImportContentsChanges::ImportMessageChange(ULONG cValue, LPSPropValue lpPropArray, ULONG ulFlags, LPMESSAGE * lppMessage){
	HRESULT hr = hrSuccess; 

	SizedSPropTagArray(1, sptSourceKey) = { 1, {PR_SOURCE_KEY} };
	
	LPSPropValue lpFolderSourceKey = NULL;
	LPSPropValue lpPropPCL = NULL;
	LPSPropValue lpPropCK = NULL;
	ULONG cbEntryId = 0;
	LPENTRYID lpEntryId = NULL;
	LPSPropValue lpPassedEntryId = NULL;

	LPSPropValue lpMessageSourceKey = PpropFindProp(lpPropArray, cValue, PR_SOURCE_KEY);
	LPSPropValue lpMessageFlags = PpropFindProp(lpPropArray, cValue, PR_MESSAGE_FLAGS);
	LPSPropValue lpMessageAssociated = PpropFindProp(lpPropArray, cValue, 0x67AA000B); //PR_ASSOCIATED
	
	LPSPropValue lpRemotePCL = PpropFindProp(lpPropArray, cValue, PR_PREDECESSOR_CHANGE_LIST);
	LPSPropValue lpRemoteCK = PpropFindProp(lpPropArray, cValue, PR_CHANGE_KEY);

	ULONG ulObjType = 0;
	ULONG ulCount = 0;
	bool bAssociatedMessage = false;
	bool bConflict = false;
	IMessage *lpMessage = NULL;
	ECMessage *lpECMessage = NULL;

	std::string strChangeList;
	ULONG ulPos = 0;
	ULONG ulSize;
	ULONG ulNewFlags = 0;

	hr = m_lpFolder->GetProps((LPSPropTagArray)&sptSourceKey,0, &ulCount, &lpFolderSourceKey);
	if(hr != hrSuccess)
		goto exit;

	if(lpFolderSourceKey->ulPropTag != PR_SOURCE_KEY) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	if(lpMessageSourceKey != NULL) {
		hr = m_lpFolder->GetMsgStore()->lpTransport->HrEntryIDFromSourceKey(m_lpFolder->GetMsgStore()->m_cbEntryId, m_lpFolder->GetMsgStore()->m_lpEntryId, lpFolderSourceKey->Value.bin.cb, lpFolderSourceKey->Value.bin.lpb, lpMessageSourceKey->Value.bin.cb, lpMessageSourceKey->Value.bin.lpb, &cbEntryId, &lpEntryId);
		if(hr != MAPI_E_NOT_FOUND && hr != hrSuccess)
			goto exit;
	} else {
	    // Source key not specified, therefore the message must be new since this is the only thing
	    // we can do if there is no sourcekey. Z-Push uses this, while offline ICS does not (it always
	    // passes a source key)
	    ulFlags |= SYNC_NEW_MESSAGE;
		hr = MAPI_E_NOT_FOUND;
	}

	if((hr == MAPI_E_NOT_FOUND) && ((ulFlags & SYNC_NEW_MESSAGE) == 0)) {
		// This is a change, but we don't already have the item. This can only mean
		// that the item has been deleted on our side. 
		hr = SYNC_E_OBJECT_DELETED;
		goto exit;
	}

	if((lpMessageFlags != NULL && (lpMessageFlags->Value.ul & MSGFLAG_ASSOCIATED)) || (lpMessageAssociated != NULL && lpMessageAssociated->Value.b)) {
		bAssociatedMessage = true;
	}

	if(hr == MAPI_E_NOT_FOUND){
		if(bAssociatedMessage == true){
		    ulNewFlags = MAPI_ASSOCIATED;
		}else{
		    ulNewFlags = 0;
		}

		lpPassedEntryId = PpropFindProp(lpPropArray, cValue, PR_ENTRYID);

		// Create the message with the passed entry ID
		if(lpPassedEntryId)
			hr = m_lpFolder->CreateMessageWithEntryID(&IID_IMessage, ulNewFlags, lpPassedEntryId->Value.bin.cb, (LPENTRYID)lpPassedEntryId->Value.bin.lpb, &lpMessage);
		else
			hr = m_lpFolder->CreateMessage(&IID_IMessage, ulNewFlags, &lpMessage);

		if(hr != hrSuccess)
			goto exit;
	}else{
		hr = m_lpFolder->OpenEntry(cbEntryId, lpEntryId, &IID_IMessage, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpMessage);
		if(hr == MAPI_E_NOT_FOUND) {
			// The item was soft-deleted; sourcekey is known, but we cannot open the item. It has therefore been deleted.
			hr = SYNC_E_OBJECT_DELETED;
			goto exit;
		}

		if(hr != hrSuccess)
			goto exit;

		if(lpRemoteCK && HrGetOneProp(lpMessage, PR_PREDECESSOR_CHANGE_LIST, &lpPropPCL) == hrSuccess){
			strChangeList.assign((char *)lpPropPCL->Value.bin.lpb, lpPropPCL->Value.bin.cb);
			ulPos = 0;
			while(ulPos < strChangeList.size()){
				ulSize = strChangeList.at(ulPos);
				if(ulSize <= sizeof(GUID)){
					break;
				}else if(lpRemoteCK->Value.bin.cb > sizeof(GUID) && memcmp(strChangeList.substr(ulPos+1, ulSize).c_str(), lpRemoteCK->Value.bin.lpb, sizeof(GUID)) == 0){
					if(ulSize == lpRemoteCK->Value.bin.cb && memcmp(strChangeList.substr(ulPos+1, ulSize).c_str(), lpRemoteCK->Value.bin.lpb, ulSize) == 0){
						//remote changekey in our changelist
						//we already have this change
						hr = SYNC_E_IGNORE;
						goto exit;
					}
				}
				ulPos += ulSize + 1;
			}
		}
		
		// Check for conflicts except for associated messages, take always the lastone
		if(bAssociatedMessage == false && lpRemotePCL && HrGetOneProp(lpMessage, PR_CHANGE_KEY, &lpPropCK) == hrSuccess){
			strChangeList.assign((char *)lpRemotePCL->Value.bin.lpb, lpRemotePCL->Value.bin.cb);
			ulPos = 0;
			while(ulPos < strChangeList.size()){
				ulSize = strChangeList.at(ulPos);
				if(ulSize <= sizeof(GUID)){
					break;
				}else if(lpPropCK->Value.bin.cb > sizeof(GUID) && memcmp(strChangeList.substr(ulPos+1, ulSize).c_str(), lpPropCK->Value.bin.lpb, sizeof(GUID)) == 0){
					/* How this works
					 *
					 * We get
					 * 1) the remote (exporter's) predecessor change list
					 * 2) the item that will be modified (importer's) change key
					 *
					 * We then look at the remote change list, and find a change with the same GUID as the local change
					 *
					 * If the trailing part (which increases with each change) is higher locally than is on the remote change list,
					 * then there's a conflict since we have a newer version than the remote server has, and the remote server is sending
					 * us a change.
					 * 
					 */
					unsigned int ulRemoteChangeNumber = 0;
					unsigned int ulLocalChangeNumber = 0;

					ulRemoteChangeNumber = *(unsigned int *)strChangeList.substr(ulPos+1+sizeof(GUID), sizeof(unsigned int)).c_str(); 
					ulLocalChangeNumber = *(unsigned int *)(lpPropCK->Value.bin.lpb+16);

					// We have a conflict if we have a newer change locally than the remove server is sending us.
					bConflict = ulLocalChangeNumber > ulRemoteChangeNumber;

					if(bConflict)
						break;
				}
				ulPos += ulSize + 1;
			}
		}
	}

	if(bConflict){
		if(CreateConflictMessage(lpMessage) == MAPI_E_NOT_FOUND){
			CreateConflictFolders();
			CreateConflictMessage(lpMessage);
		}
	}

	hr = lpMessage->QueryInterface(IID_ECMessage, (void **)&lpECMessage);
	if(hr != hrSuccess)
		goto exit;
		
	hr = lpECMessage->HrSetSyncId(m_ulSyncId);
	if(hr != hrSuccess)
		goto exit;

	// Mark as ICS object
	hr = lpECMessage->SetICSObject(TRUE);
	if(hr != hrSuccess)
		goto exit;

	hr = lpMessage->SetProps(cValue, lpPropArray, NULL);
	if(hr != hrSuccess)
		goto exit;
		
	*lppMessage = lpMessage;

exit:
	if(lpECMessage)
		lpECMessage->Release();
		
	if(lpFolderSourceKey)
		MAPIFreeBuffer(lpFolderSourceKey);

	if(lpPropPCL)
		MAPIFreeBuffer(lpPropPCL);

	if(lpPropCK)
		MAPIFreeBuffer(lpPropCK);

	if(lpEntryId)
		MAPIFreeBuffer(lpEntryId);

	return hr;
}

//ulFlags = SYNC_SOFT_DELETE, SYNC_EXPIRY
HRESULT ECExchangeImportContentsChanges::ImportMessageDeletion(ULONG ulFlags, LPENTRYLIST lpSourceEntryList){
	HRESULT hr = hrSuccess;
	ENTRYLIST EntryList;
	LPSPropValue lpPropVal = NULL;
	ULONG ulSKNr, ulCount;
	SizedSPropTagArray(1, spt) = { 1, {PR_SOURCE_KEY} };
	EntryList.lpbin = NULL;
	EntryList.cValues = 0;
	
	hr = m_lpFolder->GetProps((LPSPropTagArray)&spt,0, &ulCount, &lpPropVal);
	if(hr != hrSuccess)
		goto exit;


	MAPIAllocateBuffer(sizeof(SBinary)* lpSourceEntryList->cValues, (LPVOID*)&EntryList.lpbin);
	for(ulSKNr = 0; ulSKNr < lpSourceEntryList->cValues; ulSKNr++){
		hr = m_lpFolder->GetMsgStore()->lpTransport->HrEntryIDFromSourceKey(m_lpFolder->GetMsgStore()->m_cbEntryId, m_lpFolder->GetMsgStore()->m_lpEntryId, lpPropVal->Value.bin.cb, lpPropVal->Value.bin.lpb, lpSourceEntryList->lpbin[ulSKNr].cb, lpSourceEntryList->lpbin[ulSKNr].lpb, &EntryList.lpbin[EntryList.cValues].cb, (LPENTRYID*)&EntryList.lpbin[EntryList.cValues].lpb);
		if(hr == MAPI_E_NOT_FOUND){
			hr = hrSuccess;
			continue;
		}
		if(hr != hrSuccess)
			goto exit;
		EntryList.cValues++;
	}
	
	if(EntryList.cValues == 0)
		goto exit;

	hr = m_lpFolder->GetMsgStore()->lpTransport->HrDeleteObjects(ulFlags & SYNC_SOFT_DELETE ? 0 : DELETE_HARD_DELETE, &EntryList, m_ulSyncId);
	if(hr != hrSuccess)
		goto exit;

exit:
	if(EntryList.lpbin){
		for(ulSKNr = 0; ulSKNr < EntryList.cValues; ulSKNr++){
			MAPIFreeBuffer(EntryList.lpbin[ulSKNr].lpb);
		}
		MAPIFreeBuffer(EntryList.lpbin);
	}
	return hr;
}

HRESULT ECExchangeImportContentsChanges::ImportPerUserReadStateChange(ULONG cElements, LPREADSTATE lpReadState){
	HRESULT hr = hrSuccess;	
	LPSPropValue lpPropVal = NULL;
	ULONG ulSKNr, ulCount, cbEntryId;
	LPENTRYID lpEntryId = NULL;
	SizedSPropTagArray(1, spt) = { 1, {PR_SOURCE_KEY} };
	
	hr = m_lpFolder->GetProps((LPSPropTagArray)&spt,0, &ulCount, &lpPropVal);
	if(hr != hrSuccess)
		goto exit;

	for(ulSKNr = 0; ulSKNr < cElements; ulSKNr++){
		hr = m_lpFolder->GetMsgStore()->lpTransport->HrEntryIDFromSourceKey(m_lpFolder->GetMsgStore()->m_cbEntryId, m_lpFolder->GetMsgStore()->m_lpEntryId , lpPropVal->Value.bin.cb, lpPropVal->Value.bin.lpb, lpReadState[ulSKNr].cbSourceKey, lpReadState[ulSKNr].pbSourceKey, &cbEntryId, &lpEntryId);
		if(hr == MAPI_E_NOT_FOUND){
			hr = hrSuccess;
			continue; // Message is delete or moved
		}

		if(hr != hrSuccess) {
			goto exit;
		}

		hr = m_lpFolder->GetMsgStore()->lpTransport->HrSetReadFlag(cbEntryId, lpEntryId, lpReadState[ulSKNr].ulFlags & MSGFLAG_READ ? 0 : CLEAR_READ_FLAG, m_ulSyncId);
		if(hr != hrSuccess)
			goto exit;

		if(lpEntryId){
			MAPIFreeBuffer(lpEntryId);
			lpEntryId = NULL;
		}
	}

exit:
	if(lpEntryId)
		MAPIFreeBuffer(lpEntryId);

	return hr;
}

HRESULT ECExchangeImportContentsChanges::ImportMessageMove(ULONG cbSourceKeySrcFolder, BYTE FAR * pbSourceKeySrcFolder, ULONG cbSourceKeySrcMessage, BYTE FAR * pbSourceKeySrcMessage, ULONG cbPCLMessage, BYTE FAR * pbPCLMessage, ULONG cbSourceKeyDestMessage, BYTE FAR * pbSourceKeyDestMessage, ULONG cbChangeNumDestMessage, BYTE FAR * pbChangeNumDestMessage){
	return MAPI_E_NO_SUPPORT;
}

HRESULT ECExchangeImportContentsChanges::CreateConflictMessage(LPMESSAGE lpMessage){
	HRESULT hr = hrSuccess;

	LPMAPIFOLDER lpRootFolder = NULL;
	LPMAPIFOLDER lpConflictFolder = NULL;
	LPMESSAGE lpConflictMessage = NULL;
	LPSPropValue lpPropAdditionalREN = NULL;
	LPSPropValue lpConflictItems = NULL;
	LPSPropValue lpEntryIdProp = NULL;
	LPSBinary lpEntryIds = NULL;
	ULONG ulCount = 0;
	ULONG ulObjType = 0;
	SizedSPropTagArray(5, excludeProps) = { 5, {PR_ENTRYID, PR_CONFLICT_ITEMS, PR_SOURCE_KEY, PR_CHANGE_KEY, PR_PREDECESSOR_CHANGE_LIST} };

	//open the conflicts folder
	hr = m_lpFolder->GetMsgStore()->OpenEntry(0, NULL, &IID_IMAPIFolder, 0, &ulObjType, (LPUNKNOWN*)&lpRootFolder);
	if(hr != hrSuccess)
		goto exit;

	hr = HrGetOneProp(lpRootFolder, PR_ADDITIONAL_REN_ENTRYIDS, &lpPropAdditionalREN);
	if(hr != hrSuccess)
		goto exit;

	if(lpPropAdditionalREN->Value.MVbin.cValues == 0 || lpPropAdditionalREN->Value.MVbin.lpbin[0].cb == 0){
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = m_lpFolder->GetMsgStore()->OpenEntry(lpPropAdditionalREN->Value.MVbin.lpbin[0].cb, (LPENTRYID)lpPropAdditionalREN->Value.MVbin.lpbin[0].lpb, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpConflictFolder);
	if(hr != hrSuccess)
		goto exit;

	//create the conflict message
	hr = lpConflictFolder->CreateMessage(NULL, 0, &lpConflictMessage);
	if(hr != hrSuccess)
		goto exit;

	hr = lpMessage->CopyTo(0, NULL, (LPSPropTagArray)&excludeProps, 0, NULL, &IID_IMessage, lpConflictMessage, 0, NULL);
	if(hr != hrSuccess)
		goto exit;

	//set the entryid from original message in PR_CONFLICT_ITEMS of conflict message
	hr = HrGetOneProp(lpMessage, PR_ENTRYID ,&lpEntryIdProp);
	if(hr != hrSuccess)
		goto exit;

	hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID*)&lpConflictItems);
	if(hr != hrSuccess)
		goto exit;

	lpConflictItems->ulPropTag = PR_CONFLICT_ITEMS;
	lpConflictItems->Value.MVbin.cValues = 1;
	lpConflictItems->Value.MVbin.lpbin = &lpEntryIdProp->Value.bin;

	hr = HrSetOneProp(lpConflictMessage, lpConflictItems);
	if(hr != hrSuccess)
		goto exit;

	hr = lpConflictMessage->SaveChanges(KEEP_OPEN_READWRITE);
	if(hr != hrSuccess)
		goto exit;

	if(lpEntryIdProp){
		MAPIFreeBuffer(lpEntryIdProp);
		lpEntryIdProp = NULL;
	}

	if(lpConflictItems){
		MAPIFreeBuffer(lpConflictItems);
		lpConflictItems = NULL;
	}

	//add the entryid from the conflict message to the PR_CONFLICT_ITEMS of the original message
	hr = HrGetOneProp(lpConflictMessage, PR_ENTRYID, &lpEntryIdProp);
	if(hr != hrSuccess)
		goto exit;

	if(hrSuccess != HrGetOneProp(lpMessage, PR_CONFLICT_ITEMS, &lpConflictItems)){
		hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID*)&lpConflictItems);
		if(hr != hrSuccess)
			goto exit;

		lpConflictItems->ulPropTag = PR_CONFLICT_ITEMS;
		lpConflictItems->Value.MVbin.cValues = 0;
		lpConflictItems->Value.MVbin.lpbin = NULL;
	}
	
	hr = MAPIAllocateBuffer(sizeof(SBinary)*(lpConflictItems->Value.MVbin.cValues+1), (LPVOID*)&lpEntryIds);
	if(hr != hrSuccess)
		goto exit;

	for(ulCount = 0; ulCount < lpConflictItems->Value.MVbin.cValues; ulCount++){
		lpEntryIds[ulCount].cb = lpConflictItems->Value.MVbin.lpbin[ulCount].cb;
		lpEntryIds[ulCount].lpb = lpConflictItems->Value.MVbin.lpbin[ulCount].lpb;
	}
	lpEntryIds[ulCount].cb = lpEntryIdProp->Value.bin.cb;
	lpEntryIds[ulCount].lpb = lpEntryIdProp->Value.bin.lpb;

	lpConflictItems->Value.MVbin.lpbin = lpEntryIds;
	lpConflictItems->Value.MVbin.cValues++;

	hr = HrSetOneProp(lpMessage, lpConflictItems);
	if(hr != hrSuccess)
		goto exit;

	hr = lpMessage->SaveChanges(KEEP_OPEN_READWRITE);
	if(hr != hrSuccess)
		goto exit;


exit:
	if(lpRootFolder)
		lpRootFolder->Release();

	if(lpConflictFolder)
		lpConflictFolder->Release();

	if(lpConflictMessage)
		lpConflictMessage->Release();

	if(lpPropAdditionalREN)
		MAPIFreeBuffer(lpPropAdditionalREN);

	if(lpConflictItems)
		MAPIFreeBuffer(lpConflictItems);

	if(lpEntryIdProp)
		MAPIFreeBuffer(lpEntryIdProp);

	if(lpEntryIds)
		MAPIFreeBuffer(lpEntryIds);

	return hr;
}


HRESULT ECExchangeImportContentsChanges::CreateConflictFolders(){
	HRESULT hr = hrSuccess;
	LPMAPIFOLDER lpRootFolder = NULL;
	LPMAPIFOLDER lpParentFolder = NULL;
	LPMAPIFOLDER lpInbox = NULL;
	LPMAPIFOLDER lpConflictFolder = NULL;
	LPSPropValue lpAdditionalREN = NULL;
	LPSPropValue lpNewAdditionalREN = NULL;
	LPSPropValue lpIPMSubTree = NULL;
	LPENTRYID lpEntryId = NULL;
	ULONG cbEntryId = 0;
	ULONG ulObjType = 0;
	ULONG ulCount = 0;

	hr = m_lpFolder->OpenEntry(0, NULL, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpRootFolder);
	if(hr != hrSuccess)
		goto exit;

	hr = m_lpFolder->GetMsgStore()->GetReceiveFolder("IPM", 0, &cbEntryId, &lpEntryId, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = m_lpFolder->OpenEntry(cbEntryId, lpEntryId, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpInbox);
	if(hr != hrSuccess)
		goto exit;

	hr = HrGetOneProp(&m_lpFolder->GetMsgStore()->m_xMsgStore, PR_IPM_SUBTREE_ENTRYID, &lpIPMSubTree);
	if(hr != hrSuccess)
		goto exit;

	hr = m_lpFolder->OpenEntry(lpIPMSubTree->Value.bin.cb, (LPENTRYID)lpIPMSubTree->Value.bin.lpb, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpParentFolder);
	if(hr != hrSuccess)
		goto exit;

	HrGetOneProp(lpRootFolder, PR_ADDITIONAL_REN_ENTRYIDS, &lpAdditionalREN);

	//make new PR_ADDITIONAL_REN_ENTRYIDS
	hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID*)&lpNewAdditionalREN);
	if(hr != hrSuccess)
		goto exit;

	lpNewAdditionalREN->ulPropTag = PR_ADDITIONAL_REN_ENTRYIDS;
	lpNewAdditionalREN->Value.MVbin.cValues = (lpAdditionalREN && lpAdditionalREN->Value.MVbin.cValues<4)?4:lpAdditionalREN->Value.MVbin.cValues;
	hr = MAPIAllocateMore(sizeof(SBinary)*lpNewAdditionalREN->Value.MVbin.cValues, lpNewAdditionalREN, (LPVOID*)&lpNewAdditionalREN->Value.MVbin.lpbin);
	if(hr != hrSuccess)
		goto exit;

	//copy from original PR_ADDITIONAL_REN_ENTRYIDS
	if(lpAdditionalREN){
		for(ulCount = 0; ulCount < lpAdditionalREN->Value.MVbin.cValues; ulCount++){
			lpNewAdditionalREN->Value.MVbin.lpbin[ulCount] = lpAdditionalREN->Value.MVbin.lpbin[ulCount];
		}
	}

	hr = CreateConflictFolder(_("Sync Issues"), lpNewAdditionalREN, 1, lpParentFolder, &lpConflictFolder);
	if(hr != hrSuccess)
		goto exit;
	
	hr = CreateConflictFolder(_("Conflicts"), lpNewAdditionalREN, 0, lpConflictFolder, NULL);
	if(hr != hrSuccess)
		goto exit;
	
	hr = CreateConflictFolder(_("Local Failures"), lpNewAdditionalREN, 2, lpConflictFolder, NULL);
	if(hr != hrSuccess)
		goto exit;
	
	hr = CreateConflictFolder(_("Server Failures"), lpNewAdditionalREN, 3, lpConflictFolder, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = HrSetOneProp(lpRootFolder, lpNewAdditionalREN);
	if(hr != hrSuccess)
		goto exit;

	hr = HrSetOneProp(lpInbox, lpNewAdditionalREN);
	if(hr != hrSuccess)
		goto exit;

exit:
	if(lpRootFolder)
		lpRootFolder->Release();
	if(lpParentFolder)
		lpParentFolder->Release();
	if(lpInbox)
		lpInbox->Release();

	if(lpConflictFolder)
		lpConflictFolder->Release();
	if(lpAdditionalREN)
		MAPIFreeBuffer(lpAdditionalREN);
	if(lpNewAdditionalREN)
		MAPIFreeBuffer(lpNewAdditionalREN);
	if(lpIPMSubTree)
		MAPIFreeBuffer(lpIPMSubTree);
	if(lpEntryId)
		MAPIFreeBuffer(lpEntryId);

	return hr;
}

HRESULT ECExchangeImportContentsChanges::CreateConflictFolder(char * lpszName, LPSPropValue lpAdditionalREN, ULONG ulMVPos, LPMAPIFOLDER lpParentFolder, LPMAPIFOLDER * lppConflictFolder){
	HRESULT hr = hrSuccess;
	LPMAPIFOLDER lpConflictFolder = NULL;
	LPSPropValue lpEntryId = NULL;
	SPropValue sPropValue;
	ULONG ulObjType = 0;

	if(lpAdditionalREN->Value.MVbin.lpbin[ulMVPos].cb > 0 && hrSuccess == lpParentFolder->OpenEntry(lpAdditionalREN->Value.MVbin.lpbin[ulMVPos].cb, (LPENTRYID)lpAdditionalREN->Value.MVbin.lpbin[ulMVPos].lpb, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpConflictFolder)){
		if(lppConflictFolder)
			*lppConflictFolder = lpConflictFolder;
		goto exit;
	}

	hr = lpParentFolder->CreateFolder(FOLDER_GENERIC , lpszName, NULL, &IID_IMAPIFolder, OPEN_IF_EXISTS , &lpConflictFolder);
	if(hr != hrSuccess)
		goto exit;

	sPropValue.ulPropTag = PR_FOLDER_DISPLAY_FLAGS;
	sPropValue.Value.bin.cb = 6;
	sPropValue.Value.bin.lpb = (LPBYTE)"\x01\x04\x0A\x80\x1E\x00";

	hr = HrSetOneProp(lpConflictFolder, &sPropValue);
	if(hr != hrSuccess)
		goto exit;

	hr = HrGetOneProp(lpConflictFolder, PR_ENTRYID, &lpEntryId);
	if(hr != hrSuccess)
		goto exit;

	hr = MAPIAllocateMore(lpEntryId->Value.bin.cb, lpAdditionalREN, (LPVOID*)&lpAdditionalREN->Value.MVbin.lpbin[ulMVPos].lpb);
	if(hr != hrSuccess)
		goto exit;
	memcpy(lpAdditionalREN->Value.MVbin.lpbin[ulMVPos].lpb, lpEntryId->Value.bin.lpb, lpEntryId->Value.bin.cb);
	lpAdditionalREN->Value.MVbin.lpbin[ulMVPos].cb = lpEntryId->Value.bin.cb;

	if(lppConflictFolder)
		*lppConflictFolder = lpConflictFolder;

exit:
	if((hr != hrSuccess || lppConflictFolder == NULL) && lpConflictFolder)
		lpConflictFolder->Release();
	
	if(lpEntryId)
		MAPIFreeBuffer(lpEntryId);

	return hr;
}

ULONG ECExchangeImportContentsChanges::xExchangeImportContentsChanges::AddRef(){
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::AddRef", "");
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->AddRef();
}

ULONG ECExchangeImportContentsChanges::xExchangeImportContentsChanges::Release()
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::Release", "");
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->Release();
}

HRESULT ECExchangeImportContentsChanges::xExchangeImportContentsChanges::QueryInterface(REFIID refiid, void **pInterface)
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::QueryInterface", "%s", DBGGUIDToString(refiid).c_str());
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->QueryInterface(refiid, pInterface);
}

HRESULT ECExchangeImportContentsChanges::xExchangeImportContentsChanges::GetLastError(HRESULT hError, ULONG ulFlags, LPMAPIERROR * lppMapiError)
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::GetLastError", "");
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->GetLastError(hError, ulFlags, lppMapiError);
}

HRESULT ECExchangeImportContentsChanges::xExchangeImportContentsChanges::Config(LPSTREAM lpStream, ULONG ulFlags){
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::Config", "");
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->Config(lpStream, ulFlags);
}

HRESULT ECExchangeImportContentsChanges::xExchangeImportContentsChanges::UpdateState(LPSTREAM lpStream){
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::UpdateState", "");
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->UpdateState(lpStream);
}

HRESULT ECExchangeImportContentsChanges::xExchangeImportContentsChanges::ImportMessageChange(ULONG cValue, LPSPropValue lpPropArray, ULONG ulFlags, LPMESSAGE * lppMessage){
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::ImportMessageChange", "lpPropArray\n%s", PropNameFromPropArray(cValue,lpPropArray).c_str());
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->ImportMessageChange(cValue, lpPropArray, ulFlags, lppMessage);
}

HRESULT ECExchangeImportContentsChanges::xExchangeImportContentsChanges::ImportMessageDeletion(ULONG ulFlags, LPENTRYLIST lpSourceEntryList){
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::ImportMessageDeletion", "lpSourceEntryList\n%s", EntryListToString(lpSourceEntryList).c_str());
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->ImportMessageDeletion(ulFlags, lpSourceEntryList);
}

HRESULT ECExchangeImportContentsChanges::xExchangeImportContentsChanges::ImportPerUserReadStateChange(ULONG cElements, LPREADSTATE lpReadState){
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::ImportPerUserReadStateChange", "");
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->ImportPerUserReadStateChange(cElements, lpReadState);
}

HRESULT ECExchangeImportContentsChanges::xExchangeImportContentsChanges::ImportMessageMove(ULONG cbSourceKeySrcFolder, BYTE FAR * pbSourceKeySrcFolder, ULONG cbSourceKeySrcMessage, BYTE FAR * pbSourceKeySrcMessage, ULONG cbPCLMessage, BYTE FAR * pbPCLMessage, ULONG cbSourceKeyDestMessage, BYTE FAR * pbSourceKeyDestMessage, ULONG cbChangeNumDestMessage, BYTE FAR * pbChangeNumDestMessage){
	TRACE_MAPI(TRACE_ENTRY, "IExchangeImportContentsChanges::ImportMessageMove", "");
	METHOD_PROLOGUE_(ECExchangeImportContentsChanges, ExchangeImportContentsChanges);
	return pThis->ImportMessageMove(cbSourceKeySrcFolder, pbSourceKeySrcFolder, cbSourceKeySrcMessage, pbSourceKeySrcMessage, cbPCLMessage, pbPCLMessage, cbSourceKeyDestMessage, pbSourceKeyDestMessage, cbChangeNumDestMessage, pbChangeNumDestMessage);
}

