/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "guid.h"
#include "ECGuid.h"
#include "ZarafaICS.h"

#include "ECABContainer.h"
#include "IECImportAddressbookChanges.h"
#include "ECMsgStore.h"

#include "ECExportAddressbookChanges.h"

#include <edkmdb.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

ECExportAddressbookChanges::ECExportAddressbookChanges(ECMsgStore *lpStore) {
    m_lpMsgStore = lpStore;
	m_lpChanges = NULL;
	m_lpRawChanges = NULL;
	m_ulChangeId = 0;
	m_ulThisChange = 0;
}

ECExportAddressbookChanges::~ECExportAddressbookChanges() {
	if(m_lpRawChanges)
		MAPIFreeBuffer(m_lpRawChanges);
	if(m_lpChanges)
		MAPIFreeBuffer(m_lpChanges);
}

HRESULT ECExportAddressbookChanges::QueryInterface(REFIID iid, void **lppUnk) {
    if(iid == IID_ECUnknown) {
        *lppUnk = (void *)this;
        AddRef();
        return hrSuccess;
    } else if(iid == IID_IECExportAddressbookChanges) {
        *lppUnk = (void *)&this->m_xECExportAddressbookChanges;
        AddRef();
        return hrSuccess;
    }
    
    return MAPI_E_INTERFACE_NOT_SUPPORTED;
}

HRESULT	ECExportAddressbookChanges::Config(LPSTREAM lpStream, ULONG ulFlags, IECImportAddressbookChanges *lpCollector)
{
    HRESULT hr = hrSuccess;
	LARGE_INTEGER lint = {{ 0, 0 }};
    ABEID abeid;
    ULONG ulNewSyncId = 0; // this is, in fact, ignored. This is because we just take the highest id that we see in the changes list.
	STATSTG sStatStg;
	ULONG ulCount = 0;
	ULONG ulProcessed = 0;
	ULONG ulRead = 0;
	ABEID *lpEid = NULL;
	int n = 0;

	// Read state from stream
	hr = lpStream->Stat(&sStatStg, 0);
	if(hr != hrSuccess)
		goto exit;

	hr = lpStream->Seek(lint, STREAM_SEEK_SET, NULL);
	if (hr != hrSuccess)
		goto exit;

	if(sStatStg.cbSize.QuadPart < 8) {
	    m_ulChangeId = 0;
		m_setProcessed.clear();
	} else {
		m_setProcessed.clear();

		hr = lpStream->Read(&m_ulChangeId, sizeof(ULONG), &ulRead);
		if(hr != hrSuccess)
			goto exit;

		hr = lpStream->Read(&ulCount, sizeof(ULONG), &ulRead);
		if(hr != hrSuccess)
			goto exit;

		while(ulCount) {
			hr = lpStream->Read(&ulProcessed, sizeof(ULONG), &ulRead);
			if(hr != hrSuccess)
				goto exit;

			m_setProcessed.insert(ulProcessed);

			ulCount--;
		}
	}

	// ulFlags ignored
    m_lpImporter = lpCollector;
    m_lpImporter->AddRef();

	memset(&abeid, 0, sizeof(ABEID));
	
	abeid.ulType = MAPI_ABCONT;
	memcpy(&abeid.guid, &MUIDECSAB, sizeof(GUID));
	abeid.ulId = 1; // 1 is the first container
    
    // The parent source key is the entryid of the AB container that we're sync'ing
    SBinary sSourceKeyFolder;
    sSourceKeyFolder.lpb = (BYTE *)&abeid;
    sSourceKeyFolder.cb = sizeof(ABEID);

	if(m_lpChanges)
		MAPIFreeBuffer(m_lpChanges);
	m_lpChanges = NULL;

	if(m_lpRawChanges)
		MAPIFreeBuffer(m_lpRawChanges);
	m_lpRawChanges = NULL;

    hr = m_lpMsgStore->lpTransport->HrGetChanges(sSourceKeyFolder, 0, m_ulChangeId, ICS_SYNC_AB, 0, NULL, &ulNewSyncId, &m_ulChanges, &m_lpRawChanges);
    if(hr != hrSuccess)
        goto exit;

	/*
	 * HOSTED FIXME:
	 * The server sends MAPI_ABCONT to indicate the change is about a company,
	 * even though the company is a MAPI_DISTLIST. But since that is handled as
	 * Group, we have a problem...
	 * Preferably we should pass USEROBJECT_TYPE_* as type which
	 * is far more flexible but it is too late for that now since we can't add
	 * it to the 6.00 clients anymore, and we can't break compatibility.
	 */

	/*
	 * Sort the changes:
	 * Users must exist before the company, but before the group they are member of
	 * Groups must exist after the company and the users which are the members
	 * Companies must exist after the users and groups, so create that last
	 */
	MAPIAllocateBuffer(sizeof(ICSCHANGE) * m_ulChanges, (void **)&m_lpChanges);

	for(unsigned int i=0; i<m_ulChanges;i++) {
		if(m_setProcessed.find(m_lpRawChanges[i].ulChangeId) != m_setProcessed.end())
			continue;
		if(m_lpRawChanges[i].sSourceKey.cb < sizeof(ABEID))
			continue;
		lpEid = (ABEID *)m_lpRawChanges[i].sSourceKey.lpb;

		if(lpEid->ulType == MAPI_MAILUSER) {
			m_lpChanges[n] = m_lpRawChanges[i];
			n++;
		}
	}

	for(unsigned int i=0; i<m_ulChanges;i++) {
		if(m_setProcessed.find(m_lpRawChanges[i].ulChangeId) != m_setProcessed.end())
			continue;
		if(m_lpRawChanges[i].sSourceKey.cb < sizeof(ABEID))
			continue;
		lpEid = (ABEID *)m_lpRawChanges[i].sSourceKey.lpb;

		if(lpEid->ulType == MAPI_DISTLIST) {
			m_lpChanges[n] = m_lpRawChanges[i];
			n++;
		}
	}

	for(unsigned int i=0; i<m_ulChanges;i++) {
		if(m_setProcessed.find(m_lpRawChanges[i].ulChangeId) != m_setProcessed.end())
			continue;
		if(m_lpRawChanges[i].sSourceKey.cb < sizeof(ABEID))
			continue;
		lpEid = (ABEID *)m_lpRawChanges[i].sSourceKey.lpb;

		if(lpEid->ulType == MAPI_ABCONT) {
			m_lpChanges[n] = m_lpRawChanges[i];
			n++;
		}
	}

	m_ulChanges = n;

    // Next change is first one
    m_ulThisChange = 0;
exit:

    return hr;
}

HRESULT ECExportAddressbookChanges::Synchronize(ULONG *lpulSteps, ULONG *lpulProgress)
{	
    HRESULT hr = hrSuccess;
    IMAPIProp *lpDestObj = NULL;
    IMAPIProp *lpSrcObj = NULL;
	PABEID eid = NULL;
    
    // Check if we're already done
    if(m_ulThisChange >= m_ulChanges)
        return hrSuccess;
    
    if(m_lpChanges[m_ulThisChange].sSourceKey.cb < sizeof(ABEID)) {
        hr = MAPI_E_INVALID_PARAMETER;
        goto exit;
    }

	eid = (PABEID)m_lpChanges[m_ulThisChange].sSourceKey.lpb;
        
    switch(m_lpChanges[m_ulThisChange].ulChangeType) {
    case ICS_AB_CHANGE:
    case ICS_AB_NEW: {
		hr = m_lpImporter->ImportABChange(eid->ulType, m_lpChanges[m_ulThisChange].sSourceKey.cb, (LPENTRYID)m_lpChanges[m_ulThisChange].sSourceKey.lpb);
        break;
    }
    case ICS_AB_DELETE: {
        hr = m_lpImporter->ImportABDeletion(eid->ulType, m_lpChanges[m_ulThisChange].sSourceKey.cb, (LPENTRYID)m_lpChanges[m_ulThisChange].sSourceKey.lpb);
        break;
    }
    default:
        hr = MAPI_E_INVALID_PARAMETER;
        goto exit;
    }
    
	if(hr == SYNC_E_IGNORE)
		hr = hrSuccess;

    if(hr != hrSuccess)
        goto exit;

	// Mark the change as processed
	m_setProcessed.insert(m_lpChanges[m_ulThisChange].ulChangeId);

    m_ulThisChange++;

    if(lpulSteps)
        *lpulSteps = m_ulChanges;
        
    if(lpulProgress)
        *lpulProgress = m_ulThisChange;
        
    if(m_ulThisChange >= m_ulChanges)
        hr = hrSuccess;
    else
        hr = SYNC_W_PROGRESS;
         
exit:
    if(lpDestObj)
        lpDestObj->Release();
        
    if(lpSrcObj)
        lpSrcObj->Release();
        
    return hr;
}

HRESULT ECExportAddressbookChanges::UpdateState(LPSTREAM lpStream)
{
	HRESULT hr = hrSuccess;
	LARGE_INTEGER zero = {{0,0}};
	ULARGE_INTEGER uzero = {{0,0}};
	ULONG ulCount = 0;
	ULONG ulWritten = 0;
	ULONG ulProcessed = 0;
	std::set<ULONG>::iterator iterProcessed;

	if(m_ulThisChange == m_ulChanges) {
		// All changes have been processed, we can discard processed changes and go to the next server change ID
		m_setProcessed.clear();

		// The last change ID we received is always the highest change ID, unless there were 0 changes, then we stay
		// at the same change count
		if(m_ulChanges != 0)
			m_ulChangeId = m_lpRawChanges[m_ulThisChange-1].ulChangeId;
	}

	hr = lpStream->Seek(zero, STREAM_SEEK_SET, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = lpStream->SetSize(uzero);
	if(hr != hrSuccess)
		goto exit;

	// Write the change ID
	hr = lpStream->Write(&m_ulChangeId, sizeof(ULONG), &ulWritten);
	if(hr != hrSuccess)
		goto exit;

	ulCount = m_setProcessed.size();

	// Write the number of processed IDs to follow
	hr = lpStream->Write(&ulCount, sizeof(ULONG), &ulWritten);
	if(hr != hrSuccess)
		goto exit;

	// Write the processed IDs
	for(iterProcessed = m_setProcessed.begin(); iterProcessed != m_setProcessed.end(); iterProcessed++) {
		ulProcessed = *iterProcessed;

		hr = lpStream->Write(&ulProcessed, sizeof(ULONG), &ulWritten);
		if(hr != hrSuccess)
			goto exit;
	}

	// All done
exit:
    return hr;
}


ULONG ECExportAddressbookChanges::xECExportAddressbookChanges::AddRef()
{
    METHOD_PROLOGUE_(ECExportAddressbookChanges, ECExportAddressbookChanges);
    TRACE_MAPI(TRACE_ENTRY, "IECExportAddressbookChanges::AddRef", "");
    return pThis->AddRef();
}

ULONG ECExportAddressbookChanges::xECExportAddressbookChanges::Release()
{
    METHOD_PROLOGUE_(ECExportAddressbookChanges, ECExportAddressbookChanges);
    TRACE_MAPI(TRACE_ENTRY, "IECExportAddressbookChanges::Release", "");
    return pThis->Release();
}

HRESULT ECExportAddressbookChanges::xECExportAddressbookChanges::QueryInterface(REFIID refiid, void **pInterface)
{
    METHOD_PROLOGUE_(ECExportAddressbookChanges, ECExportAddressbookChanges);
    TRACE_MAPI(TRACE_ENTRY, "IECExportAddressbookChanges::QueryInterface", "");
    return pThis->QueryInterface(refiid, pInterface);
}

HRESULT ECExportAddressbookChanges::xECExportAddressbookChanges::Config(LPSTREAM lpStream, ULONG ulFlags, IECImportAddressbookChanges *lpCollector)
{
    METHOD_PROLOGUE_(ECExportAddressbookChanges, ECExportAddressbookChanges);
    TRACE_MAPI(TRACE_ENTRY, "IECExportAddressbookChanges::Config", "");
    return pThis->Config(lpStream, ulFlags, lpCollector);
}

HRESULT ECExportAddressbookChanges::xECExportAddressbookChanges::Synchronize(ULONG *lpulSteps, ULONG *lpulProgress)
{
    METHOD_PROLOGUE_(ECExportAddressbookChanges, ECExportAddressbookChanges);
    TRACE_MAPI(TRACE_ENTRY, "IECExportAddressbookChanges::Synchronize", "");
    return pThis->Synchronize(lpulSteps, lpulProgress);
}

HRESULT ECExportAddressbookChanges::xECExportAddressbookChanges::UpdateState(LPSTREAM lpStream)
{
    METHOD_PROLOGUE_(ECExportAddressbookChanges, ECExportAddressbookChanges);
    TRACE_MAPI(TRACE_ENTRY, "IECExportAddressbookChanges::UpdateState", "");
    return pThis->UpdateState(lpStream);
}

