/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <mapiutil.h>
#include <edkguid.h>

#ifndef NO_GETTEXT
#include <libintl.h>
#define _(string) dcgettext ("zarafa", string, LC_MESSAGES)
#else
#define _(string) string
#endif

#include "Mem.h"
#include "ECMessage.h"
#include "ECMsgStore.h"
#include "ECMAPITable.h"
#include "ECMAPIFolder.h"
#include "ECMAPIProp.h"
#include "WSTransport.h"

#include "ECTags.h"

#include "guid.h"
#include "ECGuid.h"

#include "ECDebug.h"
#include "ECDebugClient.h"

#include "mapiext.h"

#include "freebusytags.h"
#include "CommonUtil.h"
#include "Util.h"
#include "ClientUtil.h"
#include "ZarafaUtil.h"

#include "WSUtil.h" // used for UnWrapServerClientStoreEntry
#include "ECExportAddressbookChanges.h"

#include "ProviderUtil.h"
#include "EntryPoint.h"

#include "stringutil.h"
#include "ECOfflineState.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// FIXME: from libserver/ECMAPI.h
#define MSGFLAG_DELETED                           ((ULONG) 0x00000400)


const static SizedSPropTagArray(NUM_RFT_PROPS, sPropRFTColumns) =
{
	NUM_RFT_PROPS,
	{
		PR_ROWID,
		PR_INSTANCE_KEY,
		PR_ENTRYID,
		PR_RECORD_KEY,
		PR_MESSAGE_CLASS
	}
};

ECMsgStore::ECMsgStore(char *lpszProfname, LPMAPISUP lpSupport, WSTransport *lpTransport, BOOL fModify, ULONG ulProfileFlags, BOOL fIsSpooler, BOOL fIsDefaultStore, BOOL bOfflineStore) : ECMAPIProp(NULL, MAPI_STORE, fModify, NULL, "IMsgStore") {

	TRACE_MAPI(TRACE_ENTRY, "ECMsgStore::ECMsgStore","");

	this->lpSupport = lpSupport;
	lpSupport->AddRef();

	this->lpTransport = lpTransport;
	lpTransport->AddRef();

	m_lpNotifyClient = NULL;

	// Add our property handlers
	HrAddPropHandlers(PR_ENTRYID,			GetPropHandler,			DefaultSetPropComputed, (void *)this);
	HrAddPropHandlers(PR_RECORD_KEY,		GetPropHandler,			DefaultSetPropComputed, (void *)this);
	HrAddPropHandlers(PR_SEARCH_KEY,		GetPropHandler,			DefaultSetPropComputed, (void *)this);
	HrAddPropHandlers(PR_USER_NAME	,		GetPropHandler,			DefaultSetPropComputed, (void *)this);
	HrAddPropHandlers(PR_USER_ENTRYID,		GetPropHandler,			DefaultSetPropComputed, (void *)this);

	HrAddPropHandlers(PR_MAILBOX_OWNER_NAME,	GetPropHandler,		DefaultSetPropComputed, (void *)this);
	HrAddPropHandlers(PR_MAILBOX_OWNER_ENTRYID,	GetPropHandler,		DefaultSetPropComputed, (void *)this);
	HrAddPropHandlers(PR_USER_NAME,				GetPropHandler,		DefaultSetPropComputed, (void *)this);
	HrAddPropHandlers(PR_USER_ENTRYID,			GetPropHandler,		DefaultSetPropComputed, (void *)this);

	HrAddPropHandlers(PR_RECEIVE_FOLDER_SETTINGS,	GetPropHandler,		DefaultSetPropIgnore,	(void*) this, FALSE, FALSE);

	HrAddPropHandlers(PR_MESSAGE_SIZE,				GetPropHandler,		DefaultSetPropComputed,	(void*) this, FALSE, FALSE);
	HrAddPropHandlers(PR_MESSAGE_SIZE_EXTENDED,		GetPropHandler,		DefaultSetPropComputed,	(void*) this, FALSE, FALSE);

	HrAddPropHandlers(PR_QUOTA_WARNING_THRESHOLD,	GetPropHandler,		DefaultSetPropComputed,	(void*) this, FALSE, FALSE);
	HrAddPropHandlers(PR_QUOTA_SEND_THRESHOLD,		GetPropHandler,		DefaultSetPropComputed,	(void*) this, FALSE, FALSE);
	HrAddPropHandlers(PR_QUOTA_RECEIVE_THRESHOLD,	GetPropHandler,		DefaultSetPropComputed,	(void*) this, FALSE, FALSE);

	HrAddPropHandlers(PR_STORE_OFFLINE,	GetPropHandler,		DefaultSetPropComputed, (void *)this);

	// only on admin store? how? .. now checked on server in ECTableManager
	HrAddPropHandlers(PR_EC_STATSTABLE_SYSTEM,		GetPropHandler,		DefaultSetPropComputed, (void*) this, FALSE, TRUE);
	HrAddPropHandlers(PR_EC_STATSTABLE_SESSIONS,	GetPropHandler,		DefaultSetPropComputed, (void*) this, FALSE, TRUE);
	HrAddPropHandlers(PR_EC_STATSTABLE_USERS,		GetPropHandler,		DefaultSetPropComputed, (void*) this, FALSE, TRUE);

	// Basically a workaround because we can't pass 'this' in the superclass constructor.
	SetProvider(this);

	this->lpNamedProp = new ECNamedProp(lpTransport);

	this->m_ulProfileFlags = ulProfileFlags;

	m_fIsSpooler = fIsSpooler;
	m_fIsDefaultStore = fIsDefaultStore;
	m_bOfflineStore = bOfflineStore;

	this->lpfnCallback = NULL;
	this->isTransactedObject = FALSE;

	this->m_ulClientVersion = 0;

	GetClientVersion(&this->m_ulClientVersion); //Ignore errors

	ASSERT(lpszProfname != NULL);
	if(lpszProfname)
		this->m_strProfname = lpszProfname;
}



ECMsgStore::~ECMsgStore() {

	TRACE_MAPI(TRACE_ENTRY, "ECMsgStore::~ECMsgStore","");

	if(lpTransport)
		lpTransport->HrLogOff();

	// remove all advices
	if(m_lpNotifyClient)
		m_lpNotifyClient->ReleaseAll();

	// destruct
	if(m_lpNotifyClient)
		m_lpNotifyClient->Release();

	if(lpNamedProp)
		delete lpNamedProp;

	if(lpStorage) {
		// Release our propstorage since it is registered on lpTransport
		lpStorage->Release();
		lpStorage = NULL;
	}

	if(lpTransport)
		lpTransport->Release();

	if(lpSupport)
		lpSupport->Release();

	TRACE_MAPI(TRACE_RETURN, "ECMsgStore::~ECMsgStore","");
}

//FIXME: remove duplicate profilename
HRESULT GetIMsgStoreObject(BOOL bOffline, std::string strProfname, BOOL bModify, ECMapProvider* lpmapProviders, IMAPISupport *lpMAPISup, ULONG cbEntryId, LPENTRYID lpEntryId, LPMDB* lppIMsgStore)
{
	HRESULT hr = hrSuccess;
	PROVIDER_INFO sProviderInfo;
	LPPROFSECT lpProfSect = NULL;
	LPSPropValue lpsPropValue = NULL;
	char *lpszProfileName = NULL;


	hr = lpMAPISup->OpenProfileSection((LPMAPIUID)&MUID_PROFILE_INSTANCE, 0, &lpProfSect);
	if(hr != hrSuccess)
		goto exit;

	hr = HrGetOneProp(lpProfSect, PR_PROFILE_NAME, &lpsPropValue);
	if(hr != hrSuccess)
		goto exit;

	// Set ProfileName
	lpszProfileName = lpsPropValue->Value.lpszA;

	hr = GetProviders(lpmapProviders, lpMAPISup, lpszProfileName, 0, &sProviderInfo);
	if (hr != hrSuccess)
		goto exit;

		hr = sProviderInfo.lpMSProviderOnline->Logon(lpMAPISup, 0, lpszProfileName, cbEntryId, lpEntryId, (bModify)?(MAPI_BEST_ACCESS|MDB_NO_DIALOG):MDB_NO_DIALOG, NULL, NULL, NULL, NULL, NULL, lppIMsgStore);
	if (hr != hrSuccess)
		goto exit;

exit:
	if(lpsPropValue)
		MAPIFreeBuffer(lpsPropValue);

	if(lpProfSect)
		lpProfSect->Release();

	return hr;
}

HRESULT ECMsgStore::QueryInterface(REFIID refiid, void **lpvoid)
{
	HRESULT hr = hrSuccess;

	if(refiid == IID_ECMsgStore || refiid == IID_ECMAPIProp || refiid == IID_ECUnknown) {
		AddRef();
		*lpvoid = this;
		return hrSuccess;
	}
	if(refiid == IID_IMsgStore || refiid == IID_IMAPIProp || refiid == IID_IUnknown) {
		AddRef();
		*lpvoid = &this->m_xMsgStore;
		return hrSuccess;
	}

	if(refiid == IID_IExchangeManageStore) {
		if (m_bOfflineStore == FALSE){			
			AddRef();
			*lpvoid = &this->m_xExchangeManageStore;
			return hrSuccess;
		} else {
		}
	}
	if(refiid == IID_IMSLogon) {
		// No addref because MSLogon is coupled to IMsgStore's lifetime
		*lpvoid = &this->m_xMSLogon;
		return hrSuccess;
	}
	if(refiid == IID_IECServiceAdmin) {
		AddRef();
		*lpvoid = &this->m_xECServiceAdmin;
		return hrSuccess;
	}
	if(refiid == IID_IECSpooler) {
		AddRef();
		*lpvoid = &this->m_xECSpooler;
		return hrSuccess;
	}

	if(refiid == IID_IECSecurity) {
		AddRef();
		*lpvoid = &this->m_xECSecurity;
		return hrSuccess;
	}

	if (refiid == IID_IProxyStoreObject) {
		AddRef();
		*lpvoid = &this->m_xProxyStoreObject;
		return hrSuccess;
	}


	if(refiid == IID_ECMsgStoreOnline)
	{
		if (m_bOfflineStore == FALSE){			
			*lpvoid = &this->m_xMsgStore;
			AddRef();
			return hrSuccess;
		} 

		hr = GetIMsgStoreObject(FALSE, this->m_strProfname, fModify, &g_mapProviders, lpSupport, m_cbEntryId, m_lpEntryId, (LPMDB*)lpvoid);
		if (hr != hrSuccess)
			return hr;

		// Add the child because mapi lookto the ref count if you work through ProxyStoreObject
		ECMsgStore *lpChild = NULL;

		if ( ((LPMDB)*lpvoid)->QueryInterface(IID_ECMsgStore, (void**)&lpChild) != hrSuccess)
			return MAPI_E_INTERFACE_NOT_SUPPORTED;
		
		AddChild(lpChild);

		lpChild->Release();

		return hrSuccess;
	}


	// is admin store?
	if(refiid == IID_IECMultiStoreTable && IsPublicStore() == false) {
		AddRef();
		*lpvoid = &this->m_xECMultiStoreTable;
		return hrSuccess;
	}
	
	if(refiid == IID_IECLicense) {
	    AddRef();
	    *lpvoid = &this->m_xECLicense;
	    return hrSuccess;
	}

	if (refiid == IID_IECServerBehavior) {
		AddRef();
		*lpvoid = &this->m_xECServerBehavior;
		return hrSuccess;
	}

	if (refiid == IID_IECServerMimic) {
		AddRef();
		*lpvoid = &this->m_xECServerMimic;
		return hrSuccess;
	}

	return MAPI_E_INTERFACE_NOT_SUPPORTED;
}

HRESULT ECMsgStore::QueryInterfaceProxy(REFIID refiid, void **lpvoid)
{
	if (refiid == IID_IProxyStoreObject) // block recusive proxy calls
		return MAPI_E_INTERFACE_NOT_SUPPORTED;

	if(refiid == IID_IMsgStore || refiid == IID_IMAPIProp || refiid == IID_IUnknown) {
		AddRef();
		*lpvoid = &this->m_xMsgStoreProxy;
		return hrSuccess;
	}

	return QueryInterface(refiid, lpvoid);
}

ULONG ECMsgStore::Release()
{
	// If a parent has requested a callback when we're going down, do it now.
	if(m_cRef == 1 && this->lpfnCallback) {
		lpfnCallback(this->lpCallbackObject, this);
	}

	return ECUnknown::Release();
}

HRESULT ECMsgStore::HrSetReleaseCallback(ECUnknown *lpObject, RELEASECALLBACK lpfnCallback)
{
	this->lpCallbackObject = lpObject;
	this->lpfnCallback = lpfnCallback;

	return hrSuccess;
}

HRESULT	ECMsgStore::Create(char * lpszProfname, LPMAPISUP lpSupport, WSTransport *lpTransport, BOOL fModify, ULONG ulProfileFlags, BOOL fIsSpooler, BOOL fIsDefaultStore, BOOL bOfflineStore, ECMsgStore **lppECMsgStore) {
	HRESULT hr = hrSuccess;

	ECMsgStore *lpStore = new ECMsgStore(lpszProfname, lpSupport, lpTransport, fModify, ulProfileFlags, fIsSpooler, fIsDefaultStore, bOfflineStore);

	hr = lpStore->QueryInterface(IID_ECMsgStore, (void **)lppECMsgStore);

	if(hr != hrSuccess)
		delete lpStore;

	return hr;
}

HRESULT ECMsgStore::SetProps(ULONG cValues, LPSPropValue lpPropArray, LPSPropProblemArray *lppProblems)
{
	HRESULT hr = hrSuccess;

	hr = ECMAPIProp::SetProps(cValues, lpPropArray, lppProblems);
	if (hr != hrSuccess)
		goto exit;

	hr = ECMAPIProp::SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;

exit:

	return hr;
}

HRESULT ECMsgStore::DeleteProps(LPSPropTagArray lpPropTagArray, LPSPropProblemArray FAR * lppProblems)
{
	HRESULT hr = hrSuccess;

	hr = ECMAPIProp::DeleteProps(lpPropTagArray, lppProblems);
	if (hr != hrSuccess)
		goto exit;

	hr = ECMAPIProp::SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;

exit:

	return hr;
}

HRESULT ECMsgStore::SaveChanges(ULONG ulFlags)
{
	return hrSuccess;
}

HRESULT ECMsgStore::OpenProperty(ULONG ulPropTag, LPCIID lpiid, ULONG ulInterfaceOptions, ULONG ulFlags, LPUNKNOWN FAR * lppUnk)
{
	HRESULT hr = MAPI_E_INTERFACE_NOT_SUPPORTED;

	if (lpiid == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(ulPropTag == PR_RECEIVE_FOLDER_SETTINGS) {
		if (*lpiid == IID_IMAPITable && IsPublicStore() == false)
			// Non supported function for publicfolder
			hr = GetReceiveFolderTable(0, (LPMAPITABLE*)lppUnk);
	} else if(ulPropTag == PR_CONTENTS_SYNCHRONIZER) {
	    if (*lpiid == IID_IECExportAddressbookChanges) {
	        ECExportAddressbookChanges *lpEEAC = new ECExportAddressbookChanges(this);
	        
	        hr = lpEEAC->QueryInterface(*lpiid, (void **)lppUnk);
	    }
        else
            hr = MAPI_E_INTERFACE_NOT_SUPPORTED;
	} else if(ulPropTag == PR_EC_STATSTABLE_SYSTEM) {
		if (*lpiid == IID_IMAPITable)
			hr = OpenStatsTable(TABLETYPE_STATS_SYSTEM, (LPMAPITABLE*)lppUnk);
	} else if(ulPropTag == PR_EC_STATSTABLE_SESSIONS) {
		if (*lpiid == IID_IMAPITable)
			hr = OpenStatsTable(TABLETYPE_STATS_SESSIONS, (LPMAPITABLE*)lppUnk);
	} else if(ulPropTag == PR_EC_STATSTABLE_USERS) {
		if (*lpiid == IID_IMAPITable)
			hr = OpenStatsTable(TABLETYPE_STATS_USERS, (LPMAPITABLE*)lppUnk);
	} else if(ulPropTag == PR_EC_STATSTABLE_COMPANY) {
		if (*lpiid == IID_IMAPITable)
			hr = OpenStatsTable(TABLETYPE_STATS_COMPANY, (LPMAPITABLE*)lppUnk);
	} else
		hr = ECMAPIProp::OpenProperty(ulPropTag, lpiid, ulInterfaceOptions, ulFlags, lppUnk);

exit:
	return hr;
}

HRESULT ECMsgStore::OpenStatsTable(unsigned int ulTableType, LPMAPITABLE *lppTable)
{
	HRESULT hr = hrSuccess;
	WSTableView *lpTableView = NULL;
	ECMAPITable *lpTable = NULL;

	if (!lppTable) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// notifications? set 1st param: m_lpNotifyClient
	hr = ECMAPITable::Create(NULL, 0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	// open store table view, no entryid req.
	hr = lpTransport->HrOpenMiscTable(ulTableType, 0, 0, NULL, this, &lpTableView);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->HrSetTableOps(lpTableView, true);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryInterface(IID_IMAPITable, (void **)lppTable);	
	if (hr != hrSuccess)
	    goto exit;
	    
    AddChild(lpTable);
    
exit:
	if (lpTable)
		lpTable->Release();

	if (lpTableView)
		lpTableView->Release();

	return hr;
}

HRESULT ECMsgStore::InternalAdvise(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulEventMask, LPMAPIADVISESINK lpAdviseSink, ULONG *lpulConnection){
	HRESULT hr = hrSuccess;
	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	if(m_ulProfileFlags & EC_PROFILE_FLAGS_NO_NOTIFICATIONS) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	if(lpAdviseSink == NULL || lpulConnection == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	ASSERT(m_lpNotifyClient != NULL && (lpEntryID != NULL || this->m_lpEntryId != NULL));

	if(lpEntryID == NULL) {

		// never sent the client store entry
		hr = UnWrapServerClientStoreEntry(this->m_cbEntryId, this->m_lpEntryId, &cbUnWrapStoreID, &lpUnWrapStoreID);
		if(hr != hrSuccess)
			goto exit;

		cbEntryID = cbUnWrapStoreID;
		lpEntryID = lpUnWrapStoreID;
	}

	if(m_lpNotifyClient->RegisterAdvise(cbEntryID, (LPBYTE)lpEntryID, ulEventMask, lpAdviseSink, lpulConnection) != S_OK)
		hr = MAPI_E_NO_SUPPORT;

	if(hr != hrSuccess)
		goto exit;

	m_setAdviseConnections.insert(*lpulConnection);

exit:
	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT ECMsgStore::Advise(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulEventMask, LPMAPIADVISESINK lpAdviseSink, ULONG *lpulConnection){
	HRESULT hr = hrSuccess;
	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	if(m_ulProfileFlags & EC_PROFILE_FLAGS_NO_NOTIFICATIONS) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	if(lpAdviseSink == NULL || lpulConnection == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	ASSERT(m_lpNotifyClient != NULL && (lpEntryID != NULL || this->m_lpEntryId != NULL));

	if(lpEntryID == NULL) {

		// never sent the client store entry
		hr = UnWrapServerClientStoreEntry(this->m_cbEntryId, this->m_lpEntryId, &cbUnWrapStoreID, &lpUnWrapStoreID);
		if(hr != hrSuccess)
			goto exit;

		cbEntryID = cbUnWrapStoreID;
		lpEntryID = lpUnWrapStoreID;
	}

	if(m_lpNotifyClient->Advise(cbEntryID, (LPBYTE)lpEntryID, ulEventMask, lpAdviseSink, lpulConnection) != S_OK)
		hr = MAPI_E_NO_SUPPORT;

	m_setAdviseConnections.insert(*lpulConnection);

exit:
	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT ECMsgStore::Unadvise(ULONG ulConnection) {
	HRESULT hr = hrSuccess;

	if(m_ulProfileFlags & EC_PROFILE_FLAGS_NO_NOTIFICATIONS) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	ASSERT(m_lpNotifyClient != NULL);

	m_lpNotifyClient->Unadvise(ulConnection);
exit:
	return hr;
}

HRESULT ECMsgStore::Reload(void *lpParam, ECSESSIONID sessionid)
{
	HRESULT hr = hrSuccess;
	ECMsgStore *lpThis = (ECMsgStore *)lpParam;
	std::set<ULONG>::iterator iter;

	for(iter = lpThis->m_setAdviseConnections.begin(); iter != lpThis->m_setAdviseConnections.end(); iter++) {
		lpThis->m_lpNotifyClient->Reregister(*iter);
	}

	return hr;
}

//FIXME: accept two different entryids
HRESULT ECMsgStore::CompareEntryIDs(ULONG cbEntryID1, LPENTRYID lpEntryID1, ULONG cbEntryID2, LPENTRYID lpEntryID2, ULONG ulFlags, ULONG *lpulResult) {
	HRESULT hr = hrSuccess;
	BOOL fTheSame = FALSE;

	PEID peid1 = (PEID)lpEntryID1;
	PEID peid2 = (PEID)lpEntryID2;
	PEID lpStoreId = (PEID)m_lpEntryId;

	// Apparently BlackBerry CALHelper.exe needs this
	if((cbEntryID1 == 0 && cbEntryID2 != 0) || (cbEntryID1 != 0 && cbEntryID2 == 0)) {
		fTheSame = FALSE;
		goto exit;
	}

	if(lpEntryID1 == NULL || lpEntryID2 == NULL || lpulResult == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Check if one or both of the entry identifiers contains the store guid.
	if(memcmp(&lpStoreId->guid, &peid1->guid, sizeof(GUID)) != 0 ||
		memcmp(&lpStoreId->guid, &peid2->guid, sizeof(GUID)) != 0)
	{
		hr = MAPI_E_UNKNOWN_ENTRYID;
		goto exit;
	}

	if(cbEntryID1 != cbEntryID2)
		goto exit;

	if(memcmp(peid1->abFlags, peid2->abFlags, 4) != 0)
		goto exit;

	if(peid1->ulVersion != peid2->ulVersion)
		goto exit;

	if(peid1->ulType != peid2->ulType)
		goto exit;

	if(peid1->ulVersion == 0) {

		if(cbEntryID1 != sizeof(EID_V0))
			goto exit;

		if( ((EID_V0*)lpEntryID1)->ulId != ((EID_V0*)lpEntryID2)->ulId )
			goto exit;

	}else {
		if(cbEntryID1 != CbNewEID(""))
			goto exit;

		if(peid1->uniqueId != peid2->uniqueId) //comp. with the old ulId
			goto exit;
	}

	fTheSame = TRUE;

exit:
	if(lpulResult)
		*lpulResult = fTheSame;

	return hr;
}

HRESULT ECMsgStore::OpenEntry(ULONG cbEntryID, LPENTRYID lpEntryID, LPCIID lpInterface, ULONG ulFlags, ULONG *lpulObjType, LPUNKNOWN *lppUnk)
{
	HRESULT				hr = hrSuccess;
	LPENTRYID			lpRootEntryID = NULL;
	ULONG				cbRootEntryID = 0;
	BOOL				fModifyObject = FALSE;
	ECMAPIFolder*		lpMAPIFolder = NULL;
	ECMessage*			lpMessage = NULL;
	IECPropStorage*		lpPropStorage = NULL;
	WSMAPIFolderOps*	lpFolderOps = NULL;
	unsigned int		ulObjType = 0;

	// Check input/output variables
	if(lpulObjType == NULL || lppUnk == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(ulFlags & MAPI_MODIFY) {
		if(!fModify) {
			hr = MAPI_E_NO_ACCESS;
			goto exit;
		} else
			fModifyObject = TRUE;
	}

	if(ulFlags & MAPI_BEST_ACCESS)
		fModifyObject = fModify;

	if(cbEntryID == 0) {
		hr = lpTransport->HrGetStore(m_cbEntryId, m_lpEntryId, 0, NULL, &cbRootEntryID, &lpRootEntryID);
		if(hr != hrSuccess)
			goto exit;

		lpEntryID = lpRootEntryID;
		cbEntryID = cbRootEntryID;

	}else {

		hr = HrCompareEntryIdWithStoreGuid(cbEntryID, lpEntryID, &GetStoreGuid());
		if(hr != hrSuccess)
			goto exit;

		if(!(ulFlags & MAPI_DEFERRED_ERRORS)) {
	        hr = lpTransport->HrCheckExistObject(cbEntryID, lpEntryID, (ulFlags&(SHOW_SOFT_DELETES))); 
		    if(hr != hrSuccess) 
			    goto exit; 
		}
	}

	hr = HrGetObjTypeFromEntryId(cbEntryID, (LPBYTE)lpEntryID, &ulObjType);
	if(hr != hrSuccess)
		goto exit;


	switch( ulObjType ) {
	case MAPI_FOLDER:
		hr = lpTransport->HrOpenFolderOps(cbEntryID, lpEntryID, &lpFolderOps);

		if(hr != hrSuccess)
			goto exit;

		hr = ECMAPIFolder::Create(this, fModifyObject, lpFolderOps, &lpMAPIFolder);

		if(hr != hrSuccess)
			goto exit;

		hr = lpTransport->HrOpenPropStorage(m_cbEntryId, m_lpEntryId, cbEntryID, lpEntryID, (ulFlags&SHOW_SOFT_DELETES)?MSGFLAG_DELETED:0, &lpPropStorage);

		if(hr != hrSuccess)
			goto exit;

		hr = lpMAPIFolder->HrSetPropStorage(lpPropStorage, !(ulFlags & MAPI_DEFERRED_ERRORS));

		if(hr != hrSuccess)
			goto exit;

		hr = lpMAPIFolder->SetEntryId(cbEntryID, lpEntryID);

		if(hr != hrSuccess)
			goto exit;

		AddChild(lpMAPIFolder);

		if(lpInterface)
			hr = lpMAPIFolder->QueryInterface(*lpInterface,(void **)lppUnk);
		else
			hr = lpMAPIFolder->QueryInterface(IID_IMAPIFolder, (void **)lppUnk);

		if(lpulObjType)
			*lpulObjType = MAPI_FOLDER;

		break;
	case MAPI_MESSAGE:
		hr = ECMessage::Create(this, FALSE, fModifyObject, 0, FALSE, NULL, &lpMessage); // fixme ulFlags for MAPI_ASSOCIATED

		if(hr != hrSuccess)
			goto exit;

		// SC: TODO: null or my entryid ?? (this is OpenEntry(), so we don't need it?)
		// parent only needed on create new item .. ohwell
		hr = lpTransport->HrOpenPropStorage(m_cbEntryId, m_lpEntryId, cbEntryID, lpEntryID, (ulFlags&SHOW_SOFT_DELETES)?MSGFLAG_DELETED:0, &lpPropStorage);

		if(hr != hrSuccess)
			goto exit;

		hr = lpMessage->SetEntryId(cbEntryID, lpEntryID);

		if(hr != hrSuccess)
			goto exit;

		hr = lpMessage->HrSetPropStorage(lpPropStorage, false);

		if(hr != hrSuccess)
			goto exit;

		AddChild(lpMessage);

		if(lpInterface)
			hr = lpMessage->QueryInterface(*lpInterface,(void **)lppUnk);
		else
			hr = lpMessage->QueryInterface(IID_IMessage, (void **)lppUnk);

		if(lpulObjType)
			*lpulObjType = MAPI_MESSAGE;

		break;
	default:
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

exit:
	if(lpFolderOps)
		lpFolderOps->Release();

	if(lpMAPIFolder)
		lpMAPIFolder->Release();

	if(lpMessage)
		lpMessage->Release();

	if(lpPropStorage)
		lpPropStorage->Release();

	if(lpRootEntryID)
		MAPIFreeBuffer(lpRootEntryID);

	return hr;
}

HRESULT ECMsgStore::SetReceiveFolder(LPTSTR lpszMessageClass, ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID)
{
	HRESULT hr = hrSuccess;

	// Non supported function for publicfolder
	if(IsPublicStore() == TRUE) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	hr = lpTransport->HrSetReceiveFolder(this->m_cbEntryId, this->m_lpEntryId, lpszMessageClass, cbEntryID, lpEntryID);

exit:

	return hr;
}

// If the open store a publicstore
BOOL ECMsgStore::IsPublicStore()
{
	BOOL fPublicStore = FALSE;

	if(memcmp(&this->m_guidMDB_Provider, (LPBYTE) &ZARAFA_STORE_PUBLIC_GUID, sizeof(MAPIUID) ) == 0)
		fPublicStore = TRUE;

	return fPublicStore;
}

// As the store is a delegate store
BOOL ECMsgStore::IsDelegateStore()
{
	BOOL fDelegateStore = FALSE;

	if(memcmp(&this->m_guidMDB_Provider, (LPBYTE) &ZARAFA_STORE_DELEGATE_GUID, sizeof(MAPIUID) ) == 0)
		fDelegateStore = TRUE;

	return fDelegateStore;
}

HRESULT ECMsgStore::GetReceiveFolder(LPTSTR lpszMessageClass, ULONG ulFlags, ULONG *lpcbEntryID, LPENTRYID *lppEntryID, LPTSTR *lppszExplicitClass)
{
	HRESULT		hr = hrSuccess;
	ULONG		cbEntryID = 0;
	LPENTRYID	lpEntryID = NULL;

	// Non supported function for publicfolder
	if(IsPublicStore() == TRUE) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	// Check input/output variables
	if(lpcbEntryID == NULL || lppEntryID == NULL) { // lppszExplicitClass may NULL
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpTransport->HrGetReceiveFolder(this->m_cbEntryId, this->m_lpEntryId, lpszMessageClass, &cbEntryID, &lpEntryID, lppszExplicitClass);

	if(hr != hrSuccess)
		goto exit;

	if(lpEntryID) {
		*lpcbEntryID = cbEntryID;
		*lppEntryID = lpEntryID;
	} else {
		*lpcbEntryID = 0;
		*lppEntryID = NULL;
	}

exit:

	return hr;
}

HRESULT ECMsgStore::GetReceiveFolderTable(ULONG ulFlags, LPMAPITABLE *lppTable)
{
	HRESULT			hr = hrSuccess;
	ECMemTableView*	lpView = NULL;
	ECMemTable*		lpMemTable = NULL;
	LPSRowSet		lpsRowSet = NULL;
	SPropValue		sPropValue;
	unsigned int	i;

	// Non supported function for publicfolder
	if(IsPublicStore() == TRUE) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	// Check input/output variables
	if(lppTable == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = ECMemTable::Create((LPSPropTagArray)&sPropRFTColumns, PR_ROWID, &lpMemTable);//PR_INSTANCE_KEY
	if(hr != hrSuccess)
		goto exit;

	//Get the receivefolder list from the server
	hr = lpTransport->HrGetReceiveFolderTable(this->m_cbEntryId, this->m_lpEntryId, &lpsRowSet);
	if(hr != hrSuccess)
		goto exit;

	for(i=0; i<lpsRowSet->cRows; i++)
	{
		sPropValue.ulPropTag = PR_ROWID;
		sPropValue.Value.ul = i;

		hr = lpMemTable->HrModifyRow(ECKeyTable::TABLE_ROW_ADD, NULL, lpsRowSet->aRow[i].lpProps, NUM_RFT_PROPS);
		if(hr != hrSuccess)
			goto exit;

	}

	hr = lpMemTable->HrGetView(&lpView);
	if(hr != hrSuccess)
		goto exit;

	hr = lpView->QueryInterface(IID_IMAPITable, (void **)lppTable);
	if(hr != hrSuccess)
		goto exit;

exit:
	if(lpsRowSet)
		FreeProws(lpsRowSet);

	if(lpView)
		lpView->Release();

	if(lpMemTable)
		lpMemTable->Release();

	return hr;
}

HRESULT ECMsgStore::StoreLogoff(ULONG *lpulFlags) {
	HRESULT hr = hrSuccess;

	return hr;
}

HRESULT ECMsgStore::AbortSubmit(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulFlags)
{
	HRESULT hr = hrSuccess;

	// Non supported function for publicfolder
	if(IsPublicStore() == TRUE) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	// Check input/output variables
	if(lpEntryID == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpTransport->HrAbortSubmit(cbEntryID, lpEntryID);

exit:
	return hr;
}

HRESULT ECMsgStore::GetOutgoingQueue(ULONG ulFlags, LPMAPITABLE *lppTable)
{
	HRESULT			hr = hrSuccess;
	ECMAPITable*	lpTable = NULL;

	WSTableOutGoingQueue*	lpTableOps = NULL;

	// Only supported by the MAPI spooler
	/*if(this->IsSpooler() == false) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}*/

	// Check input/output variables
	if(lppTable == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = ECMAPITable::Create(this->m_lpNotifyClient, 0, &lpTable);

	if(hr != hrSuccess)
		goto exit;

	hr = this->lpTransport->HrOpenTableOutGoingQueueOps(this->m_cbEntryId, this->m_lpEntryId, this, &lpTableOps);

	if(hr != hrSuccess)
		goto exit;

	hr = lpTable->HrSetTableOps(lpTableOps, !(ulFlags & MAPI_DEFERRED_ERRORS));

	if(hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryInterface(IID_IMAPITable, (void **)lppTable);

	AddChild(lpTable);

exit:
	if(lpTable)
		lpTable->Release();
	if(lpTableOps)
		lpTableOps->Release();

	return hr;
}

HRESULT ECMsgStore::SetLockState(LPMESSAGE lpMessage, ULONG ulLockState)
{
	HRESULT			hr = hrSuccess;
	LPSPropValue	lpsPropArray = NULL;
	SPropTagArray	sPropTagArray;
	ULONG			cValue = 0;
	ULONG			ulSubmitFlag = 0;

	// Only supported by the MAPI spooler
	/*if(this->IsSpooler() == false) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}*/

	sPropTagArray.cValues = 1;
	sPropTagArray.aulPropTag[0] = PR_SUBMIT_FLAGS;

	hr = lpMessage->GetProps(&sPropTagArray, 0, &cValue, &lpsPropArray);
	if(HR_FAILED(hr))
		goto exit;

	if(cValue == 1 && lpsPropArray != NULL && PROP_TYPE(lpsPropArray->ulPropTag) != PT_ERROR) {
		ulSubmitFlag = lpsPropArray->Value.l;
	}else{
	}

	if(lpsPropArray) {
		ECFreeBuffer(lpsPropArray);
		lpsPropArray = NULL;
	}

	// set the lock state, if the message is already in the correct state
	// just get outta here
	if (ulLockState & MSG_LOCKED)
	{
		if (!(ulSubmitFlag & SUBMITFLAG_LOCKED))
			ulSubmitFlag |= SUBMITFLAG_LOCKED;
		else
			goto exit;
	} else { // unlock
		if (ulSubmitFlag & SUBMITFLAG_LOCKED)
			ulSubmitFlag &= ~SUBMITFLAG_LOCKED;
		else
			goto exit;
	}

	hr = ECAllocateBuffer(sizeof(SPropValue)*1, (void**)&lpsPropArray);
	if (hr != hrSuccess)
		goto exit;

	lpsPropArray[0].ulPropTag = PR_SUBMIT_FLAGS;
	lpsPropArray[0].Value.l = ulSubmitFlag;

	hr = lpMessage->SetProps(1, lpsPropArray, NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMessage->SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;

exit:

	if(lpsPropArray)
		ECFreeBuffer(lpsPropArray);

	return hr;
}

HRESULT ECMsgStore::FinishedMsg(ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID)
{
	HRESULT		hr = hrSuccess;
	ULONG		ulObjType = 0;
	LPMESSAGE	lpMessage = NULL;

	// Only supported by the MAPI spooler
	/*if(this->IsSpooler() == false) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}*/

	// Check input/output variables
	if(lpEntryID == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Delete the message from the local outgoing queue
	hr = lpTransport->HrFinishedMessage(cbEntryID, lpEntryID, EC_SUBMIT_LOCAL);
	if(hr != hrSuccess)
		goto exit;

	hr = OpenEntry(cbEntryID, lpEntryID, &IID_IMessage, MAPI_MODIFY,  &ulObjType, (IUnknown**)&lpMessage);
	if(hr != hrSuccess)
		goto exit;

	// Unlock the message
	hr = SetLockState(lpMessage, MSG_UNLOCKED);
	if(hr != hrSuccess)
		goto exit;

	// Information: DoSentMail released object lpMessage
	hr = lpSupport->DoSentMail(0, lpMessage);

	if(hr != hrSuccess)
		goto exit;

	lpMessage = NULL;

exit:
	if(lpMessage)
		lpMessage->Release();

	return hr;
}

HRESULT ECMsgStore::NotifyNewMail(LPNOTIFICATION lpNotification)
{
	HRESULT		hr = hrSuccess;

	// Only supported by the MAPI spooler
	/*if(this->IsSpooler() == false) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}*/

	// Check input/output variables
	if(lpNotification == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(lpNotification->info.newmail.lpParentID == NULL) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	hr = lpTransport->HrNotify(lpNotification);

exit:
	return hr;
}

HRESULT	ECMsgStore::GetPropHandler(ULONG ulPropTag, void* lpProvider, ULONG ulFlags, LPSPropValue lpsPropValue, void *lpParam, void *lpBase)
{
	HRESULT hr = hrSuccess;

	ECMsgStore *lpStore = (ECMsgStore *)lpParam;

	switch(ulPropTag) {
		case PR_SEARCH_KEY:
		case PR_ENTRYID:
			{
				ULONG cbWrapped = 0;
				LPENTRYID lpWrapped = NULL;

				lpsPropValue->ulPropTag = ulPropTag;

				hr = lpStore->GetWrappedStoreEntryID(&cbWrapped, &lpWrapped);

				if(hr == hrSuccess) {
					ECAllocateMore(cbWrapped, lpBase, (LPVOID *)&lpsPropValue->Value.bin.lpb);
					memcpy(lpsPropValue->Value.bin.lpb, lpWrapped, cbWrapped);
					lpsPropValue->Value.bin.cb = cbWrapped;

					MAPIFreeBuffer(lpWrapped);
				} else {
					hr = MAPI_E_NOT_FOUND;
				}
				break;
			}
		case PR_RECORD_KEY:
			lpsPropValue->ulPropTag = PR_RECORD_KEY;
			lpsPropValue->Value.bin.cb = sizeof(MAPIUID);
			ECAllocateMore(sizeof(MAPIUID), lpBase, (void **)&lpsPropValue->Value.bin.lpb);
			memcpy(lpsPropValue->Value.bin.lpb, &lpStore->GetStoreGuid(), sizeof(MAPIUID));
			break;

		case PR_RECEIVE_FOLDER_SETTINGS:
			lpsPropValue->ulPropTag = PR_RECEIVE_FOLDER_SETTINGS;
			lpsPropValue->Value.x = 1;
			break;
		case PR_MESSAGE_SIZE:
			hr = lpStore->HrGetRealProp(PR_MESSAGE_SIZE_EXTENDED, ulFlags, lpBase, lpsPropValue);
			if(hr != hrSuccess) {
				break;
			}
			lpsPropValue->ulPropTag = PR_MESSAGE_SIZE;
			lpsPropValue->Value.ul = (ULONG)lpsPropValue->Value.li.QuadPart;
			break;
		case PR_MESSAGE_SIZE_EXTENDED:
			lpsPropValue->ulPropTag = PR_MESSAGE_SIZE_EXTENDED;
			hr = lpStore->HrGetRealProp(PR_MESSAGE_SIZE_EXTENDED, ulFlags, lpBase, lpsPropValue);
			break;
		case PR_QUOTA_WARNING_THRESHOLD:
			lpsPropValue->ulPropTag = PR_QUOTA_WARNING_THRESHOLD;
			hr = lpStore->HrGetRealProp(PR_QUOTA_WARNING_THRESHOLD, ulFlags, lpBase, lpsPropValue);
			break;
		case PR_QUOTA_SEND_THRESHOLD:
			lpsPropValue->ulPropTag = PR_QUOTA_SEND_THRESHOLD;
			hr = lpStore->HrGetRealProp(PR_QUOTA_SEND_THRESHOLD, ulFlags, lpBase, lpsPropValue);
			break;
		case PR_QUOTA_RECEIVE_THRESHOLD:
			lpsPropValue->ulPropTag = PR_QUOTA_RECEIVE_THRESHOLD;
			hr = lpStore->HrGetRealProp(PR_QUOTA_RECEIVE_THRESHOLD, ulFlags, lpBase, lpsPropValue);
			break;

		case PR_MAILBOX_OWNER_NAME:
			if(lpStore->IsPublicStore() == TRUE || lpStore->IsOfflineStore() == TRUE) {
				hr = MAPI_E_NOT_FOUND;
				break;
			}
			lpsPropValue->ulPropTag = PR_MAILBOX_OWNER_NAME;
			hr = lpStore->HrGetRealProp(PR_MAILBOX_OWNER_NAME, ulFlags, lpBase, lpsPropValue);
			break;
		case PR_MAILBOX_OWNER_ENTRYID:
			if(lpStore->IsPublicStore() == TRUE || lpStore->IsOfflineStore() == TRUE) {
				hr = MAPI_E_NOT_FOUND;
				break;
			}
			lpsPropValue->ulPropTag = PR_MAILBOX_OWNER_ENTRYID;
			hr = lpStore->HrGetRealProp(PR_MAILBOX_OWNER_ENTRYID, ulFlags, lpBase, lpsPropValue);
			break;
		case PR_STORE_OFFLINE:
			// Delegate stores are always online, so ignore this property
			if(lpStore->IsDelegateStore() == TRUE) {
				hr = MAPI_E_NOT_FOUND;
				break;
			}
			lpsPropValue->ulPropTag = PR_STORE_OFFLINE;
			lpsPropValue->Value.b = !!lpStore->IsOfflineStore();
			break;
		case PR_USER_NAME:
			lpsPropValue->ulPropTag = PR_USER_NAME;
			hr = lpStore->HrGetRealProp(PR_USER_NAME, ulFlags, lpBase, lpsPropValue);
			break;
		case PR_USER_ENTRYID:
			lpsPropValue->ulPropTag = PR_USER_ENTRYID;
			hr = lpStore->HrGetRealProp(PR_USER_ENTRYID, ulFlags, lpBase, lpsPropValue);
			break;

		case PR_EC_STATSTABLE_SYSTEM:
		case PR_EC_STATSTABLE_SESSIONS:
		case PR_EC_STATSTABLE_USERS:
		case PR_EC_STATSTABLE_COMPANY:
			lpsPropValue->ulPropTag = ulPropTag;
			lpsPropValue->Value.x = 1;
			break;

		default:
			hr = MAPI_E_NOT_FOUND;
			break;
	}

exit:
	return hr;
}

HRESULT ECMsgStore::SetEntryId(ULONG cbEntryId, LPENTRYID lpEntryId)
{
	HRESULT hr = hrSuccess;

	ASSERT(m_lpNotifyClient == NULL);

	hr = ECGenericProp::SetEntryId(cbEntryId, lpEntryId);
	if(hr != hrSuccess)
		goto exit;

	if(! (m_ulProfileFlags & EC_PROFILE_FLAGS_NO_NOTIFICATIONS)) {
		// Create Notifyclient
		hr = ECNotifyClient::Create(MAPI_STORE, this, m_ulProfileFlags, lpSupport, &m_lpNotifyClient);
		ASSERT(m_lpNotifyClient != NULL);
		if(hr != hrSuccess)
			goto exit;
	}

exit:
	return hr;
}

GUID ECMsgStore::GetStoreGuid()
{
	PEID peid = (PEID)this->m_lpEntryId;

	return peid->guid;
}

HRESULT ECMsgStore::GetWrappedStoreEntryID(ULONG* lpcbWrapped, LPENTRYID* lppWrapped)
{
	HRESULT hr = hrSuccess;

	//hr = WrapStoreEntryID(0, ZARAFA_DLL_NAME, CbEID(peid), (LPENTRYID)peid, &cbWrapped, &lpWrapped);
	hr = lpSupport->WrapStoreEntryID(this->m_cbEntryId, this->m_lpEntryId, lpcbWrapped, lppWrapped);
	if (hr != hrSuccess)
		goto exit;
exit:
	return hr;
}

// This is a function special for the spooler to get the right store entryid
// This is a problem of the master outgoing queue
HRESULT ECMsgStore::GetWrappedServerStoreEntryID(ULONG cbEntryId, LPBYTE lpEntryId, ULONG* lpcbWrapped, LPENTRYID* lppWrapped)
{
	HRESULT		hr = hrSuccess;
	ULONG		cbStoreID = 0;
	LPENTRYID	lpStoreID = NULL;

	entryId		sEntryId;
	sEntryId.__ptr = lpEntryId;
	sEntryId.__size = cbEntryId;

	hr = WrapServerClientStoreEntry(lpTransport->GetServerName(), &sEntryId, &cbStoreID, &lpStoreID);
	if(hr != hrSuccess)
		goto exit;

	hr = lpSupport->WrapStoreEntryID(cbStoreID, lpStoreID, lpcbWrapped, lppWrapped);

exit:
	if (lpStoreID)
		ECFreeBuffer(lpStoreID);

	return hr;
}

/*
 * Implementation of the IExchangeManageStore interface
 */

HRESULT ECMsgStore::CreateStoreEntryID(LPSTR lpszMsgStoreDN, LPSTR lpszMailboxDN, ULONG	ulFlags, ULONG *lpcbEntryID, LPENTRYID *lppEntryID)
{
	HRESULT		hr = hrSuccess;
	ULONG		cbStoreEntryID = 0;
	LPENTRYID	lpStoreEntryID = NULL;
	string		strRedirServer;
	WSTransport	*lpTmpTransport = NULL;

	hr = lpTransport->HrResolveUserStore(lpszMailboxDN, NULL, &cbStoreEntryID, &lpStoreEntryID, &strRedirServer);
	if (hr == MAPI_E_UNABLE_TO_COMPLETE)
	{
		hr = lpTransport->CreateAndLogonAlternate(strRedirServer.c_str(), &lpTmpTransport);
		if (hr != hrSuccess)
			goto exit;

		hr = lpTmpTransport->HrResolveUserStore(lpszMailboxDN, NULL, &cbStoreEntryID, &lpStoreEntryID);
		if (hr != hrSuccess)
			goto exit;

		hr = lpTmpTransport->HrLogOff();
	}
	if(hr != hrSuccess)
		goto exit;

	hr = WrapStoreEntryID(0, ZARAFA_DLL_NAME, cbStoreEntryID, lpStoreEntryID, lpcbEntryID, lppEntryID);

exit:
	if (lpTmpTransport)
		lpTmpTransport->Release();

	if(lpStoreEntryID)
		MAPIFreeBuffer(lpStoreEntryID);

	return hr;
}

HRESULT ECMsgStore::EntryIDFromSourceKey(ULONG cFolderKeySize, BYTE *lpFolderSourceKey,	ULONG cMessageKeySize, BYTE *lpMessageSourceKey, ULONG *lpcbEntryID, LPENTRYID *lppEntryID)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrEntryIDFromSourceKey(this->m_cbEntryId, this->m_lpEntryId, cFolderKeySize, lpFolderSourceKey, cMessageKeySize, lpMessageSourceKey, lpcbEntryID, lppEntryID);

	return hr;
}

HRESULT ECMsgStore::GetRights(ULONG cbUserEntryID, LPENTRYID lpUserEntryID, ULONG cbEntryID, LPENTRYID lpEntryID, ULONG *lpulRights)
{
	HRESULT hr = hrSuccess;

	hr = MAPI_E_NOT_FOUND;

	return hr;
}

HRESULT ECMsgStore::GetMailboxTable(LPSTR lpszServerName, LPMAPITABLE *lppTable, ULONG ulFlags)
{
	HRESULT hr = hrSuccess;

	hr = MAPI_E_NOT_FOUND;

	return hr;
}

HRESULT ECMsgStore::GetPublicFolderTable(LPSTR lpszServerName, LPMAPITABLE *lppTable, ULONG ulFlags)
{
	HRESULT hr = hrSuccess;

	hr = MAPI_E_NOT_FOUND;

	return hr;
}

HRESULT ECMsgStore::Logoff(ULONG *lpulFlags) {
	HRESULT hr = hrSuccess;


	return hr;
}

HRESULT ECMsgStore::OpenStatusEntry(LPCIID lpInterface, ULONG ulFlags, ULONG *lpulObjType, LPVOID *lppEntry) {
	HRESULT hr = hrSuccess;

	hr = MAPI_E_NO_SUPPORT;

	return hr;
}

/////////////////////////////
// IECServiceAdmin
//

// Create the freebusy folder on a private store
HRESULT CreatePrivateFreeBusyData(LPMAPIFOLDER lpRootFolder, LPMAPIFOLDER lpInboxFolder, LPMAPIFOLDER lpCalendarFolder)
{
	// Global
	HRESULT				hr = hrSuccess;
	LPSPropValue		lpPropValue = NULL;
	LPSPropValue		lpFBPropValue = NULL;
	ULONG				cValues = 0;
	ULONG				cCurValues = 0;
	ULONG				ulObjType = 0;

	// Freebusy Data folder
	LPMAPIFOLDER		lpFBFolder		= NULL;

	// localfreebusy message
	IMessage*			lpFBMessage		= NULL;


	///////////////////////////////
	// Freebusy mv propery
	// This property will be fill in on an other place
	hr = ECAllocateBuffer(sizeof(SPropValue), (void**)&lpFBPropValue);
	if(hr != hrSuccess)
		goto exit;

	memset(lpFBPropValue, 0, sizeof(SPropValue));

	lpFBPropValue->ulPropTag = PR_FREEBUSY_ENTRYIDS;
	lpFBPropValue->Value.MVbin.cValues = 4;
	hr = ECAllocateMore(sizeof(SBinary)*lpFBPropValue->Value.MVbin.cValues, lpFBPropValue, (void**)&lpFBPropValue->Value.MVbin.lpbin);
	if(hr != hrSuccess)
		goto exit;

	memset(lpFBPropValue->Value.MVbin.lpbin, 0, sizeof(SBinary)*lpFBPropValue->Value.MVbin.cValues);


	//////////////////////////////////////////////////////
	// Create Freebusy Data into the rootfolder
	// Folder for "freebusy data" settings
	// Create the folder
	hr = lpRootFolder->CreateFolder(FOLDER_GENERIC , "Freebusy Data", NULL, &IID_IMAPIFolder, OPEN_IF_EXISTS , &lpFBFolder);
	if(hr != hrSuccess)
		goto exit;

	// Get entryid of localfreebusy
	hr = HrGetOneProp(lpFBFolder, PR_ENTRYID, &lpPropValue);
	if(hr != hrSuccess)
		goto exit;

	//Fill in position 3 of the FBProperty, with free/busy data folder entryid
	hr = ECAllocateMore(lpPropValue->Value.bin.cb, lpFBPropValue, (void**)&lpFBPropValue->Value.MVbin.lpbin[3].lpb);
	if(hr != hrSuccess)
		goto exit;

	lpFBPropValue->Value.MVbin.lpbin[3].cb = lpPropValue->Value.bin.cb;
	memcpy(lpFBPropValue->Value.MVbin.lpbin[3].lpb, lpPropValue->Value.bin.lpb, lpPropValue->Value.bin.cb);

	ECFreeBuffer(lpPropValue); lpPropValue = NULL;

	//////////////////////////////////////////////////////
	// Create localfreebusy message
	// Default setting of free/busy
	hr = lpFBFolder->CreateMessage(&IID_IMessage, 0, &lpFBMessage);
	if(hr != hrSuccess)
		goto exit;

	cValues = 6;
	cCurValues = 0;
	hr = ECAllocateBuffer(sizeof(SPropValue)*cValues, (void**)&lpPropValue);
	if(hr != hrSuccess)
		goto exit;

	lpPropValue[cCurValues].ulPropTag = PR_MESSAGE_CLASS;
	lpPropValue[cCurValues++].Value.lpszA = "IPM.Microsoft.ScheduleData.FreeBusy";

	lpPropValue[cCurValues].ulPropTag = PR_SUBJECT;
	lpPropValue[cCurValues++].Value.lpszA = "LocalFreebusy";

	lpPropValue[cCurValues].ulPropTag = PR_FREEBUSY_NUM_MONTHS;
	lpPropValue[cCurValues++].Value.ul = ECFREEBUSY_DEFAULT_PUBLISH_MONTHS;

	lpPropValue[cCurValues].ulPropTag = PR_DECLINE_RECURRING_MEETING_REQUESTS;
	lpPropValue[cCurValues++].Value.b = false;

	lpPropValue[cCurValues].ulPropTag = PR_DECLINE_CONFLICTING_MEETING_REQUESTS;
	lpPropValue[cCurValues++].Value.b = false;

	lpPropValue[cCurValues].ulPropTag = PR_PROCESS_MEETING_REQUESTS;
	lpPropValue[cCurValues++].Value.b = false;


	hr = lpFBMessage->SetProps(cCurValues, lpPropValue, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = lpFBMessage->SaveChanges(KEEP_OPEN_READONLY);
	if(hr != hrSuccess)
		goto exit;

	ECFreeBuffer(lpPropValue); lpPropValue = NULL;

	// Get entryid of localfreebusy
	hr = HrGetOneProp(lpFBMessage, PR_ENTRYID, &lpPropValue);
	if(hr != hrSuccess)
		goto exit;

	//Fill in position 1 of the FBProperty, with free/busy message entryid
	hr = ECAllocateMore(lpPropValue->Value.bin.cb, lpFBPropValue, (void**)&lpFBPropValue->Value.MVbin.lpbin[1].lpb);
	if(hr != hrSuccess)
		goto exit;

	lpFBPropValue->Value.MVbin.lpbin[1].cb = lpPropValue->Value.bin.cb;
	memcpy(lpFBPropValue->Value.MVbin.lpbin[1].lpb, lpPropValue->Value.bin.lpb, lpPropValue->Value.bin.cb);

	ECFreeBuffer(lpPropValue); lpPropValue = NULL;
	lpFBMessage->Release(); lpFBMessage = NULL;

	/////////////////////////////////////////////////////////
	// Create Associated localfreebusy message
	hr = lpCalendarFolder->CreateMessage(&IID_IMessage, MAPI_ASSOCIATED, &lpFBMessage);
	if(hr != hrSuccess)
		goto exit;

	cValues = 3;
	cCurValues = 0;
	hr = ECAllocateBuffer(sizeof(SPropValue)*cValues, (void**)&lpPropValue);
	if(hr != hrSuccess)
		goto exit;

	lpPropValue[cCurValues].ulPropTag = PR_MESSAGE_CLASS;
	lpPropValue[cCurValues++].Value.lpszA = "IPM.Microsoft.ScheduleData.FreeBusy";

	lpPropValue[cCurValues].ulPropTag = PR_SUBJECT;
	lpPropValue[cCurValues++].Value.lpszA = "LocalFreebusy";

	lpPropValue[cCurValues].ulPropTag = PR_FREEBUSY_NUM_MONTHS;
	lpPropValue[cCurValues++].Value.ul = ECFREEBUSY_DEFAULT_PUBLISH_MONTHS;


	hr = lpFBMessage->SetProps(cCurValues, lpPropValue, NULL);
	if(hr != hrSuccess)
		goto exit;

	ECFreeBuffer(lpPropValue); lpPropValue = NULL;

	hr = lpFBMessage->SaveChanges(KEEP_OPEN_READONLY);
	if(hr != hrSuccess)
		goto exit;

	// Get entryid of associated localfreebusy
	hr = HrGetOneProp(lpFBMessage, PR_ENTRYID, &lpPropValue);
	if(hr != hrSuccess)
		goto exit;

	//Fill in position 0 of the FBProperty, with associated localfreebusy message entryid
	hr = ECAllocateMore(lpPropValue->Value.bin.cb, lpFBPropValue, (void**)&lpFBPropValue->Value.MVbin.lpbin[0].lpb);
	if(hr != hrSuccess)
		goto exit;

	lpFBPropValue->Value.MVbin.lpbin[0].cb = lpPropValue->Value.bin.cb;
	memcpy(lpFBPropValue->Value.MVbin.lpbin[0].lpb, lpPropValue->Value.bin.lpb, lpPropValue->Value.bin.cb);

	ECFreeBuffer(lpPropValue); lpPropValue = NULL;
	lpFBMessage->Release(); lpFBMessage = NULL;


	/////////////////////////////////////////////////////////
	// Add freebusy entryid on Inbox folder
	//
	hr = lpInboxFolder->SetProps(1, lpFBPropValue, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = lpInboxFolder->SaveChanges(KEEP_OPEN_READWRITE);
	if(hr != hrSuccess)
		goto exit;

	/////////////////////////////////////////////////////////
	// Add freebusy entryid on the root folder
	//
	hr = lpRootFolder->SetProps(1, lpFBPropValue, NULL);
	if(hr != hrSuccess)
		goto exit;

	hr = lpRootFolder->SaveChanges(KEEP_OPEN_READWRITE);
	if(hr != hrSuccess)
		goto exit;

exit:
	//Global
	if(lpPropValue)
		ECFreeBuffer(lpPropValue);

	if(lpFBPropValue)
		ECFreeBuffer(lpFBPropValue);

	// Freebusy Data folder
	if(lpFBFolder)
		lpFBFolder->Release();

	// localfreebusy message
	if(lpFBMessage)
		lpFBMessage->Release();

	return hr;
}

HRESULT ECMsgStore::CreateStore(ULONG ulStoreType, ULONG cbUserId, LPENTRYID lpUserId, ULONG* lpcbStoreId, LPENTRYID* lppStoreId, ULONG* lpcbRootId, LPENTRYID *lppRootId)
{
	HRESULT				hr				= hrSuccess;
	ECMsgStore			*lpecMsgStore	= NULL;
	ECMAPIFolder		*lpMapiFolderRoot= NULL;
	LPMAPIFOLDER		lpFolderRoot	= NULL;	// Root container
	LPMAPIFOLDER		lpFolderRootST	= NULL;	// IPM_SUBTREE
	LPMAPIFOLDER		lpFolderRootNST	= NULL;	// NON_IPM_SUBTREE
	LPMAPIFOLDER		lpMAPIFolder	= NULL; // Temp folder
	LPMAPIFOLDER		lpMAPIFolder2	= NULL; // Temp folder
	IECPropStorage		*lpStorage		= NULL;
	ECMAPIFolder		*lpECMapiFolderInbox = NULL;
	LPMAPIFOLDER		lpFBFolder		= NULL;
	LPMAPIFOLDER		lpInboxFolder = NULL;
	LPMAPIFOLDER		lpCalendarFolder = NULL;


	LPSPropValue		lpPropValue = NULL;
	SPropValue			sPropValue;
	ULONG				cValues = 0;
	ULONG				ulObjType = 0;

	IECSecurity*		lpECSecurity = NULL;

	ECPERMISSION		sPermission;

	LPECUSER			lpECUser = NULL;
	LPECCOMPANY			lpECCompany = NULL;
	LPECGROUP			lpECGroup = NULL;

	std::string			strBuffer;
	ULONG				ulSize;

	GUID				guidStore;
	ULONG				cbStoreId = 0;
	LPENTRYID			lpStoreId = NULL;
	ULONG				cbRootId = 0;
	LPENTRYID			lpRootId = NULL;

	LPSPropValue lpPropValueRss = NULL;


	//Check input
	if(ulStoreType != ECSTORE_TYPE_PUBLIC && ulStoreType != ECSTORE_TYPE_PRIVATE) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(CoCreateGuid(&guidStore) != S_OK) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	// Create store entryid
	hr = HrCreateEntryId(guidStore, MAPI_STORE, &cbStoreId, &lpStoreId);
	if(hr != hrSuccess)
		goto exit;

	// create root entryid
	hr = HrCreateEntryId(guidStore, MAPI_FOLDER, &cbRootId, &lpRootId);
	if(hr != hrSuccess)
		goto exit;


	// Create the messagestore
	hr = lpTransport->HrCreateStore(ulStoreType, cbUserId, lpUserId, cbStoreId, lpStoreId, cbRootId, lpRootId);
	if(hr != hrSuccess)
		goto exit;

	// Open the created messagestore
	hr = ECMsgStore::Create("", lpSupport, lpTransport, true, MAPI_BEST_ACCESS, false, false, false, &lpecMsgStore);
	if(hr != hrSuccess){
		//FIXME: wat te doen met de aangemaakte store ?
		goto exit;
	}

	if(ulStoreType == ECSTORE_TYPE_PRIVATE) {
		memcpy(&lpecMsgStore->m_guidMDB_Provider, &ZARAFA_SERVICE_GUID, sizeof(MAPIUID));
	} else {
		memcpy(&lpecMsgStore->m_guidMDB_Provider, &ZARAFA_STORE_PUBLIC_GUID, sizeof(MAPIUID));
	}

	// Get user or company information depending on the store type.
	if (ulStoreType == ECSTORE_TYPE_PRIVATE) {
		hr = lpTransport->HrGetUser(cbUserId, lpUserId, &lpECUser);
		if (hr != hrSuccess)
			goto exit;
	} else if (ABEID_ID(lpUserId) == 1) {
		/* Public store, ownership set to group EVERYONE*/
		hr = lpTransport->HrGetGroup(cbUserId, lpUserId, &lpECGroup);
		if (hr != hrSuccess)
			goto exit;
	} else {
		/* Public store, ownership set to company */
		hr = lpTransport->HrGetCompany(cbUserId, lpUserId, &lpECCompany);
		if (hr != hrSuccess)
			goto exit;
	}

	// Get a propstorage for the message store
	hr = lpTransport->HrOpenPropStorage(0, NULL, cbStoreId, lpStoreId, 0, &lpStorage);
	if(hr != hrSuccess)
		goto exit;

	// Set up the message store to use this storage
	hr = lpecMsgStore->HrSetPropStorage(lpStorage, TRUE);
	if(hr != hrSuccess)
		goto exit;

	hr = lpecMsgStore->SetEntryId(cbStoreId, lpStoreId);
	if(hr != hrSuccess)
		goto exit;

	// Open rootfolder
	hr = lpecMsgStore->OpenEntry(cbRootId, lpRootId, &IID_ECMAPIFolder, MAPI_MODIFY , &ulObjType, (LPUNKNOWN *)&lpMapiFolderRoot);
	if(hr != hrSuccess)
		goto exit;

	// Set IPC folder
	if(ulStoreType == ECSTORE_TYPE_PRIVATE) {
		hr = lpecMsgStore->SetReceiveFolder("IPC", 0, cbRootId, lpRootId);
		if(hr != hrSuccess)
			goto exit;
	}

	/////////////////
	// Create Folder IPM_SUBTREE into the rootfolder
	hr = lpMapiFolderRoot->QueryInterface(IID_IMAPIFolder, (void**)&lpFolderRoot);
	if(hr != hrSuccess)
		goto exit;

	hr = CreateSpecialFolder(lpFolderRoot, lpecMsgStore, "IPM_SUBTREE", "", PR_IPM_SUBTREE_ENTRYID, 0, NULL, &lpFolderRootST);
	if(hr != hrSuccess)
		goto exit;


	if(ulStoreType == ECSTORE_TYPE_PUBLIC) { // Public folder action

		//////////////
		// Create Folder NON_IPM_SUBTREE into the rootfolder
		hr = CreateSpecialFolder(lpFolderRoot, lpecMsgStore, "NON_IPM_SUBTREE", "", PR_NON_IPM_SUBTREE_ENTRYID, 0, NULL, &lpFolderRootNST);
		if(hr != hrSuccess)
			goto exit;

		////////////////////////////////////////
		//Free busy time folder
		//
		hr = CreateSpecialFolder(lpFolderRootNST, lpecMsgStore, "SCHEDULE+ FREE BUSY", "", PR_SPLUS_FREE_BUSY_ENTRYID, 0, NULL, &lpMAPIFolder);
		if(hr != hrSuccess)
			goto exit;


		// Set acl's on the folder
		sPermission.ulRights = ecRightsReadAny|ecRightsFolderVisible;
		sPermission.ulState = RIGHT_NEW|RIGHT_AUTOUPDATE_DENIED;
		sPermission.ulType = ACCESS_TYPE_GRANT;
		sPermission.sUserId.cb = cbUserId;
		sPermission.sUserId.lpb = (unsigned char*)lpUserId;

		hr = lpMAPIFolder->QueryInterface(IID_IECSecurity, (void**)&lpECSecurity);
		if(hr != hrSuccess)
			goto exit;

		hr = lpECSecurity->SetPermissionRules(1, &sPermission);
		if(hr != hrSuccess)
			goto exit;

		lpECSecurity->Release();
		lpECSecurity = NULL;

		hr = CreateSpecialFolder(lpMAPIFolder, lpecMsgStore, "Zarafa 1", "", PR_FREE_BUSY_FOR_LOCAL_SITE_ENTRYID, 0, NULL, &lpMAPIFolder2);
		if(hr != hrSuccess)
			goto exit;

		// Set acl's on the folder
		sPermission.ulRights = ecRightsAll;//ecRightsReadAny| ecRightsCreate | ecRightsEditOwned| ecRightsDeleteOwned | ecRightsCreateSubfolder | ecRightsFolderVisible;
		sPermission.ulState = RIGHT_NEW|RIGHT_AUTOUPDATE_DENIED;
		sPermission.ulType = ACCESS_TYPE_GRANT;
		sPermission.sUserId.cb = cbUserId;
		sPermission.sUserId.lpb = (unsigned char*)lpUserId;

		hr = lpMAPIFolder2->QueryInterface(IID_IECSecurity, (void**)&lpECSecurity);
		if(hr != hrSuccess)
			goto exit;

		hr = lpECSecurity->SetPermissionRules(1, &sPermission);
		if(hr != hrSuccess)
			goto exit;

		lpMAPIFolder->Release();
		lpMAPIFolder = NULL;
		lpMAPIFolder2->Release();
		lpMAPIFolder2 = NULL;
		lpECSecurity->Release();
		lpECSecurity = NULL;


		// Set acl's on the IPM subtree folder
		sPermission.ulRights = ecRightsReadAny| ecRightsCreate | ecRightsEditOwned| ecRightsDeleteOwned | ecRightsCreateSubfolder | ecRightsFolderVisible;
		sPermission.ulState = RIGHT_NEW|RIGHT_AUTOUPDATE_DENIED;
		sPermission.ulType = ACCESS_TYPE_GRANT;
		sPermission.sUserId.cb = cbUserId;
		sPermission.sUserId.lpb = (unsigned char*)lpUserId;

		hr = lpFolderRootST->QueryInterface(IID_IECSecurity, (void**)&lpECSecurity);
		if(hr != hrSuccess)
			goto exit;

		hr = lpECSecurity->SetPermissionRules(1, &sPermission);
		if(hr != hrSuccess)
			goto exit;

		lpECSecurity->Release();
		lpECSecurity = NULL;

		//////////////////////////////////////
		// indicate, validity of the entry identifiers of the folders in a message store
		// Public folder have default the mask
		//
		cValues = 2;
		ECAllocateBuffer(sizeof(SPropValue)*cValues, (void**)&lpPropValue);
		lpPropValue->ulPropTag = PR_VALID_FOLDER_MASK;
		lpPropValue->Value.ul = FOLDER_FINDER_VALID | FOLDER_IPM_SUBTREE_VALID | FOLDER_IPM_INBOX_VALID | FOLDER_IPM_OUTBOX_VALID | FOLDER_IPM_WASTEBASKET_VALID | FOLDER_IPM_SENTMAIL_VALID | FOLDER_VIEWS_VALID | FOLDER_COMMON_VIEWS_VALID;

		// Set store displayname
		lpPropValue[1].ulPropTag = PR_DISPLAY_NAME;
		strBuffer = "Public folder"; //FIXME: set here the right public folder name?
		ulSize = strBuffer.size() + 1;

		ECAllocateMore(ulSize, lpPropValue, (void**)&lpPropValue[1].Value.lpszA);
		memcpy(lpPropValue[1].Value.lpszA, strBuffer.c_str(), ulSize);

		// Set the property into the store
		hr = lpecMsgStore->SetProps(cValues, lpPropValue, NULL);
		if(hr != hrSuccess)
			goto exit;

		if(lpPropValue){ ECFreeBuffer(lpPropValue); lpPropValue = NULL; }

	}else if(ulStoreType == ECSTORE_TYPE_PRIVATE) { //Private folder

		//////////////
		// Create Folder COMMON_VIEWS into the rootfolder
		// folder holds views that are standard for the message store
		hr = CreateSpecialFolder(lpFolderRoot, lpecMsgStore, "IPM_COMMON_VIEWS", "", PR_COMMON_VIEWS_ENTRYID, 0, NULL, NULL);
		if(hr != hrSuccess)
			goto exit;

		//////////////
		// Create Folder VIEWS into the rootfolder
		// Personal: folder holds views that are defined by a particular user
		hr = CreateSpecialFolder(lpFolderRoot, lpecMsgStore, "IPM_VIEWS", "", PR_VIEWS_ENTRYID, 0, NULL, NULL);
		if(hr != hrSuccess)
			goto exit;

		//////////////
		// Create Folder FINDER_ROOT into the rootfolder
		// Place for searchfolders
		hr = CreateSpecialFolder(lpFolderRoot, lpecMsgStore, "FINDER_ROOT", "", PR_FINDER_ENTRYID, 0, NULL, NULL);
		if(hr != hrSuccess)
			goto exit;

		////////////////////////////////////
		// Create Shortcuts
		// Shortcuts for the favorites
		hr = CreateSpecialFolder(lpFolderRoot, lpecMsgStore, _("Shortcut"), "", PR_IPM_FAVORITES_ENTRYID, 0, NULL, NULL);
		if(hr != hrSuccess)
			goto exit;

		////////////////////////////////////
		// Create Schedule folder
		hr = CreateSpecialFolder(lpFolderRoot, lpecMsgStore, "Schedule", "", PR_SCHEDULE_FOLDER_ENTRYID, 0, NULL, NULL);
		if(hr != hrSuccess)
			goto exit;

		//////////////////////////////////////////////
		// Create folders into IPM_SUBTREE
		// Folders like: Inbox, outbox, tasks, agenda, notes, trashcan, send items, contacts,concepts
		//

		////////////////////////////////////
		// Create Inbox
		hr = CreateSpecialFolder(lpFolderRootST, NULL, _("Inbox"), "", 0, 0, NULL, &lpInboxFolder);
		if(hr != hrSuccess)
			goto exit;

		// Get entryid of the folder
		hr = HrGetOneProp(lpInboxFolder, PR_ENTRYID, &lpPropValue);
		if(hr != hrSuccess)
			goto exit;

		hr = lpecMsgStore->SetReceiveFolder(NULL, 0, lpPropValue->Value.bin.cb, (LPENTRYID)lpPropValue->Value.bin.lpb);
		if(hr != hrSuccess)
			goto exit;

		hr = lpecMsgStore->SetReceiveFolder("IPM", 0, lpPropValue->Value.bin.cb, (LPENTRYID)lpPropValue->Value.bin.lpb);
		if(hr != hrSuccess)
			goto exit;

		hr = lpecMsgStore->SetReceiveFolder("REPORT.IPM", 0, lpPropValue->Value.bin.cb, (LPENTRYID)lpPropValue->Value.bin.lpb);
		if(hr != hrSuccess)
			goto exit;

		if(lpPropValue){ ECFreeBuffer(lpPropValue); lpPropValue = NULL; }

		hr = lpInboxFolder->QueryInterface(IID_ECMAPIFolder, (void**)&lpECMapiFolderInbox);
		if(hr != hrSuccess)
			goto exit;


		////////////////////////////////////
		// Create Outbox
		hr = CreateSpecialFolder(lpFolderRootST, lpecMsgStore, _("Outbox"), "", PR_IPM_OUTBOX_ENTRYID, 0, NULL, NULL);
		if(hr != hrSuccess)
			goto exit;

		////////////////////////////////////
		// Create Trashcan
		hr = CreateSpecialFolder(lpFolderRootST, lpecMsgStore, _("Deleted Items"), "", PR_IPM_WASTEBASKET_ENTRYID, 0, NULL, NULL);
		if(hr != hrSuccess)
			goto exit;

		////////////////////////////////////
		// Create Sent Items
		hr = CreateSpecialFolder(lpFolderRootST, lpecMsgStore, _("Sent Items"), "", PR_IPM_SENTMAIL_ENTRYID, 0, NULL, NULL);
		if(hr != hrSuccess)
			goto exit;

		////////////////////////////////////
		// Create Contacts
		hr = CreateSpecialFolder(lpFolderRootST, lpECMapiFolderInbox, _("Contacts"), "", PR_IPM_CONTACT_ENTRYID, 0, "IPF.Contact", &lpMAPIFolder);
		if(hr != hrSuccess)
			goto exit;

		hr = SetSpecialEntryIdOnFolder(lpMAPIFolder, lpMapiFolderRoot, PR_IPM_CONTACT_ENTRYID, 0);
		if(hr != hrSuccess)
			goto exit;

		lpMAPIFolder->Release(); lpMAPIFolder = NULL;

		////////////////////////////////////
		// Create calendar
		hr = CreateSpecialFolder(lpFolderRootST, lpECMapiFolderInbox, _("Calendar"), "", PR_IPM_APPOINTMENT_ENTRYID, 0, "IPF.Appointment", &lpCalendarFolder);
		if(hr != hrSuccess)
			goto exit;

		hr = SetSpecialEntryIdOnFolder(lpCalendarFolder, lpMapiFolderRoot, PR_IPM_APPOINTMENT_ENTRYID, 0);
		if(hr != hrSuccess)
			goto exit;

		////////////////////////////////////
		// Create Drafts
		hr = CreateSpecialFolder(lpFolderRootST, lpECMapiFolderInbox, _("Drafts"), "", PR_IPM_DRAFTS_ENTRYID, 0, "IPF.Note", &lpMAPIFolder);
		if(hr != hrSuccess)
			goto exit;

		hr = SetSpecialEntryIdOnFolder(lpMAPIFolder, lpMapiFolderRoot, PR_IPM_DRAFTS_ENTRYID, 0);
		if(hr != hrSuccess)
			goto exit;

		lpMAPIFolder->Release(); lpMAPIFolder = NULL;

		////////////////////////////////////
		// Create journal
		hr = CreateSpecialFolder(lpFolderRootST, lpECMapiFolderInbox, _("Journal"), "", PR_IPM_JOURNAL_ENTRYID, 0, "IPF.Journal", &lpMAPIFolder);
		if(hr != hrSuccess)
			goto exit;

		hr = SetSpecialEntryIdOnFolder(lpMAPIFolder, lpMapiFolderRoot, PR_IPM_JOURNAL_ENTRYID, 0);
		if(hr != hrSuccess)
			goto exit;

		lpMAPIFolder->Release(); lpMAPIFolder = NULL;

		////////////////////////////////////
		// Create Notes
		hr = CreateSpecialFolder(lpFolderRootST, lpECMapiFolderInbox, _("Notes"), "", PR_IPM_NOTE_ENTRYID, 0, "IPF.StickyNote", &lpMAPIFolder);
		if(hr != hrSuccess)
			goto exit;

		hr = SetSpecialEntryIdOnFolder(lpMAPIFolder, lpMapiFolderRoot, PR_IPM_NOTE_ENTRYID, 0);
		if(hr != hrSuccess)
			goto exit;

		lpMAPIFolder->Release(); lpMAPIFolder = NULL;

		////////////////////////////////////
		// Create Tasks
		hr = CreateSpecialFolder(lpFolderRootST, lpECMapiFolderInbox, _("Tasks"), "", PR_IPM_TASK_ENTRYID, 0, "IPF.Task", &lpMAPIFolder);
		if(hr != hrSuccess)
			goto exit;

		hr = SetSpecialEntryIdOnFolder(lpMAPIFolder, lpMapiFolderRoot, PR_IPM_TASK_ENTRYID, 0);
		if(hr != hrSuccess)
			goto exit;

		lpMAPIFolder->Release(); lpMAPIFolder = NULL;

		////////////////////////////////////
		// Create Junk mail (position 5(4 in array) in the mvprop PR_ADDITIONAL_REN_ENTRYIDS)
		hr = CreateSpecialFolder(lpFolderRootST, lpECMapiFolderInbox, _("Junk E-mail"), "", PR_ADDITIONAL_REN_ENTRYIDS, 4, "IPF.Note", &lpMAPIFolder);
		if(hr != hrSuccess)
			goto exit;

		hr = SetSpecialEntryIdOnFolder(lpMAPIFolder, lpMapiFolderRoot, PR_ADDITIONAL_REN_ENTRYIDS, 4);
		if(hr != hrSuccess)
			goto exit;

		lpMAPIFolder->Release(); lpMAPIFolder = NULL;

		//////////////////////////////////////
		// Create Freebusy folder data
		hr = CreatePrivateFreeBusyData(lpFolderRoot, lpInboxFolder, lpCalendarFolder);
		if(hr != hrSuccess)
			goto exit;

		lpCalendarFolder->Release(); lpCalendarFolder = NULL;
		lpECMapiFolderInbox->Release();	lpECMapiFolderInbox = NULL;


		//////////////////////////////////////
		// Create Outlook 2007 RSS feeds folder
		hr = lpFolderRootST->CreateFolder(FOLDER_GENERIC , _("RSS Feeds"), _("RSS Feed comment"), &IID_IMAPIFolder, OPEN_IF_EXISTS , &lpMAPIFolder);
		if(hr != hrSuccess)
			goto exit;
		
		// Get entryid of the folder
		hr = HrGetOneProp(lpMAPIFolder, PR_ENTRYID, &lpPropValueRss);
		if(hr != hrSuccess)
			goto exit;

		sPropValue.ulPropTag = PR_CONTAINER_CLASS;
		sPropValue.Value.lpszA = "IPF.Note.OutlookHomepage";

		// Set container class for rss feed
		hr = lpMAPIFolder->SetProps(1, &sPropValue, NULL);
		if(hr != hrSuccess)
			goto exit;

		lpMAPIFolder->Release(); lpMAPIFolder = NULL;

        
		// Build special olk2007 entryid
		strBuffer.append("\x01\x80", 2); //RSS Feeds type
		strBuffer.append(1, ((lpPropValueRss->Value.bin.cb+4)&0xFF));
		strBuffer.append(1, ((lpPropValueRss->Value.bin.cb+4)>>8)&0xFF);
		strBuffer.append("\x01\x00", 2);
		strBuffer.append(1, (lpPropValueRss->Value.bin.cb&0xFF));
		strBuffer.append(1, (lpPropValueRss->Value.bin.cb>>8)&0xFF);
		strBuffer.append((char*)lpPropValueRss->Value.bin.lpb, lpPropValueRss->Value.bin.cb);
		strBuffer.append("\x00\x00\x00\x00", 4);
		
		sPropValue.ulPropTag = PR_IPM_OL2007_ENTRYIDS;
		sPropValue.Value.bin.cb = strBuffer.size();
		sPropValue.Value.bin.lpb = (LPBYTE)strBuffer.c_str();

		// Set on root folder
		hr = lpMapiFolderRoot->SetProps(1, &sPropValue, NULL);
		if(hr != hrSuccess)
			goto exit;

		// Set on inbox folder
		hr = lpInboxFolder->SetProps(1, &sPropValue, NULL);
		if(hr != hrSuccess)
			goto exit;

		if(lpPropValue){ ECFreeBuffer(lpPropValue); lpPropValue = NULL; }
		if(lpPropValueRss){ ECFreeBuffer(lpPropValueRss); lpPropValueRss = NULL; }		
		
		lpInboxFolder->Release(); lpInboxFolder = NULL;

		//////////////////////////////////////
		// indicate, validity of the entry identifiers of the folders in a message store
		//
		cValues = 1;
		ECAllocateBuffer(sizeof(SPropValue)*cValues, (void**)&lpPropValue);
		lpPropValue[0].ulPropTag = PR_VALID_FOLDER_MASK;
		lpPropValue[0].Value.ul = FOLDER_VIEWS_VALID | FOLDER_COMMON_VIEWS_VALID|FOLDER_FINDER_VALID | FOLDER_IPM_INBOX_VALID | FOLDER_IPM_OUTBOX_VALID | FOLDER_IPM_SENTMAIL_VALID | FOLDER_IPM_SUBTREE_VALID | FOLDER_IPM_WASTEBASKET_VALID;

		// Set the property into the store
		hr = lpecMsgStore->SetProps(cValues, lpPropValue, NULL);
		if(hr != hrSuccess)
			goto exit;

		if(lpPropValue){ ECFreeBuffer(lpPropValue); lpPropValue = NULL; }
	}

	*lpcbStoreId = cbStoreId;
	*lppStoreId = lpStoreId;

	*lpcbRootId = cbRootId;
	*lppRootId = lpRootId;

exit:
	if(lpFolderRoot)
		lpFolderRoot->Release();

	if(lpECUser)
		ECFreeBuffer(lpECUser);

	if(lpECGroup)
		ECFreeBuffer(lpECGroup);

	if(lpECCompany)
		ECFreeBuffer(lpECCompany);

	if(lpPropValue)
		ECFreeBuffer(lpPropValue);

	if (lpPropValueRss)
		ECFreeBuffer(lpPropValueRss);

	if(lpECMapiFolderInbox)
		lpECMapiFolderInbox->Release();

	if(lpStorage)
		lpStorage->Release();

	if(lpecMsgStore)
		lpecMsgStore->Release();

	if(lpFolderRootST)
		lpFolderRootST->Release();

	if(lpFolderRootNST)
		lpFolderRootNST->Release();

	if(lpMapiFolderRoot)
		lpMapiFolderRoot->Release();

	if(lpECSecurity)
		lpECSecurity->Release();

	if(lpMAPIFolder)
		lpMAPIFolder->Release();

	if(lpInboxFolder)
		lpInboxFolder->Release();

	if(lpCalendarFolder)
		lpCalendarFolder->Release();

	return hr;
}

HRESULT ECMsgStore::HookStore(ULONG cbUserId, LPENTRYID lpUserId, LPGUID lpGuid)
{
	HRESULT			hr = hrSuccess;

	hr = lpTransport->HrHookStore(cbUserId, lpUserId, lpGuid, 0);

	return hr;
}

HRESULT ECMsgStore::RemoveStore(LPGUID lpGuid)
{
	HRESULT			hr = hrSuccess;

	hr = lpTransport->HrRemoveStore(lpGuid, 0);

	return hr;
}

HRESULT ECMsgStore::ResolveStore(ULONG ulStoreId, ULONG *lpulUserID, ULONG* lpcbStoreID, LPENTRYID* lppStoreID)
{
	HRESULT			hr = hrSuccess;
	ULONG			cbStoreEntryID;
	LPENTRYID		lpStoreEntryID;

	hr = lpTransport->HrResolveStore(ulStoreId, lpulUserID, &cbStoreEntryID, &lpStoreEntryID);
	if (hr != hrSuccess)
		goto exit;

	hr = WrapStoreEntryID(0, ZARAFA_DLL_NAME, cbStoreEntryID, lpStoreEntryID, lpcbStoreID, lppStoreID);

exit:
	if(lpStoreEntryID)
		MAPIFreeBuffer(lpStoreEntryID);

	return hr;
}

HRESULT ECMsgStore::GetStore(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpulStoreId)
{
	HRESULT			hr = hrSuccess;

	hr = lpTransport->HrGetStore(cbUserId, lpUserId, lpulStoreId);

	return hr;
}

HRESULT ECMsgStore::SetSpecialEntryIdOnFolder(LPMAPIFOLDER lpFolder, ECMAPIProp *lpFolderPropSet, unsigned int ulPropTag, unsigned int ulMVPos)
{
	HRESULT hr = hrSuccess;
	LPSPropValue	lpPropValue = NULL;
	LPSPropValue	lpPropMVValue = NULL;
	LPSPropValue	lpPropMVValueNew = NULL;

	// Get entryid of the folder
	hr = HrGetOneProp(lpFolder, PR_ENTRYID, &lpPropValue);
	if(hr != hrSuccess)
		goto exit;

	if( (PROP_TYPE(ulPropTag)&MV_FLAG) == MV_FLAG)
	{
		ECAllocateBuffer(sizeof(SPropValue), (void**)&lpPropMVValueNew);
		memset(lpPropMVValueNew, 0, sizeof(SPropValue));

		hr = HrGetOneProp(lpFolder, ulPropTag, &lpPropMVValue);
		if(hr != hrSuccess) {
			lpPropMVValueNew->Value.MVbin.cValues = (ulMVPos+1);
			ECAllocateMore(sizeof(SBinary)*lpPropMVValueNew->Value.MVbin.cValues, lpPropMVValueNew, (void**)&lpPropMVValueNew->Value.MVbin.lpbin);
			memset(lpPropMVValueNew->Value.MVbin.lpbin, 0, sizeof(SBinary)*lpPropMVValueNew->Value.MVbin.cValues);

			for(unsigned int i=0; i <lpPropMVValueNew->Value.MVbin.cValues; i++)
			{
				if(ulMVPos == i)
					lpPropMVValueNew->Value.MVbin.lpbin[i] = lpPropValue->Value.bin;
			}
		}else{
			lpPropMVValueNew->Value.MVbin.cValues = (lpPropMVValue->Value.MVbin.cValues < ulMVPos)? lpPropValue->Value.bin.cb : ulMVPos+1;
			ECAllocateMore(sizeof(SBinary)*lpPropMVValueNew->Value.MVbin.cValues, lpPropMVValueNew, (void**)&lpPropMVValueNew->Value.MVbin.lpbin);

			memset(lpPropMVValueNew->Value.MVbin.lpbin, 0, sizeof(SBinary)*lpPropMVValueNew->Value.MVbin.cValues);

			for(unsigned int i=0; i <lpPropMVValueNew->Value.MVbin.cValues; i++)
			{
				if(ulMVPos == i)
					lpPropMVValueNew->Value.MVbin.lpbin[i] = lpPropValue->Value.bin;
				else
					lpPropMVValueNew->Value.MVbin.lpbin[i] = lpPropMVValue->Value.MVbin.lpbin[i];
			}
		}

		lpPropMVValueNew->ulPropTag = ulPropTag;

		// Set the property into the right folder
		hr = lpFolderPropSet->SetProps(1, lpPropMVValueNew, NULL);
		if(hr != hrSuccess)
			goto exit;
	}else{
		// Set the property tag value
		lpPropValue->ulPropTag = ulPropTag;

		// Set the property into the right folder
		hr = lpFolderPropSet->SetProps(1, lpPropValue, NULL);
		if(hr != hrSuccess)
			goto exit;
	}

exit:
	if(lpPropValue)
		ECFreeBuffer(lpPropValue);

	return hr;
}

HRESULT ECMsgStore::CreateSpecialFolder(LPMAPIFOLDER lpFolderParent, ECMAPIProp *lpFolderPropSet, LPTSTR lpszFolderName, LPTSTR lpszFolderComment, unsigned int ulPropTag, unsigned int ulMVPos, LPTSTR lpszContainerClass, LPMAPIFOLDER *lppMAPIFolder)
{
	HRESULT hr = hrSuccess;
	LPMAPIFOLDER	lpMAPIFolder = NULL;
	LPSPropValue	lpPropValue = NULL;

	// Add a referention at the folders
	lpFolderParent->AddRef();
	if(lpFolderPropSet) { lpFolderPropSet->AddRef(); }

	if(lpFolderParent == NULL){
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Create the folder
	hr = lpFolderParent->CreateFolder(FOLDER_GENERIC , lpszFolderName, lpszFolderComment, &IID_IMAPIFolder, OPEN_IF_EXISTS , &lpMAPIFolder);
	if(hr != hrSuccess)
		goto exit;

	// Set the special property
	if(lpFolderPropSet) {
		hr = SetSpecialEntryIdOnFolder(lpMAPIFolder, lpFolderPropSet, ulPropTag, ulMVPos);
		if(hr != hrSuccess)
			goto exit;

	}

	if(lpszContainerClass && strlen(lpszContainerClass) > 1)
	{
		ECAllocateBuffer(sizeof(SPropValue)*1, (void**)&lpPropValue);

		lpPropValue[0].ulPropTag = PR_CONTAINER_CLASS;
		ECAllocateMore(strlen(lpszContainerClass) + 1, lpPropValue, (void**)&lpPropValue[0].Value.lpszA);
		memcpy(lpPropValue[0].Value.lpszA, lpszContainerClass, strlen(lpszContainerClass) + 1);

		// Set the property
		hr = lpMAPIFolder->SetProps(1, lpPropValue, NULL);
		if(hr != hrSuccess)
			goto exit;

		if(lpPropValue){ ECFreeBuffer(lpPropValue); lpPropValue = NULL; }
	}

	if(lppMAPIFolder) {
		hr = lpMAPIFolder->QueryInterface(IID_IMAPIFolder, (void**)lppMAPIFolder);
		if(hr != hrSuccess)
			goto exit;
	}

exit:
	if(lpPropValue)
		ECFreeBuffer(lpPropValue);

	if(lpMAPIFolder)
		lpMAPIFolder->Release();

	if(lpFolderParent)
		lpFolderParent->Release();

	if(lpFolderPropSet)
		lpFolderPropSet->Release();

	return hr;
}

HRESULT ECMsgStore::CreateUser(LPECUSER lpECUser, ULONG *lpcbUserId, LPENTRYID *lppUserId)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrCreateUser(lpECUser, lpcbUserId, lppUserId);

	return hr;
}

HRESULT ECMsgStore::SetUser(LPECUSER lpECUser)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrSetUser(lpECUser);

	return hr;
}

HRESULT ECMsgStore::GetUser(ULONG cbUserId, LPENTRYID lpUserId, LPECUSER *lppECUser)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrGetUser(cbUserId, lpUserId, lppECUser);

	return hr;
}

HRESULT ECMsgStore::DeleteUser(ULONG cbUserId, LPENTRYID lpUserId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrDeleteUser(cbUserId, lpUserId);

	return hr;
}

HRESULT ECMsgStore::ResolveUserName(char *lpszUserName, ULONG *lpcbUserId, LPENTRYID *lppUserId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrResolveUserName(lpszUserName, lpcbUserId, lppUserId);

	return hr;
}

HRESULT ECMsgStore::GetSendAsList(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpcSenders, LPECUSER *lppSenders)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrGetSendAsList(cbUserId, lpUserId, lpcSenders, lppSenders);

	return hr;
}

HRESULT ECMsgStore::AddSendAsUser(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbSenderId, LPENTRYID lpSenderId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrAddSendAsUser(cbUserId, lpUserId, cbSenderId, lpSenderId);

	return hr;
}

HRESULT ECMsgStore::DelSendAsUser(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbSenderId, LPENTRYID lpSenderId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrDelSendAsUser(cbUserId, lpUserId, cbSenderId, lpSenderId);

	return hr;
}

HRESULT ECMsgStore::ResolveGroupName(char *lpszGroupName, ULONG *lpcbGroupId, LPENTRYID *lppGroupId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrResolveGroupName(lpszGroupName, lpcbGroupId, lppGroupId);

	return hr;
}

HRESULT ECMsgStore::CreateGroup(LPECGROUP lpECGroup, ULONG *lpcbGroupId, LPENTRYID *lppGroupId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrCreateGroup(lpECGroup, lpcbGroupId, lppGroupId);

	return hr;
}

HRESULT ECMsgStore::SetGroup(LPECGROUP lpECGroup)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrSetGroup(lpECGroup);

	return hr;
}

HRESULT ECMsgStore::GetGroup(ULONG cbGroupId, LPENTRYID lpGroupId, LPECGROUP *lppECGroup)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrGetGroup(cbGroupId, lpGroupId, lppECGroup);

	return hr;
}

HRESULT ECMsgStore::DeleteGroup(ULONG cbGroupId, LPENTRYID lpGroupId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrDeleteGroup(cbGroupId, lpGroupId);

	return hr;
}


//Group and user functions
HRESULT ECMsgStore::DeleteGroupUser(ULONG cbGroupId, LPENTRYID lpGroupId, ULONG cbUserId, LPENTRYID lpUserId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrDeleteGroupUser(cbGroupId, lpGroupId, cbUserId, lpUserId);

	return hr;
}

HRESULT ECMsgStore::AddGroupUser(ULONG cbGroupId, LPENTRYID lpGroupId, ULONG cbUserId, LPENTRYID lpUserId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrAddGroupUser(cbGroupId, lpGroupId, cbUserId, lpUserId);

	return hr;
}

HRESULT ECMsgStore::GetUserListOfGroup(ULONG cbGroupId, LPENTRYID lpGroupId, ULONG *lpcUsers, LPECUSER *lppsUsers)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrGetUserListOfGroup(cbGroupId, lpGroupId, lpcUsers, lppsUsers);

	return hr;
}

HRESULT ECMsgStore::GetGroupListOfUser(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpcGroups, LPECGROUP *lppsGroups)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrGetGroupListOfUser(cbUserId, lpUserId, lpcGroups, lppsGroups);

	return hr;
}

HRESULT ECMsgStore::CreateCompany(LPECCOMPANY lpECCompany, ULONG *lpcbCompanyId, LPENTRYID *lppCompanyId)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrCreateCompany(lpECCompany, lpcbCompanyId, lppCompanyId);

	return hr;
}

HRESULT ECMsgStore::DeleteCompany(ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrDeleteCompany(cbCompanyId, lpCompanyId);

	return hr;
}

HRESULT ECMsgStore::SetCompany(LPECCOMPANY lpECCompany)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrSetCompany(lpECCompany);

	return hr;
}

HRESULT ECMsgStore::GetCompany(ULONG cbCompanyId, LPENTRYID lpCompanyId, LPECCOMPANY *lppECCompany)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrGetCompany(cbCompanyId, lpCompanyId, lppECCompany);

	return hr;
}

HRESULT ECMsgStore::ResolveCompanyName(char *lpszCompanyName, ULONG *lpcbCompanyId, LPENTRYID *lppCompanyId)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrResolveCompanyName(lpszCompanyName, lpcbCompanyId, lppCompanyId);

	return hr;
}

HRESULT ECMsgStore::GetCompanyList(ULONG *lpcCompanies, LPECCOMPANY *lppsCompanies)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrGetCompanyList(lpcCompanies, lppsCompanies);

	return hr;
}

HRESULT ECMsgStore::AddCompanyToRemoteViewList(ULONG cbSetCompanyId, LPENTRYID lpSetCompanyId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrAddCompanyToRemoteViewList(cbSetCompanyId, lpSetCompanyId, cbCompanyId, lpCompanyId);

	return hr;
}

HRESULT ECMsgStore::DelCompanyFromRemoteViewList(ULONG cbSetCompanyId, LPENTRYID lpSetCompanyId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrDelCompanyFromRemoteViewList(cbSetCompanyId, lpSetCompanyId, cbCompanyId, lpCompanyId);

	return hr;
}

HRESULT ECMsgStore::GetRemoteViewList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcCompanies, LPECCOMPANY *lppsCompanies)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrGetRemoteViewList(cbCompanyId, lpCompanyId, lpcCompanies, lppsCompanies);

	return hr;
}

HRESULT ECMsgStore::AddUserToRemoteAdminList(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrAddUserToRemoteAdminList(cbUserId, lpUserId, cbCompanyId, lpCompanyId);

	return hr;
}

HRESULT ECMsgStore::DelUserFromRemoteAdminList(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrDelUserFromRemoteAdminList(cbUserId, lpUserId, cbCompanyId, lpCompanyId);

	return hr;
}

HRESULT ECMsgStore::GetRemoteAdminList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcUsers, LPECUSER *lppsUsers)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrGetRemoteAdminList(cbCompanyId, lpCompanyId, lpcUsers, lppsUsers);

	return hr;
}

HRESULT ECMsgStore::SyncUsers(ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->HrSyncUsers(cbCompanyId, lpCompanyId);

	return hr;
}

HRESULT ECMsgStore::GetQuota(ULONG cbUserId, LPENTRYID lpUserId, LPECQUOTA* lppsQuota)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->GetQuota(cbUserId, lpUserId, lppsQuota);

	return hr;
}

HRESULT ECMsgStore::SetQuota(ULONG cbUserId, LPENTRYID lpUserId, LPECQUOTA lpsQuota)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->SetQuota(cbUserId, lpUserId, lpsQuota);

	return hr;
}

HRESULT ECMsgStore::AddQuotaRecipient(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG cbRecipientId, LPENTRYID lpRecipientId, ULONG ulType)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->AddQuotaRecipient(cbCompanyId, lpCompanyId, cbRecipientId, lpRecipientId, ulType);

	return hr;
}

HRESULT ECMsgStore::DeleteQuotaRecipient(ULONG cbCompanyId, LPENTRYID lpCmopanyId, ULONG cbRecipientId, LPENTRYID lpRecipientId, ULONG ulType)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->DeleteQuotaRecipient(cbCompanyId, lpCmopanyId, cbRecipientId, lpRecipientId, ulType);

	return hr;
}

HRESULT ECMsgStore::GetQuotaRecipients(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpcUsers, LPECUSER *lppsUsers)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->GetQuotaRecipients(cbUserId, lpUserId, lpcUsers, lppsUsers);

	return hr;
}

HRESULT ECMsgStore::GetQuotaStatus(ULONG cbUserId, LPENTRYID lpUserId, LPECQUOTASTATUS* lppsQuotaStatus)
{
	HRESULT	hr = hrSuccess;

	hr = lpTransport->GetQuotaStatus(cbUserId, lpUserId, lppsQuotaStatus);

	return hr;
}

HRESULT ECMsgStore::PurgeSoftDelete(ULONG ulDays)
{
    HRESULT hr = hrSuccess;
    
    hr = lpTransport->HrPurgeSoftDelete(ulDays);
    
    return hr;
}

HRESULT ECMsgStore::GetServerDetails(LPECSVRNAMELIST lpServerNameList, ULONG ulFlags, LPECSERVERLIST* lppsServerList)
{
	HRESULT hr = hrSuccess;
	
	hr = lpTransport->HrGetServerDetails(lpServerNameList, ulFlags, lppsServerList);
	
	return hr;
}

HRESULT ECMsgStore::OpenUserStoresTable(ULONG ulFlags, LPMAPITABLE *lppTable)
{
	HRESULT hr = hrSuccess;
	WSTableView *lpTableView = NULL;
	ECMAPITable *lpTable = NULL;

	if (!lppTable) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// notifications? set 1st param: m_lpNotifyClient
	hr = ECMAPITable::Create(NULL, 0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	// open store table view, no entryid req.
	hr = lpTransport->HrOpenMiscTable(TABLETYPE_USERSTORES, ulFlags, 0, NULL, this, &lpTableView);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->HrSetTableOps(lpTableView, true);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryInterface(IID_IMAPITable, (void **)lppTable);	
	if (hr != hrSuccess)
	    goto exit;
	    
	AddChild(lpTable);
    
exit:
	if (lpTable)
		lpTable->Release();

	if (lpTableView)
		lpTableView->Release();

	return hr;
}

HRESULT ECMsgStore::ResolvePseudoUrl(char *lpszPseudoUrl, char **lppszServerPath, bool *lpbIsPeer)
{
	HRESULT hr = hrSuccess;

	hr = lpTransport->HrResolvePseudoUrl(lpszPseudoUrl, lppszServerPath, lpbIsPeer);

	return hr;
}

// This is almost the same as getting a 'normal' outgoing table, except we pass NULL as PEID for the store
HRESULT ECMsgStore::GetMasterOutgoingTable(ULONG ulFlags, IMAPITable ** lppOutgoingTable)
{
	HRESULT hr = hrSuccess;
	ECMAPITable *lpTable = NULL;
	WSTableOutGoingQueue *lpTableOps = NULL;

	hr = ECMAPITable::Create(this->m_lpNotifyClient, 0, &lpTable);

	if(hr != hrSuccess)
		goto exit;

	hr = this->lpTransport->HrOpenTableOutGoingQueueOps(0, NULL, this, &lpTableOps);

	if(hr != hrSuccess)
		goto exit;

	hr = lpTable->HrSetTableOps(lpTableOps, !(ulFlags & MAPI_DEFERRED_ERRORS));

	if(hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryInterface(IID_IMAPITable, (void **)lppOutgoingTable);

	AddChild(lpTable);

exit:
	if(lpTable)
		lpTable->Release();
	if(lpTableOps)
		lpTableOps->Release();

	return hr;
}

HRESULT ECMsgStore::DeleteFromMasterOutgoingTable(ULONG cbEntryId, LPENTRYID lpEntryId, ULONG ulFlags)
{
	HRESULT hr = hrSuccess;

	// Check input/output variables
	if(lpEntryId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = this->lpTransport->HrFinishedMessage(cbEntryId, lpEntryId, EC_SUBMIT_MASTER | ulFlags);

exit:

	return hr;
}


//////////////////////////
// MAPIOfflineMgr
//

HRESULT ECMsgStore::SetCurrentState(ULONG ulFlags, ULONG ulMask, ULONG ulState, void* pReserved)
{
	HRESULT hr = hrSuccess;

	hr = MAPI_E_NO_SUPPORT;

	return hr;
}

HRESULT ECMsgStore::GetCapabilities(ULONG *pulCapabilities)
{
	HRESULT hr = hrSuccess;

	hr = MAPI_E_NO_SUPPORT;

	return hr;
}

HRESULT ECMsgStore::GetCurrentState(ULONG* pulState)
{
	HRESULT hr = hrSuccess;

	hr = MAPI_E_NO_SUPPORT;

	return hr;
}

HRESULT ECMsgStore::Advise(ULONG ulFlags, MAPIOFFLINE_ADVISEINFO* pAdviseInfo, ULONG* pulAdviseToken)
{
	HRESULT hr = hrSuccess;

	hr = MAPI_E_NO_SUPPORT;

	return hr;
}

HRESULT ECMsgStore::Unadvise(ULONG ulFlags,ULONG ulAdviseToken)
{
	HRESULT hr = hrSuccess;

	hr = MAPI_E_NO_SUPPORT;

	return hr;
}

//////////////////////////
// ProxyStoreObject
//

HRESULT ECMsgStore::UnwrapNoRef(LPVOID *ppvObject) 
{
	HRESULT hr = hrSuccess;

	if (ppvObject == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Because the function UnwrapNoRef return a non referenced object, QueryInterface isn't needed.
	*ppvObject = &this->m_xMsgStoreProxy;	
	
exit:

	return hr;
}

//////////////////////////
// ECMultiStoreTable
//

// open a table with given entryids and columns.
// entryids can be from any store
HRESULT ECMsgStore::OpenMultiStoreTable(LPENTRYLIST lpMsgList, ULONG ulFlags, LPMAPITABLE *lppTable) {
	HRESULT		hr = hrSuccess;
	ECMAPITable	*lpTable = NULL;
	WSTableView	*lpTableOps = NULL;

	if (!lpMsgList || !lppTable) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// no notifications on this table
	hr = ECMAPITable::Create(NULL, ulFlags, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	// open a table on the server, with content specified in lpMsgList
	// TODO: my entryid ?
	hr = lpTransport->HrOpenMultiStoreTable(lpMsgList, ulFlags, 0, NULL, this, &lpTableOps);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->HrSetTableOps(lpTableOps, !(ulFlags & MAPI_DEFERRED_ERRORS));
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryInterface(IID_IMAPITable, (void **)lppTable);

	// add child really needed?
	AddChild(lpTable);
	
exit:
	if (lpTable)
		lpTable->Release();

	if (lpTableOps)
		lpTableOps->Release();

	return hr;
}

HRESULT ECMsgStore::LicenseAuth(unsigned char *lpData, unsigned int ulSize, unsigned char **lppResponse, unsigned int * lpulResponseData)
{
    HRESULT hr = hrSuccess;
    
    hr = lpTransport->HrLicenseAuth(lpData, ulSize, lppResponse, lpulResponseData);
    
    return hr;
}

HRESULT ECMsgStore::LicenseCapa(char ***lppszCapas, unsigned int *lpulSize)
{
    HRESULT hr = hrSuccess;
    
    hr = lpTransport->HrLicenseCapa(lppszCapas, lpulSize);
    
    return hr;
}

 
//////////////////////////
// ECServerBehavior
//
HRESULT ECMsgStore::GetServerBehavior(ULONG *lpulBehavior)
{
	HRESULT	hr = hrSuccess;
	ULONG	ulBehavior = 0;

	if (lpulBehavior == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpTransport->HrGetServerBehavior(&ulBehavior);

	*lpulBehavior = ulBehavior;

exit:
	return hr;
}

//////////////////////////
// ECServerBehavior
//
HRESULT ECMsgStore::SetServerBehavior(ULONG ulBehavior)
{
	return lpTransport->HrSetServerBehavior(ulBehavior);
}

//////////////////////////////////
// IMsgStore interface

ULONG ECMsgStore::xMsgStore::AddRef() {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	return pThis->AddRef();
}

ULONG ECMsgStore::xMsgStore::Release() {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::Release", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	ULONG ulReturn = pThis->Release();
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::Release", "%d", ulReturn);
	return ulReturn;
}

HRESULT ECMsgStore::xMsgStore::QueryInterface(REFIID refiid, void **pInterface) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::QueryInterface", "%s", DBGGUIDToString(refiid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->QueryInterface(refiid, pInterface);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::Advise(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulEventMask, LPMAPIADVISESINK lpAdviseSink, ULONG *lpulConnection) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::Advise", "entryid=%s, eventmask=%08X", Util::bin2hex(cbEntryID, (BYTE*)lpEntryID).c_str(), ulEventMask);
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	return pThis->Advise(cbEntryID, lpEntryID, ulEventMask, lpAdviseSink, lpulConnection);
}

HRESULT ECMsgStore::xMsgStore::Unadvise(ULONG ulConnection) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::Unadvise", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	return pThis->Unadvise(ulConnection);
}

HRESULT ECMsgStore::xMsgStore::CompareEntryIDs(ULONG cbEntryID1, LPENTRYID lpEntryID1, ULONG cbEntryID2, LPENTRYID lpEntryID2, ULONG ulFlags, ULONG *lpulResult) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::CompareEntryIDs", "flags: %d\ncb=%d  entryid1: %s\n cb=%d entryid2: %s", ulFlags, cbEntryID1, Util::bin2hex(cbEntryID1, (BYTE*)lpEntryID1).c_str(), cbEntryID2, Util::bin2hex(cbEntryID2, (BYTE*)lpEntryID2).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->CompareEntryIDs(cbEntryID1, lpEntryID1, cbEntryID2, lpEntryID2, ulFlags, lpulResult);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::CompareEntryIDs", "%s result=%s", GetMAPIErrorDescription(hr).c_str(), (lpulResult)?((*lpulResult == TRUE)?"TRUE":"FALSE") : "NULL");
	return hr;
}

HRESULT ECMsgStore::xMsgStore::OpenEntry(ULONG cbEntryID, LPENTRYID lpEntryID, LPCIID lpInterface, ULONG ulFlags, ULONG *lpulObjType, LPUNKNOWN *lppUnk) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::OpenEntry", "interface=%s, cb=%d, EntryId=%s", (lpInterface)?DBGGUIDToString(*lpInterface).c_str():"", cbEntryID, Util::bin2hex(cbEntryID, (BYTE*)lpEntryID).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->OpenEntry(cbEntryID, lpEntryID, lpInterface, ulFlags, lpulObjType, lppUnk);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::OpenEntry", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::SetReceiveFolder(LPTSTR lpszMessageClass, ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::SetReceiveFolder", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr =pThis->SetReceiveFolder(lpszMessageClass, ulFlags, cbEntryID, lpEntryID);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::SetReceiveFolder", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::GetReceiveFolder(LPTSTR lpszMessageClass, ULONG ulFlags, ULONG *lpcbEntryID, LPENTRYID *lppEntryID, LPTSTR *lppszExplicitClass) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::GetReceiveFolder", "%s", (lpszMessageClass)?lpszMessageClass:"NULL");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->GetReceiveFolder(lpszMessageClass, ulFlags, lpcbEntryID, lppEntryID,lppszExplicitClass);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::GetReceiveFolder", "%s, entryid=%s", GetMAPIErrorDescription(hr).c_str(), (hr == hrSuccess && lppEntryID)?Util::bin2hex(*lpcbEntryID, (BYTE*)*lppEntryID).c_str():"NULL");
	return hr;
}

HRESULT ECMsgStore::xMsgStore::GetReceiveFolderTable(ULONG ulFlags, LPMAPITABLE *lppTable) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::GetReceiveFolderTable", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->GetReceiveFolderTable(ulFlags, lppTable);;
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::GetReceiveFolderTable", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::StoreLogoff(ULONG *lpulFlags) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::StoreLogoff", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	return pThis->StoreLogoff(lpulFlags);
}

HRESULT ECMsgStore::xMsgStore::AbortSubmit(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulFlags) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::AbortSubmit", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->AbortSubmit(cbEntryID, lpEntryID, ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::AbortSubmit", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::GetOutgoingQueue(ULONG ulFlags, LPMAPITABLE *lppTable) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::GetOutgoingQueue", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->GetOutgoingQueue(ulFlags, lppTable);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::GetOutgoingQueue", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::SetLockState(LPMESSAGE lpMessage,ULONG ulLockState) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::SetLockState", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->SetLockState(lpMessage, ulLockState);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::SetLockState", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::FinishedMsg(ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::FinishedMsg", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->FinishedMsg(ulFlags, cbEntryID, lpEntryID);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::FinishedMsg", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::NotifyNewMail(LPNOTIFICATION lpNotification) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::NotifyNewMail", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->NotifyNewMail(lpNotification);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::NotifyNewMail", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::GetLastError(HRESULT hError, ULONG ulFlags, LPMAPIERROR * lppMapiError) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::GetLastError", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->GetLastError(hError, ulFlags, lppMapiError);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::GetLastError", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::SaveChanges(ULONG ulFlags) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::SaveChanges", "flags=%d", ulFlags);
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr =  pThis->SaveChanges(ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::SaveChanges", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::GetProps(LPSPropTagArray lpPropTagArray, ULONG ulFlags, ULONG FAR * lpcValues, LPSPropValue FAR * lppPropArray) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::GetProps", "%d %s", ulFlags, PropNameFromPropTagArray(lpPropTagArray).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->GetProps(lpPropTagArray, ulFlags, lpcValues, lppPropArray);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::GetProps", "%s: %s", GetMAPIErrorDescription(hr).c_str(), PropNameFromPropArray(*lpcValues, *lppPropArray).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::GetPropList(ULONG ulFlags, LPSPropTagArray FAR * lppPropTagArray) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::GetPropList", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->GetPropList(ulFlags, lppPropTagArray);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::GetPropList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::OpenProperty(ULONG ulPropTag, LPCIID lpiid, ULONG ulInterfaceOptions, ULONG ulFlags, LPUNKNOWN FAR * lppUnk) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::OpenProperty", "PropTag=%s, lpiid=%s", PropNameFromPropTag(ulPropTag).c_str(), DBGGUIDToString(*lpiid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->OpenProperty(ulPropTag, lpiid, ulInterfaceOptions, ulFlags, lppUnk);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::OpenProperty", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::SetProps(ULONG cValues, LPSPropValue lpPropArray, LPSPropProblemArray FAR * lppProblems) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::SetProps", "%s", PropNameFromPropArray(cValues, lpPropArray).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->SetProps(cValues, lpPropArray, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::SetProps", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::DeleteProps(LPSPropTagArray lpPropTagArray, LPSPropProblemArray FAR * lppProblems) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::DeleteProps", "%s", PropNameFromPropTagArray(lpPropTagArray).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->DeleteProps(lpPropTagArray, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::DeleteProps", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::CopyTo(ULONG ciidExclude, LPCIID rgiidExclude, LPSPropTagArray lpExcludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpInterface, LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray FAR * lppProblems) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::CopyTo", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->CopyTo(ciidExclude, rgiidExclude, lpExcludeProps, ulUIParam, lpProgress, lpInterface, lpDestObj, ulFlags, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::CopyTo", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::CopyProps(LPSPropTagArray lpIncludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpInterface, LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray FAR * lppProblems) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::CopyProps", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->CopyProps(lpIncludeProps, ulUIParam, lpProgress, lpInterface, lpDestObj, ulFlags, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::CopyProps", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::GetNamesFromIDs(LPSPropTagArray * pptaga, LPGUID lpguid, ULONG ulFlags, ULONG * pcNames, LPMAPINAMEID ** pppNames) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::GetNamesFromIDs", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->GetNamesFromIDs(pptaga, lpguid, ulFlags, pcNames, pppNames);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::GetNamesFromIDs", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStore::GetIDsFromNames(ULONG cNames, LPMAPINAMEID * ppNames, ULONG ulFlags, LPSPropTagArray * pptaga) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStore::GetIDsFromNames", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStore);
	HRESULT hr = pThis->GetIDsFromNames(cNames, ppNames, ulFlags, pptaga);
	TRACE_MAPI(TRACE_RETURN, "IMsgStore::GetIDsFromNames", "%s\n%s", GetMAPIErrorDescription(hr).c_str(), MapiNameIdListToString(cNames, ppNames, (hr==hrSuccess)?(*pptaga):NULL).c_str());
	return hr;
}

/*
 *
 * IExchangeManageStore interface
 *
 *
 */

ULONG ECMsgStore::xExchangeManageStore::AddRef() {
	TRACE_MAPI(TRACE_ENTRY, "IExchangeManageStore::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore, ExchangeManageStore);
	return pThis->AddRef();
}

ULONG ECMsgStore::xExchangeManageStore::Release() {
	TRACE_MAPI(TRACE_ENTRY, "IExchangeManageStore::Release", "");
	METHOD_PROLOGUE_(ECMsgStore, ExchangeManageStore);
	ULONG ulRef = pThis->Release();
	TRACE_MAPI(TRACE_RETURN, "IExchangeManageStore::Release", "%d", ulRef);
	return ulRef;
}

HRESULT ECMsgStore::xExchangeManageStore::QueryInterface(REFIID refiid, void **pInterface)
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeManageStore::QueryInterface", "%s", DBGGUIDToString(refiid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, ExchangeManageStore);
	HRESULT hr = pThis->QueryInterface(refiid, pInterface);
	TRACE_MAPI(TRACE_RETURN, "IExchangeManageStore::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xExchangeManageStore::CreateStoreEntryID(LPSTR lpszMsgStoreDN, LPSTR lpszMailboxDN, ULONG	ulFlags, ULONG *lpcbEntryID, LPENTRYID *lppEntryID)
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeManageStore::CreateStoreEntryID","msgStoreDN=%s , MailboxDN=%s , flags=0x%08X", (lpszMsgStoreDN)?lpszMsgStoreDN: "NULL", (lpszMailboxDN)?lpszMailboxDN:"NULL", ulFlags);
	METHOD_PROLOGUE_(ECMsgStore, ExchangeManageStore);
	HRESULT hr = pThis->CreateStoreEntryID(lpszMsgStoreDN, lpszMailboxDN, ulFlags, lpcbEntryID, lppEntryID);
	TRACE_MAPI(TRACE_RETURN, "IExchangeManageStore::CreateStoreEntryID","%s, cb=%d, data=%s", GetMAPIErrorDescription(hr).c_str(), (lpcbEntryID)?*lpcbEntryID:0, (lppEntryID)?Util::bin2hex(*lpcbEntryID, (BYTE*)*lppEntryID).c_str():"NULL");
	return hr;
}

HRESULT ECMsgStore::xExchangeManageStore::EntryIDFromSourceKey(ULONG cFolderKeySize, BYTE *lpFolderSourceKey,	ULONG cMessageKeySize, BYTE *lpMessageSourceKey, ULONG *lpcbEntryID, LPENTRYID *lppEntryID)
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeManageStore::EntryIDFromSourceKey","");
	METHOD_PROLOGUE_(ECMsgStore, ExchangeManageStore);
	return pThis->EntryIDFromSourceKey(cFolderKeySize, lpFolderSourceKey, cMessageKeySize, lpMessageSourceKey, lpcbEntryID, lppEntryID);
}

HRESULT ECMsgStore::xExchangeManageStore::GetRights(ULONG cbUserEntryID, LPENTRYID lpUserEntryID, ULONG cbEntryID, LPENTRYID lpEntryID, ULONG *lpulRights)
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeManageStore::GetRights","");
	METHOD_PROLOGUE_(ECMsgStore, ExchangeManageStore);
	return pThis->GetRights(cbUserEntryID, lpUserEntryID, cbEntryID, lpEntryID, lpulRights);
}

HRESULT ECMsgStore::xExchangeManageStore::GetMailboxTable(LPSTR lpszServerName, LPMAPITABLE *lppTable, ULONG ulFlags)
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeManageStore::GetMailboxTable","");
	METHOD_PROLOGUE_(ECMsgStore, ExchangeManageStore);
	return pThis->GetMailboxTable(lpszServerName, lppTable, ulFlags);
}

HRESULT ECMsgStore::xExchangeManageStore::GetPublicFolderTable(LPSTR lpszServerName, LPMAPITABLE *lppTable, ULONG ulFlags)
{
	TRACE_MAPI(TRACE_ENTRY, "IExchangeManageStore::GetPublicFolderTable","");
	METHOD_PROLOGUE_(ECMsgStore, ExchangeManageStore);
	return pThis->GetPublicFolderTable(lpszServerName, lppTable, ulFlags);
}

HRESULT __stdcall ECMsgStore::xMSLogon::QueryInterface(REFIID iid, void ** ppvObject)
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::QueryInterface", "%s", DBGGUIDToString(iid).c_str());
	METHOD_PROLOGUE_(ECMsgStore , MSLogon);
	HRESULT hr = pThis->QueryInterface(iid, ppvObject);
	TRACE_MAPI(TRACE_RETURN, "IMSLogon::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

ULONG __stdcall ECMsgStore::xMSLogon::AddRef()
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore , MSLogon);
	return pThis->m_cRef;

}

ULONG __stdcall ECMsgStore::xMSLogon::Release()
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::Release", "");
	METHOD_PROLOGUE_(ECMsgStore , MSLogon);
	// Ignore release, because MSLogon is coupled to store's lifetime (see queryinterface for IMsgStore also)
	return pThis->m_cRef;
}

HRESULT ECMsgStore::xMSLogon::GetLastError(HRESULT hResult, ULONG ulFlags, LPMAPIERROR *lppMAPIError)
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::GetLastError", "");
	METHOD_PROLOGUE_(ECMsgStore, MSLogon);
	HRESULT hr = pThis->GetLastError(hResult, ulFlags, lppMAPIError);
	TRACE_MAPI(TRACE_RETURN, "IMSLogon::GetLastError", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMSLogon::Logoff(ULONG *lpulFlags)
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::Logoff", "");
	METHOD_PROLOGUE_(ECMsgStore, MSLogon);
	return pThis->Logoff(lpulFlags);
}

HRESULT ECMsgStore::xMSLogon::OpenEntry(ULONG cbEntryID, LPENTRYID lpEntryID, LPCIID lpInterface, ULONG ulFlags, ULONG *lpulObjType, LPUNKNOWN *lppUnk)
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::OpenEntry", "interface=%s, flags=%d, cb=%d, EntryId=%s", (lpInterface)?DBGGUIDToString(*lpInterface).c_str():"", ulFlags, cbEntryID, Util::bin2hex(cbEntryID, (BYTE*)lpEntryID).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MSLogon);
	HRESULT hr = pThis->OpenEntry(cbEntryID, lpEntryID, lpInterface, ulFlags, lpulObjType, lppUnk);
	TRACE_MAPI(TRACE_RETURN, "IMSLogon::OpenEntry", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMSLogon::CompareEntryIDs(ULONG cbEntryID1, LPENTRYID lpEntryID1, ULONG cbEntryID2, LPENTRYID lpEntryID2, ULONG ulFlags, ULONG *lpulResult)
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::CompareEntryIDs", "flags: %d\ncb=%d  entryid1: %s\n cb=%d entryid2: %s", ulFlags, cbEntryID1, Util::bin2hex(cbEntryID1, (BYTE*)lpEntryID1).c_str(), cbEntryID2, Util::bin2hex(cbEntryID2, (BYTE*)lpEntryID2).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MSLogon);
	HRESULT hr = pThis->CompareEntryIDs(cbEntryID1, lpEntryID1, cbEntryID2, lpEntryID2, ulFlags, lpulResult);
	TRACE_MAPI(TRACE_RETURN, "IMSLogon::CompareEntryIDs", "%s, result=%s", GetMAPIErrorDescription(hr).c_str(), (lpulResult)?((*lpulResult == TRUE)?"TRUE":"FALSE") : "NULL");
	return hr;
}

HRESULT ECMsgStore::xMSLogon::Advise(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulEventMask, LPMAPIADVISESINK lpAdviseSink, ULONG *lpulConnection)
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::Advise", "");
	METHOD_PROLOGUE_(ECMsgStore, MSLogon);
	return pThis->Advise(cbEntryID, lpEntryID, ulEventMask, lpAdviseSink, lpulConnection);
}

HRESULT ECMsgStore::xMSLogon::Unadvise(ULONG ulConnection)
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::Unadvise", "");
	METHOD_PROLOGUE_(ECMsgStore, MSLogon);
	return pThis->Unadvise(ulConnection);
}

HRESULT ECMsgStore::xMSLogon::OpenStatusEntry(LPCIID lpInterface, ULONG ulFlags, ULONG *lpulObjType, LPVOID *lppEntry)
{
	TRACE_MAPI(TRACE_ENTRY, "IMSLogon::OpenStatusEntry", "");
	METHOD_PROLOGUE_(ECMsgStore, MSLogon);
	HRESULT hr = pThis->OpenStatusEntry(lpInterface, ulFlags, lpulObjType, lppEntry);
	TRACE_MAPI(TRACE_RETURN, "IMSLogon::OpenStatusEntry", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

/////////////////////////////
// IECServiceAdmin
//
HRESULT __stdcall ECMsgStore::xECServiceAdmin::QueryInterface(REFIID iid, void ** ppvObject)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::QueryInterface", "%s", DBGGUIDToString(iid).c_str());
	METHOD_PROLOGUE_(ECMsgStore , ECServiceAdmin);
	HRESULT hr = pThis->QueryInterface(iid, ppvObject);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

ULONG __stdcall ECMsgStore::xECServiceAdmin::AddRef()
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore , ECServiceAdmin);
	return pThis->AddRef();

}

ULONG __stdcall ECMsgStore::xECServiceAdmin::Release()
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::Release", "");
	METHOD_PROLOGUE_(ECMsgStore , ECServiceAdmin);
	ULONG ulRef = pThis->Release();
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::Release", "%d", ulRef);
	return ulRef;

}

HRESULT ECMsgStore::xECServiceAdmin::CreateStore(ULONG ulStoreType, ULONG cbUserId, LPENTRYID lpUserId, ULONG* lpcbStoreId, LPENTRYID* lppStoreId, ULONG* lpcbRootId, LPENTRYID *lppRootId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::CreateStore", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->CreateStore(ulStoreType, cbUserId, lpUserId, lpcbStoreId, lppStoreId, lpcbRootId, lppRootId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::CreateStore", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::HookStore(ULONG cbUserId, LPENTRYID lpUserId, LPGUID lpGuid)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::HookStore", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->HookStore(cbUserId, lpUserId, lpGuid);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::HookStore", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::RemoveStore(LPGUID lpGuid)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::RemoveStore", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->RemoveStore(lpGuid);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::RemoveStore", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::ResolveStore(ULONG ulStoreId, ULONG *lpulUserID, ULONG* lpcbStoreID, LPENTRYID* lppStoreID)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::ResolveStore", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->ResolveStore(ulStoreId, lpulUserID, lpcbStoreID, lppStoreID);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::ResolveStore", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetStore(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpulStoreId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetStore", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetStore(cbUserId, lpUserId, lpulStoreId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetStore", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::CreateUser(LPECUSER lpECUser, ULONG *lpcbUserId, LPENTRYID *lppUserId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::CreateUser", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->CreateUser(lpECUser, lpcbUserId, lppUserId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::CreateUser", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::SetUser(LPECUSER lpECUser)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::SetUser", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->SetUser(lpECUser);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::SetUser", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetUser(ULONG cbUserId, LPENTRYID lpUserId, LPECUSER *lppECUser)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetUser", "userid=%d", ABEID_ID(lpUserId));
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetUser(cbUserId, lpUserId, lppECUser);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetUser", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::DeleteUser(ULONG cbUserId, LPENTRYID lpUserId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::DeleteUser", "user=%d", ABEID_ID(lpUserId));
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->DeleteUser(cbUserId, lpUserId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::DeleteUser", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetUserList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcUsers, LPECUSER* lpsUsers)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetUserList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetUserList(cbCompanyId, lpCompanyId, lpcUsers, lpsUsers);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetUserList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetSendAsList(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpcSenders, LPECUSER *lppSenders)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetSendAsList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetSendAsList(cbUserId, lpUserId, lpcSenders, lppSenders);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetSendAsList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::AddSendAsUser(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbSenderId, LPENTRYID lpSenderId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::AddSendAsUser", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->AddSendAsUser(cbUserId, lpUserId, cbSenderId, lpSenderId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::AddSendAsUser", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::DelSendAsUser(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbSenderId, LPENTRYID lpSenderId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::DelSendAsUser", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->DelSendAsUser(cbUserId, lpUserId, cbSenderId, lpSenderId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::DelSendAsUser", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::ResolveGroupName(char *lpszGroupName, ULONG *lpcbGroupId, LPENTRYID *lppGroupId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::ResolveGroupName", "groupname=%s", (lpszGroupName)?lpszGroupName:"NULL");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->ResolveGroupName(lpszGroupName, lpcbGroupId, lppGroupId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::ResolveGroupName", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::ResolveUserName(char *lpszUserName, ULONG *lpcbUserId, LPENTRYID *lppUserId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::ResolveUserName", "username=%s", (lpszUserName)?lpszUserName:"NULL");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->ResolveUserName(lpszUserName, lpcbUserId, lppUserId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::ResolveUserName", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::CreateGroup(LPECGROUP lpECGroup, ULONG *lpcbGroupId, LPENTRYID *lppGroupId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::CreateGroup", "%s", (lpECGroup->lpszGroupname)?lpECGroup->lpszGroupname:"NULL");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->CreateGroup(lpECGroup, lpcbGroupId, lppGroupId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::CreateGroup", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::SetGroup(LPECGROUP lpECGroup)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::SetGroup", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->SetGroup(lpECGroup);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::SetGroup", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetGroup(ULONG cbGroupId, LPENTRYID lpGroupId, LPECGROUP *lppECGroup)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetGroup", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetGroup(cbGroupId, lpGroupId, lppECGroup);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetGroup", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::DeleteGroup(ULONG cbGroupId, LPENTRYID lpGroupId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::DeleteGroup", "%d", ABEID_ID(lpGroupId));
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->DeleteGroup(cbGroupId, lpGroupId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::DeleteGroup", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetGroupList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcGroups, LPECGROUP *lppsGroups)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetGroupList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetGroupList(cbCompanyId, lpCompanyId, lpcGroups, lppsGroups);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetGroupList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

//Group and user functions
HRESULT ECMsgStore::xECServiceAdmin::DeleteGroupUser(ULONG cbGroupId, LPENTRYID lpGroupId, ULONG cbUserId, LPENTRYID lpUserId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::DeleteGroupUser", "group=%d, user=%d", ABEID_ID(lpGroupId), ABEID_ID(lpUserId));
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->DeleteGroupUser(cbGroupId, lpGroupId, cbUserId, lpUserId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::DeleteGroupUser", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::AddGroupUser(ULONG cbGroupId, LPENTRYID lpGroupId, ULONG cbUserId, LPENTRYID lpUserId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::AddGroupUser", "group=%d, user=%d", ABEID_ID(lpGroupId), ABEID_ID(lpUserId));
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->AddGroupUser(cbGroupId, lpGroupId, cbUserId, lpUserId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::AddGroupUser", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetUserListOfGroup(ULONG cbGroupId, LPENTRYID lpGroupId, ULONG *lpcUsers, LPECUSER *lppsUsers)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetUserListOfGroup", "group=%d", ABEID_ID(lpGroupId));
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetUserListOfGroup(cbGroupId, lpGroupId, lpcUsers, lppsUsers);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetUserListOfGroup", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetGroupListOfUser(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpcGroups, LPECGROUP *lppsGroups)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetGroupListOfUser", "user=%d", ABEID_ID(lpUserId));
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetGroupListOfUser(cbUserId, lpUserId, lpcGroups, lppsGroups);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetGroupListOfUser", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::CreateCompany(LPECCOMPANY lpECCompany, ULONG *lpcbCompanyId, LPENTRYID *lppCompanyId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::CreateCompany", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->CreateCompany(lpECCompany, lpcbCompanyId, lppCompanyId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::CreateCompany", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::DeleteCompany(ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::DeleteCompany", "%d", ABEID_ID(lpCompanyId));
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->DeleteCompany(cbCompanyId, lpCompanyId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::DeleteCompany", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr; 
}

HRESULT ECMsgStore::xECServiceAdmin::SetCompany(LPECCOMPANY lpECCompany)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::SetCompany", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->SetCompany(lpECCompany);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::SetCompany", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetCompany(ULONG cbCompanyId, LPENTRYID lpCompanyId, LPECCOMPANY *lppECCompany)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetCompany", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetCompany(cbCompanyId, lpCompanyId, lppECCompany);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetCompany", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::ResolveCompanyName(char *lpszCompanyName, ULONG *lpcbCompanyId, LPENTRYID *lppCompanyId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::ResolveCompanyName", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->ResolveCompanyName(lpszCompanyName, lpcbCompanyId, lppCompanyId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::ResolveCompanyName", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetCompanyList(ULONG *lpcCompanies, LPECCOMPANY *lppsCompanies)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetCompanyList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetCompanyList(lpcCompanies, lppsCompanies);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetCompanyList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::AddCompanyToRemoteViewList(ULONG cbSetCompanyId, LPENTRYID lpSetCompanyId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::AddCompanyToRemoteViewList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->AddCompanyToRemoteViewList(cbSetCompanyId, lpSetCompanyId, cbCompanyId, lpCompanyId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::AddCompanyToRemoteViewList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::DelCompanyFromRemoteViewList(ULONG cbSetCompanyId, LPENTRYID lpSetCompanyId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::DelCompanyFromRemoteViewList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->DelCompanyFromRemoteViewList(cbSetCompanyId, lpSetCompanyId, cbCompanyId, lpCompanyId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::DelCompanyFromRemoteViewList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetRemoteViewList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcCompanies, LPECCOMPANY *lppsCompanies)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetRemoteViewList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetRemoteViewList(cbCompanyId, lpCompanyId, lpcCompanies, lppsCompanies);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetRemoteViewList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::AddUserToRemoteAdminList(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::AddCompanyToRemoteAdminList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->AddUserToRemoteAdminList(cbUserId, lpUserId, cbCompanyId, lpCompanyId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::AddCompanyToRemoteAdminList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::DelUserFromRemoteAdminList(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::DelCompanyFromRemoteAdminList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->DelUserFromRemoteAdminList(cbUserId, lpUserId, cbCompanyId, lpCompanyId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::DelCompanyFromRemoteAdminList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetRemoteAdminList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcUsers, LPECUSER *lppsUsers)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetRemoteAdminList", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetRemoteAdminList(cbCompanyId, lpCompanyId, lpcUsers, lppsUsers);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetRemoteAdminList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::SyncUsers(ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::SyncUsers", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->SyncUsers(cbCompanyId, lpCompanyId);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::SyncUsers", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::GetQuota(ULONG cbUserId, LPENTRYID lpUserId, LPECQUOTA* lppsQuota)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetQuota", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->GetQuota(cbUserId, lpUserId, lppsQuota);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::GetQuota", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::SetQuota(ULONG cbUserId, LPENTRYID lpUserId, LPECQUOTA lpsQuota)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::SetQuota", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	HRESULT hr = pThis->SetQuota(cbUserId, lpUserId, lpsQuota);
	TRACE_MAPI(TRACE_RETURN, "IECServiceAdmin::SetQuota", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServiceAdmin::AddQuotaRecipient(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG cbRecipientId, LPENTRYID lpRecipientId, ULONG ulType)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::AddQuotaRecipient", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	return pThis->AddQuotaRecipient(cbCompanyId, lpCompanyId, cbRecipientId, lpRecipientId, ulType);
}

HRESULT ECMsgStore::xECServiceAdmin::DeleteQuotaRecipient(ULONG cbCompanyId, LPENTRYID lpCmopanyId, ULONG cbRecipientId, LPENTRYID lpRecipientId, ULONG ulType)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::DeleteQuotaRecipient", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	return pThis->DeleteQuotaRecipient(cbCompanyId, lpCmopanyId, cbRecipientId, lpRecipientId, ulType);
}

HRESULT ECMsgStore::xECServiceAdmin::GetQuotaRecipients(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpcUsers, LPECUSER *lppsUsers)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetQuotarecipients", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	return pThis->GetQuotaRecipients(cbUserId, lpUserId, lpcUsers, lppsUsers);
}

HRESULT ECMsgStore::xECServiceAdmin::GetQuotaStatus(ULONG cbUserId, LPENTRYID lpUserId, LPECQUOTASTATUS* lppsQuotaStatus)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetQuotaStatus", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	return pThis->GetQuotaStatus(cbUserId, lpUserId, lppsQuotaStatus);
}

HRESULT ECMsgStore::xECServiceAdmin::PurgeSoftDelete(ULONG ulDays)
{
    TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::PurgeSoftDelete", "");
    METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
    return pThis->PurgeSoftDelete(ulDays);
}

HRESULT ECMsgStore::xECServiceAdmin::GetServerDetails(LPECSVRNAMELIST lpServerNameList, ULONG ulFlags, LPECSERVERLIST* lppsServerList)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::GetServerDetails", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	return pThis->GetServerDetails(lpServerNameList, ulFlags, lppsServerList);
}

HRESULT ECMsgStore::xECServiceAdmin::OpenUserStoresTable(ULONG ulFlags, LPMAPITABLE *lppTable)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::OpenUserStoresTable", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	return pThis->OpenUserStoresTable(ulFlags, lppTable);
}

HRESULT ECMsgStore::xECServiceAdmin::ResolvePseudoUrl(char *lpszPseudoUrl, char **lppszServerPath, bool *lpbIsPeer)
{
	TRACE_MAPI(TRACE_ENTRY, "IECServiceAdmin::ResolvePseudoUrl", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServiceAdmin);
	return pThis->ResolvePseudoUrl(lpszPseudoUrl, lppszServerPath, lpbIsPeer);
}

///////////////////////
// IECSpooler
HRESULT ECMsgStore::xECSpooler::QueryInterface(REFIID iid , void** ppvObject)
{
	TRACE_MAPI(TRACE_ENTRY, "IECSpooler::QueryInterface", "%s", DBGGUIDToString(iid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, ECSpooler);
	HRESULT hr = pThis->QueryInterface(iid, ppvObject);
	TRACE_MAPI(TRACE_RETURN, "IECSpooler::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

ULONG ECMsgStore::xECSpooler::AddRef()
{
	TRACE_MAPI(TRACE_ENTRY, "IECSpooler::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore, ECSpooler);
	return pThis->AddRef();
}

ULONG ECMsgStore::xECSpooler::Release()
{
	TRACE_MAPI(TRACE_ENTRY, "IECSpooler::Release", "");
	METHOD_PROLOGUE_(ECMsgStore, ECSpooler);
	return pThis->Release();
}

HRESULT ECMsgStore::xECSpooler::GetMasterOutgoingTable(ULONG ulFlags, IMAPITable ** lppOutgoingTable)
{
	TRACE_MAPI(TRACE_ENTRY, "IECSpooler::GetMasterOutgoingTable", "");
	METHOD_PROLOGUE_(ECMsgStore, ECSpooler);
	return pThis->GetMasterOutgoingTable(ulFlags, lppOutgoingTable);
}

HRESULT ECMsgStore::xECSpooler::DeleteFromMasterOutgoingTable(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulFlags)
{
	TRACE_MAPI(TRACE_ENTRY, "IECSpooler::DeleteFromMasterOutgoingTable", "");
	METHOD_PROLOGUE_(ECMsgStore, ECSpooler);
	HRESULT hr = pThis->DeleteFromMasterOutgoingTable(cbEntryID, lpEntryID, ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IECSpooler::DeleteFromMasterOutgoingTable", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

/////////////////////////////////////////////////////
// Interface IMAPIOfflineMgr
//
ULONG ECMsgStore::xMAPIOfflineMgr::AddRef()
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::AddRef", "");
	return pThis->AddRef();
}

ULONG ECMsgStore::xMAPIOfflineMgr::Release()
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Release", "");

	return pThis->Release();
}

HRESULT ECMsgStore::xMAPIOfflineMgr::QueryInterface(REFIID iid, void **lpvoid)
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::QueryInterface", "%s", DBGGUIDToString(iid).c_str());
	HRESULT hr = pThis->QueryInterface(iid,lpvoid);
	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::SetCurrentState(ULONG ulFlags, ULONG ulMask, ULONG ulState, void* pReserved)
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::SetCurrentState", "flags=0x%08X, mask=0x%08X, state=0x%08X", ulFlags, ulMask, ulState);
	HRESULT hr = pThis->SetCurrentState(ulFlags, ulMask, ulState, pReserved);
	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::SetCurrentState", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::GetCapabilities(ULONG *pulCapabilities)
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::GetCapabilities", "");
	HRESULT hr = pThis->GetCapabilities(pulCapabilities);
	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::GetCapabilities", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::GetCurrentState(ULONG* pulState)
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::GetCurrentState", "");
	HRESULT hr = pThis->GetCurrentState(pulState);
	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::GetCurrentState", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::Placeholder1(void* ulTest)
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Placeholder1", "");

	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Placeholder1", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::Advise(ULONG ulFlags, MAPIOFFLINE_ADVISEINFO* pAdviseInfo, ULONG* pulAdviseToken)
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Advise", "0x%08X", ulFlags);
	HRESULT hr = pThis->Advise(ulFlags, pAdviseInfo, pulAdviseToken);
	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Advise", "%s, AdviseToken=0x%08X", GetMAPIErrorDescription(hr).c_str(), (pulAdviseToken)?pulAdviseToken:0);
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::Unadvise(ULONG ulFlags, ULONG ulAdviseToken)
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Unadvise", "0x%08X", ulFlags);
	HRESULT hr = pThis->Unadvise(ulFlags, ulAdviseToken);
	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Unadvise", "%s, AdviseToken=0x%08X", GetMAPIErrorDescription(hr).c_str(), ulAdviseToken);
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::Placeholder2()
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Placeholder2", "");

	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Placeholder2", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::Placeholder3()
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Placeholder3", "");

	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Placeholder3", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::Placeholder4()
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Placeholder4", "");

	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Placeholder4", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}
HRESULT ECMsgStore::xMAPIOfflineMgr::Placeholder5()
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Placeholder5", "");

	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Placeholder5", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}
HRESULT ECMsgStore::xMAPIOfflineMgr::Placeholder6()
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Placeholder6", "");

	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Placeholder6", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}
HRESULT ECMsgStore::xMAPIOfflineMgr::Placeholder7()
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Placeholder7", "");

	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Placeholder7", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMAPIOfflineMgr::Placeholder8()
{
	METHOD_PROLOGUE_(ECMsgStore, MAPIOfflineMgr);
	TRACE_MAPI(TRACE_ENTRY, "IMAPIOfflineMgr::Placeholder8", "");

	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IMAPIOfflineMgr::Placeholder8", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

/////////////////////////////////////////////////////
// Interface IProxyStoreObject
//
ULONG ECMsgStore::xProxyStoreObject::AddRef() 
{
	METHOD_PROLOGUE_(ECMsgStore, ProxyStoreObject);
	TRACE_MAPI(TRACE_ENTRY, "IProxyStoreObject::AddRef", "");
	ULONG ulRef = pThis->AddRef();
	TRACE_MAPI(TRACE_RETURN, "IProxyStoreObject::AddRef", "%d", ulRef);
	return ulRef;
}

ULONG ECMsgStore::xProxyStoreObject::Release()
{
	METHOD_PROLOGUE_(ECMsgStore, ProxyStoreObject);
	TRACE_MAPI(TRACE_ENTRY, "IProxyStoreObject::Release", "");
	ULONG ulRef = pThis->Release();
	TRACE_MAPI(TRACE_RETURN, "IProxyStoreObject::Release", "%d", ulRef);
	return ulRef;
}

HRESULT ECMsgStore::xProxyStoreObject::QueryInterface(REFIID iid, void **lpvoid)
{
	METHOD_PROLOGUE_(ECMsgStore, ProxyStoreObject);
	TRACE_MAPI(TRACE_ENTRY, "IProxyStoreObject::QueryInterface", "%s", DBGGUIDToString(iid).c_str());
	HRESULT hr = pThis->QueryInterface(iid,lpvoid);
	TRACE_MAPI(TRACE_RETURN, "IProxyStoreObject::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xProxyStoreObject::PlaceHolder1()
{
	METHOD_PROLOGUE_(ECMsgStore, ProxyStoreObject);
	TRACE_MAPI(TRACE_ENTRY, "IProxyStoreObject::PlaceHolder1", "");
	
	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IProxyStoreObject::PlaceHolder1", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xProxyStoreObject::UnwrapNoRef(LPVOID *ppvObject)
{
	METHOD_PROLOGUE_(ECMsgStore, ProxyStoreObject);
	TRACE_MAPI(TRACE_ENTRY, "IProxyStoreObject::UnwrapNoRef", "");
	HRESULT hr = pThis->UnwrapNoRef(ppvObject);
	TRACE_MAPI(TRACE_RETURN, "IProxyStoreObject::UnwrapNoRef", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xProxyStoreObject::PlaceHolder2()
{
	METHOD_PROLOGUE_(ECMsgStore, ProxyStoreObject);
	TRACE_MAPI(TRACE_ENTRY, "IProxyStoreObject::PlaceHolder2", "");
	
	HRESULT hr = MAPI_E_NO_SUPPORT;

	TRACE_MAPI(TRACE_RETURN, "IProxyStoreObject::PlaceHolder2", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

//////////////////////////////////
// IMsgStoreProxy interface

ULONG ECMsgStore::xMsgStoreProxy::AddRef() {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	return pThis->AddRef();
}

ULONG ECMsgStore::xMsgStoreProxy::Release() {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::Release", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	ULONG ulReturn = pThis->Release();
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::Release", "%d", ulReturn);
	return ulReturn;
}

HRESULT ECMsgStore::xMsgStoreProxy::QueryInterface(REFIID refiid, void **pInterface) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::QueryInterface", "%s", DBGGUIDToString(refiid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->QueryInterfaceProxy(refiid, pInterface);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::Advise(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulEventMask, LPMAPIADVISESINK lpAdviseSink, ULONG *lpulConnection) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::Advise", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	return pThis->Advise(cbEntryID, lpEntryID, ulEventMask, lpAdviseSink, lpulConnection);
}

HRESULT ECMsgStore::xMsgStoreProxy::Unadvise(ULONG ulConnection) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::Unadvise", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	return pThis->Unadvise(ulConnection);
}

HRESULT ECMsgStore::xMsgStoreProxy::CompareEntryIDs(ULONG cbEntryID1, LPENTRYID lpEntryID1, ULONG cbEntryID2, LPENTRYID lpEntryID2, ULONG ulFlags, ULONG *lpulResult) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::CompareEntryIDs", "flags: %d\ncb=%d  entryid1: %s\n cb=%d entryid2: %s", ulFlags, cbEntryID1, Util::bin2hex(cbEntryID1, (BYTE*)lpEntryID1).c_str(), cbEntryID2, Util::bin2hex(cbEntryID2, (BYTE*)lpEntryID2).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->CompareEntryIDs(cbEntryID1, lpEntryID1, cbEntryID2, lpEntryID2, ulFlags, lpulResult);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::CompareEntryIDs", "%s result=%s", GetMAPIErrorDescription(hr).c_str(), (lpulResult)?((*lpulResult == TRUE)?"TRUE":"FALSE") : "NULL");
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::OpenEntry(ULONG cbEntryID, LPENTRYID lpEntryID, LPCIID lpInterface, ULONG ulFlags, ULONG *lpulObjType, LPUNKNOWN *lppUnk) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::OpenEntry", "interface=%s", (lpInterface)?DBGGUIDToString(*lpInterface).c_str():"");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->OpenEntry(cbEntryID, lpEntryID, lpInterface, ulFlags, lpulObjType, lppUnk);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::OpenEntry", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::SetReceiveFolder(LPTSTR lpszMessageClass, ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::SetReceiveFolder", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr =pThis->SetReceiveFolder(lpszMessageClass, ulFlags, cbEntryID, lpEntryID);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::SetReceiveFolder", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::GetReceiveFolder(LPTSTR lpszMessageClass, ULONG ulFlags, ULONG *lpcbEntryID, LPENTRYID *lppEntryID, LPTSTR *lppszExplicitClass) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::GetReceiveFolder", "%s", (lpszMessageClass)?lpszMessageClass:"NULL");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->GetReceiveFolder(lpszMessageClass, ulFlags, lpcbEntryID, lppEntryID,lppszExplicitClass);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::GetReceiveFolder", "%s, entryid=%s", GetMAPIErrorDescription(hr).c_str(), (hr == hrSuccess && lppEntryID)?Util::bin2hex(*lpcbEntryID, (BYTE*)*lppEntryID).c_str():"NULL");
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::GetReceiveFolderTable(ULONG ulFlags, LPMAPITABLE *lppTable) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::GetReceiveFolderTable", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->GetReceiveFolderTable(ulFlags, lppTable);;
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::GetReceiveFolderTable", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::StoreLogoff(ULONG *lpulFlags) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::StoreLogoff", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	return pThis->StoreLogoff(lpulFlags);
}

HRESULT ECMsgStore::xMsgStoreProxy::AbortSubmit(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulFlags) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::AbortSubmit", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->AbortSubmit(cbEntryID, lpEntryID, ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::AbortSubmit", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::GetOutgoingQueue(ULONG ulFlags, LPMAPITABLE *lppTable) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::GetOutgoingQueue", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->GetOutgoingQueue(ulFlags, lppTable);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::GetOutgoingQueue", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::SetLockState(LPMESSAGE lpMessage,ULONG ulLockState) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::SetLockState", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->SetLockState(lpMessage, ulLockState);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::SetLockState", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::FinishedMsg(ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::FinishedMsg", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->FinishedMsg(ulFlags, cbEntryID, lpEntryID);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::FinishedMsg", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::NotifyNewMail(LPNOTIFICATION lpNotification) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::NotifyNewMail", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->NotifyNewMail(lpNotification);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::NotifyNewMail", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::GetLastError(HRESULT hError, ULONG ulFlags, LPMAPIERROR * lppMapiError) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::GetLastError", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->GetLastError(hError, ulFlags, lppMapiError);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::GetLastError", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::SaveChanges(ULONG ulFlags) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::SaveChanges", "flags=%d", ulFlags);
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr =  pThis->SaveChanges(ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::SaveChanges", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::GetProps(LPSPropTagArray lpPropTagArray, ULONG ulFlags, ULONG FAR * lpcValues, LPSPropValue FAR * lppPropArray) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::GetProps", "%s", PropNameFromPropTagArray(lpPropTagArray).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->GetProps(lpPropTagArray, ulFlags, lpcValues, lppPropArray);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::GetProps", "%s: %s", GetMAPIErrorDescription(hr).c_str(), PropNameFromPropArray(*lpcValues, *lppPropArray).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::GetPropList(ULONG ulFlags, LPSPropTagArray FAR * lppPropTagArray) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::GetPropList", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->GetPropList(ulFlags, lppPropTagArray);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::GetPropList", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::OpenProperty(ULONG ulPropTag, LPCIID lpiid, ULONG ulInterfaceOptions, ULONG ulFlags, LPUNKNOWN FAR * lppUnk) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::OpenProperty", "PropTag=%s, lpiid=%s", PropNameFromPropTag(ulPropTag).c_str(), DBGGUIDToString(*lpiid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->OpenProperty(ulPropTag, lpiid, ulInterfaceOptions, ulFlags, lppUnk);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::OpenProperty", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::SetProps(ULONG cValues, LPSPropValue lpPropArray, LPSPropProblemArray FAR * lppProblems) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::SetProps", "%s", PropNameFromPropArray(cValues, lpPropArray).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->SetProps(cValues, lpPropArray, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::SetProps", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::DeleteProps(LPSPropTagArray lpPropTagArray, LPSPropProblemArray FAR * lppProblems) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::DeleteProps", "%s", PropNameFromPropTagArray(lpPropTagArray).c_str());
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->DeleteProps(lpPropTagArray, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::DeleteProps", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::CopyTo(ULONG ciidExclude, LPCIID rgiidExclude, LPSPropTagArray lpExcludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpInterface, LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray FAR * lppProblems) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::CopyTo", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->CopyTo(ciidExclude, rgiidExclude, lpExcludeProps, ulUIParam, lpProgress, lpInterface, lpDestObj, ulFlags, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::CopyTo", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::CopyProps(LPSPropTagArray lpIncludeProps, ULONG ulUIParam, LPMAPIPROGRESS lpProgress, LPCIID lpInterface, LPVOID lpDestObj, ULONG ulFlags, LPSPropProblemArray FAR * lppProblems) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::CopyProps", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->CopyProps(lpIncludeProps, ulUIParam, lpProgress, lpInterface, lpDestObj, ulFlags, lppProblems);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::CopyProps", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::GetNamesFromIDs(LPSPropTagArray * pptaga, LPGUID lpguid, ULONG ulFlags, ULONG * pcNames, LPMAPINAMEID ** pppNames) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::GetNamesFromIDs", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->GetNamesFromIDs(pptaga, lpguid, ulFlags, pcNames, pppNames);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::GetNamesFromIDs", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xMsgStoreProxy::GetIDsFromNames(ULONG cNames, LPMAPINAMEID * ppNames, ULONG ulFlags, LPSPropTagArray * pptaga) {
	TRACE_MAPI(TRACE_ENTRY, "IMsgStoreProxy::GetIDsFromNames", "");
	METHOD_PROLOGUE_(ECMsgStore, MsgStoreProxy);
	HRESULT hr = pThis->GetIDsFromNames(cNames, ppNames, ulFlags, pptaga);
	TRACE_MAPI(TRACE_RETURN, "IMsgStoreProxy::GetIDsFromNames", "%s\n%s", GetMAPIErrorDescription(hr).c_str(), MapiNameIdListToString(cNames, ppNames, (hr==hrSuccess)?(*pptaga):NULL).c_str());
	return hr;
}

//////////////////////////////////
// IECMultiStoreTable interface

ULONG ECMsgStore::xECMultiStoreTable::AddRef() {
	TRACE_MAPI(TRACE_ENTRY, "IECMultiStoreTable::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore, ECMultiStoreTable);
	return pThis->AddRef();
}

ULONG ECMsgStore::xECMultiStoreTable::Release() {
	TRACE_MAPI(TRACE_ENTRY, "IECMultiStoreTable::Release", "");
	METHOD_PROLOGUE_(ECMsgStore, ECMultiStoreTable);
	ULONG ulReturn = pThis->Release();
	TRACE_MAPI(TRACE_RETURN, "IECMultiStoreTable::Release", "%d", ulReturn);
	return ulReturn;
}

HRESULT ECMsgStore::xECMultiStoreTable::QueryInterface(REFIID refiid, void **pInterface) {
	TRACE_MAPI(TRACE_ENTRY, "IECMultiStoreTable::QueryInterface", "%s", DBGGUIDToString(refiid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, ECMultiStoreTable);
	HRESULT hr = pThis->QueryInterfaceProxy(refiid, pInterface);
	TRACE_MAPI(TRACE_RETURN, "IECMultiStoreTable::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECMultiStoreTable::OpenMultiStoreTable(LPENTRYLIST lpMsgList, ULONG ulFlags, LPMAPITABLE *lppTable) {
	TRACE_MAPI(TRACE_ENTRY, "IECMultiStoreTable::OpenMultiStoreTable", "");
	METHOD_PROLOGUE_(ECMsgStore, ECMultiStoreTable);
	HRESULT hr = pThis->OpenMultiStoreTable(lpMsgList, ulFlags, lppTable);
	TRACE_MAPI(TRACE_RETURN, "IECMultiStoreTable::OpenMultiStoreTable", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

////////////////////////////////////
// IECLicense interface
ULONG ECMsgStore::xECLicense::AddRef() {
	TRACE_MAPI(TRACE_ENTRY, "IECLicense::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore, ECLicense);
	return pThis->AddRef();
}

ULONG ECMsgStore::xECLicense::Release() {
	TRACE_MAPI(TRACE_ENTRY, "IECLicense::Release", "");
	METHOD_PROLOGUE_(ECMsgStore, ECLicense);
	ULONG ulReturn = pThis->Release();
	TRACE_MAPI(TRACE_RETURN, "IECLicense::Release", "%d", ulReturn);
	return ulReturn;
}

HRESULT ECMsgStore::xECLicense::QueryInterface(REFIID refiid, void **pInterface) {
	TRACE_MAPI(TRACE_ENTRY, "IECLicense::QueryInterface", "%s", DBGGUIDToString(refiid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, ECLicense);
	HRESULT hr = pThis->QueryInterfaceProxy(refiid, pInterface);
	TRACE_MAPI(TRACE_RETURN, "IECLicense::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECLicense::LicenseAuth(unsigned char *lpData, unsigned int ulSize, unsigned char **lppResponse, unsigned int *lpulResponseSize) {
	TRACE_MAPI(TRACE_ENTRY, "IECLicense::LicenseAuth","");
	METHOD_PROLOGUE_(ECMsgStore, ECLicense);
	HRESULT hr = pThis->LicenseAuth(lpData, ulSize, lppResponse, lpulResponseSize);
	TRACE_MAPI(TRACE_RETURN, "IECLicense::LicenseAuth","");
	return hr;
}

HRESULT ECMsgStore::xECLicense::LicenseCapa(char ***lppszData, unsigned int *lpulSize) {
	TRACE_MAPI(TRACE_ENTRY, "IECLicense::LicenseCapa","");
	METHOD_PROLOGUE_(ECMsgStore, ECLicense);
	HRESULT hr = pThis->LicenseCapa(lppszData, lpulSize);
	TRACE_MAPI(TRACE_RETURN, "IECLicense::LicenseCapa","");
	return hr;
}


//////////////////////////////////
// IECServerBehavior interface

ULONG ECMsgStore::xECServerBehavior::AddRef() {
	TRACE_MAPI(TRACE_ENTRY, "IECServerBehavior::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServerBehavior);
	return pThis->AddRef();
}

ULONG ECMsgStore::xECServerBehavior::Release() {
	TRACE_MAPI(TRACE_ENTRY, "IECServerBehavior::Release", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServerBehavior);
	ULONG ulReturn = pThis->Release();
	TRACE_MAPI(TRACE_RETURN, "IECServerBehavior::Release", "%d", ulReturn);
	return ulReturn;
}

HRESULT ECMsgStore::xECServerBehavior::QueryInterface(REFIID refiid, void **pInterface) {
	TRACE_MAPI(TRACE_ENTRY, "IECServerBehavior::QueryInterface", "%s", DBGGUIDToString(refiid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, ECServerBehavior);
	HRESULT hr = pThis->QueryInterfaceProxy(refiid, pInterface);
	TRACE_MAPI(TRACE_RETURN, "IECServerBehavior::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServerBehavior::GetServerBehavior(ULONG *lpulBehavior) {
	TRACE_MAPI(TRACE_ENTRY, "IECServerBehavior::GetServerBehavior", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServerBehavior);
	HRESULT hr = pThis->GetServerBehavior(lpulBehavior);
	TRACE_MAPI(TRACE_RETURN, "IECServerBehavior::GetServerBehavior", "%u, %s", lpulBehavior ? *lpulBehavior : 0, GetMAPIErrorDescription(hr).c_str());
	return hr;
}

//////////////////////////////////
// IECServerMimic interface

ULONG ECMsgStore::xECServerMimic::AddRef() {
	TRACE_MAPI(TRACE_ENTRY, "IECServerMimic::AddRef", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServerMimic);
	return pThis->AddRef();
}

ULONG ECMsgStore::xECServerMimic::Release() {
	TRACE_MAPI(TRACE_ENTRY, "IECServerMimic::Release", "");
	METHOD_PROLOGUE_(ECMsgStore, ECServerMimic);
	ULONG ulReturn = pThis->Release();
	TRACE_MAPI(TRACE_RETURN, "IECServerMimic::Release", "%d", ulReturn);
	return ulReturn;
}

HRESULT ECMsgStore::xECServerMimic::QueryInterface(REFIID refiid, void **pInterface) {
	TRACE_MAPI(TRACE_ENTRY, "IECServerMimic::QueryInterface", "%s", DBGGUIDToString(refiid).c_str());
	METHOD_PROLOGUE_(ECMsgStore, ECServerMimic);
	HRESULT hr = pThis->QueryInterfaceProxy(refiid, pInterface);
	TRACE_MAPI(TRACE_RETURN, "IECServerMimic::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMsgStore::xECServerMimic::SetServerBehavior(ULONG ulBehavior) {
	TRACE_MAPI(TRACE_ENTRY, "IECServerMimic::SetServerBehavior", "%u", ulBehavior);
	METHOD_PROLOGUE_(ECMsgStore, ECServerMimic);
	HRESULT hr = pThis->SetServerBehavior(ulBehavior);
	TRACE_MAPI(TRACE_RETURN, "IECServerMimic::SetServerBehavior", "");
	return hr;
}
