/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "stringutil.h"
#include "ZarafaUser.h"

#include <sstream>

objectdetails_t::objectdetails_t(userobject_type_t objtype) : m_type(objtype) {}
objectdetails_t::objectdetails_t() : m_type(USEROBJECT_TYPE_USER) {}
	
unsigned int objectdetails_t::GetPropInt(const std::string &propname) {
	item_iterator item = m_mapProps.find(propname);
    if(item != m_mapProps.end()) {
        return atoi(item->second.c_str());
    } else return 0;
}

bool objectdetails_t::GetPropBool(const std::string &propname) {
	item_iterator item = m_mapProps.find(propname);
    if(item != m_mapProps.end()) {
        return atoi(item->second.c_str());
    } else return false;
}

std::string	objectdetails_t::GetPropString(const std::string &propname) {
	item_iterator item = m_mapProps.find(propname);
    if(item != m_mapProps.end()) {
        return item->second;
    } else return std::string();
}
	
void objectdetails_t::SetPropInt(const std::string &propname, unsigned int value) {
    m_mapProps[propname] = stringify(value);
}

void objectdetails_t::SetPropBool(const std::string &propname, bool value) {
    m_mapProps[propname] = value ? "1" : "0";
}

void objectdetails_t::SetPropString(const std::string &propname, const std::string &value) {
    m_mapProps[propname] = value;
}

void objectdetails_t::AddPropString(const std::string &propname, const std::string &value) {
	m_mapMVProps[propname].push_back(value);
}

std::list<std::string> objectdetails_t::GetPropList(const std::string &propname) {
	mvitem_iterator mvitem = m_mapMVProps.find(propname);
	if (mvitem != m_mapMVProps.end()) return mvitem->second;
	else return std::list<std::string>();
}

void objectdetails_t::ClearPropList(const std::string &propname) {
	m_mapMVProps[propname].clear();
}


userobject_type_t objectdetails_t::GetType() {
    return m_type;
}

void objectdetails_t::CopyFrom(objectdetails_t &from) {
    item_iterator iterProps;
    
    for(iterProps = from.m_mapProps.begin(); iterProps != from.m_mapProps.end(); iterProps++) {
        this->m_mapProps[iterProps->first] = iterProps->second;
    }
}

serverdetails_t::serverdetails_t(const std::string &servername)
: m_strServerName(servername)
, m_ulHttpPort(0)
, m_ulSslPort(0)
{ }

void serverdetails_t::SetHostAddress(const std::string &hostaddress) {
	m_strHostAddress = hostaddress;
}

void serverdetails_t::SetFilePath(const std::string &filepath) {
	m_strFilePath = filepath;
}

void serverdetails_t::SetHttpPort(unsigned port) {
	m_ulHttpPort = port;
}

void serverdetails_t::SetSslPort(unsigned port) {
	m_ulSslPort = port;
}

const std::string& serverdetails_t::GetServerName() const {
	return m_strServerName;
}

const std::string& serverdetails_t::GetHostAddress() const {
	return m_strHostAddress;
}

unsigned serverdetails_t::GetHttpPort() const {
	return m_ulHttpPort;
}

unsigned serverdetails_t::GetSslPort() const {
	return m_ulSslPort;
}

std::string serverdetails_t::GetFilePath() const {
	if (!m_strFilePath.empty())
		return "file://"+m_strFilePath;
	return std::string();
}

std::string serverdetails_t::GetHttpPath() const {
	if (!m_strHostAddress.empty() && m_ulHttpPort > 0) {
		std::ostringstream oss;
		oss << "http://" << m_strHostAddress << ":" << m_ulHttpPort << "/zarafa";
		return oss.str();
	}
	return std::string();	
}

std::string serverdetails_t::GetSslPath() const {
	if (!m_strHostAddress.empty() && m_ulSslPort > 0) {
		std::ostringstream oss;
		oss << "https://" << m_strHostAddress << ":" << m_ulSslPort << "/zarafa";
		return oss.str();
	}
	return std::string();	
}
