/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef ZARAFAUSER_H
#define ZARAFAUSER_H

#include <map>
#include <list>
#include <string>
#include "ECDefs.h"

typedef std::string objectid_t;

/* Object properties
 * Prefix: OB_PROP
 * Type: S - String
 *       B - Boolean
 *       I - Integer
 * Name: Property name
 */
#define OB_PROP_S_FULLNAME		"fullname"
#define OB_PROP_S_LOGIN			"loginname"
#define OB_PROP_S_PASSWORD		"password"
#define OB_PROP_S_COMPANYNAME	"companyname"
#define OB_PROP_I_ADMINLEVEL	"adminlevel"
#define OB_PROP_B_NONACTIVE		"isnonactive"
#define OB_PROP_S_EMAIL			"emailaddress"
#define OB_PROP_S_SYSADMIN		"sysadmin"
#define OB_PROP_L_CERTIFICATE	"certificate"
#define OB_PROP_L_SENDAS		"sendas"
#define OB_PROP_S_EXCH_DN		"exchangedn"
#define OB_PROP_S_EXTERNID		"externid"
#define OB_PROP_S_SERVERNAME	"servername"
#define OB_PROP_S_TELEPHONE		"telephone"
#define OB_PROP_S_DEPARTMENT		"department"
#define OB_PROP_S_LOCATION		"location"
#define OB_PROP_S_FAX			"fax"

typedef std::map<std::string, std::string>::iterator item_iterator;
typedef std::map<std::string, std::list<std::string> >::iterator mvitem_iterator;

/** One user's / group's details
 */
class objectdetails_t {
public:
	objectdetails_t(userobject_type_t objtype);
	objectdetails_t();
	
	unsigned int 	GetPropInt(const std::string &propname);
	bool			GetPropBool(const std::string &propname);
	std::string		GetPropString(const std::string &propname);
	
	void			SetPropInt(const std::string &propname, unsigned int value);
	void			SetPropBool(const std::string &propname, bool value);
	void			SetPropString(const std::string &propname, const std::string &alue);

	/* "mv" props */
	void			AddPropString(const std::string &propname, const std::string &value);
	std::list<std::string> GetPropList(const std::string &propname);
	void			ClearPropList(const std::string &propname);

	void			CopyFrom(objectdetails_t &from);

	userobject_type_t GetType();
	
private:
	userobject_type_t m_type;
	std::map<std::string, std::string> m_mapProps;
	std::map<std::string, std::list<std::string> > m_mapMVProps;
};

/** Quota Details
 */
class quotadetails_t {
public:
	quotadetails_t() : 
		bUseDefaultQuota(true), bIsUserDefaultQuota(false),
		llWarnSize(0), llSoftSize(0), llHardSize(0) {}

	bool bUseDefaultQuota;
	bool bIsUserDefaultQuota; /* Default quota for users within company */
	long long llWarnSize;
	long long llSoftSize;
	long long llHardSize;
};

/** Server Details
*/
class serverdetails_t {
public:
	serverdetails_t(const std::string &servername = std::string());

	void SetHostAddress(const std::string &hostaddress);
	void SetFilePath(const std::string &filepath);
	void SetHttpPort(unsigned port);
	void SetSslPort(unsigned port);

	const std::string&	GetServerName() const;
	const std::string&	GetHostAddress() const;
	unsigned			GetHttpPort() const;
	unsigned			GetSslPort() const;

	std::string	GetFilePath() const;
	std::string	GetHttpPath() const;
	std::string	GetSslPath() const;

private:
	std::string m_strServerName;
	std::string m_strHostAddress;
	std::string m_strFilePath;
	unsigned	m_ulHttpPort;
	unsigned	m_ulSslPort;
};

#endif
