/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "ZarafaUtil.h"
#include "mapicode.h"
#include "../common/stringutil.h"
#include "../common/base64.h"

#include <mapidefs.h>
#include "ECGuid.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

bool IsZarafaEntryId(ULONG cb, LPBYTE lpEntryId)
{
	bool	bZarafaEntry = false;
	EID*	peid = NULL;

	if(lpEntryId == NULL)
		goto exit;

	peid = (PEID)lpEntryId;

	if( (cb == sizeof(EID) && peid->ulVersion == 1) ||
		(cb == sizeof(EID_V0) && peid->ulVersion == 0 ) )
		bZarafaEntry = true;

	//TODO: maybe also a check on objType
exit:
	return bZarafaEntry;
}

bool ValidateZarafaEntryId(ULONG cb, LPBYTE lpEntryId, unsigned int ulCheckType)
{

	bool	bOk = false;
	EID*	peid = NULL;

	if(lpEntryId == NULL)
		goto exit;

	peid = (PEID)lpEntryId;

	if( ((cb == sizeof(EID) && peid->ulVersion == 1) ||
		 (cb == sizeof(EID_V0) && peid->ulVersion == 0 ) ) &&
		 peid->ulType == ulCheckType)
	{
		bOk = true;
	}

exit:
	return bOk;
}

/**
 * Validate a zarafa entryid list on a specific mapi object type
 * 
 * @param lpMsgList		Pointer to an ENTRYLIST structure that contains the number 
 *						of items to validate and an array of ENTRYID structures identifying the items.
 * @param ulCheckType	Contains the type of the objects in the lpMsgList. 
 *
 * @return bool			true if all the items in the lpMsgList matches with the object type
 */
bool ValidateZarafaEntryList(LPENTRYLIST lpMsgList, unsigned int ulCheckType)
{

	bool	bOk = true;
	EID*	peid = NULL;

	if(lpMsgList == NULL) {
		bOk = false;
		goto exit;
	}

	for(ULONG i=0; i < lpMsgList->cValues; i++)
	{
		peid = (PEID)lpMsgList->lpbin[i].lpb;

		if( !(((lpMsgList->lpbin[i].cb == sizeof(EID) && peid->ulVersion == 1) ||
			 (lpMsgList->lpbin[i].cb == sizeof(EID_V0) && peid->ulVersion == 0 ) ) &&
			 peid->ulType == ulCheckType))
		{
			bOk = false;
			goto exit;
		}
	}

exit:
	return bOk;
}

ECRESULT GetStoreGuidFromEntryId(ULONG cb, LPBYTE lpEntryId, LPGUID lpguidStore)
{
	ECRESULT er = erSuccess;
	EID*	peid = NULL;

	if(lpEntryId == NULL || lpguidStore == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	peid = (PEID)lpEntryId;

	if(!((cb == sizeof(EID) && peid->ulVersion == 1) ||
		 (cb == sizeof(EID_V0) && peid->ulVersion == 0 )) )
	{
		er = ZARAFA_E_INVALID_ENTRYID;
		goto exit;
	}

	memcpy(lpguidStore, &peid->guid, sizeof(GUID));

exit:
	return er;
}

ECRESULT GetObjTypeFromEntryId(ULONG cb, LPBYTE lpEntryId, unsigned int* lpulObjType)
{
	ECRESULT er = erSuccess;
	EID*	peid = NULL;

	if(lpEntryId == NULL || lpulObjType == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	peid = (PEID)lpEntryId;

	if(!((cb == sizeof(EID) && peid->ulVersion == 1) ||
		 (cb == sizeof(EID_V0) && peid->ulVersion == 0 )) )
	{
		er = ZARAFA_E_INVALID_ENTRYID;
		goto exit;
	}


	*lpulObjType = peid->ulType;

exit:
	return er;
}

ECRESULT GetObjTypeFromEntryId(entryId sEntryId,  unsigned int* lpulObjType) {
    return GetObjTypeFromEntryId(sEntryId.__size, sEntryId.__ptr, lpulObjType);
}

ECRESULT GetStoreGuidFromEntryId(entryId sEntryId, LPGUID lpguidStore) {
    return GetStoreGuidFromEntryId(sEntryId.__size, sEntryId.__ptr, lpguidStore);
}

HRESULT HrGetStoreGuidFromEntryId(ULONG cb, LPBYTE lpEntryId, LPGUID lpguidStore)
{
	ECRESULT er = GetStoreGuidFromEntryId(cb, lpEntryId, lpguidStore);

	return ZarafaErrorToMAPIError(er);
}

HRESULT HrGetObjTypeFromEntryId(ULONG cb, LPBYTE lpEntryId, unsigned int* lpulObjType)
{
	ECRESULT er = GetObjTypeFromEntryId(cb, lpEntryId, lpulObjType);

	return ZarafaErrorToMAPIError(er);
}

ECRESULT ABEntryIDToID(ULONG cb, LPBYTE lpEntryId, unsigned int* lpulID, std::string* lpstrExternId, unsigned int* lpulType)
{
	ECRESULT		er = erSuccess;
	PABEID			lpABEID = NULL;
	unsigned int	ulID = 0;
	std::string		strExternId;

	if(lpEntryId == NULL || lpulID == NULL || cb < CbNewABEID("")) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	lpABEID = (PABEID)lpEntryId;
	ulID = lpABEID->ulId;

	if (lpABEID->ulVersion == 1)
		strExternId = base64_decode(lpABEID->szExId);

	*lpulID = ulID;

	if (lpstrExternId)
		*lpstrExternId = strExternId;

	if (lpulType)
		*lpulType = lpABEID->ulType;

exit:
	return er;
}

ECRESULT ABEntryIDToID(ULONG cb, LPBYTE lpEntryId, unsigned int* lpulID, unsigned int* lpulType)
{
	return ABEntryIDToID(cb, lpEntryId, lpulID, NULL, lpulType);
}

ECRESULT ABEntryIDToID(entryId* lpsEntryId, unsigned int* lpulID, std::string* lpstrExternId, unsigned int* lpulType)
{
	ECRESULT er = erSuccess;

	if(lpsEntryId == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = ABEntryIDToID(lpsEntryId->__size, lpsEntryId->__ptr, lpulID, lpstrExternId, lpulType);

exit:
	return er;
}

ECRESULT ABEntryIDToID(entryId* lpsEntryId, unsigned int* lpulID, unsigned int* lpulType)
{
	ECRESULT er = erSuccess;

	if(lpsEntryId == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = ABEntryIDToID(lpsEntryId->__size, lpsEntryId->__ptr, lpulID, NULL, lpulType);

exit:
	return er;
}

HRESULT HrABEntryIDToID(ULONG cb, LPBYTE lpEntryId, unsigned int* lpulID, unsigned int* lpulType)
{
	HRESULT hr = S_OK;

	if(lpEntryId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = ZarafaErrorToMAPIError( ABEntryIDToID(cb, lpEntryId, lpulID, lpulType) );

exit:
	return hr;
}

ECRESULT SIEntryIDToID(ULONG cb, LPBYTE lpInstanceId, LPGUID guidServer, unsigned int *lpulInstanceId, unsigned int *lpulTag)
{
	ECRESULT er = erSuccess;
	LPSIEID lpInstanceEid = NULL;

	if (lpInstanceId == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	lpInstanceEid = (LPSIEID)lpInstanceId;

	if (guidServer)
		memcpy(guidServer, (LPBYTE)lpInstanceEid + sizeof(SIEID), sizeof(GUID));
	if (lpulInstanceId)
		*lpulInstanceId = lpInstanceEid->ulId;
	if (lpulTag)
		*lpulTag = lpInstanceEid->ulType;

exit:
	return er;
}

HRESULT HrSIEntryIDToID(ULONG cb, LPBYTE lpInstanceId, LPGUID guidServer, unsigned int *lpulInstanceId, unsigned int *lpulTag)
{
	HRESULT hr = S_OK;

	if(lpInstanceId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = ZarafaErrorToMAPIError( SIEntryIDToID(cb, lpInstanceId, guidServer, lpulInstanceId, lpulTag) );

exit:
	return hr;
}

ECRESULT ABIDToEntryID(struct soap *soap, unsigned int ulID, const std::string& strExternId, unsigned long ulType, entryId *lpsEntryId)
{
	ECRESULT		er			= erSuccess;
	PABEID			lpUserEid	= NULL;
	std::string		strEncExId  = base64_encode((unsigned char*)strExternId.c_str(), strExternId.size());
	unsigned int	ulLen       = 0;

	if (lpsEntryId == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	ulLen = CbNewABEID(strEncExId.c_str());
	lpUserEid = (PABEID)s_alloc<char>(soap, ulLen);
	memset(lpUserEid, 0, ulLen);
	lpUserEid->ulId = ulID;
	lpUserEid->ulType = ulType;
	memcpy(&lpUserEid->guid, MUIDECSAB_SERVER, sizeof(GUID));

	// If the externid is non-empty, we'll create a V1 entry id.
	if (!strExternId.empty())
	{
		lpUserEid->ulVersion = 1;
		// avoid FORTIFY_SOURCE checks in strcpy to an address that the compiler thinks is 1 size large
		memcpy(lpUserEid->szExId, strEncExId.c_str(), strEncExId.length()+1);
	}

	lpsEntryId->__size = ulLen;
	lpsEntryId->__ptr = (unsigned char*)lpUserEid;

exit:
	return er;
}

ECRESULT SIIDToEntryID(struct soap *soap, LPGUID guidServer, unsigned int ulInstanceId, unsigned int ulTag, entryId *lpsInstanceId)
{
	ECRESULT er = erSuccess;
	LPSIEID lpInstanceEid = NULL;
	ULONG ulSize = 0;

	if (lpsInstanceId == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	ulSize = sizeof(SIEID) + sizeof(GUID);

	lpInstanceEid = (LPSIEID)s_alloc<char>(soap, ulSize);
	memset(lpInstanceEid, 0, sizeof(ulSize));

	lpInstanceEid->ulId = ulInstanceId;
	lpInstanceEid->ulType = ulTag;
	memcpy(&lpInstanceEid->guid, MUIDECSI_SERVER, sizeof(GUID));

	memcpy((char *)lpInstanceEid + sizeof(SIEID), guidServer, sizeof(GUID));

	lpsInstanceId->__size = ulSize;
	lpsInstanceId->__ptr = (unsigned char *)lpInstanceEid;

exit:
	return er;
}

ECRESULT SIEntryIDToID(entryId* sInstanceId, LPGUID guidServer, unsigned int *lpulInstanceId, unsigned int *lpulTag)
{
	ECRESULT er = erSuccess;

	if (sInstanceId == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = SIEntryIDToID(sInstanceId->__size, sInstanceId->__ptr, guidServer, lpulInstanceId, lpulTag);

exit:
	return er;
}

// NOTE: when using this function, we can never be sure that USEROBJECT_TYPE_USER is correct; the object
// MAY be USEROBJECT_TYPE_NONACTIVE. You must handle this in your code so that if it fails with USEROBJECT_TYPE_USER,
// you must retry with USEROBJECT_TYPE_NONACTIVE.
ECRESULT MAPITypeToType(unsigned int ulMAPIType, userobject_type_t *lpsUserObjType)
{
	ECRESULT			er = erSuccess;
	userobject_type_t	sUserObjType = USEROBJECT_TYPE_USER;

	if (lpsUserObjType == NULL)
	{
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	switch (ulMAPIType)
	{
	case MAPI_MAILUSER:
		sUserObjType = USEROBJECT_TYPE_USER;
		break;

	case MAPI_DISTLIST:
		sUserObjType = USEROBJECT_TYPE_GROUP;
		break;
		
	case MAPI_ABCONT:
		sUserObjType = USEROBJECT_TYPE_COMPANY;
		break;
		
	default:
		er = ZARAFA_E_INVALID_TYPE;
		goto exit;
	}

	*lpsUserObjType = sUserObjType;

exit:
	return er;
}

ECRESULT TypeToMAPIType(userobject_type_t sUserObjType, unsigned int *lpulMAPIType)
{
	ECRESULT		er = erSuccess;
	unsigned int	ulMAPIType = 0;

	if (lpulMAPIType == NULL)
	{
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	// Check for correctness of mapping!
	switch (sUserObjType)
	{
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		ulMAPIType = MAPI_MAILUSER;
		break;

	case USEROBJECT_TYPE_GROUP:
		ulMAPIType = MAPI_DISTLIST;
		break;

	case USEROBJECT_TYPE_COMPANY:
		/*
		 * HOSTED FIXME: This argument is fake and wrong, but is the only way to
		 * inform the client about a change of a company without breaking compatibility
		 * with 6.00 clients. Preferably we should pass USEROBJECT_TYPE_* as type which
		 * is far more flexible but it is too late for that now.
		 */
		ulMAPIType = MAPI_ABCONT;
		break;
		
    case USEROBJECT_TYPE_ADDRESSLIST:
        ulMAPIType = MAPI_ABCONT;
        break;

	default:
		er = ZARAFA_E_INVALID_TYPE;
		goto exit;
	}

	*lpulMAPIType = ulMAPIType;

exit:
	return er;
}
