/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "Zarafa.h"
#include "ZarafaCode.h"

#include <mapidefs.h>
#include <mapitags.h>
#include "mapiext.h"

#include <EMSAbTag.h>

#include "SOAPUtils.h"
#include "ECABObjectTable.h"
#include "ECSecurity.h"

#include "ECSession.h"
#include "ECSessionManager.h"
#include "stringutil.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

ECABObjectTable::ECABObjectTable(ECSession *lpSession, unsigned int ulABId, unsigned int ulABType, unsigned int ulABParentId, unsigned int ulABParentType) : ECGenericObjectTable(lpSession, ulABType)
{
	ECODAB* lpODAB = new ECODAB;
	lpODAB->ulABId = ulABId;
	lpODAB->ulABType = ulABType;
	lpODAB->ulABParentId = ulABParentId;
	lpODAB->ulABParentType = ulABParentType;

	// Set dataobject
	m_lpObjectData = lpODAB;

	// Set callback function for queryrowdata
	m_lpfnQueryRowData = QueryRowData;

	// Set the default sort order for Address Book entries
	struct sortOrder sObjectType[] = {
		{
			/*
			 * The Global Address Book must be the first entry
			 * in getHierarchyTable() otherwise it will not
			 * be set as default directory!
			 *
			 * To support future hierarchy additions like
			 * company subcontainers, or other global containers
			 * besides the GAB, we should sort the table using the
			 * PR_EMS_AB_HIERARCHY_PATH property which is formatted
			 * as followed for the GAB + 2 example companies:
			 *
			 * 		/Global Address Book
			 *		/Global Address Book/Company A
			 *		/Global Address Book/Company B
			 */
			PR_EMS_AB_HIERARCHY_PATH,
			EC_TABLE_SORT_ASCEND,
		},
	};

	struct sortOrderArray sDefaultSortOrder = {
		sObjectType,
		sizeof(sObjectType) / sizeof(sObjectType[0]),
	};

	SetSortOrder(&sDefaultSortOrder, 0, 0);
}

ECABObjectTable::~ECABObjectTable()
{
	if(m_lpObjectData)
		delete (ECODAB*)m_lpObjectData;
}

ECRESULT ECABObjectTable::GetColumnsAll(ECListInt* lplstProps)
{
	ECRESULT		er = erSuccess;
	ECObjectTableMap::iterator		iterObjects;
	ECODAB *lpODAB = (ECODAB*)m_lpObjectData;

	pthread_mutex_lock(&m_hLock);

	ASSERT(lplstProps != NULL);
	
	//List always empty
	lplstProps->clear();

	// Add some generated and standard properties
	lplstProps->push_back(PR_DISPLAY_NAME);
	lplstProps->push_back(PR_ENTRYID);
	lplstProps->push_back(PR_DISPLAY_TYPE);
	lplstProps->push_back(PR_INSTANCE_KEY);
	lplstProps->push_back(PR_RECORD_KEY);
	lplstProps->push_back(PR_OBJECT_TYPE);
	lplstProps->push_back(PR_AB_PROVIDER_ID);
	lplstProps->push_back(PR_LAST_MODIFICATION_TIME);
	lplstProps->push_back(PR_CREATION_TIME);

	if(lpODAB->ulABType == MAPI_ABCONT) {
	    // Hierarchy table
    	lplstProps->push_back(PR_CONTAINER_FLAGS);
    	lplstProps->push_back(PR_DEPTH);
    	lplstProps->push_back(PR_EMS_AB_CONTAINERID);
	} else {
	    // Contents table
	    lplstProps->push_back(PR_SMTP_ADDRESS);
	}

	pthread_mutex_unlock(&m_hLock);


	return er;
}

ECRESULT ECABObjectTable::ReloadTableMVData(ECObjectTableList* lplistRows, ECListInt* lplistMVPropTag)
{
	ECRESULT			er = erSuccess;
	ECListIntIterator	iterListMVPropTag;

	ECObjectTableList::iterator	iterListRows;

	ASSERT(lplistMVPropTag->size() <2); //FIXME: Limit of one 1 MV column

	// scan for MV-props and add rows

	// Add items to list

	return er;
}


ECRESULT ECABObjectTable::GetMVRowCount(unsigned int ulObjId, unsigned int *lpulCount)
{
	ECRESULT er = erSuccess;
    ECObjectTableList			listRows;
	ECObjectTableList::iterator iterListRows;
	ECObjectTableMap::iterator		iterIDs;

	ECListInt::iterator	iterListMVPropTag;

	pthread_mutex_lock(&m_hLock);

	*lpulCount = 0;

	pthread_mutex_unlock(&m_hLock);

	return er;
}

ECRESULT ECABObjectTable::QueryRowData(ECGenericObjectTable *lpThis, struct soap *soap, ECSession *lpSession, ECObjectTableList* lpRowList, struct propTagArray *lpsPropTagArray, void* lpObjectData, struct rowSet **lppRowSet, bool bTableData, bool bTableLimit)
{
	ECRESULT er = erSuccess;
	ECODAB *lpODAB = (ECODAB*)lpObjectData;
	struct rowSet	*lpsRowSet = NULL;

	ASSERT(lpRowList != NULL);

	if (lpODAB->ulABType == MAPI_ABCONT)
		er = lpSession->GetUserManagement()->QueryHierarchyRowData(soap, lpRowList, lpsPropTagArray, &lpsRowSet);
	else if (lpODAB->ulABType == MAPI_MAILUSER || lpODAB->ulABType == MAPI_DISTLIST)
		er = lpSession->GetUserManagement()->QueryContentsRowData(soap, lpRowList, lpsPropTagArray, &lpsRowSet);
	else
		er = ZARAFA_E_INVALID_TYPE;
	if(er != erSuccess)
		goto exit;
		
	*lppRowSet = lpsRowSet;
	
exit:

	return er;
}

ECRESULT ECABObjectTable::GetHierarchyCompanies(list<localobjectdetails_t> **lppCompanies, unsigned int ulFlags)
{
	ECRESULT er = erSuccess;
	list<localobjectdetails_t> *lpCompanies = NULL;

	er = lpSession->GetSecurity()->GetViewableCompanyIds(&lpCompanies, ulFlags);
	if (er != erSuccess)
		goto exit;

	/*
	 * This looks a bit hackish, and it probably can be considered as such.
	 * If the size of lpObjects is 1 then the user can only view a single company space (namely his own).
	 * There are 2 reasons this could happen, either when working in a non-hosted environment or
	 * when the company does not have enough permissions to view any other company spaces.
	 * Having a Global Address Book _and_ a container for the company will only look strange and
	 * confusing for the user. So lets delete the entry altogether.
	 */
	if (lpCompanies->size() == 1)
		lpCompanies->clear();

	*lppCompanies = lpCompanies;

exit:
	return er;
}

bool IsSystemObject(localobjectdetails_t details) 
{ 
	return (details.GetType() == USEROBJECT_TYPE_GROUP && details.ulId == ZARAFA_UID_EVERYONE) ||
		   (details.GetType() == USEROBJECT_TYPE_USER && details.ulId == ZARAFA_UID_SYSTEM); 
} 

// Loads either a contents table or a hierarchy table for an IABContainer
ECRESULT ECABObjectTable::LoadContainer(unsigned int ulContainerId, unsigned int ulType, list<localobjectdetails_t> **lppObjects)
{
	ECRESULT er = erSuccess;
	list<localobjectdetails_t> *lpObjects = NULL;
	list<localobjectdetails_t> *lpObjectsAlt = NULL;
	list<localobjectdetails_t> *lpALObjects = NULL;
	list<localobjectdetails_t> *lpObjectsCompany = NULL;
	list<localobjectdetails_t> *lpObjectsNonactive = NULL;
	list<localobjectdetails_t> *lpObjectsGroup = NULL;
	unsigned int ulCompanyId = 0;
	userobject_type_t ulObjType;
	char *szHideEveryone = NULL;
	bool bHideEveryone = false;

	if (ulContainerId == 0) {
		/*
		 * Zarafa Address Book
		 * The Zarafa Address Book only contains a single container,
		 * which is the Global Address Book.
		 */
		switch (ulType) {
		case MAPI_ABCONT:
			lpObjects = new std::list<localobjectdetails_t>();
			lpObjects->push_back(localobjectdetails_t(1, USEROBJECT_TYPE_COMPANY));
			break;
		default:
			er = ZARAFA_E_INVALID_PARAMETER;
			goto exit;
		}
	} else if (ulContainerId == 1) {
		/*
		 * Global Address Book
		 * The Global Address Book contains all visible (nonactive) users, groups and companies.
		 */
		switch (ulType) {
		case MAPI_ABCONT:
			/* Make companies visible through GetHierarchyTable().
			 *
			 * When in offline mode we only support the Global Address Book container,
			 * there are 2 reasons for this:
			 * 1) Currently offline mode does not assign users to companies, which means the lists are empty
			 *    except for the company itself and the System user.
			 * 2) Outlook generates the hierarchy view once during startup, when permissions change and
			 *    during sync one of the companies becomes unavailable, it will still be present
			 *    in the hierarchy view. The user will not be allowed to open it, so it isn't a real security risk,
			 *    but since we still have issue (1) open, we might as well disable the hierarchy view
			 *    containers completely. */
			er = GetHierarchyCompanies(&lpObjects, USERMANAGEMENT_IDS_ONLY);
			if (er != erSuccess)
				goto exit;

			/* Only regular users have a company and access to addresslists */
			if (lpSession->GetSecurity()->GetUserId() != ZARAFA_UID_SYSTEM) {
				er = lpSession->GetSecurity()->GetUserCompany(&ulCompanyId);
				if (er != erSuccess)
					goto exit;
				
    	        er = lpSession->GetUserManagement()->GetObjectListAndSync(USEROBJECT_TYPE_ADDRESSLIST, ulCompanyId,
    	                                                                  &lpALObjects, 0);
	            if (er != erSuccess)
            	    goto exit;

            	lpObjects->merge(*lpALObjects);
			}
			break;
		case MAPI_DISTLIST:
		case MAPI_MAILUSER:
		    // Load the contents table of the global address book: users, groups and viewable companies
			er = lpSession->GetSecurity()->GetUserCompany(&ulCompanyId);
			if (er != erSuccess)
				goto exit;

			// Hide group everyone when requested
			if (lpSession->GetSecurity()->GetAdminLevel() == 0) {
				szHideEveryone = lpSession->GetSessionManager()->GetConfig()->GetSetting("hide_everyone");
				if (szHideEveryone)
					bHideEveryone = parseBool(szHideEveryone);
			}

            // Get users and groups for our company
            er = LoadCompany(ulCompanyId, ulType, &lpObjects, false || bHideEveryone);
            if (er != erSuccess)
                goto exit;

			break;
		default:
			er = ZARAFA_E_INVALID_PARAMETER;
			goto exit;
		}
	} else {
	    // Other containers
	    
	    if(ulType == MAPI_MAILUSER || ulType == MAPI_DISTLIST) {
	        // Contents tables
	        
	        er = lpSession->GetUserManagement()->GetItemObjectType(ulContainerId, &ulObjType);
	        if(er != erSuccess)
	            goto exit;
	            
            if(ulObjType == USEROBJECT_TYPE_ADDRESSLIST) {
                // Address list
                er = lpSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(OBJECTRELATION_ADDRESSLIST_MEMBER, USEROBJECT_TYPE_USER,
                                                                                  ulContainerId, &lpObjects, USERMANAGEMENT_IDS_ONLY);
                if (er != erSuccess)
                    goto exit;

                er = lpSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(OBJECTRELATION_ADDRESSLIST_MEMBER, USEROBJECT_TYPE_NONACTIVE,
                                                                                  ulContainerId, &lpObjectsAlt, USERMANAGEMENT_IDS_ONLY);
                if (er != erSuccess)
                    goto exit;

                lpObjects->merge(*lpObjectsAlt);
            } else if(ulObjType == USEROBJECT_TYPE_COMPANY) {
                // Company: users, groups of that company
                er = LoadCompany(ulContainerId, ulType, &lpObjects, true);
                if (er != erSuccess)
                    goto exit;
                    
            } else {
                er = ZARAFA_E_INVALID_PARAMETER;
                goto exit;
            }

	    } else {
	        er = ZARAFA_E_INVALID_PARAMETER;
	        goto exit;
        }
	}

	*lppObjects = lpObjects;

exit:
    if (lpObjectsAlt)
        delete lpObjectsAlt;

    if (lpALObjects)
        delete lpALObjects;
        
	if (lpObjectsCompany)
		delete lpObjectsCompany;

	if (lpObjectsGroup)
		delete lpObjectsGroup;

	if (lpObjectsNonactive)
		delete lpObjectsNonactive;

	if (er != erSuccess && lpObjects)
		delete lpObjects;

	return er;
}

// This is just a helper function since the company contents are loaded in 3 different places
ECRESULT ECABObjectTable::LoadCompany(unsigned int ulCompanyId, unsigned int ulType, list<localobjectdetails_t> **lppObjects, bool bFilterSystem)
{
	ECRESULT er = erSuccess;
	list<localobjectdetails_t> *lpObjects = NULL;
	list<localobjectdetails_t> *lpObjectsNonactive = NULL;
	list<localobjectdetails_t> *lpObjectsGroup = NULL;
	list<localobjectdetails_t> *lpObjectsCompany = NULL;
	list<localobjectdetails_t>::iterator iterObject;

	switch (ulType) {
	case MAPI_MAILUSER:
		er = lpSession->GetUserManagement()->GetObjectListAndSync(USEROBJECT_TYPE_USER, ulCompanyId,
																  &lpObjects, USERMANAGEMENT_IDS_ONLY);
		if (er != erSuccess)
			goto exit;

		er = lpSession->GetUserManagement()->GetObjectListAndSync(USEROBJECT_TYPE_NONACTIVE, ulCompanyId,
																  &lpObjectsNonactive, USERMANAGEMENT_IDS_ONLY);
		if (er != erSuccess)
			goto exit;

		/* Add non-active to 'objects' list */
		lpObjects->merge(*lpObjectsNonactive);

		/* HACK: TableManager does handle multiple member types nicely, because of that we need to insert groups here as well */
		er = lpSession->GetUserManagement()->GetObjectListAndSync(USEROBJECT_TYPE_GROUP, ulCompanyId,
																  &lpObjectsGroup, USERMANAGEMENT_IDS_ONLY);
		if (er != erSuccess)
			goto exit;

		/* Add groups to 'objects' list */
		lpObjects->merge(*lpObjectsGroup);

		/* Remove special users Everyone and SYSTEM, these are only visible in GAB and not company container */
		if (bFilterSystem)
			lpObjects->remove_if(IsSystemObject);

		/*
		 * The company container always contains itself as distlist,
		 * unless ulCompanyId is 0 which is the default company and
		 * that one isn't visible.
		 */
		if (ulCompanyId != 0)
			lpObjects->push_front(localobjectdetails_t(ulCompanyId, USEROBJECT_TYPE_COMPANY));
		break;
	default:
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	*lppObjects = lpObjects;

exit:
	if (lpObjectsCompany)
		delete lpObjectsCompany;

	if (lpObjectsGroup)
		delete lpObjectsGroup;

	if (lpObjectsNonactive)
		delete lpObjectsNonactive;

	if (er != erSuccess && lpObjects)
		delete lpObjects;

	return er;
}


// Loads either contents table or hierarchy table for an IDistList
ECRESULT ECABObjectTable::LoadGroup(unsigned int ulGroupId, unsigned int ulType, list<localobjectdetails_t> **lppObjects)
{
	ECRESULT er = erSuccess;
	list<localobjectdetails_t> *lpObjects = NULL;
	list<localobjectdetails_t> *lpObjectsNonactive = NULL;
	userobject_type_t ulObjType;
	
	er = lpSession->GetUserManagement()->GetItemObjectType(ulGroupId, &ulObjType);
	if (er != erSuccess)
	    goto exit;
	
	switch (ulType) {
	case MAPI_MAILUSER:
	    if(ulObjType == USEROBJECT_TYPE_GROUP) {
            // Contents table
            er = lpSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(OBJECTRELATION_GROUP_MEMBER, USEROBJECT_TYPE_USER,
                                                                              ulGroupId, &lpObjects, USERMANAGEMENT_IDS_ONLY);
            if (er != erSuccess)
                goto exit;

            er = lpSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(OBJECTRELATION_GROUP_MEMBER, USEROBJECT_TYPE_NONACTIVE,
                                                                              ulGroupId, &lpObjectsNonactive, USERMANAGEMENT_IDS_ONLY);
            if (er != erSuccess)
                goto exit;

            /* Add non-active to 'objects' list */
            lpObjects->merge(*lpObjectsNonactive);

        } else if(ulObjType == USEROBJECT_TYPE_COMPANY) {
            /* The company group contains only users */
            er = lpSession->GetUserManagement()->GetObjectListAndSync(USEROBJECT_TYPE_USER, ulGroupId,
                                                                      &lpObjects, USERMANAGEMENT_IDS_ONLY);
            if (er != erSuccess)
                goto exit;

            er = lpSession->GetUserManagement()->GetObjectListAndSync(USEROBJECT_TYPE_NONACTIVE, ulGroupId,
                                                                      &lpObjectsNonactive, USERMANAGEMENT_IDS_ONLY);
            if (er != erSuccess)
                goto exit;

            /* Add non-active to 'objects' list */
            lpObjects->merge(*lpObjectsNonactive);

			lpObjects->remove_if(IsSystemObject);
		} else {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		break;
	default:
	    // Hierarchy table
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (ulGroupId == ZARAFA_UID_EVERYONE)
		lpObjects->remove_if(IsSystemObject);

	*lppObjects = lpObjects;

exit:
	if (lpObjectsNonactive)
		delete lpObjectsNonactive;

	if (er != erSuccess && lpObjects)
		delete lpObjects;

	return er;
}

ECRESULT ECABObjectTable::Load()
{
	ECRESULT er = erSuccess;
	ECODAB *lpODAB = (ECODAB*)m_lpObjectData;
	sObjectTableKey sRowItem;

	std::list<localobjectdetails_t> *lpObjects = NULL;
	std::list<localobjectdetails_t>::iterator iterObjects;
	std::list<unsigned int> lstObjects;

	// If the GAB is disabled, don't show any entries except the top-level object
	if(lpODAB->ulABParentId != 0 && parseBool(lpSession->GetSessionManager()->GetConfig()->GetSetting("enable_gab")) == false && lpSession->GetSecurity()->GetAdminLevel() == 0)
	    goto exit;

    er = lpSession->GetSecurity()->IsUserObjectVisible(lpODAB->ulABParentId);
    if (er != erSuccess)
        goto exit;

    // We could be loading tables for IABContainers or for IDistLists
	if (lpODAB->ulABParentType == MAPI_ABCONT) {
		er = LoadContainer(lpODAB->ulABParentId, lpODAB->ulABType, &lpObjects);
		if (er != erSuccess)
			goto exit;
	} else if (lpODAB->ulABParentType == MAPI_DISTLIST) {
        er = LoadGroup(lpODAB->ulABParentId, lpODAB->ulABType, &lpObjects);
        if (er != erSuccess)
            goto exit;
	} else {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	for (iterObjects = lpObjects->begin(); iterObjects != lpObjects->end(); iterObjects++) {
		/* Only add visible items */
		if (lpSession->GetSecurity()->IsUserObjectVisible(iterObjects->ulId) != erSuccess)
			continue;
			
        lstObjects.push_back(iterObjects->ulId);
    }
    
    UpdateRows(ECKeyTable::TABLE_ROW_ADD, &lstObjects, 0);

exit:
	if(lpObjects)
		delete lpObjects;

	return er;
}
