/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef EC_ATTACHMENT_STORAGE
#define EC_ATTACHMENT_STORAGE

#include "ECDatabase.h"
#include <list>
#include <string>

class ECAttachmentStorage {
public:
	ECAttachmentStorage(ECDatabase *lpDatabase, bool bFileStorage, unsigned int ulCompressionLevel);
	virtual ~ECAttachmentStorage();

	static ECRESULT CreateAttachmentStorage(ECDatabase *lpDatabase, ECConfig *lpConfig, ECAttachmentStorage **lppAttachmentStorage);

	/* Are attachments stored in the database or on disk. */
	virtual bool IsFileStorage();

	/* Single Instance Attachment wrappers (should not be overriden by subclasses) */
	virtual bool ExistAttachment(ULONG ulObjId);
	virtual ECRESULT LoadAttachment(struct soap *soap, ULONG ulObjId, int *lpiSize, unsigned char **lppData);
	virtual ECRESULT SaveAttachment(ULONG ulObjId, bool bDeleteOld, int iSize, unsigned char *lpData, ULONG *lpulInstanceId);
	virtual ECRESULT SaveAttachment(ULONG ulObjId, bool bDeleteOld, ULONG ulAttachId, ULONG *lpulInstanceId);
	virtual ECRESULT CopyAttachment(ULONG ulObjId, ULONG ulNewObjId);
	virtual ECRESULT DeleteAttachments(std::list<ULONG> lstDeleteObjects);
	virtual ECRESULT DeleteAttachment(ULONG ulObjId);
	virtual ECRESULT GetSize(ULONG ulObjId, ULONG *lpulSize);

	/* Convert ObjectId (hierarchyid) into Instance Id */
	virtual ECRESULT GetSingleInstanceId(ULONG ulObjId, ULONG *lpulAttachId, ULONG *lpulTag = NULL);
	virtual ECRESULT GetSingleInstanceIds(std::list<ULONG> &lstObjIds, std::list<ULONG> *lstAttachIds);

	/* Request parents for partitulca Single Instance Id */
	virtual ECRESULT GetSingleInstanceParents(ULONG ulAttachId, std::list<ULONG> *lplstObjIds);

	/* Single Instance Attachment handlers (must be overridden by subclasses) */
	virtual bool ExistAttachmentInstance(ULONG ulAttachId) = 0;

protected:
	/* Single Instance Attachment handlers (must be overridden by subclasses) */
	virtual ECRESULT LoadAttachmentInstance(struct soap *soap, ULONG ulAttachId, int *lpiSize, unsigned char **lppData) = 0;
	virtual ECRESULT SaveAttachmentInstance(ULONG ulAttachId, int iSize, unsigned char *lpData) = 0;
	virtual ECRESULT CopyAttachmentInstance(ULONG ulAttachId, ULONG ulNewAttachId) = 0;
	virtual ECRESULT DeleteAttachmentInstances(std::list<ULONG> &lstDeleteAttachments) = 0;
	virtual ECRESULT DeleteAttachmentInstance(ULONG ulAttachId) = 0;
	virtual ECRESULT GetSizeInstance(ULONG ulAttachId, ULONG *lpulSize, bool *lpbCompressed = NULL) = 0;

private:
	/* Add reference between Object and Single Instance */
	virtual ECRESULT SetSingleInstanceId(ULONG ulObjId, ULONG ulAttachId, ULONG ulTag);

	/* Count the number of times an attachment is referenced */
	virtual ECRESULT IsOrphanedSingleInstance(ULONG ulAttachId, bool *bOrphan);
	virtual ECRESULT GetOrphanedSingleInstances(std::list<ULONG> &lstAttachments, std::list<ULONG> *lplstOrphanedAttachments);

protected:
	ECDatabase *m_lpDatabase;
	bool m_bFileStorage;
	bool m_bFileCompression;
	std::string m_CompressionLevel;
};

class ECDatabaseAttachment : public ECAttachmentStorage {
public:
	ECDatabaseAttachment(ECDatabase *lpDatabase);
	virtual ~ECDatabaseAttachment();

protected:
	/* Single Instance Attachment handlers */
	virtual bool ExistAttachmentInstance(ULONG ulAttachId);
	virtual ECRESULT LoadAttachmentInstance(struct soap *soap, ULONG ulAttachId, int *lpiSize, unsigned char **lppData);
	virtual ECRESULT SaveAttachmentInstance(ULONG ulAttachId, int iSize, unsigned char *lpData);
	virtual ECRESULT CopyAttachmentInstance(ULONG ulAttachId, ULONG ulNewAttachId);
	virtual ECRESULT DeleteAttachmentInstances(std::list<ULONG> &lstDeleteAttachments);
	virtual ECRESULT DeleteAttachmentInstance(ULONG ulAttachId);
	virtual ECRESULT GetSizeInstance(ULONG ulAttachId, ULONG *lpulSize, bool *lpbCompressed = NULL);
};

class ECFileAttachment : public ECAttachmentStorage {
public:
	ECFileAttachment(ECDatabase *lpDatabase, std::string basepath, unsigned int ulCompressionLevel);
	virtual ~ECFileAttachment();

protected:
	/* Single Instance Attachment handlers */
	virtual bool ExistAttachmentInstance(ULONG ulAttachId);
	virtual ECRESULT LoadAttachmentInstance(struct soap *soap, ULONG ulAttachId, int *lpiSize, unsigned char **lppData);
	virtual ECRESULT SaveAttachmentInstance(ULONG ulAttachId, int iSize, unsigned char *lpData);
	virtual ECRESULT CopyAttachmentInstance(ULONG ulAttachId, ULONG ulNewAttachId);
	virtual ECRESULT DeleteAttachmentInstances(std::list<ULONG> &lstDeleteAttachments);
	virtual ECRESULT DeleteAttachmentInstance(ULONG ulAttachId);
	virtual ECRESULT GetSizeInstance(ULONG ulAttachId, ULONG *lpulSize, bool *lpbCompressed = NULL);

private:
	std::string CreateAttachmentFilename(ULONG ulAttachId, bool bCompressed);

	std::string m_basepath;
};

#endif
