/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <mapitags.h>
#include <mapidefs.h>
#include <mapiutil.h>

#include <libintl.h>

#include "ECMAPI.h"
#include "stringutil.h"
#include "SOAPUtils.h"
#include "soapH.h"
#include "ECGenProps.h"
#include "Zarafa.h"
#include "ECDefs.h"
#include "ECUserManagement.h"
#include "ECSecurity.h"
#include "ECSessionManager.h"

#include <edkmdb.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define _(string) dcgettext("zarafa", string, LC_MESSAGES)

ECGenProps::ECGenProps()
{
	// Nothing to do
}

ECGenProps::~ECGenProps()
{
	// Nothing to do
}

ECRESULT ECGenProps::GetMVPropSubquery(unsigned int ulPropTagRequested, std::string &subquery) 
{
	ECRESULT er = erSuccess;
	unsigned int ulType = PROP_TYPE(ulPropTagRequested);

	//skip MV_INSTANCE
	switch(ulType) {
		case PT_MV_I2:
		case PT_MV_LONG:
			subquery = "SELECT concat(count(*), ':', group_concat(length(val_ulong),':',val_ulong ORDER BY subquery.orderid SEPARATOR '')) FROM mvproperties AS subquery WHERE subquery.hierarchyid=hierarchy.id AND subquery.type="+stringify(PROP_TYPE(ulPropTagRequested))+" AND subquery.tag="+stringify(PROP_ID(ulPropTagRequested))+" GROUP BY subquery.hierarchyid";
			break;
		case PT_MV_R4:
		case PT_MV_DOUBLE:
		case PT_MV_APPTIME:
			subquery = "SELECT concat(count(*), ':', group_concat(length(val_double),':',val_double ORDER BY subquery.orderid SEPARATOR '')) FROM mvproperties AS subquery WHERE subquery.hierarchyid=hierarchy.id AND subquery.type="+stringify(PROP_TYPE(ulPropTagRequested))+" AND subquery.tag="+stringify(PROP_ID(ulPropTagRequested))+" GROUP BY subquery.hierarchyid";
			break;
		case PT_MV_CURRENCY:
		case PT_MV_SYSTIME:
			subquery = "SELECT group_concat(count(*),':',length(val_hi),':',val_hi ORDER BY subquery.orderid SEPARATOR '')) FROM mvproperties AS subquery WHERE subquery.hierarchyid=hierarchy.id AND subquery.type="+stringify(PROP_TYPE(ulPropTagRequested))+" AND subquery.tag="+stringify(PROP_ID(ulPropTagRequested))+" GROUP BY subquery.hierarchyid";
			subquery += "),(SELECT group_concat(count(*),':',length(val_lo),':',val_lo ORDER BY subquery.orderid SEPARATOR '')) FROM mvproperties AS subquery WHERE subquery.hierarchyid=hierarchy.id AND subquery.type="+stringify(PROP_TYPE(ulPropTagRequested))+" AND subquery.tag="+stringify(PROP_ID(ulPropTagRequested))+" GROUP BY subquery.hierarchyid";
			break;
		case PT_MV_BINARY:
		case PT_MV_CLSID:
			subquery = "SELECT concat(count(*), ':', group_concat(length(val_binary),':',val_binary ORDER BY subquery.orderid SEPARATOR '')) FROM mvproperties AS subquery WHERE subquery.hierarchyid=hierarchy.id AND subquery.type="+stringify(PROP_TYPE(ulPropTagRequested))+" AND subquery.tag="+stringify(PROP_ID(ulPropTagRequested))+" GROUP BY subquery.hierarchyid";
			break;
		case PT_MV_STRING8:
		case PT_MV_UNICODE:
			subquery = "SELECT concat(count(*), ':', group_concat(char_length(val_string),':',val_string ORDER BY subquery.orderid SEPARATOR '')) FROM mvproperties AS subquery WHERE subquery.hierarchyid=hierarchy.id AND subquery.type="+stringify(PROP_TYPE(ulPropTagRequested))+" AND subquery.tag="+stringify(PROP_ID(ulPropTagRequested))+" GROUP BY subquery.hierarchyid";
			break;
		case PT_MV_I8:
			subquery = "SELECT concat(count(*), ':', group_concat(length(val_longint),':',val_longint ORDER BY subquery.orderid SEPARATOR '')) FROM mvproperties AS subquery WHERE subquery.hierarchyid=hierarchy.id AND subquery.type="+stringify(PROP_TYPE(ulPropTagRequested))+" AND subquery.tag="+stringify(PROP_ID(ulPropTagRequested))+" GROUP BY subquery.hierarchyid";
			break;
		default:
			er = ZARAFA_E_NOT_FOUND;
			break;
	}
	

	return er;
}

ECRESULT ECGenProps::GetPropSubquery(unsigned int ulPropTagRequested, std::string &subquery) 
{
	ECRESULT er = erSuccess;

	switch(ulPropTagRequested) {
	case PR_CONTENT_COUNT:
		// MSGFLAG_READ = 0x01, MSGFLAG_ASSOCIATED = 0x40, MSGFLAG_DELETED = 0x400
		subquery = "SELECT if(hierarchy.flags = 2, "
					"(SELECT count(*) FROM searchresults WHERE searchresults.folderid=hierarchy.id), "
					"(SELECT count(*) FROM hierarchy AS subquery WHERE subquery.parent=hierarchy.id AND subquery.flags & " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED) + " = 0 AND subquery.type=5))";
		break;
		
	case PR_CONTENT_UNREAD:
		subquery = "SELECT if(hierarchy.flags = 2, "
					"(SELECT count(*) FROM searchresults WHERE searchresults.flags & " + stringify(MSGFLAG_READ) + " = 0 AND searchresults.folderid=hierarchy.id), " 
					"(SELECT count(*) FROM hierarchy AS subquery WHERE subquery.parent=hierarchy.id AND subquery.flags & " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED|MSGFLAG_READ) + " = 0 AND subquery.type=5))";
		break;

	case PR_ASSOC_CONTENT_COUNT:
		subquery = "SELECT count(*) FROM hierarchy AS subquery WHERE subquery.parent=hierarchy.id AND subquery.flags & " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED) + " = " + stringify(MSGFLAG_ASSOCIATED) + " AND subquery.type=5";
		break;

	case PR_DELETED_MSG_COUNT:
		subquery = "SELECT count(*) FROM hierarchy AS subquery WHERE subquery.parent=hierarchy.id AND subquery.flags & " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED) + " = " + stringify(MSGFLAG_DELETED) + " AND subquery.type=5";
		break;

    case PR_DELETED_FOLDER_COUNT:
		subquery = "SELECT count(*) FROM hierarchy AS subquery WHERE subquery.parent=hierarchy.id AND subquery.flags & " + stringify(MSGFLAG_DELETED) + " = " + stringify(MSGFLAG_DELETED) + " AND subquery.type=3";
		break;

	case PR_DELETED_ASSOC_MSG_COUNT:
		subquery = "SELECT count(*) FROM hierarchy AS subquery WHERE subquery.parent=hierarchy.id AND subquery.flags & " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED) + " = " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED) + " AND subquery.type=5";
		break;

	case PR_SUBFOLDERS:
		subquery = "SELECT subquery.id FROM hierarchy as subquery WHERE subquery.parent=hierarchy.id AND subquery.flags & " + stringify(MSGFLAG_DELETED) + " = 0 AND subquery.type=3 LIMIT 1";
		break;

	case PR_HASATTACH:
		subquery = "SELECT subquery.id FROM hierarchy as subquery WHERE subquery.parent=hierarchy.id AND subquery.type=7 LIMIT 1";
		break;
		
	case PR_MESSAGE_FLAGS:
		// get flags plus 0x10 for MSGFLAG_HASATTACH
		subquery = "SELECT hierarchy.flags + if((SELECT subquery.id FROM hierarchy as subquery WHERE subquery.parent=hierarchy.id AND subquery.type=7 LIMIT 1), 16, 0)";
		break;
//FIXME: Slow query missing storeid
	case PR_PARENT_DISPLAY:
		subquery = "SELECT properties.val_string FROM properties JOIN hierarchy as subquery ON properties.hierarchyid=subquery.parent WHERE subquery.id=hierarchy.id AND properties.tag=0x3001"; // PR_DISPLAY_NAME of parent
		break;

    case PR_ATTACH_DATA_BIN:
		subquery = "SELECT singleinstances.hierarchyid FROM `singleinstances` WHERE singleinstances.hierarchyid=hierarchy.id LIMIT 1";
        break;

    case PR_LAST_MODIFICATION_TIME:
        subquery = "UNIX_TIMESTAMP(hierarchy.modtime)";
        break;
        
    case PR_CREATION_TIME:
        subquery = "UNIX_TIMESTAMP(hierarchy.createtime)";
        break;

    case PR_EC_OUTGOING_FLAGS:
        subquery = "SELECT outgoingqueue.flags FROM outgoingqueue where outgoingqueue.hierarchy_id = hierarchy.id and outgoingqueue.flags & 1 = 1";
        break;

	case PR_FOLDER_TYPE:
		subquery = "IF((hierarchy.type=3), hierarchy.flags&"+stringify(FOLDER_GENERIC|FOLDER_ROOT|FOLDER_SEARCH)+", NULL)";
		break;

	default:
		er = ZARAFA_E_NOT_FOUND;
		break;
	}

	return er;
}

//WARNING: the type of the properies must be the same!
ECRESULT ECGenProps::GetPropSubstitute(unsigned int ulPropTagRequested, unsigned int *lpulPropTagRequired)
{
	ECRESULT er = erSuccess;

	unsigned int ulPropTagRequired = 0;

	switch(ulPropTagRequested) {
		case PR_NORMALIZED_SUBJECT:
			ulPropTagRequired = PR_SUBJECT;
			break;
		default:
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
	}

	*lpulPropTagRequired = ulPropTagRequired;

exit:
	return er;
}

// This should be synchronized with GetPropComputed
ECRESULT ECGenProps::IsPropComputed(unsigned int ulPropTag, unsigned int ulObjType)
{
	ECRESULT		er = erSuccess;

	switch(ulPropTag) {
		case PR_LAST_MODIFICATION_TIME:
		case PR_CREATION_TIME:
		case PR_SUBFOLDERS:
		case PR_HASATTACH:
		case PR_ATTACH_DATA_BIN:
		case PR_MSG_STATUS:
		case PR_NORMALIZED_SUBJECT:
		case PR_SOURCE_KEY:
		case PR_PARENT_SOURCE_KEY:
		case PR_EC_IMAP_ID:
		    er = erSuccess;
		    break;
		default:
			er = ZARAFA_E_NOT_FOUND;
			break;
	}

	return er;
}

// This should be synchronized with GetPropComputedUncached
ECRESULT ECGenProps::IsPropComputedUncached(unsigned int ulPropTag, unsigned int ulObjType) {
    ECRESULT er = erSuccess;
    
    switch(ulPropTag) {
        case PR_LONGTERM_ENTRYID_FROM_TABLE:
		case PR_ENTRYID:
		case PR_PARENT_ENTRYID: 
		case PR_STORE_ENTRYID:
		case PR_USER_NAME:
		case PR_MAILBOX_OWNER_NAME:
		case PR_USER_ENTRYID:
		case PR_MAILBOX_OWNER_ENTRYID:
		case PR_EC_MAILBOX_OWNER_ACCOUNT:
		case PR_EC_HIERARCHYID:
		case PR_INSTANCE_KEY:
		case PR_RECORD_KEY:
		case PR_OBJECT_TYPE:
		case PR_CONTENT_COUNT:
		case PR_CONTENT_UNREAD:
		case PR_RIGHTS:
		case PR_ACCESS_LEVEL:
		case PR_ACCESS:
		case PR_ROW_TYPE:
		case PR_MAPPING_SIGNATURE:
		    er = erSuccess;
		    break;
		case PR_DISPLAY_NAME: // only the store property is generated
		    if(ulObjType == MAPI_STORE)
    			er = erSuccess;
            else
                er = ZARAFA_E_NOT_FOUND;
			break;
        default:
            er = ZARAFA_E_NOT_FOUND;
            break;
    }
    
    return er;
}

// These are properties that are never written to the 'properties' table; ie they are never directly queried. This
// is not the same as the generated properties, as they may access data in the database to *create* a generated
// property. 
ECRESULT ECGenProps::IsPropRedundant(unsigned int ulPropTag, unsigned int ulObjType) {
    ECRESULT er = erSuccess;
    
    switch(ulPropTag) {
		case PR_CREATION_TIME:			// is in hierarchy table
		case PR_ACCESS:					// generated from ACLs
		case PR_USER_NAME:				// generated from owner (hierarchy)
		case PR_MAILBOX_OWNER_NAME:		// generated from owner (hierarchy)
		case PR_USER_ENTRYID:			// generated from owner (hierarchy)
		case PR_MAILBOX_OWNER_ENTRYID:	// generated from owner (hierarchy)
		case PR_EC_MAILBOX_OWNER_ACCOUNT: // generated from owner (hierarchy)
		case PR_EC_HIERARCHYID:			// generated from hierarchy
		case PR_SUBFOLDERS:				// generated from hierarchy
		case PR_HASATTACH:				// generated from hierarchy
		case PR_LONGTERM_ENTRYID_FROM_TABLE: // generated from hierarchy
		case PR_ENTRYID:				// generated from hierarchy
		case PR_PARENT_ENTRYID: 		// generated from hierarchy
		case PR_STORE_ENTRYID:			// generated from store id
		case PR_INSTANCE_KEY:			// table data only
		case PR_RECORD_KEY:				// generated from hierarchy
		case PR_OBJECT_TYPE:			// generated from hierarchy
		case PR_CONTENT_COUNT:			// generated from hierarchy
		case PR_CONTENT_UNREAD:			// generated from hierarchy
		case PR_RIGHTS:					// generated from security system
		case PR_ACCESS_LEVEL:			// generated from security system
		case PR_PARENT_SOURCE_KEY:		// generated from ics system
		case PR_FOLDER_TYPE:			// generated from hierarchy (CreateFolder)
		case PR_EC_IMAP_ID:				// generated for each new mail and updated on move by the server
		    er = erSuccess;
		    break;
		default:
			er = ZARAFA_E_NOT_FOUND;
			break;
    }
    
    return er;
}

// NOTE: This currently only works for single-column values (!)
ECRESULT ECGenProps::GetPropComputed(struct soap *soap, ECSession* lpSession, unsigned int ulPropTagRequested, unsigned int ulObjId,unsigned int ulOrderId, unsigned int ulParentId, unsigned int ulObjType, unsigned int ulFlags, unsigned int ulStoreId, char *szColData, struct propVal *lpPropVal)
{
	ECRESULT		er = erSuccess;
	char*			lpszColon = NULL;
	unsigned int	ulParent = 0;

	switch(ulPropTagRequested) {
	case PR_SUBFOLDERS: 
		lpPropVal->ulPropTag = PR_SUBFOLDERS;
		lpPropVal->__union = SOAP_UNION_propValData_b;

		if(szColData == NULL) {
			lpPropVal->Value.b = FALSE;
		} else {
			lpPropVal->Value.b = TRUE;
		}
		break;

	case PR_HASATTACH: 
		lpPropVal->ulPropTag = PR_HASATTACH;
		lpPropVal->__union = SOAP_UNION_propValData_b;

		if(szColData == NULL) {
			lpPropVal->Value.b = FALSE;
		} else {
			lpPropVal->Value.b = TRUE;
		}
		break;
	case PR_ATTACH_DATA_BIN:
		// You can never actually see the attachment data in tables. It is loaded afterwards
		// by ns__readChunk when required by the client.
		if(szColData == NULL) {
		    // No attachment data found in the LOB table
		    er = ZARAFA_E_NOT_FOUND;
        } else {
            // Give ZARAFA_E_NOT_ENOUGH_MEMORY to show that there is data, but we're not sending it (yet)
    		lpPropVal->Value.ul = ZARAFA_E_NOT_ENOUGH_MEMORY;
	    	lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(PR_ATTACH_DATA_BIN));
	    	lpPropVal->__union = SOAP_UNION_propValData_ul;
        }
		break;
	case PR_NORMALIZED_SUBJECT:
		
		if(szColData == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			break;
		}

		lpszColon = strchr(szColData, ':');
		if(lpszColon != NULL && (lpszColon - szColData) >1 && (lpszColon - szColData) < 4)
		{
			lpszColon++; // skip ':'
			if(strlen(lpszColon)+1 <= strlen(szColData) && lpszColon[0] == ' ')
				lpszColon++;// skip space

			lpPropVal->Value.lpszA = s_alloc<char>(soap, strlen(lpszColon)+1);

			strcpy(lpPropVal->Value.lpszA, lpszColon);
			lpPropVal->ulPropTag = PR_NORMALIZED_SUBJECT;
			lpPropVal->__union = SOAP_UNION_propValData_lpszA;
		}else{
			er = ZARAFA_E_NOT_FOUND;
			break;
		}
		break;
	case PR_MSG_STATUS:
		lpPropVal->ulPropTag = PR_MSG_STATUS;
		lpPropVal->__union = SOAP_UNION_propValData_ul;

		if(szColData == NULL)
			lpPropVal->Value.ul = 0;
		else
			lpPropVal->Value.ul = atoui(szColData);
		break;
	case PR_LAST_MODIFICATION_TIME:
	case PR_CREATION_TIME:
		if(szColData == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			break;
		}
		lpPropVal->ulPropTag = ulPropTagRequested;
		// Special handling as the data is in unix-timestamp format
		lpPropVal->Value.hilo = s_alloc<hiloLong>(soap);
 		lpPropVal->__union = SOAP_UNION_propValData_hilo;
		UnixTimeToFileTime(atoi(szColData), &lpPropVal->Value.hilo->hi, &lpPropVal->Value.hilo->lo);
		break;
	case PR_SOURCE_KEY:
		lpPropVal->ulPropTag = PR_SOURCE_KEY;
		lpPropVal->__union = SOAP_UNION_propValData_bin;
		lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);

		er = lpSession->GetSessionManager()->GetCacheManager()->GetPropFromObject(PROP_ID(PR_SOURCE_KEY), ulObjId, soap, (unsigned int*)&lpPropVal->Value.bin->__size, &lpPropVal->Value.bin->__ptr);
		break;
	case PR_PARENT_SOURCE_KEY:
		lpPropVal->ulPropTag = PR_PARENT_SOURCE_KEY;
		lpPropVal->__union = SOAP_UNION_propValData_bin;
		lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);

		er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, &ulParent, NULL, NULL, NULL);
		if(er != erSuccess)
			break;

		er = lpSession->GetSessionManager()->GetCacheManager()->GetPropFromObject(PROP_ID(PR_SOURCE_KEY), ulParent, soap, (unsigned int*)&lpPropVal->Value.bin->__size, &lpPropVal->Value.bin->__ptr);
		break;
    case PR_EC_IMAP_ID:
		lpPropVal->ulPropTag = PR_EC_IMAP_ID;
		lpPropVal->__union = SOAP_UNION_propValData_ul;

		if(szColData == NULL)
			lpPropVal->Value.ul = ulObjId;
		else
			lpPropVal->Value.ul = atoui(szColData);
		break;
	default:
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

exit:

	return er;
}

// All in memory properties
ECRESULT ECGenProps::GetPropComputedUncached(struct soap *soap, ECSession* lpSession, unsigned int ulPropTag, unsigned int ulObjId, unsigned int ulOrderId, unsigned int ulStoreId, struct propVal *lpPropVal)
{
	ECRESULT		er = erSuccess;
	unsigned int	ulRights = 0;
	bool			bOwner = false;
	bool			bAdmin = false;
	unsigned int	ulObjType = 0;
	unsigned int	ulFlags = 0;
	unsigned int	ulUserId = 0;
	unsigned int	ulParentId = 0;
	char*			lpStoreName = NULL;

	struct propValArray sPropValArray = {0, 0};
	struct propTagArray sPropTagArray = {0, 0};

	switch(ulPropTag) {
		case PR_LONGTERM_ENTRYID_FROM_TABLE:
		case PR_ENTRYID:
		case PR_PARENT_ENTRYID: 
		case PR_STORE_ENTRYID:
		{
			entryId sEntryId;

			if (ulPropTag == PR_PARENT_ENTRYID) {
				er = lpSession->GetSessionManager()->GetCacheManager()->GetParent(ulObjId, &ulParentId);
				if(er != erSuccess)
					goto exit;

                er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(ulParentId, NULL, NULL, NULL, &ulObjType);
                if(er != erSuccess)
                    goto exit;
                    
                if(ulObjType == MAPI_FOLDER) {
                    ulObjId = ulParentId;
                } // else PR_PARENT_ENTRYID == PR_ENTRYID
					
			}else if (ulPropTag == PR_STORE_ENTRYID) {
				ulObjId = ulStoreId;
			}

			er = lpSession->GetSessionManager()->GetCacheManager()->GetEntryIdFromObject(ulObjId, soap, &sEntryId);
			if(er != erSuccess) {
				// happens on recipients, attachments and msg-in-msg .. TODO: add strict type checking?
				//ASSERT(FALSE);
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}
			lpPropVal->ulPropTag = ulPropTag;

			lpPropVal->__union = SOAP_UNION_propValData_bin;
			lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);

			lpPropVal->Value.bin->__ptr = sEntryId.__ptr;
			lpPropVal->Value.bin->__size = sEntryId.__size;

			break;
		}
		case PR_USER_ENTRYID:
			sPropTagArray.__ptr = new unsigned int[1];
			sPropTagArray.__ptr[0] = PR_ENTRYID;
			sPropTagArray.__size = 1;

			ulUserId = lpSession->GetSecurity()->GetUserId();
			if(	lpSession->GetUserManagement()->GetProps(soap, ulUserId, &sPropTagArray, &sPropValArray) == erSuccess &&
				sPropValArray.__ptr && sPropValArray.__ptr[0].ulPropTag == PR_ENTRYID)
			{
				lpPropVal->__union = sPropValArray.__ptr[0].__union;
				lpPropVal->ulPropTag = PR_USER_ENTRYID;
				lpPropVal->Value.bin = sPropValArray.__ptr[0].Value.bin; // memory is allocated in GetUserData(..)
			}else{
				er = ZARAFA_E_NOT_FOUND;
				break;
			}
			break;
		case PR_USER_NAME:
			sPropTagArray.__ptr = new unsigned int[1];
			sPropTagArray.__ptr[0] = PR_DISPLAY_NAME;
			sPropTagArray.__size = 1;

			ulUserId = lpSession->GetSecurity()->GetUserId();
			if(	lpSession->GetUserManagement()->GetProps(soap, ulUserId, &sPropTagArray, &sPropValArray) == erSuccess &&
				sPropValArray.__ptr && sPropValArray.__ptr[0].ulPropTag == PR_DISPLAY_NAME)
			{
				lpPropVal->__union = sPropValArray.__ptr[0].__union;
				lpPropVal->ulPropTag = PR_USER_NAME;
				lpPropVal->Value.lpszA = sPropValArray.__ptr[0].Value.lpszA;// memory is allocated in GetUserData(..)
			}else{
				er = ZARAFA_E_NOT_FOUND;
				break;
			}
			break;
		case PR_DISPLAY_NAME:
			er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, NULL, NULL, NULL, &ulObjType);
			if (er != erSuccess) {
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}

			if(ulObjType != MAPI_STORE) {
			    er = ZARAFA_E_NOT_FOUND;
			    goto exit;
	        }
        
			er = GetStoreName(soap, lpSession, ulStoreId, ulObjId, &lpStoreName);
			if(er != erSuccess)
				break;
		
			lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			lpPropVal->ulPropTag = PR_DISPLAY_NAME;
			lpPropVal->Value.lpszA = lpStoreName;

		break;
		case PR_MAILBOX_OWNER_NAME:
			sPropTagArray.__ptr = new unsigned int[1];
			sPropTagArray.__ptr[0] = PR_DISPLAY_NAME;
			sPropTagArray.__size = 1;

			if(lpSession->GetSecurity()->GetStoreOwner(ulStoreId, &ulUserId) == erSuccess &&
				lpSession->GetUserManagement()->GetProps(soap, ulUserId, &sPropTagArray, &sPropValArray) == erSuccess &&
				sPropValArray.__ptr && sPropValArray.__ptr[0].ulPropTag == PR_DISPLAY_NAME)
			{
				lpPropVal->__union = sPropValArray.__ptr[0].__union;
				lpPropVal->ulPropTag = PR_MAILBOX_OWNER_NAME;
				lpPropVal->Value.lpszA = sPropValArray.__ptr[0].Value.lpszA; // memory is allocated in GetUserData(..)
			}else{
				er = ZARAFA_E_NOT_FOUND;
				break;
			}
		break;
		case PR_MAILBOX_OWNER_ENTRYID:
			sPropTagArray.__ptr = new unsigned int[1];
			sPropTagArray.__ptr[0] = PR_ENTRYID;
			sPropTagArray.__size = 1;

			if(lpSession->GetSecurity()->GetStoreOwner(ulStoreId, &ulUserId) == erSuccess &&
				lpSession->GetUserManagement()->GetProps(soap, ulUserId, &sPropTagArray, &sPropValArray) == erSuccess &&
				sPropValArray.__ptr && sPropValArray.__ptr[0].ulPropTag == PR_ENTRYID)
			{
				lpPropVal->__union = sPropValArray.__ptr[0].__union;
				lpPropVal->ulPropTag = PR_MAILBOX_OWNER_ENTRYID;
				lpPropVal->Value.bin = sPropValArray.__ptr[0].Value.bin;// memory is allocated in GetUserData(..)
			}else{
				er = ZARAFA_E_NOT_FOUND;
				break;
			}
			break;
		case PR_EC_MAILBOX_OWNER_ACCOUNT:
			sPropTagArray.__ptr = new unsigned int[1];
			sPropTagArray.__ptr[0] = PR_ACCOUNT;
			sPropTagArray.__size = 1;

			if (lpSession->GetSecurity()->GetStoreOwner(ulStoreId, &ulUserId) == erSuccess &&
				lpSession->GetUserManagement()->GetProps(soap, ulUserId, &sPropTagArray, &sPropValArray) == erSuccess &&
				sPropValArray.__ptr && sPropValArray.__ptr[0].ulPropTag == PR_ACCOUNT) {

				lpPropVal->__union = sPropValArray.__ptr[0].__union;
				lpPropVal->ulPropTag = PR_EC_MAILBOX_OWNER_ACCOUNT;
				lpPropVal->Value.lpszA = sPropValArray.__ptr[0].Value.lpszA; // memory is allocated in GetUserData(..)
			} else {
				er = ZARAFA_E_NOT_FOUND;
				break;
			}
			break;
        case PR_EC_HIERARCHYID:
			lpPropVal->ulPropTag = ulPropTag;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			
			lpPropVal->Value.ul = ulObjId;
            break;
		case PR_INSTANCE_KEY:
			lpPropVal->ulPropTag = ulPropTag;
			lpPropVal->__union = SOAP_UNION_propValData_bin;
			
			lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
			lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ULONG)*2);
			
			lpPropVal->Value.bin->__size = sizeof(ULONG)*2;
			memcpy(lpPropVal->Value.bin->__ptr, &ulObjId, sizeof(ULONG));
			memcpy(lpPropVal->Value.bin->__ptr+sizeof(ULONG), &ulOrderId, sizeof(ULONG));
			break;
		case PR_RECORD_KEY:
			lpPropVal->ulPropTag = ulPropTag;
			lpPropVal->__union = SOAP_UNION_propValData_bin;
			
			lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
			lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ULONG));
			
			lpPropVal->Value.bin->__size = sizeof(ULONG);
			memcpy(lpPropVal->Value.bin->__ptr, &ulObjId, sizeof(ULONG));
			break;
		case PR_OBJECT_TYPE:
			er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, NULL, NULL, NULL, &ulObjType);
			if(er != erSuccess) {
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}
			lpPropVal->ulPropTag = PR_OBJECT_TYPE;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			lpPropVal->Value.ul = ulObjType;
			break;
		case PR_CONTENT_COUNT:
		case PR_CONTENT_UNREAD:
			// create here the properties on a message, the folder properties are created with function GetPropSubquery(..)
			er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, NULL, NULL, &ulFlags, &ulObjType);
			if(er != erSuccess || ulObjType != MAPI_MESSAGE) {
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}
			
			lpPropVal->ulPropTag = ulPropTag;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			
			if(ulPropTag == PR_CONTENT_COUNT)
				lpPropVal->Value.ul = 1;
			else
				lpPropVal->Value.ul = (ulFlags&MSGFLAG_READ)?0:1;

			break;
		case PR_RIGHTS:
			er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, NULL, NULL, NULL, &ulObjType);
			if(er != erSuccess || ulObjType != MAPI_FOLDER)
			{
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}

			lpPropVal->ulPropTag = PR_RIGHTS;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
		
			if(lpSession->GetSecurity()->IsStoreOwner(ulObjId) == erSuccess || lpSession->GetSecurity()->IsObjectAdmin(ulObjId) == erSuccess)
				lpPropVal->Value.ul = ecRightsAll;
			else if(lpSession->GetSecurity()->GetObjectPermission(ulObjId, &ulRights) == hrSuccess)
				lpPropVal->Value.ul = ulRights;
			else
				lpPropVal->Value.ul = 0;

			break;
		case PR_ACCESS:
			if(ulObjType == MAPI_STORE || ulObjType == MAPI_ATTACH)
			{
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}

			er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, NULL, NULL, NULL, &ulObjType);
			if(er != erSuccess)
			{
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}

			lpPropVal->ulPropTag = PR_ACCESS;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			lpPropVal->Value.ul = 0;

			// if the requested object is from the owners store, return all permissions	
			if (lpSession->GetSecurity()->IsStoreOwner(ulObjId) == erSuccess ||
				lpSession->GetSecurity()->IsObjectAdmin(ulObjId) == erSuccess) {
				switch(ulObjType) {
					case MAPI_FOLDER:
						lpPropVal->Value.ul = MAPI_ACCESS_READ | MAPI_ACCESS_MODIFY | MAPI_ACCESS_DELETE;
						if(ulFlags != FOLDER_SEARCH) //FOLDER_GENERIC, FOLDER_ROOT 
							lpPropVal->Value.ul |= MAPI_ACCESS_CREATE_HIERARCHY | MAPI_ACCESS_CREATE_CONTENTS | MAPI_ACCESS_CREATE_ASSOCIATED;
						break;
					case MAPI_MESSAGE:
						lpPropVal->Value.ul = MAPI_ACCESS_READ | MAPI_ACCESS_MODIFY | MAPI_ACCESS_DELETE;
						break;
					case MAPI_ATTACH:
					case MAPI_STORE:
					default:
						er = ZARAFA_E_NOT_FOUND;
						break;
				}

				break;
			}

			ulRights = 0;

			lpSession->GetSecurity()->GetObjectPermission(ulObjId, &ulRights);// skip error checking, ulRights = 0

			if(lpSession->GetSecurity()->IsOwner(ulObjId) == erSuccess)
				bOwner = true;
			else
				bOwner = false;

			switch(ulObjType) {
				case MAPI_FOLDER:
					if( (ulRights&ecRightsReadAny)==ecRightsReadAny)
						lpPropVal->Value.ul |= MAPI_ACCESS_READ;

					if( bOwner == true || (ulRights&ecRightsFolderAccess) == ecRightsFolderAccess)
						lpPropVal->Value.ul |= MAPI_ACCESS_DELETE | MAPI_ACCESS_MODIFY;
					
					if(ulFlags != FOLDER_SEARCH) //FOLDER_GENERIC, FOLDER_ROOT 
					{
						if( (ulRights&ecRightsCreateSubfolder) == ecRightsCreateSubfolder)
							lpPropVal->Value.ul |= MAPI_ACCESS_CREATE_HIERARCHY;
							
						if( (ulRights&ecRightsCreate) == ecRightsCreate )
							lpPropVal->Value.ul |= MAPI_ACCESS_CREATE_CONTENTS;	
							
						if( (ulRights&ecRightsFolderAccess) == ecRightsFolderAccess)
							lpPropVal->Value.ul |= MAPI_ACCESS_CREATE_ASSOCIATED;
					}

					break;
				case MAPI_MESSAGE:
					if( (ulRights&ecRightsReadAny)==ecRightsReadAny)
						lpPropVal->Value.ul |= MAPI_ACCESS_READ;

					if( (ulRights&ecRightsEditAny)==ecRightsEditAny || (bOwner == true && (ulRights&ecRightsEditOwned) == ecRightsEditOwned) )
						lpPropVal->Value.ul |= MAPI_ACCESS_MODIFY;

					if( (ulRights&ecRightsDeleteAny) == ecRightsDeleteAny || (bOwner == true && (ulRights&ecRightsDeleteOwned) == ecRightsDeleteOwned) )
						lpPropVal->Value.ul |= MAPI_ACCESS_DELETE;

					break;
				case MAPI_ATTACH:
				case MAPI_STORE:
					er = ZARAFA_E_NOT_FOUND;
					break;
				default:
					er = ZARAFA_E_NOT_FOUND;
					break;
			}
			break;
		case PR_ACCESS_LEVEL:
			{
				lpPropVal->ulPropTag = PR_ACCESS_LEVEL;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				lpPropVal->Value.ul = 0;

				ulRights = 0;

				//FIXME: if store only open with read rights,
				//access level = 0
				if(lpSession->GetSecurity()->IsObjectAdmin(ulObjId) == erSuccess)
					bAdmin = true; // Admin of all stores
				else if(lpSession->GetSecurity()->IsStoreOwner(ulObjId) == erSuccess)
					bAdmin = true; // Admin of your one store
				else {
					lpSession->GetSecurity()->GetObjectPermission(ulObjId, &ulRights); // skip error checking, ulRights = 0

					if(lpSession->GetSecurity()->IsOwner(ulObjId) == erSuccess)
						bOwner = true;
					else
						bOwner = false;
				}

				if(bAdmin == true || ulRights&ecRightsCreate || ulRights&ecRightsEditAny || ulRights&ecRightsDeleteAny || ulRights&ecRightsCreateSubfolder || 
					(bOwner == true  && (ulRights&ecRightsEditOwned || ulRights&ecRightsDeleteOwned)) )
				{
					lpPropVal->Value.ul = MAPI_MODIFY;
				}
			}
			break;
        case PR_ROW_TYPE:
			lpPropVal->ulPropTag = ulPropTag;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			lpPropVal->Value.ul = TBL_LEAF_ROW;
            break;

        case PR_DEPTH: // always 0 unless using CONVENIENT_DEPTH (hierarchy table only)
			lpPropVal->ulPropTag = ulPropTag;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			lpPropVal->Value.ul = 0;
            break;

		case PR_MAPPING_SIGNATURE:
			lpPropVal->ulPropTag = ulPropTag;

			lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
			lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(GUID));

			lpPropVal->__union = SOAP_UNION_propValData_bin;
			lpPropVal->Value.bin->__size = sizeof(GUID);

			er = lpSession->GetServerGUID((GUID*)lpPropVal->Value.bin->__ptr);
			if(er != erSuccess){
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}
			break;

		default:
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
	}

exit:
	if(sPropTagArray.__ptr)
		delete[] sPropTagArray.__ptr;

	if(soap == NULL && sPropValArray.__ptr) // soap != NULL gsoap will cleanup the memory
		delete[] sPropValArray.__ptr;

	return er;
}

ECRESULT ECGenProps::GetStoreName(struct soap *soap, ECSession* lpSession, unsigned int ulStoreId, unsigned int ulObjId, char** lppStoreName)
{
	ECRESULT			er = erSuccess;
	unsigned int		ulUserId = 0;
	unsigned int	    ulCompanyId = 0;
	struct propValArray sPropValArray = {0, 0};
	struct propTagArray sPropTagArray = {0, 0};

	string				strFormat;
	char*				lpStoreName = NULL;

	er = lpSession->GetSecurity()->GetStoreOwner(ulStoreId, &ulUserId);
	if (er != erSuccess)
		goto exit;

	// get the companyid to which the logged in user belongs to.
	er = lpSession->GetSecurity()->GetUserCompany(&ulCompanyId);
	if (er != erSuccess)
		goto exit;

	// When the userid belongs to a company or group everybody, the store is considered a public store.
	if(ulUserId == ZARAFA_UID_EVERYONE || ulUserId == ulCompanyId) {
	    strFormat = _("Public Folders");
	} else {
        sPropTagArray.__ptr = new unsigned int[3];
        sPropTagArray.__ptr[0] = PR_DISPLAY_NAME;
        sPropTagArray.__ptr[1] = PR_ACCOUNT;
        sPropTagArray.__ptr[2] = PR_COMPANY_NAME;
        sPropTagArray.__size = 3;

        er = lpSession->GetUserManagement()->GetProps(soap, ulUserId, &sPropTagArray, &sPropValArray);
        if (er != erSuccess || !sPropValArray.__ptr) {
            er = ZARAFA_E_NOT_FOUND;
            goto exit;
        }

        strFormat = string(lpSession->GetSessionManager()->GetConfig()->GetSetting("storename_format"));

        for (int i = 0; i < sPropValArray.__size; i++) {
            string sub;
            size_t pos = 0;

            switch (sPropValArray.__ptr[i].ulPropTag) {
            case PR_DISPLAY_NAME:
                sub = "%f";
                break;
            case PR_ACCOUNT:
                sub = "%u";
                break;
            case PR_COMPANY_NAME:
                sub = "%c";
                break;
            default:
                break;
            }

            if (sub.empty())
                continue;

            while ((pos = strFormat.find(sub, pos)) != string::npos)
                strFormat.replace(pos, sub.size(), sPropValArray.__ptr[i].Value.lpszA);
        }

        strFormat = string(_("Inbox -")) + " " + strFormat;
    }
    
	lpStoreName = s_alloc<char>(soap, strFormat.size() + 1);
	strcpy(lpStoreName, strFormat.c_str());

	*lppStoreName = lpStoreName;

exit:
	if(sPropTagArray.__ptr)
		delete[] sPropTagArray.__ptr;

	return er;
}

ECRESULT ECGenProps::GetComputedDepth(struct soap *soap, ECSession* lpSession, unsigned int ulFolderId, unsigned int ulObjId, struct propVal *lpPropVal){
	ECRESULT		er = erSuccess;
	unsigned int ulObjType;

	lpPropVal->ulPropTag = PR_DEPTH;
	lpPropVal->__union = SOAP_UNION_propValData_ul;
	lpPropVal->Value.ul = 0;

	while(ulObjId != ulFolderId && lpPropVal->Value.ul < 50){
		er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, &ulObjId, NULL, NULL, &ulObjType);
		if(er != erSuccess) {
			// should never happen
			ASSERT(FALSE);
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		if(ulObjType != MAPI_FOLDER){
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		lpPropVal->Value.ul++;
	}

exit:
	return er;
}
