/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef ECICS_H
#define ECICS_H

#include "ECSession.h"

// This class is used to pass SOURCEKEYs internally between parts of the server backend. You can use it as a char* to get the data, use size() to get the size,
// and have various ways of creating new SOURCEKEYs, including using a GUID and an ID, which is used for zarafa-generated source keys.

class SOURCEKEY {
public:
    SOURCEKEY() { lpData = NULL; ulSize = 0; };
    SOURCEKEY(const SOURCEKEY &s) { 
        if(&s == this) return; 
        if(s.ulSize == 0) { 
            ulSize = 0; 
            lpData = NULL;
        } else { 
            lpData = new char[s.ulSize]; 
            memcpy(lpData, s.lpData, s.ulSize); 
            ulSize = s.ulSize;
        } 
    }
    SOURCEKEY(unsigned int ulSize, char *lpData) { 
		if (lpData) {
			this->lpData = new char[ulSize];
			memcpy(this->lpData, lpData, ulSize);
		} else {
			this->lpData = NULL;
		}
		this->ulSize = ulSize;
    }
    SOURCEKEY(GUID guid, unsigned int ulId) { 
        ulSize = sizeof(GUID) + sizeof(unsigned int); 
        lpData = new char [ulSize]; 
        memcpy(lpData, (char *)&guid, sizeof(guid)); 
        memcpy(lpData+sizeof(GUID), &ulId, sizeof(unsigned int)); 
    }
    SOURCEKEY(GUID guid, unsigned long long ullId) { 
        ulSize = sizeof(GUID) + sizeof(unsigned long long); 
        lpData = new char [ulSize]; 
        memcpy(lpData, (char *)&guid, sizeof(guid)); 
        memcpy(lpData+sizeof(GUID), &ullId, sizeof(unsigned long long)); 
    }
    ~SOURCEKEY() { 
        if(lpData) delete [] lpData; 
    }
    
    SOURCEKEY&  operator= (const SOURCEKEY &s) { 
        if(&s == this) return *this; 
        if(lpData) delete [] lpData; 
        lpData = new char[s.ulSize]; 
        memcpy(lpData, s.lpData, s.ulSize); 
        ulSize = s.ulSize; 
        return *this; 
    }
    
    bool operator == (const SOURCEKEY &s) {
        if(this == &s)
            return true;
        if(ulSize != s.ulSize)
            return false;
        return memcmp(lpData, s.lpData, s.ulSize) == 0;
    }
    
    operator unsigned char *() { return (unsigned char *)lpData; }
    
    unsigned int 	size() { return ulSize; }
	bool			empty() { return ulSize == 0; } 
private:
    char *lpData;
    unsigned int ulSize;
    
};

ECRESULT AddChange(BTSession *lpecSession, unsigned int ulSyncId, SOURCEKEY sSourceKey, SOURCEKEY sParentSourceKey, unsigned int ulChange, SOURCEKEY *lpMovedFromSourceKey = NULL, unsigned int ulFlags = 0, bool fForceNewChangeKey = false, std::string * lpstrChangeKey = NULL, std::string * lpstrChangeList = NULL);
ECRESULT AddABChange(BTSession *lpecSession, unsigned int ulChange, SOURCEKEY sSourceKey, SOURCEKEY sParentSourceKey);
ECRESULT GetChanges(struct soap *soap, ECSession *lpSession, SOURCEKEY sSourceKeyFolder, unsigned int ulSyncId, unsigned int ulChangeId, unsigned int ulChangeType, unsigned int ulFlags, struct restrictTable *lpsRestrict, unsigned int *lpulMaxChangeId, icsChangesArray **lppChanges);
void* CleanupSyncsTable(void* lpTmpMain);
void* CleanupChangesTable(void* lpTmpMain);

#endif
