/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "ECStatsCollector.h"
#include "stringutil.h"

using namespace std;

ECStatsCollector::ECStatsCollector() {
	pthread_mutex_init(&m_StringsLock, NULL);

	AddStat(SCN_SERVER_STARTTIME, SCDT_TIMESTAMP, "Started at", "Time when the server was started");
	AddStat(SCN_SERVER_LAST_CACHECLEARED, SCDT_TIMESTAMP, "Last cache purge", "Time when the cache was purged (SIGUSR2)");
	AddStat(SCN_SERVER_LAST_CONFIGRELOAD, SCDT_TIMESTAMP, "Last configfile reload", "Time when the configuration file was reloaded / logrotation (SIGHUP)");
	AddStat(SCN_SERVER_CONNECTIONS, SCDT_LONGLONG, "Connections", "Number of handled incoming connections");
	AddStat(SCN_THREAD_COUNTER, SCDT_LONGLONG, "Threads", "Current number of present threads");
	AddStat(SCN_MAX_SOCKET_NUMBER, SCDT_LONGLONG, "Highest socket", "Highest socket number used");
	AddStat(SCN_REDIRECT_COUNT, SCDT_LONGLONG, "Redirections", "Number of redirected requests");
	AddStat(SCN_SEARCHFOLDER_COUNT, SCDT_LONGLONG, "Searchfolders", "Total number of searchfolders");
	AddStat(SCN_SEARCHFOLDER_THREADS, SCDT_LONGLONG, "Searchfolder threads", "Current number of running searchfolder threads");

	AddStat(SCN_DATABASE_CONNECTS, SCDT_LONGLONG, "SQL Connect", "Number of connections made to SQL server");
	AddStat(SCN_DATABASE_SELECTS, SCDT_LONGLONG, "SQL Select", "Number of SQL Select commands executed");
	AddStat(SCN_DATABASE_INSERTS, SCDT_LONGLONG, "SQL Insert", "Number of SQL Insert commands executed");
	AddStat(SCN_DATABASE_UPDATES, SCDT_LONGLONG, "SQL Update", "Number of SQL Update commands executed");
	AddStat(SCN_DATABASE_DELETES, SCDT_LONGLONG, "SQL Delete", "Number of SQL Delete commands executed");
	AddStat(SCN_DATABASE_FAILED_CONNECTS, SCDT_LONGLONG, "Failed SQL Connect", "Number of failed connections made to SQL server");
	AddStat(SCN_DATABASE_FAILED_SELECTS, SCDT_LONGLONG, "Failed SQL Select", "Number of failed SQL Select commands");
	AddStat(SCN_DATABASE_FAILED_INSERTS, SCDT_LONGLONG, "Failed SQL Insert", "Number of failed SQL Insert commands");
	AddStat(SCN_DATABASE_FAILED_UPDATES, SCDT_LONGLONG, "Failed SQL Update", "Number of failed SQL Update commands");
	AddStat(SCN_DATABASE_FAILED_DELETES, SCDT_LONGLONG, "Failed SQL Delete", "Number of failed SQL Delete commands");
	AddStat(SCN_DATABASE_LAST_FAILED, SCDT_TIMESTAMP, "Last Failed SQL Time", "Timestamp of last failed SQL command");

	AddStat(SCN_LOGIN_PASSWORD, SCDT_LONGLONG, "Password Logins", "Number of logins through password authentication");
	AddStat(SCN_LOGIN_SSL, SCDT_LONGLONG, "SSL Logins", "Number of logins through SSL certificate authentication");
	AddStat(SCN_LOGIN_SSO, SCDT_LONGLONG, "SSO Logins", "Number of logins through Single Sign-on");
	AddStat(SCN_LOGIN_SOCKET, SCDT_LONGLONG, "Unix socket Logins", "Number of logins through Unix socket");
	AddStat(SCN_LOGIN_DENIED, SCDT_LONGLONG, "Authentication Failed", "Number of failed logins");

	AddStat(SCN_SESSIONS_CREATED, SCDT_LONGLONG, "Sessions Created", "Number of created sessions");
	AddStat(SCN_SESSIONS_DELETED, SCDT_LONGLONG, "Sessions Deleted", "Number of deleted sessions");
	AddStat(SCN_SESSIONS_TIMEOUT, SCDT_LONGLONG, "Sessions Timed-out", "Number of timed-out sessions");

	AddStat(SCN_SESSIONGROUPS_CREATED, SCDT_LONGLONG, "SessionGroups Created", "Number of created sessiongroups");
	AddStat(SCN_SESSIONGROUPS_DELETED, SCDT_LONGLONG, "SessionGroups Deleted", "Number of deleted sessiongroups");

	AddStat(SCN_LDAP_CONNECTS, SCDT_LONGLONG, "LDAP Connect", "Number of connections made to LDAP server");
	AddStat(SCN_LDAP_RECONNECTS, SCDT_LONGLONG, "LDAP Re-Connect", "Number of re-connections made to LDAP server");
	AddStat(SCN_LDAP_CONNECT_FAILED, SCDT_LONGLONG, "Failed LDAP Connect", "Number of failed connections made to LDAP server");
	AddStat(SCN_LDAP_CONNECT_TIME, SCDT_LONGLONG, "Total LDAP Connect time", "Total duration of connections made to LDAP server");
	AddStat(SCN_LDAP_CONNECT_TIME_MAX, SCDT_LONGLONG, "Max LDAP Connect time", "Longest connection time made to LDAP server");
	
	/* maybe usesless because SCN_LOGIN_* */
	AddStat(SCN_LDAP_AUTH_LOGINS, SCDT_LONGLONG, "LDAP Authentications", "Number of LDAP authentications");
	AddStat(SCN_LDAP_AUTH_DENIED, SCDT_LONGLONG, "LDAP Authentication Failed", "Number of failed authentications");
	AddStat(SCN_LDAP_AUTH_TIME, SCDT_LONGLONG, "Total Authentication Time", "Total authentication time");
	AddStat(SCN_LDAP_AUTH_TIME_MAX, SCDT_LONGLONG, "Max LDAP Connect time", "Longest duration of authentication made to LDAP server");

	AddStat(SCN_LDAP_SEARCH, SCDT_LONGLONG, "LDAP Search", "Number of searches made to LDAP server");
	AddStat(SCN_LDAP_SEARCH_FAILED, SCDT_LONGLONG, "Failed LDAP Search", "Number of failed searches made to LDAP server");
	AddStat(SCN_LDAP_SEARCH_TIME, SCDT_LONGLONG, "Total LDAP Search Time", "Total duration of LDAP searches");
	AddStat(SCN_LDAP_SEARCH_TIME_MAX, SCDT_LONGLONG, "Max LDAP Search Time", "Longest duration of LDAP search");
	
}

ECStatsCollector::~ECStatsCollector() {
	SCMap::iterator iSD;

	for (iSD = m_StatData.begin(); iSD != m_StatData.end(); iSD++) {
		pthread_mutex_destroy(&iSD->second.lock);
	}
	pthread_mutex_destroy(&m_StringsLock);
}

void ECStatsCollector::AddStat(SCName index, SCType type, const char *name, const char *description) {
	ECStat newStat;

	newStat.data.ll = 0;		// reset largest data var in union
	newStat.type = type;
	pthread_mutex_init(&newStat.lock, NULL);
	newStat.name = name;
	newStat.description = description;

	m_StatData[index] = newStat;
}

void ECStatsCollector::Increment(SCName name, float inc) {
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_FLOAT);

	pthread_mutex_lock(&iSD->second.lock);
	iSD->second.data.f += inc;
	pthread_mutex_unlock(&iSD->second.lock);
}

void ECStatsCollector::Increment(SCName name, int inc) {
	Increment(name, (LONGLONG)inc);
}

void ECStatsCollector::Increment(SCName name, LONGLONG inc) {
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_LONGLONG);

	pthread_mutex_lock(&iSD->second.lock);
	iSD->second.data.ll += inc;
	pthread_mutex_unlock(&iSD->second.lock);
}

void ECStatsCollector::Set(SCName name, float set) {
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_FLOAT);

	pthread_mutex_lock(&iSD->second.lock);
	iSD->second.data.f = set;
	pthread_mutex_unlock(&iSD->second.lock);
}

void ECStatsCollector::Set(SCName name, LONGLONG set) {
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_LONGLONG);

	pthread_mutex_lock(&iSD->second.lock);
	iSD->second.data.ll = set;
	pthread_mutex_unlock(&iSD->second.lock);
}

void ECStatsCollector::Set(SCName name, time_t set) {
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_TIMESTAMP);

	pthread_mutex_lock(&iSD->second.lock);
	iSD->second.data.ts = set;
	pthread_mutex_unlock(&iSD->second.lock);
}

void ECStatsCollector::Min(SCName name, float min)
{
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_FLOAT);

	pthread_mutex_lock(&iSD->second.lock);
	if (iSD->second.data.f > min)
		iSD->second.data.f = min;
	pthread_mutex_unlock(&iSD->second.lock);	
}

void ECStatsCollector::Min(SCName name, LONGLONG min)
{
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_LONGLONG);

	pthread_mutex_lock(&iSD->second.lock);
	if (iSD->second.data.ll > min)
		iSD->second.data.ll = min;
	pthread_mutex_unlock(&iSD->second.lock);	
}

void ECStatsCollector::Min(SCName name, time_t min)
{
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_TIMESTAMP);

	pthread_mutex_lock(&iSD->second.lock);
	if (iSD->second.data.ts > min)
		iSD->second.data.ts = min;
	pthread_mutex_unlock(&iSD->second.lock);	
}


void ECStatsCollector::Max(SCName name, float max)
{
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_FLOAT);

	pthread_mutex_lock(&iSD->second.lock);
	if (iSD->second.data.f < max)
		iSD->second.data.f = max;
	pthread_mutex_unlock(&iSD->second.lock);	
}

void ECStatsCollector::Max(SCName name, LONGLONG max)
{
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_LONGLONG);

	pthread_mutex_lock(&iSD->second.lock);
	if (iSD->second.data.ll < max)
		iSD->second.data.ll = max;
	pthread_mutex_unlock(&iSD->second.lock);	
}

void ECStatsCollector::Max(SCName name, time_t max)
{
	map<SCName, ECStat>::iterator iSD = m_StatData.find(name);
	if (iSD == m_StatData.end())
		return;

	ASSERT(iSD->second.type == SCDT_TIMESTAMP);

	pthread_mutex_lock(&iSD->second.lock);
	if (iSD->second.data.ts < max)
		iSD->second.data.ts = max;
	pthread_mutex_unlock(&iSD->second.lock);	
}



void ECStatsCollector::Set(const std::string &name, const std::string &description, const std::string &value)
{
	ECStrings data;

	data.description = description;
	data.value = value;

	pthread_mutex_lock(&m_StringsLock);
	m_StatStrings[name] = data;
	pthread_mutex_unlock(&m_StringsLock);
}

void ECStatsCollector::Remove(const std::string &name)
{
	pthread_mutex_lock(&m_StringsLock);
	m_StatStrings.erase(name);
	pthread_mutex_unlock(&m_StringsLock);
}

std::string ECStatsCollector::GetValue(SCMap::iterator iSD)
{
	std::string rv;
	char timestamp[32];			// man page: at least 26
	int len;

	switch(iSD->second.type) {
	case SCDT_FLOAT:
		rv = stringify_float(iSD->second.data.f);
		break;
	case SCDT_LONGLONG:
		rv = stringify_int64(iSD->second.data.ll);
		break;
	case SCDT_TIMESTAMP:
		if (iSD->second.data.ts > 0) {
			ctime_r(&iSD->second.data.ts, timestamp);
			len = strlen(timestamp);
			timestamp[len-1] = '\0'; // chop off trailing \n
			rv = timestamp;
		}
		break;
	}

	return rv;
}

std::string ECStatsCollector::GetValue(SCName name) {
	std::string rv;
	SCMap::iterator iSD = m_StatData.find(name);

	if (iSD != m_StatData.end())
		rv = GetValue(iSD);

	return rv;
}

void ECStatsCollector::ForEachStat(void(callback)(const std::string &, const std::string &, const std::string &, void*), void *obj)
{
	SCMap::iterator iSD;

	for (iSD = m_StatData.begin(); iSD != m_StatData.end(); iSD++) {
		pthread_mutex_lock(&iSD->second.lock);
		callback(iSD->second.name, iSD->second.description, GetValue(iSD), obj);
		pthread_mutex_unlock(&iSD->second.lock);
	}
}

void ECStatsCollector::ForEachString(void(callback)(const std::string &, const std::string &, const std::string &, void*), void *obj)
{
	std::map<std::string, ECStrings>::iterator iSS;

	pthread_mutex_lock(&m_StringsLock);
	for (iSS = m_StatStrings.begin(); iSS != m_StatStrings.end(); iSS++) {
		callback(iSS->first, iSS->second.description, iSS->second.value, obj);
	}
	pthread_mutex_unlock(&m_StringsLock);
}

void ECStatsCollector::Reset() {
	SCMap::iterator iSD;

	for (iSD = m_StatData.begin(); iSD != m_StatData.end(); iSD++) {
		// reset largest var in union
		iSD->second.data.ll = 0;
	}
}

void ECStatsCollector::Reset(SCName name) {
	SCMap::iterator iSD = m_StatData.find(name);

	if (iSD != m_StatData.end()) {
		// reset largest var in union
		iSD->second.data.ll = 0;
	}
}
