/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef ECUSERMANAGEMENT_H
#define ECUSERMANAGEMENT_H

#include <list>
#include <map>

#include "ZarafaCode.h"
#include "ZarafaUser.h"
#include "ECConfig.h"
#include "ECSession.h"
#include "ECLogger.h"
#include "ECDefs.h"
#include "plugin.h"

class localobjectdetails_t : public objectdetails_t {
public:
    localobjectdetails_t() : objectdetails_t(), ulId(0) {};
	localobjectdetails_t(unsigned int id, userobject_type_t type) : objectdetails_t(type), ulId(id) {};
	localobjectdetails_t(unsigned int id, objectdetails_t &details) : objectdetails_t(details), ulId(id) {};

	bool operator== (const localobjectdetails_t &obj) { return ulId == obj.ulId; };
	bool operator< (const localobjectdetails_t &obj) { return ulId < obj.ulId; } ;

	unsigned int ulId;
};

// Use for ulFlags
#define USERMANAGEMENT_IDS_ONLY 1		// Return only local userID (the ulId field). 'details' is undefined in this case

// Use for ulLicenseStatus in CheckUserLicense()
#define USERMANAGEMENT_LIMIT_ACTIVE_USERS		0x1	/* Limit reached, but not yet exceeded */
#define USERMANAGEMENT_LIMIT_NONACTIVE_USERS	0x2 /* Limit reached, but not yet exceeded */
#define USERMANAGEMENT_EXCEED_ACTIVE_USERS		0x4 /* Limit exceeded */
#define USERMANAGEMENT_EXCEED_NONACTIVE_USERS	0x8 /* Limit exceeded */

#define USERMANAGEMENT_BLOCK_CREATE_ACTIVE_USER		( USERMANAGEMENT_LIMIT_ACTIVE_USERS | USERMANAGEMENT_EXCEED_ACTIVE_USERS )
#define USERMANAGEMENT_BLOCK_CREATE_NONACTIVE_USER	( USERMANAGEMENT_LIMIT_NONACTIVE_USERS | USERMANAGEMENT_EXCEED_NONACTIVE_USERS )
#define USERMANAGEMENT_USER_LICENSE_EXCEEDED		( USERMANAGEMENT_EXCEED_ACTIVE_USERS | USERMANAGEMENT_EXCEED_NONACTIVE_USERS )

class ECUserManagement {
public:
	ECUserManagement(BTSession *lpSession, ECPluginFactory *lpPluginFactory, ECConfig *lpConfig, ECLogger *lpLogger);
	virtual ~ECUserManagement();

	// Authenticate a user
	virtual ECRESULT	AuthUserAndSync(char *szUsername, char *szPassword, unsigned int *lpulUserId);

	virtual ECRESULT	GetItemObjectType(unsigned int ulObjectId, userobject_type_t *lpulType);
	virtual ECRESULT	GetItemCompany(unsigned int ulObjectId, userobject_type_t ulType, unsigned int *ulCompanyId);
	// Get data for an object, with on-the-fly delete of the specified object id
	virtual ECRESULT	GetObjectDetailsAndSync(unsigned int ulObjectId, objectdetails_t *lpDetails);
	// Get quota details for a user object
	virtual ECRESULT	GetQuotaDetailsAndSync(unsigned int ulObjectId, quotadetails_t *lpDetails, bool bGetUserDefault = false);
	// Set quota details for a user object
	virtual ECRESULT	SetQuotaDetailsAndSync(unsigned int ulObjectId, quotadetails_t sDetails);
	// Get objectlist for company, with on-the-fly delete/create of users and groups
	virtual ECRESULT	GetCompanyObjectListAndSync(userobject_type_t type, unsigned int ulCompanyId, std::list<localobjectdetails_t> **lppObjects, unsigned int ulFlags = 0);
	// Get objectlist for entire server, with on-the-fly delete/create of users and groups
	virtual ECRESULT	GetObjectListAndSync(userobject_type_t type, unsigned int ulCompanyId, std::list<localobjectdetails_t> **lppObjects, unsigned int ulFlags = 0);
	// Get merged object list for entire server, with on-the-fly delete/create of users and groups
	virtual ECRESULT	GetMergedObjectListAndSync(unsigned int ulListTypes, unsigned int ulCompanyId, std::list<localobjectdetails_t> **lppObjects, unsigned int ulFlags = 0);
	// Get subobjects in an object, with on-the-fly delete of the specified parent object
	virtual ECRESULT	GetSubObjectsOfObjectAndSync(userobject_relation_t relation, userobject_type_t childtype, unsigned int ulParentId, std::list<localobjectdetails_t> **lppObjects, unsigned int ulFlags = 0);
	// Get parent to which an object belongs, with on-the-fly delete of the specified child object id
	virtual ECRESULT	GetParentObjectsOfObjectAndSync(userobject_relation_t relation, userobject_type_t parenttype, unsigned int ulChildId, std::list<localobjectdetails_t> **lppGroups, unsigned int ulFlags = 0);

	// Set data for a single user, with on-the-fly delete of the specified user id
	virtual ECRESULT	SetObjectDetailsAndSync(unsigned int ulObjectId, objectdetails_t sDetails, std::list<std::string> *lpRemoveProps );

	// Add a member to a group, with on-the-fly delete of the specified group id
	virtual ECRESULT	AddSubObjectToObjectAndSync(userobject_relation_t relation, unsigned int ulParentId, unsigned int ulChildId);
	virtual ECRESULT	DeleteSubObjectFromObjectAndSync(userobject_relation_t relation, unsigned int ulParentId, unsigned int ulChildId);

	// Resolve a user name to a user id, with on-the-fly create of the specified user
	virtual ECRESULT	ResolveObjectAndSync(userobject_type_t type, char *szName, unsigned int *lpulObjectId);

	// Get a local object ID for a part of a name
	virtual ECRESULT	SearchObjectAndSync(userobject_type_t type, char *szSearchString, unsigned int ulFlags, unsigned int *lpulId);

	// Create an object
	virtual ECRESULT	CreateObjectAndSync(userobject_type_t type, objectdetails_t details, unsigned int *ulId);
	// Delete an object
	virtual ECRESULT	DeleteObjectAndSync(unsigned int ulObjectId);
	// Either modify or create an object with a specific object id and type (used for synchronize)
	virtual ECRESULT	CreateOrModifyObject(const string &strExternId, userobject_type_t type, objectdetails_t details, unsigned int ulPreferredId, std::list<std::string> *lpRemoveProps);
	virtual ECRESULT	ModifyExternId(unsigned int ulObjectId, const string &strExternId);
	// Get MAPI property data for a group or user/group/company id, with on-the-fly delete of the specified user/group/company
	virtual ECRESULT	GetProps(struct soap *soap, unsigned int ulObjectId, struct propTagArray *lpPropTagArray, struct propValArray *lpPropValArray);
	virtual ECRESULT	GetContainerProps(struct soap *soap, unsigned int ulObjectId, struct propTagArray *lpPropTagArray, struct propValArray *lpPropValArray);
	// Do the same for a whole set of items
	virtual ECRESULT	QueryContentsRowData(struct soap *soap, ECObjectTableList *lpRowList, struct propTagArray *lpPropTagArray, struct rowSet **lppRowSet);
	virtual ECRESULT	QueryHierarchyRowData(struct soap *soap, ECObjectTableList *lpRowList, struct propTagArray *lpPropTagArray, struct rowSet **lppRowSet);

	/* Convert a user loginname to username and companyname */
	virtual ECRESULT	ConvertLoginToUserAndCompany(objectdetails_t *lpDetails);
	virtual ECRESULT	ConvertUserAndCompanyToLogin(objectdetails_t *lpDetails);
	/* convert SendAs extern id's to local id's */
	virtual ECRESULT	ConvertExternIDsToLocalIDs(objectdetails_t *lpDetails);
	/* calls externid->localid and user/company->login conversions */
	virtual ECRESULT	UpdateUserDetails(objectdetails_t *lpDetails);

	virtual ECRESULT	GetUserCount(unsigned int *lpulUsers, unsigned int *lpulNonActives); // returns active users and non-active users (so you may get ulUsers=3, ulNonActives=5)
	virtual ECRESULT	GetPublicStoreDetails(objectdetails_t *lpDetails);

	virtual ECRESULT	GetServerDetails(const std::string &strServer, serverdetails_t *lpDetails);

	/* Check if the user license status */
	ECRESULT	CheckUserLicense(unsigned int *lpulLicenseStatus);

	// Returns true if ulId is a local ID (so either SYSTEM or EVERYONE)
	bool		IsLocalItem(unsigned int ulId);

	// Create a v1 based AB SourceKey
	ECRESULT GetABSourceKeyV1(unsigned int ulUserId, SOURCEKEY *lpsSourceKey);

private:
	// Get local details
	ECRESULT	GetLocalObjectDetails(unsigned int ulId, objectdetails_t *lpDetails);

	// Get remote details
	ECRESULT	GetExternalObjectDetails(unsigned int ulId, objectdetails_t *lpDetails);

	// Get userinfo from cache
	ECRESULT	GetItemInfo(unsigned int ulId, userobject_type_t *lpulType, std::string *lpStr, std::string *lpSignature = NULL);
	ECRESULT	GetExternalId(unsigned int ulId, userobject_type_t *lpulType, std::string *lpExternId, std::string *lpSignature = NULL);
	ECRESULT	GetLocalId(const string &strExternId, userobject_type_t ulType, unsigned int *lpulId, std::string *lpSignature = NULL);
	ECRESULT	GetItemType(unsigned int ulId, userobject_type_t *lpulType);

	// Get userid from usertable or create a new user/group if it doesn't exist yet
	ECRESULT	GetLocalObjectIdOrCreate(userobject_type_t type, objectsignature_t signature, unsigned int *lpulId);

	// Get a list of local item ID's in the database plus any local items (SYSTEM, EVERYONE)
	ECRESULT	GetLocalItemIdList(unsigned int ulCompanyId, std::list<unsigned int> **lppUsers, userobject_type_t type);

	// Converts the data in user/group/company details fields into property value array for content tables and MAPI_MAILUSER and MAPI_DISTLIST objects
	ECRESULT	ConvertObjectDetailsToProps(struct soap *soap, unsigned int ulId, userobject_type_t type, objectdetails_t *lpObjectDetails, struct propTagArray *lpPropTags, struct propValArray *lpPropVals);
	// Converts the data in company/addresslist details fields into property value array for hierarchy tables and MAPI_ABCONT objects
	ECRESULT	ConvertContainerObjectDetailsToProps(struct soap *soap, unsigned int ulId, userobject_type_t type, objectdetails_t *lpObjectDetails, struct propTagArray *lpPropTags, struct propValArray *lpPropVals);
	// Create GlobalAddressBook properties
	ECRESULT	ConvertABContainerToProps(struct soap *soap, unsigned int ulId, struct propTagArray *lpPropTagArray, struct propValArray *lpPropValArray);

	ECRESULT	MoveOrCreateLocalObject(userobject_type_t ulType, objectsignature_t signature, unsigned int *lpulObjectId);
	ECRESULT	CreateLocalObjectSimple(objectsignature_t signature, userobject_type_t type, unsigned int ulPreferredId);
	ECRESULT	CreateLocalObject(userobject_type_t type, objectsignature_t signature, unsigned int *lpulId);
	ECRESULT	MoveOrDeleteLocalObject(unsigned int ulObjectId, userobject_type_t ulType);
	ECRESULT	MoveLocalObject(unsigned int ulObjectId, userobject_type_t ulType, unsigned int ulCompanyId, const std::string &strNewUserName);
	ECRESULT	DeleteLocalObject(unsigned int ulObjectId, userobject_type_t ulType);

	ECRESULT	GetUserAndCompanyFromLoginName(const std::string &strLoginName, string *lpstrUserName, string *lpstrCompanyName);

	// Process the modification of a user-object
	ECRESULT	ProcessModification(unsigned int ulId, userobject_type_t ulType, const std::string &newsignature);

	ECRESULT	CreateABEntryID(struct soap *soap, unsigned int ulObjId, unsigned int ulType, struct propVal *lpPropVal);

	// Get the user count
	virtual ECRESULT	GetLocalUserCount(unsigned int *lpulUsers, unsigned int *lpulNonActives);
	virtual ECRESULT	GetExternUserCount(unsigned int *lpulUsers, unsigned int *lpulNonActives);

protected:
	ECPluginFactory 	*m_lpPluginFactory;
	BTSession			*m_lpSession;
	ECLogger			*m_lpLogger;
	ECConfig			*m_lpConfig;

};

#define ZARAFA_UID_EVERYONE 1
#define ZARAFA_UID_SYSTEM 2

#define ZARAFA_ACCOUNT_SYSTEM "SYSTEM"
#define ZARAFA_FULLNAME_SYSTEM "SYSTEM"
#define ZARAFA_ACCOUNT_EVERYONE "Everyone"
#define ZARAFA_FULLNAME_EVERYONE "Everyone"

#define ZARAFA_ADDRESS_BOOK			"Zarafa Address Book"
#define ZARAFA_GLOBAL_ADDRESS_BOOK	"Global Address Book"

#endif
