/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

// -*- Mode: c++ -*-
#ifndef __DBBASE_H
#define __DBBASE_H

#include "plugin.h"
#include <stdexcept>
#include <string>

#include "ECDatabase.h"
#include "ECDefs.h"

// Table names
#define DB_OBJECT_TABLE				"object"
#define DB_OBJECTPROPERTY_TABLE		"objectproperty"
#define DB_OBJECT_RELATION_TABLE	"objectrelation"

// Object properties
#define OP_MODTIME			"modtime"
#define OP_LOGINNAME		"loginname" 
#define OP_PASSWORD			"password"
#define OP_ISADMIN			"isadmin"
#define OP_ISNONACTIVE		"isnonactive"
#define OP_FULLNAME			"fullname"
#define OP_FIRSTNAME		"firstname"
#define OP_EMAILADDRESS		"emailaddress"
#define OP_HARDQUOTA		"hardquota"
#define OP_SOFTQUOTA		"softquota"
#define OP_WARNQUOTA		"warnquota"
#define OP_USEDEFAULTQUOTA	"usedefaultquota"
#define OP_UD_HARDQUOTA		"userhardquota"
#define OP_UD_SOFTQUOTA		"usersoftquota"
#define OP_UD_WARNQUOTA		"userwarnquota"
#define OP_UD_USEDEFAULTQUOTA	"userusedefaultquota"
#define OP_GROUPNAME		"groupname"
#define OP_COMPANYID		"companyid"
#define OP_COMPANYNAME		"companyname"
#define OP_COMPANYADMIN		"companyadmin"
// link to find unix user
#define OP_PAMUID			"pamuid"

class DB_RESULT_AUTOFREE {
public:
    DB_RESULT_AUTOFREE(ECDatabase *lpDatabase) {
        m_lpDatabase = lpDatabase;
        m_lpResult = NULL;
    };
    ~DB_RESULT_AUTOFREE() {
        if(m_lpDatabase && m_lpResult)
            m_lpDatabase->FreeResult(m_lpResult);
    };
    operator DB_RESULT () const {
        return m_lpResult;
    };
    DB_RESULT * operator & () {
        // Assume overwrite will happen soon
        if(m_lpDatabase && m_lpResult)
            m_lpDatabase->FreeResult(m_lpResult);
        m_lpResult = NULL;
        return &m_lpResult;
    };
private:
    DB_RESULT		m_lpResult;
    ECDatabase *	m_lpDatabase;
};

class DBPlugin : public UserPlugin {
public:
	DBPlugin(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed);
	virtual ~DBPlugin();

	virtual void InitPlugin() throw(std::exception);

// from UserPlugin
public:
	virtual list<objectsignature_t> getAllObjects(userobject_type_t type, const objectid_t &company) throw(std::exception);
	virtual auto_ptr<objectdetails_t> getObjectDetails(userobject_type_t type, const objectid_t &userid) throw(std::exception);
	virtual list<objectsignature_t> getSubObjectsForObject(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype) throw(std::exception);
	virtual list<objectsignature_t> getParentObjectsForObject(userobject_relation_t relation, userobject_type_t childtype, const objectid_t &childobject, userobject_type_t parenttype) throw(std::exception);

	virtual int changeObject(userobject_type_t type, const objectid_t &objectid, objectdetails_t details, std::list<std::string> *lpDeleteProps) throw(std::exception);
	virtual objectsignature_t createObject(userobject_type_t type, objectdetails_t details) throw(std::exception);
	virtual void deleteObject(userobject_type_t type, const objectid_t &id) throw(std::exception);

	virtual void addSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception);
	virtual void deleteSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception);

	// Quota
	virtual auto_ptr<quotadetails_t> getQuota(userobject_type_t type, const objectid_t &id, bool bGetUserDefault) throw(std::exception);
	virtual int setQuota(userobject_type_t type, const objectid_t &id, quotadetails_t quotadetails) throw(std::exception);

protected:
	virtual list<objectsignature_t> searchObjects(string match, userobject_type_t type, char *search_props[], char *return_prop, unsigned int ulFlags) throw(std::exception);

	virtual ECRESULT CreateMD5Hash(const std::string &strData, std::string* lpstrResult);

	// extension for details, objectid may only be USEROBJECT_TYPE_USER
	void addSendAsToDetails(const objectid_t &objectid, objectdetails_t *lpDetails);

	ECDatabase *m_lpDatabase;
};

#endif
