/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef DBUSERPLUGIN_H
#define DBUSERPLUGIN_H

#include <stdexcept>
#include <string>

#include "plugin.h"
#include "DBBase.h"

class DBUserPlugin : public DBPlugin
{

public:
	DBUserPlugin(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed);
	virtual ~DBUserPlugin();

	virtual void InitPlugin() throw(std::exception);

// from UserPlugin
public:

	// Reading information
	virtual objectsignature_t 			resolveName(userobject_type_t type, const string &name, const string &company) throw(std::exception);
	virtual objectsignature_t 			authenticateUser(const string &username, const string &companyname, const string &password) throw(std::exception);
	virtual auto_ptr<objectdetails_t>   getObjectDetails(userobject_type_t type, const objectid_t &objectid) throw (std::exception);
	virtual auto_ptr<map<objectid_t, objectdetails_t> > getObjectDetails(userobject_type_t type, list<objectid_t> &objectids) throw (std::exception);
	virtual list<objectsignature_t>		searchObject(userobject_type_t type, const string &match, unsigned int ulFlags) throw(std::exception);
	virtual int setQuota(userobject_type_t type, const objectid_t &id, quotadetails_t quotadetails) throw(std::exception);
	virtual auto_ptr<objectdetails_t>	getPublicStoreDetails() throw(std::exception);
	virtual auto_ptr<serverdetails_t>	getServerDetails(const string &server) throw(std::exception);
	virtual void						getUserCount(unsigned int *lpulUsers, unsigned int *lpulNonActives) throw (std::exception);

	virtual void addSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception);
};

extern "C" {
	extern UserPlugin* getUserPluginInstance(pthread_mutex_t*, ECConfig*, ECLogger*, IECStatsCollector *, bool, bool);
	extern void deleteUserPluginInstance(UserPlugin*);
}

#endif
