/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef LDAPCACHE_H
#define LDAPCACHE_H

#include <memory>
#include <list>
#include <map>
#include <string>

#include "ECDefs.h"

class LDAPUserPlugin;

struct cache_entry_t {
	std::string object;
	std::string name;

	cache_entry_t()
		{}

	cache_entry_t(std::string object, std::string name)
		: object(object), name(name) {}

	bool operator < (const cache_entry_t& b) const { return this->object < b.object; }
};

typedef std::list<std::string> dn_list_t;
typedef std::map<std::string, cache_entry_t> dn_cache_t;

class LDAPCache {
private:
	/* Protect mutex from being overriden */
	pthread_mutex_t m_hMutex;
	pthread_mutexattr_t m_hMutexAttrib;

	std::auto_ptr<dn_cache_t> m_lpCompanyCache;		/* USEROBJECT_TYPE_COMPANY */
	std::auto_ptr<dn_cache_t> m_lpGroupCache;		/* USEROBJECT_TYPE_GROUP */
	std::auto_ptr<dn_cache_t> m_lpNonactiveCache;	/* USEROBJECT_TYPE_NONACTIVE */
	std::auto_ptr<dn_cache_t> m_lpUserCache;		/* USEROBJECT_TYPE_USER */
	std::auto_ptr<dn_cache_t> m_lpAddressListCache; /* USEROBJECT_TYPE_ADDRESSLIST */

public:
	LDAPCache();
	~LDAPCache();

	bool isObjectTypeCached(userobject_type_t ulType);
	void setObjectDNCache(userobject_type_t ulType, std::auto_ptr<dn_cache_t> lpCache);
	std::auto_ptr<dn_cache_t> getObjectDNCache(LDAPUserPlugin *lpPlugin, userobject_type_t ulType);

	/* Helper functions for cached items */
	static cache_entry_t getParentForDN(const std::auto_ptr<dn_cache_t> &lpCache, const std::string &dn);
	static std::auto_ptr<dn_list_t> getChildrenForDN(const std::auto_ptr<dn_cache_t> &lpCache, const std::string &dn);

	static std::auto_ptr<dn_list_t> GetChildrenForObject(const std::auto_ptr<dn_cache_t> &lpCache, const std::string &name);
	static std::string getDNForObject(const std::auto_ptr<dn_cache_t> &lpCache, const std::string &externid);
	static std::string getDNForObjectName(const std::auto_ptr<dn_cache_t> &lpCache, const std::string &name);
	static bool isDNInList(const std::auto_ptr<dn_list_t> &lpList, const std::string &dn);
};

#endif /* LDAPCACHE_H */
