/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <cctype>
#include <algorithm>

#include <mapi.h>
#include <mapix.h>
#include <mapicode.h>
#include <mapidefs.h>
#include <mapiutil.h>
#include <inetmapi.h>
#include <mapiext.h>

#include <CommonUtil.h>
#include <ECTags.h>
#include "ECChannel.h"
#include "LMTP.h"

using namespace std;


LMTP::LMTP(ECChannel *lpChan, char *szServerPath, ECLogger *lpLog, ECConfig *lpConf) {
    m_lpChannel = lpChan;
    m_lpLogger = lpLog;
    m_lpConfig = lpConf;
    m_strPath = szServerPath;
}

LMTP::~LMTP() {
}

HRESULT LMTP::HrSplitInput(const string &strInput, vector<string> &lstWords) {

	HRESULT hr = hrSuccess;
	string::size_type beginPos = 0;
	string::size_type findPos;
	string strTemp;

	// don't use strTemp.begin() as destination (or should have resized strTemp to match strInput size)
	std::transform(strInput.begin(), strInput.end(), back_insert_iterator<string>(strTemp), ::toupper);

	if ((findPos = strTemp.find("MAIL FROM:", beginPos)) != string::npos) {
	   	strTemp = "MAIL FROM:";
		lstWords.push_back(strInput.substr(beginPos, findPos+strTemp.length()));
		beginPos = findPos + strTemp.length();
	}
	else if ((findPos = strTemp.find("RCPT TO:", beginPos)) != string::npos) {
	   	strTemp = "RCPT TO:";
		lstWords.push_back(strInput.substr(beginPos, findPos + strTemp.length()));
		beginPos = findPos + strTemp.length();
	}

	while ( (findPos = strInput.find(" ", beginPos)) != string::npos) {
		
		if(strInput.substr(beginPos, findPos - beginPos).empty()==false)
			lstWords.push_back(strInput.substr(beginPos, findPos - beginPos));
		
		beginPos = findPos + 1;
	}

	if (beginPos < strInput.size()) {
		lstWords.push_back(strInput.substr(beginPos));
    }
	
	return hr;
}

HRESULT LMTP::HrGetCommand(string strCommand, LMTP_Command &eCommand) {
	
	HRESULT hr = hrSuccess;
	
	std::transform(strCommand.begin(), strCommand.end(), strCommand.begin(), ::toupper);
	
	if (strCommand.compare("LHLO") == 0) {
		eCommand = LMTP_Command_LHLO;
	} else if (strCommand.compare("MAIL FROM:") == 0  || strCommand.compare("MAIL FROM: ") == 0) {
		eCommand = LMTP_Command_MAIL_FROM;
	} else if (strCommand.compare("RCPT TO:") == 0 || strCommand.compare("RCPT TO: ") == 0) {
		eCommand = LMTP_Command_RCPT_TO;
	} else if (strCommand.compare("DATA") == 0) {
		eCommand = LMTP_Command_DATA;
	}  else if (strCommand.compare("RSET") == 0) {
		eCommand = LMTP_Command_RSET;
	} else if (strCommand.compare("QUIT") == 0) {
		eCommand = LMTP_Command_QUIT;
	} else {
		hr = MAPI_E_CALL_FAILED;
	}
	
	return hr;
}


HRESULT LMTP::HrResponse(string strResponse) {
	
	return m_lpChannel->HrWriteLine(strResponse);
}

HRESULT LMTP::HrCommandLHLO(string &strIP) {

	return hrSuccess;
}

HRESULT LMTP::HrCommandMAILFROM(string *strFrom) {	
	
	HRESULT hr = hrSuccess;
	std::string strAddress;
	
	// strFrom is only checked for syntax
	hr = HrParseAddress(*strFrom, &strAddress);
	
	return hr;

}

HRESULT LMTP::HrCommandRCPTTO(string *strTo, string *strUnresolved) {
	HRESULT hr = hrSuccess;

	hr = HrParseAddress(*strTo, strUnresolved);
	
	if(hr == hrSuccess) {
	    m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Parsed '%s' to recipient address '%s'", strTo->c_str(), strUnresolved->c_str());
    } else {
	    m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Invalid recipient address '%s'", strTo->c_str());
    }
	
	return hr;
}



HRESULT LMTP::HrCommandDATA(FILE *tmp) {
	HRESULT hr=hrSuccess;
	std::string inBuffer;
	int offset;

	HrResponse("354 2.1.5 Start mail input; end with <CRLF>.<CRLF>");

	// Now the mail body needs to be read line by line until <CRLF>.<CRLF> is encountered
	while (1) {
		hr = m_lpChannel->HrReadLine(&inBuffer);
		if(hr != hrSuccess)
		    break;
		
		if(inBuffer == ".")
		    break;

		offset = 0;
		if (inBuffer[0] == '.')
			offset = 1;			// "remove" escape point, since it wasn't the end of mail marker

		fwrite((char *)inBuffer.c_str() + offset, 1, inBuffer.size() - offset, tmp);

		// The data from HrReadLine does not contain the CRLF, so add that here
		fwrite("\r\n", 1, 2, tmp);
	}

	return hr;
}

HRESULT LMTP::HrCommandQUIT() {
	
	return hrSuccess;
}

HRESULT LMTP::HrParseAddress(std::string strInput, std::string *strAddress)
{
    size_t pos1;
    size_t pos2;
    
    pos1 = strInput.find('<');
    pos2 = strInput.find('>');
    
    if(pos1 == std::string::npos || pos2 == std::string::npos) {
        return MAPI_E_NOT_FOUND;
    }
    
    *strAddress = strInput.substr(pos1+1, pos2-pos1-1);
    
    return hrSuccess;
}
